#ifndef __CONTEXT_H__
#define __CONTEXT_H__
	
#include "xbefore.h"
#include <X11/Xlib.h>
#include "xafter.h"

#include "expr.h"
#include "linked.h"
#include "ident.h"

class Context;
class Context
{
	static LinkedList<void*> contexts;
	friend class Net;
	static void DisplayWillClose(Display* dpy);
	
	Context* super;

	struct GCList
	{
		Display* dpy;
		GC gc;
		Text colorname,fontname;
		XID font;
		XFontStruct* fontstr;

		GCList(Display* d,GC g): dpy(d),gc(g),font(0),fontstr(0) {}
		int const operator ==(const GCList& l) { return dpy==l.dpy; }
		int const operator !=(const GCList& l) { return dpy!=l.dpy; }
	};
	
	struct EnvList
	{
		Ident name;
		Expression* expr;

		EnvList(char* n,Expression* e): name(n),expr(e) {}
 	  	//AH Changed Start function added
 	  	EnvList(Ident& i,Expression* e): name(i.to_charPtr()),expr(e) {}
                //AH Changed End
		int const operator ==(const EnvList& l) { return name==l.name; }
		int const operator !=(const EnvList& l) { return name!=l.name; }
	};
	
	LinkedList<EnvList> envlist;
	OpaqueIterator<EnvList> enviter;
	LinkedList<GCList> gclist;
	OpaqueIterator<GCList> gciter;
	
	void FreeDisplay(Display* dpy);
	void ChangeGC(GCList* gcl,bool draw_xor);
	int AllocColor(Display* dpy,char* color);
 	//AH Changed Start function added
	int AllocColor(Display* dpy, Text& t) {return AllocColor(dpy, t.to_charPtr());}
 	//AH Changed End
	GCList* GetGCList(Display* dpy,bool draw_xor);

public:
	bool draw_xor;
	Text color,font;
	int relative,cpos,spline;
	bool dashed,filled,rotatable,scalable,condition,shadow;
	double thick;
	
	Context(Context* super = 0);
	~Context();
	
	void Evaluate(ExprData& ed);
	GC GetGC(Display* dpy,bool draw_xor);
	Value* GetVar(char* name);
	void SetZoom(Display* dpy,GC gc,double zoom);
	void TextExtents(Display* dpy,char* text,int& asc,int &desc,int& width);
	istream& Input(istream& is);
};

typedef Context* ContextP;

inline istream& operator >>(istream& is,Context& c) {	return c.Input(is); }

#endif
