#ifndef __MODULE_H__
#define __MODULE_H__

#include <iostream.h>

#include "linked.h"
#include "event.h"

class Net;
class Page;
class Editor;
class Object;

#define MODULE_NOTIFY(module_class) \
static Module* Instanciate(Net& n) { return new module_class(n); } \
static ModuleNotification notif(Instanciate);

class Module;
typedef Module* (*ModuleInstanciator)(Net&);
struct ModuleNotification {
	ModuleNotification(ModuleInstanciator mi);
};


class Module
{
	friend class Net;
	static Iterator<ModuleInstanciator> Modules();

protected:
	Net& net;
	char* name;

public:
	char* Name() { return name; }
	
	// gibt zurueck,ob das Modul fuer das Netz "net" zustaendig ist
	// bei Rueckgabe von "false" erfolgt ein sofortiges "delete" der Instanz
	  virtual bool Init() { return false; }
	
	// informiert ein Modul ueber ein gerade erzeugtes Pulldown-Menu 
	// Rueckgabewert: "true" = weitere Module nicht mehr informieren
	virtual bool CreatePulldown(Editor*, char*, Widget /*parent*/) 
	  { return false; };
	  
	// informiert ein Modul ueber einen gerade erzeugten Menu-Eintrag
	// Rueckgabewert: "true" = weitere Module nicht mehr informieren
	virtual bool CreateEntry(Editor*, char*, WidgetClass, Widget /*parent*/) 
	  { return false; };

	// Editor wurde ueber einer Page des Netzes geoeffnet. Rueckgabewert
	// sagt aus,ob autom. ein Menuepunkt zum Aufruf erzeugt werden soll.
	virtual bool Open(Editor&,Page&) { return false; }
	
	// Ein Editor wechselt die Page. Rueckgabewert = Menuepunkt ja oder nein?
	virtual bool Change(Editor&,Page& /*before*/,Page& /*now*/) { return false; }
	
	// ein Editor wird geschlossen.
	virtual void Close(Editor&,Page&) {}
	
	// Modul wurde durch autom. erzeugten Menuepunkt aufgerufen.
	// Rueckgabewert: 
	virtual void Call(Editor&,Page&) {}
	
	// Menupunkt wurde ausgewaehlt. before = true: Vor der Bearbeitung,sonst 
	// dannach. Rueckgabewert: "true" = Nicht weiterverarbeiten. 
	virtual bool Menu(Editor&,Page&,Ident&,bool /*before*/) { return false; }
	
	// Editorfenster soll neugezeichnet werden. before = true: Vor der Bearb.
	// sonst dannach. Rueckgabewert: "true" = Nicht weiterverarbeiten. 
	virtual bool Draw(Editor&,Page&,Display*,Drawable,bool /*before*/)
    { return false; }

  // Auf ein Objekt wurde Doppelgeklickt.
  // Rueckgabewert: "true" = Nicht weiterverarbeiten. 
  virtual bool DoubleClick(Editor&,Page&,Object*) { return false; }

	// Event wurde ausgeloest. before = true: Vor der Bearbeitung,sonst 
	// dannach. Rueckgabewert: "true" = Nicht weiterverarbeiten. 
	virtual bool Event(Editor&,Page&,EventInfo&,bool /*before*/) { return false;}

        // Wird von Net::RequestCB  aufgerufen, damit die Module
        // modifizierte Attribute auswerten koennen.
        // Wird auch von Net::AddCB aufgerufen, damit die Module
        // das modifizierte Netz auswerten koennen. 
        virtual void Refresh(Editor*,Page&) {}

  // Modul beansprucht das Netz fr sich
  virtual bool ModuleLock() { return false; }
  // Wenn true, werden menu-events zuerst an dieses Modul gemeldet
  virtual bool MenuLock() { return false; }
  
	Module(Net& n,char* nam): net(n),name(nam) {}
	virtual ~Module() {}
};

#endif
