#ifndef __NET_H__
#define __NET_H__

#include <iostream.h>

#include "linked.h"
#include "ident.h"
#include "position.h"
#include "text.h"
#include "dialog.h"
#include "filesel.h"
#include "struct.h"
#include "array.h"
#include "module.h"

#define NoFilename "Unnamed.net"

class Editor;
class NetClass;
class Object;

class Net;
class Net
{
	friend class Page;
	friend class TimeNETLink;
	friend class Tokengame;
	friend class Decomposition;
	friend class HCPN2EDSPN;
	friend class PathDecomp;
	friend class MSIPDecomp;

	static LinkedList<void*> nets;

	LinkedList<Module*> modules;
	void InitModules();
	void ExitModules();
        void RefreshModules();

	Array<Editor*,Page*> pages,undopages;
	Page* prime,*undo,*initial;
	Editor* curr_editor;
	
	NetClass* netcl;
	Struct attr;

	bool modified,locked;
	Text filename;

	void RequestCB(Dialog<Net,Object*>&,Object*,Ident&);
	Text input;
	Dialog<Net,Object*> request,message;

	void AskForDisplay(Widget shell);
	void AskForExpression(Widget shell);
	void Kill(Widget shell,bool ask = true);
	void Exit(Widget shell,bool ask = true);

	FileSelector<Net,bool> fsel;
	void SaveAsCB(bool,char* file,bool ok);
	void SaveAsAndQuitCB(bool, char* filename, bool ok);
	void LoadCB(bool,char* file,bool ok);
	void AddCB(bool,char* file,bool ok);

	int EditorCount() { return int(pages); } 

	void EditorOpened(Editor& ed);
	void EditorClosed(Editor& ed);

	void PageKilled(Page* p);

	friend class Editor;
        
	// R.D. Canged: char* Title(Editor& editor); geht unter Linux schief!
	Text Title(Editor& editor);
	void Menu(Editor&,Ident& name);
	void Draw(Editor& e,Display* dpy,Drawable d);
	void Event(Editor& e,EventInfo& ev);

public:
	void ObjectDialog(Widget parent,Object* ob,bool hidden, bool readonly);
	void NetDialog(Widget parent,bool hidden);

        void UndoCopy(); // ehemals private, wird von SPNLControl gebraucht
        void Undo(); // ehemals private, wird von SPNLControl gebraucht
  
	static Iterator<void*> Nets() { return Iterator<void*>(nets); }
	static Net* GetNet(Ident& name);

	char* FileName() { return filename.to_charPtr(); }
	void SetFileName(char* name) { filename = name; }
 	//AH Changed Start: function inserted
 	void SetFileName(Ident& i) { SetFileName(i.to_charPtr()); }
 	//AH Changed End
	bool Modified() { return modified; }
	bool Locked();
	Struct& Attribute() { return attr; }
	Value* Attr(char* name);
	Text FullName() { return Text(modified?"* ":"")+filename; }
	Page* Prime() { return prime; }
	NetClass* Class() { return netcl; }
	Iterator<Module*> Modules() { return Iterator<Module*>(modules); }
	
	Page* GetPage(Editor* ed);
	Editor* OpenEditor(Page* p = 0,char* dpy_str = 0, char* path = NULL);
 	//AH Changed Start: function inserted
 	Editor* OpenEditor(Page* p = 0,Text t = Text((const char*)0), char* path = NULL)\
 	  { return OpenEditor(p,t.to_charPtr(), path);}
 	//AH Changed End
	void ChangeEditor(Editor& ed,Page* p);
	void CloseEditor(Editor& ed);

	bool Lock(bool on);
	void Update(Page* p = 0);

	void UpdateButtonStates(Editor& editor,Page& page);

	Net(NetClass* nc,char* fname = 0);
	Net(Net&);
	Net* Copy();
	~Net();

	static Net* Load(char* fpath);
 	//AH Changed Start: function inserted
 	static Net* Load(Text t) {return Load(t.to_charPtr());}
 	//AH Changed End
	bool Save(Page* root = 0);

	istream& Input(istream&);
	ostream& Output(ostream&);
};

typedef Net* NetP;

inline istream& operator >>(istream& is,Net& n) {	return n.Input(is); }
inline ostream& operator <<(ostream& os,Net& n) {	return n.Output(os); }

#endif
