#ifndef __NETCL_H__
#define __NETCL_H__

#include <iostream.h>

#include "ident.h"
#include "text.h"
#include "linked.h"
#include "struct.h"
#include "net.h"
#include "page.h"
#include "wrapper.h"
#include "objbar.h"

class Editor;
class Attrib;
class ObjectClass;
class ConnectorClass;
class NodeClass;
class Node;
class Value;

class NetClass;

template<class T> class ClassGroup
{
	friend class NetClass;
	Ident name;
	LinkedList<T*> members;
	ClassGroup(char* n,T* oc = 0): name(n) { if (oc) Add(oc); }
 	//AH Changed Start constructor added
 	ClassGroup(Ident i,T* oc = 0): name(i.to_charPtr()) { if (oc) Add(oc); }
 	//AH Changed End
	Iterator<T*> Members() { return Iterator<T*>(members); }
	void Add(T* oc) { members.Append(oc); }
	bool IsIn(T* oc) { return members.IsIn(oc); }
 public:
	int operator ==(const ClassGroup<T>& g) { return name==g.name; }
	int operator !=(const ClassGroup<T>& g) { return name!=g.name; }
};

	
class NetClass
{
	static LinkedList<void*> netcls;
	
	LinkedList<Attrib*> types;
	LinkedList<ObjectClass*> objcls;
	LinkedList<NodeClass*> nodecls;
	LinkedList<ConnectorClass*> conncls;
	Text path;
	Ident name;
	Struct attr;
	bool subclass,checknames,builtin;

	LinkedList<ClassGroup<NodeClass> > nodegroups;
	LinkedList<ClassGroup<ConnectorClass> > conngroups;
	
	void AddNodeClass(NodeClass* oc);
	void AddConnectorClass(ConnectorClass* oc);
	
	friend class Editor;
	void EnterEditor(ObjectBar<NodeClass>& nodebar,ObjectBar<ConnectorClass>& connbar);

	friend class ConnectorClass;
	LinkedList<NodeClass*> NodeClassesOfName(Ident& id);

	friend class Page;
	Page cutbuffer;
	
	static NetClass* Load(char* path,char* name);
	
	friend int main(int argc,char* argv[]);
	static int LoadExternals(char* path);

public:
	NetClass(char* input = 0,char* path = 0);
 	//AH Changed Start constructor added
 	NetClass(char* input = 0, Text t = Text((const char*)0) );
 	//AH Changed End
	~NetClass();

	static Iterator<void*> NetClasses() { return Iterator<void*>(netcls); }
	static NetClass* GetNetClass(char* id,char* path = 0);
 	//AH Changed Start function added
 	static NetClass* GetNetClass(Ident& i,char* path = 0) { return GetNetClass(i.to_charPtr(), path );}
 	//AH Changed End

	Ident Name() { return name; }
	//AH Changed const_cast<char*> inserted
	char* Label() { return const_cast<char*>(name.Label()); }
 	//AH Changed Start char* Path() { return path; }
	char* Path() { return path.to_charPtr(); }
 	//AH Changed End
	Struct& Attributes() { return attr; }
	bool IsSubclass() { return subclass; }
	bool IsBuiltin() { return builtin; }
	bool CheckNames() { return checknames; }
	
	Iterator<Attrib*> Types() { return Iterator<Attrib*>(types); }
	Iterator<ObjectClass*> ObjectClasses()
		{ return Iterator<ObjectClass*>(objcls); }
	Iterator<NodeClass*> NodeClasses() { return Iterator<NodeClass*>(nodecls); }
	Iterator<ConnectorClass*> ConnectorClasses()
		{ return Iterator<ConnectorClass*>(conncls); }

	Value* GetType(const Ident id);
	NodeClass* GetNodeClass(const Ident id);
	ConnectorClass* GetConnectorClass(const Ident id);
	ObjectClass* GetObjectClass(const Ident id);
	
	ConnectorClass *FittingConnectorClass(Node* in,Node* out);
	
	istream& Input(istream&);
//ostream& Output(istream&);
};

typedef NetClass *NetClassP;

inline istream& operator >>(istream& is,NetClass& n) {	return n.Input(is); }
// inline ostream& operator >>(ostream& is,NetClass& n) {	return n.Output(is); }

#endif
