#ifndef __NODE_H__
#define __NODE_H__

#include "object.h"
#include "connect.h"
#include "net.h"
#include "nodecl.h"
class translator;                                 

class Node: public Object
{
  friend class Decomposition; //to modify subpage
	NodeClass* nodecl;
	LinkedList<Position*> dummy;
	LinkedList<Node*> aliases;
	LinkedList<Connector*> outputs;
	LinkedList<Connector*> inputs;
	Position pos;
	double angle;
	Node* real;
	Page* subpage;
	
	void MoveConnections(Position rel);
	void RearrangeConnections();
	void SetReal(Node* r);

	friend class Connector;
	void AddInput(Connector* c);
	void RemoveInput(Connector* c);
	void AddOutput(Connector* c);
	void RemoveOutput(Connector* c);
	
	void AddAlias(Node* n);
	void RemoveAlias(Node* n);
	Node* CreateAlias(Page* parent);
	bool IsConnected(Node* n);

	friend class Editor;
	void EnterEditor(Editor* ed);
	void LeaveEditor();

	friend class Page;
	friend class MSIPDecomp; //for UpdateHierarchy
	friend class PathDecomp;//and UpdateVirtualConnections
	void UpdateHierarchy();
	void UpdateVirtualConnections();
	void Renumber();
	double Nearest(Position& ps,Position& nearest);
	Position RearrangePoint(Position& ps);

  friend class Translator;
  friend class TemplateIncluder;
  friend class SymbolTranslator;
  
	Node(Node& n,Page* parent = 0,bool portcopy = false);
	Node* Copy(Page* parent);

public:
	Page* SubPage() { return subpage; }
	Node* Real() { return real; }
	bool Virtual() { return real!=0; }

	bool IsNode() { return true; }
	NodeClass* Class() { return nodecl; }
	Position Mid() { return pos; }
	double Angle() { return angle; }
	
	Node(NodeClass* const, Page* parent, char* name = 0, 
	  Position p = Position(0,0), double angle = 0);
	
	~Node();
	void RemoveNodeAndAliases();//AH New: remove a node and all aliases of it

	Node* GetAlias(Page* parent);

	int NumInputs() { return inputs.Length(); }
	int NumOutputs() { return outputs.Length(); }
	Iterator<Connector*> Inputs() { return Iterator<Connector*>(inputs); }
	Iterator<Connector*> Outputs() { return Iterator<Connector*>(outputs); }
	
	int NumAliases() { return aliases.Length(); }
	Iterator<Node*> Aliases() { return Iterator<Node*>(aliases); }
	
	istream& Input(istream&);
	ostream& Output(ostream&);
	
	bool Meltable(Node* n);
	void Melt(Node* n);

	void SetPos(const Position p) { pos = p; RearrangeConnections(); }
	void SetScale(const Position s) { scale = s; RearrangeConnections(); }
	void SetAngle(const double x) { angle = x; RearrangeConnections(); }
	void SetClass(NodeClass* cl);
  void Move(const Position rel); 

	ShapeData GetShapeData(double zoom, bool draft);
	ExprData GetExprData(bool mono, bool draft);
};

typedef Node *NodeP;

#endif
