#ifndef __OBJECT_H__
#define __OBJECT_H__

#include <iostream.h>

#include "linked.h"
#include "attrib.h"
#include "objcl.h"
#include "net.h"
#include "event.h"
#include "shape.h"
#include "expr.h"
#include "metrics.h"

class Editor;
class Page;
class Struct;
class Node;

class Object;
class Object
{
	static unsigned global_id;

protected:
	bool selected;
	Page* page;
	ObjectClass* objcl;
	Struct* attr;
	Position scale;
	Position* curr;
	LinkedList<Position*> relatives;
	unsigned id;
	Ident name;

	static unsigned NewID() { return global_id++; }

	istream& Input(istream& is,Ident& id);
	
	friend class Page;
	virtual void Renumber() { id = NewID(); }
	void AutoName();

	virtual bool Event(Display* dpy,EventInfo& ev);
	void Draw(Display* dpy,Drawable d,double zoom,bool draft);
		
public:
	virtual void Move(const Position rel) = 0;// ehemals private, wird von SPNLControl gebraucht
	void SetName(char* n) { name = n; }
	Ident& Name() { return name; }
	Text FullName();
	unsigned ID() { return id; }
	Struct* Attribute() { return attr; }
	Value* Attr(char* name);
	virtual Page* SubPage() { return 0; }
	virtual bool Virtual() = 0;
	virtual bool IsNode() = 0;
	ObjectClass* Class() { return objcl; }
	Page* GetPage() { return page; }
	Position GetRelative(int pos);
	
	Position Scale() { return scale; }
	virtual Position Mid() = 0;
	virtual double Angle() { return 0; }
	
	Object(Page* parent,ObjectClass* cl,char* name,Struct* st);
	Object(Object& o,Page* parent,Struct* attr);
	virtual ~Object();
	
	int IsUnder(Display* dpy,Position ps);
        Metrics GetMetrics(Display* dpy);

	virtual istream& Input(istream&) = 0;
	virtual ostream& Output(ostream&) = 0;
	
	void Select(bool dosel); /*toggle selection, redraw page*/
	void MarkSelected(bool dosel); /*toggle selection, dont redraw page*/
	bool Selected() { return selected; }
	
	virtual void SetScale(const Position s) { scale = s; }
	virtual void SetClass(ObjectClass*);
	void SetRelative(int pos, Position p);
	
	virtual ShapeData GetShapeData(double zoom, bool draft) = 0;
	virtual ExprData GetExprData(bool mono, bool draft) = 0;
};

typedef Object *ObjectP;

inline istream& operator >>(istream& is,Object& o) {	return o.Input(is); }
inline ostream& operator <<(ostream& os,Object& o) {	return o.Output(os); }

#endif

