/*********************************************************
posint - class for non-negative integers, scope [0,1,2..oo[
         is used by reachability.cc to represent markings

         val - integer value, 0 when dontcare == true
	 max - maximum allowed value for val
	 min - minimum allowed value for val
	 dontcare - true for undefined/don't care
	       is displayed as `*'

*********************************************************/
#ifndef __POSINT_H__
#define __POSINT_H__

#include <strstream.h>

#include "constants.h"
#include "value.h"
#include "repeater.h"
#include "float.h"
#include "wrapper.h"

class PosInt;
class PosInt: public Value, public Wrapper<PosInt,bool>
{
protected:
	int val,max,min;
	Widget form,text;
	Repeater up_r,dn_r;

	bool dontcare;
	
	void Edit(Widget,bool,XtPointer) {}
	void Arrow(Widget w,bool,XtPointer);
	
	void UpdateWidget();

public:
	PosInt(NetClass* nc,bool hidden = false);
	PosInt(int v = 0);
	PosInt(const PosInt& i);
	~PosInt() { 	UncatchDestroy(&form,&text); }
	
	Value* Copy() const { return new PosInt(*this); }
	Value* Cast() const { return new Float(val); }

	Widget CreateWidget(Widget parent);
	bool EvalWidget();
	
	istream& Parse(istream& is);
	istream& Input(istream& is);
	ostream& Output(ostream& os) const { return os<<val; }
	
 	int to_int() const { return val; }
 	bool to_bool() const { return bool(val); }
 	double to_double() const { return double(val); }
 	char* to_charPtr() const; 
	
	Value& operator =(const int i);
 	Value& operator =(const PosInt& i) { return *this=i.to_int(); }
 	Value& operator =(const Value& v) { return *this=v.to_int(); }
	
	bool operator ==(int v) { return val==v; }
	bool operator !=(int v) { return val!=v; }
	bool operator >(int v)  { return val>v; }
	bool operator <(int v)  { return val<v; }
	PosInt operator -()      { return PosInt(-val); }
	PosInt operator !() 		 { return PosInt(!val); }
	PosInt operator -(int v)  { return PosInt(val-v); }
	PosInt operator +(int v)  { return PosInt(val+v); }
	PosInt operator *(int v)  { return PosInt(val*v); }
	PosInt operator /(int v)  { return PosInt(val/v); }
	
	PosInt& operator ++()  { ++val; return *this; }
};

typedef PosInt* PosIntP;

#endif
