#ifndef __POSITION_H__
#define __POSITION_H__

#include "xbefore.h"
#include <X11/Xlib.h>
#include "xafter.h"

#include <iostream.h>
#include "scope.h"

#include "constants.h"

const double undefined = 1e-301;

class Position;
class Position
{
  //return the number which is between the two others in size
	double _Limit(double x,double x1,double x2) const
	{ return x2>x1?(x<x1?x1:x>x2?x2:x):(x<x2?x2:x>x1?x1:x); }

public:
	double x,y;
	Position(): x(undefined),y(undefined) {}
	Position(int a,int b): x(a),y(b) {}
	Position(double a,double b): x(a),y(b) {}
	Position(const XPoint* p): x(p->x),y(p->y) {}
	Position(const XPoint p): x(p.x),y(p.y) {}
	
	bool Defined() const { return x!=undefined || y!=undefined; }
	
	Position Abs() const { return Position(fabs(x),fabs(y)); }
	
	double Angle() const { return atan2(y,x); }
	Position Rotate(double angle) const 
		{	double c = cos(angle),s = sin(angle); return Position(x*c-y*s,x*s+y*c); }
	
	double Distance(const Position p) const 
		{ return sqrt((p.x-x)*(p.x-x)+(p.y-y)*(p.y-y));}
	double Length() const { return sqrt(x*x+y*y); }	
        double X() const { return x; }
        double Y() const { return y; }

	Position Round() const { return Position(int(x+.5),int(y+.5)); }
	Position Grid(double grid) const { return grid==0?*this:
		Position(grid*int(x/grid+.5),grid*int(y/grid+.5)); }
	
	Position XLimit(double x1,double x2) const { return Position(_Limit(x,x1,x2),y); }
	Position YLimit(double y1,double y2) const { return Position(x,_Limit(y,y1,y2)); }
	
	Position XYMin(const Position p) const { return Position(x<p.x?x:p.x,y<p.y?y:p.y); }
	
	operator XPoint() const { XPoint xp={short(x+0.5),short(y+0.5)}; return xp; }
 	//AH Changed Start: operators bool() char*() replaced
 	bool to_bool() const { return x!=0 || y!=0; }
 	char* to_charPtr() const { static char buf[256]; ostrstream os(buf,256); os<<'('<<x<<','<<y<<')'<<'\0'; return buf; }
 	//AH Changed End
	
	Position operator *(const double d) const { return Position(x*d,y*d); }
	Position operator *(const Position s) const { return Position(x*s.x,y*s.y); }
	Position operator /(const Position s) const { return Position(x/s.x,y/s.y); }
	Position operator /(const double d) const { return Position(x/d,y/d); }
	Position operator +(const Position s) const { return Position(x+s.x,y+s.y); }
	Position operator -(const Position s) const { return Position(x-s.x,y-s.y); }
	
	Position& operator +=(const Position s) { x+=s.x; y+=s.y; return *this; }
	Position& operator -=(const Position s) { x-=s.x; y-=s.y; return *this; }
	Position& operator *=(const Position s) { x*=s.x; y*=s.y; return *this; }
	Position& operator /=(const Position s) { x/=s.x; y/=s.y; return *this; }
	
	Position& operator +=(const double d) { x+=d; y+=d; return *this; }
	Position& operator -=(const double d) { x-=d; y-=d; return *this; }
	Position& operator *=(const double d) { x*=d; y*=d; return *this; }
	Position& operator /=(const double d) { x/=d; y/=d; return *this; }
	
	Position operator -() const { return Position(-x,-y); }
	
	int const operator ==(const Position p) const { return x==p.x && y==p.y; }
	int const operator !=(const Position p) const { return x!=p.x || y!=p.y; }
	
};

typedef Position* PositionP;

inline ostream& operator <<(ostream& os,const Position& p)
{
	return os<<'('<<p.x<<','<<p.y<<')';
}

inline istream& operator >>(istream& is,Position& p)
{
	Scope sp(is,2,2);
	while (is && sp.Get())
		switch (sp.field)
		{
			case 1: if (!(is>>p.x)) cerr<<"x-coord. expected"<<endl; break;
			case 2: if (!(is>>p.y)) cerr<<"y-coord. expected"<<endl;
		}
	return is;
}

#endif
