#ifndef __SHAPE_H__
#define __SHAPE_H__

#include "xbefore.h"
#include <X11/Xlib.h>
#include "xafter.h"
#include <iostream.h>

#include "position.h"
#include "expr.h"
#include "context.h"
#include "metrics.h"

class Struct;

struct ShapeData
{
	double zoom;
	Position pos,scale;
	double angle;
	Iterator<Position*> iter;
	Iterator<Position*> rels;
	bool node;
	
	ShapeData(double z,Position p,Position s,double a,
		Iterator<Position*>& pi,Iterator<Position*>& re,bool n):
		zoom(z),
		pos(p),
		scale(s),
		angle(a),
		iter(pi),
		rels(re),
		node(n) {}
};

#define SHAPE_NOTIFY(shape_name,shape_class) \
static Shape* Instanciate(istream& is, Context* super) { return new shape_class(is,super); } \
static ShapeNotification notif(shape_name,Instanciate);

class Shape;
typedef Shape* (*ShapeInstanciator)(istream& is, Context* super);
struct ShapeNotification {
	ShapeNotification(char* name,ShapeInstanciator si);
};

		
class Shape
{
 protected:
	Context context;
	
	virtual istream& Input(istream& is,int field) = 0;
	virtual void DoDraw(Display* dpy,Drawable d,GC gc,ShapeData& sd,ExprData& ed) = 0;

	Position Absolute(ShapeData& p,Position r);

protected:
	Shape(Context* super = 0): context(super) {}

public:
	static Shape *GetShape(istream&,Context* super = 0);

	virtual ~Shape() {}

	virtual char* Type() = 0;
	virtual double Distance(ShapeData&,Position& ps,Position& nearest) = 0;
	virtual int Inside(Display*,Position,ShapeData&,ExprData&) = 0;
        virtual Metrics GetMetrics(Display*,ShapeData&,ExprData&) { return Metrics(); }
	virtual Position PointToMid(ShapeData&,Position&);
	
	void Draw(Display* dpy,Drawable d,
		ShapeData&,ExprData&,bool draw_xor);

	Context& GetContext() { return context; }
	
	int RelativePos() { return context.relative+(context.relative>=0?1:0); }
	
	istream& Input(istream&);
//	ostream& Output(ostream&);
};

inline istream& operator >>(istream& is,Shape& v) {	return v.Input(is); }
//inline ostream& operator <<(ostream& os,Shape& v) {	return v.Output(os); }

#endif
