#ifndef __TEXT_H__
#define __TEXT_H__

#include <strstream.h>
extern "C" {
#include <ctype.h>
#include <stdlib.h>
}
#include "value.h"

class Text;
class Text: public Value
{
protected:
	Widget scroll,text;
	short rows;
	char* p;
	
	//AH CHANGED:  char* Sure(char* s) ...
	const char* Sure(char* s) const { return s?s:""; }

public:
	Text(NetClass* nc = 0,bool hidden = false);
	Text(const char* s);
	Text(const Text& t);
	~Text() { UncatchDestroy(&scroll,&text); if (p) delete[] p; }
	Value* Copy() const { return new Text(*this); }
	Value* Cast() const { return Copy(); }
	
	bool Composite() const { return true; }
	bool Resizable() const { return rows>1; }
	bool Defined() const { return p!=0; }
	
	int Length() const { return p?strlen(p):0; }
	Text Left(int i) const;
	
	Widget CreateWidget(Widget parent);
	bool EvalWidget();
	
	istream& Parse(istream& is);
	istream& Input(istream& is);
	ostream& Output(ostream& os) const;
	
 	char* to_charPtr() const { return p; }
 	bool to_bool() const { return p!=0; }
 	double to_double() const { return p && isdigit(*p)?atof(p):0.0; }
	
	Value& operator =(const char* s);
	Value& operator =(const double d);
	Value& operator =(const int i);
	Value& operator =(const Text& s) { return *this=s.p; }
 	Value& operator =(const Value& v) { return *this=v.to_charPtr(); }
	//AH New:
	Text StripPath(char slash) const;  /*strips path fom text: 
				       path/path/path/name.xx.StripPath('/') => "name.xx"*/
	Text FirstPathPart(char slash) const; /*get first part of a path: 
					  "apath/bpath/cpath/name.xx".FirstPathPart('/') => "apath"*/

	Text StripLastPathPart(char slash) const; /* strip last part of a part
					       "path/path/path/name.xx".StripPath('/') => "path/path/path"*/
	Text StripFirstPathPart(char slash) const; /*strip first part of a path: 
					       "apath/bpath/cpath/name.xx".FirstPathPart('/') => "bpath/cpath/name.xx*/
	Text BetweenC(char c1, char c2) const; /*returns text between first occurence of c1
					   and last occurence of c2:
					   "(local cond) (global cond)".BetweenC('(',')')
					   => "local cond) (global cond"*/
	Text BetweenS(char* s1, char* s2) const;/*returns text between first occurence of string s1
					   and first occurence of string s2:
					   "#(local cond)# #(global cond)#".BetweenS("#(",")#")
					   => "local cond"*/
	Text ReplaceChar(char c1, char c2) const; /*replace all occurences of c1 with c2 */
	bool IsIn(const Text& t) const;   /*  is contained in t */
	bool IsIn(const char* c) const;   /*  is contained in c */
	bool Contains(const Text& t) const; /*contains t*/
	bool Contains(const char* c) const; /*contains c*/
	bool Contains(const char c) const; /*contains c*/
	bool BeginsWith(const Text& t) const ; /*begins with t*/
	bool BeginsWith(const char* c) const; 

	bool operator ==(const char* v) const;
 	bool operator ==(const Value& v) const { return *this==v.to_charPtr(); }
	bool operator !=(const char* v) const;
 	bool operator !=(const Text& t ) const {return *this!=t.to_charPtr(); }
	bool operator >(const char* v) const;
	bool operator <(const char* v) const; // alphabetischer String-Vergleich
	Text operator -() const; // String invertieren
	Text operator -(const char* v) const; // String-Subtraktion
	Text operator +(const char* v) const; // String-Addition
 	Text operator +(const Text& t) const { return *this+t.to_charPtr(); }
 	Text operator +(const Value& v) const { return *this+v.to_charPtr(); }
	Text operator *(const char* v) const;
	Text operator /(const char* v) const { return *this-v; } // Abbildung auf -

	Text& operator +=(const char* s);
	Text& operator +=(const char c);
 	Text& operator +=(Text t) { return *this+=t.to_charPtr(); }
 	Text& operator +=(const double d);

private:
	friend class Operation;
	Value* Eq(Value& v);
	Value* Gt(Value& v);
	Value* Lt(Value& v);
	Value* Neg() { return new Text(-*this); }
	Value* Not(); // String-Laenge
 	Value* Sub(Value& v) { return new Text(*this-v.to_charPtr()); }
 	Value* Add(Value& v) { return new Text(*this+v.to_charPtr()); }
 	Value* Mul(Value& v) { return new Text(*this*v.to_charPtr()); }
 	Value* Div(Value& v) { return new Text(*this/v.to_charPtr()); }
	
 	Value* And(Value& v) { return new Text(*this+v.to_charPtr()); }
 	Value* Or(Value& v) { return new Text(*this+v.to_charPtr()); }
};

typedef Text* TextP;

#endif
