#ifndef __WRAPPER_H__
#define __WRAPPER_H__

#include "xbefore.h"
#include <X11/Intrinsic.h>
#include "xafter.h"

typedef void* WrapperID;

// For InputCB(): XtInputMask is one or more of 
// XtInputReadMask|XtInputWriteMask|XtInputExceptMask

template<class T,class U> class Wrapper
{
	T* instance;
	typedef void (T::*SCB)(Widget w,U udata,XtPointer calld);
	typedef bool (T::*TCB)(U udata);
	typedef bool (T::*ICB)(int fd,U udata);

	struct CBInfo
	{
		enum { Standard,WinMgr,Timer,Input } type;
		Wrapper<T,U>& wr;
		union { Widget w; unsigned long interval; int fd; };
		union { String cbn; XtIntervalId tid; XtInputId iid; };
		union { SCB cb; TCB tcb; ICB icb; };
		U udata;
		
		CBInfo(Wrapper<T,U>& w,Widget wid,String cbname,SCB c,U ud,bool s): 
			type(s?Standard:WinMgr),wr(w),w(wid),cbn(cbname),cb(c),udata(ud) {}
		
		CBInfo(Wrapper<T,U>& w,unsigned long i,TCB c,U ud): 
			type(Timer),wr(w),tcb(c),interval(i),udata(ud) {}

		CBInfo(Wrapper<T,U>& w,int f,ICB c,U ud):
			type(Input),wr(w),fd(f),icb(c),udata(ud) {}
	};

	static void _Callback(Widget w,XtPointer cld,XtPointer calld);
	static void _TimeOut(XtPointer cld,XtIntervalId*);
	static void _Input(XtPointer cld,int*,XtInputId*);
	static void _Destroy(Widget w,XtPointer cld,XtPointer calld);

public:
	Wrapper(T* i): instance(i) {}
	
	WrapperID AddCB(Widget w,String cbname,SCB callback,U udata);
	WrapperID TimerCB(unsigned long interval,TCB callback,U udata);
	WrapperID InputCB(int fd,XtInputMask cond,ICB callback,U udata);
	WrapperID WinMgrCB(Widget shell,String atomname,SCB callback,U udata);
	void RemoveCB(WrapperID id);
};

#endif
