// -----------------------------------------------------------------
// Importe
// -----------------------------------------------------------------

#include "tokengam.h"
// Tokengameklasse

// -----------------------------------------------------------------
// EDSPNTokengame-Klassendeclaration
// -----------------------------------------------------------------

class EDSPNTokengame: public Tokengame, virtual public CommModule, 
                      public Parser
{
protected:

  bool firsttime;

  ObjectP           ParserObject;
  TextP             ParserText;
  bool              RecursionError;
  LinkedList<NodeP> Recursion;
  // Fuer die Fehlerausgaben des Parser

  LinkedList<NodeP> Immediates;
  LinkedList<NodeP> Timed;
  // Liste schaltfaehiger Transitionen

  bool BackwardEnable, ForwardEnable;
  // Statusvariablen

  signed int  Steps, StepsCnt;
  // fuer Backward/Forward

  int  HighestPriority;
  // hoechste Prioritaet der Immediates

  bool ConfigurationChange;
  // Hat sich die Konfiguration veraendert?

  bool ParseAllMarkings();
  NodeP FindMarkingParameter(charP Name);

  void MakeReadyToFire(NodeP node);

  void CollectEnabledTrans();
  bool Enabled(NodeP node);
  void ShowEnabledTrans();

  void SaveParams(StructP params);
  Text Int2Steps(int value);
  int  Steps2Int(StructP params);
  void OnDialogButton(Ident& button);
  // Dialog-Hilfsfunktionen

  struct Stellen { char* Name; long Marken; };
  Stellen* ListeStellen;

// -----------------------------------------------------------------
public:

  EDSPNTokengame(Net& n): Tokengame(n), CommModule(n,"Tokengame"), 
                          Parser(),
                          //AH Changed Start => (ObjectP)(TextP)
                          ParserObject((ObjectP)NULL), 
                          ParserText((TextP)NULL), RecursionError(false),
                          //AH Changed End
                          BackwardEnable(false), ForwardEnable(false), 
                          Steps(1), StepsCnt(1),
                          HighestPriority(0), 
                          ConfigurationChange(true) {};

  ~EDSPNTokengame() {};

  bool Init() {return (net.Prime()->Class()->Name() == "EDSPN");}; 

  bool Open(Editor& e, Page&)
    { 
      this->ed= &e;

      MenuWidget= this->ed->CreateEntry("Menu",xmPushButtonWidgetClass,
					this->ed->Box(), NULL,CallBackID);
      if (MenuWidget)
	this->ed->ButtonState("Menu", false);
      
      return false;
    }

  bool CreateEntry(Editor* ed, char* name, WidgetClass, Widget parent) 
    {
      if (!strcmp(name,"Check Structure"))
	{// "Tokengame" hinter "Check Structure" erzeugen

	  WrapperID ID;
	  ToggleWidget= ed->CreateEntry(Name(),xmToggleButtonWidgetClass,
					parent,NULL, ID);
	}
      return false;
    };
  // Modulfunktionen

  void   ParserError(const charP message, charP string, int length);
  long   ParseMarking(charP string, int length);
  double ParseParameter(charP string, int length);
  double ParseRealParameter(charP string, int length);
  long   ParseIntegerParameter(charP string, int length);
  // Parser-Hilfsfunktionen

  bool  IsPlace(NodeP);
  bool  IsTransition(NodeP);
  bool  IsImmediate(NodeP node);
  bool  IsInhibitor(ConnectorP conn);
  bool  IsParameter(NodeP node);
  bool  IsMarkPar(NodeP node);
  bool  IsDelayPar(NodeP node);
  TextP GetMarking(NodeP);
  int   GetPriority(NodeP node);
  TextP Get_md_enable(NodeP node);
  TextP GetMultiplicity(ConnectorP c);
  int   GetMarkPar(NodeP);
  TextP GetDelayPar(NodeP);
  TextP GetColor(NodeP);
  // Tokengame-Hilfsfunktionen

  void  StartTokengame();
  void  EndTokengame();

  bool  EventEnable();
  void  OpenOptionsMenu();

  bool  SelectEnabledTrans(PageP pg);
  bool  DeselectEnabledTrans(PageP pg);

  void  OpenFireMenu(Node& node);
  void  CopyMarkings();
  bool  IsMarkAb(NodeP node);
  void  FireTrans(NodeP node, int, bool undo);
  // Tokengame-Hauptfunktionen

};
