// -----------------------------------------------------------------
// Debug
// -----------------------------------------------------------------
#define FAUS

// -----------------------------------------------------------------
// Importe
// -----------------------------------------------------------------

#include <sys/time.h>

#include "stdlib.h"
#include "HCPNtokengam.h"
#include "fkomm.h"

// -----------------------------------------------------------------
// Konstanten
// -----------------------------------------------------------------

// maximale Zeilenlaenge der Zusatzdatei fuer die Attribute der
// externen Transitionen.
const unsigned int ZUSATZ_MAX=300;




// **********************************************************************
// In dieser Datei ist die Klasse FAnlage deklariert. Diese unterstuetzt
// die speziellen Funktionen externer Transitionen zum steuern externer
// Prozesse.
// Dabei wird die Klasse Uebergang dazu genutzt Uebergaenge zu vermerken.
// Die Klasse FTrans dient dazu den externen Transitionen zusaetzliche
// Attribute zuzuordnen.
// **********************************************************************




// *********** Klassendeklaration Uebergang (Anfang) ******************

// Die Klasse Uebergang spezifiziert Uebergaenge.
class Uebergang
{
protected:
  // Zeiger auf den Knoten der Transition 
  NodeP         knoten;

  // Nummer des Uebergangs der Transition
  unsigned int  binding;
  
  // Konstruktor
  Uebergang()
    {}

public:
  // Konstruktor
  // Uebergeben wird der Zeiger auf den Knoten der Transition und die
  // Nummer des Uebergangs der Transition.
  Uebergang(NodeP K, unsigned int B)
    : knoten(K), binding(B)
    {}

  // Liefert den Zeiger auf den Knoten der Transition
  NodeP Knoten()
    {
      return knoten;
    }

  // Liefert die Nummer des Uebergangs der Transition.
  unsigned int Binding()
    {
      return binding;
    }
};

// *********** Klassendeklaration Uebergang (Ende) ******************



// *********** Klassendeklaration FTrans (Anfang) ******************

// Die Klasse FTrans dient dazu den externen Transitionen zusaetzliche
// Attribute zuzuordnen.
class FTrans: public Uebergang
{
private:
  // Beschreibung des verknuepftes Auftrags.
  // (Klassenname, die Indentifikationsnummer und den Auftrag der
  // ausfuehrenden Einheit)
  Text auftrag;

  // Gibt den Transitionspfad des Uebergangs an
  Text name;
  
public:
  // aktiv ist true, wenn ein Auftrag begonnen wurde und noch nicht
  // abgeschlossen ist.
  bool          aktiv;

  // abbruch ist true, wenn der Auftrag fehlerhaft abgebrochen wurde
  // und daher erneut gestartet werden muss.
  bool          abbruch;

public:
  // Die Zeit die ein Auftrag zur Ausfuehrung benoetigt hat.
  double        schaltzeit;

  // Konstruktor
  // Uebergeben wird die Beschreibung des zugeordneten Auftrags,
  // der Transitionspfad des Uebergangs, die Nummer des Uebergangs
  // der Transition und ein Zeiger auf den Knoten der Transition.
  FTrans(const char *Auftrag,const char *Name, unsigned int Bindingnr, NodeP Knoten);

  // Destruktor
 ~FTrans();

  // Liefert true, wenn der Knoten des uebergebenen Zeigers mit dem Knoten
  // knoten in der Instanz uebereinstimmt.
  bool          IstKnoten(NodeP Knoten);

  // Liefert die Beschreibung des verknuepftes Auftrags.
  const char   *Auftrag();

  // Die Startanforderung des verknueften Auftrags wird vermerkt.
  // Damit wird der Auftrag aktiv.
  // Der Rueckgabewert ist true, wenn der Auftrag aktiviert werden
  // konnte.
  // Falls ein fehlhafter Abbruch vorlag, so wird dieser Zustand
  // aus dem Auftragsstatus geloescht.
  bool          Start(unsigned int bindingnr);

  // Mit dieser Methode wird das Ende des Auftrags gemeldet.
  // Damit ist der verknuepfte Auftrag nicht mehr aktiv.
  // Als Rueckgabewert wird die Nummer des Uebergangs geliefert. 
  unsigned int  Ende();

  // Der fehlerhafte Abbruch des Auftrags wird vermerkt.
  // Der Auftrag ist beendet.
  void          Abbruch();

  // Liefert true, wenn der Auftrag fehlerhaft abgebrochen wurde.
  bool          Abgebrochen();

  // Liefert true, wenn der Auftrag aktiv ist.
  bool          Aktiv();

  // Liefert den Transitionspfad des Uebergangs an
  const char   *Name();
};

// *********** Klassendeklaration FTrans (Ende) ******************




// *********** Klassendeklaration FAnlage (Anfang) ******************

// Die Klasse FAnlage unterstuetzt
// die speziellen Funktionen externer Transitionen zum steuern externer
// Prozesse.
class FAnlage: public HCPNTokengame, virtual public CommModule
{ 
private:
  timeval zeit_alt;

  // diff_zeit enthaelt immer die Zeit in Millisekunden 
  // seit der letzten Meldung des
  // Simulatorprozesses in der sich das Steuerungsmodul im
  // Steuerungsmodus befunden hat. 
  unsigned long diff_zeit;

  // ExtStrg ist true, wenn externe Prozesse gesteuert werden sollen.
  // Vorgesehen fuer eine Umschaltung von der Steuerung zu einen Markenspiel.
  bool ExtStrg;

  // einstellungs_modus ist true, wenn Einstellungen am Steuerungsprozess
  // vorgenommen werden. Es ist in dieser Zeit keine Steuerung moeglich.
  bool einstellungs_modus;

  // Kom ist eine Schnittstelle mit der Zeichenketten durch Dateien ausgetauscht
  // werden koennen.
  T_FKommunikation *Kom;

  // aktiv ist true, wenn die Steuerung mit dem Steuerungsprozess moeglich ist.
  bool aktiv;

  // kommaktiv ist true, wenn Kommunikation ueber die entsprechenden
  // Dateien moeglich ist.
  bool kommaktiv;

  // Pfadname der Kommunikationsdatei Agnes => Steuerungsprozess
  Text agn2fan_pfad;

  // Pfadname der Kommunikationsdatei Steuerungsprozess => Agnes
  Text fan2agn_pfad;

  // Pfadname der Lockdatei. Mit dieser Datei wird sichergestellt, dass
  // immer nur ein Steuerungsprozess auf die Steuerungsschnittstelle
  // zugreift.
  Text flock_pfad;

  // Liste der Zusatzattribute fuer externe Transitionen
  LinkedList<FTrans *>    TransZusatz;

  // Liste der Transitionen die im Simulatorprozess gefeuert werden
  // sollen.
  LinkedList<Uebergang *> FeuerListe;

  // Diese Methode prueft, ob Rueckmeldungen von Auftraegen vorliegen.
  // Wenn gerade zeitbehaftete Transitionen schaltfaehig sind, werden
  // die Uebergaenge deren Auftraege Rueckmeldungen geliefert vorne an
  // an die Liste FeuerListe angefuegt.
  // Es wird false zurueckgeliefert, wenn ein Fehler aufgetreten ist.
  bool FAuftragLesen();

  // Die Zusatzattribute der externen Transionen werden mit dieser
  // Methode aus der Datei zusatz eingelesen.
  void Lade_ExtZusatz(ifstream& zusatz);

  // Schreibt alle im Netz verwendeten Auftragebeschreibungen
  // in <netname>.auftr Datei.
  void Erzeuge_AuftrTab(Text auftr_tab_pfad);

protected:
  // Process-ID des gestarteten Steuerungsprozesses
  signed int pid2;

public:
  // Konstruktor
  // Uebergeben wird eine Referenz auf das aktuelle Netz.
  FAnlage(Net& n): HCPNTokengame(n),
                   CommModule(n,"Control"),
                   ExtStrg(true),
                   Kom(0),
                   pid2(0) {}

  // Destruktor
  ~FAnlage() {}

  bool Init(); //AH NEW: bestimmt, ob Modul angezeigt wird

  // Liefert true, wenn bei dem uebergebenen Knoten um
  // eine zeitlose Transition handelt
  bool IsImmediateTrans(NodeP node);

  // Liefert true, wenn bei dem uebergebenen Knoten um
  // eine zeitbehaftete Transition handelt
  bool IsTimedTrans(NodeP node);

  // Startet die Petri-Netz-Steuerung.
  void StartTokengame();

  // Beendet die Petri-Netz-Steuerung.
  void EndTokengame();

  // Definiert ein Fenster um Optionen der Steuerung einzustellen.
  void OpenOptionsMenu();

  // Diese Methode wird periodisch mit dem Takt der
  // Blinkfrequenz aufgerufen. 
  // Hiermit werden die Auftraege schaltfaehiger Transitionen
  // an den Steuerungsprozess weitergeleitet, Rueckmeldungen
  // des Prozesses empfangen und ausgwertet und dem Simulationsprozess
  // mitgeteilt welche Transition er schalten soll.
  bool DoPeriodic();

  // Diese Methode benachrichtigt den Simulatorprozess, dass
  // der angegebene Uebergang geschaltet werden soll.
  // Uebergeben wird der Zeiger auf die Transition des Uebergangs,
  // die Nummer des Uebergangs und ein Flag das angibt, ob die
  // uebergebene Transitione rueckwaerts geschaltet werden soll.
  // Wenn ExtStrg true ist und die angegebene Transition eine
  // externe Transition ist wird dem Simulationsprozess ihre
  // Schaltzeit uebergeben. In jedem Fall wird dem
  // Simulatorprozess jedoch der Knotenname und die Nummer
  // des Uebergangs mitgeteilt. 
  void FireTrans(NodeP, int, bool);
  
  // Diese Methode bereitet die uebergebene Transition zum
  // Schalten vor.
  // Uebergeben wird der Zeiger auf die Transition des Uebergangs
  // und die Nummer des Uebergangs.
  // Handelt es sich bei der Transition um eine interne
  // Transition wird der Uebergang in die Liste FeuerListe
  // eingetragen. Ist es eine externe Transition wird sie
  // aktiviert und der Steuerungsprozess benachrichtigt.
  void Fire(FireItemP item, intP nr);

  // Set abhaengig vom aktuellen Modus des Moduls
  // das Verhaeltnis von Schaltzeitzaehlern zu der
  // Blinkfrequenz der Transitionen.
  void SetSpeed();

  // Reagiert auf das Schalten des Knopfes button.
  void ExtendDialogButton(Ident& button);

};

MODULE_NOTIFY(FAnlage)
