#include "HCPN2EDSPN.h"

#define DEBUG

HCPN2EDSPN::HCPN2EDSPN()
{}

// ----------------------------------------------------------------------------

HCPN2EDSPN::~HCPN2EDSPN() 
{
}


int HCPN2EDSPN::FindHCPNPlaceNb(const Text &name)
{
  if(!hcpnplaces) return -1;
  for(unsigned int i=0; i<nb_hcpnplaces; i++)
    {
      if(hcpnplaces[i]->Name() == name)
	return i;
    }
  return -1;
}

//--------------------------------------
//       LoadUnfoldFile
//       load .unfold file that was created by
//       MSAnalysis/bin/Uncolor
//       Fill structures for HCPN- and EDSPN places /transitions 
//--------------------------------------

void HCPN2EDSPN::LoadUnfoldFile(char *filename)
{
  char buffer[10000];
  unsigned int uintbuffer;
  unsigned int i;
  ifstream unfold_file(filename);
  
  //--------------------------------------
  //           read hcpn places
  //--------------------------------------
  while (!unfold_file.eof())
    { 
      unfold_file>>buffer;
      if (strstr(buffer,"_Places_"))
	break;
    }
  
  unfold_file>>nb_hcpnplaces;
  hcpnplaces = new HCPNPlace*[nb_hcpnplaces];
  
  for(i=0; i<nb_hcpnplaces; i++)
    {
      int nbcolors;
      unfold_file>>buffer; //read path + name of place
      unfold_file>>nbcolors; //number of unfolded places for this place
      HCPNPlace* p = new HCPNPlace(Text(buffer), nbcolors);
      hcpnplaces[i] = p;
    }
#ifdef DEBUG
  cout<<"There are "<< nb_hcpnplaces<< "hcpn places"<<endl;
  for(i=0; i<nb_hcpnplaces; i++)
    {
      cout<< hcpnplaces[i]->Name()<< " " << hcpnplaces[i]->NbColors()<<endl;
    }
#endif 
  
  //--------------------------------------
  //           read  hcpn transitions
  //--------------------------------------
  
  int nbbinds;
  int counttrans = 0;
  
  while (!unfold_file.eof())
    { 
      unfold_file>>buffer;
      if (strstr(buffer,"_Transitions_"))
	break;
    }
  unfold_file>>nb_hcpntransitions;
  hcpntransitions = new HCPNTransition*[nb_hcpntransitions];
  
  while (!unfold_file.eof())
    { 
      unfold_file>>buffer;
      if (strstr(buffer,"_Immediate_"))
	break;
    }
  
  unfold_file>>nb_hcpnimmediate;
  
  for(i=0; i<nb_hcpnimmediate; i++, counttrans++)
    {
      unfold_file>>buffer; //read path + name of transition
      unfold_file>>nbbinds; //number of bindings
      HCPNTransition* t = new HCPNTransition(Text(buffer), nbbinds, H_IMMEDIATE);
      hcpntransitions[counttrans] = t;
    }
  
  while (!unfold_file.eof())
    { 
      unfold_file>>buffer;
      if (strstr(buffer,"_Exclusive_"))
	break;
    }
  
  unfold_file>>nb_hcpnexclusive;
  
  for(i=0; i<nb_hcpnexclusive; i++, counttrans++)
    {
      unfold_file>>buffer; //read path + name of transition
      unfold_file>>nbbinds; //number of bindings
      HCPNTransition* t = new HCPNTransition(Text(buffer), nbbinds, H_EXCLUSIVE);
      hcpntransitions[counttrans] = t;
    }

  while (!unfold_file.eof())
    { 
      unfold_file>>buffer;
      if (strstr(buffer,"_Infinite_"))
	break;
    }
  
  unfold_file>>nb_hcpninfinite;
  
  for(i=0; i<nb_hcpninfinite; i++, counttrans++)
    {
      unfold_file>>buffer; //read path + name of transition
      unfold_file>>nbbinds; //number of bindings
      HCPNTransition* t = new HCPNTransition(Text(buffer), nbbinds, H_INFINITE);
      hcpntransitions[counttrans] = t;
    }
#ifdef DEBUG
  cout<<"There are "<< nb_hcpntransitions<< "hcpn transitions"<<endl;
  cout<<"Imm: " << nb_hcpnimmediate<< " Exc: "<<nb_hcpnexclusive<< " Inf: "<< nb_hcpninfinite<< endl;
  for(i=0; i<nb_hcpntransitions; i++)
    {
      cout<< hcpntransitions[i]->Name()<< " " << hcpntransitions[i]->NbBinds()<<" "<< hcpntransitions[i]->Type()  <<endl;
    }
#endif 

  //--------------------------------------
  //        read incidence matrix
  //--------------------------------------

  while (!unfold_file.eof())
    { 
      unfold_file>>buffer;
      if (strstr(buffer,"_Incidence_"))
	break;
    }

  unfold_file>>nb_edspntransitions;
  unfold_file>>nb_edspnplaces;

  edspntransitions = new EDSPNTransition*[nb_edspntransitions];
  edspnplaces = new EDSPNPlace*[nb_edspnplaces];
  
  incidence_matrix = new  int*[nb_edspntransitions];
  for(i=0; i< nb_edspntransitions; i++)
    {
      incidence_matrix[i] = new int[nb_edspnplaces];
    }

  for(unsigned int t = 0; t <  nb_edspntransitions; t++)
    for(unsigned int p = 0; p <  nb_edspnplaces; p++)
      unfold_file>>incidence_matrix[t][p];

   //--------------------------------------
  //        read post matrix
  //--------------------------------------

  while (!unfold_file.eof())
    { 
      unfold_file>>buffer;
      if (strstr(buffer,"_Post_"))
	break;
    }

  unfold_file>>nb_edspntransitions;
  unfold_file>>nb_edspnplaces;

  post_matrix = new  int*[nb_edspntransitions];
  for(i=0; i< nb_edspntransitions; i++)
    {
      post_matrix[i] = new int[nb_edspnplaces];
    }

  for(unsigned int t = 0; t <  nb_edspntransitions; t++)
    for(unsigned int p = 0; p <  nb_edspnplaces; p++)
      unfold_file>>post_matrix[t][p];

  //--------------------------------------
  //        read pre matrix
  //--------------------------------------

  while (!unfold_file.eof())
    { 
      unfold_file>>buffer;
      if (strstr(buffer,"_Pre_"))
	break;
    }

  unfold_file>>nb_edspntransitions;
  unfold_file>>nb_edspnplaces;

  pre_matrix = new  int*[nb_edspntransitions];
  for(i=0; i< nb_edspntransitions; i++)
    {
      pre_matrix[i] = new int[nb_edspnplaces];
    }

  for(unsigned int t = 0; t <  nb_edspntransitions; t++)
    for(unsigned int p = 0; p <  nb_edspnplaces; p++)
      unfold_file>>pre_matrix[t][p];


  //--------------------------------------
  //        read unfolded places
  //--------------------------------------
  
  while (!unfold_file.eof())
    { 
      unfold_file>>buffer;
      if (strstr(buffer,"_UnfoldedPlaces_"))
	break;
    }

  Text edspnname;
  Text hcpnname;
  Text tokenname;
  bool complement;
  unsigned int nbcorrHCPN;
  unsigned int nb_color = 0;
  Text oldname = "";

  for(i=0; i< nb_edspnplaces; i++)
    {
      unfold_file>>buffer; //name of corr. HCPN  place
      Text name = buffer;

      
      if(name == oldname)
	nb_color++;
      else
	nb_color= 0;
      oldname = name;

      edspnname = name;
      unfold_file>>buffer; // "original" or "complement"
      if(strstr(buffer,"complement"))
	{
	  complement = true;
	  edspnname += "COMP";
	}
      else
	 complement = false;
      unfold_file>>buffer; //name of corr. HCPN color
      tokenname = buffer;
      if (tokenname != "e")
	edspnname += tokenname;
      int* postcol = new  int[nb_edspntransitions]; // column of post matrix
      int* precol = new  int[nb_edspntransitions]; // column of pre matrix
      for(unsigned int j=0; j< nb_edspntransitions; j++)
	{
	  postcol[j] = post_matrix[j][i];
	  precol[j] = pre_matrix[j][i];
	}
      
      
      nbcorrHCPN =  FindHCPNPlaceNb(name); //number of corr. HCPN place
      EDSPNPlace* p = new EDSPNPlace(edspnname, nbcorrHCPN, nb_color,
				     complement, tokenname, 0, //marking = 0
				     postcol, precol, i);
      edspnplaces[i] = p;
    } 

  //read initial marking
  while (!unfold_file.eof())
    { 
      unfold_file>>buffer;
      if (strstr(buffer,"_InitialMarking_"))
	break;
    }
  
  unfold_file>>uintbuffer;  unfold_file>>uintbuffer; // read over vector dimensions

  for(i=0; i< nb_edspnplaces; i++)
    {
      unfold_file>>uintbuffer;
      edspnplaces[i]->SetMarking(uintbuffer);
    }


#ifdef DEBUG
  cout<<"EDSPNplaces"<<endl;
  for(i=0; i< nb_edspnplaces; i++)
    {
      cout<<"name: "<<edspnplaces[i]->Name();
      cout<<" hcpn nb: "<<edspnplaces[i]->GetHCPNNumber();
      if(edspnplaces[i]->IsComp()) cout<<" COMP"; else cout<<" NOCOMP";
      cout<<" tokenname: "<<edspnplaces[i]->TokenName()<<endl;
      cout<<" mark: "<<edspnplaces[i]->GetMarking()<<endl;
      cout<<"postcol: ";
      int* pptr = edspnplaces[i]->GetPost();
      for(unsigned int j= 0; j< nb_edspntransitions; j++)
	cout<<pptr[j]<<" ";
      cout<<endl;
      cout<<"precol: ";
      pptr = edspnplaces[i]->GetPre();
      for(unsigned int j= 0; j< nb_edspntransitions; j++)
	cout<<pptr[j]<<" ";
      cout<<endl;
    }
#endif  

  //--------------------------------------
  //     fill structure for EDSPNTransitions
  //--------------------------------------
  int edspntranscount = 0;
  for(i=0; i< nb_hcpntransitions; i++)
    {
      hcpnname = hcpntransitions[i]->Name();
      nbcorrHCPN = i;
      for(unsigned int j=0; j< hcpntransitions[i]->NbBinds(); j++)
	{
	  if( hcpntransitions[i]->NbBinds()>1)
	    edspnname = hcpnname + Float(j+1);
	  else
	    edspnname = hcpnname;
	  int* postrow = new  int[nb_edspnplaces]; // row of post matrix
	  for(unsigned int k=0; k< nb_edspnplaces; k++)
	    postrow[k] = post_matrix[edspntranscount][k];
	  int* prerow = new  int[nb_edspnplaces]; // row of pre matrix
	  for(unsigned int k=0; k< nb_edspnplaces; k++)
	    prerow[k] = pre_matrix[edspntranscount][k];
	  
	  EDSPNTransition* t = new EDSPNTransition(edspnname,
						   nbcorrHCPN,
						   j, // number of binding
						   postrow, prerow, edspntranscount);
	  edspntransitions[edspntranscount] = t;
	  edspntranscount++;
	}
    }
#ifdef DEBUG
  cout<<"EDSPNtransitions"<<endl;
  for(i=0; i< nb_edspntransitions; i++)
    {
      cout<<"name: "<<edspntransitions[i]->Name();
      cout<<" hcpn nb: "<<edspntransitions[i]->GetHCPNNumber();
      cout<<" type: "<<edspntransitions[i]->Type()<<endl;
      cout<<"postrow: ";
      int* pptr = edspntransitions[i]->GetPost();
      for(unsigned int j= 0; j< nb_edspnplaces; j++)
	cout<<pptr[j]<<" ";
      cout<<endl;
      cout<<"prerow: ";
      pptr = edspntransitions[i]->GetPre();
      for(unsigned int j= 0; j< nb_edspnplaces; j++)
	cout<<pptr[j]<<" ";
      cout<<endl;
    }
#endif  
}

//--------------------------------------
//     FillStructures
//     edspnplaces / edspntransitions   
//--------------------------------------



// ----------------------------------------------------------------------------

// calculate the new size (width x height) of a 
// page belonging to subtransition S
// if all subtransitions on the page are 
// exploded and filled into the page
// also calculate the displacement
// that has to be added to a place, transition or subtransition
// on the page when all subtransitions have been exploded
void HCPN2EDSPN::RecalcSize(SubTrans *S)
{
  unsigned int i;
  //Iterator<Node*> nodes = S->Subnodes();
  Node* node;
  Page* p = S->SubPage();

  //initialize width and height with
  //original values 
  S->flattenedWidth = S->Width();
  S->flattenedHeight = S->Height();
  
#ifdef DEBUG
  cout << "Subtransition " << S->Path()
       << S->Width() <<" x "<<S->Height()
       << " Min = "<<S->Minimum()
       << " Max = "<<S->Maximum()<<endl;
#endif

  // list of all subtransitions that are
  // are below S in the hierarchy
  LinkedList<SubTrans*> childs = S->Children();

  //no subtransitions in this page? => return;
  if (childs.Length() == 0)
    return;
  
  //calculate new size of all inferior subtransitions
  OpaqueIterator<SubTrans*> i1(childs);
  for (SubTrans** o=i1; o; o=++i1) 
    {
      RecalcSize(*o);
      //additional space needed when exploding subtrans *o
      //in page of S 
      S->flattenedWidth += (*o)->flattenedWidth;
      S->flattenedHeight += (*o)->flattenedHeight;
    }
  
#ifdef DEBUG
  cout<<"New size of"<<S->Path()<<" : "<<S->flattenedWidth
      <<" x "<<S->flattenedHeight<<endl;
#endif

  //now calculate displacements for all objects
  // on the page belonging to S
  OpaqueIterator<SubTrans*> i2(childs);
  for (SubTrans** s2=i2; s2; s2=++i2) 
    {
      Position subpos = (*s2)->Pos();
      
      //for all places:
      // add a displacement  if they are right or under
      // subtransition (*s2)
      // displacement = size of unfolded subtransition
      for(i=0; i< nb_edspnplaces; i++)
	{
	  Text hcpnname;
	  EDSPNPlace *ep = edspnplaces[i];
	  
	  hcpnname = hcpnplaces[(ep->GetHCPNNumber())]->Name();
	  node = p->FindNode(hcpnname);
	  
	  //place on this page?
	  if(hcpnname.StripLastPathPart('\\') != S->Path())
	    continue;
	  
	  if(node->Mid().x > subpos.x)
	    {
#ifdef DEBUG
	      cout<<"Place "<<ep->Name()<<" is right of " 
		  <<"Subtrans " <<(*s2)->Path()<<endl;
#endif
	      ep->displacement.x += (*s2)->flattenedWidth;
	    }
	  if(node->Mid().y > subpos.y)
	    {
#ifdef DEBUG
	      cout<<"Place "<<ep->Name()<<" is under " 
		  <<"Subtrans " <<(*s2)->Path()<<endl;
#endif
	      ep->displacement.y += (*s2)->flattenedHeight;
	    }
	}
      
      //calculate displacement for transitions
      for(i=0; i< nb_edspntransitions; i++)
	{
	  Text hcpnname;
	  EDSPNTransition *et = edspntransitions[i];
	  
	  hcpnname = hcpntransitions[(et->GetHCPNNumber())]->Name();
	  node = p->FindNode(hcpnname);
	  
	  if(hcpnname.StripLastPathPart('\\') != S->Path())
	    continue;
	  
	  if(node->Mid().x > subpos.x)
	    {
#ifdef DEBUG
	      cout<<"Trans "<<et->Name()<<" is right of " 
		  <<"Subtrans " <<(*s2)->Path()<<endl;
#endif
	      et->displacement.x += (*s2)->flattenedWidth;
	    }
	  if(node->Mid().y > subpos.y)
	    {
#ifdef DEBUG
	      cout<<"Trans "<<et->Name()<<" is under " 
		  <<"Subtrans " <<(*s2)->Path()<<endl;
#endif
	      et->displacement.y += (*s2)->flattenedHeight;
	    }
	}

      //calculate displacement for subtransitions
      OpaqueIterator<SubTrans*> i3(childs);
      for (SubTrans** s3=i3; s3; s3=++i3) 
	{
	  if((*s3)->Pos().x > subpos.x)
	    {
#ifdef DEBUG
	      cout<<"SubTrans "<<(*s3)->Path()<<" is right of " 
		  <<"Subtrans " <<(*s2)->Path()<<endl;
#endif
	      (*s3)->displacement += (*s2)->flattenedWidth;
	    }
	  if((*s3)->Pos().y > subpos.y)
	    {
#ifdef DEBUG
	      cout<<"SubTrans "<<(*s3)->Path()<<" is under " 
		  <<"Subtrans " <<(*s2)->Path()<<endl;
#endif
	      (*s3)->displacement += (*s2)->flattenedHeight;
	    }
	}
    }
}

//walk recursively through the hierarchy and 
// calculate the final displacements for all objects
void HCPN2EDSPN::RecalcPositions(SubTrans *S, Position displace)
{
  unsigned int i;
  Node* node;
  Page* p = S->SubPage();
  LinkedList<SubTrans*> childs = S->Children();
 
  Position newdisplace = displace + S->displacement;

  for(i=0; i< nb_edspnplaces; i++)
    {
      Text hcpnname;
      EDSPNPlace *ep = edspnplaces[i];
      
      hcpnname = hcpnplaces[(ep->GetHCPNNumber())]->Name();
      node = p->FindNode(hcpnname);
      
      //place on this page?
      if(hcpnname.StripLastPathPart('\\') != S->Path())
	continue;

      ep->displacement += newdisplace;
    }
  
  for(i=0; i< nb_edspntransitions; i++)
    {
      Text hcpnname;
      EDSPNTransition *et = edspntransitions[i];
      
      hcpnname = hcpntransitions[(et->GetHCPNNumber())]->Name();
      node = p->FindNode(hcpnname);
      
      if(hcpnname.StripLastPathPart('\\') != S->Path())
	continue;
      
      et->displacement += newdisplace;
    }

  OpaqueIterator<SubTrans*> i1(childs);
  for (SubTrans** o=i1; o; o=++i1) 
      RecalcPositions(*o, newdisplace);
}


void HCPN2EDSPN::FillStructures(Text &errortext, Page *p)
{
  unsigned int i;
  bool showflag = true;

  id_number = 0;

  errortext = "";

  //=================================
  //       Subtransitions
  //=================================

  SubTrans *S = new SubTrans(p, 0, 0);

#ifdef DEBUG
  S->PrintHierarchy();
#endif

  //======================================
  //Calculate position-displacements for subnets
  //======================================
  
  RecalcSize(S);
  RecalcPositions(S);

  //=================================
  //       Places
  //=================================

  for(i=0; i< nb_edspnplaces; i++)
    {
      Text hcpnname;
      Node* node;
      EDSPNPlace *ep = edspnplaces[i];

      id_number++;
      ep->id = id_number; //=====> identification number

      hcpnname = hcpnplaces[(ep->GetHCPNNumber())]->Name();
      node = p->FindNode(hcpnname);
      if(!node)
	{
	  errortext += Text("HCPN2EDSPN::FillStructure : Cant find ") + hcpnname + "\n";
	  return;
	}

      ep->scale = node->Scale();//=====> scale
      ep->relatives = node->GetRelative(0); //=====>relatives
      ep->orientation = node->Angle(); //=====>orientation
      ep->position= node->Mid();
      unsigned int nbcolor = ep->GetNbColor();
      ep->position.x = nbcolor*40+ep->position.x; // #
      ep->position.y = nbcolor*10+ep->position.y; //  # =====>position
    }

  //=================================
  //       Transitions
  //=================================

  for(i=0; i< nb_edspntransitions; i++)
    {
      Text hcpnname;
      Node* node;
      EDSPNTransition *et = edspntransitions[i];

      id_number++;

      et->id = id_number; //=====> identification number
      
      hcpnname = hcpntransitions[(et->GetHCPNNumber())]->Name();
      node = p->FindNode(hcpnname);
      if(!node)
	{
	  errortext += Text("HCPN2EDSPN::FillStructure : Cant find ") + hcpnname + "\n";
	  return;
	}

      //read global condition
      Text globalCondition = (*node->Attr("global_guard")).to_charPtr();

      //read binding
      int bind_nb = et->GetBindNumber();
      Text bind = (*((List*)(node->Attr("bindings")))->GetEntry( bind_nb)).to_charPtr(); //brrr!

#ifdef DEBUG
      cout<< "Bind "<< bind_nb << " of hcpntrans. " <<hcpnname <<": "<< bind<<endl; 
#endif
     
      //parse binding
      Text localCondition;
      List inArcs;
      List outArcs;
      Text weight;
      int delaytype; //D_IMMEDIATE, D_EXPONENTIAL...
      Text delay;
      
      TokenizeBinding(bind, localCondition, inArcs, outArcs, weight,
		      delaytype, delay);

      if (localCondition == "-" || localCondition =="") //no condition
	localCondition = Text();
      if (globalCondition == "-" || globalCondition =="") //no condition
	globalCondition = Text();
      
#ifdef DEBUG
      cout<< "Local cond: " << localCondition<< endl;
      if(inArcs.Length())
	{
	  cout<<"Inarcs: "  <<endl;
	  for (int j=0; j<inArcs.Length(); j++)
	    cout<<*(inArcs.GetEntry(j))<<endl;
	}
      if(outArcs.Length())
	{
	  cout<<"outarcs: "  <<endl;
	  for (int j=0; j<outArcs.Length(); j++)
	    cout<<*(outArcs.GetEntry(j))<<endl;
	}
      cout<<"Weight: "<<weight<<endl;
      cout<< "DelayType : " << delaytype<< endl;
      cout<< "Delay: "<< delay<< endl;
#endif
       
      et->scale = node->Scale();//=====> scale
      et->sel_col = "blue"; //=====> sel_col
      et->fire_pol = "PRS"; //=====> fire_pol
      et->relatives = node->GetRelative(0); //=====>relatives
      et->orientation = node->Angle(); //=====>orientation
      et->position= node->Mid();

      et->position.x = bind_nb*40+et->position.x; // #
      et->position.y = bind_nb*10+et->position.y; //  # =====>position

      switch(delaytype)//=====> IsMD
	{
	case D_IMMEDIATE:	  
	case D_EXPONENTIAL:
	case D_DETERMINISTIC:
	  if (delay.Contains("#")) //marking dependent expression
	    {
	      et->IsMD = true;
	      errortext += "Warning: marking dependent weight/delay in transition ";
	      errortext +=   et->Name() +"\n";
	      errortext += "Please correct manually \n";
	    }
	  else
	    et->IsMD = false;
	 break;
	default:
	  break;
	}

      if(localCondition.to_bool() || globalCondition.to_bool()) //=====> md_enable
	{
	  switch(delaytype)
	    {
	    case D_IMMEDIATE:
	      errortext += "Warning: imm. transition ";
	      errortext += et->Name() +
		" contains a marking dependent enabling function \n";
	      errortext += "Please correct manually \n";
	      if(localCondition.to_bool())
		et->md_enable= globalCondition + " & " + localCondition;
	      else
		et->md_enable= globalCondition;
	      break;
	    default:
	      if(localCondition.to_bool() || globalCondition.to_bool())
		{
		  errortext += "Warning: transition ";
		  errortext +=  hcpnname +
		       " contains a guard:\n";
		  if(globalCondition.to_bool())
		    {
		      errortext += "global guard: ";
		      errortext += globalCondition + "\n";
		    }
		  if(localCondition.to_bool())
		    {
		      errortext += "local guard: ";
		      errortext += localCondition + "\n";
		    }
		  if(showflag)
		    {
		      errortext += " In EDSPNs, guards are only defined for imm. transitions.\n";
		      errortext += " Although it would be possible in many cases to replace \n";
		      errortext += " guards by net constructions, this is not implemented yet.\n";
		      showflag = false;
		    }
		}
	      break;     
	    }
	}

      switch(delaytype)  //=====> type
	{ 
	case D_IMMEDIATE:
	  et->type =  E_IMMEDIATE;
	  et->weight = delay; //thats correct    //=====> weight
	  et->priority = (int)(*node->Attr("priority")).to_double();  //brrr! //=====> priority
	  break;
	case D_EXPONENTIAL:
	  et->type = E_EXPONENTIAL; 
	  et->delay = delay;   //=====> delay
	  et->inf_serv =   //=====> inf_serv
	    (Text("infinite")==(*node->Attr("server_type")).to_charPtr());
	  break;
	case D_DETERMINISTIC:
	  et->type = E_DETERMINISTIC;
	  et->delay = delay; 
	   break;
	case D_UNIFORM:
	case D_GENERAL:
	  et->type = E_GENERAL;
	  et->pmffunc = delay;  //=====> pmffunc
	  break;
	}
    }
}

// ----------------------------------------------------------------------------
// classify places and tranitions: on page S? 
// for transitions that are onpage: 
// classify transitions:  input? (arc from buffer place to transition)
//                        output?(arc from transition to buffer place)
//                        interior?(connected to on page place)
void HCPN2EDSPN::ClassifyNodes(SubTrans *S)
{
  unsigned int i, j;

  S->p_onpage = new EDSPNPlace*[nb_edspnplaces];
  S->t_onpage = new EDSPNTransition*[nb_edspntransitions];
  S->interior = new EDSPNTransition*[nb_edspntransitions];
  S->input = new EDSPNTransition*[nb_edspntransitions];
  S->output = new EDSPNTransition*[nb_edspntransitions];

  //classify places into onpage / not on page
  for(i=0; i< nb_edspnplaces; i++)
    {
      Text hcpnname;
      EDSPNPlace *ep = edspnplaces[i];
      
      hcpnname = hcpnplaces[(ep->GetHCPNNumber())]->Name();
     
      //place on/below this page?
      if(!hcpnname.BeginsWith(S->Path()))
	{
	  S->p_onpage[i] = 0;
	  continue;
	}
      S->p_onpage[i] = ep;
    }
  
  //classify transitions into onpage, 
  //input, output, interior (not disjunct)
  for(i=0; i< nb_edspntransitions; i++)
    {
      Text hcpnname;
      EDSPNTransition *et = edspntransitions[i];
      
      hcpnname = hcpntransitions[(et->GetHCPNNumber())]->Name();
      S->input[i] = 0;
      S->output[i] = 0;
      S->interior[i] = 0;      

      if(!hcpnname.BeginsWith(S->Path()))
	{
	  S->t_onpage[i] = 0;
	  continue; //not on page
	}

      S->t_onpage[i] = et; //on page
      int* postrow = et->GetPost();
      int* prerow = et->GetPre();
      for(j=0; j< nb_edspnplaces; j++)
	{
	  if(!(S->p_onpage[j])) 
	    { 
	      // trans connected to place j which is not on page?
	      if(postrow[j]>0) 
		S->output[i] = et; //input trans
	      else if (prerow[j]>0) 
		S->input[i] = et; //output trans
	    }
	  else
	    {
	      // trans connected to place j which is on page?
	      if(postrow[j]!=0 || prerow[j]!=0)
		S->interior[i] = et; //interior trans
	    }
	}
    }

#ifdef DEBUG
  cout<<"Print classification of all transitions for "<<S->Path()<<endl;
  for(i=0; i< nb_edspntransitions; i++)
    {
      EDSPNTransition *et = edspntransitions[i];
      
      cout<<"Transition" << et->Name()<<" is ";
      if(!(S->t_onpage[i]))
	{
	  cout<<"not on page";
	}
      else 
	{
	  if (S->input[i])
	    {
	      cout<<"I";
	    }
	  if (S->output[i])
	    {
	      cout<<"O";
	    }
	  if(S->interior[i])
	    cout<<"interior";
	}
	cout<<"."<<endl;
    }
#endif
}

// ----------------------------------------------------------------------------
// FUNCTIONS OF CLASS SubTrans
// ----------------------------------------------------------------------------
// descend through all subpages
// and store their path name
// and the minimum and maximum position 
// of the objects contained in the subpage
SubTrans::SubTrans(Page* subpage, Node* parentnode, SubTrans* parenttrans)
{
  Node* node;
  Position akt, min, max, subtranspos;
  bool firsttime = true;
  Text pathname, subname;
  Iterator<Node*> nodes = subpage->Nodes();

  if(parenttrans)
    pathname = parenttrans->Path();
  else
    pathname = "";
  if(parentnode)
    {
      subtranspos = parentnode->Mid();
      if(pathname!="")
	subname = pathname + "\\" + parentnode->Name();
      else
	subname = parentnode->Name();
    }
  else
    subname = "";

  //walk through all nodes contained in this subtrans
  //compute minimum and maximum position of objects
  for (node=nodes; node; node=++nodes)
    { 
      if (node->Virtual()) continue;
      if(firsttime)
	{
	  min = max = node->Mid();
	  firsttime = false;
	}
      akt = node->Mid();
      if(min.x>akt.x) min.x = akt.x;
      else if (max.x < akt.x) max.x = akt.x;
      if(min.y>akt.y) min.y = akt.y;
      else if (max.y < akt.y) max.y = akt.y; 
    }

  this->path = subname;
  this->pos  = subtranspos;
  this->min  = min; 
  this->max  = max;
  this->subpage = subpage;
  this->parenttrans = parenttrans;
		 
  //now find subtransitions and perform recursive call
  for (node=nodes; node; node=++nodes)
    { 
      if (node->Virtual()) continue;
      if (node->Class()->Name()=="SubTrans")
	{ 
	  SubTrans *N = new SubTrans(node->SubPage(), node, this);
	  this->AppendSubTrans(N);
	}
    }
}

// ----------------------------------------------------------------------------
#ifdef DEBUG
//debugging only:
//print structures
void SubTrans::PrintHierarchy(Text tab)
{
  cout<<tab.to_charPtr()<<"Pathname   : "<< Path() << endl;
  cout<<tab.to_charPtr()<<"Size (x,y) :  ("<< Width()<<","<< Height()<<")"<<endl;

  Iterator<Node*> nodes = Subnodes();
  Node* node;
  cout<<tab.to_charPtr()<<"Contains the following objects:"<<endl;
  for (node=nodes; node; node=++nodes)
    {
      cout<<tab.to_charPtr()<<node->Class()->Name()<<" ";
      cout<<node->Name()<<" ";
      if(node->Virtual())
	cout<<"VIRTUAL";
      cout<<endl;
    }
  LinkedList<SubTrans*> childs = Children();
  cout<<tab.to_charPtr()<<"Contains "<<childs.Length()<<" subtrans."<<endl;
  OpaqueIterator<SubTrans*> i(childs);
  for (SubTrans** o=i; o; o=++i) 
    {
      (*o)->PrintHierarchy(tab+"  ");
    }
}

#endif


