#ifndef HCPN2EDSPN_H
#define HCPN2EDSPN_H
// ----------------------------------------------------------------------------
// HCPN2EDSPN -- Module to convert a HCPN to an EDSPN
// helper classes:
// EDSPNNode, EDSPNTransition, EDSPNPlace
// HCPNTransition, HCPNPlace
// SubTrans
// ----------------------------------------------------------------------------
// includes
// ----------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
//for Motif
#include "xbefore.h"
#include <Xm/ToggleB.h>
#include <Xm/PushB.h>
#include "xafter.h"
//from Agnes/inc
#include "linked.h"
#include "commmod.h"
#include "module.h"
#include "main.h"
#include "editor.h"
#include "net.h"
#include "node.h"
#include "netcl.h"
#include "dialog.h"
#include "wrapper.h"
#include "list.h"
#include "position.h"
//for generation of .table-file
#include "interface.h"
//base class
#include "parser.h"
// ----------------------------------------------------------------------------
// class declarations
// ----------------------------------------------------------------------------
// class for SubTransitions
// contains pointers to the related pages / objects
// and information about positions
class EDSPNTransition;
class EDSPNPlace;
class SubTrans
{
  Text path; //pathname of Subtrans;
  Position pos; //position of transition on page
  Position min; // position of uppermost /leftmost object in transitions subpage 
  Position max; // position of downmost / rightmost object
  
  Page* subpage; //pointer to subpage
  
  SubTrans* parenttrans; //pointer to parent subtransition

  LinkedList<SubTrans*> children; //list of subtransitions on subpage
 public:
  SubTrans(Text path, Position pos,
	   Position min, Position max, Page* subpage,
	   SubTrans* parenttrans):
    path(path),
    pos(pos),
    min(min),
    max(max),
    subpage(subpage),
    parenttrans(parenttrans)
    { 
      children =  LinkedList<SubTrans*>();
      flattenedWidth = flattenedHeight = 0;
      displacement= Position();
    } 

  SubTrans(Page* subpage, Node* parentnode, SubTrans* parenttrans);
  
  double flattenedWidth; //heigth an width when all subpages
  double flattenedHeight; // have been exploded
  Position displacement; // displacement on toppage when all
                         // subtrans. have been exploded

  EDSPNPlace **p_onpage; //place is on subpage
  EDSPNTransition **t_onpage; //transition is on subpage
  EDSPNTransition **interior; //interior  transition of subpage
  EDSPNTransition **input; //input transition of subpage
  EDSPNTransition **output; //output transition of subpage

  LinkedList<SubTrans*> Children() {return children;} 
  Text Path() {return path;}
  double Width() { return max.x - min.x;  }
  double Height() { return max.y - min.y;  }
  Position Minimum(){return min;}
  Position Maximum(){return max;}
  Position Pos(){return pos;}
  Page* ParentPage() { if (parenttrans) return parenttrans->SubPage(); else return 0;}
  Page* SubPage() { return subpage;}
  Iterator<Node*> Subnodes() { return subpage->Nodes();}
  SubTrans* ParentSub() {return parenttrans; }
  void AppendSubTrans(SubTrans* sub) {children.Append(sub);}
  void PrintHierarchy(Text tab=Text());
};
// ----------------------------------------------------------------------------
//class for HCPN places
class HCPNPlace
{
  Text name;  //name of a product or elementary HCPN-place
  unsigned int nbcolors; //number of different colors that can occur in this place
 public:
  HCPNPlace(Text name, unsigned int nbcolors): name(name), nbcolors(nbcolors) {}
  Text Name() {return name;}
  unsigned int NbColors() {return nbcolors;}
  bool operator == (HCPNPlace p) const { return name==p.name && nbcolors==p.nbcolors; } 
};
// ----------------------------------------------------------------------------
//class for HCPN transitions
//Agnes-node classification of transitions
#define N_TIMED    0
#define N_IMMEDIATE 1
//MSAnalysis-classification of transitions
#define H_IMMEDIATE 0
#define H_EXCLUSIVE 1
#define H_INFINITE  2
//delay-type-classification of trans:: from parser.h
//#define D_IMMEDIATE 0
//#define D_EXPONENTIAL 1
//#define D_DETERMINISTIC  2
//#define D_UNIFORM 4
//#define D_GENERAL 5
//Agnes-node classification of transitions: from parser.h
//#define N_TIMED    0
//#define N_IMMEDIATE 1
class HCPNTransition
{
  Text name;  //name of a  HCPN-transitions
  unsigned int nbbinds; //number of bindings of transition
  unsigned int type; // IMMEDIATE, EXCLUSIVE or INFINITE
 public:
  HCPNTransition(Text name, unsigned int nbbinds, unsigned int type): name(name), nbbinds(nbbinds), type(type) {}
  Text Name() {return name;}
  unsigned int NbBinds() {return nbbinds;}
  unsigned int Type() {return type; }
  bool operator == (HCPNTransition t) const { return name==t.name && nbbinds==t.nbbinds && type ==t.type; } 
};
// ----------------------------------------------------------------------------
//class EDSPNNode: base class for EDSPNTransition and EDSPNPlace
#define E_TRANSITION 0
#define E_PLACE 1
class EDSPNNode
{
  Text name;  //name of EDSPN-node
  short type; //type of node, transition or place 
  unsigned int nbcorr_HCPN; // number of corresponding HCPN node
  unsigned int spec_HCPN; /* HCPN specification:
			     color number for places,
			     binding number for transitions*/
  int* post; /* column of post matrix belonging to place
			   or row of post matrix belonging to trans */
  int* pre; /* column of pre matrix belonging to place
			   or row of pre matrix belonging to trans */
  unsigned int index; /*number of place/transition and index of matrix row or column*/

 public:
  EDSPNNode(Text name, 
	    short type,
	    unsigned int nbcorr_HCPN,
	    unsigned int spec_HCPN,
	    int* post,
	    int* pre,
	    unsigned int index):
    name(name), 
    type(type),
    nbcorr_HCPN(nbcorr_HCPN),
    spec_HCPN(spec_HCPN),
    post(post),
    pre(pre),
    index(index){}
	   
  virtual ~EDSPNNode(){};
  //variables used for the net file
  unsigned int id; //identification number
  Position scale; //scale of place, is simply copied from the hcpn-place
  Position relatives; //position of name relative to place (copied from hcpn)
  Position position; //position of place
  double orientation; //orientation of place (copied from hcpn)
  Position displacement;

  Node* hcpnnode; //to store a pointer to a node

  Text Name() {return name;};
  void SetName(Text n) {name = n;};
  unsigned int GetHCPNNumber() {return nbcorr_HCPN;};
  int* GetPost() {return post;}
  int* GetPre() {return pre;}
  short GetType() {return type;}
  unsigned int GetIndex() {return index;}
  unsigned int GetSpec() {return spec_HCPN;}
  void SetPost(int* p) { post = p;}
  void SetPre(int* p) { pre = p;}
  virtual bool operator == (EDSPNNode n) const { return name==n.name 
						   && nbcorr_HCPN == n.nbcorr_HCPN
						   && spec_HCPN == n.spec_HCPN;} 
};
// ----------------------------------------------------------------------------
//class for EDSPN places
class EDSPNPlace: public EDSPNNode
{
  bool complement; //was created to replace a HCPN place with capacity
  Text tokenname; // name of the color this place represents 
  unsigned int marking; //marking of place
 public:
  EDSPNPlace(Text name, unsigned int nbcorr_HCPN,  unsigned int nb_color,
	     bool complement, Text tokenname, unsigned int marking,
	     int* postcol, int* precol, unsigned int index): 
    EDSPNNode(name, E_PLACE, nbcorr_HCPN, nb_color, postcol, precol, index),
    complement(complement),
    tokenname(tokenname),
    marking(marking)
    {
    displacement = Position();}

  //methods

  Text TokenName() {return tokenname;}
  unsigned int GetNbColor() {return GetSpec();}
  bool IsComp() {return complement;}
  unsigned int GetMarking() {return marking;}
  void SetMarking(unsigned int m) {marking = m;}
  bool operator == (EDSPNPlace p) const { return (EDSPNNode)*this==(EDSPNNode)p;} 
};
// ----------------------------------------------------------------------------
//class for EDSPN transitions
#define E_IMMEDIATE        0
#define E_EXPONENTIAL      1
#define E_GENERAL          2
#define E_DETERMINISTIC    3
class EDSPNTransition: public EDSPNNode
{
 public:
  EDSPNTransition(Text name, unsigned int nbcorr_HCPN, 
		  unsigned int  bind_no,  int* postrow, int* prerow, unsigned int index):
    EDSPNNode (name, E_TRANSITION, nbcorr_HCPN, bind_no, postrow, prerow, index)
    {
      displacement = Position();}

  //variables used for the net file
  unsigned int type; // IMMEDIATE, EXPONENTIAL, GENERAL, DETERMINISTIC
 
  //attributes  
  Text weight; //imm
  bool IsMD; //imm, exp, det
  unsigned int priority; //imm
  Text fire_pol; //all:    "PRS" or "PRD"
  Text sel_col; //all
  Text md_enable; //imm 
  Text delay; //exp, det
  bool inf_serv; // exp
  Text pmffunc; //gen

  unsigned int Type() {return type; }
  unsigned int GetBindNumber() {return GetSpec();}
  bool operator == (EDSPNTransition t) const { return (EDSPNNode)*this==(EDSPNNode)t;} 
};
// ----------------------------------------------------------------------------
// class HCPN2EDSPN, class to support conversion of a HCPN net to a EDSPN (unfold)
class HCPN2EDSPN: public Parser
{
protected:
 
  unsigned int nb_hcpnplaces; // number of hcpn places
  HCPNPlace** hcpnplaces; // vector of hcpn places 
  unsigned int nb_hcpntransitions; // number of hcpn trans.
  HCPNTransition** hcpntransitions; // vector of hcpn trans.
  
  unsigned int nb_hcpnimmediate; // number of immediate,
  unsigned int nb_hcpnexclusive; // exclusive
  unsigned int nb_hcpninfinite; // and infinite hcpn transitions

  unsigned int nb_edspnplaces; // number of edspn places 
  EDSPNPlace** edspnplaces; // vector of edspn places
  unsigned int nb_edspntransitions;  // number of edspn trans.
  EDSPNTransition** edspntransitions; // vector of edspn trans

  int** incidence_matrix; //incidence matrix of edspn net edspntransitions x edspn_places
  int** post_matrix; //incidence matrix of edspn net edspntransitions x edspn_places
  int** pre_matrix; //incidence matrix of edspn net edspntransitions x edspn_places
   
  unsigned int id_number; //running id number of objects
public:

  void FillStructures(Text &errortext, Page *p);
  void RecalcSize(SubTrans *S);
  void RecalcPositions(SubTrans *S, Position displace=Position());
  void LoadUnfoldFile(char* filename);
  
  void ClassifyNodes(SubTrans* S); //classify nodes into input/output-transitions etc.
  int FindHCPNPlaceNb(const Text &name); //get number of HCPN place named name

  HCPN2EDSPN();
  virtual ~HCPN2EDSPN();
};

#endif
