#include "HCPNUnfolder.h"

#ifndef NoExplicitInstanciation
#define NoExplicitInstanciation
#include "dialog.cc"
#include "filesel.cc"
#include "wrapper.cc"
FSELINSTANCIATE(HCPNUnfolder,Editor*)
DIAINSTANCIATE(HCPNUnfolder,bool)
#endif

#define DEBUG

  //MODULE_NOTIFY(HCPNUnfolder) deactivated

HCPNUnfolder::HCPNUnfolder(Net& n): 
  CommModule(n,"HCPNUnfolder"), 
  Wrapper<HCPNUnfolder,bool>(this),
  fsel(this)
{}

// ----------------------------------------------------------------------------

HCPNUnfolder::~HCPNUnfolder() 
{
  CloseFilesel();
  DestroyFilesel();
}

bool HCPNUnfolder::Open(Editor& e, Page& p)
{// Wird aufgerufen, wenn Init() true zurueckgeliefert hat

  this->ed= &e;
  this->p= &p;
  // Den Editor merken, in dem das Tokengame laeuft

 WrapperID dummy;
 
 this->ed->CreateEntry("Export",
		       xmPushButtonWidgetClass, this->ed->Shell(), NULL, dummy);

  return false;
  // keinen automatischen Menueeintrag erzeugen, dies wurde schon gemacht!
}

// ----------------------------------------------------------------------------

bool HCPNUnfolder::Menu(Editor& editor, Page& page, Ident& name, bool before)
{// Wird aufgerufen, wenn ein Menuepunkt angeklickt wurde
  Text netname = Text(net.FileName()) - ".net";  
  Text stdout_filename = netname + ".out";   
  Text unfold_filename = netname + ".unfold";

  if (name =="Unfold..." && before)
    {
      CreateLogwin(stdout_filename.to_charPtr());
      OpenLogwin(editor, "Unfold...", "Ok");
      
      ofstream stdout_file(stdout_filename.to_charPtr());
      
      stdout_file<<"HCPN2EDSPN Unfolder V0.1"<<endl;
      
      stdout_file<<"Creating table file..."<<flush;
      if (!CreateTableFile(netname))
	{
	  stdout_file<<"\t Unable to create table file!"<<endl;
	  return true;
	}
      stdout_file<<"done"<<endl;

      stdout_file<<"Calling uncolor to create .unfold-file...";
      if (!CallUncolor(netname))
	{
	  stdout_file<<"\t Error when calling uncolor!"<<endl;
	  return true;
	}
      stdout_file<<"done"<<endl;

      stdout_file<<"reading .unfold-file...";
      LoadUnfoldFile(unfold_filename.to_charPtr()); //initialize HCPN2EDSPN
      stdout_file<<"done"<<endl;
           
      Text errortext;
      stdout_file<<"filling structures...";
      FillStructures(errortext, p);
      stdout_file<<"done"<<endl;
      stdout_file<<errortext.to_charPtr();

      Text title= Text("Unfold Net ")+net.FullName()+" to...";
      char *path= editor.CurrentPath();
      fsel.Open(editor.Shell(),title.to_charPtr(),&HCPNUnfolder::SaveEDSPN, &editor,
		path ? path : model_dir.to_charPtr(), /*search dir */
		((net.FullName())-".net"+"*").to_charPtr(), /*file mask: netname* */
		((net.FullName())-".net"+"_EDSPN.net").to_charPtr()); /*selection: netname_EDSPN.net */

      // DestroyLogwin();

      return true;
      // keine weitere Behandlung fuer Export
    }

    return (false); // nicht fuer dieses Modul
}

// ----------------------------------------------------------------------------

bool HCPNUnfolder::CreateTableFile(Text& netname)
{
  Error error(*LastEditor());
  
  Text table_file = netname + ".table";
 
  if ((table_file).to_charPtr()[0]!='/') table_file = model_dir + "/" + table_file;
 
  ofstream file(table_file.to_charPtr());
  AnalyserInterface::DescribeSymbols(net.Prime(), file, true); 
  AnalyserInterface::DescribePage(net.Prime(), file, true, error);
  file.close();

  if (error || !file) return false;

  return true;
}

// ----------------------------------------------------------------------------

bool HCPNUnfolder::CallUncolor(Text& netname)
{
  Text DialogContense;
  Text ana_dir= tnet_dir + "/MSAnalysis/bin/";
  Text params = ana_dir + "uncolor" + " -N " + netname - ".net";
  int exitcode;

  pid = Exec(params.to_charPtr(), 0, 0);
  
  // if(WaitForProcess(pid, exitcode, 30)) // with timeout
  if(WaitForProcess(pid, exitcode))
    return true;
  else if (exitcode == CANTRUN)
    {
      DialogContense= "I was unable to run \n";
      DialogContense+= params + "\n";
      DialogContense+= "The program does not exist";
    }
  else if (exitcode == UNNORMALEXIT)
    {
      DialogContense= "I was unable to run \n";
      DialogContense+= params + "\n";
      DialogContense+= "The program has been killed before termination";
    }
  else if (exitcode == TIMEOUT)
    {
      DialogContense= "Waited too long, program\n";
      DialogContense+= params + "\n";
      DialogContense+= "will be killed";
      kill(pid, SIGINT);
    }
  else 
    {
      DialogContense= "The program call \n" ;
      DialogContense+= params + "\n";
      DialogContense+= "returned error number " ;
      DialogContense+= double(exitcode);
      DialogContense+= "\n";
    }
  OpenMsgDialog(*this->ed, "Unfold_Error",
		DialogContense.to_charPtr(),true,"Exit");
  return false;
}

void HCPNUnfolder::SaveEDSPN(Editor *ed, char *filename, bool ok)
{
  unsigned int i;
  if(!ok) return;
  ofstream EDSPNfile(filename);
  EDSPNfile<<"EDSPN ("<<endl;
  EDSPNfile<<"("<<endl;
  if (nb_edspnplaces && nb_edspntransitions)
    {
      EDSPNfile<<"objects = ("<<endl;
      for(i=0; i< nb_edspnplaces; i++)
	{
	  EDSPNPlace* ep = edspnplaces[i];
	  EDSPNfile<<"Place (name = ";
	  EDSPNfile<<ep->Name().ReplaceChar('\\','_').to_charPtr();  /* slashes for path
								       are replaced by _-signs*/
	  EDSPNfile<<";";
	  //write ID-Number
	  EDSPNfile<<"  id = "<<ep->id<<"; ";
	  EDSPNfile.setf(ios::fixed, ios::floatfield);
	  EDSPNfile.precision(1);
	  EDSPNfile<<"scale = "<<ep->scale<<"; ";	 
	  EDSPNfile<<"attributes = {marking = \"";
	  EDSPNfile<<ep->GetMarking()<<"\";}; ";
	  EDSPNfile<<"relatives = ("<<ep->relatives<<"); ";
	  EDSPNfile<<"position = ("<<ep->position.x+ep->displacement.x
		   <<","<<ep->position.y+ep->displacement.y<<"); ";
	  EDSPNfile<<"orientation = "<<ep->orientation<<")";
	  if(nb_edspntransitions == 0 && i ==  nb_edspnplaces - 1)
	    EDSPNfile<<endl;
	  else
	    EDSPNfile<<","<<endl;
	}
      for(i=0; i< nb_edspntransitions; i++)
	{
	  EDSPNTransition *et = edspntransitions[i];
	  switch (et->type)
	    {
	    case E_IMMEDIATE: 
	      EDSPNfile<<"ImmTrans (name = ";
	      break;
	    case E_EXPONENTIAL:
	      EDSPNfile<<"ExpTrans (name = ";
	      break;
	    case E_GENERAL:
	      EDSPNfile<<"GenTrans(name = ";
	      break;
	    case E_DETERMINISTIC:
	      EDSPNfile<<"DetTrans (name = ";
	    }
	  EDSPNfile<<et->Name().ReplaceChar('\\','_').to_charPtr();
	  EDSPNfile<<";";
	  EDSPNfile<<"  id = "<<et->id<<"; ";
	  EDSPNfile<<"scale = "<<et->scale<<"; ";
	  EDSPNfile<<"attributes = {";
	  switch(et->type)
	    { 
	    case E_IMMEDIATE:
	      EDSPNfile<<"weight = \""
		       <<et->weight.to_charPtr()<<"\";";
	      if(et->IsMD)
		EDSPNfile<<" IsMD = true;";
	      else
		EDSPNfile<<" IsMD = false;";
	      EDSPNfile<<" priority = "<<et->priority<<";";
	      EDSPNfile<<" fire_pol = "<<et->fire_pol.to_charPtr()<<";";
	      if(et->md_enable.to_bool())
		EDSPNfile<<" md_enable = \""<<et->md_enable.to_charPtr()<<"\";"; 
	      else
		EDSPNfile<<" md_enable = \"\";";
	      break;
	    case E_EXPONENTIAL:
	      EDSPNfile<<"delay = \""
		       <<et->delay.to_charPtr()<<"\";";
	      if(et->IsMD)
		EDSPNfile<<" IsMD = true;";
	      else
		EDSPNfile<<" IsMD = false;";
	      if(et->inf_serv)
		EDSPNfile<<" inf_serv = true;";
	      else
		EDSPNfile<<" inf_serv = false;";
	      EDSPNfile<<" fire_pol = "<<et->fire_pol.to_charPtr()<<";";
	      break;
	    case E_GENERAL:
	      EDSPNfile<<"pmfunc = \""
		       <<et->pmffunc.to_charPtr()<<"\";";
	      EDSPNfile<<" fire_pol = "<<et->fire_pol.to_charPtr()<<";";
	      break;   
	    case E_DETERMINISTIC:
	      EDSPNfile<<"delay = \""
		       <<et->delay.to_charPtr()<<"\";";
	      if(et->IsMD)
		EDSPNfile<<" IsMD = true;";
	      else
		EDSPNfile<<" IsMD = false;";
	      EDSPNfile<<" fire_pol = "<<et->fire_pol.to_charPtr()<<";";
	      break;
	    }
	  EDSPNfile<<" selcol = \""<<et->sel_col.to_charPtr()<<"\";}; ";
	  EDSPNfile<<"relatives = ("<<et->relatives<<"); ";
	  EDSPNfile<<"position = ("<<et->position.x+et->displacement.x
		   <<","<<et->position.y+et->displacement.y<<"); ";
	  EDSPNfile<<"orientation = "<<et->orientation<<")";
	  if(nb_edspnplaces == 0 && i ==  nb_edspnplaces - 1)
	    EDSPNfile<<endl;
	  else
	    EDSPNfile<<","<<endl;
	}

      //now write arcs
      NodeClass *placeclass, *transclass;
      placeclass = p -> GetNodeClass("EPlace"); //place and transition
      transclass = p -> GetNodeClass("TimedTrans");// templates
      Iterator<Position*> prels =  placeclass->Relatives(); /*standard positions
							     of relatives*/
      Iterator<Position*> trels =  transclass->Relatives(); /*standard positions
							     of relatives*/
      double zoom = 1.0; //zoom fixed to 1.0

      unsigned int arccounter = 0;
      bool firstarc = true;
	for(unsigned int t = 0; t<nb_edspntransitions; t++)
	  for(unsigned int p = 0; p<nb_edspnplaces; p++) {
	    int tpweight = post_matrix[t][p]; //arcs from transition to place
	    if (tpweight)
	      {
		arccounter++;
		id_number++;
		if(firstarc)
		  firstarc = false;
		else
		  EDSPNfile<<","<<endl;
		
		EDSPNfile<<"Arc (name = A"<<arccounter<<"; id = "<<id_number
			 <<"; scale = (0.25,0.25); attributes = {multi = \"";
		
		EDSPNfile<<tpweight;
		EDSPNfile<<"\"; IsMD = false;}; relatives = ((0,0),(-50,0));"
			 <<" input = ";
		
		EDSPNPlace *ep = edspnplaces[p];
		EDSPNTransition *et = edspntransitions[t];
		
		EDSPNfile<<et->id<<"; output = "<<ep->id;
		
		EDSPNfile<<"; positions = (";
		
		Position placepos = ep->position + ep->displacement;
		Position transpos = et->position + et->displacement;
		Position placescale = ep->scale;
		Position transscale = et->scale;
		Position start, end;
		
		ShapeData p_sd =
		  ShapeData(zoom, placepos, placescale, ep->orientation, prels , prels , true);
		ShapeData t_sd =
		  ShapeData(zoom, transpos, transscale, et->orientation, trels , trels , true);
		
		placeclass->Nearest(p_sd, transpos ,end);
		transclass->Nearest(t_sd, placepos ,start);
		
		EDSPNfile<<start<<","<<end<<"))";
	      }
	    
	    int ptweight = pre_matrix[t][p]; //arcs from place to transition
	    if (!ptweight) continue;
	    arccounter++;
	    id_number++;
	    if(firstarc)
	      firstarc = false;
	    else
	      EDSPNfile<<","<<endl;
	    
	    EDSPNfile<<"Arc (name = A"<<arccounter<<"; id = "<<id_number
		     <<"; scale = (0.25,0.25); attributes = {multi = \"";

	    EDSPNfile<<ptweight;
	    EDSPNfile<<"\"; IsMD = false;}; relatives = ((0,0),(-50,0));"
		     <<" input = ";
	    
	    EDSPNPlace *ep = edspnplaces[p];
	    EDSPNTransition *et = edspntransitions[t];
	    
	    EDSPNfile<<ep->id<<"; output = "<<et->id;
	   
	    EDSPNfile<<"; positions = (";
	    
	    Position placepos = ep->position + ep->displacement;
	    Position transpos = et->position + et->displacement;
	    Position placescale = ep->scale;
	    Position transscale = et->scale;
	    Position start, end;
	    
	    ShapeData p_sd =
	      ShapeData(zoom, placepos, placescale, ep->orientation, prels , prels , true);
	    ShapeData t_sd =
	      ShapeData(zoom, transpos, transscale, et->orientation, trels , trels , true);
	    
	    placeclass->Nearest(p_sd, transpos ,start);
	    transclass->Nearest(t_sd, placepos ,end);
	 
	    EDSPNfile<<start<<","<<end<<"))";

	  }
	EDSPNfile<<endl<<")"; /*close object list*/
    }
  EDSPNfile<<")"<<endl<<endl;
  EDSPNfile<<")"<<endl;
  EDSPNfile.close();
}


bool HCPNUnfolder::CreateEntry(Editor *ed, char *name, WidgetClass, Widget parent)
    {
      WrapperID ID;

      if (!strcmp(name,"SaveAs"))
	{// Unfold hinter SaveAs erzeugen

	  ed->CreateEntry("Unfold...",
			  xmPushButtonWidgetClass, parent, NULL, ID);
	}
      return false;
    };

