// -----------------------------------------------------------------
// Debug Definitionen
// -----------------------------------------------------------------
//#define __OUTLIST__ 
// Ausgabe der Stellen und Transitionslisten

// -----------------------------------------------------------------
// Importe
// -----------------------------------------------------------------

#include <signal.h>
// Fuer Signale

#include <iostream.h>
// Fuer Ein-/Ausgabe 

#include "interface.h"
// Fuer die Erstellung der .table-Datei

#include "tokengam.h"
// Tokengameklasse

// -----------------------------------------------------------------
// Definitionen
// -----------------------------------------------------------------

#define NETFILENAME   ".net"
#define TABLEFILENAME ".table"
#define OUTFILENAME   ".tok"
#define LISTFILENAME  ".list"
#define FIREFILENAME  ".fire"
// Dateikennungen

#define MAX_LENGTH 256
// maximale Laenge eines Transitionsnamens
// mit Pfadangabe und Indexnummern feuernder bindings
// Bei Aenderungen auch den Slave.cc aendern!

// -----------------------------------------------------------------
// HCPNTokengame-Klassendeclaration
// -----------------------------------------------------------------

class HCPNTokengame: public Tokengame, virtual public CommModule
{ 
protected:

  Text        netname;

  int         pid;
  // Process-ID des gestarteten Slave

  bool        AniEnable, BackwardEnable, 
              ForwardEnable, 
              RealTime, Received;
  // Statusvariablen

  signed int  Steps, StepsCnt, Speed, SpeedCnt, LowestSpeed;
  // Schrittzaehler und Geschwindigkeiten fuer Automatismen 
  // (Backward/Forward/Animation)

  NodeP node;
  // die als naechstes zu schaltende Transition

  Dialog<HCPNTokengame,bool> dialog;
  // Fire-Dialogbox
  void DialogCallback(Dialog<HCPNTokengame,bool>& dia,bool,Ident& btn);
  // dazugehoeriger CallBack
  StructP setup;
  // Fuer die Parameterliste

  NodeP Fastest;
  int*  FastestB;
  double FastestTimePercentage;
  // vom Slave: schnellste Transition und binding und delay

  void ConnectionWithSlave(Text& netname);
  // Kontakt mit dem Slave herstellen
    
  Text Bool2Text(bool value);
  Text Int2Text(int value);
  Text Int2Steps(int value);
  int  Steps2Int(StructP params);

  virtual void SetSpeed();
  // Setzt die Geschwindigkeit je nach "RealTime" nach
  // der derzeit schnellsten Transition oder auf 0

  bool ReadInput(ifstream& infile);
  // Empfaengt und liest die Listen des Slave ein

  NodeP FindNode(PageP pg, charP PathName, bool SearchForTrans,
		 bool SelectSubTrans);
  // Liefert den Knoten mit Namen SearchName 
  // (z.B. "\SubTrans_1\...\SubTrans_n\SearchName")

  void OnDialogButton(Ident& button);
  virtual void ExtendDialogButton(Ident&) {};
  // Wenn ein Dialog-Button angeklickt wird

public:

  HCPNTokengame(Net& n): Tokengame(n),CommModule(n,"Tokengame"), 
			 netname(""),
                         pid(0),  
                         AniEnable(false), 
                         BackwardEnable(false), ForwardEnable(false), 
                         RealTime(false), Received(false), 
                         Steps(1), StepsCnt(1), Speed(0), SpeedCnt(0), 
                         LowestSpeed(20),
                         dialog(this), 
                         FastestB(0) {}

  ~HCPNTokengame() {}

  bool Init() {return (net.Prime()->Class()->Name() == "HCPN");}; 

  bool  IsPlace(NodeP node);
  bool  IsTransition(NodeP node);
  TextP GetMarking(NodeP node);
  TextP GetColor(NodeP node);

  void StartTokengame();
  void EndTokengame();

  bool EventEnable(){return (!AniEnable && !BackwardEnable && !ForwardEnable);};

  void OpenOptionsMenu();

  bool SelectEnabledTrans(PageP);
  bool DeselectEnabledTrans(PageP);

  void OpenFireMenu(Node& n);
  void FireTrans(NodeP, int, bool);
  
  virtual void Fire(FireItemP item, intP binding);
};
