// ----------------------------------------------------------------------------
// Importe
// ----------------------------------------------------------------------------
#include <sys/time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <iostream.h>
#include <builtin.h>
#include <typeinfo>

#include "HCPN2EDSPN.h"
#include "MSIPDecomp.h"

#define DEBUG
#define EQCLASSDEBUG

MODULE_NOTIFY(MSIPDecomp)

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//             USER INTERFACE FUNCTIONS      
// ----------------------------------------------------------------------------
//called from InitDecomposition. tries to find directory with old decomposition
void MSIPDecomp::StartDecomposition()
{
  Text bsname = dirname+ "/" + purename+"BS.net";
  Text unfold_filename = netname + ".unfold";
  ifstream bs_file(bsname.to_charPtr()); //basic skeleton net 
  ifstream u_file(unfold_filename.to_charPtr()); //unfold file
  
  if(bs_file && u_file)
    {//decomposition exists
      Text BoxContents;
      BoxContents+= (Text("There is an old decomposition of this net. \nUse it?")).to_charPtr();
      bs_file.close(); u_file.close();
      OpenMsgDialog(*this->ed, "Decomposition_Question", 
		    BoxContents.to_charPtr(), false, "\"Yes, Load\" \"No, Recompute\"");
    }
  else
    ComputeDecomposition();
}
// ----------------------------------------------------------------------------
void MSIPDecomp::EndDecomposition()
{ 
  net.ChangeEditor(*ed,net.Prime()); //switch to top page
  net.UpdateButtonStates(*ed, *(net.Prime()));
}
// ----------------------------------------------------------------------------
//when decomposition button is clicked
void MSIPDecomp::OpenOptionsMenu()
{
  Text DialogContense;
  DialogContense+= Text("[enum ( \"Display Aggregated\" Aggregated, ");
  DialogContense+= Text("\"Display Extended\" Extended ) ");
  DialogContense+= Text("\"Doubleclick on subtransition: \"displaymode= ");
  DialogContense+= displaymode + ";";
  DialogContense+= Text("]");
  CreateDialog(DialogContense.to_charPtr());
  OpenDialog(*this->ed, "DecompositionOptions", "Analyze Close");
}

// ----------------------------------------------------------------------------
//when any dialog button is clicked
void MSIPDecomp::OnDialogButton(Ident& button)
{
  Struct* params= GetParams();

  if (button=="Yes, Load")
    { //load decomposition
      CloseDialog();
      DestroyDialog();
      LoadDecomposition();
    }
  else if (button=="No, Recompute")
    { //compute decomposition
      CloseDialog();
      DestroyDialog();
      ComputeDecomposition();
    }
  else if ((button=="Exit Decomp."))
    { //exit
      DeInitDecomposition(); 
      CloseDialog();
      DestroyDialog();
    }
  else if(button=="Close")
    { //close decomposition window
      displaymode = (*params->GetMember("displaymode")).to_charPtr();
      CloseDialog();
      DestroyDialog();
    }
  else if((button=="Ok") || (button=="Cancel") || (button=="Exit"))
    { //close other dialog boxes
      CloseDialog();
      DestroyDialog();
    }
  else if(button=="Analyze")
    { //open analysis dialog
      CloseDialog();
      DestroyDialog();
      CreateDialog(AnalysisParameters().to_charPtr());
      SetCurrentDialog(1);
      OpenDialog(*ed, Name(), const_cast<char*>("Start Cancel"));
    }
  else if(button=="Start")
    { //start analysis
      Struct *paramcopy = (Struct*)params->Copy();
      AnalyzeDecomposition(paramcopy);
      SetCurrentDialog(1);
      CloseDialog();
      DestroyDialog();   
    }
}
// ----------------------------------------------------------------------------
//when any logwin button is clicked
void MSIPDecomp::OnLogwinButton(Ident& button)
{
 if(button=="Stop")
   { //stop analysis, kill program masterrg and slaves
      killpg(getpgid(pid), SIGKILL); //kill process group
      DestroyLogwin();
    }
 if((button=="Ok") || (button=="Close"))
    {
      DestroyLogwin();
    }
}
// ----------------------------------------------------------------------------
//load existing decomposition
void MSIPDecomp::LoadDecomposition()
{
  Text bsname = dirname+ "/" + purename+"BS.net";
  Text unfold_filename = netname + ".unfold";
  //read net and build sub trans hierarchy
  top = new SubTrans(p, 0, 0);
  //load unfold file
  LoadUnfoldFile(unfold_filename.to_charPtr()); //init HCPN2EDSPN
  FillStructures();
  //load basic skeleton
  BasicSkeleton=0;
  BasicSkeleton=BasicSkeleton->Load(bsname.to_charPtr());
  if(!BasicSkeleton)
    {
      errortext = Text("Can't load ") + bsname;
      OpenMsgDialog(*this->ed, "Decomposition_Error", 
		    errortext.to_charPtr(), false, "\"Exit\"");
      DeInitDecomposition();
      return;
    }
  //load low level systems
  LinkedList<SubTrans*> childs = top->Children();
  OpaqueIterator<SubTrans*> i2(childs);
  for (SubTrans** o=i2; o; o=++i2) 
    {
      Text lowname = dirname+ "/" + purename+ (*o)->Path() + ".net";
      Net *n=0;
      n=n->Load(lowname.to_charPtr());
       if(!n)
	 {
	   errortext = Text("Can't load ") + lowname;
	   OpenMsgDialog(*this->ed, "Decomposition_Error", 
			 errortext.to_charPtr(), false, "\"Exit\"");
	   DeInitDecomposition();
	   return;
	 }
      LowLevelSystems.Set((*o), n); //store net in an array
      ClassifyNodes((*o));//build input, interior, input, output arrays
    } 
}
// ----------------------------------------------------------------------------
//compute decomposition: check for subtranitions, call uncolor,
// initialize HCPN2EDSPN,  create equivalence classes,
//create directory, low level systems and basic skeleton 
void MSIPDecomp::ComputeDecomposition()
{
  Text bsname = dirname+ "/" + purename+"BS.net";
  Text DialogContense;
  Text stdout_filename =  netname + OUTFILENAME;
  Text unfold_filename = netname + ".unfold";

  ofstream stdout_file(stdout_filename.to_charPtr());
      
  CreateLogwin(stdout_filename.to_charPtr());
  OpenLogwin(*ed, "Decompose...", "Ok");
  //check for subtransitions
  stdout_file<<"Looking for subtransitions on the top page...";
  top = new SubTrans(p, 0, 0);
  if (top->Children().Length()==0)
    {
      stdout_file<<"No subtransitions!"<<endl;
      DeInitDecomposition();
      return;
    }
  stdout_file<<top->Children().Length()<<" subs found."<<endl;
  //call uncolor
  stdout_file<<"Creating table file..."<<flush;
  if (!CreateTableFile(netname, net))
    {
      stdout_file<<"\t Unable to create table file!"<<endl;
      DeInitDecomposition();
      return;
    }
  stdout_file<<"done"<<endl;

  stdout_file<<"Calling uncolor to create .unfold-file...";
  if (!CallUncolor(netname))
    {
      stdout_file<<"\t Error when calling uncolor!"<<endl;
      return;
    }    
  stdout_file<<"done"<<endl;

  stdout_file<<"reading .unfold-file...";
  LoadUnfoldFile(unfold_filename.to_charPtr()); //Init HCPN2EDSPN
  FillStructures(); //add position, scale... information
  stdout_file<<"done"<<endl;
  //calculate eq. classes
  stdout_file<<"calculating equivalence classes...";
  LinkedList<SubTrans*> childs = top->Children();
  OpaqueIterator<SubTrans*> i1(childs);
  for (SubTrans** o=i1; o; o=++i1) 
    {
      EClasslist_T *s_classes =	new EClasslist_T(); //one classlist for every subpage    
      CalcClasses(*o, *s_classes); //calculate classlist
      eclassarray.Set((*o), s_classes); //store classes in an array
    }
  stdout_file<<"done"<<endl;
  //compute SIPs
  stdout_file<<"computing SIPS for each class...";
  for (SubTrans** o=i1; o; o=++i1) 
    {//for all subtransitions
      ComputeSIPS(*o, *eclassarray[*o]);
    }
  stdout_file<<"done"<<endl;
  //make directory
  Text makedir = "mkdir ";
  makedir= makedir + dirname;
  system(makedir.to_charPtr());
 
  stdout_file<<makedir.to_charPtr()<<endl;
  stdout_file<<"creating basicSkeleton/ low level systems...";
  //create nets
  OpaqueIterator<SubTrans*> i2(childs);
  for (SubTrans** o=i2; o; o=++i2) 
    { 
      Text lowname = dirname+ "/" + purename+ (*o)->Path() + ".net";
      Net *n=CreateLowLevelSystem(lowname, (*o)->Path());
      LowLevelSystems.Set((*o), n); //store net in an array
    }  

  BasicSkeleton = CreateBasicSkeleton(bsname);

  stdout_file<<"done"<<endl;
  stdout_file<<"Click on the substitution transitions to see"<<endl;
  stdout_file<<"the aggregated subnets. You may move objects"<<endl;
  stdout_file<<"and save the modified net. Other actions are"<<endl;
  stdout_file<<"not allowed while decomposition is active."<<endl;
}
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//        FUNCTIONS FOR ANALYSIS
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//used to set up the analyze dialog box
Text MSIPDecomp::AnalysisParameters()
{
  return
    "[ int \"Max. Iter. of Approx. Algorithm\" m_max_iter = 20;"
    "  int \"Max. Iter. for adjusting throughputs in BS\" t_max_iter = 10;"
    "  text(0) \"Epsilon for Approx. Algorithm\" t_eps = \"1e-4\";"
    "  enum(\"Iterative (SOR)\"iter, \"Direct (Gauss/Seidel)\" direct)"
    "       \"Solution Method\" method = direct; "
    "  int \"Max. Iter. for SOR\" max_iter = 10000;"
    "  text(0) \"Epsilon for solving eq. system\" epsilon = \"1e-7\";"
    "  bool \"Verbose Mode\" debug = false; ]";
}
// ----------------------------------------------------------------------------
//when clicking "Analyze": create file ".delays" with initial delays and start
//external program MSAnalysis/bin/masterrg for iterative analysis
void MSIPDecomp::AnalyzeDecomposition(Struct* params)
{
  Text bsname = dirname+ "/" + purename+"BS.net";
  Text DialogContense;
  Text stdout_filename =  netname + OUTFILENAME;
  Text unfold_filename = netname + ".unfold";

  ofstream stdout_file(stdout_filename.to_charPtr());
   
  CreateLogwin(stdout_filename.to_charPtr());
  OpenLogwin(*ed, "Decompose...analyzing", "Ok");
  //create table files for BS and LSi
  stdout_file<<"Creating .table files..."<<flush;

  CreateTableFile(bsname, *BasicSkeleton);
  
  LinkedList<SubTrans*> childs = top->Children();
  OpaqueIterator<SubTrans*> i2(childs);
  for (SubTrans** o=i2; o; o=++i2) 
    { 
      Text lowname = dirname+ "/" + purename+ (*o)->Path() + ".net";
      Net *lnet = *LowLevelSystems[(*o)];
      CreateTableFile(lowname, *lnet);
    }  
  stdout_file<<"done"<<endl;
  //create .delays file
  stdout_file<<"Creating .delays file with initial transition delays...";
  Text delaysfilename = dirname+ "/" + purename + ".delays";
  ofstream delaysfile(delaysfilename.to_charPtr());
  delaysfile<<"subtransitions "<<childs.Length()<<endl;
  delaysfile<<"t_eps "<<(*(params->GetMember("t_eps"))).to_charPtr()<<endl;
  delaysfile<<"t_max_iter "<<(*params->GetMember("t_max_iter"))<<endl;
  delaysfile<<"m_max_iter "<<(*params->GetMember("m_max_iter"))<<endl;
  delaysfile<<"epsilon "<<(*(params->GetMember("epsilon"))).to_charPtr()<<endl;
  delaysfile<<"max_iter "<<(*params->GetMember("max_iter"))<<endl;
  delaysfile<<"flags ";
  if (Text((*params->GetMember("debug")).to_charPtr()) == "true")
    delaysfile<<" -v";
  if (Text((*params->GetMember("method")).to_charPtr()) == "iter")
    delaysfile<<" -i";
  delaysfile<<endl;
  for (SubTrans** o=i2; o; o=++i2) 
    {
      delaysfile<<(*o)->Path().to_charPtr();
      unsigned int count=0;
      for(unsigned int i=0; i< nb_edspntransitions; i++)
	if((*o)->input[i] || (*o)->output[i]) //is input or output of subtrans
	  count++;
      delaysfile<<" "<<count<<endl;
      for(unsigned int i=0; i< nb_edspntransitions; i++)
	{
	  if((*o)->input[i] || (*o)->output[i]) //is input or output of subtrans
	    {
	      EDSPNTransition *et=edspntransitions[i];
	      Text hcpnname = hcpntransitions[(et->GetHCPNNumber())]->Name();
	      Node* hnode = p->FindNode(hcpnname);
	      delaysfile<<et->Name().StripFirstPathPart('\\').to_charPtr()<<" "<<et->GetBindNumber()<<" ";
	      double delay = ReadExpDelay(hnode, et->GetBindNumber());
	      delaysfile<<delay<<endl;
	    }
	}
    }
  stdout_file<<"done"<<endl;
  //run masterrg
  Text command = tnet_dir + "/MSAnalysis/bin/masterrg "+ " -N " + netname + " -D" + dirname;
  Text analysis_filename =  netname + ".analyze";
  stdout_file<<"Calling "<<command<<endl;
  
  ofstream analysis_file(analysis_filename.to_charPtr());   
  CreateLogwin(analysis_filename.to_charPtr());
  OpenLogwin(*ed, "Iterative Approximation", "Close Stop");


  pid = Exec(command.to_charPtr(),analysis_filename.to_charPtr() , 0);
  stdout_file<<"Iterative approximation algorithm started"<<endl;
  delete params;
}
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//        FUNCTIONS FOR CREATING BASIC SKELETON AND LOW LEVEL SYSTEMS
// ----------------------------------------------------------------------------
//create low level system for subtransition path, name of netfile = lowname
Net *MSIPDecomp::CreateLowLevelSystem(Text lowname, Text path)
{
  Net *cnet = p->GetNet()->Copy();
  cnet->SetFileName(lowname.to_charPtr());
  Page *cprime = cnet->Prime();
  
  LinkedList<SubTrans*> childs = top->Children();
  OpaqueIterator<SubTrans*> i1(childs);
  for (SubTrans** o=i1; o; o=++i1) 
    {
      if ((*o)->Path() == path) 
	CreateExtension(*o, **eclassarray[*o], cprime);   //subsystem s_i is extended ...
      else
	CreateAggregation(*o, **eclassarray[*o], cprime); //while the others are aggregated
    } 
  cnet->Save();
  return cnet;
} 
// ----------------------------------------------------------------------------
// create basic skeleton net in file named bsname
Net *MSIPDecomp::CreateBasicSkeleton(Text bsname)
{
  Net *cnet = p->GetNet()->Copy();
  cnet->SetFileName(bsname.to_charPtr());
  Page *cprime = cnet->Prime();
  
  LinkedList<SubTrans*> childs = top->Children();
  OpaqueIterator<SubTrans*> i1(childs);
  for (SubTrans** o=i1; o; o=++i1) 
    {
      //aggregate all subsystems
      CreateAggregation(*o, **eclassarray[*o], cprime);
    } 
  cnet->Save();
  return cnet;
} 
// ----------------------------------------------------------------------------
//calculate position of the new place as average of the positions
//of its in- and output transitions
Position MSIPDecomp::CalcMSIPPosition(EDSPNPlace *newnode)
{
  Position newpos = Position(0,0);	    
  int divisor = 0;
  for(unsigned int tt=0; tt<nb_edspntransitions; tt++)
    {
      if(newnode->GetPost()[tt])
	{
	  divisor++;
	  newpos = newpos + edspntransitions[tt]->position;
	}
      if(newnode->GetPre()[tt])
	{
	  divisor++;
	  newpos = newpos + edspntransitions[tt]->position;
	}
    }
  return newpos / divisor;
}
// ----------------------------------------------------------------------------
//put results on  page cprime:
//throughput of all interface transitions
void MSIPDecomp::PutTPResults(SubTrans* S, Page *cprime)
{
  NodeClass *resultclass = cprime -> GetNodeClass("Result");
  for(unsigned int tt=0; tt<nb_edspntransitions; tt++)
    {
      if(S->input[tt] || S->output[tt])
	{
	  //new result
	  Text resultname = Text("TP") +
	    ((edspntransitions[tt]->Name()).StripFirstPathPart('\\')).ReplaceChar('\\','_');
	  Position resultpos = Position(20,20);
	  Node* result =
	    new Node(resultclass, cprime,
		     resultname.to_charPtr(), resultpos);
	  //attributes for result  
	  Text attributes = "{ attributes = { def = \"TP[\\";
	  attributes +=  (edspntransitions[tt]->Name());
	  attributes +="]\"; val = \"\";};}";
	  istrstream is(attributes.to_charPtr(),100);
	  is>>*result; 
	}
    }
}
// ----------------------------------------------------------------------------
// add the msips in s_classes to subpage
void MSIPDecomp::AddMSIPS(EClasslist_T &s_classes, Page *subpage)
{ //create msip places and respective bindings 
  NodeClass *eplaceclass = subpage -> GetNodeClass("EPlace");

  for (unsigned int i=0; i<(unsigned int)s_classes.Length(); i++) 
    { //for all classes
      EClass_T *class_i = *s_classes[i];
      NodeList_T *new_nodes  = class_i->new_nodes;
      for(int l=0; l<new_nodes->Length(); l++)
	  { //for all MSIPS in class_i
	    EDSPNPlace *newplace= static_cast<EDSPNPlace*>(*(*new_nodes)[l]);
	    double orientation = 1.0;
	    Position newpos = CalcMSIPPosition(newplace);
	    //create new placenode 
	    Node* placenode =  
	      new Node(eplaceclass, subpage,
		       newplace->Name().to_charPtr(),
		       newpos, orientation);
	    Text attributes= "{ attributes = { marking = \"";
	    attributes= attributes +  Float(newplace->GetMarking());
	    attributes += "\";};}";	    
	    istrstream is(attributes.to_charPtr(),100);
	    is>>*placenode;  
	    //draw arcs, create bindings
	    for(unsigned int tt=0; tt<nb_edspntransitions; tt++)
	      {
		if(!newplace->GetPost()[tt]) continue;
		EDSPNNode* input=edspntransitions[tt];	
		//draw connector from input transition to new place
		DrawConnection(input->Name().StripFirstPathPart('\\').to_charPtr(),
			       newplace->Name().to_charPtr(), subpage);
		AddConnection(input->hcpnnode, 0, Text("e"), 0, placenode);
	      }
	   for(unsigned int tt=0; tt<nb_edspntransitions; tt++)
	      {
		if(!newplace->GetPre()[tt]) continue;
		EDSPNNode* output=edspntransitions[tt];
		//draw connector from place to output transition
		DrawConnection(newplace->Name().to_charPtr(),
			       output->Name().StripFirstPathPart('\\').to_charPtr(), subpage);
		AddConnection(output->hcpnnode, 0, Text("e"), placenode, 0);
	      }
	  }
    } 
}
// ----------------------------------------------------------------------------
// put unfolded input and output transitions on subpage.
// bufferonly triggers if the transition should be connected 
// to buffers only (aggregation) or to all places (extension)
void MSIPDecomp::PutInputOutput(SubTrans *S, Page *subpage, bool bufferonly)
{
  for(unsigned int i=0; i< nb_edspntransitions; i++) {
    if(!(S->output[i]) && !(S->input[i])) //neither output nor input
      continue;

    EDSPNTransition *et = edspntransitions[i];
    int bindnr = et->GetBindNumber();
    Text hcpnname = hcpntransitions[(et->GetHCPNNumber())]->Name();
    Node* hnode = p->FindNode(hcpnname); 
    Text subsubname = (hcpnname.StripFirstPathPart('\\')).StripLastPathPart('\\');
    
    Text trelpath = (et->Name()).StripFirstPathPart('\\');
    Text tpurename = trelpath.StripPath('\\'); 
    Position newpos = hnode->Mid(); //position: same as HCPN-trans
    NodeClass *nclass = hnode->Class();
    
    //create transition node
    Node* trans =  new Node(nclass, subpage->FindPage(subsubname), 
			    tpurename.to_charPtr(),
			    newpos, et->orientation);
    et->hcpnnode = trans; //store pointer to node 
    
    //draw connections from transition to places 
    for(unsigned int j=0; j< nb_edspnplaces; j++)
      {
	EDSPNPlace *ep = edspnplaces[j];
	Text hname = hcpnplaces[(ep->GetHCPNNumber())]->Name().StripFirstPathPart('\\');
	if(bufferonly && S->p_onpage[j])
	  continue;
	if(et->GetPost()[j]>0) //output trans
	  {
	    //draw connector from place to transition
	    DrawConnection(trelpath, hname, subpage);
	    //create a binding
	    CreateConnection(trans, 0, hnode, bindnr, //create binding 0 
			     ep->TokenName(), 0, subpage->FindNode(hname));
	  }
	if(et->GetPre()[j]>0) //input trans 
	  {
	    //draw connector from transition to place
	    DrawConnection(hname, trelpath, subpage);
	    //create a binding
	    CreateConnection(trans, 0, hnode, bindnr, 
			     ep->TokenName(), subpage->FindNode(hname), 0);
	  }	  
      }
  }
}  
// ----------------------------------------------------------------------------
// remove original subsystem for subtransition S.
// restore all interface transitions and put MSIPs on the page. 
// create result parameters for every page and every transition
void MSIPDecomp::CreateAggregation(SubTrans* S, EClasslist_T &s_classes, Page *cprime)
{
  unsigned int i;
  Node* subnode = cprime->FindNode(S->Path());
  Page* subpage = subnode->SubPage(); 
 
  //delete all objects on the subpage
  //except buffer places, subtransitions
 //delete transitions
 for(i=0; i< nb_edspntransitions; i++) 
   {
     EDSPNTransition *et = edspntransitions[i];
     if(!(S->t_onpage[i])) continue;
     Text hname = hcpntransitions[(et->GetHCPNNumber())]->Name().StripFirstPathPart('\\');
     delete subpage->FindNode(hname);
   }
 //delete places
 for(i=0; i< nb_edspnplaces; i++) 
   {
     EDSPNPlace *ep = edspnplaces[i];
     if(!S->p_onpage[i]) continue;
     Text hname = hcpnplaces[(ep->GetHCPNNumber())]->Name().StripFirstPathPart('\\');
     (subpage->FindNode(hname))->RemoveNodeAndAliases();
   }

 //re-create input and output transitions, but unfolded
 PutInputOutput(S, subpage, true); //connect only buffers = true 
 AddMSIPS(s_classes,subpage);
 PutTPResults(S, cprime);
}
// ----------------------------------------------------------------------------
// create extended subnet: original net + msips
void MSIPDecomp::CreateExtension(SubTrans* S, EClasslist_T &s_classes, Page *cprime)
{
  unsigned int i;
  Node* subnode = cprime->FindNode(S->Path());
  Page* subpage = subnode->SubPage(); 
 //delete input and output transitions

 for(i=0; i< nb_edspntransitions; i++) 
   {
     EDSPNTransition *et = edspntransitions[i];
     if(!(S->output[i]) && !(S->input[i])) //neither output nor input
       continue;
     Text hname = hcpntransitions[(et->GetHCPNNumber())]->Name().StripFirstPathPart('\\');
     delete subpage->FindNode(hname);
   }
 //re-create input and output transitions, but unfolded

 PutInputOutput(S, subpage, false); //connect only buffers = false
 
 //add msips
 AddMSIPS(s_classes,subpage);
 PutTPResults(S, cprime); //add result parameters for throughput
 //add result: probability that msip input places are sufficiently marked
 NodeClass *resultclass= cprime -> GetNodeClass("Result");
 for(unsigned int tt=0; tt<nb_edspntransitions; tt++)
   {
     if(!S->input[tt] && !S->output[tt])
       continue;
     //new result for probability
     Text resultname = Text("P") + 
       edspntransitions[tt]->Name().StripFirstPathPart('\\').ReplaceChar('\\','_');
     Position resultpos = Position(20,20);
     Node *result =
       new Node(resultclass, cprime,
		resultname.to_charPtr(), resultpos);
     Text attributes = "{ attributes = { def = \"P[";
     bool first = true;
     for (i=0; i<(unsigned int)s_classes.Length(); i++)
       {//for all classes
	 EClass_T *class_i = *s_classes[i];
	 NodeList_T *new_nodes  = class_i->new_nodes;
	 for(int l=0; l<new_nodes->Length(); l++)
	   { //for all MSIPS
	     EDSPNPlace *newnode= static_cast<EDSPNPlace*>(*(*new_nodes)[l]);   
	     if(newnode->GetPre()[tt]) 
	       {//msip is in preset of transition tt
		 if(!first) attributes+=" & ";
		 first = false;
		 attributes+=Text("#\\") + S->Path() + "\\" + (newnode->Name()).StripPath('\\');
		 attributes+=">0";
	       }
	   }//for all MSIPS
       }//for all classes
     if(first) //no msips in preset of transition
       attributes+="TRUE";
     attributes +="]\"; val = \"\";};}";
     istrstream is(attributes.to_charPtr(),100);
     is>>*result; 
   }
}
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//             FUNCTIONS FOR COMPUTING MSIPS
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// build the incidence matrix of class_i after class_i->list 
//   has been filled 
void MSIPDecomp::FillIncMatrix(EClass_T *class_i)
{
  unsigned int nb_places = 0;
  unsigned int nb_transitions = 0;
  NodeList_T *nodelist = class_i->list; //list of EDSPNNodes
  Inc_Matrix_T *matrix = class_i->matrix; //incidence matrix
  //count places and transitions in the nodelist
  //and store pointers to the nodes in the matrix structure
  //(to identify a node with its index)
 for (int i=0; i<nodelist->Length(); i++) 
    {//count places and transitions, store pointers
      EDSPNNode *node = *(*nodelist)[i];
      if (node->GetType()==E_PLACE)
	{
	  matrix->places.Set(nb_places, node);//store in place array
	  nb_places++;	
	}
      else
	{
	  matrix->transitions.Set(nb_transitions, node);
	  nb_transitions++;
	}
    }
 //set matrix size and allocate memory
 matrix->rows = nb_transitions;
 matrix->columns = nb_places;
 matrix->V = new int*[nb_transitions]; //values
 for (unsigned int row=0; row< nb_transitions; row++) 
   {
     matrix->V[row] = new int[nb_places];
   }
 //fill matrix
 unsigned int t_counter =0;
 for (unsigned int tt=0; tt< nb_edspntransitions; tt++) 
   {//for all transitions 
     EDSPNTransition *et= edspntransitions[tt];
     if(!nodelist->IsIn(et)) //is et in the class?
       continue;
     unsigned int p_counter =0;
     for (unsigned int pp=0; pp< nb_edspnplaces; pp++) 
       {//for all places
	 EDSPNPlace *ep= edspnplaces[pp];
	 if(!nodelist->IsIn(ep))
	   continue;
	 //incidence matrix = row tt of post vector (ep) - row tt of pre vector (ep)
	 matrix->V[t_counter][p_counter] = ep->GetPost()[tt] - ep->GetPre()[tt];
	 p_counter++;
       }
     t_counter++;
   }
}

// ----------------------------------------------------------------------------
// compute marking for structurally implicit place: set up an input file for
// lp_solve, let it run and read output file (value of objective function)  
unsigned int MSIPDecomp::GetMSIPMarking(int* postcol, int* precol,
					Text basefilename)
{
  bool first;
  Text lpfilename = basefilename + ".lp";
  ofstream lpfile(lpfilename.to_charPtr());
  //v = minimum y.m0 + MU
  //LINES: 1
  lpfile<<"min: ";
  first = true;
  for (unsigned int pp=0; pp<nb_edspnplaces; pp++) 
    {
      EDSPNPlace *ep = edspnplaces[pp];
      if(ep->GetMarking())
	{
	  if(!first)
	     lpfile<<" + ";
	  first = false;
	  lpfile<<ep->GetMarking()<<" * y"<<pp<<" ";
	}
    }
  lpfile<<" + MU ; "<<endl;
  //subject to...
  //y.C = C[p,.]
  //LINES: nb_edspntransitions
  first=true;
  for (unsigned int tt=0; tt<nb_edspntransitions; tt++) 
    {
      for (unsigned int pp=0; pp<nb_edspnplaces; pp++) 
	{
	  if(incidence_matrix[tt][pp])
	    {
	      if(!first)
		lpfile<<" + ";
	      first = false;
	      lpfile<<incidence_matrix[tt][pp]<<" * y"<<pp<<" ";
	    }
	}
      lpfile<<" = "<<postcol[tt] - precol [tt]<<" ; "<<endl;
    }
  //y.Pre[.,t] + MU >= Pre[p,t] f.a.t in p*
  //(scalar product y with column t of pre is equal to element t of the 
  // new places pre - vector. For all transition in the postset of p, 
  // all transitions t with Pre[p,t] >0)
  //LINES: number of transitions in postset
  first = true;
  for (unsigned int tt=0; tt<nb_edspntransitions; tt++) 
    {
      if(precol[tt] > 0)
	{
	  for (unsigned int pp=0; pp<nb_edspnplaces; pp++) 
	    {
	      if(pre_matrix[tt][pp])
		{
		  if(!first)
		    lpfile<<" + ";
		  first = false;
		  lpfile<<pre_matrix[tt][pp]<<" * y"<<pp<<" ";
		}
	    }
	  lpfile<<" + MU >= "<<precol[tt]<<" ;"<<endl;
	}
    }
  /*always assumed by lp_solve
  //y>=0
  //LINES: nb_edspnplaces
  for (unsigned int pp=0; pp<nb_edspnplaces; pp++) 
    lpfile<<"y"<<pp<<" >= 0 ; "<<endl;
  */
  lpfile<<"int MU;"<<endl; // MU is an integer (?)
  lpfile.close();
  Text progname = tnet_dir + "/Agnes/scripts/lp_solvescript";
  Text resultfilename = basefilename + ".out";
  Text command = progname + Text(" ") + lpfilename + Text("  ") + resultfilename;
  // pid = Exec(command.to_charPtr(), 0, 0);
  //int exitcode;
  //WaitForProcess(pid, exitcode);
  system(command.to_charPtr());
  
  ifstream resfile(resultfilename.to_charPtr());
  char buffer[10000];
  bool found = false;
  while(!resfile.eof())
    {
      resfile>>buffer;
      if (strstr(buffer,"function:"))
	{
	  found =true;
	  break;
	}
      if (strstr(buffer,"infeasible"))
	{
	  found =false;
	  break;
	}
    }
  if(!found)
    {
      cout<<"MSIPDecomp::GetMSIPMarking: lp_solve found no solution for file";
      cout<<lpfilename<<endl;
      return 0;
    }
  unsigned int marking;
  resfile>>marking;
  resfile.close();
  //remove files created
  Text delcommand = Text("'rm' -f ") + resultfilename;
  system(delcommand.to_charPtr());
  delcommand = Text("'rm' -f ") + lpfilename;
  system(delcommand.to_charPtr());
  return marking;
}

// ----------------------------------------------------------------------------
// for class_i, compute invariants and store in invmatrix.
void MSIPDecomp::GetInvariants(EClass_T *class_i, Inc_Matrix_T* &invmatrix,
			       Text basefilename)
{
  //write incidence matrix file (transpose: places x trans!)
  Inc_Matrix_T *matrix = class_i->matrix; //incidence matrix
  Text incfilename = basefilename + ".Minc";

  ofstream incfile(incfilename.to_charPtr());
  incfile<<matrix->columns<<" "<<matrix->rows<<endl;
  for (unsigned int col=0; col<matrix->columns ; col++)
    {
      for (unsigned int row=0; row< matrix->rows; row++) 
	incfile<<matrix->V[row][col]<<" ";
      incfile<<endl;
    }
  incfile.close();
  //call external program to compute invariants
  Text progname = tnet_dir + "/Shared/bin/callinvars ";
  Text resultfilename = basefilename + ".Mres";
  Text command = progname + incfilename + Text("  ") + resultfilename;
  pid = Exec(command.to_charPtr(), 0, 0);
  int exitcode;
  WaitForProcess(pid, exitcode);
  
  //read invariants
  ifstream invfile(resultfilename.to_charPtr());
  invmatrix = new Inc_Matrix_T;
  invfile>>invmatrix->rows;
  invfile>>invmatrix->columns;
  invmatrix->V = new int*[invmatrix->rows];
  for (unsigned int row=0; row<invmatrix->rows ; row++) 
    {
      invmatrix->V[row] = new int[invmatrix->columns];
      for (unsigned int col=0; col<invmatrix->columns; col++) 
	invfile>>invmatrix->V[row][col];
    }
  invfile.close();  
  //remove files created
  Text delcommand = Text("'rm' -f ") + incfilename;
  system(delcommand.to_charPtr());
  delcommand = Text("'rm' -f ") + resultfilename;
  system(delcommand.to_charPtr());
#ifdef DEBUG
  cout<<invmatrix->columns<<" invariants for "<<basefilename<<endl;
  for (unsigned int invcol = 0; invcol<invmatrix->columns; invcol++)
    {
      for (unsigned int pp=0; pp<invmatrix->rows ; pp++) 
	{//for all places in invariant
	  unsigned int placeindex = (*(matrix->places)[pp])->GetIndex(); //index translation
	  if(invmatrix->V[pp][invcol])
	    cout<<invmatrix->V[pp][invcol]<<"*"<<edspnplaces[placeindex]->Name()<<" ";
	}
      cout<<endl;
    }
#endif
}

// ----------------------------------------------------------------------------
//compute a marking structurally implicit place, using an invariant given by
// column invcol of a matrix of invariants invmatrix. 
// The new place has the incidence vector l = Invariant .NetIncidenceMatrix
// The marking of the new place is computed by a call to the external program 
// lpsolve
EDSPNNode *MSIPDecomp::ComputeMSIP(SubTrans *S, EClass_T *class_i, int i,
				   Inc_Matrix_T *invmatrix, unsigned int invcol)
{
  //incidence vector is assembled of a post and a pre vector
  int* postcol= new int[nb_edspntransitions];
  int* precol = new int[nb_edspntransitions];

  Inc_Matrix_T *matrix = class_i->matrix; //incidence matrix

  for (unsigned int tt=0; tt< nb_edspntransitions; tt++) 
    postcol[tt] = precol[tt] = 0;
  
  //perform vector-matrix product
  //incidence_matrix has size  nb_edspntransitions x nb_edspnplaces
  //invariant matrix has size places_in_class x nb_invariants
  for (unsigned int pp=0; pp<invmatrix->rows ; pp++) 
    {//for all places in invariant
      unsigned int placeindex = (*(matrix->places)[pp])->GetIndex(); //index translation
      for (unsigned int tt=0; tt< nb_edspntransitions; tt++) 
	{//for all trans	  
	  int addval= invmatrix->V[pp][invcol] * incidence_matrix[tt][placeindex];
	  if(addval>0)
	    postcol[tt]+=addval;
	  else
	    precol[tt]-=addval;
	}
    }
  
  //delete connections to interior transitions
  for(int tt=0; tt<(matrix->transitions).Length() ; tt++) 
    {//for all transitions in class
      unsigned int transindex = (*(matrix->transitions)[tt])->GetIndex();
      postcol[transindex]=precol[transindex] = 0;
    }
  
  Text basefilename = model_dir + Text("/")+ S->Path() + Float(i) + Text("_") + Float(invcol);
  unsigned int marking = GetMSIPMarking(postcol, precol, basefilename);

  //compute new places
  Text name = S->Path() + Text("CLASS") + Float(i) + Text("_") + Float(invcol);
  EDSPNPlace *newplace = new
    EDSPNPlace( name, 
		555, //nbcorr_HCPN, 
		555, //nb_color,
		false, //complement,
		"e", //tokenname,
		marking, //marking,
		postcol,
		precol,
		555 //index
		);

#ifdef DEBUG
  cout<<"Created Place "<<name.to_charPtr()<<"Marking: "<<marking<<endl;
  cout<<"Post Connections: ";
  for (unsigned int tt=0; tt< nb_edspntransitions; tt++) 
    if(postcol[tt]) cout<<edspntransitions[tt]->Name()<<" "; 
  cout<<endl<<"Pre Connections: ";
  for (unsigned int tt=0; tt< nb_edspntransitions; tt++) 
    if(precol[tt]) cout<<edspntransitions[tt]->Name()<<" "; 
  cout<<endl;
#endif 
  return newplace;
}

// ----------------------------------------------------------------------------
//compute structurally implicit places for Subtransition S
//with the corresponding equivalence classes S s_classes.
//Every class class_i in s_classes contains a list of nodes that build 
//the class. This function fills the incidence matrix class_i->matrix
//and computes new Structurally Implicit Places class_i->new_nodes  
void MSIPDecomp::ComputeSIPS(SubTrans *S, EClasslist_T *s_classes)
{
  for (int i=0; i<s_classes->Length(); i++) 
    {//for all classes
      EClass_T *class_i = *(*s_classes)[i];
      NodeList_T *nodelist = class_i->list; //list of EDSPNNodes
      Inc_Matrix_T *invmatrix;
      //fill the incidence matrix of class i
      FillIncMatrix(class_i);
      if(nodelist->Length() == 1) //special case: single place
	{	 
	  //create a false "invariant" that will produce a MSIP
	  //that is a copy of the single place in the class
	  invmatrix = new Inc_Matrix_T;
	  invmatrix->rows = 1;
	  invmatrix->columns = 1;
	  invmatrix->V = new int*[invmatrix->rows];
	  for (unsigned int pp=0; pp<invmatrix->rows; pp++) 
	    {
	      invmatrix->V[pp] = new int[invmatrix->columns];
	      invmatrix->V[pp][0] = 1;
	    }
	}
      else
	{
	  //calculate invariants for place i
	  Text basefilename = model_dir + Text("/")+ S->Path() +Float(i);
	  GetInvariants(class_i, invmatrix, basefilename);
	}
      for (unsigned int invcol=0; invcol<invmatrix->columns; invcol++) 
	{ //for all invariants
	  EDSPNNode *newplace = ComputeMSIP(S, class_i, i, invmatrix, invcol);
	  class_i->new_nodes->Append(newplace);		    
	}
    } 
}
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//             FUNCTIONS FOR EQUIVALENCE CLASS CREATION
// ----------------------------------------------------------------------------
//find places that are connected to transition trans_nb. add places to 
//equivalence class curr_class
void MSIPDecomp::FindPlace(unsigned int trans_nb, 
			   EClass_T *curr_class,
			   SubTrans* S)
{
  unsigned int j;
  bool found = false;
  EDSPNTransition *et;
  int *postrow, *prerow;

  et = edspntransitions[trans_nb];
  postrow = et->GetPost();
  prerow = et->GetPre();

  //find places connected to transition
  for(j=0; j< nb_edspnplaces; j++) 
    {
      if(!S->p_onpage[j]) continue; //not on page
      if(postrow[j]>0 || prerow[j] >0 )  
	{
	  EDSPNPlace *ep = edspnplaces[j];
	  if (curr_class->list->IsIn(ep)) //do not run cyclic
	    continue;
	  else
	    {
	      found = true;
	      curr_class->list->Append(ep); //append place
	      FindTransition(j, curr_class, S); //find transition 
	    }
	}
    }
}

// ----------------------------------------------------------------------------
//find transitions that are connected to place place_nb. add transitions to 
//equivalence class curr_class
void MSIPDecomp::FindTransition(unsigned int place_nb,
				EClass_T *curr_class,
				SubTrans* S)
{
  unsigned int j;
  bool found = false;
  EDSPNPlace *ep;
  int  *precol, *postcol; 

  ep = edspnplaces[place_nb];
  precol = ep->GetPre();
  postcol = ep->GetPost();

  //find transition connected to place
  for(j=0; j< nb_edspntransitions; j++) 
    {
      if (!S->t_onpage[j]) continue; //not on page
      if(precol[j]>0 || postcol[j]> 0) 
	{
	  EDSPNTransition *et = edspntransitions[j];
	  if(S->input[j] || S->output[j]) //only internal transitions
	    continue;
	  if (curr_class->list->IsIn(et)) //do not run cyclic
	    continue;
	  else
	    {
	      found = true;
	      curr_class->list->Append(et); //append transition
	      FindPlace(j, curr_class, S); //find place 
	    }
	}
    }
}

// ----------------------------------------------------------------------------
//test if node is in equivalence classes s_classes
bool MSIPDecomp::IsInClasses(EClasslist_T &s_classes, EDSPNNode* node)
{
  int i,j;
 
  for (i=0; i<s_classes.Length(); i++) //for all classes
    {
      EClass_T *class_i = *s_classes[i];
      NodeList_T *nodelist = class_i->list;
      for (j=0; j<nodelist->Length(); j++) //elements of class_i
	{
	  if(*(*nodelist)[j] == node) //found element
	    return true;
	}
    }
    return false;
}

// ----------------------------------------------------------------------------
//debugging only: print an eq class
void MSIPDecomp::PrintClass(EClass_T *eclass)
{
  int j;
  NodeList_T *nodelist = eclass->list;
  for (j=0; j<nodelist->Length(); j++) //elements of eclass_i
    {
      EDSPNNode *en = *(*nodelist)[j];
      cout<<" "<<en->Name();
      if(j!=nodelist->Length()-1)
	cout<<", ";
    }
}
// ----------------------------------------------------------------------------
//debugging only: print all eq classes
void MSIPDecomp::PrintClasses(EClasslist_T &s_classes)
{
  int i;
  for (i=0; i<s_classes.Length(); i++) //for all classes
    {
      cout<<"CLASS "<<i;
       EClass_T *class_i = *s_classes[i];
       PrintClass(class_i);
       cout<<endl;
    }
}
// ----------------------------------------------------------------------------
//calculate all equivalence classes for subtransition S
void MSIPDecomp::CalcClasses(SubTrans* S, EClasslist_T &s_classes)
{
  unsigned int j;
#ifdef DEBUG
  cout<<"###############################"<<endl;
  cout<<"Calculating EqClasses for"<<endl;
  cout<<"Subtrans "<<S->Path()<<endl;
  cout<<"###############################"<<endl;
#endif

  ClassifyNodes(S);//build input, interior, input, output arrays

  for(j=0; j< nb_edspnplaces; j++)
	{
	  if((S->p_onpage[j])) //is on page
	    { 
	      EDSPNPlace *ep = edspnplaces[j];
	      if(IsInClasses(s_classes, ep)) //is already in an equivalence class
		continue;
#ifdef EQCLASSDEBUG
	    cout<<"Calculating eq class starting with "<<ep->Name()<<endl;
#endif
	    EClass_T *new_class = new EClass_T();
	    new_class->list->Append(ep);
	    FindTransition(j, new_class, S); //recursively build eq class
	    s_classes.Append(new_class);
	    }
	}
#ifdef DEBUG
 cout<<"Classes for "<<S->Path()<<endl;
 PrintClasses(s_classes);
#endif
}
