// ----------------------------------------------------------------------------
// Importe
// ----------------------------------------------------------------------------
#include <sys/time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <iostream.h>
#include <builtin.h>
#include <typeinfo>

#include "HCPN2EDSPN.h"
#include "PathDecomp.h"

#define DEBUG
#define PATHDEBUG

MODULE_NOTIFY(PathDecomp)
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//             USER INTERFACE FUNCTIONS      
// ----------------------------------------------------------------------------
//called from InitDecomposition. tries to find directory with old decomposition
void PathDecomp::StartDecomposition()
{
  Text bsname = dirname+ "/" + purename+"BS.net";
  Text unfold_filename = netname + ".unfold";

  //look for old decomposed net to use
   
  ifstream bs_file(bsname.to_charPtr());
  ifstream u_file(unfold_filename.to_charPtr());
  
  if(bs_file && u_file)
    {//decomposition exists
      Text BoxContents;
      BoxContents+= (Text("There is an old decomposition of this net. \nUse it?")).to_charPtr();
      bs_file.close(); u_file.close();
      OpenMsgDialog(*this->ed, "Decomposition_Question", 
		    BoxContents.to_charPtr(), false, "\"Yes, Load\" \"No, Recompute\"");
    }
  else
    ComputeDecomposition();
}
// ----------------------------------------------------------------------------
void PathDecomp::EndDecomposition()
{
  net.ChangeEditor(*ed,net.Prime()); //switch to top page
  net.UpdateButtonStates(*ed, *(net.Prime()));
}

// ----------------------------------------------------------------------------
//when decomposition button is clicked
void PathDecomp::OpenOptionsMenu()
{
  Text DialogContense;
  
  DialogContense+= Text("[enum ( \"Display Aggregated\" Aggregated, ");
  DialogContense+= Text("\"Display Extended\" Extended ) ");
  DialogContense+= Text("\"Doubleclick on subtransition: \"displaymode= ");
  DialogContense+= displaymode + ";";
  DialogContense+= Text("]");
  CreateDialog(DialogContense.to_charPtr());

  OpenDialog(*this->ed, "DecompositionOptions", 
	     "Close");
}
// ----------------------------------------------------------------------------
//when any dialog button is clicked
void PathDecomp::OnDialogButton(Ident& button)
{// Wird bei Mausklick auf die Dialog-Buttons aufgerufen

  Struct* params= GetParams();

  if (button=="Yes, Load")
    { //load decomposition
      CloseDialog();
      DestroyDialog();
      LoadDecomposition();
    }
  if (button=="No, Recompute")
    { //compute decomposition
      CloseDialog();
      DestroyDialog();
      ComputeDecomposition();
    }
  if ((button=="Exit Decomp."))
    {// exit
      DeInitDecomposition();
      CloseDialog();
      DestroyDialog();
    }
  if(button=="Close")
    { //close decomposition window
      displaymode = (*params->GetMember("displaymode")).to_charPtr();
      CloseDialog();
      DestroyDialog();
    }
  if((button=="Ok") || (button=="Cancel") || (button=="Exit"))
    {//close other dialog boxes
      CloseDialog();
      DestroyDialog();
    }
}
// ----------------------------------------------------------------------------
//load existing decomposition
void PathDecomp::LoadDecomposition()
{
  Text bsname = dirname+ "/" + purename+"BS.net";
  Text unfold_filename = netname + ".unfold";  
  //read net and build sub trans hierarchy
  top = new SubTrans(p, 0, 0);
  //load unfold file
  LoadUnfoldFile(unfold_filename.to_charPtr()); //init HCPN2EDSPN
  FillStructures();
  //load basic skeleton
  BasicSkeleton=0;
  BasicSkeleton=BasicSkeleton->Load(bsname.to_charPtr());
  if(!BasicSkeleton)
    {
      errortext += Text("Can't load ") + bsname;
      OpenMsgDialog(*this->ed, "Decomposition_Error", 
		    errortext.to_charPtr(), false, "\"Exit\"");
      DeInitDecomposition();
      return;
    }
  //load low level systems
  LinkedList<SubTrans*> childs = top->Children();
  OpaqueIterator<SubTrans*> i2(childs);
  for (SubTrans** o=i2; o; o=++i2) 
    {
      Text lowname = dirname + "/" + purename+ (*o)->Path() + ".net";
      Net *n=0;
      n=n->Load(lowname.to_charPtr());
       if(!n)
	 {
	   errortext += Text("Can't load ") + lowname;
	   OpenMsgDialog(*this->ed, "Decomposition_Error", 
			 errortext.to_charPtr(), false, "\"Exit\"");
	   DeInitDecomposition();
	   return;
	 }
      LowLevelSystems.Set((*o), n); //store net in an array
      ClassifyNodes((*o));//build input, interior, input, output arrays
    } 
}
// ----------------------------------------------------------------------------
//compute decomposition: check for subtranitions, call uncolor,
// initialize HCPN2EDSPN,  create paths,
//create directory, low level systems and basic skeleton 
void PathDecomp::ComputeDecomposition()
{
  Text bsname = dirname + "/" + purename+"BS.net";
  Text DialogContense;
  Text stdout_filename =  netname + OUTFILENAME;
  Text unfold_filename = netname + ".unfold";

  ofstream stdout_file(stdout_filename.to_charPtr());
      
  CreateLogwin(stdout_filename.to_charPtr());
  OpenLogwin(*ed, "Decompose...", "Ok");
  //check for subtransitions
  stdout_file<<"Looking for subtransitions on the top page...";
  top = new SubTrans(p, 0, 0);
  if (top->Children().Length()==0)
    {
      stdout_file<<"No subtransitions!"<<endl;
      DeInitDecomposition();
      return;
    }
  stdout_file<<top->Children().Length()<<" subs found."<<endl;
  //call uncolor
  stdout_file<<"Creating table file..."<<flush;
  if (!CreateTableFile(netname, net))
    {
      stdout_file<<"\t Unable to create table file!"<<endl;
      DeInitDecomposition();
      return;
    }
  stdout_file<<"done"<<endl;

  stdout_file<<"Calling uncolor to create .unfold-file...";
  if (!CallUncolor(netname))
    {
      stdout_file<<"\t Error when calling uncolor!"<<endl;
      return;
    }    
  stdout_file<<"done"<<endl;

  stdout_file<<"reading .unfold-file...";
  LoadUnfoldFile(unfold_filename.to_charPtr()); //init HCPN2EDSPN
  FillStructures(); //add position, scale... information
  stdout_file<<"done"<<endl;
  //calculate paths
  stdout_file<<"calculating paths...";
  LinkedList<SubTrans*> childs = top->Children();
  OpaqueIterator<SubTrans*> i1(childs);
  paths = new Pathlist_T*[childs.Length()];
  int i=0;
  for (SubTrans** o=i1; o; o=++i1) 
    {
      Pathlist_T *s_path = new Pathlist_T(); //one pathlist for every subpage
      CalcPaths(*o, *s_path);
      paths[i] = s_path;
      i++;
    }
  stdout_file<<"done"<<endl;
  //make directory
  Text makedir = "mkdir ";
  makedir= makedir + dirname;
  system(makedir.to_charPtr());

  stdout_file<<makedir.to_charPtr()<<endl;
  stdout_file<<"creating basicSkeleton/ low level systems...";
  //create nets
  OpaqueIterator<SubTrans*> i2(childs);
  for (SubTrans** o=i2; o; o=++i2) 
    {
      Text lowname = dirname + "/" + purename+ (*o)->Path() + ".net";
      Net *n=CreateLowLevelSystem(lowname, (*o)->Path());
      LowLevelSystems.Set((*o), n); //store net in an array
    } 
 
  BasicSkeleton = CreateBasicSkeleton(bsname);
  
  stdout_file<<"done"<<endl;
  stdout_file<<"Click on the substitution transitions to see"<<endl;
  stdout_file<<"the aggregated subnets. You may move objects"<<endl;
  stdout_file<<"and save the modified net. Other actions are"<<endl;
  stdout_file<<"not allowed while decomposition is active."<<endl;
}
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//        FUNCTIONS FOR ANALYSIS
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

//not yet implemented

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//        FUNCTIONS FOR CREATING BASIC SKELETON AND LOW LEVEL SYSTEMS
// ----------------------------------------------------------------------------
//create low level system for subtransition path, name of netfile = lowname
Net *PathDecomp::CreateLowLevelSystem(Text lowname, Text path)
{
  Net *cnet = p->GetNet()->Copy();
  cnet->SetFileName(lowname.to_charPtr());
  Page *cprime = cnet->Prime();
  
  LinkedList<SubTrans*> childs = top->Children();
  OpaqueIterator<SubTrans*> i1(childs);
  int i=0;
  for (SubTrans** o=i1; o; o=++i1, i++) 
    {
      if ((*o)->Path() == path) 
	CreateExtension(*o, *paths[i], cprime);   //subsystem s_i is extended ...
      else
	CreateAggregation(*o, *paths[i], cprime); //while the others are aggregated
    } 
  cnet->Save();
  return cnet;
} 
// ----------------------------------------------------------------------------
// create basic skeleton net in file named bsname
Net *PathDecomp::CreateBasicSkeleton(Text bsname)
{
  Net *cnet = p->GetNet()->Copy();
  cnet->SetFileName(bsname.to_charPtr());
  Page *cprime = cnet->Prime();
  
  LinkedList<SubTrans*> childs = top->Children();
  OpaqueIterator<SubTrans*> i1(childs);
  int i=0;
  for (SubTrans** o=i1; o; o=++i1) 
    {
      //aggregate all subsystems
      CreateAggregation(*o, *paths[i], cprime);
      i++;
    } 
  cnet->Save();
  return cnet;
}

// ----------------------------------------------------------------------------
// add places for s_path to subpage
void PathDecomp::AddNewPlaces(SubTrans *S, Pathlist_T &spath, Page *subpage)
{ //create msip places and respective bindings 
  NodeClass *eplaceclass = subpage -> GetNodeClass("EPlace");
  NodeClass *resultclass = subpage -> GetNodeClass("Result");
  for (unsigned int i=0; i<(unsigned int)spath.Length(); i++) 
    { //for all paths
      Path_T *path_i = *spath[i];     
      EDSPNNode* input= *path_i->First(); //input transition
      EDSPNNode* output=*path_i->Last(); //output transition
      
      Text placename = S->Path() + Text("_PATH") + Float(i); //name for new place
      Position newpos = 
	Position ( (input->position.x  + output->position.x) /2,
		   (input->position.y  + output->position.y) /2);
      //new place 
      Node* placenode =  
	new Node(eplaceclass,  subpage,
		 placename.to_charPtr(), newpos,
		 input->orientation);
      Text attributes= "{ attributes = { marking = \"";
      attributes= attributes +  Float(PathMarking(path_i));
      attributes += "\";};}";
      
      istrstream is(attributes.to_charPtr(),100);
      is>>*placenode;  
      //new result
      Text resultname = placename+"Res";
      Position resultpos = Position(20,20);
      Node* resultnode =
	new Node(resultclass, subpage->Prime(),
		 resultname.to_charPtr(), resultpos);

      //attributes for result 
      attributes = "{ attributes = { def = \"E[#\\";
      attributes += S->Path() + "\\" + placename;
      attributes +="]\"; val = \"\";};}";
      istrstream is2(attributes.to_charPtr(),100);
      is2>>*resultnode; 

      //draw connector from input transition to new place, add binding
      DrawConnection(input->Name().StripFirstPathPart('\\'),
		     placename, subpage);    
      AddConnection(input->hcpnnode, 0, Text("e"), 0, placenode);
      //draw connector from new place to output transition, add binding
      DrawConnection(placename,
		     output->Name().StripFirstPathPart('\\'), subpage);  
      AddConnection(output->hcpnnode, 0, Text("e"), placenode, 0);
    } 
}
// ----------------------------------------------------------------------------
// put unfolded input and output transitions on subpage.
// foraggregation triggers if the transition should be connected 
// to buffers only (aggregation) or to all places (extension)
void PathDecomp::PutInputOutput(SubTrans *S, Page *subpage, 
				Pathlist_T &spath, bool foraggregation)
{
  NodeClass *ttransclass =  subpage-> GetNodeClass("TimedTrans");
  NodeClass *itransclass = subpage -> GetNodeClass("ImmTrans");

  for(unsigned int i=0; i< nb_edspntransitions; i++) {
    if(!(S->output[i]) && !(S->input[i])) //neither output nor input
      continue;
    
    EDSPNTransition *et = edspntransitions[i];
    int bindnr = et->GetBindNumber();
    Text hcpnname = hcpntransitions[(et->GetHCPNNumber())]->Name();
    Node* hnode = p->FindNode(hcpnname); 
    Text subsubname = (hcpnname.StripFirstPathPart('\\')).StripLastPathPart('\\');
    
    Text trelpath = (et->Name()).StripFirstPathPart('\\');
    Text tpurename = trelpath.StripPath('\\'); 
    Position newpos = hnode->Mid(); //position: same as HCPN-trans
    NodeClass *nclass;
    
    //set transition type class
    if(foraggregation)
      {
	if(!IsIOofPath(spath, (EDSPNNode *)et))
	  {
	    if(!(S->output[i] && S->input[i])) //only IO transitions accepted
	      continue;
	    nclass = hnode->Class(); //for transitions that are in no path, copy nodeclass
	  }
	else if(S->input[i])
	  nclass=itransclass; //immediate transitions for input trans
	else
	  nclass=ttransclass; //timed transitions for output trans
      }
    else
      nclass = hnode->Class();
    

    //create transition node
    Node* trans =  new Node(nclass, subpage->FindPage(subsubname), 
			    tpurename.to_charPtr(),
			    newpos, et->orientation);
    et->hcpnnode = trans; //store pointer to node 
    
    //draw connections from transition to places 
    for(unsigned int j=0; j< nb_edspnplaces; j++)
      {
	EDSPNPlace *ep = edspnplaces[j];
	Text hname = hcpnplaces[(ep->GetHCPNNumber())]->Name().StripFirstPathPart('\\');
	if(foraggregation && S->p_onpage[j])
	  continue;
	if(et->GetPost()[j]>0) //output trans
	  {
	    //draw connector from place to transition
	    DrawConnection(trelpath, hname, subpage);
	    //create a binding
	    CreateConnection(trans, 0, hnode, bindnr, //create binding 0 
			     ep->TokenName(), 0, subpage->FindNode(hname));
	  }
	if(et->GetPre()[j]>0) //input trans 
	  {
	    //draw connector from transition to place
	    DrawConnection(hname, trelpath, subpage);
	    //create a binding
	    CreateConnection(trans, 0, hnode, bindnr, 
			     ep->TokenName(), subpage->FindNode(hname), 0);
	  }	  
      }
  }
}  

// ----------------------------------------------------------------------------
//remove original subsystem for subtransition S and create a 
//new place for every path in pathlist spath.
//add a result parameter for every new place
void PathDecomp::CreateAggregation(SubTrans* S, Pathlist_T &spath, Page *cprime)
{
  unsigned int i;
  Node* subnode = cprime->FindNode(S->Path());
  Page* subpage = subnode->SubPage(); 
 
  //delete all objects on the subpage
  //except buffer places, subtransitions
  //delete transitions
 for(i=0; i< nb_edspntransitions; i++) 
   {
     EDSPNTransition *et = edspntransitions[i];
     if(!(S->t_onpage[i])) continue;
     Text hname = hcpntransitions[(et->GetHCPNNumber())]->Name().StripFirstPathPart('\\');
     delete subpage->FindNode(hname);
   }
 //delete places
 for(i=0; i< nb_edspnplaces; i++) 
   {
     EDSPNPlace *ep = edspnplaces[i];
     if(!S->p_onpage[i]) continue;
     Text hname = hcpnplaces[(ep->GetHCPNNumber())]->Name().StripFirstPathPart('\\');
     (subpage->FindNode(hname))->RemoveNodeAndAliases();
   }
 //re-create input and output transitions, but unfolded
 PutInputOutput(S, subpage, spath, true); //aggregation =true 
 AddNewPlaces(S, spath,subpage);
 //PutTPResults(S, cprime);
}
// ----------------------------------------------------------------------------
//for subtransition S create a 
//new place for every path in pathlist spath.
//add a result parameter for every new place
void PathDecomp::CreateExtension(SubTrans* S, Pathlist_T &spath, Page *cprime)
{
  unsigned int i;
  Node* subnode = cprime->FindNode(S->Path());
  Page* subpage = subnode->SubPage(); 
 //delete input and output transitions
 for(i=0; i< nb_edspntransitions; i++) 
   {
     EDSPNTransition *et = edspntransitions[i];
     if(!(S->output[i]) && !(S->input[i])) //neither output nor input
       continue;
     Text hname = hcpntransitions[(et->GetHCPNNumber())]->Name().StripFirstPathPart('\\');
     delete subpage->FindNode(hname);
   }
 //re-create input and output transitions, but unfolded
 PutInputOutput(S, subpage, spath, false); //aggregation = false

 //add msips
 AddNewPlaces(S, spath,subpage);
}
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
//             FUNCTIONS FOR PATHS CREATION
// ----------------------------------------------------------------------------
//find places that are in postset of transition trans_nb. add places to curr_path
void PathDecomp::FindPlace(unsigned int trans_nb, Path_T *curr_path,
			   Pathlist_T &spath,
			   SubTrans* S)
{
  unsigned int j;
  bool found = false;
  EDSPNTransition *et;
  int* postrow;

  et = edspntransitions[trans_nb];
  postrow = et->GetPost();

  //find places connected to out-arc of transition
  for(j=0; j< nb_edspnplaces; j++) 
    {
      if(!S->p_onpage[j]) continue; //not on page
      if(postrow[j]>0) 
	{
	  EDSPNPlace *ep = edspnplaces[j];
	  if (curr_path->IsIn(ep)) //do not run cyclic
	    continue;
	  else
	    {
	      found = true;
	      Path_T *new_path
		= new Path_T(*curr_path); //copy current path...
	      new_path->Append(ep); //... and append place
	      FindTransition(j, new_path, spath, S); //find transition 
	    }
	}
    }
}
// ----------------------------------------------------------------------------
//find transitions that are in postset of place place_nb. add trans. to curr_path
void PathDecomp::FindTransition(unsigned int place_nb,
				Path_T *curr_path,
				Pathlist_T &spath,
				SubTrans* S)
{
  unsigned int j;
  bool found = false;
  EDSPNPlace *ep;
  int* precol; 

  ep = edspnplaces[place_nb];
  precol = ep->GetPre();

  //find transition connected to out arc of place
  for(j=0; j< nb_edspntransitions; j++) 
    {
      if (!S->t_onpage[j]) continue; //not on page
      if(precol[j]>0) 
	{
	  EDSPNTransition *et = edspntransitions[j];
	  if (curr_path->IsIn(et)) //do not run cyclic
	    continue;
	  else
	    {
	      found = true;
	      Path_T *new_path 
		= new Path_T(*curr_path); //copy current path...
	      new_path->Append(et); //... and append transition
	      if(S->output[j]) //output transition of subpage
		CheckAndAppend(new_path, spath);
	      
	      FindPlace(j, new_path, spath, S); //find place 
	    }
	}
    }
}
// ----------------------------------------------------------------------------
//check if a check_path is a longest path, if yes append to list spath
void PathDecomp::CheckAndAppend(Path_T *check_path, Pathlist_T &spath)
{
  if (LongestPath(check_path, spath))
    {
      spath.Append(check_path);
#ifdef PATHDEBUG
      cout<<spath.Length();
      PrintPath(check_path); cout<<endl;
#endif
    }
}
// ----------------------------------------------------------------------------
// sum up markings of all places in path 
unsigned int PathDecomp::PathMarking(Path_T *path)
{
  int j;
  unsigned int marking = 0;
  for (j=0; j<path->Length(); j++) //elements of path
    { 
	  EDSPNPlace *ep = static_cast<EDSPNPlace*> (*(*path)[j]); //downcast
	  marking += ep->GetMarking();
    }
  return marking;
}
// ----------------------------------------------------------------------------
// is node input / output trans. of any path in spath?
bool PathDecomp::IsIOofPath(Pathlist_T &spath, EDSPNNode* node)
{
  int i;
  for (i=0; i<spath.Length(); i++) //for all paths
    {
      Path_T *path_i = *spath[i];
      if(*(*path_i).First() == node)
	return true;
      if(*(*path_i).Last() == node)
	return true;
    }
    return false;
}
// ----------------------------------------------------------------------------
//is node in any path of list spath?
bool PathDecomp::IsInPaths(Pathlist_T &spath, EDSPNNode* node)
{
  int i,j;
  for (i=0; i<spath.Length(); i++) //for all paths
    {
      Path_T *path_i = *spath[i];

      for (j=0; j<path_i->Length(); j++) //elements of path_i
	{
	  if(*(*path_i)[j] == node) //found element
	    return true;
	}
    }
    return false;
}
// ----------------------------------------------------------------------------
// are the input output transitions of check_path unique? (no other
// path in spath having the same io transitions)
bool PathDecomp::UniquePath(Path_T *check_path, Pathlist_T &spath)
{
  int i;
  bool unique = true;
  int pindex = spath.Index(check_path); /*index of check_path in paths
					  or -1 if not contained*/
#ifdef DEBUG
  cout<<"Path goes from " << (*(*check_path).First())->Name()
      <<" to "<<(*(*check_path).Last())->Name()<<endl;
#endif

  for (i=0; i<spath.Length(); i++) //for all paths
    {
      if (i==pindex) 
	continue;
      Path_T *path_i = *spath[i];
      
#ifdef DEBUG
      cout<<"Compare with path from"<<(*(*path_i).First())->Name()
	  <<" to " <<(*(*path_i).Last())->Name()<<endl;
#endif
      if(*(*check_path).First() == *(*path_i).First()
	 && *(*check_path).Last() == *(*path_i).Last())
	{
	  cout<<"Path removed: equal input and output trans. "<<endl;
	  unique=false;
	  break;
	}   
    }
  if (unique)
    return true;
  else
    return false;
}
// ----------------------------------------------------------------------------
// is check_path a longest path? (not contained in any other path of spath)
bool PathDecomp::LongestPath(Path_T *check_path, Pathlist_T &spath)
{
  int i,j,k;
  bool contained = false;
  int pindex = spath.Index(check_path); /*index of check_path in paths
					  or -1 if not contained*/
  for (i=0; i<spath.Length(); i++) //for all paths
    {
      if (i==pindex) 
	continue;
      Path_T *path_i = *spath[i];

      bool first_match = true;
      for (j=0; j<path_i->Length(); j++) //elements of path_i
	{
	  if(path_i->Length() - j < check_path->Length()) //lengths dont match
	    {
	      first_match = false;
	      break;
	    }
	  if(*(*path_i)[j] == *(*check_path)[0]) //found first common element
	    break;
	}
      if(!first_match)
	continue;
      for (k = 0; k<check_path->Length();k++, j++) //elements of path_i
	{
	  if(*(*path_i)[j] != *(*check_path)[k])
	    break;
	}
      if(k==check_path->Length())
	{
	  contained = true;
	  break;
	}
    }

  if (contained)
    return false;
  else
    return true;
}
// ----------------------------------------------------------------------------
//debugging only:print a path
void PathDecomp::PrintPath(Path_T *path)
{
  int j;
  for (j=0; j<path->Length(); j++) //elements of path_i
    {
      EDSPNNode *en = *(*path)[j];
      cout<<" "<<en->Name();
      if(j!=path->Length()-1)
	cout<<"->";
    }
}
// ----------------------------------------------------------------------------
//debugging only:print all paths
void PathDecomp::PrintPaths(Pathlist_T &spath)
{
  int i;
  for (i=0; i<spath.Length(); i++) //for all paths
    {
      cout<<"PATH "<<i;
       Path_T *path_i = *spath[i];
       PrintPath(path_i);
       if(!LongestPath(path_i, spath))
	 cout<<"IS NOT A LONGEST PATH";
       cout<<endl;
    }
}
// ----------------------------------------------------------------------------
//calculate all paths for subtransition S
void PathDecomp::CalcPaths(SubTrans* S, Pathlist_T &spath)
{
  unsigned int i;
 
#ifdef DEBUG
  cout<<"###############################"<<endl;
  cout<<"Calculating Paths for"<<endl;
  cout<<"Subtrans "<<S->Path()<<endl;
  cout<<"###############################"<<endl;
#endif

  ClassifyNodes(S); //build input, interior, input, output arrays
 
 for(i=0; i< nb_edspntransitions; i++)
      {
	EDSPNTransition *et = edspntransitions[i];
	if(S->output[i] && S->input[i] && !S->interior[i]) continue; // no IO transitions
	if(S->input[i]) //transition is input transition only
	  {
#ifdef PATHDEBUG
	    cout<<"Calculating paths starting with "<<et->Name()<<endl;
#endif
	    Path_T *new_path = new Path_T();
	    new_path->Append(et);
	    FindPlace(i, new_path, spath, S); //Find places conn. to out arc
	  }
      }
#ifdef DEBUG
 cout<<"Paths for "<<S->Path()<<endl;
 PrintPaths(spath);
#endif

 //remove longest paths and identical paths
 for (int i=0; i<spath.Length(); i++) //for all paths
   {
     Path_T *path_i = *spath[i];
     if(!LongestPath(path_i, spath))
       { spath.Remove(path_i); i--;}
     /*
       //remove paths with identical input and output transitions
     else if(!UniquePath(path_i, spath))
       { spath.Remove(path_i); i--;}
     */
   }  
}
