#ifndef PATHDECOMP_H
#define PATHDECOMP_H

#include "decomposition.h"
// ----------------------------------------------------------------------------
// defines
// ----------------------------------------------------------------------------
#define NETFILENAME   ".net"
#define TABLEFILENAME ".table"
#define OUTFILENAME   ".decomp"
#define DIREXTENSION "DIRP"
typedef LinkedList<EDSPNNode*> Path_T; 
typedef LinkedList<Path_T*> Pathlist_T;
// ----------------------------------------------------------------------------
// class declaration
// pathwise decomposition
// ----------------------------------------------------------------------------
class PathDecomp: virtual public Decomposition
{ 
protected:
  void OnDialogButton(Ident& button);

  Pathlist_T **paths; /*for each subtransition
					       a linked list of paths,
					       each of which is a
					       linked list of EDSPNNodes
					     */
  
  void CalcPaths(SubTrans* S, Pathlist_T &spath);
  void PrintPath(Path_T *path);
  void PrintPaths(Pathlist_T &spath);
  void CheckAndAppend(Path_T *check_path,
		      Pathlist_T &spath);

  void FindPlace(unsigned int trans_nb, Path_T *curr_path,
		 Pathlist_T &spath,
		 SubTrans* S);
  void FindTransition(unsigned int place_nb, Path_T *curr_path,
		      Pathlist_T &spath,
		      SubTrans* S);
  bool LongestPath(Path_T *check_path,
		  Pathlist_T &spath);
  bool UniquePath(Path_T *check_path,
		 Pathlist_T &spath);

  Net *CreateBasicSkeleton(Text bsname);
  Net *CreateLowLevelSystem(Text lowname, Text path);
  void CreateAggregation(SubTrans* S, Pathlist_T &spath,  Page *cprime);
  void CreateExtension(SubTrans* S, Pathlist_T &spath, Page *cprime);  
  void AddNewPlaces(SubTrans *S, Pathlist_T &spath, Page *subpage);
  void PutInputOutput(SubTrans *S, Page *subpage, 
		      Pathlist_T &spath, bool foraggregation);


  unsigned int PathMarking(Path_T *path);
  bool IsInPaths(Pathlist_T &spath, EDSPNNode* node);
  bool IsIOofPath(Pathlist_T &spath, EDSPNNode* node);

  void  OpenOptionsMenu();
 
  void ComputeDecomposition();
  void LoadDecomposition();

  void StartDecomposition();
  void EndDecomposition();
 public:
 PathDecomp(Net& n):
    Decomposition(n),
    CommModule(n,"PathDecomp")
    {
      dirname = netname + DIREXTENSION;
    }
  
  ~PathDecomp() {}
  bool Init() {return (net.Prime()->Class()->Name() == "HCPN");};

};

#endif
