#include "decomposition.h"

Decomposition::Decomposition(Net& n): 
  CommModule(n,"Decomposition"), 
  Wrapper<Decomposition,bool>(this),
  CallBackID(0), 
  ButtonWidget(0),
  MenuWidget(0),
  SubWidget(0),
  ToggleWidget(0),
  DecompositionEnable(false), 
  DisplayingSubPage(0),
  pid(0),
  top(0),
  displaymode( "Aggregated"),
  BasicSkeleton(0)
{
  netname = Text(net.FileName()) - ".net";
  purename = netname.StripPath('/');
}
// ----------------------------------------------------------------------------
Decomposition::~Decomposition() 
{
  //delete menu-icon
  if (ButtonWidget) this->ed->DestroyEntry(ButtonWidget, CallBackID);
}
// ----------------------------------------------------------------------------
// open is called every time a hcpn net is opened
bool Decomposition::Open(Editor& e, Page& page)
{
  this->p= &page;
  this->ed= &e;
 
  WrapperID dummy;

  //menu entry named "Decomposition" in menu "Modules"
  MenuWidget = XtNameToWidget(this->ed->Mods(), "Decomposition");
  if (!MenuWidget)
    {
      //menu entry "Decomposition" opens a submenu 
      MenuWidget=this->ed->CreateEntry("Decomposition",xmCascadeButtonWidgetClass,this->ed->Mods(),NULL,dummy);
      SubWidget = ed->CreatePulldown("DecompositionSubMenu", MenuWidget);
    }
  else
    XtVaGetValues(MenuWidget, XmNsubMenuId, &SubWidget, NULL);
  
  //create a toggle button in the submenu, named Name() (e.g. "MSIPDecomp")
  ToggleWidget = XtNameToWidget(this->ed->Mods(), Name());
  if (!ToggleWidget)
    {
      ToggleWidget= this->ed->CreateEntry(Name(), xmToggleButtonWidgetClass,
					  SubWidget, NULL, dummy);
    }  
  //button above drawing area
  ButtonWidget = XtNameToWidget(this->ed->Box(), "DecompositionButton");
  if (!ButtonWidget)
    {
      ButtonWidget= this->ed->CreateEntry("DecompositionButton",xmPushButtonWidgetClass,
					  this->ed->Box(), NULL,CallBackID);
      this->ed->ButtonState("DecompositionButton", false);        
    }
  return false;
}
// ----------------------------------------------------------------------------
// called on selection of any menu item 
bool Decomposition::Menu(Editor& editor, Page& page, Ident& name, bool before)
{
  if (name==Name() && before)
    {//this decomposition module was selected
      if (DecompositionEnable==false)  
	{// activate decomposition
	  if(page.SuperPage()) //page with superpage
	    {
	      net.ChangeEditor(editor,net.prime); //switch to prime page
	      net.UpdateButtonStates(editor, *net.prime);
	    }
	  DecompositionEnable=true;
	  XmToggleButtonSetState(this->ToggleWidget,true,true); //mark toggle button
	  InitDecomposition();
	}
      else
 	{ //deactivate decomposition
	  Text DialogContense;
 	  DialogContense+= Text("Decomposition exits").to_charPtr();	
	  OpenMsgDialog(*this->ed, "Decomposition_Exit", 
 		DialogContense.to_charPtr(), false, "\"Exit Decomp.\" Cancel");
	}
      return (true);
    }
  else if (DecompositionEnable && name=="Leave" && before)
    { // button "Leave" pressed to leave a subpage
      if (PageP superpage= page.SuperPage())	
	  LeaveSubPage(editor, page);	
      return (true);
    }
  else if (DecompositionEnable && name=="DecompositionButton" && before) 
    { //button "Decomposition" above drawing area pressed
      OpenOptionsMenu();  
      return (true);
    }
 else if (DecompositionEnable && name=="Save" && before) 
    { //Save button or menu entry
      SaveSubnet(page);  
      return (true);
    }
  else if (DecompositionEnable && before)
    { //block any other menu selection and open error dialog
      Text DialogContense;
      DialogContense+= (Text("Decomposition active, action not possible.")).to_charPtr();
      OpenMsgDialog(*this->ed, "Decomposition_Warning", 
 		    DialogContense.to_charPtr(), false, "\"Exit Decomp.\" Cancel");
      return (true);
    }
  
  if (DecompositionEnable) 
    return (true); 
  else
    return (false); 
}
// ----------------------------------------------------------------------------
// called on mouse activities
bool Decomposition::Event(Editor& ed, Page& pg, EventInfo& ev, bool before)
{
  if (DecompositionEnable && ev.event==EventDouble && ev.button==ButtonSelect && before)
    {// left mouse button doubleclick
      pg.SetCurrOb(ed.Dpy(), ev.pos); // get object under mouse
      ObjectP obj=  pg.GetCurrOb();
      if (obj)
	{
	  Page* subpage;
	  Node* node= (NodeP)obj;
	  if (obj->IsNode() && (subpage= node->SubPage()))
	    {// is it a subransition?
	      SubTransDoubleClick(ed, node); //enter subpage
	    }
	  else
	    {// open object dialog window
	      EditorSettings set= ed.Settings();
	      Widget shell= ed.Shell();
	      net.ObjectDialog(shell,obj,false,true);
	    }
	}
      return true;
    }
  else if (DecompositionEnable && ev.button==ButtonSelect && before)
    {//left mouse button single click
      //handed on to Page::Event to enable moving of objects
      pg.Event(ed, ev); 
      return true;
    }

  if (DecompositionEnable) 
    return (true); 
  else
    return (false); 
}
// ----------------------------------------------------------------------------
void Decomposition::InitDecomposition()
{
  this->ed->ButtonState("DecompositionButton", true);
  StartDecomposition();
}
// ----------------------------------------------------------------------------
void Decomposition::DeInitDecomposition()
{
  DecompositionEnable= false;
  EndDecomposition();
  XmToggleButtonSetState(this->ToggleWidget,false,true);
  this->ed->ButtonState("DecompositionButton", false);
  this->ed->Update();
}
// ----------------------------------------------------------------------------
// when doubleclicking a subtransition,
//   display aggregated or extended  subsystem
void Decomposition::SubTransDoubleClick(Editor &ed, Node* node)
{
  if(DisplayingSubPage) //entering a subsubpage
    {
      Page *subpage= node->SubPage();
      subpage->net = &net;
      net.ChangeEditor(ed,subpage);
      net.UpdateButtonStates(ed, *subpage);
      DisplayingSubPage++;
      return;
    }
  Net *n=0;
  if(displaymode == "Aggregated")
    {//all aggregated subnets are in the basic skeleton
      if(!BasicSkeleton) return;
      n=BasicSkeleton;
    }
  else if (displaymode == "Extended")
    {//extended subnet is in one of the low level systems
      LinkedList<SubTrans*> childs = top->Children();
      OpaqueIterator<SubTrans*> i2(childs);
      for (SubTrans** o=i2; o; o=++i2) 
	{
	  n=*LowLevelSystems[*o];
	  if((*o)->Path() == node->Name())
	    break; //contains extended subsystem
	} 
    }
  if(!n) return;
  Page *primepage = n->Prime(); //n is BS or a LS
  //find subtransition node in n 
  Node *subnode = primepage->GetNode(node->Name().to_charPtr());
  Page *subpage= subnode->SubPage(); 
  
  //bend pointers of subpage
  subpage->net = &net;
  subpage->parent = node;
  storepage = subpage; //store pointers to 
  storenode = subnode; //recover when pressing leave
  storenet  = n;

  //display subpage
  DisplayingSubPage++;
  net.ChangeEditor(ed,subpage);
  net.Update(subpage);
  net.UpdateButtonStates(ed, *subpage);

}
// ----------------------------------------------------------------------------
// leave a subpage where an extended or aggreated subnet is showed
void Decomposition::LeaveSubPage(Editor &ed, Page& page)
{
  if(DisplayingSubPage>1)
    {
      Page* superpage= page.SuperPage();
      page.net=storenet;
      net.ChangeEditor(ed,superpage);
      net.UpdateButtonStates(ed, *superpage);
      DisplayingSubPage--;
      return;
    }
  //restore pointers
  storepage->net = storenet;
  storepage->parent = storenode;
  
  net.ChangeEditor(ed,net.Prime()); //show prime page
  net.Update(net.Prime());
  net.UpdateButtonStates(ed, *(net.Prime()));
  DisplayingSubPage--;
}
// ----------------------------------------------------------------------------
//  the user is allowed to move objects on 
//  the aggregated/extended subpages. when the save-button
//  is pressed, the associated nets 
//  (basic skeleton , low level systems) have to be updated
void Decomposition::SaveSubnet(Page& page)
{
  if(displaymode == "Aggregated")
    {//saving aggregated subnets
      BasicSkeleton->Save(); //already contains changes 
      LinkedList<SubTrans*> childs = top->Children();
      OpaqueIterator<SubTrans*> i2(childs);
      for (SubTrans** o=i2; o; o=++i2) 
	{
	  Net *n=*LowLevelSystems[*o];
	  if((*o)->Path() == storenode->Name())
	    continue; //don't modify this one: contains extended subnet
	  Page *primepage = n->Prime();
	  Node *subnode = primepage->GetNode(storenode->Name().to_charPtr());
	  Page *subpage = subnode->subpage; //get old subpage
	  
	  //delete old nodes and connectors
	  Iterator<Node*> niter1(subpage->nodes);
	  for (Node* n=niter1; n; n=++niter1)
	    delete n;
	  
	  Iterator<Connector*> citer1(subpage->connectors);
	  for (Connector* c=citer1; c; c=++citer1)
	    delete c;
	  
	  //copy new nodes and connectors
	  Iterator<Node*> niter(page.nodes);
	  for (Node* n=niter; n; n=++niter)
	    {
	      Node* node = n->Copy(subpage);
	      node=node;
	    }
	  
	  Iterator<Connector*> citer(page.connectors);
	  for (Connector* c=citer; c; c=++citer)
	    c->Copy(subpage);
	  
	  n->Save(); //save net with modified subpage
	}
    }
  else if (displaymode == "Extended")
    {//save extended subnet
      LinkedList<SubTrans*> childs = top->Children();
      OpaqueIterator<SubTrans*> i2(childs);
      for (SubTrans** o=i2; o; o=++i2) 
	{
	  Net *n=*LowLevelSystems[*o];
	  if(!((*o)->Path() == storenode->Name()))
	    continue; //don't modify this one: contains aggregated subnet
	  //already contains changes
	  n->Save(); //save net with modified subpage
	}
    }
}
// ----------------------------------------------------------------------------
// opens a dialog box that shows the reason for an error of 
//an external program
void Decomposition::HandleProgramError(int exitcode, int pid, Text params)
{
  Text DialogContense;
  if (exitcode == CANTRUN)
    {
      DialogContense= "I was unable to run \n";
      DialogContense+= params + "\n";
      DialogContense+= "The program does not exist";
    }
  else if (exitcode == UNNORMALEXIT)
    {
      DialogContense= "I was unable to run \n";
      DialogContense+= params + "\n";
      DialogContense+= "The program was killed before termination\n";
      DialogContense+= "or aborted (segmentation fault).\n";
    }
  else if (exitcode == TIMEOUT)
    {
      DialogContense= "Waited too long, program\n";
      DialogContense+= params + "\n";
      DialogContense+= "will be killed";
      kill(pid, SIGINT);
    }
  else 
    {
      DialogContense= "The program call \n" ;
      DialogContense+= params + "\n";
      DialogContense+= "returned error number " ;
      DialogContense+= double(exitcode);
      DialogContense+= "\n";
    }
  OpenMsgDialog(*this->ed, "Decomposition_Error",
		DialogContense.to_charPtr(),true,"Exit");
}
// ----------------------------------------------------------------------------
// call MSAnalysis/bin/uncolor to generate .unfold file
bool Decomposition::CallUncolor(Text& netname)
{
  Text ana_dir= tnet_dir + "/MSAnalysis/bin/";
  Text params = ana_dir + "uncolor" + " -N " + netname - ".net";
  int exitcode, pid;
  pid = Exec(params.to_charPtr(), 0, 0);
  if(WaitForProcess(pid, exitcode))
    return true;
  else
    {
      HandleProgramError(exitcode, pid, params);
      return false;
    }
}
// ----------------------------------------------------------------------------
// create the table file that is used by the MSAnalysis programs
bool Decomposition::CreateTableFile(Text& netname, Net &net)
{
  Error error(*LastEditor());
  Text table_file = netname - ".net";
  table_file = table_file + ".table";
  if ((table_file).to_charPtr()[0]!='/') table_file = model_dir + "/" + table_file;
 
  ofstream file(table_file.to_charPtr());
  AnalyserInterface::DescribeSymbols(net.Prime(), file, true); 
  AnalyserInterface::DescribePage(net.Prime(), file, true, error);
  file.close();

  if (error || !file) return false;
  return true;
}
// ----------------------------------------------------------------------------
// additional to the information in the .unfold file, store 
// agnes - information like position, scale etc. for places and trans.
void Decomposition::FillStructures()
{
  unsigned int i;
  for(i=0; i< nb_edspnplaces; i++)
    {//for all places
      Text hcpnname;
      Node* node;
      EDSPNPlace *ep = edspnplaces[i];

      hcpnname = hcpnplaces[(ep->GetHCPNNumber())]->Name();
      node = p->FindNode(hcpnname);
    
      ep->scale = node->Scale();//=====> scale
      ep->relatives = node->GetRelative(0); //=====>relatives
      ep->orientation = node->Angle(); //=====>orientation
      ep->position= node->Mid();
    }
  for(i=0; i< nb_edspntransitions; i++)
    {//for all transitions
      Text hcpnname;
      Node* node;
      EDSPNTransition *et = edspntransitions[i];

      hcpnname = hcpntransitions[(et->GetHCPNNumber())]->Name();
      node = p->FindNode(hcpnname);
    
      et->scale = node->Scale();//=====> scale
      et->relatives = node->GetRelative(0); //=====>relatives
      et->orientation = node->Angle(); //=====>orientation
      et->position= node->Mid();
    }
}

// ----------------------------------------------------------------------------
//create arcs from fromnode to tonode. start search
//for nodes on subpage. revert reverts arc directions 
void Decomposition::DrawConnection(Text frompath, Text topath,
				   Page *subpage, bool revert /*=false*/)
{
  if(!subpage)
    return;
  Text firstpath = frompath.FirstPathPart('\\');
  if(firstpath==topath.FirstPathPart('\\'))
    {//level the paths 
      DrawConnection(frompath.StripFirstPathPart('\\'), topath.StripFirstPathPart('\\'),
		     (subpage->FindNode(firstpath))->SubPage(), revert);
      return;
    }
  if((!firstpath.to_bool()) && (topath.FirstPathPart('\\').to_bool()))
    {//let firstpath always be the longer path
      DrawConnection(topath, frompath, subpage, true);
      return;
    }
  Node *fromnode = subpage->FindNode(frompath);
  if(!fromnode)
    {
      cout<<"DrawConnection: couldn't find from"<<frompath.to_charPtr()<<endl;
      return;
    }
  Node *tonode = subpage->FindNode(topath);
  if(!tonode)
    {
      cout<<"DrawConnection: couldn't find to"<<topath.to_charPtr()<<endl;
      return;
    }
  NodeClass *fromclass = fromnode->Class();
  NodeClass *toclass = tonode->Class();
  ConnectorClass *cclass;

  if((fromclass == subpage -> GetNodeClass("EPlace")))
    cclass = subpage ->GetConnectorClass("EArc");
  else if ((fromclass == subpage -> GetNodeClass("PPlace")))
    cclass = subpage ->GetConnectorClass("PArc");
  else if((toclass == subpage -> GetNodeClass("EPlace")))
    cclass = subpage ->GetConnectorClass("EArc");
  else if ((toclass == subpage -> GetNodeClass("PPlace")))
    cclass = subpage ->GetConnectorClass("PArc");
  else{
    cout<<"DrawConnection: node types don't match"<<endl;
    return;}
    
  Text restpath = frompath;
  firstpath =  restpath.FirstPathPart('\\');
  Page *actpage = subpage;
  Node *acttonode = tonode;
  
  while (firstpath.to_bool())
    {
      Node *strans = actpage->GetNode(firstpath.to_charPtr());
      Connector* conn2 = new Connector(cclass,actpage);
      if(revert) 
	conn2->SetConnections(acttonode,strans);
      else
	conn2->SetConnections(strans, acttonode);
      conn2->Rearrange();
      strans->UpdateHierarchy();
      strans->UpdateVirtualConnections();
      actpage = strans->SubPage();
      acttonode = actpage->GetNode(topath.to_charPtr());
      restpath = restpath.StripFirstPathPart('\\');
      firstpath =  restpath.FirstPathPart('\\');
    }
  
  Connector* conn1 = new Connector(cclass,actpage);
  if(revert)
    conn1->SetConnections(acttonode, fromnode);
  else
    conn1->SetConnections(fromnode,acttonode);
  conn1->Rearrange();
}
// ----------------------------------------------------------------------------
//  Add connections to an existing binding
//  Binding and transition are given by transnode/bindnr.
void Decomposition::AddConnection(Node* transnode, //add connection to this transition
				  int bindnr, //and this bind_nr
				  Text tokens,
				  Node* inplace,  
				  Node* outplace)
{
  int type;
  Text global_guard, selcol, priority, server_type, distr;
  List *bindlist; 
  Text unused;
  Text newbind, bind, placename, attributes;
  Text localCondition, weight, delay;
  List inArcs, outArcs;
  int delaytype;
 
  TokenizeTransition(transnode, type, global_guard, bindlist, unused,
		     selcol, priority, server_type, distr);

  if(!(bindlist->Length()>bindnr))
    {
      cerr<<"Transition "<<transnode->Name()<<" has no binding nr. "<<bindnr<<endl;
      return;
    }
  //the corresponding binding of the transition 
  bind = bindlist->GetEntry(bindnr)->to_charPtr(); 
  //parse binding
  TokenizeBinding(bind, localCondition, inArcs, outArcs, weight,
		  delaytype, delay);
  
  if(inplace)
    {
      Text *in = new Text( Text(inplace->Name().to_charPtr())+ Text(" = ") + tokens);
      inArcs.Append(in);
    }
  else if(outplace)
    {
      Text *out = new Text(Text(outplace->Name().to_charPtr())+ Text(" = ") + tokens);
      outArcs.Append(out);
    }
      
  AssembleBinding(bind, localCondition, inArcs, outArcs, weight,
		  delaytype, delay);
 
 List* newbindlist = new List();
  
  for(int j=0; j< bindlist->Length(); j++)
    {
      if(j==bindnr)
	newbindlist->Append((Value *)&bind);
      else
	newbindlist->Append(bindlist->GetEntry(j));
    }

  AssembleTransition(transnode, type, global_guard, newbindlist, unused,
		     selcol, priority, server_type, distr);
}
// ----------------------------------------------------------------------------
//  Create a new binding. All properties except the 
//  connections are copied from an existing transition copynode
void Decomposition::CreateConnection(Node* transnode, //create connection to this transition
				     int bindnr, //and this bind_nr
				     Node* copynode, //copy properties of this transition
				     int copybindnr, //and this binding
				     Text tokens,
				     Node* inplace,  
				     Node* outplace)
{
  int type;
  Text global_guard, selcol, priority, server_type, distr;
  List *bindlist, *correspbindlist; 
  Text unused;
  Text newbind, bind, placename, attributes;
  Text localCondition, weight, delay;
  List inArcs, outArcs;
  int delaytype;
  
  TokenizeTransition(transnode, type, global_guard, bindlist, unused,
		     selcol, priority, server_type, distr);
  
  if(bindlist->Length()!=bindnr)
    {
      if(bindlist->Length()>bindnr) //binding exists
	{
	  AddConnection(transnode, bindnr, tokens, inplace, outplace);
	  return;
	}
      else //there is a gap between existing bindings and desired bind nr
	{//cant handle this
	  cerr<<"CreateConnection: Transition "<<transnode->Name();
	  cerr<<"has only "<<bindlist->Length()<<"bindings, "<<"can't append bind nr"<<bindnr<<endl;
	  return;
	}
    }

  TokenizeTransition(copynode, type, global_guard, correspbindlist, unused,
		     selcol, priority, server_type, distr);
  
  if(!(correspbindlist->Length()>copybindnr))
    {
      cerr<<"CreateConnection: Transition "<<copynode->Name()<<" has no binding nr. "<<copybindnr<<endl;
      return;
    }
  //the corresponding binding of the transition 
  bind = correspbindlist->GetEntry(bindnr)->to_charPtr(); 
  //parse binding
  TokenizeBinding(bind, localCondition, inArcs, outArcs, weight,
		  delaytype, delay);

  inArcs.Clear(); outArcs.Clear(); 
  if(inplace)
    {
      Text *in = new Text( Text(inplace->Name().to_charPtr())+ Text(" = ") + tokens);
      inArcs.Append(in);
    }
  else if(outplace)
    {
      Text *out = new Text(Text(outplace->Name().to_charPtr())+ Text(" = ") + tokens);
      outArcs.Append(out);
    }
  
  AssembleBinding(bind, localCondition, inArcs, outArcs, weight,
		  delaytype, delay);
  
  bindlist->Append((Value *)&bind);
		   
  AssembleTransition(transnode, type, global_guard, bindlist, unused,
		     selcol, priority, server_type, distr);
}

