#ifndef DECOMPOSITION_H
#define DECOMPOSITION_H
// -----------------------------------------------------------------
// Base-class for net-decomposition (MSIPDecomp, PathDecomp)
// -----------------------------------------------------------------
// includes
// -----------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>

#include "xbefore.h"
#include <Xm/ToggleB.h>
#include <Xm/CascadeB.h>
#include <Xm/PushB.h>
#include "xafter.h"

#include "linked.h"
#include "commmod.h"
#include "module.h"
#include "main.h"
#include "editor.h"
#include "net.h"
#include "node.h"
#include "netcl.h"
#include "dialog.h"
#include "wrapper.h"
#include "list.h"
#include "HCPN2EDSPN.h"

// -----------------------------------------------------------------
// class declaration
// -----------------------------------------------------------------
class Decomposition: virtual public CommModule, public Wrapper<Decomposition,bool>,
		     public HCPN2EDSPN
{
protected:
  EditorP ed; //editor and page
  PageP p; //where decomposition was called

  WrapperID   CallBackID;
  Widget      ButtonWidget, MenuWidget, SubWidget, ToggleWidget;

  bool DecompositionEnable; //decomposition enabled?
  unsigned int DisplayingSubPage; //hierarchy level

  Text netname; //name and path of net without .net extension, eg. "/home/Models/mynet"
  Text purename; //netname without path, eg. "mynet"
  Text dirname; //name of working directory 

  bool errorflag; //error occured
  Text errortext; //explanation of error

  int       pid; //to store a process id
  SubTrans* top; //top subtransition

  Page *storepage; //when displaying subnets:
  Node *storenode; //to temporarily store a subpage, a node,
  Net  *storenet;  // and a net

  Text displaymode; //"Aggregated" or "Extended"

  Net *BasicSkeleton; /* basic skeleton net*/
  Array<SubTrans*,Net*> LowLevelSystems; /*array containing low-level-system-nets
					   for each SubTrans*/

  // -----------------------------------------------------------------
  virtual void  OpenOptionsMenu() {}; //when clicking on the decomp. button
  virtual void  StartDecomposition() {}; //start chooses compute or load
  virtual void  ComputeDecomposition() {}; //(re)compute a net decomposition
  virtual void  LoadDecomposition() {}; //load old decomposition
  virtual void  EndDecomposition() {}; //ends decomposition
  
  void SubTransDoubleClick(Editor &ed, Node* node); //when doubleclicking on subtrans
  void LeaveSubPage(Editor &ed, Page& page); //when clicking on leave button
  void SaveSubnet(Page& page); //when clicking on save button

  bool CreateTableFile(Text& netname, Net &net); //create a table file for MSAnalysis progs
  void HandleProgramError(int exitcode, int pid, Text params); //handle errors of ex. progs
  bool CallUncolor(Text& netname); //call external program for .unfold-file
  void FillStructures(); //add "Agnes-information" to the structures read from unfold-file
 
  void DrawConnection(Text frompath, Text topath,       //draw arcs
		      Page *subpage, bool revert=false);
  void CreateConnection(Node* transnode, int bindnr, // create binding with connection
			Node* copynode, int copybindnr, 
			Text tokens, Node* inplace, Node* outplace);
  void AddConnection(Node* transnode, int bindnr, //add connection to binding
		     Text tokens, Node* inplace, Node* outplace);
  
// -----------------------------------------------------------------
public:

  Decomposition(Net& n);
  ~Decomposition();
 
  void InitDecomposition();
  void DeInitDecomposition();

  virtual bool  Init() {return true;}; 
  virtual bool  Open(Editor&,Page&);
  virtual bool  Change(Editor&,Page&,Page&) {return false;};
  virtual bool  Menu(Editor&,Page&,Ident&,bool);
  virtual bool  Event(Editor&,Page&,EventInfo&,bool);
  //be the first module to evaluate menu calls
  virtual bool MenuLock() {return DecompositionEnable;}; 

};

#endif
