
#include "evalmod.h"
#include "main.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <signal.h>

typedef char* charP;

 //AH Changed Start replaced numerous charP(_bla_) (char *) by _bla_.to_charPtr()
 // inserted  .to_bool() .to_double()
 //AH Changed End 

//-----------------------------------------------------------------------------
// EvalSettings enthlt die Strukturdefinition aller Auswertungs-Parameter
// Default: keine Parameter
//
Text EvalModule::EvalSettings()
{ return "[]"; }

//-----------------------------------------------------------------------------
// EvalSettings ergibt die Strukturdefinition aller Experiment-Parameter
//
Text EvalModule::ExperimentSettings()
{
  return
    "[ text (0) \"Experiment Variable\" var = \"\";"
    "  text (0) \"Start Value\" start = \"\";"
    "  text (0) \"End Value\" end = \"\";"
    "  text (0) \"Step Size\" step = \"\";"
    "  enum (\"Linear\" lin, \"Logarithmic\" log) \"Step Mode\" mode = lin;"
    "  text \"Result File\" file = \"experiment.log\" ]";
}

//-----------------------------------------------------------------------------
// Die Methode Call wird beim Anwhlen des Menpunkts aufgerufen. Die Dialog-
// Fenster werden erzeugt und die Auswertungs-Parameter werden abgefragt
//
void EvalModule::Call(Editor& editor, Page&)
{
  if (!CurrentDialog()) {
    CreateDialog(EvalSettings().to_charPtr());
    CreateDialog(ExperimentSettings().to_charPtr());
    CreateFilesel(model_dir.to_charPtr());
  }

  SetCurrentDialog(1);

  //AH Changed const_cast<char*> added
  OpenDialog(editor, Name(),
             AllowExperiments() ? const_cast<char*>("Start Experiment Load Save Cancel")
                                : const_cast<char*>("Start Load Save Cancel"));
}

//-----------------------------------------------------------------------------
// OnDialogButton wird beim Bettigen eines Dialog-Buttons aufgerufen.
// Daraufhin wird die entsprechende Aktion ausgefhrt.
//
void EvalModule::OnDialogButton(Ident& button)
{
  if (button=="Start") {
    if (!StartExperiment())
      StartEvaluation();
    
    SetCurrentDialog(1);
    CloseDialog();
  }
  
  else if (button=="Experiment") {
    SetCurrentDialog(2);
    OpenDialog(*LastEditor(), "Experiment Settings", "Ok Load Save Clear Cancel");
  }

  else if (button=="Ok" || button=="Cancel") {
    CloseDialog();
    if (CurrentDialog()>2) {
      DestroyDialog();
      SetCurrentDialog(1);
    }
  }

  else if (button=="Load" || button=="Save") {
    save_options = (button == "Save");
    OpenFilesel(*LastEditor(), "Select Option File", "*.opt");
  }

  else if (button=="Keep" || button=="Kill") {
    DestroyDialog();
    SetCurrentDialog(1);
    DestroyLogwin();

    if (button=="Kill")
      kill(pid, SIGINT);
  }
}

//-----------------------------------------------------------------------------
// OnLogwinButton wird beim Bettigen des Logwindow-Buttons aufgerufen. 
// Daraufhin wird das Fenster geschlossen oder der Kill-Dialog aufgerufen.
//
void EvalModule::OnLogwinButton(Ident& button)
{
  if (button=="Close") {
    int dummy;
    if (waitpid(pid, &dummy, WNOHANG))
      DestroyLogwin();
    else {
      CreateDialog();
      OpenMsgDialog(*LastEditor(), "Warning",
                    "Process still running. Kill it or keep it alive?",
                    true, "Keep Kill Cancel");
    }
  }
}

//-----------------------------------------------------------------------------
// OnFilesel wird nach dem Auswhlen eines Files aufgerufen. Daraufhin werden
// die Parameter geladen oder gespeichert.
//
void EvalModule::OnFilesel(char* name, bool ok)
{
  if (!ok) return;

  Text filename = Text(name) - ".opt" + ".opt";

  SetCurrentDialog(1); Struct* evals = GetParams(); 
  SetCurrentDialog(2); Struct* exps = GetParams();

  if (save_options) {
    ofstream file(filename.to_charPtr());
    evals->Output(file);
    exps->Output(file);
  }
  else {
    ifstream file(filename.to_charPtr());
    
    evals->Input(file);
    exps->Input(file);
  }
}

//-----------------------------------------------------------------------------
// StartEvaluation startet die externe Auswertungskomponente. Die Ergebnisse
// werden mit dem CommandLogwindow eingelesen
//
void EvalModule::StartEvaluation()
{
  Text command;
  Text stdout_file, stdin_file;
  Text net_file, result_file;
  Text logwin_title, logwin_buttons = "Close";
  bool succ;

  succ = EvalCall(false, command, stdout_file, stdin_file, net_file,
                  result_file, logwin_title, logwin_buttons);

  ofstream outf(stdout_file.to_charPtr(), ios::trunc);outf.close();
  pid = Exec(command.to_charPtr(), stdout_file.to_charPtr(), stdin_file.to_charPtr());

  if (stdout_file.to_bool()) {
    CreateLogwin(stdout_file.to_charPtr());
    OpenLogwin(*LastEditor(), logwin_title.to_charPtr(), logwin_buttons.to_charPtr());
  }

  if (result_file != stdout_file) 
    CreateLogwin(result_file.to_charPtr());
}
  
//-----------------------------------------------------------------------------
// StartExperiment fhrt ein Experiment durch. Dazu wird ein Script-File
// erzeugt, das nacheinander fr alle Variablen die Auswertungskomponente
// aufruft. Das Scriptfile ist so konzipiert, da es sich jederzeit abbrechen
// und neustarten lt. Die Ergebnisse werden mittels grep & cut aus dem
// Ausgabefile extrahiert. Am Ende des Experiments werden die Ergebnisse
// graphisch ausgegeben.
//
bool EvalModule::StartExperiment()
{
  SetCurrentDialog(2);

  Text var = Param("var");
  double start = Param("start").to_double();
  double end = Param("end").to_double();
  double step = Param("step").to_double();
  bool logarithmic = (Param("mode") == "log");
  Text results_file = model_dir + "/" + Param("file");

  SetCurrentDialog(1);

  if (var=="") return false;
  
  if (end<=start || step<=0 || logarithmic && step<=1) {
    OpenMsgDialog(*LastEditor(), "Experiment Error",
                  "Illegal Experiment Settings");
    return true;
  }

  if (!SetValue(var, start)) {
     OpenMsgDialog(*LastEditor(), "Experiment Error",
                  "Illegal Experiment Variable");
    return true;
  }

  Text command, cmd;
  Text stdout_file, stdin_file;
  Text net_file, result_file;
  Text logwin_title, logwin_buttons = "Close";
  bool succ;
  
  int current_index = 0;
  double current_x = start;
  Text script = "#!/usr/bin/csh\n";

  while (current_x <= end) {
    SetValue(var, current_x);
    succ = EvalCall(true, command, stdout_file, stdin_file, net_file,
                    result_file, logwin_title, logwin_buttons);
    if (!succ) continue;

    Text index; index = current_index;
    Text tmpfile = Text(net_file) + ".tmp" + index; 
    rename(net_file.to_charPtr(), tmpfile.to_charPtr());

    cmd = Text("if (-r ") + tmpfile + ") then\n"
        + "\\cp " + tmpfile + " " + net_file + "\n";
                  
    if (stdin_file.to_bool()) cmd = cmd + stdin_file + " > ";
    cmd += command;
    if (stdout_file.to_bool()) cmd = cmd + " > " + stdout_file + "\n";

    cmd = cmd + "if ($status) then\n"
              + "echo \\!clear\ncat " + result_file + " >> " + results_file
              + "\nexit\nendif\n";
  
    script = script + cmd;

    if (current_index == 0) {
      cmd = Text("echo \\# ") + var + " `grep \\\\!setattribute "
          + result_file + " | cut -f2 -d\\ ` > " + results_file + "\n";
      script = script + cmd;
    }

    Text cx; cx = current_x;
    cmd = Text("echo ") + cx + " `grep \\\\!setattribute "
        + result_file + " | cut -f2 -d\\\"` >> " + results_file + "\n"
        + "\\rm " + tmpfile + "\n"
        + "endif\n";
    
    script = script + cmd;

    if (logarithmic) current_x *= step;
    else current_x += step;
    current_index++;
  }

  cmd = Text("exec ") + tnet_dir + "/Agnes/scripts/plot " + results_file +"\n"
      + "\\rm $0\n"; 

  script = script + cmd;

  Text script_file = model_dir + "/experiment_script";
  ofstream file( script_file.to_charPtr());
  file<< script.to_charPtr();
  file.close();
  
  chmod(script_file.to_charPtr(), S_IRWXU);
  pid = Exec(script_file.to_charPtr());
  
  ofstream rfile( results_file.to_charPtr()); rfile.close();
  CreateLogwin(results_file.to_charPtr());
  OpenLogwin(*LastEditor(), "Experiment Results", logwin_buttons.to_charPtr());

  return true;
}
