#ifndef KOMM
#define KOMM 0

#include <fstream.h>
#include <string.h>

const int kommando_max=500;



// **********************************************************************
// Mit der Klasse T_FKommunikation koennen Nachrichten in Form von
// Zeichenketten zwischen Prozessen ausgetauscht werden. Einzelne Nachrichten
// werden durch das Zeichen \n voneinander getrennt. Dadurch ist das
// sequentielle Lesen der Nachrichten moeglich.
// Die Zeichenketten werden in Dateien uebergeben,
// wobei fuer jede der beiden Datenrichtungen jeweils eine Datei genutzt wird.
// **********************************************************************

// *********** Klassendeklaration T_FKommunikation (Anfang) ******************

class T_FKommunikation
{
 private:
  // empfang_name ist ein Zeiger auf den Pfadnamen der Datei aus der Daten
  // eines anderen Prozesses gelesen werden sollen.
  char *empfang_name;
   
  // sende_name ist ein Zeiger auf den Pfadnamen der Datei in die Daten
  // fuer einen anderen Prozess geschrieben werden sollen.
  char *sende_name;
  
  // empf_puf enthaelt alle Nachrichten die nach dem letzten Aufruf von
  // Nachrichten_Lesen aus der entsprechenden Datei gelesen wurden.
  char empf_puf[kommando_max+1];

  // Empf_Akt ist ein Zeiger auf die aktuelle Nachricht innerhalb des
  // Puffers empf_puf.
  char *Empf_Akt;
  
 public:
  // empf_puf enthaelt alle Nachrichten die seit dem letzten erfolgreichem
  // Schreiben in die Datei sende_name zum senden gesammelt wurden.
  // Nachrichten_Senden 
  char *sende_puf;

  // Konstruktor
  // Uebergeben wird der Name der Datei aus dem Daten von
  // einem anderen Prozess empfangen werden und der Name der Datei in
  // der Daten an einen anderen Prozess gesendet werden.
  T_FKommunikation(const char Empfang_name[],
		   const char Sende_name[]);
  
  // Destruktor 
  ~T_FKommunikation();
  
  // Liefert einen Zeiger auf die naechste noch nicht gelesene
  // Nachricht im Puffer empf_puf. Falls es keine ungelesenen
  // Nachrichten mehr gibt ist der Rueckgabewert 0.
  char *Nachricht();
  

  // Der Puffer empf_puf wird geloescht.
  // Alle Nachrichten aus der Datei empfang_name werden in Puffer empf_puf
  // geschrieben und der Inhalt der Datei anschliessend geloescht.
  // Liefert die Laenge der empfangenen Nachrichten in Bytes zurueck;
  int Nachrichten_Lesen();
  

  // Die Nachricht s wird an den Puffer sende_puf angehangen
  void Nachricht(const char s[]);
 

  // Wenn die Datei sende_name leer ist werden alle Nachrichten aus
  // dem Puffer sende_puf in diese Datei geschrieben und
  // der anschiessend der Pufferinhalt geloescht.
  // Der Rueckgabewert ist true, wenn die Datei leer war und das
  // Schreiben in die Datei erfolgreich war.
  bool Nachrichten_Senden();

};

// *********** Klassendeklaration T_FKommunikation (Ende) ******************



#endif 
