// -----------------------------------------------------------------
// Debug Definitionen
// -----------------------------------------------------------------

// #define __OUTSTRING__ 
// Ausgabe der Parser-Strings

// -----------------------------------------------------------------
// Importe
// -----------------------------------------------------------------

#include "parser.h"

//------------------------------------------------------------------------
//Hilfsfunktionen
//------------------------------------------------------------------------

void Parser::Out(const charP message, charP string, int length)
{
  cout << message << endl;
  for (int i=0; i< length; i++)
    cout << string[i];
  cout << endl << ends << flush;
}

//------------------------------------------------------------------------

void Parser::DestroyBlancs(charP string)
{
  if (!string || !*string)
    return;

  int length= (int)strlen(string);
  for (int i=0; i<length; i++)
    if (string[i]==' ')
      {
	memmove(string+i, string+i+1, length-i);
	length--;
      }
}

//------------------------------------------------------------------------

int Parser::SearchOpeningBracket(charP string, int endpos)
{
  int i;
  int bracecnt= 0;
  for (i= endpos; i>=0; i--)
    {
      if      (string[i] == ')') bracecnt --;
      else if (string[i] == '(') bracecnt ++;
      if      (!bracecnt) break;
    }
  return i;
  // search for opening bracket:
  // ...(..(..(..(...)..)..)..)
  //    ^
}

//------------------------------------------------------------------------

int Parser::SearchOpeningBracket2(charP string, int endpos)
{
  int i;
  int bracecnt= 0;
  for (i= endpos; i>=0; i--)
    {
      if      (string[i] == '}') bracecnt --;
      else if (string[i] == '{') bracecnt ++;
      if      (!bracecnt) break;
    }
  return i;
  // search for opening bracket:
  // ...{..{..{..{...}..}..}..}
  //    ^
}

//------------------------------------------------------------------------

int Parser::SearchLOperator(charP string, int endpos)
{
#ifdef __OUTSTRING__
  Out ("SearchLOperator:",
       string,endpos+1); 
#endif
  
  bool op_found= false;
  int op_index= -1;
  for (int i= endpos; i>=0; i--)
    {
      if (((!strncmp(string+i, "AND", 3) && (i<=endpos-2))) ||
	  ((!strncmp(string+i, "OR",  2) && (i<=endpos-1))))
	{// "AND" | "OR" found?

	  op_found= true;
	  op_index= i;
	  // save position of last operand
	}
      else if (op_found && op_index==i+3)
	break;
    }
  return ((op_index>0)? op_index: -1);
  // search for logic operator
  // ...AND... | ...OR...
  //    ^           ^
}

//------------------------------------------------------------------------

int Parser::SearchOperator(charP string, int endpos)
{
#ifdef __OUTSTRING__
  Out ("SearchOperator:",
       string,endpos+1);
#endif

  bool op_found= false;
  int op_index= -1;
  for (int i= endpos; i>=0; i--)
    {
      if (  (string[i] == '*') ||
	    (string[i] == '/') ||
	    (string[i] == '^') ||
	  (((string[i] == '+') || 
	    (string[i] == '-')) &&
	   !IsSignOfExponent(string,i)))
	{// "+" | "-" | "*" | "/" | "^" found?

	  op_found= true;
	  op_index= i;
	  // save position of last operand
	}
      else if (op_found && op_index==i+1)
	break;
    }
  return ((op_index>0)? op_index: -1);
  // search for operator before expression
  // ...+... | ...-... | ...*... | .../... | ...^...
  //    ^         ^         ^         ^         ^
}

//------------------------------------------------------------------------

int Parser::SearchLowestLOperator(charP string, int endpos)
{
  bool and_found= false;
  bool or_found= false;
  int index, and_index=0, or_index=0;

  do
    {
      if (string[endpos] == ')')
	{// "...)" <-- logic condition is in brackets?
	  
	  index= SearchOpeningBracket(string,endpos);
	  // "...(...)" <-- expression in brackets
	  //     ^
	  if (index==-1)
	    {
	      error= true;
	      ParserError("'(' is missing:", 
			  string, endpos+1);
	      return -1;
	      // error, no opening bracket found
	    }
	}
      else
	index= endpos;
      
      index= SearchLOperator(string, index);
      
      if (index!=-1)
	if (!or_found && (string[index]=='O'))
	  {
	    or_found= true;
	    or_index= index;
	  }
	else if (!and_found && (string[index]=='A'))
	  {
	    and_found= true;
	    and_index= index;
	  }

    } while ((endpos= index-1)>0);
  // search the rest of the whole string

  if (or_found)
    return or_index;
  // lowest priority:
  // "OR" found?

  else if (and_found)
    return and_index;
  // highest priority:
  // "AND" found?

  return -1;
  // nothing found
}

//------------------------------------------------------------------------

int Parser::SearchLowestOperator(charP string, int endpos)
{
  bool pow_found= false;
  bool muldiv_found= false;
  bool plusmin_found= false;
  int index, pow_index=0, muldiv_index=0, plusmin_index=0;

  do
    {
      if (string[endpos] == ')')
	{// "...)" <-- expression is in brackets?
	  
	  index= SearchOpeningBracket(string,endpos);
	  // "...(...)" <-- expression in brackets
	  //     ^
	  if (index==-1)
	    {
	      error= true;
	      ParserError("'(' is missing:", 
			  string, endpos+1);
	      return -1;
	      // error, no opening bracket found
	    }
	}
      else if (type==reward_def && string[endpos] == '}')
	{// "...}" <-- rew_item is in brackets? (nur fuer <reward_def>)

	  index= SearchOpeningBracket2(string,endpos);
	  // "...{...}" <-- rew_item in brackets
	  //     ^
	  if (index==-1)
	    {
	      error= true;
	      ParserError("'{' is missing:", 
			  string, endpos+1);
	      return -1;
	      // error, no opening bracket found
	    }
	}
      else
	index= endpos;
      
      index= SearchOperator(string, index);
      
      if (index!=-1)
	if (!plusmin_found && ((string[index]=='+') || (string[index]=='-')))
	  {
	    plusmin_found= true;
	    plusmin_index= index;
	  }
	else if (!muldiv_found && ((string[index]=='*') || (string[index]=='/')))
	  {
	    muldiv_found= true;
	    muldiv_index= index;
	  }
	else if (!pow_found && (string[index]=='^'))
	  {
	    pow_found= true;
	    pow_index= index;
	  }

    } while ((endpos= index-1)>0);
  // search the rest of the whole string

  if (plusmin_found)
    return plusmin_index;
  // lowest priority:
  // "+" | "-" found?

  else if (muldiv_found)
    return muldiv_index;
  // "*" | "/" found?

  else if (pow_found)
    return pow_index;
  // highest priority:
  // "^" found?

  return -1;
  // nothing found
}

//------------------------------------------------------------------------

bool Parser::IsSignOfExponent(charP string, int endpos)
{
  if (!endpos)
    return false;

  if ((string[endpos-1]!='E') && (string[endpos-1]!='e'))
    return false;
  // no exponent sign found?

  int i;
  for (i= endpos-1; i>=0; i--)
    if ((string[i] == '+') ||
	(string[i] == '-') ||
	(string[i] == '*') ||
	(string[i] == '/') ||
	(string[i] == '^') ||
	(string[i] == '(') ||
	((type==reward_def) && (string[i] == '{')))
      break;
  // search for beginning of expression

  if (isalpha(string[i+1]))
    return false;
  // exponent sign is part of variable-name

  return true;
}

//------------------------------------------------------------------------
// <pmf_definition>
//------------------------------------------------------------------------

void Parser::ParsePMFDefinition(charP string, int length)
{
#ifdef __OUTSTRING__
  Out("ParsePMFDefinition:",
      string,length);
#endif

  int endpos;

  if (!string || !*string || (length<=0) || length>(int)strlen(string))
    {
      error= true;
      ParserError("missing <pmf_definition>!",
		  "", 0);
      return;
      // error, <pmf_definition> missing
    }
  else
    endpos= length-1;
  // NULL or "" <-- no expression or wrong value

  if (!strncmp(string, "DETERMINISTIC(", 14) && string[endpos]==')')
    {// "DETERMINISTIC(" <real_constant> ")"

      ParseRealConstant(string+14, length-15);
    }
  else if (!strncmp(string, "UNIFORM(", 8) && string[endpos]==')')
    {// "UNIFORM(" ... ")"

      charP pointer;
      if ((pointer= strchr(string, ',')) && (pointer<=string+endpos))
	{
	  int k= pointer-string;
	  // "UNIFORM(" <real_constant> "," <real_constant> ")"
	  //                             ^
	  
	  ParseRealConstant(string+8, k-8);
	  ParseRealConstant(string+k+1, endpos-(k+1));
	}
      else
	{
	  error= true;
	  ParserError("',' is missing:", 
		      string, length);
	  return;
	  // error, no punctuation found
	}
    }
  else
    {// <pmf_expression>

      ParsePMFExpression(string, length);
    }
}
//------------------------------------------------------------------------
// <pmf_expression>
//------------------------------------------------------------------------

void Parser::ParsePMFExpression(charP string, int length)
{
#ifdef __OUTSTRING__
  Out("ParsePMFExpression:",
      string,length);
#endif

  int endpos;

  if (!string || !*string || (length<=0) || length>(int)strlen(string))
    {
      error= true;
      ParserError("missing <pmf_expression>!",
		  "", 0);
      return;
      // error, <pmf_expression> missing
    }
  else
    endpos= length-1;
  // NULL or "" <-- no expression or wrong value

  for (int i= endpos-1; i>=0; i--)
    if ((string[i]==']') && ((string[i+1]=='+') || (string[i+1]=='-')))
      {
	// <pmf_expression> "+" <pmf_expression> |
	// <pmf_expression> "-" <pmf_expression>

	ParsePMFExpression(string,i+1);
	ParsePMFExpression(string+i+2, length-(i+2));
	return;
      }

  charP pointer;
  if ((pointer= strstr(string, "I[")) && pointer<= string+endpos)
    {// <impulse>

      ParseImpulse(string, length);
    }
  else if ((pointer= strstr(string, "R[")) && pointer<= string+endpos)
    {// <rectangle>

      ParseRectangle(string, length);
    }
  else
    {
      error= true;
      ParserError("<impulse> | <rectangle> is missing:", 
		  string, length);
      return;
      // error, no impulse or rectangle found
    }
}

//------------------------------------------------------------------------
// <impulse>
//------------------------------------------------------------------------

void Parser::ParseImpulse(charP string, int length)
{
#ifdef __OUTSTRING__
  Out("ParseImpulse:",
      string,length);
#endif

  int endpos;

  if (!string || !*string || (length<=0) || length>(int)strlen(string))
    {
      error= true;
      ParserError("missing <impulse>!",
		  "", 0);
      return;
      // error, <impulse> missing
    }
  else
    endpos= length-1;
  // NULL or "" <-- no expression or wrong value

  if (string[endpos]!=']')
    {
      error= true;
      ParserError("']' is missing:", 
		  string, length);
      return;
      // error, no closing bracket found
    }

  charP pointer;
  if ((pointer= strstr(string, "I[")) && (pointer<=string+endpos))
    { 
      int k= pointer-string;
      
      if (pointer==string)
	{
	  // "I[" <real_constant> "]"
	  //  ^
	  
	  ParseRealConstant(string+2, length-3);
	}
      else
	{
	  // <real_constant> ["*"] "I[" <real_constant> "]"
	  //                        ^
	  
	  ParseRealConstant(string+k+2, length-k-3);
	  
	  if (string[k-1]=='*')
	    ParseRealConstant(string, k-1);
	  else
	    ParseRealConstant(string, k);
	}
    }
  else
    {
      error= true;
      ParserError("'I[' is missing:", 
		  string, length);
      return;
      // error, no opening bracket found
    }
}

//------------------------------------------------------------------------
// <rectangle>
//------------------------------------------------------------------------

void Parser::ParseRectangle(charP string, int length)
{
#ifdef __OUTSTRING__
  Out("ParseRectangle:",
      string,length);
#endif

  int endpos;

  if (!string || !*string || (length<=0) || length>(int)strlen(string))
    {
      error= true;
      ParserError("missing <rectangle>!",
		  "", 0);
      return;
      // error, <rectangle> missing
    }
  else
    endpos= length-1;
  // NULL or "" <-- no expression or wrong value

  if (string[endpos]!=']')
    {
      error= true;
      ParserError("']' is missing:", 
		  string, length);
      return;
      // error, no closing bracket found
    }

  charP pointer;
  if ((pointer= strchr(string,',')) && (pointer<=string+endpos))
    {
      int j= pointer-string;

      if ((pointer= strstr(string, "R[")) && (pointer<=string+j))
	{ 
	  int k= pointer-string;

	  if (pointer==string)
	    {
	      // "R[" <real_constant> "," <real_constant> "]"
	      //  ^
	      
	      ParseRealConstant(string+2, j-2);
	      ParseRealConstant(string+j+1, length-j-2);

	    }
	  else
	    {
	      // <expolynomial> ["*"] "R[" <real_constant> "," <real_constant> "]"
	      //                       ^
	      
	      ParseRealConstant(string+k+2, j-(k+2));
	      ParseRealConstant(string+j+1, length-j-2);

	      if (string[k-1]=='*')
		ParseExpolynomial(string, k-1);
	      else
		ParseExpolynomial(string, k);
	    }
	}
      else
	{
	  error= true;
	  ParserError("'R[' is missing:", 
		      string, length);
	  return;
	  // error, no opening bracket found
	}
    }
  else
    {
      error= true;
      ParserError("',' is missing:", 
		  string, length);
      return;
      // error, no punctuation found
    }
}

//------------------------------------------------------------------------
// <expolynomial>
//------------------------------------------------------------------------

void Parser::ParseExpolynomial(charP string, int length)
{
#ifdef __OUTSTRING__
  Out("ParseExpolynomial:",
      string,length);
#endif

  int endpos;

  if (!string || !*string || (length<=0) || length>(int)strlen(string))
    {
      error= true;
      ParserError("missing <expolynomial>!",
		  "", 0);
      return;
      // error, <expolynomial> missing
    }
  else
    endpos= length-1;
  // NULL or "" <-- no expression or wrong value

  int index= SearchLowestOperator(string, endpos);
  if (error)
    return;

  if (index==-1 || (string[index]!='+' && string[index]!='-'))
    {// no operator found or "*" | "/" | "^"

      if ((string[0]=='(') && (string[endpos]==')'))
	{
	  ParseExpolynomial(string+1, length-2); 
	  // "(" <expolynomial> ")"
	 
	  return;
	}

      else if (string[0]=='-')
	{
	  ParseExpolynomial(string+1, length-1);
	  // "-" <expolynomial>

	  return;
	}
      else
	{
	  charP pointer;
	  if (!strncmp(string+endpos-1, "x)", 2))
	    {
	      if ((pointer= strstr(string, "e^(-")) && (pointer<=string+endpos))
		{
		  int k= pointer-string;
		  // ...["*"]"e^(-" <real_constant> ["*"]"x)"
		  //          ^
		  
		  if (string[endpos-2]=='*')
		    ParseRealConstant(pointer+4, length-7-k);
		  else
		    ParseRealConstant(pointer+4, length-6-k);
	      
		  if ((k-1>0) && (string[k-1]=='*'))
		    length= k-1;
		  else
		    length= k;
		  endpos= length-1;
		}
	      else
		{
		  error= true;
		  ParserError("'e^(-' is missing:", 
			      string, length);
		  return;
		  // error, no exponent found
		}
	    }
	  
	  if ((pointer= strchr(string,'x')) && (pointer<=string+endpos)) 
	    {
	      int k= pointer-string;
	      //  <real_constant> ["*"]"x"["^" <integer_constant>] ...
	      //                        ^
	      
	      if ((k-1>0) && (string[k-1]=='*'))
		ParseRealConstant(string, k-1);
	      else
		ParseRealConstant(string, k);
	      
	      if ((k+1<endpos) && (string[k+1]=='^'))
		ParseIntegerConstant(string+k+2, length-(k+2));

	      else if ((k+1==endpos) && (string[k+1]=='^'))
		{
		  error= true;
		  ParserError("missing <integer_constant> after '^':", 
			      string, length);
		  return;
		  // error, no power sign found
		}
	      else if (k<endpos)
		{
		  error= true;
		  ParserError("'^' is missing:", 
			      string, length);
		  return;
		  // error, no power sign found
		}
	    }
	  else
	    {
	      ParseRealConstant(string, length);
	      // <real_constant> 
	    }
	  return;
	}
    }
  // <expolynomial> numop <expolynomial>
  
  char numop= string[index];

  switch (numop)
    {
    case '+':
    case '-':
      ParseExpolynomial(string,index); 
      ParseExpolynomial(string+index+1,endpos-index);
      // <expolynomial> "+" <expolynomial> |
      // <expolynomial> "-" <expolynomial>

      return;
      break;
    }
}

//------------------------------------------------------------------------
// <if_expr>
//------------------------------------------------------------------------

double Parser::ParseIfExpression(charP string, int length)
{
#ifdef __OUTSTRING__
  Out("ParseIfExpression:",
      string,(int)strlen(string));
#endif

  int endpos;

  if (!string || !*string || (length<=0) || length>(int)strlen(string))
    {
      error= true;
      ParserError("missing <if_expr>!",
		  "", 0);
      return 0.0;
      // error, <if_expr> missing
    }
  else
    endpos= length-1;
  // NULL or "" <-- no expression or wrong value

  charP h_string= string;
  do
    {
      if (!strncmp(h_string,"IF", 2) && (h_string<string+endpos))
	{// "IF" found?
	  
	  IsMDDelay= true;
	  IsMDArcMult= true;

	  charP pointer;
	  if ((pointer= strchr(h_string,':')))
	    {
	      int k= pointer-h_string;
	      // "IF"...":"...
	      //         ^

	      bool is_true= ParseLogicCondition(h_string+2, k-2);
		
	      if (((pointer= strstr(h_string+k,"IF")) ||
		   (pointer= strstr(h_string+k,"ELSE"))) && 
		  (pointer<=string+endpos))
		{
		  if (is_true)
		    return (ParseExpression(h_string+k+1, pointer-(h_string+k+1)));
		  h_string= pointer;
		}
	      else
		{
		  error= true;
		  ParserError("keyword must be 'ELSE' after <if_expr>:", 
			      h_string, (int)strlen(h_string));
		  return 0.0;
		}
	    }
	  else
	    {
	      error= true;
	      ParserError("':' is missing in <if_expr>:",
			  h_string, (int)strlen(h_string));
	      return 0.0;
	    }
	}
      else if (!strncmp(h_string, "ELSE", 4) && (h_string<string+endpos-2))
	{// "ELSE" found?
	  
	  IsMDDelay= true;
	  IsMDArcMult= true;

	  return ParseExpression(h_string+4,string+length-(h_string+4));
	}
      else if (h_string==string)
	{
	  return ParseExpression(h_string, length);
	  // only one expression
	}
      else
	{
	  error= true;
	  ParserError("keyword must be 'ELSE' after <if_expr>:",
		      h_string, (int)strlen(h_string));
	  return 0.0;
	}
    } while (h_string<string+length);
  // search the rest of the whole string

  return 0.0;  // to keep the compiler happy!!!
}

//------------------------------------------------------------------------
// <expression>
//------------------------------------------------------------------------

double Parser::ParseExpression(charP string, int length)
{
#ifdef __OUTSTRING__
  Out("ParseExpression:",
      string,length);
#endif

  int endpos;

  if (!string || !*string || (length<=0) || length>(int)strlen(string))
    {
      error= true;
      ParserError("missing <expression>!",
		  "", 0);
      return 0.0;
      // error, <expression> missing
    }
  else
    endpos= length-1;
  // NULL or "" <-- no expression or wrong value

  int index= SearchLowestOperator(string, endpos);
  if (error)
    return 0.0;

  if (index==-1)
    {// no operator found

      return ((string[0]=='-')?
	      -1*ParseExpression(string+1, length-1):
	      // "-" <expression>

	      ((string[0]=='(') && (string[endpos]==')'))? 
	      ParseExpression(string+1, length-2): 
	      // "(" <expression> ")"

	      ParseRealValue(string, length)
	      // <real_value>
	      );
    }

  // <expression> numop <expression>

  IsMDDelay= true;
  IsMDArcMult= true;
  IsMDDelayPar= true;
  
  char numop= string[index];
  
  double value;
  switch (numop)
    {
    case '+':
      return (ParseExpression(string,index) + 
	      ParseExpression(string+index+1,endpos-index));
      // <expression> "+" <expression>

      break;
    case '-':
      return (ParseExpression(string,index) - 
	      ParseExpression(string+index+1,endpos-index));
      // <expression> "-" <expression>

      break;
    case '*':
      return (ParseExpression(string,index) * 
	      ParseExpression(string+index+1,endpos-index));
      // <expression> "*" <expression>

      break;
    case '/':
      value= ParseExpression(string+index+1,endpos-index);
      
      if (type!=reward_def)
	{
	  if (value == 0.0)
	    {
	      error= true;
	      ParserError("division by 0 error:",
			  string, length);
	      return 0.0;
	      // error, division by 0
	    }
	}
      else value= 1.0;
      // do not evaluate, check syntax only (nur fuer <reward_def>)
      
      return (ParseExpression(string,index) / 
	      value);
      // <expression> "/" <expression>

      break;
    case '^':
      return (pow(ParseExpression(string,index), 
		  ParseExpression(string+index+1,endpos-index)));
      // <expression> "^" <expression>

      break;
    }            

  return 0.0;  // to keep the compiler happy!!!
}

// -----------------------------------------------------------------
// <real_value>
// -----------------------------------------------------------------

double Parser::ParseRealValue(charP string, int length)
{
#ifdef __OUTSTRING__
  Out("ParseRealValue:",
      string,length);
#endif

  int endpos;
  
  if (!string || !*string || (length<=0) || length>(int)strlen(string))
    {
      error= true;
      ParserError("missing <real_value>!",
		  "", 0);
      return 0.0;
      // error, <real_value> missing
    }
  else
    endpos= length-1;
  // NULL or "" <-- no expression or wrong value
  
  if (type==reward_def && length>=3 && 
      (string[0]=='P' || string[0]=='E') && 
      string[1]=='{' && string[endpos]=='}')
    return ParseRewItem(string+2, length-3, string[0]=='P');
  // "P{" ... "}" | "E{" ... "}" (nur fuer <reward_def>)

  if (isalpha(string[0]))
    {
      IsMDDelayPar= true;
      IsMDArcMult= true;
      
      return ParseParameter(string, length);
      // <real_parameter> | <integer_parameter>
    }

  bool real_const= false;
  for (int i= endpos; i>=0; i--)
    {
      if (string[i]=='.')
        // auskommentiert, da sonst alle Ausdrcke mit e/E als real 
        // erkannt werden (ArminZ 4.Mai01)
	// || (string[i]=='E') || (string[i]=='e'))
	real_const= true;
    }
  if (real_const)
    return ParseRealConstant(string, length);
  // <real_constant>
  
  else
    return (double)ParseIntegerValue(string, length);
  // <integer_value>
}

//------------------------------------------------------------------------
// <integer_value>
// -----------------------------------------------------------------

long Parser::ParseIntegerValue(charP string, int length)
{
#ifdef __OUTSTRING__
  Out("ParseIntegerValue:",
      string,length);
#endif

  int endpos;
  
  if (!string || !*string || (length<=0) || length>(int)strlen(string))
    {
      error= true;
      ParserError("missing <integer_value>!",
		  "", 0);
      return 0;
      // error, <integer_value> missing
    }
  else
    endpos= length-1;
  // NULL or "" <-- no expression or wrong value
  
  if (isalpha(string[0]))
    {
      IsMDArcMult= true;
      IsMDDelayPar= true;
      
      return ParseIntegerParameter(string, length);
    }
  // <integer_parameter>

  else if (type!=param_def && string[0]=='#')
    {
      IsMDDelay= true;
      IsMDArcMult= true;

      return ParseMarking(string,length);
    }
  // <marking> (ausser Parameterdefinition)

  else
    return ParseIntegerConstant(string, length);
  // <integer_constant>
}

//------------------------------------------------------------------
// <real_constant>
// -----------------------------------------------------------------

double Parser::ParseRealConstant(charP string, int length)
{
#ifdef __OUTSTRING__
  Out("ParseRealConstant:",
      string,length);
#endif

  if (!string || !*string || (length<=0) || length>(int)strlen(string))
    {
      error= true;
      ParserError("missing <real_constant>!",
		  "", 0);
      return 0.0;
      // error, <real_constant> missing
    }
  // NULL or "" <-- no expression or wrong value

  bool point_found= false;
  bool e_found= false;
  for (int i=0; i<length; i++)
    {
     if (string[i]=='.')
       if (point_found || e_found)
	 {
	   error= true;
	   break;
	 }// only one point allowed, in front of exponent sign
	else 
	  {
	    point_found= true;
	    continue;
	  }
     else if ((string[i]=='E') || (string[i]=='e'))
       if (e_found || (i==0) || (i+1==length))
	 {
	   error= true;
	   break;
	 }// only one exponent sign allowed, not at the beginning or the end
       else 
	 {
	   e_found= true;
	   continue;
	 }
     else if ((string[i]=='+')  || (string[i]=='-'))
       if ((i==0) && (string[i]=='+'))
	 {
	   error= true;
	   break;
	 }// negative sign only in front of mantisse allowed
     
       else if ( (i-1 >=0) &&
		(string[i-1]!='E') && (string[i-1]!='e') )
	 {
	   error= true;
	   break;
	 }// or sign only after exponent sign allowed
       else
	 continue;
     
     else if (!isdigit(string[i]))
       {
	 error= true;
	 break;
       }// the rest must be digits
   }
  if (!point_found && !e_found)
    error= true;

  if (error)
    {
      error= true;
      ParserError("illegal <real_constant>:",
		  string, length);
      return 0.0;
    }
  
  return atof(string);
}

// -----------------------------------------------------------------
// <rew_item>
// -----------------------------------------------------------------

double Parser::ParseRewItem(charP string, int length, bool isP)
{
// Im string steht nur der Inhalt von "P{"..."}" | "E{"..."}"

#ifdef __OUTSTRING__
  Out("ParseRewItem:",
      string,length);
#endif

  if (!string || !*string || (length<=0) || length>(int)strlen(string))
    {
      error= true;
      ParserError("missing <rew_item>!",
		  "", 0);
      return 0.0;
      // error, <rew_item> missing
    }
  // NULL or "" <-- no reward item or wrong value

  int i;
  for (i=0; i<length; i++)
    if ((string[i]=='I') && (i+1<length) && (string[i+1]=='F')) break;
  // ...IF...
  //    ^

  if (i<length)
    if (isP)
      {
	ParseLogicCondition(string+i+2,length-i-2);
	ParseLogicCondition(string,i);
	// "P{" <logic_condition> "IF" <logic_condition> "}"
      }
    else
      {
	ParseLogicCondition(string+i+2,length-i-2);
	ParseMarkFunc(string,i);
	// "E{" <mark_func> "IF" <logic_condition> "}"
      }
  else
    if (isP)
      {
	ParseLogicCondition(string,length);
	// "P{" <logic_condition> "}"
      }
    else
      {
	ParseMarkFunc(string,length);
	// "E{" <mark_func> "}"
      }

  return 1.0;
  // do not evaluate, check syntax only (nur fuer <reward_def>)
}

// -----------------------------------------------------------------
// <integer_constant>
// -----------------------------------------------------------------

long Parser::ParseIntegerConstant(charP string, int length)
{
#ifdef __OUTSTRING__
  Out("ParseIntegerConstant:",
      string,length);
#endif

  int endpos= length-1;

  if (!string || !*string || (length<=0) || length>(int)strlen(string))
    {
      error= true;
      ParserError("missing <integer_constant>!",
		  "", 0);
      return 0;
      // error, <integer_constant> missing
    }
  // NULL or "" <-- no integer constant or wrong value

  for (int i= endpos; i>=0; i--)
    if (!isdigit(string[i]))
      {
	error= true;
	ParserError("illegal <integer_constant>:",
		    string, length);
	return 0;
      }
  return atol(string);
}

//------------------------------------------------------------------------
// <logic_condition>
//------------------------------------------------------------------------

bool Parser::ParseLogicCondition(charP string, int length)
{
#ifdef __OUTSTRING__
  Out("ParseLogicCondition:",
      string,length);
#endif

  int endpos;

  if (!string || !*string || (length<=0) || length>(int)strlen(string))
    {
      error= true;
      ParserError("missing <logic_condition>!",
		  "", 0);
      return true;
      // error, <logic_condition> missing
    }
  else
    endpos= length-1;
  // NULL or "" <-- no logic condition or wrong value

  int index= SearchLowestLOperator(string, endpos);
  if (error)
    return false;

  if (index==-1)
    {// no operator found
      return ((!strncmp(string,"NOT",3) && (length>3))?
	      !ParseLogicCondition(string+3, length-3):
	      // "NOT" <logic_condition>

	      ((string[0]=='(') && (string[endpos]==')'))? 
	      ParseLogicCondition(string+1, length-2): 
	      // "(" <logic_condition> ")"

	      ParseComparison(string, length)
	      // <comparison>
	      );
    }

  // <logic_condition> "OR"  <logic_condition> |
  // <logic_condition> "AND" <logic_condition>

  char numop= string[index];
  
  int i;
  (numop=='O')? i=2: i=3;
  // i is length of operator ("OR" | "AND")
  
  switch (numop)
    {
    case 'A':
      return (ParseLogicCondition(string,index) &
	      ParseLogicCondition(string+index+i,endpos-(index+i)+1));
      // <logic_condition> "AND" <logic_condition>

      break;
    case 'O':
      return (ParseLogicCondition(string,index) |
	      ParseLogicCondition(string+index+i,endpos-(index+i)+1));
      // <logic_condition> "OR" <logic_condition>

      break;
    }            

  return false;  // to keep the compiler happy!!!
}

//------------------------------------------------------------------------
// <comparison>
//------------------------------------------------------------------------

bool Parser::ParseComparison(charP string, int length)
{
  int endpos;

  if (!string || !*string || (length<=0) || length>(int)strlen(string))
    {
      error= true;
      ParserError("missing <comparison>!",
		  "", 0);
      return true;
      // error, <comparison> missing
    }
  else
    endpos= length-1;
  // NULL or "" <-- no logic condition or wrong value

  for (int i=0; i<length; i++)
    {
      if (string[i]=='=')
	return (ParseMarkFunc(string,i) == ParseMarkFunc(string+i+1,endpos-i));
      // <mark_func> "=" <mark_func>
      
      else if ((string[i]=='/') && ((i+1 <= endpos) && string[i+1]=='='))
	return (ParseMarkFunc(string,i) != ParseMarkFunc(string+i+2,endpos-i-1));
      // <mark_func> "/=" <mark_func>
      
      else if (string[i]=='>')
	if ((i+1 <= endpos) && string[i+1]=='=')
	  return (ParseMarkFunc(string,i) >= ParseMarkFunc(string+i+2,endpos-i-1));
	else
	  return (ParseMarkFunc(string,i) > ParseMarkFunc(string+i+1,endpos-i));
      // <mark_func> ">=" <mark_func> | 
      // <mark_func> ">" <mark_func>
      
      else if (string[i]=='<')
	if ((i+1 <= endpos) && string[i+1]=='=')
	  return (ParseMarkFunc(string,i) <= ParseMarkFunc(string+i+2,endpos-i-1));
        else
	  return (ParseMarkFunc(string,i) < ParseMarkFunc(string+i+1,endpos-i));
      // <mark_func> "<=" <mark_func> | 
      // <mark_func> "<" <mark_func>
    }
  
  error= true;
  ParserError("no <comp_oper> in <comparison>:",
	      string, length);
  return false;
}

//------------------------------------------------------------------------
// <marc_func>
//------------------------------------------------------------------------

long Parser::ParseMarkFunc(charP string, int length)
{
#ifdef __OUTSTRING__
  Out("ParseMarkFunc:",
      string,length);
#endif

  int   endpos;

  if (!string || !*string || (length<=0) || length>(int)strlen(string))
    {
      error= true;
      ParserError("missing <mark_func>!",
		  "", 0);
      return 0;
      // error, <mark_func> missing
    }
  else
    endpos= length-1;
  // NULL or "" <-- no MarkFunc or wrong value

  int index= SearchLowestOperator(string, endpos);
  if (error)
    return 0;

  if (index==-1)
    {// no operator found
      return (((string[0]=='(') && (string[endpos]==')'))? 
	      ParseMarkFunc(string+1, length-2):
	      // "(" <mark_func> ")"
 
	      ParseIntegerValue(string, length)
	      // <integer_value>
	      );
    }

  // <mark_func> numop <mark_func>

  char numop= string[index];

  long value;
  switch (numop)
    {
    case '+':
      return (ParseMarkFunc(string,index) + 
	      ParseMarkFunc(string+index+1,endpos-index));
      // <mark_func> "+" <mark_func>

      break;
    case '-':
      return (ParseMarkFunc(string,index) - 
	      ParseMarkFunc(string+index+1,endpos-index));
      // <mark_func> "-" <mark_func>

      break;
    case '*':
      return (ParseMarkFunc(string,index) * 
	      ParseMarkFunc(string+index+1,endpos-index));
      // <mark_func> "*" <mark_func>

      break;
    case '/':
      value= ParseMarkFunc(string+index+1,endpos-index);
      
      if (type!=reward_def)
	{
	  if (value==0)
	    {
	      error= true;
	      ParserError("division by 0 error:",
			  string, length);
	      return 0;
	      // error, division by 0
	    }
	}
      else value= 1;
      // do not evaluate, check syntax only (nur fuer <reward_def>)

      return (ParseMarkFunc(string,index) / 
	      value);
      // <mark_func> "/" <mark_func>

      break;
    case '^':
      return ((long)pow(ParseMarkFunc(string,index), 
			ParseMarkFunc(string+index+1,endpos-index)));
      // <mark_func> "^" <mark_func>

      break;
    }            

  return 0;  // to keep the compiler happy!!!
}

//------------------------------------------------------------------------
//------------------------------------------------------------------------
//------------------------------------------------------------------------
// AH NEW: some new functions to support decomposition.cc
// the functions are used to assemble und deassemble HCPN-transitions.
//------------------------------------------------------------------------

// read out exponential delay of transition node trans
double Parser::ReadExpDelay(Node *trans, unsigned int bindnr)
{
  int type;
  Text global_guard, selcol, priority, server_type, distr;
  List *bindlist; 
  Text unused;
  Text newbind, bind, placename, attributes;
  Text localCondition, weight, delay;
  List inArcs, outArcs;
  int delaytype;

  if(trans->Class()->Name()!="TimedTrans")
    {
      error= true;
      Text errortext = Text("ReadExpDelay: node") + trans->Name()
	+ Text(" is no timed transition, but ")
	+ trans->Class()->Name();
      ParserError(errortext.to_charPtr(), "", 0);
      return 0;
    }

  TokenizeTransition(trans, type, global_guard, bindlist, unused,
		     selcol, priority, server_type, distr);

  if(!(bindlist->Length()>(int)bindnr))
    {
      error= true;
      Text errortext = Text("ReadExpDelay: Transition ") + trans->Name()
	+ Text(" has no binding nr. ") + Float(bindnr);
      ParserError(errortext.to_charPtr(), "", 0);
      return 0;
    }

  bind = bindlist->GetEntry(bindnr)->to_charPtr(); 
  TokenizeBinding(bind, localCondition, inArcs, outArcs, weight,
		  delaytype, delay);

  if(delaytype != D_EXPONENTIAL)
    {
      error= true;
      Text errortext = "ReadExpDelay: no exponential delay.";
      ParserError(errortext.to_charPtr(), "", 0);
      return 0;
    }

  return ParseRealConstant(delay.to_charPtr(), delay.Length());
}

//------------------------------------------------------------------------
// tokenize HCPN-transition trans  
void Parser::TokenizeTransition(Node *trans, int &type, Text &global_guard, List* &bindlist, Text &unused,
				Text &selcol,
				Text &priority,
				Text &server_type, Text &distr) 
{
  if(trans->Class()->Name()=="TimedTrans")
    type = N_TIMED;
  else if (trans->Class()->Name()=="ImmTrans")
    type = N_IMMEDIATE;
  else
    {
      error= true;
      Text errortext = Text("TokenizeTransition: node given is of type ")
	+ trans->Class()->Name();
      ParserError(errortext.to_charPtr(), "", 0);
      return;
    }
  bindlist = (List*)(trans->Attr("bindings"))->Copy(); //binding list of the transition
  global_guard = (*trans->Attr("global_guard")).to_charPtr();
  unused = (*trans->Attr("unused")).to_charPtr();
  selcol = (*trans->Attr("selcol")).to_charPtr();

  if(type == N_IMMEDIATE)
    {
      priority = (*trans->Attr("priority")).to_charPtr();
      server_type="";
      distr="";
    }
  else 
    {
      priority ="";
      server_type = (*trans->Attr("server_type")).to_charPtr();
      distr = (*trans->Attr("distr")).to_charPtr();
    }
}
//------------------------------------------------------------------------
// assemble transition trans from the given arguments. the old transition
// will be overwritten.
void Parser::AssembleTransition(Node *trans, int type, Text global_guard, List* bindlist, Text unused,
			Text selcol,
			Text priority,
			Text server_type, Text distr)
{
  Text attributes;
  attributes= "{ attributes = { global_guard = ";
  attributes+="\"";
  attributes+= global_guard;
  attributes+="\"";
 
  if (type==N_IMMEDIATE)
    {
      attributes+= "; priority = \"";
      attributes+=  priority + "\"";
    }

  attributes+= "; bindings = { ";
  int no_of_binds = bindlist->Length();
  for(int l=0; l<no_of_binds; l++)
    {
      attributes+="\"";
      attributes+=(*(bindlist->GetEntry(l))).to_charPtr();
      if(l+1==no_of_binds)
	attributes+="\"";
      else
	attributes+="\", ";
    }
  attributes+= "};";

  if (type==N_TIMED)
    {
      attributes+=" server_type = ";
      attributes+=server_type+";";
      attributes+=" distr = ";
      attributes+=distr+";";
    }

  attributes+= " unused = ";
  attributes+=unused+";";
  attributes+=" selcol = \"";
  attributes+=selcol+"\";};}";

  istrstream is(attributes.to_charPtr(),10000);
  is>>*trans;  
}
//------------------------------------------------------------------------
// tokenize a HCPN-binding
void Parser::TokenizeBinding(Text bind, Text &localCondition, List &inArcs, 
		     List &outArcs, Text &weight, int &delaytype, Text &delay)
{ 
  Text restbind;
  Text token;
  Text inarc, outarc;

  //read local condition
  localCondition=bind.FirstPathPart(';'); //anything left from first arc definition
  localCondition =  localCondition.BetweenC('(',')'); //get condition enclosed in brackets;
  if(localCondition.to_bool())
    restbind =  bind - "(" - localCondition.to_charPtr() -")";
  else
    restbind =  bind;

  //read arcs
  while(true)
    {
      token = restbind.FirstPathPart(','); //get first arc def
      if(!token.to_bool()) break;
      restbind = restbind - token.to_charPtr() - ",";
      token = token + "##";
      if ((inarc = token.BetweenS("in#","##")).to_bool())
	inArcs.Append(new Text(inarc));
      else if ((outarc = token.BetweenS("out#","##")).to_bool())
	outArcs.Append(new Text(outarc));
    }
  token = restbind.FirstPathPart(';'); //get last arc def
  restbind = restbind - token.to_charPtr() -";";
  token = token + "##";
  if ((inarc = token.BetweenS("in#","##")).to_bool())
    inArcs.Append(new Text(inarc));
  else if ((outarc = token.BetweenS("out#","##")).to_bool())
    outArcs.Append(new Text(outarc));

  //read weight
  weight = restbind.FirstPathPart(';');
  restbind = restbind - weight.to_charPtr() -";";

  //read delay;
  restbind = restbind + "##";
  if((delay = restbind.BetweenS("Immediate","##")).to_bool())
    delaytype = D_IMMEDIATE;
  else if((delay = restbind.BetweenS("Exponential","##")).to_bool())
    delaytype = D_EXPONENTIAL;
  else if((delay = restbind.BetweenS("Deterministic","##")).to_bool())
    delaytype = D_DETERMINISTIC;
  else if((delay = restbind.BetweenS("Uniform","##")).to_bool())
    delaytype = D_UNIFORM;
  else if((delay = restbind.BetweenS("General","##")).to_bool())
    delaytype = D_GENERAL;

  delay = delay.BetweenC('(',')');
  return;
}
//------------------------------------------------------------------------
//assemble HCPN-binding
void Parser::AssembleBinding(Text &bind, Text localCondition, List inArcs, 
		     List outArcs, Text weight, int delaytype, Text delay)
{ 
  Text restbind;
  Text token;
  Text inarc, outarc;

  //build new binding, local condition
  if(localCondition.to_bool() && localCondition!="")
    { 
      bind="(";
      bind+=localCondition+") ";
    }
  else //no local condition
    bind ="";

  bool first=true;
  //inarcs
  for(int j=0; j< inArcs.Length(); j++)
    {
      if(!first) bind+=", ";
      first=false;
      bind+="in#";
      bind+=inArcs.GetEntry(j)->to_charPtr();	
    }

  //outarcs
  for(int j=0; j< outArcs.Length(); j++)
    {
      if(!first) bind+=", ";
      first=false;
      bind+="out#";
      bind+=outArcs.GetEntry(j)->to_charPtr();	
    }
  
  //weight
  bind +=";";
  bind += weight;
  bind+="; ";

  //delay
  switch(delaytype)
    {
    case D_IMMEDIATE:
      bind+="Immediate"; break;
    case D_EXPONENTIAL:
      bind+="Exponential"; break;
    case D_DETERMINISTIC:
      bind+="Deterministic"; break;
    case D_UNIFORM:
     bind+="Uniform"; break;
    case D_GENERAL:
      bind+="General"; break;
    }

  bind+="(";
  bind+=delay;
  bind+=")";

  return;
}
