#ifndef PARSER_H
#define PARSER_H
// -----------------------------------------------------------------
// Importe
// -----------------------------------------------------------------

#include <iostream.h>
// fuer Ausgaben

#include <string.h>
// fuer Strings

#include <math.h>
// fuer pow()

#include <ctype.h>
// fuer isdigit()

#include <stdlib.h>
// fuer atol() und atof()
//AH new:
#include "node.h"
#include "list.h"
#include "text.h"

// -----------------------------------------------------------------
// Definitionen
// -----------------------------------------------------------------

//AH New: delay type class. of transitions
#define D_IMMEDIATE 0
#define D_EXPONENTIAL 1
#define D_DETERMINISTIC  2
#define D_UNIFORM 4
#define D_GENERAL 5
//Agnes node class. of transitions
#define N_TIMED    0
#define N_IMMEDIATE 1

#define charP char*
// Pointerabkuerzung

enum parser_type {
  none, md_exp_delay, md_det_delay, md_weight, 
  md_enable, md_arc_mult, reward_def, param_def, pmf_def};
// alle Parser-Typen

// -----------------------------------------------------------------
// Parser Klassendeklaration
// -----------------------------------------------------------------

class Parser
{
// -----------------------------------------------------------------
// Hilfsfunktionen
// -----------------------------------------------------------------

 private:

  void DestroyBlancs(charP string);
  int  SearchOpeningBracket(charP string, int endpos);
  int  SearchOpeningBracket2(charP string, int endpos);
  int  SearchLOperator(charP string, int endpos);
  int  SearchOperator(charP string, int endpos);
  int  SearchLowestLOperator(charP string, int endpos);
  int  SearchLowestOperator(charP string, int endpos);
  bool IsSignOfExponent(charP string, int endpos);
  void Out(const charP message, charP string, int length);

  double h_Parse_md_weight(charP string) 
    {
      if (!string || !*string)
	return 0.0;
      
      int length= (int)strlen(string);
      if (string[length-1]==';')
	length--;
      
      return (ParseIfExpression (string, length));
    };

  bool h_Parse_md_enable(charP string)
    {
      if (!string || !*string)
	return true;
      
      int length= (int)strlen(string);
      if (string[length-1]==';')
	length--;
      
      return (ParseLogicCondition (string, length));
    };
  
  double h_Parse_param_def(charP string)
    {
      if (!string || !*string)
	return 0.0;
      
      int length= (int)strlen(string);
      if (string[length-1]==';')
	length--;
      
      return (ParseExpression(string, length));
    };

  void   h_Parse_pmf_def(charP string)
    {
      if (!string || !*string)
	return;
      
      int length= (int)strlen(string);
      if (string[length-1]==';')
	length--;
      
      ParsePMFDefinition(string, length);
    };

 public:

  //---------------------------------------------------------------------
  // Folgende Attribute koennen nach dem Parsen gelesen werden.
  //---------------------------------------------------------------------

  enum parser_type type;
  // Art des zu parsenden Ausdrucks

  bool error;
  // Trat ein Fehler auf,
  // dann ist das Ergebnis ungueltig

  bool IsMDDelay;
  // <constant> | <delayparameter> | sonst
  // false,       false,             true
  // war <md_exp_delay> | <md_det_delay> | <md_weight> markierungsabhaengig?

  bool IsMDArcMult;
  // <constant> | sonst
  // false,       true
  // war <md_arc_mult> markierungsabhaengig?

  bool IsMDDelayPar;
  // <constant> | sonst
  // false,       true
  // war <param_def> markierungsabhaengig?
  //---------------------------------------------------------------------
  // Parser fuer die Syntax von TimeNET Definitionen.
  // Folgende Funktionen stehen zum Parsen zur Verfuegung
  // und erwarten den String als Parameter.
  // Saemtliche Funktionen parsen nach festgelegter EBNF fuer das 
  // TimeNET Dateiformat "<model>.TN".
  //---------------------------------------------------------------------

  double Parse_md_exp_delay(charP string) 
    {type= md_exp_delay;IsMDDelay= false;error= false;
     DestroyBlancs(string); return h_Parse_md_weight(string);};
  // <md_exp_delay>

  double Parse_md_det_delay(charP string) 
    {type= md_det_delay;IsMDDelay= false;error= false;
     DestroyBlancs(string); return h_Parse_md_weight(string);};
  // <md_det_delay>

  double Parse_md_weight(charP string) 
    {type= md_weight;IsMDDelay= false;error= false;
     DestroyBlancs(string); return h_Parse_md_weight(string);};
  // <md_weight>

  double Parse_md_enable(charP string) 
    {type= md_enable;error= false;
     DestroyBlancs(string); return h_Parse_md_enable(string);};
  // <md_enable>

  double Parse_md_arc_mult(charP string) 
    {type= md_arc_mult;IsMDArcMult= false;error= false;
     DestroyBlancs(string); return h_Parse_md_weight(string);};
  // <md_arc_mult>

  double Parse_reward_def(charP string) 
    {type= reward_def;error= false;
     DestroyBlancs(string); return h_Parse_param_def(string);}
  // <md_reward_def>

  double Parse_param_def(charP string) 
    {type= param_def;IsMDDelayPar= false;error= false;
     DestroyBlancs(string); return h_Parse_param_def(string);}
  // <md_param_def>

  void   Parse_pmf_def(charP string) 
    {type= pmf_def;error= false;
     DestroyBlancs(string); h_Parse_pmf_def(string);}
  // <md_pmf_def>

  //---------------------------------------------------------------------
  // Folgende Funktionen stehen auch zum Parsen zur Verfuegung
  // und erwarten den String und seine Laenge als Parameter.
  // ParseRewItem testet nur die Syntax, ebenso die void-Funktionen.
  // type entsprechend setzen, da die Syntax kontextabhaengig ist.
  // error und IsMD mit false initialisieren.
  // Der String darf keine Leerzeichen enthalten.
  //---------------------------------------------------------------------

  void ParsePMFDefinition(charP string, int length);
  // <pmf_definition>

  void ParsePMFExpression(charP string, int length);
  // <pmf_expression>

  void ParseImpulse(charP string, int length);
  // <impulse>

  void ParseRectangle(charP string, int length);
  // <rectangle>

  void ParseExpolynomial(charP string, int length);
  // <expolynomial>

  double ParseIfExpression(charP string, int length);
  // <if_expr>

  double ParseExpression(charP string, int length);
  // <expression>

  double ParseRealValue(charP string, int length);
  // <real_value>
  
  double ParseRealConstant(charP string, int length);
  // <real_constant>

  double ParseRewItem(charP string, int length, bool isP);
  // <rew_item> (nur der Inhalt von "P{" ... "}"  | "E{" ... "}")

  bool ParseLogicCondition(charP string, int length);
  // <logic_condition>

  bool ParseComparison(charP string, int length);
  // <comparison>

  long ParseMarkFunc(charP string, int length);
  // <mark_func>

  long ParseIntegerValue(charP string, int length);
  // <integer_value>

  long ParseIntegerConstant(charP string, int length);
  // <integer_constant>

  //---------------------------------------------------------------------
  // AH NEW: these functions give some rough support for assembling
  // and disassembling of HCPNtransitions. 
  //---------------------------------------------------------------------
  double ReadExpDelay(Node *trans, unsigned int bindnr);
  void TokenizeTransition(Node *trans, int &type, Text &global_guard,
			  List* &bindlist, Text &unused,
			  Text &selcol,	Text &priority,
			  Text &server_type, Text &distr);

  void AssembleTransition(Node *trans, int type, Text global_guard,
			  List* bindlist, Text unused,
			  Text selcol,Text priority,
			  Text server_type, Text distr);

  void TokenizeBinding(Text bind, Text &localCondition, 
		       List &inArcs, List &outArcs, 
		       Text &weight, int &delaytype, Text &delay);
  
  void AssembleBinding(Text &bind, Text localCondition,
		       List inArcs, List outArcs,
		       Text weight, int delaytype, Text delay);
  
  //---------------------------------------------------------------------
  // Folgende Funktionen muessen selber definiert werden und
  // werden vom Parser im Bedarfsfall aufgerufen. 
  //---------------------------------------------------------------------

  virtual void ParserError(const charP, charP, int) {};
  // error
  // Input:  Fehlernachricht, Name und Laenge des fehlerhaften Strings
  // Output: eine angemessene Ausgabe

  virtual long ParseMarking(charP , int ) {return 1;};
  // <marking>
  // Input:  Name und Laenge einer Stelle ("#P1", 3)
  // Output: Markierung der Stelle: P1

  virtual double ParseParameter(charP , int ) {return 1.0;};
  // <real_parameter> | <integer_parameter>
  // Input:  Name und Laenge eines Parameter ("paramdef", 8)
  // Output: Aufruf von einer der beiden folgenden Funktionen,
  //         je nach Typ des Parameter

  virtual double ParseRealParameter(charP , int ) {return 1.0;};
  // <real_parameter>
  // Input:  Name und Laenge eines RealParameter ("rparamdef", 9)
  // Output: ausgewerteter Inhalt des <real_parameter>

  virtual long ParseIntegerParameter(charP , int ) {return 1;};
  // <integer_parameter>
  // Input:  Name und Laenge eines IntegerParameter ("iparamdef", 9)
  // Output: ausgewerteter Inhalt des <integer_parameter>
};
#endif
