// -----------------------------------------------------------------
// Importe
// -----------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>

#include "xbefore.h"
#include <Xm/ToggleB.h>
#include <Xm/PushB.h>
#include "xafter.h"
// Fuer das X-Toolkit

#include "linked.h"
#include "commmod.h"
#include "module.h"
#include "main.h"
#include "editor.h"
#include "net.h"
#include "node.h"
#include "netcl.h"
#include "dialog.h"
#include "wrapper.h"
#include "list.h"
#include "parser.h"

// -----------------------------------------------------------------
// Definitionen
// -----------------------------------------------------------------

#define UPDATEINTERVALL 500
// Blinkfrequenz in ms

#define ENABLECOLOR "yellow"
#define ACTIVECOLOR "green"
// Farben des Tokengame

#define intP  int*
#define charP char*
// Pointerabkuerzung

// -----------------------------------------------------------------
// verwendete Typen
// -----------------------------------------------------------------

enum DoType {BACKWARD,FORWARD};

// -----------------------------------------------------------------
// Prototypen
// -----------------------------------------------------------------

class FireItem;
class FiredItem;
class Tokengame;

typedef FireItem*  FireItemP;
typedef FiredItem* FiredItemP;
typedef Tokengame* TokengameP;

Text ConstructName (NodeP node);

// -----------------------------------------------------------------
// Klassendeklarationen
// -----------------------------------------------------------------

class FireItem
{
public:

  NodeP node;                  
  // Zeiger auf den Knoten der schaltfaehigen Transition

  LinkedList<int*> Bindings;  
  // Index-Liste schaltfaehiger bindings der Transition

  FireItem(NodeP node) {this->node= node;};
  ~FireItem() {this->Bindings.Clear();};

  bool operator==(NodeP node) {return (this->node==node);};
  bool operator!=(NodeP node) {return (this->node!=node);};
  // Ist dieser node ein feuerbereites Item?

  void Fire(int* BIndex, TokengameP game);
  // Feuern des Fireitems
};

// -----------------------------------------------------------------

class FiredItem
{
public:

  NodeP node;                  
  // Knoten der abgefeuerten Transition
  int   binding;               
  // Index des Bindings der abgefeuerten Trans

  FiredItem(NodeP node, int binding) 
    {
      this->node= node;
      this->binding= binding;
    };

  ~FiredItem() {};

  void Do(DoType Type, TokengameP game);
  // Backward/Forward fuer abgefeuerte Transitionen
};

// -----------------------------------------------------------------

class Tokengame: virtual public CommModule, public Wrapper<Tokengame,bool>
{
protected:

  friend class FireItem;
  friend class FiredItem;
  
  bool TokenGameEnable;
  // Tokengame aktiviert?

  bool TimerEnable;
  // Timerfunktion erneut aufrufen?

  bool ShowHide;
  // Status des Blinkens

  int  RecursiveCnt;
  // fuer RestoreMarking

  EditorP ed;
  // In diesem Editor laeuft das Tokengame

  WrapperID   FlashID, CallBackID;
  Widget      MenuWidget, ToggleWidget;
  // id fuer FlashTimer, id&widget vom Menu, widget von ToggleButton

  LinkedList<TextP> InitialMarking;
  // Liste mit initialer Markierung

  LinkedList<FireItemP> ReadyToFire;
  // Liste mit schaltfaehigen Items

  LinkedList<FiredItemP> Fired;
  int FiredCount, FiredCount_save;
  // Liste mit geschalteten Items und pointer

  bool IsReadyToFire(NodeP node);
  // ist Transition feuerbereit?

  void SaveMarking(PageP pg= 0);
  // nachher: FiredCoun_savet= FiredCount;
  void RestoreMarking(PageP pg= 0); 
  // vorher: RecursiveCount= 0;, nachher: FiredCount= FiredCount_save;

  bool  FlashTimer(bool);

  void SelectTrans(NodeP node);
  void DeselectTrans(NodeP node);
  void DeselectAllTrans(PageP pg= 0);
  void DeselectColor(PageP pg= 0);
// -----------------------------------------------------------------
public:

  Tokengame(Net& n);
  ~Tokengame();
 
  void InitTokengame();
  void DeInitTokengame(bool restore);

  void SetModified() {net.modified= true;};

  virtual bool  Init() {return true;}; 
  virtual bool  Open(Editor&,Page&);
  virtual bool  Change(Editor&,Page&,Page&) {return false;};
  virtual bool  Menu(Editor&,Page&,Ident&,bool);
  virtual bool  Event(Editor&,Page&,EventInfo&,bool);
  // Modulfunktionen

          charP FileName() {return net.FileName();};

  virtual bool  IsPlace(NodeP) {return false;};
  virtual bool  IsTransition(NodeP) {return false;};
  //AH Changed Start => (TextP)
  virtual TextP GetColor(NodeP) {return (TextP)NULL;};
  virtual TextP GetMarking(NodeP) {return (TextP)NULL;};
  //AH Changed End

  virtual void  StartTokengame() {};
  virtual void  EndTokengame() {};

  virtual bool  EventEnable() {return true;};
  virtual void  OpenOptionsMenu() {};

  virtual bool  SelectEnabledTrans(PageP) {return true;};
  virtual bool  DeselectEnabledTrans(PageP) {return true;};

  //Menu uberwachen, damit kein anderes Modul aufgerufen werden kann
  virtual bool MenuLock() {return true;};

  virtual void  OpenFireMenu(Node&) {};
  virtual void  FireTrans(NodeP, int, bool) {};
  // Tokengame Funktionen

  virtual bool DoPeriodic() {return TokenGameEnable;};
};

