#include "xbefore.h"
#include <Xm/Text.h>
#include <Xm/ScrolledW.h>
#include <Xm/TextF.h>
#include "xafter.h"

extern "C" {
#include <string.h>
#include <ctype.h>
}

#include "code.h"

VALUE_NOTIFY("code",Code)


Code::Code(NetClass* nc, bool hidden):
  Text(nc,hidden),    
  Wrapper<Code,int>(this)
{
}


Code::Code(const Code& t):
  Text(t),
  Wrapper<Code,int>(this)
{
}


Code::Code(const char *s):
  Text(s),
  Wrapper<Code,int>(this)
{
}


Widget Code::CreateWidget(Widget parent)
{
	if (rows>1)
	{
		scroll = XtVaCreateManagedWidget("Text",
			xmScrolledWindowWidgetClass,parent,NULL);
	
		text = XtVaCreateManagedWidget("Scrolled",
			xmTextWidgetClass,scroll,
			XmNeditMode,XmMULTI_LINE_EDIT,
			XmNrows,rows,NULL);
 		//AH Changed Start Highlight Validate => &Highlight &Validate
 		AddCB(text,XmNvalueChangedCallback,&Code::Highlight,0);
 		AddCB(text,XmNmodifyVerifyCallback,&Code::Validate,0);
 		//AH Changed End
		CatchDestroy(&scroll,&text);
		//AH Changed const_cast,char* added
		XmTextSetString(text,const_cast<char*>(Sure(p)));
		return scroll;
	}
	else
	{
		text = XtVaCreateManagedWidget("Text",
			xmTextFieldWidgetClass,parent,NULL);
 		//AH Changed Start Highlight Validate => &Highlight &Validate
 		AddCB(text,XmNvalueChangedCallback,&Code::Highlight,0);
 		AddCB(text,XmNmodifyVerifyCallback,&Code::Validate,0);
 		//AH Changed End
		CatchDestroy(&text);
		//AH Changed const_cast,char* added
		XmTextSetString(text,const_cast<char*>(Sure(p)));
		return text;
	}
}


void Code::Validate(Widget, int, XtPointer calldata)
{
	XmTextVerifyCallbackStruct* tvcs = (XmTextVerifyCallbackStruct*)calldata;
	if (tvcs->text->length)
	{
	  // Tabs werden durch Anzahl von TABWITH Leerzeichen ersetzt!
	  char* p = tvcs->text->ptr;
	  if (*p=='\t') {
	    char* tab_text = new char[TABWIDTH];
	    int n;
	    for (n=0; n<TABWIDTH; ++n)
	      tab_text[n] = ' ';
	    tvcs->text->ptr = tab_text;
	    tvcs->text->length = TABWIDTH;
	  }
	}
}


void Code::Highlight(Widget, int, XtPointer)
{
  //cout<<"Highlight\n";

  char* string = XmTextGetString(text);
  char* p;
  char* token;
  char* com_beg;
  char* com_end;
  char vorher, nachher;  // ueberpruefen die Zeichen vor bzw. nach dem keyword
                         //  damit 'array1' oder 'e6module' nicht gehighlightet werden

  if(strlen(string))
    XmTextSetHighlight(text,0,strlen(string),XmHIGHLIGHT_NORMAL);

  if(strlen(string)>1) {
    p = string;
    while (*p) {
    search:while (!isalpha(*p)&&*p&&!(*p=='('&&*(p+1)=='*')) ++p; // Kommentar oder Token
      if (*p=='('&&*(p+1)=='*') { // Kommentarbeginn erkannt
	com_beg = p;
	p += 2;
	while (*p&&!(*p=='*'&&*(p+1)==')')) ++p; // Kommentarende suchen
	if (*p)
	  p+=2;
	com_end = p;
	XmTextSetHighlight(text,com_beg-string,com_end-string,XmHIGHLIGHT_SELECTED);
	goto search;
      }
      if (p > string)
	vorher= *(p-1);
      else
	vorher= *p;
      token = p;
      // Hier kommt garantiert ein Buchstabe oder gar nichts!
      for (;isalpha(*p)||(*p=='_')||(*p=='.');++p) 
	if (*p=='p' || *p=='t') { // Muss leider sein bei nicht reinen Alpha-Keywords
	  if (!strncmp(p,"p->t",4)) p += 3;
	  else if (!strncmp(p,"p<-t",4)) p += 3;
	  else if (!strncmp(p,"p-ot",4)) p += 3;
	  else if (!strncmp(p,"t->p",4)) p += 3;
	  else if (!strncmp(p,"t<-p",4)) p += 3;
	  else if (!strncmp(p,"to-p",4)) p += 3;
	}
      nachher= *p;
      if (!isdigit(vorher) && !isdigit(nachher) && (token!=p?spnl.IsKey(token,p-token):0))
	XmTextSetHighlight(text,token-string,p-string,XmHIGHLIGHT_SECONDARY_SELECTED);
    }
  }
  XtFree(string);
}


