#include "context.h"
#include "scope.h"
#include "text.h"
#include "ident.h"
#include "value.h"
#include "tool.h"

#include <iostream.h>

extern "C" {
#include <string.h>
}

LinkedList<void*> Context::contexts;

Context::Context(Context* s):
	super(s),
	enviter(envlist),
	gciter(gclist),
	draw_xor(false),
	color("black"),
	font("fixed"),
	relative(-1),
	cpos(50),
	spline(0),
	dashed(false),
	filled(false),
	rotatable(true),
	scalable(true),
	condition(true),
	shadow(0),
	thick(1.0)
{
	contexts.Append(this);
}


Context::~Context()
{
	contexts.Remove(this);
	for (GCList* gcl=gciter; gcl; gcl=++gciter)
		XFreeGC(gcl->dpy,gcl->gc);
}

void Context::DisplayWillClose(Display* dpy)
{
	Iterator<void*> iter(contexts);
	for (void* c=iter; c; c=++iter)
		ContextP(c)->FreeDisplay(dpy);
}


void Context::FreeDisplay(Display* dpy)
{
	for (GCList* gcl=gciter; gcl; gcl=++gciter)
		if (gcl->dpy==dpy)
		{
			if(gcl->font)
				XUnloadFont(gcl->dpy,gcl->font);
			if(gcl->fontstr)
				XFreeFontInfo(NULL,gcl->fontstr,0);
			XFreeGC(gcl->dpy,gcl->gc);
			gclist.Remove(*gcl);
			return;		
		}
}


istream& Context::Input(istream& is)
{
	Ident id(false);
	Scope dp(is);

	while (is && dp.GetId(id))
	{
		Expression* e = new Expression(is);
		if (is)
			envlist.Append(EnvList(id,e));
		else
			delete e;
	}
	
	return is;
}


Value* Context::GetVar(char* name)
{
	for (EnvList* el=enviter; el; el=++enviter)
		if (el->name==name)
			return *el->expr;
	
	return super?super->GetVar(name):0;
}


void Context::Evaluate(ExprData& ed)
{
	ed.context = super;
	for (EnvList* el=enviter; el; el=++enviter)
		el->expr->Evaluate(ed);
	
	Value* val;
 	//AH Changed Start 
 	thick = ((val = GetVar("thick")))?(*val).to_double():1.0;
 	color = ((val = GetVar("color")))?(*val).to_charPtr():"black";
 	font = ((val = GetVar("font")))?(*val).to_charPtr():"fixed";
 	spline = ((val = GetVar("spline")))?Maxi(0,(*val).to_int()):0;
 	shadow = ((val = GetVar("shadow")))?(*val).to_int():0;
 	filled = ((val = GetVar("filled")))?(*val).to_int()!=0:false;
 	relative = ((val = GetVar("relative")))?Maxi(-1,(*val).to_int()):-1;
 	rotatable = ((val = GetVar("rotatable")))?(*val).to_int()!=0:true;
 	scalable = ((val = GetVar("scalable")))?(*val).to_int()!=0:true;
 	dashed = ((val = GetVar("dashed")))?(*val).to_int()!=0:false;
 	condition = ((val = GetVar("cond")))?(*val).to_int()!=0:true;
 	cpos = ((val = GetVar("cpos")))?Maxi(0,Mini(100,(*val).to_int())):100;
}


void Context::SetZoom(Display* dpy, GC gc, double zoom)
{
	XGCValues xgcv;
	xgcv.line_width = int(thick*zoom+0.5);
	XChangeGC(dpy,gc,GCLineWidth,&xgcv);
}


Context::GCList* Context::GetGCList(Display* dpy, bool x_or)
{
	GCList* gcl;
_again:
	for (gcl=gciter; gcl; gcl=++gciter)
		if (gcl->dpy==dpy)
			break;
	
	if (!gcl)
	{
		XGCValues xgcv;
		xgcv.background = AllocColor(dpy, "gainsboro");
		GC gc = XCreateGC(dpy,DefaultRootWindow(dpy),GCBackground,&xgcv);
		gclist.Append(GCList(dpy,gc));
		goto _again;
	}

	ChangeGC(gcl,draw_xor = x_or);
	return gcl;
}


GC Context::GetGC(Display* dpy, bool x_or)
{
	return GetGCList(dpy,x_or)->gc;
}


void Context::TextExtents(Display* dpy, char* text, 
	int& asc, int &desc, int& width)
{
	GCList* gcl = GetGCList(dpy,draw_xor);	
	XCharStruct xc; int d;
	XTextExtents(gcl->fontstr,text,strlen(text),&d,&asc,&desc,&xc);
	width = xc.width;
}


int Context::AllocColor(Display* dpy, char* color)
{
	XColor exact,screen;
	XAllocNamedColor(dpy,DefaultColormap(dpy,DefaultScreen(dpy)),
		color,&exact,&screen);
	return screen.pixel;
}


void Context::ChangeGC(Context::GCList* gcl, bool draw_xor)
{
	XGCValues xgcv;
	unsigned long set = GCLineWidth+GCLineStyle+GCFillStyle+GCFunction;
	
	xgcv.line_width = unsigned(thick+0.5);
	xgcv.line_style = dashed?LineOnOffDash:LineSolid;
	xgcv.fill_style = FillSolid; 
	
	xgcv.function = draw_xor?GXxor:GXcopy;
	
	if (color!=gcl->colorname)
	{
		xgcv.foreground = AllocColor(gcl->dpy,color);
		set |= GCForeground;
		gcl->colorname = color;
	}
	
	if (font!=gcl->fontname)
	{
		int ret = 0;
 		//AH Changed Start => to_charPtr()
 		char **font_list = XListFonts(gcl->dpy,font.to_charPtr(),1,&ret); // gibts den Font?
 		//AH Changed End
		if (ret && font_list)
			XFreeFontNames(font_list);

		if (gcl->fontstr)
			XFreeFontInfo(NULL,gcl->fontstr,0);
		if (gcl->font)
			XUnloadFont(gcl->dpy, gcl->font);

 	//AH Changed Start => to_charPtr()
 		if (!ret || (gcl->font = xgcv.font = XLoadFont(gcl->dpy,font.to_charPtr()))==BadName 
 	//AH Changed End
			|| xgcv.font==BadAlloc)
		{
			cerr<<"Font "<<font<<" could not be loaded, using default."<<endl;
//			XGetGCValues(gcl->dpy,XDefaultGC(gcl->dpy,XDefaultScreen(gcl->dpy)),GCFont,&xgcv);
			gcl->font = xgcv.font = XLoadFont(gcl->dpy,"fixed");
		}
		
		set |= GCFont;
		gcl->fontname = font;
		gcl->fontstr = XQueryFont(gcl->dpy,xgcv.font);
	}

	XChangeGC(gcl->dpy,gcl->gc,set,&xgcv);
}
