#include "xbefore.h"
#include <Xm/CascadeB.h>
#include <Xm/CascadeBG.h>
#include <Xm/DrawingA.h>
#include <Xm/DrawnB.h>
#include <Xm/MainW.h>
#include <Xm/MenuShell.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/ScrolledW.h>
#include <Xm/SeparatoG.h>
#include <Xm/Separator.h>
#include <Xm/Form.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/Label.h>
#include <Xm/Frame.h>
#include <Xm/PanedW.h>
#include <Xm/ToggleB.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xmu/Editres.h>
#include "xafter.h"

extern "C" {
#include <time.h>
#include <stdlib.h>
}

#include <strstream.h>
#include <iomanip.h>

#include "main.h"
#include "tool.h"
#include "editor.h"
#include "netcl.h"
#include "value.h"
#include "struct.h"
#include "logwin.h"

LinkedList<void*> Editor::editors;

static XtResource resources[] = {
/*
{
	"agnesDir",
	XtCString,
	XtRString,
	sizeof(String),
	XtOffset(ApplicationDataP,agnes_dir),
	XtRImmediate,
	(XtPointer) "."
},
{
	"modelDir",
	XtCString,
	XtRString,
	sizeof(String),
	XtOffset(ApplicationDataP,model_dir),
	XtRImmediate,
	(XtPointer) "."
},*/
{
	"initialZoom",
	XtRInt,
	XtRInt,
	sizeof(int),
	XtOffset(ApplicationDataP,initial_zoom),
	XtRImmediate,
	(XtPointer) 100
},
{
	"initialMoveGrid",
	XtRInt,
	XtRInt,
	sizeof(int),
	XtOffset(ApplicationDataP,initial_mvgrid),
	XtRImmediate,
	(XtPointer) 10
},
{
	"initialScaleGrid",
	XtRInt,
	XtRInt,
	sizeof(int),
	XtOffset(ApplicationDataP,initial_scgrid),
	XtRImmediate,
	(XtPointer) 10
},
{
	"initialRotateGrid",
	XtRInt,
	XtRInt,
	sizeof(int),
	XtOffset(ApplicationDataP,initial_rotgrid),
	XtRImmediate,
	(XtPointer) 45
},
{
	"initialNetClass",
	XtCString,
	XtRString,
	sizeof(String),
	XtOffset(ApplicationDataP,initial_netclass),
	XtRImmediate,
	(XtPointer) "EDSPN"
},
{//the time for a double-click
	"clickDelay",
	XtRInt,
	XtRInt,
	sizeof(int),
	XtOffset(ApplicationDataP,click_delay),
	XtRImmediate,
	(XtPointer) 600
},
{//time between redraws of editor window
	"updateDelay",
	XtRInt,
	XtRInt,
	sizeof(int),
	XtOffset(ApplicationDataP,update_delay),
	XtRImmediate,
	(XtPointer) 10
},
{//time between two pollings of the mouse position when scrolling
	"scrollDelay",
	XtRInt,
	XtRInt,
	sizeof(int),
	XtOffset(ApplicationDataP,scroll_delay),
	XtRImmediate,
	(XtPointer) 10
},
{//scroll step in x direction in pixels
	"XscrollStep",
	XtRInt,
	XtRInt,
	sizeof(int),
	XtOffset(ApplicationDataP,xscrollstep),
	XtRImmediate,
	(XtPointer) 150
},
{//scroll step in y direction in pixels
	"YscrollStep",
	XtRInt,
	XtRInt,
	sizeof(int),
	XtOffset(ApplicationDataP,yscrollstep),
	XtRImmediate,
	(XtPointer) 100
},
{
	"displayGrid",
	XtRBool,
	XtRBool,
	sizeof(bool),
	XtOffset(ApplicationDataP,display_grid),
	XtRImmediate,
	(XtPointer) false
},
{
	"expertMode",
	XtRBool,
	XtRBool,
	sizeof(bool),
	XtOffset(ApplicationDataP,expert_mode),
	XtRImmediate,
	(XtPointer) false
},
{
	"rightClickAction",
	XtCString,
	XtRString,
	sizeof(String),
	XtOffset(ApplicationDataP,right_click),
	XtRImmediate,
	(XtPointer) "Rotates"
},
};

//AH Changed Start: NULL nach (XPointer) gecastet
static XrmOptionDescRec options[] = {
 	{"-zoom",".initialZoom",XrmoptionSepArg,(XPointer)NULL},
 	{"-grid",".initialMoveGrid",XrmoptionSepArg,(XPointer)NULL},
 	{"-class",".initialNetClass",XrmoptionSepArg,(XPointer)NULL},
 	{"-right",".rightClickAction",XrmoptionSepArg,(XPointer)NULL},
	{"-expert",".expertMode",XrmoptionNoArg,"true"}
};
 //AH Changed End

Editor* Editor::NewEditor(Net* n, char* display_string, int load_cmd)
{
	if (!n && !load_cmd) return 0;

	int argc = app_argc;
	char** argv = Argv(); // Kopie des argv-Vektors, da vom Toolkit modifiziert
	Text disp(display_string);
	Display* dpy; 
	
 	//AH Changed Start disp => disp.to_char_Ptr()
 	if ((dpy = XtOpenDisplay(app,disp.to_charPtr(),app_name,app_class,
	//AH Changed End
		options,XtNumber(options),&argc,argv))==0 &&
 	//AH Changed (disp += ":0") => (disp += ":0").to_charPtr()
 	(dpy=XtOpenDisplay(app,(disp += ":0").to_charPtr(),app_name,app_class,
        //AH Changed End
		options,XtNumber(options),&argc,argv))==0)
	{
		cerr<<"can't open display "<<disp<<"!"<<endl;
		delete[] argv;
		return 0;
	}
	
	if (!n) // Startup-Behandlung
	{
		if (load_cmd < argc) // Netze/Klassen der Kommandozeile rekursiv lesen
		{
		  Text loadname = argv[load_cmd];
		  NetClass* nc = NetClass::GetNetClass(loadname.to_charPtr());
			if (nc)	n = new Net(nc);
			else 
			  {
			    //add ".net" to name 
			    loadname = loadname - ".net";
			    loadname = loadname + ".net";
			    //search model_dir first
			    Text netname = model_dir + "/" + loadname;
			    ifstream fis(netname.to_charPtr());
			    if(fis) //is in modeldir 
			      fis.close();
			    else
			      {//not in modeldir
				//take absolut path
				netname = loadname;
			      }
			    n = Net::Load(netname); 
			  }
			NewEditor(0,display_string,++load_cmd);
		}
		else
			if (argc==1) // Keine Argumente in der Kommandozeile
			{
				delete[] argv;
				return new Editor(0,dpy);
			}
	}

	delete[] argv;
	return n ? new Editor(n,dpy) : 0;
}


Editor::Editor(Net* n, Display* dpy): 
	Wrapper<Editor,void*>(this),
	display(dpy),
	app_shell(0),
	about(this),
 	help((tnet_dir+"/Agnes/README").to_charPtr()),
	curr_path(model_dir),
	pixmap(0),
	net(n),
	click_id(0),
	update_id(0),
        scroll_id(0),
	settings(0,5,0.1,15,false,false),
	curr_clname("?")
{
	editors.Append(this);	


	app_shell = XtVaAppCreateShell(app_name,app_class, 
				       applicationShellWidgetClass,display,NULL);

	XtRealizeWidget(app_shell);
	
	XtVaGetApplicationResources(app_shell,&app_data,
		resources,XtNumber(resources),NULL);
		
	if (!net) // Kein Netz uebergeben!
	{
		NetClass* nc = NetClass::GetNetClass(app_data.initial_netclass);
		if (!nc) nc = NetClassP(NetClass::NetClasses().First());
    if (!nc) {
      cerr<<"No netclasses found\n";
      exit(-1);
    };
    net = new Net(nc);
	}
		
	settings.mvgrid = app_data.initial_mvgrid;
	settings.scgrid = double(app_data.initial_scgrid)/100.0;
	settings.rotgrid = app_data.initial_rotgrid;
	settings.rclick = app_data.right_click;
				
	shell = XtVaCreatePopupShell("TopLevelShell",topLevelShellWidgetClass,
		app_shell,NULL);

	// Editres protocol support turned on (not necessary)
	XtAddEventHandler(shell,(EventMask)0,True,
		(XtEventHandler)_XEditResCheckMessages,NULL);

 	//AH Changed Start Destroy => &Destroy
 	WinMgrCB(shell,"WM_DELETE_WINDOW",&Editor::Destroy,0);
 	//AH Changed End

	main = XtVaCreateManagedWidget("MainWindow",
			xmMainWindowWidgetClass,shell,XmNuserData,this,NULL);
	
	CreateMenuBar(main);
	
	cmds = XtVaCreateManagedWidget("Commands",xmRowColumnWidgetClass,main,
		XmNuserData,XtPointer(this),NULL);
	nodebar.EnterParent(cmds);
	connbar.EnterParent(cmds);

	Widget cbox = XtVaCreateManagedWidget("CommandBox",xmRowColumnWidgetClass,cmds,NULL);

	box = XtVaCreateManagedWidget("ButtonBox",xmRowColumnWidgetClass,cbox,NULL);
	CreateEntries(box,xmPushButtonWidgetClass,NULL,
//		"Leave Enter ObjInfo Remove Undo Copy Cut Paste ZoomIn ZoomOut");
		"Load Save Enter Leave ObjInfo Remove Undo Copy Cut Paste");
	pos = XtVaCreateManagedWidget("Position",xmLabelWidgetClass,box,NULL);

	msg = XtVaCreateManagedWidget("Message",xmPushButtonWidgetClass,cbox,NULL);

	Message("");
	
	view = XtVaCreateManagedWidget("View",xmScrolledWindowWidgetClass, main,NULL);
	_FixHighlight(view);
	
	darea = XtVaCreateManagedWidget("DrawingArea",
		xmDrawingAreaWidgetClass,view,NULL);
	
	XtVaGetValues(view,
		XmNhorizontalScrollBar,&hsb,
		XmNverticalScrollBar,&vsb,
		XmNclipWindow,&clip,NULL);

	XGCValues xgcv;
	XtVaGetValues(darea,XtNwidth,&width,XtNheight,&height,
		XmNbackground,&xgcv.foreground,NULL);
	xgcv.fill_style = FillSolid; 
	
	grid_gc1 = grid_gc2 = 0;
	SetZoom(double(app_data.initial_zoom)/100.0); // Kreiert die initiale Pixmap
	
	clear_gc = XCreateGC(display,pixmap,GCFillStyle+GCForeground,&xgcv);
	//AH NEW: set foreground color
	Arg args;
	XtSetArg(args,XmNforeground, 0xdd5566);
	XtSetValues(darea, &args, 1); 
	//AH NEW End

	XtVaGetValues(darea,XmNforeground,&xgcv.foreground,NULL);
	xgcv.line_style = LineSolid; /*LineOnOffDash;*/
	xgcv.function = GXxor;
	dragbox_gc = XCreateGC(display,pixmap,GCFunction+GCForeground+GCLineStyle,&xgcv);
	XtVaGetValues(darea,XmNtopShadowColor,&xgcv.foreground,NULL);
	grid_gc1 = XCreateGC(display,pixmap,GCForeground+GCLineStyle,&xgcv);
	XtVaGetValues(darea,XmNbottomShadowColor,&xgcv.foreground,NULL);
	grid_gc2 = XCreateGC(display,pixmap,GCForeground+GCLineStyle,&xgcv);
	
//	char dot_dashes[] = { 1,Mini(1,short(settings.zoom*settings.mvgrid)-1) };
//	XSetDashes(display,grid_gc1,0,dot_dashes,2);
//	XSetDashes(display,grid_gc2,0,dot_dashes,2);
	
 	AddCB(darea,XmNexposeCallback,&Editor::Redraw,0);
	AddCB(darea,XmNinputCallback,&Editor::Input,0);

 	XmMainWindowSetAreas(main,menu,cmds,(_WidgetRec *)NULL,(_WidgetRec *)NULL,view);

	net->EditorOpened(*this);

	XtPopup(shell,XtGrabNone);
}


void Editor::SetNet(Net* n)
{
	if (n==net) return;
	if (net) net->EditorClosed(*this);
	if ((net=n)) net->EditorOpened(*this);
}


Editor::~Editor()
{
	if (update_id)
		RemoveCB(update_id);
	
	if (click_id)
		RemoveCB(click_id);
	
	if(scroll_id)
	  RemoveCB(scroll_id);

	net->EditorClosed(*this);

	editors.Remove(this);
	
	if (app_shell)
		XtDestroyWidget(app_shell);
	
	if (display) 
		XtCloseDisplay(display);

	if (!editors)
		exit(0);
}


bool Editor::Exists(Editor* editor)
{
	return editors.IsIn(editor);
}


void Editor::Destroy(Widget, void*, XtPointer)
{
	delete this;
}


void Editor::Menu(Widget w, void*, XtPointer cldata)
{
	char* n = XtName(w); Ident name(n);
	
	if (name=="ZoomIn") SetZoom(settings.zoom+0.1);
	else if (name=="ZoomOut") SetZoom(settings.zoom-0.1);
	else if (name=="AboutAgnes") OpenAboutBox();
	else if (name=="ErrorWindow")
		logwindow->Open("LogWindow","Error Log",shell);
	else if (name=="HelpWindow")
		help.Open("LogWindow"," Help ",shell);
	else if (name=="DispGrid")
	{
		settings.grid_visib = !settings.grid_visib;
		Update();
	}
	else if (name=="ExpertMode")
		settings.expert_mode = !settings.expert_mode;
	else if (strchr("ZRMS",*n) && isdigit(n[1]) && n[2]==0)
	{
		if (((XmToggleButtonCallbackStruct*)cldata)->set!=True)
			return;
		
		istrstream is(GetStringValue(w,XmNlabelString));
		double val = 0.0;
		is>>val;
		switch (*n)
		{
		case 'Z': SetZoom(val/100.0); break;
		case 'M': settings.mvgrid = val; Update(); break;
		case 'S': settings.scgrid = val; break;
		case 'R': settings.rotgrid = val; break;
		}
	}
        else if (name.IsOneOf("Scrolls Rotates Scales Removes Undoes Pastes"))
        {
                if (((XmToggleButtonCallbackStruct*)cldata)->set!=True)
                  return;
                else
                  settings.rclick = name;
        }
	else
		net->Menu(*this,name);
}


void Editor::SetSettings(EditorSettings& set)
{
	SetZoom(set.zoom);
	settings = set;
	Update();
}

void Editor::SetDimension(unsigned w,unsigned h)
{
	width = w;
	height = h;
	SetZoom(settings.zoom,false);
}



void Editor::SetZoom(double z, bool check)
{
	if (z<=0.1) z = 0.1;
	if (check && z==settings.zoom) return;
	
	// Hier wird das Zentrieren um die Bildmitte vorbereitet
	int x,y,w = 0,h;
	if (settings.zoom)
	{
		GetScroll(x,y,w,h);
		x = int((x+w/2)/settings.zoom); y = int((y+h/2)/settings.zoom);
	}
				
_again:
	if (pixmap && pixmap!=BadAlloc) XFreePixmap(display,pixmap);
	pixmap = XCreatePixmap(display,RootWindowOfScreen(XtScreen(shell)),
		unsigned(width*z),unsigned(height*z),
		unsigned(DefaultDepthOfScreen(XtScreen(shell))));
	
	if (pixmap==BadAlloc)
	{
		z = settings.zoom; goto _again;
	}
	
	settings.zoom = z;
	XtVaSetValues(darea,XtNwidth,Dimension(width*z),
		XtNheight,Dimension(height*z),NULL);

	if (w) SetScroll(
			Mini(int(width*z)-w-1,Maxi(0,int(x*z-w/2))),
			Mini(int(height*z)-h-1,Maxi(0,int(y*z-h/2))));

	Coordinate();

	if (grid_gc1)
	{
//		char dot_dashes[] = { 1,Maxi(1,short(z*settings.mvgrid)-1) };
//		XSetDashes(display,grid_gc1,0,dot_dashes,2);
//		XSetDashes(display,grid_gc2,0,dot_dashes,2);
		Update();
	}
}


void Editor::Message(char *str)
{
	SetStringValue(msg,XmNlabelString,*str?str:curr_clname);
}


void Editor::Title(char *str)
{
	Text title= "TimeNET 3.0, net displayed: ";
	title+= str;
 	XtVaSetValues(shell,XmNtitle,title.to_charPtr(),NULL);
}


void Editor::Coordinate(Position ppos)
{
	char buf[64]; ostrstream ps(buf,64);
	ps<<"x: "<<double(int(100*ppos.x))/100.0<<endl
		<<"y: "<<double(int(100*ppos.y))/100.0<<endl
		<<"z: "<<int(settings.zoom*100)<<'%'<<ends;
	SetStringValue(pos,XmNlabelString,buf);
}

//establish a timer that polls the mouse position
//called from page.cc when mouse button is held down: scrolls the drawing area
//when drawing a connector, dragging the dragbox, moving objects and connectors 
void Editor::StartScrolling()
{
  if(!scroll_id)
    {
      scroll_id = TimerCB(app_data.scroll_delay,&Editor::ScrollTimer,0);
      lastmouse=MousePosition(true);
   }
}


bool Editor::ScrollTimer(void* para)
{
  //mouse position relative to upper left corner of drawing area
  Position mpos = MousePosition(true)/settings.zoom; 
  //stop scrolling (uninstall timer) when button is released
  if(saveevent.event==EventExit) return false;
  ScrollVisible(mpos); //calculate and set new position of drawing area
  EventInfo currev=saveevent; //saved in Editor::Event
  currev.event = EventMotion; 
  //limit to size of drawing area
  currev.pos = mpos.XLimit(0,width-1).YLimit(0,height-1);
  net->Event(*this,currev); //update and redraw drawing area
  return true; //true = reinstall timer automatically
}

//stop scrolling after mouse button is released
void Editor::StopScrolling()
{
  if(scroll_id)
    RemoveCB(scroll_id);
  scroll_id = 0;
}

void Editor::ScrollVisible(Position p, bool center)
{
  const int borderwidth = 6;
  int ux,uy,lx,ly;
  GetScroll(ux,uy,lx,ly); //upper left position, window width and height
  //limit p.x and p.y to size of drawing area
  p = p.XLimit(0,width-1).YLimit(0,height-1)*settings.zoom;	
  
  if (center)
    {
      ux = int(p.x)-lx/2;
      uy = int(p.y)-ly/2;
    }
  else
    {
      lx += ux-1; //right border
      ly += uy-1; //down border
      
      //mouse position p is in visible area, do nothing
      if (p.x>=ux && p.x<=lx	&& p.y>=uy && p.y<=ly)
	return;
      
      //mouse position outside border, set scroll to distance
      //between position and border 
      if (p.x<ux) //mouse position < left border
	{ //fast scrolling 
	  if (ux-int(p.x)> borderwidth)
	    ux-=app_data.xscrollstep;
	  else //slow scrolling close to border
	    ux-=app_data.xscrollstep/10;
	}
      else if (p.x>lx) //mouse position > right border
	{
	  if (int(p.x)-lx> borderwidth)
	    ux+=app_data.xscrollstep;
	  else
	    ux+=app_data.xscrollstep/10;
	}

      if (p.y<uy) 
	{	
	  if (uy-int(p.y)> borderwidth)
	    uy-=app_data.yscrollstep;
	  else
	    uy-=app_data.yscrollstep/10;
	}
      else if (p.y>ly)
	{
	  if (int(p.y)-ly> borderwidth)
	    uy+=app_data.yscrollstep;
	  else
	    uy+=app_data.yscrollstep/10;
	}
    }
	SetScroll(ux,uy);
}


void Editor::Redraw(Widget w, void*, XtPointer calldata)
{
	XmDrawingAreaCallbackStruct* dacs = (XmDrawingAreaCallbackStruct*)calldata;

	if (dacs->reason==XmCR_EXPOSE)
	{
		XExposeEvent& ee = dacs->event->xexpose;
		XCopyArea(display,pixmap,XtWindow(w),
			XDefaultGC(display,XDefaultScreen(display)),
			ee.x,ee.y,ee.width,ee.height, ee.x,ee.y);
	}		
}



void Editor::Input(Widget, void*, XtPointer calldata)
{
  XmDrawingAreaCallbackStruct* dacs = (XmDrawingAreaCallbackStruct*)calldata;
  XEvent* ev = dacs->event;
  int type = ev->type;
  int	btn = (!ev->xbutton.state && type==MotionNotify) ? 1:ev->xbutton.button;
  bool shiftkey = ev->xbutton.state & ShiftMask; /*shift-key pressed?*/

  if (type==MotionNotify && XPending(display))
    return;

  last_event = *ev; // Fuer Scroll() ...

  /*no mouse button pressed && no mouse motion: */
 /*ignore non-mouse events*/
  if (type!=ButtonPress && type!=ButtonRelease && type != MotionNotify )
    return;

  Position pos = Position(ev->xbutton.x,ev->xbutton.y)/settings.zoom;

  EventInfo currev(
		   pos.XLimit(0,width-1).YLimit(0,height-1),
		   (btn==1 ? ButtonSelect : btn==2 ? ButtonCreate:
		    btn==3 ? ButtonUser : saveevent.button),
		   (type==ButtonPress?EventInit:type==ButtonRelease?EventExit:EventMotion),
		   shiftkey
		   );
 
  switch (currev.event)
    {
    case EventInit:
      {
	if(click_id)
	  {
	    if(clicked_obj &&  clicked_obj == net->GetPage(this)->UnderMouse(Dpy(),currev.pos))
	      { //same object clicked twice in double-click time
		currev.event = EventDouble;
		net->Event(*this,currev);
		break;
	      }
	    else
	      { //not same object twice: reset timer
		RemoveCB(click_id);
		click_id = 0;
	      }
	  }
	saveevent=currev; //needed for scrolling, stores mouse-button while mousebutton is pressed
	net->Event(*this,currev);
	//the timer counts down the time the user has for a double click
	click_id = TimerCB(app_data.click_delay,&Editor::ClickTimer,0);
	//save the object that is under mouse pointer (if any)
	clicked_obj = net->GetPage(this)->UnderMouse(Dpy(),currev.pos);
	break;
      }
    case EventMotion:
      { 
	net->Event(*this,currev);
	break;
      }
    case EventExit:
      {
	net->Event(*this,currev);
	break;
      }
    default:
      break;
    }
}


void Editor::Update()
{
	if (update_id)
		RemoveCB(update_id);
 	update_id = net ? TimerCB(app_data.update_delay,&Editor::DoUpdate,NULL) : 0;
}


void Editor::Copy()
{
	XCopyArea(display,pixmap,XtWindow(darea),
		XDefaultGC(display,XDefaultScreen(display)),
		0,0,Dimension(width*settings.zoom),Dimension(height*settings.zoom),0,0);
}


void Editor::Clear()
{
	double zoom = settings.zoom;
	
	XFillRectangle(display,pixmap,clear_gc,0,0,
		Dimension(width*zoom),Dimension(height*zoom));
	
	if (settings.grid_visib && int(settings.mvgrid*zoom+0.5)>5)
	{
		for (double x=0; x<width; x += settings.mvgrid)
		{
			XDrawLine(display,pixmap,grid_gc1,short(x*zoom+0.5)+1,0,short(x*zoom+0.5)+1,
				short(zoom*(height-1)));
			XDrawLine(display,pixmap,grid_gc2,short(x*zoom+0.5),0,short(x*zoom+0.5),
				short(zoom*(height-1)));
		}
		
		for (double y=0; y<height; y += settings.mvgrid)
		{
			XDrawLine(display,pixmap,grid_gc1,0,
				short(y*zoom+0.5)+1,short(zoom*(width-1)),short(y*zoom+0.5)+1);
			XDrawLine(display,pixmap,grid_gc2,0,
				short(y*zoom+0.5),short(zoom*(width-1)),short(y*zoom+0.5));
		}
	}
}


bool Editor::DoUpdate(void*)
{
	Text full = net->Title(*this);
	Title(full);

	net->Draw(*this,display,pixmap);

	update_id = 0;
	return false;
}


NodeClass* Editor::CurrNodeClass()
{
	return nodebar.Current();
}


ConnectorClass* Editor::CurrConnectorClass()
{
	return connbar.Current();
}

void Editor::NewNetClass(NetClass* n)
{
	connbar.LeaveParent();
	nodebar.LeaveParent();
	nodebar.ClearButtons();
	connbar.ClearButtons();

	n->EnterEditor(nodebar,connbar);
	Message(curr_clname = n->Label());
	
	nodebar.EnterParent(cmds);
	connbar.EnterParent(cmds);
}

Widget Editor::CreatePulldown(char* name, Widget parent)
{
	Widget pulldown=XmCreatePulldownMenu(parent ? XtParent(parent) : menu,
                //AH Changed Start name,NULL,0);
 		name,(ArgList)NULL,0);
        	//AH Changed End
	
	// Die Tearoffs sind jedoch problematisch
	XtVaSetValues(pulldown,XmNtearOffModel,XmTEAR_OFF_ENABLED,NULL);
	
	if (parent)
		XtVaSetValues(parent,XmNsubMenuId,pulldown,NULL);
	else			
	{
		Widget cascade = XtVaCreateManagedWidget(name,
			xmCascadeButtonWidgetClass,menu,
			XmNsubMenuId,pulldown,NULL);
		if (strcmp(name,"Help")==0)
			XtVaSetValues(menu,XmNmenuHelpWidget,cascade,NULL);
	}
	
        Iterator<Module*> miter(GetNet()->modules);
        for (Module* m=miter; m; m=++miter)
          if (m->CreatePulldown(this, name, parent)) break;

	return pulldown;
}


Widget Editor::CreateEntry(char* name, 
	WidgetClass cl,Widget pulldown,void* udata, WrapperID& id)
{
	Widget w = XtVaCreateManagedWidget(name,cl,pulldown,XmNuserData,this,NULL);
	CheckLabelPixmap(w);
	
 	//AH Changed Start Menu=> &Editor::Menu
	id= AddCB(w,cl==xmToggleButtonWidgetClass?
 		  XmNvalueChangedCallback:XmNactivateCallback,&Editor::Menu,udata);
 	//AH Changed End

        Iterator<Module*> miter(GetNet()->modules);
        for (Module* m=miter; m; m=++miter)
          if (m->CreateEntry(this, name, cl,pulldown)) break;

	return w;
}


void  Editor::DestroyEntry(Widget w, WrapperID id)
{
        if (id)
	  RemoveCB(id);
	XtDestroyWidget(w);
}


Widget Editor::CreateSeparator(Widget pulldown)
{
	return XtVaCreateManagedWidget("Separator",
		xmSeparatorWidgetClass,pulldown,NULL);
}


void Editor::CreateEntries(Widget pulldown, WidgetClass cl, void* udata, 
	char* names)
{
        WrapperID dummy;
	Ident name;
	istrstream is(names);
	while (is>>name)
	{
		if (name!="_")
			CreateEntry(name,cl,pulldown,udata, dummy);
		else
			CreateSeparator(pulldown);
	}
}

	
void Editor::CreateMenuBar(Widget parent)
{
        WrapperID dummy;
	menu = XmVaCreateSimpleMenuBar(parent,"MenuBar",NULL);
	Widget netw = CreatePulldown("Net"),
		classp = CreatePulldown("NetClasses",
		CreateEntry("New",xmCascadeButtonWidgetClass,netw, NULL, dummy));

	menush = XtParent(netw);
		
//	CreateEntry("Netclasses",xmLabelWidgetClass,classp, NULL, dummy);
	Iterator<void*> iter = NetClass::NetClasses();
	for (void* nc=iter; nc; nc=++iter)
		if (!NetClassP(nc)->IsSubclass())
			SetStringValue(
				CreateEntry(NetClassP(nc)->Name(),xmPushButtonWidgetClass,classp, NULL, dummy),
				XmNlabelString,NetClassP(nc)->Label());
		
	CreateEntries(netw,xmPushButtonWidgetClass,NULL,
	"Load Add Save SaveAs _ Kill _ NewEditor NewDisplay _ Close Quit");

	CreateEntries(CreatePulldown("Edit"),xmPushButtonWidgetClass,NULL,
		"Undo _ Copy Cut Paste _ Remove");
	
	CreateEntries(CreatePulldown("Object"),xmPushButtonWidgetClass,NULL,
		"ObjInfo _ Enter Leave _ Prev Next Select SelectAll");
	
	mods = CreatePulldown("Module");
	opts = CreatePulldown("Options");

	CreateEntry("NetInfo",xmPushButtonWidgetClass,opts, NULL, dummy);
	
	Widget zoomp = CreatePulldown("Zoom");
	CreateEntries(zoomp,xmPushButtonWidgetClass,NULL,"ZoomIn ZoomOut");
	CreateEntries(zoomp,xmToggleButtonWidgetClass,NULL,"_ Z1 Z2 Z3 Z4 Z5 Z6 Z7");

	CreateEntries(CreatePulldown("Help"),xmPushButtonWidgetClass,NULL,
		"AboutAgnes HelpWindow ErrorWindow");

	Widget rclick = CreatePulldown("RightClick",
		CreateEntry("RightClick",xmCascadeButtonWidgetClass,opts, NULL, dummy));
	
	CreateEntries(rclick,xmToggleButtonWidgetClass,NULL,
		"Scrolls Rotates Scales Removes Undoes Pastes");

	Widget mgridp = CreatePulldown("MoveGrid",
		CreateEntry("MoveGrid",xmCascadeButtonWidgetClass,opts, NULL, dummy)),
	sgridp = CreatePulldown("ScaleGrid",
		CreateEntry("ScaleGrid",xmCascadeButtonWidgetClass,opts, NULL, dummy)),
	rgridp = CreatePulldown("RotateGrid",
		CreateEntry("RotateGrid",xmCascadeButtonWidgetClass,opts, NULL, dummy));
	
	CreateEntries(opts,xmToggleButtonWidgetClass,NULL,
		"_ DispGrid ExpertMode");
	XtVaSetValues(XtNameToWidget(opts,"DispGrid"),
		XmNset,settings.grid_visib = app_data.display_grid,NULL);
	XtVaSetValues(XtNameToWidget(opts,"ExpertMode"),
		XmNset,settings.expert_mode = app_data.expert_mode,NULL);
	

	CreateEntries(mgridp,xmToggleButtonWidgetClass,NULL,"M0 M1 M2 M3 M4 M5 M6");
	CreateEntries(sgridp,xmToggleButtonWidgetClass,NULL,"S0 S1 S2 S3 S4 S5 S6");
	CreateEntries(rgridp,xmToggleButtonWidgetClass,NULL,"R0 R1 R2 R3 R4 R5 R6");
	
	XtManageChild(menu);
}


void Editor::GetScroll(int& x, int& y, int& w, int& h)
{
	Dimension dw,dh;
	
	XtVaGetValues(clip,XtNwidth,&dw,XtNheight,&dh,NULL);
	XtVaGetValues(hsb,XmNvalue,&x,NULL);
	XtVaGetValues(vsb,XmNvalue,&y,NULL);

	w = int(dw); h = int(dh);
}


void Editor::SetScroll(int x, int y)
{
	XEvent event;
	XmScrollBarCallbackStruct sbcs = {XmCR_VALUE_CHANGED,&event};
		
	if (x<0) x = 0;
	if (y<0) y = 0;
		
	XtVaSetValues(hsb,XmNvalue,sbcs.value = x,NULL);
	XtCallCallbacks(hsb,XmNvalueChangedCallback,&sbcs);
	XtVaSetValues(vsb,XmNvalue,sbcs.value = y,NULL);
	XtCallCallbacks(vsb,XmNvalueChangedCallback,&sbcs);
}


void Editor::Scroll()
{
	XEvent* event = &last_event; // last_event wird bei jedem Input gespeichert
	
	if (event->type==MotionNotify)
	{
		int x,y,old_x,old_y,mn,mx,slz;
		XmScrollBarCallbackStruct sbcs = {XmCR_VALUE_CHANGED,event};

		XtVaGetValues(hsb,XmNvalue,&old_x,XmNminimum,&mn,
			XmNmaximum,&mx,XmNsliderSize,&slz,NULL);
		x = Mini(mx-slz,Maxi(mn,old_x+2*(last_x-event->xmotion.x)));

		if (x!=old_x)
		{
			XtVaSetValues(hsb,XmNvalue,sbcs.value = x,NULL);
			XtCallCallbacks(hsb,XmNvalueChangedCallback,&sbcs);
		}
	
		XtVaGetValues(vsb,XmNvalue,&old_y,XmNminimum,&mn,
			XmNmaximum,&mx,XmNsliderSize,&slz,NULL);
		y = Mini(mx-slz,Maxi(mn,old_y+2*(last_y-event->xmotion.y)));

		if (y!=old_y)
		{
			XtVaSetValues(vsb,XmNvalue,sbcs.value = y,NULL);
			XtCallCallbacks(vsb,XmNvalueChangedCallback,&sbcs);
		}
		last_x = event->xmotion.x+(x-old_x);
		last_y = event->xmotion.y+(y-old_y);
	}
	else
	{
		if (event->type!=ButtonRelease)
		{
			last_x = event->xmotion.x;
			last_y = event->xmotion.y;
		}
	}
}


Editor* Editor::GetEditor(Widget w)
{
	Editor* e;
	
	while (!XmIsMainWindow(w)) w = XtParent(w);
	XtVaGetValues(w,XmNuserData,&e,NULL);
	
	return e;
}


bool Editor::ClickTimer(void* quickexit)
{
	click_id = 0;
	return false;
}


void Editor::ButtonState(char* names,bool state)
{
	Ident name;
	Widget w;
	istrstream is(names);
	while (is>>name)
	{
 	  	//AH Changed Start .to_charPtr()
 		if((w = XtNameToWidget(box,name.to_charPtr())))
 		//AH Changed End
			XtSetSensitive(w,state?True:False);
			
		WidgetList wlist; Cardinal wc;
		XtVaGetValues(menush,XmNchildren,&wlist,XmNnumChildren,&wc,NULL);
		while (wc--)
 		  	//AH Changed Start .to_charPtr()
 			if((w = XtNameToWidget(wlist[wc],name.to_charPtr())))
			//AH Changed End
				XtSetSensitive(w,state?True:False);
	}
}


Widget Editor::NewModule(char* name)
{
        WrapperID dummy;
	Widget w = XtNameToWidget(mods,name);
	return (w && XmIsPushButton(w))?
		w:CreateEntry(name,xmPushButtonWidgetClass,mods,0, dummy);
}


void Editor::RemoveModule(char* name)
{
	Widget w = XtNameToWidget(mods,name);
	if (w && XmIsPushButton(w)) XtDestroyWidget(w);
}


void Editor::Lock(bool set)
{
	Mouse(set);
}


void Editor::Mouse(bool define, unsigned int shape)
{
	if (define)
	{
		Cursor cursor = XCreateFontCursor(display,shape);
		XDefineCursor(display,XtWindow(shell),cursor);
		XFlush(display);
	}
	else
		XUndefineCursor(display,XtWindow(shell));
}


Position Editor::MousePosition(bool relative)
{
	Window client,root; int x,y,wx,wy; unsigned mask;
	
	XQueryPointer(display,XtWindow(darea),&root,&client,&x,&y,&wx,&wy,&mask);
		
	if(relative)
	  return Position(wx,wy); //relative to upper left corner of drawing area
	else
	  return Position(x,y); //relative to upper left corner of screen
}


void Editor::OpenAboutBox()
{
   	//AH Changed Start 	about.Create("AboutDialog",NULL,shell,false);
 	about.Create("AboutDialog",(char*)NULL,shell,false);
 	//AH Changed End
	
	Widget form = XtVaCreateManagedWidget("AboutForm",
		xmFormWidgetClass,about,NULL);
	
	XtVaCreateManagedWidget("TextT",xmLabelWidgetClass,form,NULL);
	XtVaCreateManagedWidget("TextT2",xmLabelWidgetClass,form,NULL);
	XtVaCreateManagedWidget("TextT3",xmLabelWidgetClass,form,NULL);
	XtVaCreateManagedWidget("AgnesLogo",xmLabelWidgetClass,form,NULL);
	XtVaCreateManagedWidget("AgnesVersion",xmLabelWidgetClass,form,NULL);
	XtVaCreateManagedWidget("TUBLogo",xmLabelWidgetClass,form,NULL);
	XtVaCreateManagedWidget("Text1",xmLabelWidgetClass,form,NULL);
	XtVaCreateManagedWidget("Text2",xmLabelWidgetClass,form,NULL);
	XtVaCreateManagedWidget("Text3",xmLabelWidgetClass,form,NULL);
	
	about.Manage();
}


void Editor::MainLoop()
{
	if (editors)
		XtAppMainLoop(app);
}
