#include "xbefore.h"
#include <Xm/TextF.h>
#include "xafter.h"

#include <strstream.h>

#include "float.h"
#include "scope.h"
#include "constants.h"

VALUE_NOTIFY("float",Float)

Float::Float(NetClass* nc, bool hidden):
	Value(nc,hidden),
	val(0.0),
	max(MAXDOUBLE),
	min(-MAXDOUBLE),
	text(0)
{
}


Float::Float(double d): 
	val(d),
	max(MAXDOUBLE),
	min(-MAXDOUBLE),
	text(0)
{
}


Float::Float(const Float& f): 
	Value(f),
	val(f.val),
	max(f.max),
	min(f.min),
	text(0)
{
}


istream& Float::Parse(istream& is)
{
	Scope si(is,0,2);	
	min = 0.0;
	while (is && si.Get())
		switch (si.field)
		{
			case 1:
				if (!(is>>min))
					cerr<<"float (min. value) expected!"<<endl;
				break;
			case 2:
				if (is>>max)
				{
					if (max<min)
					{
						double tmp = max;
						max = min;
						min = tmp;
					}
				}
				else
					cerr<<"float (max. value) expected!"<<endl;
				break;
		}
	return is;
}


istream& Float::Input(istream& is)
{
	if (is>>val)
	{
		*this = val;
		return is;
	}
	else
		cerr<<"float expected!"<<endl;
	
	is.clear(ios::badbit);
	return is;
}


void Float::UpdateWidget()
{
        //AH Changed Start XmTextFieldSetString(text,(char*)*this);
 	XmTextFieldSetString(text,(*this).to_charPtr());
 	//AH Changed End
	XmTextFieldSetInsertionPosition(text,XmTextFieldGetLastPosition(text));
}


Widget Float::CreateWidget(Widget parent)
{
	text = XtVaCreateManagedWidget("Float",xmTextFieldWidgetClass,parent,NULL);
	CatchDestroy(&text);
	UpdateWidget();

	return text;
}


bool Float::EvalWidget()
{
	istrstream istr(XmTextFieldGetString(text));
	if (istr>>*this) UpdateWidget();

	return !istr.fail();
}


Value& Float::operator =(const double d)
{
	if ((val = d)>max) val = max; else if (val<min) val = min;
	if (text) UpdateWidget();
	return *this;
}


//AH Changed Start Float::operator char*() const
char* Float::to_charPtr() const
//AH Changed End
{
	static char buf[64]; ostrstream ostr(buf,64);
	ostr<<val<<ends;
	return buf;
}
