#include "xbefore.h"
#include <Xm/Form.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/LabelG.h>
#include "xafter.h"

#include <strstream.h>

extern "C" {
#include <ctype.h>
#include <string.h>
extern int strcasecmp(const char*,const char*);
}

#include "ident.h"
#include "scope.h"
#include "text.h"

VALUE_NOTIFY("ident",Ident)


Ident::Ident(NetClass* nc, bool hidden): 
	Value(nc,hidden),
	Wrapper<Ident,int>(this),
	csens(true),
	text(0),
	p(0),
	descr(0)
{
}


Ident::Ident(bool cs): 
	Wrapper<Ident,int>(this),
	csens(cs),
	text(0),
	p(0),
	descr(0)
{
}


Ident::Ident(const char *s, bool cs):
	Wrapper<Ident,int>(this),
	csens(cs),
	text(0),
	p(0),
	descr(0)
{
	*this = s;
}


Ident::Ident(const Ident& i): 
	Value(i),
	Wrapper<Ident,int>(this),
	text(0),
	p(0),
	descr(0)
{
	*this = i;
}


Ident::~Ident()
{
	UncatchDestroy(&text);
	if (p) delete[] p;
	if (descr) delete descr;
}


void Ident::Validate(Widget, int, XtPointer calldata)
{
	XmTextVerifyCallbackStruct* tvcs = (XmTextVerifyCallbackStruct*)calldata;
	if (tvcs->text->length)
	{
		for (char* p=tvcs->text->ptr; *p; )
			if (Valid(*p,tvcs->startPos+p-tvcs->text->ptr==0))
			{
				if (!csens) *p = tolower(*p);
				p++;
			}
			else
				strcpy(p,p+1);
		tvcs->text->length = strlen(tvcs->text->ptr);
	}
}


Value& Ident::operator =(const char* s)
{
  //AH CHANGED Start
  if (p) delete[] p;
  if (s) 
    {
      strcpy(p = new char[strlen(s)+1],s);
      if (text) XmTextSetString(text,p);
    }
  else 
    {
      p = 0;
      if (text) XmTextSetString(text,"");
    }
  return *this;
  
  //AH CHANGED End

}


Value& Ident::operator =(const Ident& i)
{
	if (descr) delete descr;
	descr = 0;
	if (i.descr) descr = new Text(*i.descr);
	csens = i.csens;
	return *this = i.p;
}


bool Ident::IsOneOf(char* names)
{
	Ident id;
	istrstream is(names);	
	
	while (is>>id)
		if (*this==id) return true;

	return false;
}


Widget Ident::CreateWidget(Widget parent)
{
	text = XtVaCreateManagedWidget("Ident",xmTextFieldWidgetClass,parent,NULL);
 	//AH Changed Start 	AddCB(text,XmNmodifyVerifyCallback,Validate,0);
 	AddCB(text,XmNmodifyVerifyCallback,&Ident::Validate,0);
 	//AH Changed End
	CatchDestroy(&text);
	//AH Changed Start const_cast<char*>
	//to convert from a const to a non const (same as char *)
	XmTextSetString(text,const_cast<char*>(Label()));
	//AH Changed End
	return text;
}


bool Ident::EvalWidget()
{
	if (text)
	{
		char *s = XmTextGetString(text);
		if (*s) *this = s; else return false;
	}
	else
		cerr<<"Evaluating nonexisting Widget!"<<endl;

	return true;
}


istream& Ident::Parse(istream& is)
{
	Scope sb(is,0,1);
	bool csens;
	while (is && sb.Get())
	{
		Ident b(false);
		if (is>>b)
			csens = b=="true";
		else
			cerr<<"bool (case sensitivity) expected."<<endl;
	}
	return is;
}


istream& Ident::Input(istream& is)
{
	char c,buf[1024],*ptr = buf;

  if (descr) { delete descr; descr = 0; }
	if (is>>c)
	{
		is.putback(c);
		if (c=='"') 
			is>>*(descr = new Text);
	}
	for (is>>c; is; is.get(c))
	{
		if (c=='?' && ptr==buf)
		{
			*this = 0; return is;
		}

		if (Valid(c,ptr==buf) && ptr<buf+255)
			*ptr++ = c;
		else
		{
			is.putback(c);
			break;
		}
	}
	is.clear();
	if (ptr==buf) is.clear(ios::badbit);
	if (is)
	{
		*ptr = 0;
		*this = buf;
	}
	return is;
}


ostream& Ident::Output(ostream& os) const
{
	return os<<(p?p:"?");
}


bool Ident::operator ==(const char *s) const
{
	return p && s && (csens?strcmp:strcasecmp)(p,s)==0;
}

bool Ident::operator >(const char *s) const
{
	return p && s && (csens?strcmp:strcasecmp)(p,s)>0;
}

bool Ident::operator <(const char *s) const
{
	return p && s && (csens?strcmp:strcasecmp)(p,s)<0;
}


bool Ident::Valid(char c, bool first)
{
        return isalpha(c) || c=='_' || c=='.' || (!first && isdigit(c));
}


Ident Ident::operator +(const char* s) const
{
	Ident ni(*this);
	return ni += s;
}


Ident& Ident::operator +=(const char* s)
{
	if (s)
	{
		char* np = new char[(p?strlen(p):0)+strlen(s)+1];
		*np = 0;
		if (p)
		{
			strcat(np,p);
			delete[] p;
		}
		
		strcat(p = np,s);
		if (text)
			XmTextSetString(text,p);
	}
	
	return *this;
}


Ident& Ident::operator +=(const char c)
{
	char s[] = {c,0};
	return *this += s;
}
