#include <iostream.h>
#include "linked.h"


template<class T> LinkedList<T>::~LinkedList()
{
	Clear();
}

// Schnelle Listen-Zuweisung...
template<class T> LinkedList<T>& LinkedList<T>::operator =(const LinkedList<T>& l)
{
	Container<T>* from = l.first,* to = first;
	while (to && from) { to->val = from->val; to = to->next; from = from->next; }
	if (from)
	{
		for (; from; from=from->next) Append(from->val); 
	}
	else
		while (to)
		{
			Container<T>* next = to->next;
			_Remove(to); to = next;
		}
	return *this;
}

/*
// Langsame Listen-Zuweisung
template<class T> LinkedList<T>& LinkedList<T>::operator =(const LinkedList<T>& l)
{
	Clear();
	for (Container<T>* from = l.first; from; from=from->next)
		Append(from->val); 
	return *this;
}
*/

template<class T> bool LinkedList<T>::operator ==(const LinkedList<T>& l)
{
	if (length!=l.length) return false;
	
	Container<T>* t,* lt;
	for (lt=first,t=first; t; t=t->next,lt=lt->next)
		if (lt->val!=t->val) return false;
	
	return true;
}


template<class T> bool LinkedList<T>::IsIn(const T& it) const
{
	for (Container<T>* t=first; t; t=t->next)
		if (t->val==it)
			return true;
	
	return false;
}


template<class T> void LinkedList<T>::Clear()
{
	for (; length; length--)
	{
		Container<T>* next = first->next;
		delete first;
		first = next;
	}
	
	last = 0;
}


template<class T> bool LinkedList<T>::Append(const T& val)
{
	Container<T>* n = new Container<T>(val,(Container<T>*)0,last);
	
	if (!n) return false;
	if (last) last->next = n;
	if (!first) first = n;
	last = n;
	length++;
	return true;
}


template<class T> Container<T>* LinkedList<T>::_Idx(int pos) const
{
	Container<T>* c = 0;

	if (pos <= length/2)
	{
		for (c=first; pos--; c=c->next); 
	}
	else // pos > length/2 && pos <= length-1
	{
		pos = length - pos - 1;
		for (c=last; pos--; c=c->prev); 
	}
	
	return c;
}


template<class T> int LinkedList<T>::Index(const T& val) const
{
	Container<T>* c; int pos = 0;
	
	for (c=first; c && !(c->val==val); pos++,c=c->next); 
	
	return pos==length?-1:pos;
}


template<class T> bool LinkedList<T>::Insert(const T& val, int pos)
{
	if (pos==length)
		return Append(val);
	else if (pos < 0 || pos > length)
	{
		cerr<<"LinkedList.Insert: can't insert at pos "<<pos<<endl;
		exit(-1);
	}
	else
	{
		Container<T>* c = _Idx(pos);

		c = new Container<T>(val,c,c?c->prev:0);
		if (c->prev)
			c->prev->next = c;
		else
			first = c;
		
		if (c->next)
			c->next->prev = c;
		else
			last = c;
			
		length++;
		return true;
	}
}


template<class T> void LinkedList<T>::_Remove(Container<T>* r)
{
	if (!r) return;
	if (r->prev)
		r->prev->next = r->next;
	else
		first = r->next;

	if (r->next)
		r->next->prev = r->prev;
	else
		last = r->prev;


	delete r;
	length--;
}


template<class T> bool LinkedList<T>::RemoveIndex(int pos)
{
	if (pos<0 || pos>=length)
	{
		cerr<<"LinkedList.RemoveIndex: can't remove pos "<<pos<<endl;
		exit(-1);
	}
	else
	{
		_Remove(_Idx(pos));
		return true;
	}
}


template<class T> bool LinkedList<T>::Replace(const T& oldval, const T& newval) 
{
	for (Container<T>* c=first; c; c=c->next)
		if (c->val==oldval)
		{
			c->val = newval;
			return true;
		}
	
	cerr<<"LinkedList.Replace: can't replace entry!"<<endl;
	exit(-1);
}


template<class T> bool LinkedList<T>::Remove(const T& val) 
{
	for (Container<T>* c=first; c; c=c->next)
		if (c->val==val)
		{
			_Remove(c);
			return true;
		}
	cerr<<"LinkedList.Remove: can't remove entry!"<<endl;
	exit(-1);
}


template<class T> T* LinkedList<T>::operator [](int pos) const
{
	if (pos<0 || pos>=length)
	{
		cerr<<"LinkedList.operator []: no position "<<pos<<endl;
		exit(-1);
	}
	
	return &_Idx(pos)->val;
}

// for g++

#define INSTANCIATE(the_class) \
template class Container< the_class >;\
template class LinkedList< the_class >;\
template class Iterator< the_class >;

#define CINSTANCIATE(the_class) \
template class Container< the_class * >;\
template class LinkedList< the_class * >;\
template class Iterator< the_class * >;

#define OINSTANCIATE(the_class) \
template class Container< the_class >;\
template class LinkedList< the_class >;\
template class OpaqueIterator< the_class >;

#ifndef NoExplicitInstanciation

#include "netcl.h"
#include "objbar.h"
#include "nodecl.h"
#include "conncl.h"
#include "context.h"
//#include "spnlstructure.h"

//CINSTANCIATE(FrameElement)
//CINSTANCIATE(SPNLElement)

INSTANCIATE(void*)
CINSTANCIATE(Editor)
CINSTANCIATE(Net)
CINSTANCIATE(Page)
CINSTANCIATE(NetClass)
CINSTANCIATE(Shape)
CINSTANCIATE(Value)
CINSTANCIATE(Position)
CINSTANCIATE(Node)
CINSTANCIATE(Connector)
CINSTANCIATE(Object)
CINSTANCIATE(ObjectClass)
CINSTANCIATE(NodeClass)
CINSTANCIATE(ConnectorClass)
CINSTANCIATE(Attrib)
CINSTANCIATE(Ident)
CINSTANCIATE(Text)
CINSTANCIATE(Module)
CINSTANCIATE(char*)
CINSTANCIATE(char)

class Uebergang;
CINSTANCIATE(Uebergang)
class FTrans;
CINSTANCIATE(FTrans)
// S.R. Fischertechnik-Anlage

class FireItem;
CINSTANCIATE(FireItem)
class FiredItem;
CINSTANCIATE(FiredItem)
CINSTANCIATE(int)
// K.S. Tokengame

class SubTrans;
class EDSPNNode;
class HCPNpage;
class HCPNdefinition;
struct EClass_T;
CINSTANCIATE(HCPNpage)
CINSTANCIATE(HCPNdefinition)
CINSTANCIATE(SubTrans)
CINSTANCIATE(EDSPNNode)
CINSTANCIATE(LinkedList<EDSPNNode*>)
CINSTANCIATE(EClass_T)
//Agnes/modules/HCPN2EDSPN, PathDecomp and MSIPDecomp


INSTANCIATE(ModuleInstanciator)
OINSTANCIATE(Position)
OINSTANCIATE(ClassGroup<NodeClass>)
OINSTANCIATE(ClassGroup<ConnectorClass>)
OINSTANCIATE(ConnectorClass::Rule)
OINSTANCIATE(Context::EnvList)
OINSTANCIATE(Context::GCList)

INSTANCIATE(ObjectBar<NodeClass>::Button*)
INSTANCIATE(ObjectBar<ConnectorClass>::Button*)

#endif
