#include "xbefore.h"
#include <Xm/List.h>
#include <Xm/Form.h>
#include <Xm/RowColumn.h>
#include <Xm/Label.h>
#include <Xm/PushBG.h>
#include <Xm/PushB.h>
#include <Xm/Frame.h>
#include <Xm/Separator.h>
#include "xafter.h"

#include "list.h"
#include "tool.h"
#include "scope.h"

VALUE_NOTIFY("list",List);


List::List(NetClass* nc, bool hidden): 
	Value(nc,hidden),
	Wrapper<List,int>(this),
	form(0),list(0),toolbar(0),
	def(0),
	maxlen(0),
	cutbuf(0),
	cutbuf_len(0)
{
}


List::List(const List& l): 
	Value(l),
	Wrapper<List,int>(this),
	form(0),list(0),toolbar(0),
	def(0),
	maxlen(l.maxlen),
	cutbuf(0),
	cutbuf_len(0)
{
	if (l.def)
		def = l.def->Copy();
	
	if (l.cutbuf)
		cutbuf = CopyTable(l.cutbuf,l.cutbuf_len);
		
	*this = l;
}


List::~List()
{
	if (def) delete def;
	UncatchDestroy(&toolbar,&form,&list);

	Iterator<Value*> iter(vals);
	for (Value* v=iter; v; v=++iter)
		delete v;

	KillTable(cutbuf,cutbuf_len);
}


istream& List::Parse(istream& is)
{
	if (def) delete def;
	
	Scope da(is);
	while (is && da.Get())
		if (!(def = Type(is))) cerr<<"cerr parsing type of List!"<<endl;

	return is;
}


istream& List::Input(istream& is)
{
	Value* val;
	LinkedList<Value*> ivals;
	Scope dp(is);
	
	while (is && dp.Get())
	{
		if (is>>*(val = def->Copy()))
			ivals.Append(val);
		else
		{
			delete val;
			cerr<<"initializer expected!"<<endl;
			dp.Skip();
			break;
		}
	}
	
	Iterator<Value*> iter(vals);
	for (val=iter; val; val=++iter)
		delete val;
	vals = ivals;
	UpdateWidget();
		
	return is;
}


ostream& List::Output(ostream& os) const
{
	os<<'{';
	Iterator<Value*> iter(vals);
	for (Value* val=iter; val; )
	{
		os<<*val;
		if ((val = ++iter)) os<<',';
	}
	
	return os<<'}';
}


void List::UpdateToolbar(int selcnt)
{
	if (toolbar)
	{
		XtSetSensitive(XtNameToWidget(toolbar,"Insert"),selcnt==1?True:False);
		XtSetSensitive(XtNameToWidget(toolbar,"Remove"),selcnt?True:False);
		XtSetSensitive(XtNameToWidget(toolbar,"Change"),selcnt?True:False);
		XtSetSensitive(XtNameToWidget(toolbar,"Cut"),selcnt?True:False);
		XtSetSensitive(XtNameToWidget(toolbar,"Copy"),selcnt?True:False);
		XtSetSensitive(XtNameToWidget(toolbar,"Paste"),cutbuf?True:False);
	}
}


void List::Selection(Widget, int, XtPointer calldata)
{
	XmListCallbackStruct& xlcs = *(XmListCallbackStruct*)calldata;
	int selcnt = xlcs.selected_item_count;
	
	if (xlcs.reason==XmCR_DEFAULT_ACTION)
	{
		istrstream is(NormalString(xlcs.item));
		is>>*def;
	}
	
	UpdateToolbar(selcnt);
}


XmString List::BuildString(Value* v, int /*cnt*/)
{
	char buf[1024];
	ostrstream str(buf,1024);
//	str.precision(5);
//	str<<cnt<<": "<<*v<<ends;
	str<<*v<<ends;
	return XmStringCreate(buf,XmSTRING_DEFAULT_CHARSET);
}


// Nummeriert neu ?
XmStringTable List::CopyTable(XmStringTable src, int len)
{
	XmStringTable newtbl = new XmString[len];
	while (len--)
		newtbl[len] = XmStringCopy(src[len]);

	return newtbl;
}


void List::KillTable(XmStringTable& tbl, int& len)
{
	while (len--)
		XmStringFree(tbl[len]);
		
	delete[] tbl;
	tbl = 0;
	len = 0;
}

/*
void List::Renumber()
{
	XmStringTable all_items; int item_cnt;
	XtVaGetValues(list,XmNitems,&all_items,XmNitemCount,&item_cnt,NULL);
	XmStringTable newtbl = CopyTable(all_items,item_cnt);
	
	XmListReplaceItemsPos(list,new_tbl,item_cnt,1);
	KillTable(newtbl,item_cnt);
}
*/

void List::MultiXmSelect(int* poslist, int cnt, int firstpos)
{
	XmListDeselectAllItems(list);
	if (poslist)
		for (int i=0; i<cnt; i++)
			XmListSelectPos(list,poslist[i],(i==cnt-1)?True:False);
	else
		for (int i=0; i<cnt; i++)
			XmListSelectPos(list,firstpos+i,(i==cnt-1)?True:False);
}


void List::Button(Widget w, int, XtPointer)
{
	Ident name(XtName(w));
	
	def->EvalWidget();
	
	int* pos_list,pos_cnt = 0,endpos;
	XmListGetSelectedPos(list,&pos_list,&pos_cnt);
	XtVaGetValues(list,XmNitemCount,&endpos,NULL);
	int firstpos = pos_cnt?pos_list[0]:0;

	if (name=="Cut" || name=="Copy")
	{
		KillTable(cutbuf,cutbuf_len);
		XmStringTable selection; int selcnt;
	
		XtVaGetValues(list,XmNselectedItems,&selection,
			XmNselectedItemCount,&selcnt,NULL);
		cutbuf = CopyTable(selection,cutbuf_len = selcnt);
		
		if (name=="Cut")
		{
			XmListDeletePositions(list,pos_list,pos_cnt);
			pos_cnt = 0;
		}
	}

	else if (name=="Change" && pos_cnt)
	{
		int i;
		XmStringTable ctbl = new XmString[pos_cnt];
		for (i=0; i<pos_cnt; i++)
			ctbl[i] = BuildString(def,i);
		
		XmListReplacePositions(list,pos_list,ctbl,pos_cnt);
		MultiXmSelect(pos_list,pos_cnt);
		
		KillTable(ctbl,i);
	}
	
	else if (name=="Paste")
	{
		XmListAddItems(list,cutbuf,cutbuf_len,firstpos);
		MultiXmSelect(0,cutbuf_len,firstpos);
	}

	else if (name=="Append")
	{
		XmString defstr = BuildString(def,endpos);
		XmListAddItems(list,&defstr,1,endpos+1);
		MultiXmSelect(0,1,endpos+1);
		XmStringFree(defstr);
	}
	else if (name=="Insert")
	{
		XmString defstr = BuildString(def,firstpos);
		XmListAddItems(list,&defstr,1,firstpos);
		MultiXmSelect(0,1,firstpos);
		XmStringFree(defstr);
	}
	else if (name=="Remove")
	{
		XmListDeletePositions(list,pos_list,pos_cnt);
		pos_cnt = 0;
	}
	
	UpdateToolbar(pos_cnt);
}


Widget List::CreateButton(Widget parent, char *name)
{
	Widget w = XtVaCreateManagedWidget(name,xmPushButtonWidgetClass,parent,NULL);
	CheckLabelPixmap(w);

 	//AH Changed Start 	AddCB(w,XmNactivateCallback,Button,0);
 	AddCB(w,XmNactivateCallback,&List::Button,0);	
 	//AH Changed End

	return w;
}


bool List::Append(Value* v)
{
	if (vals.Append(v))
	{
		UpdateWidget();
		return true;
	}
	return false;
}


bool List::Insert(Value* v, int pos)
{
	if (vals.Insert(v,pos))
	{
		UpdateWidget();
		return true;
	}
	return false;
}


bool List::Remove(Value* v)
{
	bool r = vals.Remove(v);
	if (r) delete v;
	UpdateWidget();
	return r;
}


void List::Clear(bool update)
{
	Iterator<Value*> iter(vals);
	
	for (Value* v=iter; v; v=++iter) delete v; 
	vals.Clear();
	
	if (update)
		UpdateWidget();
}



void List::UpdateWidget()
{
	if (list)
	{
		XmListDeleteAllItems(list);
		
		Iterator<Value*> iter(vals);
		int cnt = 1;
		for (Value* v=iter; v; cnt++,v=++iter)
		{
			XmString string = BuildString(v,cnt);
			XmListAddItem(list,string,0);
			XmStringFree(string);
		}
	}
}


Widget List::CreateWidget(Widget parent)
{
	form = XtVaCreateManagedWidget("List",xmFormWidgetClass,parent,NULL);
	
	toolbar = XtVaCreateManagedWidget("ListToolbar",
		xmRowColumnWidgetClass,form,NULL);
	CreateButton(toolbar,"Append");
	CreateButton(toolbar,"Insert");
	CreateButton(toolbar,"Change");
	CreateButton(toolbar,"Remove");
	CreateButton(toolbar,"Cut");
	CreateButton(toolbar,"Copy");
	CreateButton(toolbar,"Paste");

	Widget form2 = XtVaCreateManagedWidget("ListDefault",
		xmFormWidgetClass,form,NULL);
	XtVaCreateManagedWidget("Label",xmLabelWidgetClass,form2,NULL);
	def->CreateWidget(form2);
	
	XtManageChild(list = XmCreateScrolledList(form,"ListList",NULL,0));
	CatchDestroy(&list,&form,&toolbar);
	//AH Changed Start Selection => &Selection
 	AddCB(list,XmNdefaultActionCallback,&List::Selection,1);
 	AddCB(list,XmNextendedSelectionCallback,&List::Selection,0);
 	AddCB(list,XmNmultipleSelectionCallback,&List::Selection,0);
 	AddCB(list,XmNsingleSelectionCallback,&List::Selection,0);
	//AH Changed End
	
	UpdateToolbar(0);
	UpdateWidget();
	
	return form;
}


bool List::EvalWidget()
{
	Clear(false);

	XmStringTable item; int item_cnt;
	XtVaGetValues(list,XmNitems,&item,XmNitemCount,&item_cnt,NULL);
	for (int i=0; i<item_cnt; i++)
	{
		istrstream is(NormalString(item[i]));
		Value* v = def->Copy();
		if (is>>*v)
			vals.Append(v);
		else
			delete v;
	}	
	
	return true;
}


Value& List::operator =(const List& l)
{
	Clear(false);
	
	Iterator<Value*> iter(l.vals);
	for (Value* v=iter; v; v=++iter)
		vals.Append(v->Copy());
	
	UpdateWidget();
	return *this;
}


Value& List::operator =(const Value& v)
{
	if (v.IsList())
		*this = (List&)v;
	
	return *this;
}
