#include <strstream.h>
#include <fstream.h>
  
extern "C" {
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
}

#include "ident.h"
#include "attrib.h"
#include "netcl.h"
#include "scope.h"
#include "nodecl.h"
#include "conncl.h"
#include "node.h"


#ifndef NoExplicitInstanciation
template class ClassGroup<NodeClass>;
template class ClassGroup<ConnectorClass>;
#endif

LinkedList<void*> NetClass::netcls;


NetClass *NetClass::GetNetClass(char* id, char* p)
{
	Iterator<void*> iter(netcls);
	for (void* nt=iter; nt; nt=++iter)
	{
		if (NetClassP(nt)->name==id)
			return NetClassP(nt);
	}

	return Load(p,id);
}
	

NetClass* NetClass::Load(char* npath, char* name)
{	
	struct stat st;
	Text path = Text(npath)+"/"+name;
	
	// only regular files
	//AH Changed Start: to_charPtr()
	if (stat(path.to_charPtr(),&st)==0 && !S_ISREG(st.st_mode))
	  //AH Changed End
		return 0;
	
	NetClass* n = new NetClass(0,path);
	ifstream fin(path.to_charPtr());
	if (fin && fin>>*n)
		return n;

	delete n;
	return 0;
}

	
int NetClass::LoadExternals(char* path)
{
	DIR* dir = opendir(path);
	struct dirent* entry;
	NetClass* nc;
	int cnt = 0;
	
	if (!dir) return 0;
	// readdir in MT-Umgebungen durch readdir_r ersetzen?!
	for (entry=readdir(dir); entry; entry=readdir(dir))
		if (((nc = GetNetClass(entry->d_name))==0 || nc->IsBuiltin()) 
			&& Load(path,entry->d_name))
		{
			if (nc) delete nc; // externals overriden builtins
			cnt++;
		}

	closedir(dir);
	
	return cnt;
}

	
NodeClass* NetClass::GetNodeClass(const Ident id)
{
	Iterator<NodeClass*> iter(nodecls);
	
	for (NodeClass* o=iter; o; o=++iter)
		if (o->Name()==id)
			return o;
	
	return 0;
}


LinkedList<NodeClass*> NetClass::NodeClassesOfName(Ident& id)
{
	LinkedList<NodeClass*> dest;
	Iterator<NodeClass*> iter(nodecls);
	
	for (NodeClass* nc=iter; nc; nc=++iter)
		if (nc->Name()==id)
			dest.Append(nc);
	
	OpaqueIterator<ClassGroup<NodeClass> > giter(nodegroups);
	for (ClassGroup<NodeClass>* g=giter; g; g=++giter)
		if (g->name==id)
		{
			Iterator<NodeClass*> iter(g->members);
			for (NodeClass* o=iter; o; o=++iter)
					dest.Append(o);
			break;
		}
	
	return dest;
}


ConnectorClass* NetClass::GetConnectorClass(const Ident id)
{
	Iterator<ConnectorClass*> iter(conncls);
	
	for (ConnectorClass* o=iter; o; o=++iter)
		if (o->Name()==id)	
			return ConnectorClassP(o);
	
	return 0;
}


ObjectClass* NetClass::GetObjectClass(const Ident id)
{
	Iterator<ObjectClass*> iter(objcls);
	
	for (ObjectClass* o=iter; o; o=++iter)
		if (o->Name()==id)	
			return o;
	
	return 0;
}


NetClass::NetClass(char* input, char* p): 
	subclass(false),
	checknames(true),
	builtin(input!=0),
	cutbuffer(NetP(0),this,NodeP(0))
{
	netcls.Append(this);
	
	if (input)
	{
		istrstream is(input);
		is>>*this;
	}
	
	char* sl = p ? strrchr(p,'/') : 0;
	if (sl)
	{
		*sl = 0;
		path = p;
		*sl++ = '/';
		name = sl;
	}
}


//AH Changed Start constructor added
NetClass::NetClass(char* input, Text t):
 	subclass(false),
	checknames(true),
	builtin(input!=0),
	cutbuffer(NetP(0),this,NodeP(0))
{
  char *p = t.to_charPtr();
  	netcls.Append(this);
	
	if (input)
	{
		istrstream is(input);
		is>>*this;
	}
	
	char* sl = p ? strrchr(p,'/') : 0;
	if (sl)
	{
		*sl = 0;
		path = p;
		*sl++ = '/';
		name = sl;
	}

}
//AH Changed End



NetClass::~NetClass()
{
	Iterator<Attrib*> aiter(types);
	for (Attrib* a=aiter; a; a=++aiter) delete a; 
	
	Iterator<ObjectClass*> oiter(objcls);
	for (ObjectClass* o=oiter; o; o=++oiter) delete o; 
	
	netcls.Remove(this);
}


Value* NetClass::GetType(const Ident id)
{
	Iterator<Attrib*> iter(types);
	
	for (Attrib *a=iter; a; a=++iter)
		if (a->Name()==id)
			return a->Type();

	return 0;
}


void NetClass::AddNodeClass(NodeClass* oc)
{
	Ident gname = oc->GroupName();
	//AH Changed Start .to_bool()
	if (gname.to_bool())
	  //AH Changed End
	{
		OpaqueIterator<ClassGroup<NodeClass> > iter(nodegroups);
		ClassGroup<NodeClass>* g;
		
		for (g=iter; g; g=++iter)
			if (g->name==gname) break;
		if (g) g->Add(oc); else nodegroups.Append(ClassGroup<NodeClass>(gname,oc));
	}
	else
		nodegroups.Append(ClassGroup<NodeClass>(oc->Name(),oc));
	
	nodecls.Append(oc);
	objcls.Append(oc);
}


void NetClass::AddConnectorClass(ConnectorClass* oc)
{
	Ident gname = oc->GroupName();
	//AH Changed Start .to_bool()
	if (gname.to_bool())
	  //AH Changed End
	{
		OpaqueIterator<ClassGroup<ConnectorClass> > iter(conngroups);
		ClassGroup<ConnectorClass>* g;
		
		for (g=iter; g; g=++iter)
			if (g->name==gname) break;
		if (g)
			g->Add(oc);
		else
			conngroups.Append(ClassGroup<ConnectorClass>(gname,oc));
	}
	else
		conngroups.Append(ClassGroup<ConnectorClass>(oc->Name(),oc));
	
	conncls.Append(oc);
	objcls.Append(oc);
}



istream& NetClass::Input(istream& is)
{
	Ident id(false);
	if (is>>id && (id=="netclass" || id=="subclass"))
	{
		subclass = id=="subclass";
		Scope snt(is);
		while (is && snt.GetId(id))
		{
			if (id=="name")
			{
				if (!(is>>name))
					cerr<<"Name of netcl expected!"<<endl;
			}
			else if (id=="attributes")
			{
				attr.Parse(is);
			}
			else if (id=="checknames")
			{
				if (is>>id)
					checknames = id=="true";
			}
			else if (id=="types")
			{
				Scope st(is);
				while (is && st.Get())
				{
					Attrib *t = new Attrib(this);
					if (is>>*t)
						types.Append(t);
					else
						delete t;
				}
			}
			else if (id=="nodes")
			{
				Scope sn(is);
				while (is && sn.Get())
				{
					NodeClass *nc = new NodeClass(*this);
					if (is>>*nc)
						AddNodeClass(nc);
					else
						delete nc;
						
				}
			}
			else if (id=="connectors")
			{
				Scope sc(is);
				while (is && sc.Get())
				{
					ConnectorClass *cc = new ConnectorClass(*this);
					if (is>>*cc)
						AddConnectorClass(cc);
					else
						delete cc;
				}
			}
			else
			{
				cerr<<"unknown identifier '"<<id<<"'!"<<endl;
			  is.clear(ios::badbit);
			}
		}
		return is;
	}
//	else
//		cerr<<"Keyword 'netclass' expected!"<<endl;
  is.clear(ios::badbit);
	return is;
}


ConnectorClass* NetClass::FittingConnectorClass(Node* in, Node* out)
{
	Iterator<ConnectorClass*> iter(conncls);
	
	for (ConnectorClass* o=iter; o; o=++iter)
		if (o->AllowedConnection(in,out))
			return o;
	
	return 0;
}

/*
NetClass::Group* NetClass::GetGroup(ObjectClass* oc)
{
	OpaqueIterator<Group> iter(groups);
	for (Group* g=iter; g; g=++iter)
		if (g->IsIn(oc)) return g;
	return 0;
}
*/

void NetClass::EnterEditor(ObjectBar<NodeClass>& nodebar, ObjectBar<ConnectorClass>& connbar)
{
	OpaqueIterator<ClassGroup<NodeClass> > niter(nodegroups);
	
	for (ClassGroup<NodeClass>* ng=niter; ng; ng=++niter)
		nodebar.AddButton(ng->name,ng->Members());
	
	OpaqueIterator<ClassGroup<ConnectorClass> > citer(conngroups);
	for (ClassGroup<ConnectorClass>* cg=citer; cg; cg=++citer)
		connbar.AddButton(cg->name,cg->Members());
}
