#include "xbefore.h"
#include <Xm/RowColumn.h>
#include <Xm/DrawnB.h>
#include <Xm/PushB.h>
#include <Xm/MenuShell.h>
#include "xafter.h"

#include <iostream.h>

#include "objbar.h"
#include "tool.h"

//AH Changed Start: numerous &ObjectBar<T>::'s added  3x .to_charPtr() added ...
//AH Changed End


template<class T> void ObjectBar<T>::AddButton(char* name, Iterator<T*> iter)
{
	Button* b = new Button(name,iter);
	buttons.Append(b);
	if (int(buttons)==1) curr = b;
}


template<class T> void ObjectBar<T>::ButtonState(char* name, bool enable)
{
	for (Button* b=biter; b; b=++biter)
		if (b->name==name)
			b->State(enable);
}


template<class T> void ObjectBar<T>::ClearButtons()
{
	if (build_id)
		RemoveCB(build_id);
	build_id = 0;
		
	for (Button* b=biter; b; b=++biter) delete b; 
	buttons.Clear();
	curr = 0;
}


template<class T> ObjectBar<T>::~ObjectBar()
{
	ClearButtons();
}

	
template<class T> void ObjectBar<T>::EnterParent(Widget parent)
{
	if (build_id)
		RemoveCB(build_id);
	
	build_id = TimerCB(build_delay,&ObjectBar<T>::Build,(void*)parent);
}

template<class T> bool ObjectBar<T>::Build(void* udata)
{
	Widget parent = Widget(udata);
	build_id = 0;
	
	XtVaGetValues(parent,XmNbackground,&unsel,XmNtopShadowColor,&sel,NULL);
	
	if (int(buttons))
	{
		bar = XtVaCreateManagedWidget("ObjectBar",
			xmRowColumnWidgetClass,parent,NULL);

		for (Button* b=biter; b; b=++biter)
		{
			Widget box = XtVaCreateManagedWidget("Box",xmRowColumnWidgetClass,bar,NULL);
			b->button = XtVaCreateManagedWidget("Button",xmDrawnButtonWidgetClass,box,NULL);
			b->label = XtVaCreateManagedWidget("Name",xmPushButtonWidgetClass,box,NULL);
			AddCB(b->button,XmNarmCallback,&ObjectBar<T>::Click,b);
			AddCB(b->label,XmNarmCallback,&ObjectBar<T>::Popup,b);
			AddCB(b->button,XmNexposeCallback,&ObjectBar<T>::Draw,b);
		}
		if (curr) Select(curr);
	}
	
	return false;
}


template<class T> void ObjectBar<T>::LeaveParent()
{
	if (build_id)
		RemoveCB(build_id);
	build_id = 0;
	
	if (bar) XtDestroyWidget(bar);
	bar = 0;
}


template<class T> void ObjectBar<T>::Draw(Widget bw, void* butt, XtPointer)
{
	Dimension b,hlt,mh,mw,w,h;
	
	XtVaGetValues(bw,
		XmNwidth,&w,XmNheight,&h,
		XmNmarginWidth,&mw,XmNmarginHeight,&mh,
		XmNshadowThickness,&b,
		XmNhighlightThickness,&hlt,NULL);
	
	int x = w/2,y = h/2;
	if (h<w) w = h;
	b += hlt; w -= 2*b+2*mw;
	
	((Button*)butt)->curr->DrawButton(XtDisplay(bw),XtWindow(bw),x,y,w,w);
}


template<class T> void ObjectBar<T>::Click(Widget, void* butt, XtPointer calld)
{
	Button* button = (Button*)butt;
	XmPushButtonCallbackStruct* pbcs = (XmPushButtonCallbackStruct*)calld;
	
	switch (pbcs->event->xbutton.button)
	{
		case 1: if (curr==button) button->Prev(); Select(button); break;
		case 2: Select(button); break;
		case 3: if (curr==button) button->Next(); Select(button); break;
	}
}


template<class T> void ObjectBar<T>::Popdown(Widget, void* obc, XtPointer)
{
	T* o = (T*)obc;

	if (!o)
	{
		if (popup) XtDestroyWidget(popup);
		popup = 0;
	}
	else
	{
		curr->curr = o;
		Select(curr);
	}
}


template<class T> void ObjectBar<T>::Popup(Widget, void* butt, XtPointer calld)
{
	Button* button = (Button*)butt;
	
	if (int(button->alts))
	{
		XButtonEvent& ev = ((XmPushButtonCallbackStruct*)calld)->event->xbutton;
		popup = XmCreatePopupMenu(XtParent(*button),"Popup",NULL,0);

		for (T* o=button->alts; o; o=++button->alts)
		{
			Widget ent = XtVaCreateManagedWidget(
				(o->Name()).to_charPtr(),xmPushButtonWidgetClass,popup,NULL);
			SetStringValue(ent,XmNlabelString,(o->Label()).to_charPtr());
			AddCB(ent,XmNactivateCallback,&ObjectBar<T>::Popdown,o);
		}
		AddCB(popup,XmNunmapCallback,&ObjectBar<T>::Popdown,(void*)0);
		if (button!=curr) Select(button);
		ev.x_root -= ev.x; ev.y_root -= ev.y; ev.x = ev.y = 0;
		XmMenuPosition(popup,&ev);
		XtManageChild(popup);
	}
}


template<class T> void ObjectBar<T>::Button::State(bool newstate)
{
	if (state!=newstate)
	{
		XtSetSensitive(button,(state = newstate)?True:False);
		XtSetSensitive(label,state?True:False);
	}
}


template<class T> void ObjectBar<T>::Button::Next()
{
	T* o;

	for (o=alts; o; o=++alts)
		if (o==curr) break;
	
	curr = ((o = ++alts)?o:alts);
}


template<class T> void ObjectBar<T>::Button::Prev()
{
	T* o,* b = alts.Last();
	
	for (o=alts; o; b=o,o=++alts)
		if (o==curr) break;
	
	curr = b;
}


template<class T> void ObjectBar<T>::Select(void* butt)	
{
	Button* button = (Button*)butt;
	
	curr = button;
	for (Button* b=biter; b; b=++biter)
	{
		bool set = curr==b;
		XtVaSetValues(*b,XmNmappedWhenManaged,False,NULL);
		XtVaSetValues(*b,XmNshadowType,set?XmSHADOW_IN:XmSHADOW_OUT,
			XmNbackground,set?sel:unsel,XmNmappedWhenManaged,True,NULL);
		
		if (b->curr->Label()!=b->curr_label)
			SetStringValue(b->label,XmNlabelString,(b->curr->Label()).to_charPtr());
	}
}


// for g++
#ifndef NoExplicitInstanciation
#define NoExplicitInstanciation

#include "nodecl.h"
#include "conncl.h"
#include "wrapper.cc"

template class ObjectBar<NodeClass>;
template class Wrapper<ObjectBar<NodeClass>,void*>;

template class ObjectBar<ConnectorClass>;
template class Wrapper<ObjectBar<ConnectorClass>,void*>;

#endif
