#include <X11/Xlib.h>
#include "pline.h"
#include "scope.h"

#include "constants.h"

SHAPE_NOTIFY("pline",Pline)



Pline::~Pline()
{
	if (segs) delete[] segs;
	if (arcs) delete[] arcs;
	if (pnts) delete[] pnts;
	Iterator<Position*> iter(p);
	for (Position *pos=iter; pos; pos=++iter) delete pos; 
}


double Pline::Distance(ShapeData& sd, Position& ps, Position& best)
{
	double bestd = 1e300,testd;
	LinkedList<Position> l(Positions(sd));
	OpaqueIterator<Position> iter(l);

	Position* p1 = iter,*p2,test,psz = ps*sd.zoom;
	if (p1) bestd = (best = *p1).Distance(psz); else return bestd;
	
	while ((p2 = ++iter))
	{
		if (p1->x==p2->x)
			test = Position(p1->x,ps.YLimit(p1->y,p2->y).y);
		else if (p1->y==p2->y)
			test = Position(ps.XLimit(p1->x,p2->x).x,p1->y);
		else
		{
			double  m1 = (p2->y-p1->y)/(p2->x-p1->x),m2 = -1/m1,
			b1 = p1->y-m1*p1->x, b2 = psz.y-m2*psz.x;
			test = Position((b2-b1)/(m1-m2),0.0).XLimit(p1->x,p2->x);
			test.y = m1*test.x+b1;
		}
		if ((testd = test.Distance(psz))<bestd)	{ best = test; bestd = testd; }
		p1 = p2;
	}
	return bestd/sd.zoom;
}


static bool intersect_lines(Position a, Position b, Position c, Position d,Position& p)
{
  double eps = 1e-8;
  double m11 = b.x-a.x; double m12 = c.x-d.x; double m13 = c.x-a.x;
  double m21 = b.y-a.y; double m22 = c.y-d.y; double m23 = c.y-a.y;
  double my, lambda;

  if (fabs(m21)<eps) {
    if (fabs(m22)<eps || fabs(m11)<eps) return false;
    lambda = m23/m22;
    my = (m13-m12*lambda)/m11;
  }
  else {
    double q = m12-m11*m22/m21;
    if (fabs(q)<eps) return false;
    lambda = (m13-m11*m23/m21)/q;
    my = (m23-m22*lambda)/m21;
  }
  if (lambda<0 || lambda>1 || my<0 || my>1) return false;
  p = a+(b-a)*my; 
  return true;
}


Position Pline::PointToMid(ShapeData& sd, Position& ps)
{
	LinkedList<Position> l(Positions(sd));
	OpaqueIterator<Position> iter(l);

	Position* p1 = iter, *p2;
  Position test;
  double testd;

  Position best;
  double bestd = 1e300;
  bool found = false;
  
  Position psz = ps*sd.zoom;
  Position middle = sd.pos;
  
	while ((p2 = ++iter))
	{
    if (intersect_lines(*p1, *p2, psz, middle, test)) {
      found = true;
      if ((testd = test.Distance(psz))<bestd) { best = test; bestd = testd; }
    }
		p1 = p2;
	}
  if (!found) Distance(sd, ps, best);
	return best;
}


int Pline::Inside(Display*, Position pos, ShapeData& sd, ExprData&)
{
	if (int(p))
	{
		pos *= sd.zoom;
		LinkedList<Position> l(Positions(sd));
		
		sd.pos *= sd.zoom;
		Position *p; double d = pos.Distance(sd.pos);
		OpaqueIterator<Position> iter(l);
		for (p=iter; p; p=++iter)
			if (d<=p->Distance(sd.pos)) break;
		sd.pos /= sd.zoom;
		//AH Changed Start .to_bool()
		if (!pos.to_bool()) return 0;
		//AH Changed End

		Region reg = XPolygonRegion(pnts,mpnts,EvenOddRule);
		bool inside = XPointInRegion(reg,int(pos.x),int(pos.y))==True;
		XDestroyRegion(reg);
		return inside?RelativePos():0;
	}
	else // Connector-Pline (ohne eigene Punkte)
	{
		Position best;
		return Distance(sd,pos,best)<4?RelativePos():0;
	}
}


void Pline::DoDraw(Display* dpy, Drawable d, GC gc, ShapeData& sd, ExprData&)
{
	LinkedList<Position> l(Positions(sd));
	
	if (context.filled)
		XFillPolygon(dpy,d,gc,pnts,mpnts,CoordModeOrigin,Complex);
	else
		if (context.spline)
		{
			Spline(l,context.spline*sd.zoom);
			XDrawSegments(dpy,d,gc,segs,nsegs);
			XDrawArcs(dpy,d,gc,arcs,narcs);
		}
		else
			XDrawLines(dpy,d,gc,pnts,mpnts,CoordModeOrigin);
}


LinkedList<Position> Pline::Positions(ShapeData& sd)
{
	LinkedList<Position> d;
	Iterator<Position*>& iter = int(p)?piter:sd.iter;

	if (!pnts || mpnts!=int(iter))
	{
		if (pnts) delete[] pnts;
		pnts = new XPoint[mpnts = int(iter)];
	}

  XPoint *wp = pnts;
  for (Position *pos=iter; pos; pos=++iter)
  {
    Position scaled = int(p)?Absolute(sd,*pos):(*pos*sd.zoom);
    d.Append(scaled);
    *wp++ = scaled;
  }
  // Let connector lines end at relpos 0 & 1
  // idler
  if (!p) {
    if (sd.rels) {
      Position diff = *d[1] - *d[0];
      
      *d[0] += sd.rels[0]->Rotate(diff.Angle()) * sd.scale * sd.zoom;
      pnts[0] = *d[0];
    }

    if (int(sd.rels)>1) {
      int l = d.Length()-1;
      Position diff = *d[l] - *d[l-1];
   
      *d[l] += sd.rels[1]->Rotate(diff.Angle()) * sd.scale * sd.zoom;
      pnts[l] = *d[l];
    }
  }
  
	return d;
}


Metrics Pline::GetMetrics(Display*,ShapeData& sd,ExprData&)
{
  LinkedList<Position> poslist(Positions(sd));
  OpaqueIterator<Position> iter(poslist);

  Metrics akt_metrics;

  if (poslist.Length() < 2)
    cout << "Pline.cc,GetMetrics(): Pline has got less than two Positions!\n";
  for(Position* pos=iter; pos; pos=++iter)
    akt_metrics.Compare(*pos);

  return akt_metrics;
}


void Pline::Spline(LinkedList<Position>& d, double spline)
{
	if (int(d)<2) return;
	
	OpaqueIterator<Position> pos(d);
	Position a0 = *PositionP(pos),a1,a2 = *++pos;
	
	if (!segs || msegs!=int(pos))
	{
		if (segs) delete[] segs;
		segs = new XSegment[msegs = int(pos)];
	}

	if (!arcs || marcs!=int(pos))
	{
		if (arcs) delete[] arcs;
		arcs = new XArc[marcs = int(pos)];
	}

	nsegs = 0; narcs = 0;

	if (int(pos)>2) for (int i=0; i < int(pos)-2; i++)
	{
		a1 = a2; a2 = *++pos;

		double dx0 = a0.x - a1.x, dy0 = a0.y - a1.y;
		double dx2 = a2.x - a1.x, dy2 = a2.y - a1.y;
		double l1 = sqrt(dx0 * dx0 + dy0 * dy0);
		double l2 = sqrt(dx2 * dx2 + dy2 * dy2);
		if (i != int(pos)-3) l2 /= 2.0;
		double r = (l1 > l2) ? l2 : l1;
		if (r > spline) r = spline;
		if (r < 0) r = 0.0;

		double sql0 = dx0 * dx0 + dy0 * dy0;
		double sql2 = dx2 * dx2 + dy2 * dy2;
		double num = dy0 * dx2 - dy2 * dx0;
		double denom = sqrt(sql0 * sql2) - (dx0 * dx2 + dy0 * dy2);
		
		if (fabs(num) < 1.0e-6 || fabs(denom) < 1.0e-6 )
		{	
			XSegment seg = {int(a0.x+.5),int(a0.y+.5),int(a1.x+.5),int(a1.y+.5) };
			segs[nsegs++] = seg;
			a0 = a1;
		}
		else
		{
			double dist = fabs(r * num / denom);
			double l0 = dist / sqrt(sql0), l2 = dist / sqrt(sql2);
			double xt0 = a1.x + dx0 * l0, yt0 = a1.y + dy0 * l0;
			double xt2 = a1.x + dx2 * l2, yt2 = a1.y + dy2 * l2;
			double sqlt = (xt0-a1.x)*(xt0-a1.x) + (yt0-a1.y)*(yt0-a1.y);
			double xm = (xt0+xt2) / 2.0, ym = (yt0+yt2) / 2.0;
			double sqlcrt = sqrt(sqlt + r*r);
			double sqlmrt = sqrt((xm-a1.x)*(xm-a1.x) + (ym-a1.y)*(ym-a1.y));
			double xc = a1.x - (a1.x-xm) * sqlcrt / sqlmrt;
			double yc = a1.y - (a1.y-ym) * sqlcrt / sqlmrt;
			double sqlt0 = sqlength(a1.x, a1.y, xt0, yt0);
			double sqlt2 = sqlength(a1.x, a1.y, xt2, yt2);
			
			if (sql0 < sqlt0 || sql2 < sqlt2)
			{
				sql0 = sqrt(sql0); sql2 = sqrt(sql2);
				sqlt0 = sqrt(sqlt0); sqlt2 = sqrt(sqlt2);
				l0 = sqlt0 - sql0; l2 = sqlt2 - sql2;
				
				if (l0 > l2)
				{
					xt0 = a0.x; yt0 = a0.y;
					dist = sql0/sqlt2;
					xt2 = a1.x + (xt2-a1.x) * dist;
					yt2 = a1.y + (yt2-a1.y) * dist;
					r = r * sql0/sqlt0;
				}
				else
				{
					xt2 = a2.x; yt2 = a2.y;
					dist = sql2/sqlt0;
					xt0 = a1.x + (xt0-a1.x) * dist; 
					yt0 = a1.y + (yt0-a1.y) * dist; 
					r = r * sql2/sqlt2;
				}
				xc = a1.x + (xc - a1.x) * dist;
				yc = a1.y + (yc - a1.y) * dist;
			}
	
			XSegment seg = { int(a0.x+.5),int(a0.y+.5),int(xt0+.5),int(yt0+.5) };
			segs[nsegs++] = seg;
		
			int angle0 = int(.5 + atan2(yc-yt0, xt0-xc) / M_PI * 180.0 * 64.0 );
			int angle1 = int(.5 + atan2(yc-yt2, xt2-xc) / M_PI * 180.0 * 64.0 );
		
			if (angle0 < 0) angle0 += 360*64;
			if (angle1 < 0) angle1 += 360*64;
		
			if (angle1 < angle0)
			{
				int help = angle1; angle1 = angle0; angle0 = help;
			}
		
			angle1 -= angle0;
			if (angle1 > 180*64) angle1 -= 360*64;
		
			int midx = int(xc-r+.5),midy = int(yc-r+.5);
			unsigned xrad = unsigned(r*2+.5+(xc-r-midx)),
				yrad = unsigned(r*2+.5+(yc-r-midy));
			XArc arc = { midx,midy,xrad,yrad,angle0,angle1 };
			arcs[narcs++] = arc;
		
			a0.x = xt2;
			a0.y = yt2;
		}
	}

	XSegment seg = {int(a0.x+.5),int(a0.y+.5),int(a2.x+.5),int(a2.y+.5) };
	segs[nsegs++] = seg;
}


istream& Pline::Input(istream& is, int)
{
	Position* pos;
	if (is>>*(pos = new Position)) p.Append(pos); else delete pos;
	return is;
}

/*
ostream& Pline::Output(ostream& os)
{
	Iterator<Position*> iter(p);
	for (Position *pos=iter; pos; pos=++iter) os<<','<<*pos; 
}
*/

























