#include "xbefore.h"
#include <Xm/TextF.h>
#include <Xm/ArrowB.h>
#include <Xm/PushB.h>
#include <Xm/Form.h>
#include "xafter.h"

#include <strstream.h>

#include "posint.h"
#include "scope.h"

VALUE_NOTIFY("posint",PosInt)


PosInt::PosInt(NetClass* nc, bool hidden): 
	Value(nc,hidden),
	Wrapper<PosInt,bool>(this),
	val(0),
	max(MAXINT),
	min(MININT),
	form(0),
	text(0),
	up_r(300,50),
	dn_r(300,50),
	dontcare(false)
{
}


PosInt::PosInt(int v): 
	Wrapper<PosInt,bool>(this),
	max(MAXINT),
	min(MININT),
	form(0),
	text(0),
	up_r(300,50),
	dn_r(300,50),
	dontcare(false)
{
  if (v<0)
    {
      val=0;
      dontcare=true;
    }
  else
    {
      val=v;
      dontcare = false;
    }
}


PosInt::PosInt(const PosInt& i): 
	Value(i),
	Wrapper<PosInt,bool>(this),
	val(i.val),
	max(i.max),
	min(i.min),
	form(0),
	text(0),
	up_r(300,50),
	dn_r(300,50),
	dontcare(i.dontcare)
{
}


istream& PosInt::Parse(istream& is)
{
	Scope si(is,0,2);	
	min = 0;
	while (is && si.Get())
		switch (si.field)
		{
			case 1:
				if (!(is>>min))
					cerr<<"integer (min. value) expected!"<<endl;
				break;
			case 2:
				if (is>>max)
				{
					if (max<min)
					{
						int tmp = max;
						max = min;
						min = tmp;
					}
				}
				else
					cerr<<"integer (max. value) expected!"<<endl;
				break;
		}
	return is;
}


istream& PosInt::Input(istream& is)
{
  char c = 0;
  if( (is>>c) && (c=='*'))
    {
      *this = 0;
      dontcare = true;
      return is;
    }
  else 	is.putback(c);
  
  if (is>>val)
    {
      *this = val;
      return is;	
    }
  else
    {
      cerr<<"integer expected!"<<endl;
      is.clear(ios::badbit);
      return is;
    }
}


void PosInt::Arrow(Widget, bool up, XtPointer calldata)
{
	// XmPushButtonCallbackStruct* abcs = (XmPushButtonCallbackStruct*)calldata;
	// bool dclick = abcs->click_count>1;
	char* buf=XmTextFieldGetString(text);
	istrstream istr(buf);
	if(buf[0]=='*')
	  {
	    val=0;
	    dontcare=true;
	  }

	istr>>val;
	
    if(up)
      if(dontcare)
	{
	  val=0;
	  dontcare = false;
	}
      else
	{
	  if(val>=max)
	    val=max;
	  else
	    val=val+1;
	}
    else
      {
	if(val==0)
	  dontcare=true;
	else
	  {
	    val=val-1;
	  }                               
      }

	UpdateWidget();
}


void PosInt::UpdateWidget()
{
	char buf[32]; ostrstream ostr(buf,32);
	
	if(dontcare)
	  ostr<<"*"<<ends;
	else
	  ostr<<val<<ends;
	XmTextFieldSetString(text,buf);
	XmTextFieldSetInsertionPosition(text,XmTextFieldGetLastPosition(text));
}


Widget PosInt::CreateWidget(Widget parent)
{
	Widget up,dn;
	form = XtVaCreateManagedWidget("PosInt",xmFormWidgetClass,parent,NULL);
	dn_r = dn = XtVaCreateManagedWidget("Dn",xmPushButtonWidgetClass,form,NULL);
	text = XtVaCreateManagedWidget("Edit",xmTextFieldWidgetClass,form,NULL);
	up_r = up = XtVaCreateManagedWidget("Up",xmPushButtonWidgetClass,form,NULL);
	CatchDestroy(&form,&text);
 	//AH Changed Start from Arrow Arrow Edit => &Arrow &Arrow &Edit
	// &Int::Arrow &Int::Arrow &Int::Edit
 	AddCB(up,XmNarmCallback,&PosInt::Arrow,true);
 	AddCB(dn,XmNarmCallback,&PosInt::Arrow,false);
 	AddCB(text,XmNvalueChangedCallback,&PosInt::Edit,true);
 	//AH Changed End
	UpdateWidget();
	return form;
}

bool PosInt::EvalWidget()
{
	if (text)
	{
		istrstream istr(XmTextFieldGetString(text));
		if (istr>>*this) UpdateWidget();
		return !istr.fail();
	}
	cerr<<"Evaluating nonexisting Widget!"<<endl;
	return true;
}

Value& PosInt::operator =(const int i)
{
	if ((val = i)>max) val = max; else if (val<min) val = min;
	if(val<0) dontcare = true;
	if (text)
		UpdateWidget();
	
	return *this;
}


//AH Changed Start  PosInt::operator char*() const
char* PosInt::to_charPtr() const
//AH Changed End
{
	static char buf[64]; ostrstream ostr(buf,64);
	if(dontcare)
	  ostr<<"*"<<ends;
	else
	  ostr<<val<<ends;
	return buf;
}
