#include "xbefore.h"
#include <Xm/Text.h>
#include <Xm/ScrolledW.h>
#include <Xm/TextF.h>
#include "xafter.h"
#include <iostream.h>
extern "C" {
#include <stdio.h>
#include <string.h>
}
extern "C" {   
  char* rindex(const char*, int); 
  char* index(const char*, int); 
}

#include "text.h"
#include "scope.h"
#include "float.h"
#include "int.h"

VALUE_NOTIFY("text",Text)

//AH Changed Start
  //inserted some const_cast<char*>
  //AH Changed End


Text::Text(NetClass* nc, bool hidden):
	 Value(nc,hidden),
	 scroll(0),text(0),
	 rows(1),
	 p(0)
{
}

		 
Text::Text(const Text& t): 
	Value(t),
	scroll(0),text(0),
	rows(t.rows),
	p(0)
{
	*this = t.p;
}


Text::Text(const char *s): 
	scroll(0),text(0),
	rows(1),
	p(0)
{
	*this = s;
}


istream& Text::Parse(istream& is)
{
	Scope sb(is,0,1);	
	while (is && sb.Get())
		if (!(is>>rows)) cerr<<"integer (number of rows) expected!"<<endl;

	return is;
}


istream& Text::Input(istream& is)
{
	char c,*b = 0,reading = 0; long len = 1; // Nullbyte reservieren
	
	if (is>>c && c=='"')
	{
		streampos start = is.tellg();
_again:	while (is.get(c))
			switch (c)
			{
			case '"':
				if (reading)
				{
					*b = 0;
					if (text) XmTextSetString(text,p);
					return is;
				}
				else
				{
					if (p) delete [] p;
					p = b = new char[len];
					reading++;
					is.seekg(start);
					goto _again;
				}
			case '\\':
				is.get(c);
				if (c=='n') c = '\n';
				//AH Changed Start: workaround to parse both
				// "\\\\" (from motif panel) and "\\"  
				// to "\\" as needed for paths
				//parse "\\"" to "\"" instead of """
				else if (c=='"') c = '\"';
				else if (c!='\\')
				  if (reading) *b++ = '\\'; else len++;
				//AH Changed End
				// fall thru
			default:
				if (reading) *b++ = c; else len++;
			}
	}
	else
		cerr<<"\" expected!"<<endl;
	is.clear(ios::badbit);
	return is;
}


ostream& Text::Output(ostream& os) const
{
  os<<'"';
	for (char *c=p; c && *c; c++)
		switch (*c)
		{
		case '\r':
			break;
		case '\n':
			os<<"\\n";
			break;
		case '"':
		case '\\':
			os<<'\\';
			// fall thru
		default:
			os<<*c;
		}
	return os<<'"';
}


Text Text::Left(int i) const
{
  if (i>=Length()) return *this;
  if (i<0) return (const char *) 0;
 
  char* c = new char[i+1];
  strncpy(c, p, i);
  c[i]=0;

  Text t;
  t.p = c;
  return t;
}


Widget Text::CreateWidget(Widget parent)
{
	if (rows>1)
	{
		scroll = XtVaCreateManagedWidget("Text",
			xmScrolledWindowWidgetClass,parent,NULL);
	
		text = XtVaCreateManagedWidget("Scrolled",
			xmTextWidgetClass,scroll,
			XmNeditMode,XmMULTI_LINE_EDIT,
			XmNrows,rows,NULL);
		CatchDestroy(&scroll,&text);
		XmTextSetString(text,const_cast<char*>(Sure(p)));
		return scroll;
	}
	else
	{
		text = XtVaCreateManagedWidget((rows==1 ? "Text" : "Ident"),
			xmTextFieldWidgetClass,parent,NULL);
		CatchDestroy(&text);
		XmTextSetString(text,const_cast<char*>(Sure(p)));
		return text;
	}
}
  

bool Text::EvalWidget()
{
	if (text)
		*this = XmTextGetString(text);
	else
		cerr<<"Evaluating nonexisting Widget!"<<endl;
	return true;
}

Text Text::FirstPathPart(char slash) const
  /*get first part of a path: 
    "apath/bpath/cpath/name.xx".FirstPathPart('/') => apath*/
{
  if(p)
    {
      char* tmp_c = index(p,slash);
      if (tmp_c && *tmp_c != '\0')
	return Left(tmp_c-p);
      else return Text();
    }
  else return Text();
}

Text Text::StripFirstPathPart(char slash) const
  /*strip first part of a path: 
    "apath/bpath/cpath/name.xx".FirstPathPart('/') => "bpath/cpath/name.xx*/
{
  char*  tmp = new char[2];
  sprintf(tmp, "%c", slash);
  Text t = *this - this->FirstPathPart(slash).to_charPtr() - tmp;
  return t;
}

Text Text::StripPath(char slash) const
  /*strips path fom text: 
    "path/path/path/name.xx".StripPath('/') => "name.xx"*/
{
  if(p)
    {
      char* tmp_c = rindex(p,slash);
      
      if (tmp_c) 		     tmp_c++;
      if (tmp_c && *tmp_c != '\0') return Text(tmp_c);
      else			   return Text(*this);
    }
  else return Text();
}

Text Text::StripLastPathPart(char slash) const
  /*
    strip last part of a part
    "path/path/path/name.xx".StripPath('/') => "path/path/path"*/
{
  if(p)
    {
      char* tmp_c = rindex(p,slash);
      if (!tmp_c || *tmp_c == '\0') return Text("");
      int len = tmp_c-p;
      char* ret  = new char[len+1];
      strncpy(ret,p,len);
      ret[len] = '\0';
      Text t;
      t.p = ret;
      return t;
    }
  else return Text();
}


//returns str between first and last spezf. char
Text Text::BetweenS(char* s1, char* s2) const
{
  //return empty text for null pointers
  if (!p || !s1 || !s2) 
    return Text();
  
  unsigned len0 = strlen(p);
  unsigned len1 = strlen(s1);
  unsigned len2 = strlen(s2);
  
  //return empty text if lengths don'r match  
  if (len1 > len0 || len2 > len0 || len1+len2 >= len0)
    return Text();
  
  bool       ok    = false;
  char*      tmp   = p;
  unsigned   start = 0;
  unsigned   stop  = 0;
  unsigned   count = 1;
  
  while (*tmp != '\0')
    {
      if (len1 >= strlen(tmp))
	{
	  ok = false;
	  break;
	}
      if (strncmp(tmp,s1,len1) == 0)
	{
	  ok    = true;
	  start = count;
	  break;
	}
      count++;
      tmp++;
    }
  
  if (!ok) return Text();
  
  while (*tmp != '\0')
    {
      if (len2 > strlen(tmp))
	{
	  ok = false;
	  break;
	}
      if (strncmp(tmp,s2,len2) == 0)
	{
	  ok   = true;
	  stop = count;
	  break;
	}
      count++;
      tmp++;
    }
  
  if (!ok) return Text();
  
  len1  -= 1;
  stop  -= 1;
  start += len1;
  
  char* tmp1 = p;
  int   len  = stop - start;
  
  for (unsigned i=0; i<start; i++, tmp1++) ;
  
  if (len <= 0 ) return Text();
  
  char* ret  = new char[len+1];
  
  if (len == 1)
    {
      sprintf(ret,"%c%c",*tmp1,'\0');
      return Text(ret);
    }
  
  strncpy(ret,tmp1,len);
  ret[len] = '\0';
  
  Text t;
  t.p=ret;
  
  return t;
}
//---------------------------------------------------------------------------
// return Text between char c1 and char c2, 
// empty Text if impossible
Text	Text::BetweenC(char c1, char c2) const
{
  if (!p)   return Text();
  
  char* start = index(p,c1); //pointer to fist occurence of c1
  if (!start) return Text();
  
  start++;
  
  char*    tmp      = start;
  unsigned stop_pos = 0;
  unsigned count    = 1;
  
  //search c2
  while (*tmp != '\0')
    {
      if (*tmp == c2)
	stop_pos = count-1;
      count++;
      tmp++;
    }
  if (stop_pos == 0) return Text();
  
  char* ret = new char[stop_pos+1];
  strncpy(ret,start,stop_pos);
  ret[stop_pos] = '\0';
  
  Text t;
  t.p = ret;
  return t;
}

//---------------------------------------------------------------------------
// replace all occurences of c1 with c2 

Text Text::ReplaceChar(char c1, char c2) const
{
  if (!p)   return Text();

  Text t = Text(*this);

  char*    tmp = t.p;
  while (*tmp != '\0')
    {
      if ((*tmp)==c1)
	(*tmp)=c2;
      tmp++;
    }
  return t;
}

//---------------------------------------------------------------------------
bool	Text::BeginsWith(const Text& t) const
{
  return (*this).BeginsWith(t.p);
}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

bool	Text::BeginsWith(const char* c) const
{
  if(!p) return false;
  if(!c) return true;
  char *tmp = strstr(p,c);
  if (c && (tmp==p))
    return true;
  else
    return false;
}

//---------------------------------------------------------------------------
bool	Text::IsIn(const Text& t) const
{
  return (*this).IsIn(t.p);
}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

bool	Text::IsIn(const char* c) const
{
  if(!p) return true;
  if (c && strstr(c,p))
    return true;
  else
    return false;
}
//---------------------------------------------------------------------------

bool	Text::Contains(const Text& t) const
{
  return (*this).Contains(t.p);
}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

bool	Text::Contains(const char* c) const
{
  if(!c) return true;
  if (p && strstr(p,c))
    return true;
  else
    return false;
}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
bool	Text::Contains(const char c) const
{
  char* tmp;  
  bool flag;
  tmp = new char[2];
  sprintf(tmp, "%c", c);
  flag = (*this).Contains(tmp);
  delete[] tmp;
  return flag;
}

//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Value& Text::operator =(const char* s) 
{
	if (p) delete[] p;
	if (s) strcpy(p = new char [strlen(s)+1],s); else p = 0;
	if (text) XmTextSetString(text,const_cast<char*>(Sure(p)));
	return *this;
}

Value& Text::operator =(const double d)
{
  //AH Changed Start 	return *this = (char*)Float(d);
	return *this = Float(d).to_charPtr();
	//AH Changed End
}

Value& Text::operator =(const int i)
{
 //AH Changed Start 	return *this = (char*)Int(i);
	return *this = Int(i).to_charPtr();
	//AH Changed End
}


bool Text::operator ==(const char* v) const
{
	return p && v && strcmp(p,v)==0;
}

bool Text::operator !=(const char* v) const
{
	return !(*this==v);
}


bool Text::operator >(const char* v) const
{
	return p && v && strcmp(p,v)>0;
}


bool Text::operator <(const char* v) const
{
	return p && v && strcmp(p,v)<0;
}


Text Text::operator -() const
{
	char* s1 = const_cast<char*>(Sure(p)),* local = new char[strlen(s1)+1]+strlen(s1);
	*local = 0; while (*s1) *--local = *s1++;
	Text t;
	t.p = local;
	return t;
}

Text Text::operator -(const char* v) const
{
	char* p1;
	if (p && v && (p1 = strstr(p,v)))
	{
		char* local = new char[strlen(p)-strlen(v)+1];
		strncpy(local,p,p1-p); strcpy(local+(p1-p),p1+strlen(v));
		Text t;
		t.p = local;
		return t;
	}
	else return Text(*this);

}


Text Text::operator +(const char* v) const
{
	if (v)
	{
		char* s1 = const_cast<char*>(Sure(p)),* local = new char[strlen(s1)+strlen(v)+1];
		strcpy(local,s1); strcat(local,v);
		Text t;
		t.p = local;
		return t;
	}
	else return Text(*this);
}


Text& Text::operator +=(const char* s)
{
	if (s)
	{
		char* np = new char[(p?strlen(p):0)+strlen(s)+1];
		*np = 0;
		if (p)
		{
			strcat(np,p);
			delete[] p;
		}
		
		strcat(p = np,s);
		if (text)
			XmTextSetString(text,p);
	}
	
	return *this;
}


Text& Text::operator +=(const char c)
{
	char s[] = {c,0};
	return *this += s;
}


Text& Text::operator +=(const double d)
{
        static char buf[256];
        ostrstream os(buf,256);
        os<<d<<'\0';
        if (buf)
        {
                char* np = new char[(p?strlen(p):0)+strlen(buf)+1];
                *np = 0;
                if (p)
                {
                        strcat(np,p);
                        delete[] p;
                }

                strcat(p = np,buf);
                if (text)
                        XmTextSetString(text,p);
        }

        return *this;
}


Text Text::operator *(const char* v) const
{
	if (v)
	{
		char* s1 = const_cast<char*>(Sure(p)); int cnt = Maxi(0,atoi(v));
		char* local = new char[cnt*strlen(s1)+1];
		*local = 0;
		while (cnt--)
			strcat(local,s1);
		Text t;
		t.p = local;
		return t;
	}
	else return Text(*this);
}

Value* Text::Eq(Value& v)
{
  //AH Changed Start	return new Float(*this==v?1:0);
	return new Float(*this==v.to_charPtr()?1:0);
	//AH Changed End
}

Value* Text::Gt(Value& v)
{
 //AH Changed Start return new Float(*this>v?1:0);
	return new Float(*this>v.to_charPtr()?1:0);
	//AH Changed End
}

Value* Text::Lt(Value& v)
{
 //AH Changed Start	return new Float(*this<v?1:0); 
	return new Float(*this<v.to_charPtr()?1:0); 
	//AH Changed End
} 

Value* Text::Not()
{
  return new Float(strlen(Sure(p)));
}
