/*************************************************************************
*                                                                          
*  Datei         : netio.h                     
*                                                                          
*  Aufgabe       : Schnittstelle zum .TN Dateiformat
*                                                                          
*  Programmierer : Ulrich Hertlein
*                  Kolja Koischwitz
*
*  Datum         : Mai 1994
*
**************************************************************************/

#ifndef NETIO_H
#define NETIO_H

#include <stdio.h>

#ifndef _XtIntrinsic_h
typedef char* String;
#  define FALSE     0
#  define TRUE      1
#endif

#undef DELAY
#undef Bool
typedef int Bool;
typedef enum {GSPN=0,DSPN,EDSPN,CDSPN} NetType;
typedef enum {IMM=0,EXP,DET,GEN} TransitionType;
typedef enum {DELAY,DELAYPAR,MARKDEP,PMFDEF} DelayType;
typedef enum {RS,RA,RE} FiringPolicy;
typedef enum {ORIENT_0,ORIENT_45,ORIENT_90,ORIENT_M45} Orientation;

typedef struct Coord
{ 
  double x,y;

} Coord;

typedef struct Markpar
{
  String name;
  Coord pos;
  int value;

} Markpar;

typedef struct Place
{
  String name;
  Coord pos,name_pos;
  Bool is_value;
  union
  {
    int value;
    Markpar *markpar;
  } marking;

} Place;


typedef struct Delaypar
{
  String name;
  Bool is_value;
  union
  {
    double value;
    String md_delay;
  } delay;

  Coord pos;

} Delaypar;


typedef struct Arc
{
  Bool is_value;
  union
  {
    int value;
    String md_mult;
  } multipl;
  Place *place;
  
  int num_pts;
  Coord *point;

} Arc;


typedef struct Transition
{
  String name;
  TransitionType type;

  DelayType delay_type;
  union
  {
    double value;
    String md_delay;
    Delaypar *delay_par;
  } delay;

  Bool inf_server;
  FiringPolicy f_policy;
  int priority;
  String md_enable;
  
  int num_inarcs;
  Arc *in_arc;
  int num_outarcs;
  Arc *out_arc;
  int num_inharcs;
  Arc *inh_arc;

  Coord pos,name_pos,delay_pos;
  Orientation orient;

  double phase, group, group_weight;

} Transition;


typedef struct Reward
{
  String name;
  String def;

} Reward;


typedef struct NetDescr
{
  NetType type;
  String netname;
  String comment;
  int (*error_func)();

  int num_places;
  Place *place;

  int num_trans;
  Transition *trans;

  int num_delaypars;
  Delaypar *delaypar;

  int num_markpars;
  Markpar *markpar;

  int num_rewards;
  Reward *reward;

} NetDescr;


/*
 * Function prototyping
 */
extern void    Free(/* char *ptr */);
extern void   *nMalloc(/* unsigned size */);
extern void   *nRealloc(/* char *ptr, unsigned size */);
extern char   *nStrdup(/* char *str */);

extern char   *skip_white(/* char *buf */);

extern Markpar     *find_markpar(/* NetDescr *net, String name */);
extern Place       *find_place();
extern Delaypar    *find_delaypar();
extern Transition  *find_trans(/* NetDescr *net, String name */);

/* load_net()
 *
 * netname:	name of the TN file to be loaded
 *		if 'netname' does not start with a "/" (absolute filename),
 *		the contents of the env.variable MODELDIR will be prepended;
 *		extension .TN is always added
 *		if 'netname' is NULL, stdin will be read
 *		
 * error_func:	error handler to be used
 *		if 'error_func' is NULL, a default handler will be used
 *
 * Deklaration eines eigenen Error-Handlers:
 * error_handler(char *error_str, int fatal_error)
 * error_str:   Fehler-Text
 * fatal_error: TRUE  => Handler sollte exit() aufrufen,
 *              FALSE => Nur Warnung, handler sollte return-en
 */ 
#ifndef __STDC__
extern NetDescr    *load_net(/* String name, int (*error_func)() */);
extern void         show_net(/* NetDescr *net */);
extern void         free_net(/* NetDescr *net */);
#else
extern NetDescr    *load_net(String name, int (*error_func)(char*,int));
extern void         show_net(NetDescr *net);
extern void         free_net(NetDescr *net);
#endif

#endif

