/* **********************************************************************
*                                                                              
*  Programm         :     integrate.c (pmf ----> PDF Ueberfuehrung)
*                                                                              
*  Aufgabe          :      struct_time_l wird ergaenzt um integrierte power-ausdruecke
*                                                                              
*                                                                              
*  Programmierer    :     Joerg W. Mitzlaff                               
*                                                                              
*  Datum            :     Juli 1993   
*
*************************************************************************/

#include "pmfParsdfunc.h"
#include <math.h>

double pot(basis, exponent)
     double basis;
     unsigned int exponent;
{
    double retu;
    int i;
    if((basis == 0.0) && (exponent == 0))
	retu = 1.0;  /* 0^0 */
    else
      if(basis == 0.0)  /* 0^x */
	retu = 0.0;
      else
	{

	  retu=1.0;         /* x^x */
	  if(exponent>0)  /* x^0 ? */
	    for(i=0;i<exponent;i++)
	      retu *= basis;
	}
  /*  printf("                              pot: %f^%i = %f\n",basis,exponent,retu); */
    return(retu);
}

double fak(n)
     unsigned int n;
{
    int i;
    double retu;

    retu=1.0;

    if(n>1)
	for(i=2;i<=n;i++)
	    retu *= i;
    return(retu);
}

void add_or_append_one_power(p1, m, b, lambda)
     time_l *p1;
     unsigned int m;
     double b;
     double lambda;
/* Parameter: power-stru, exponent, neues oder zu addierendes b, lambda */
{
    power_m *append;
   
    
    
    append = p1->power;
    while( (append->m != m) && (append->next != NULL) ) 
	append = append->next;
    if(append->m == m)
      {
	add_or_append_all_expos(append->expo, make_n_stru(lambda,b,0.0,NULL));
/*	printf("I->x1=%f app expos in x^%i  b=%f la=%f\n",p1->x1,m,b,lambda); */
      }
    else  /* neue power anlegen: 
	     wie in integr. schleife muss offalt auf alle m=0 b's aufaddiert werden */
      {
	append_one_power(p1->power, make_m_stru(m,
						make_n_stru(lambda,b,0.0,NULL),
						NULL));
/*	printf("I->x1=%f app power x^%i  b=%f la=%f\n",p1->x1,m,b,lambda); */
      }

    
}

double integrate(p1, modus)
     time_l *p1;
     int modus;  /* erzeugt hoffendlich PMF-function 
					    double ist der offset des letzten Intervalls */
{
    double offneu, offalt; /* siehe offalt */

    time_l  *i;
    power_m *j;
    expo_n  *k;
    unsigned h;
    double zwerg,zwerg2;

    offneu = 0;
    i = p1;
    do{
      offalt = offneu;  /* jedes Intervall hat nur einen offset, und der ist gleich */
      /*	printf("offset in I:%f =%f\n",i->x1,offalt); */
      j = i->power;
      do{
	k = j->expo;
	do{
	  if(k->c!=0.0)
	    {
	    if(k->lambda==0.0)
	      {
		if(j->m==0)
		  {
		    k->b = k->b  - k->c  * i->x1; /* alle b's von x^0 bekommen offset  */
		    /*			printf("I->x1=%f setze b neu in x^%i  b=%f la=%f\n",i->x1,j->m,k->b,k->lambda); */
		    add_or_append_one_power(i, 1, k->c, 0.0);
		    
		    offneu += k->c * (i->x2 - i->x1);
		  }
		else
		  {
		    zwerg = (-1.0) * k->c * pot(i->x1,(j->m + 1)) / (j->m + 1);
		    add_or_append_one_power(i, 
					    0,
					    zwerg,
					    0.0);
		    add_or_append_one_power(i,
					    j->m + 1,
					    k->c / (j->m + 1),
					    0.0);
		    offneu +=  k->c / (j->m + 1) * (pot(i->x2,j->m+1) -  pot(i->x1,j->m+1));
		  }
	      }
	    else  /* wirds kompliziert */
	      {
		add_or_append_one_power(i, 
					0,
					k->c*fak(j->m)*exp((-1.0)*k->lambda*i->x1) / pot(k->lambda,j->m+1),
					0.0);
		add_or_append_one_power(i, 
					0,
					(-1.0)*k->c*fak(j->m) / pot(k->lambda,j->m+1),
					k->lambda);
		if(j->m>0)
		  {
		    zwerg=0.0;
		    
		    for(h=1;h <= j->m;h++)
		      zwerg += pot(k->lambda,h) * pot(i->x1,h) / fak(h);
		    
		    zwerg *= k->c * fak(j->m) * exp((-1.0)*k->lambda*i->x1) / pot(k->lambda,j->m+1);
		    
		    add_or_append_one_power(i, 
					    0,
					    zwerg,
					    0.0);
		    
		    for(h=1;h <= j->m;h++)
		      add_or_append_one_power(i,
					      h,
					      (-1.0) * k->c * fak(j->m) / ( pot(k->lambda,(j->m+1-h)) * fak(h)),
					      k->lambda);
		  }
		
		/* Offset vergroessern */
		zwerg = 0.0;
		for(h=0;h <= j->m;h++)
		  zwerg += pot(k->lambda,h) * pot(i->x1,h) * exp((-1.0)*k->lambda*i->x1) / fak(h);
		zwerg = k->c * fak(j->m) * (1 - zwerg) / pot(k->lambda,j->m+1);
		
		zwerg2=0.0;
		for(h=0;h <= j->m;h++)
		  zwerg2 += pot(k->lambda,h) * pot(i->x2,h) * exp((-1.0)*k->lambda*i->x2) / fak(h);
		zwerg2 = k->c * fak(j->m) * (1.0 - zwerg2) / pot(k->lambda,j->m+1);
		
		zwerg = zwerg2 - zwerg;
		offneu += zwerg;
	      }
	    } /*if(k->c!=0.0)*/
	  k = k->next;
	}while(k!=NULL);
	
	j = j->next;	
      }while(j!=NULL);
      
      if(i->p > 0.0)
	{
	  j = i->power;    /* x^0 finden und b-faktor vergroessern */	
	  while((j->m != 0) && (j->next != NULL))	
	    j = j->next;
	  
	  if(j->m != 0)    /* Es war kein x^0 da; fuege eins an */
	    {
	      j->next = zero_tail();
	      j->next->expo->b = i->p;  /* b ist ja null! */
	      /*	    printf("neues x^0 b + impuls\n"); */
	    }
	  else
	    {
	      j->expo->b += i->p;  
	      /*	    printf("b + impuls %f\n",i->p); */
	    }
	  
	  /* offset vergroessern */
	  if(modus==INTEGRATE_NORMAL)
	    offneu += i->p; /* pointmass erhoeht PDF-func.;
			       geht erst nach Behandl. aller power's, da p abzuziehen ist */
	  else
	    offneu += i->p * i->x1;
	}
      
      /* auf alle b-faktoren mit  x^0 und lambda=0 muss der alte offset rauf */
      j = i->power;
      do{
	k = j->expo;
	do{
	  if(j->m == 0 && k->lambda == 0.0)
	    k->b += offalt;
	  k = k->next;
	}while(k!=NULL);
	
	j = j->next;	
      }while(j!=NULL);
      
      i = i->next;
    }while(i!=NULL);
    
/*    printf("expolynomial distribution: internal structure c = pmf; b = PDF (but in memory 1-PDF)\n");*/
/*    print_expoly(p1); */

    /* berechnen von 1-PDF */
    i = p1;
    do{
      j = i->power;
      do{
	k = j->expo;
	do{

	   if((j->m == 0) && (k->lambda == 0.0))
	     k->b = 1.0 - k->b;
	   else
	     k->b = k->b * (-1.0);

	   k = k->next;
	}while(k!=NULL);
	j = j->next;	
      }while(j!=NULL);
      i = i->next;
    }while(i!=NULL); 

    return(offneu);
}


double expected_value(p1)
     time_l *p1; /* arbeitet auf einer Kopie der durch parsierung
				     erzeugten, aber noch nicht intergrierten struktur */
{
    time_l  *copy;
    time_l  *i;
    power_m *j;
    double retu;

    copy = copy_all_intervall(p1);

    i = copy;
    do{
      j = i->power;
	do{
	  /* struktur um x*f(x) ergaenzen */
	  j->m ++;
	  
	j = j->next;	
      }while(j!=NULL);
      i = i->next;
    }while(i!=NULL);
    
    retu = integrate(copy,INTEGR_FOR_EXPECTED_VALUE);
    
    free_whole_stru(copy);

    return(retu);
}


