#ifndef __TNLINK_H__
#define __TNLINK_H__


extern "C" {
#include "netio/netio.h"
}

#include "xbefore.h"
#include <Xm/CascadeB.h>
#include <Xm/PushB.h>
#include "xafter.h"

#include "commmod.h"
#include "parser.h"

//---------------------------------------------------------------------
// Definitionen
//---------------------------------------------------------------------

#define intP int*
#define boolP bool*
#define charP char*

#define NO_RESULT "no"
#define UPDATEINTERVALL 300

#define NET_ENDING  ".net"
#define TN_ENDING ".TN"
// Fuer Load & Save

//---------------------------------------------------------------------
// Klassendeclarationen
//---------------------------------------------------------------------

enum  method {mS, mC};
enum  par {SERIELL, PARALLEL};
enum  {INPUTARC= 0, OUTPUTARC, INHIBITOR};
enum  overall_solution {EMC_EXPLICIT, FILL_IN_AVOIDANCE};
enum  solution {DIRECT, ITERATIVE};
enum  distributed {SEQUENTIAL, DISTRIBUTED};
enum  init_vector {UNIFORM, RANDOM, LOAD_FILE};
enum  arithmetic {DOUBLE, ARBITRARY};
enum  partype {DELAYP, MARKP};
enum  prtype {aPRINTER, aFILE};
enum  pr_gdesctype {aPS, aEPS};
enum  pr_tdesctype {aPS2, aLATEX};
enum  embedding {aNO, aTIME, aMEMO};
enum  outputform {aCURVE, aPOINT};
enum  variant {aA, aB};
enum  mode {aLIN, aLOG};


class TimeNETLink: public CommModule, public Wrapper<TimeNETLink,bool>, public Parser
{
  bool EDSPNen;

  EditorP ed;

  bool save_options;
  StructP setup;
  Dialog<TimeNETLink,bool> dialog;
  void DialogCallback(Dialog<TimeNETLink,bool>& dia,bool,Ident& btn);

  Text  netname;
  void  SetNetName();
  // for fileIO

  WrapperID FlashID;

  int   closereason;
  int   killreason;
  // for logwin

  Widget pmfwidget, distwidget;

  bool  gp_logx;
  bool  gp_logy;
  Text  gp_title;
  Text  gp_xdescr;
  Text  gp_ydescr;
  int   gp_xvalue;
  int   gp_yvalue;
  // for gnuplot

  enum  pr_gdesctype pr_gdesc;
  enum  pr_tdesctype pr_tdesc;
  bool  pr_graph;
  bool  pr_text;
  enum  prtype pr_dest;
  Text  pr_cmd;
  // for print

  Text  exp_param;
  double exp_from;
  float exp_to;
  enum  mode exp_mode;
  float exp_step;
  // for experiment

  Text  exp2_param;
  float exp2_from;
  float exp2_to;
  enum  mode exp2_mode;
  float exp2_step;
  // for experiment

  Text  exp3_param;
  float exp3_from;
  float exp3_to;
  enum  mode exp3_mode;
  float exp3_step;
  // for experiment

  enum  overall_solution sa_overall_solu;
  enum  distributed sa_distributed;
  float sa_epsilon;
  int   sa_max_iter;
  float sa_trunc_error;
  enum  arithmetic sa_arith;
  int   sa_arb_bits;
  enum  solution sa_lin_solu;
  enum  init_vector sa_init_vec;
  bool  sa_save_vec;
  int   sa_seed_val;
  // for StationaryAnalysis

  enum  solution dtsa_solu;
  enum  embedding dtsa_embed;
  int   dtsa_max_iter;
  float dtsa_epsilon;
  int   dtsa_scale;
  // for StationaryAnalysis DTSPN

  float sp_epsilon;
  enum  solution sp_solu;
  int   sp_max_iter;
  float sp_droptol;
  int   sp_phases;
  int   sp_vphases;
  // for StationaryApproximation

  enum  par ss_distr;
  enum  method ss_method;
  bool  ss_detect;
  int   ss_conf_level;
  int   ss_error_percnt;
  int   ss_epsilon_max;
  int   ss_random_seed;
  int   ss_max_samples;
  int   ss_min_firings;
  float ss_max_modeltime;
  float ss_max_realtime;
  bool  ss_visu;
  int   ss_thresholds;
  int   ss_pilotrun;
  // for StationarySimulation

  enum  variant ta_variant;
  enum  outputform ta_output;
  float ta_transtime;
  float ta_epsilon;
  float ta_stepsize;
  float ta_internal;
  int   ta_clustersize;
  // for TransientAnalysis

  enum  solution dtta_solu;
  enum  embedding dtta_embed;
  enum  outputform dtta_output;
  int   dtta_max_iter;
  int   dtta_timesteps;
  int   dtta_stepsize;
  int   dtta_firststep;
  float dtta_epsilon;
  int   dtta_scale;
  // for TransientAnalysis DTSPN

  enum  par ts_distr;
  bool  ts_rule;
  float ts_transient_time;
  int   ts_no_sampling_points;
  int   ts_conf_level;
  float ts_epsilon_max;
  int   ts_random_seed;
  float ts_max_realtime;
  // for TransientSimulation

  int pidStationaryAnalysisDTSPN;
  void StationaryAnalysisDTSPN(Editor& ed);
  int pidStationaryAnalysis;
  void StationaryAnalysis(Editor& ed);
  int pidStationaryApproximation;
  void StationaryApproximation(Editor& ed);
  int pidStationarySimulation;
  void StationarySimulation(Editor& ed);
  int pidRestartStSimulation;
  void RestartStSimulation(Editor& ed);
  int pidTransientAnalysis;
  void TransientAnalysis(Editor& ed);
  int pidTransientAnalysisDTSPN;
  void TransientAnalysisDTSPN(Editor& ed);
  int pidTransientSimulation;
  void TransientSimulation(Editor& ed);
  int pidExperiment;
  void Experiment(Editor& ed);
  int pidExperiment2;
  void Experiment2(Editor& ed);
  int pidExperiment3;
  void Experiment3(Editor& ed);
  int pidGnuPlot;
  void GnuPlot(Editor& ed);
  int pidEstimSE;
  void EstimSE(Editor& ed);
  int pidTraps;
  void Traps(Editor& ed);
  int pidSiphons;
  void Siphons(Editor& ed);
 
  void Print(Editor& ed);
  void ClearResults(Editor&);
  void GetResults(Editor&);
  void CheckStructure(Editor& ed);

  void PMFEditor(NodeP node);
  void TokenDistribution(NodeP node);
  void ThroughPut(NodeP node);

  void DefaultParams(StructP params);
  void SaveParams(StructP params);
  // DialogParameter speichern

  NodeP FindMarkingParameter(charP Name);

  bool CheckNet(bool okwindow= false);
  // Testet das Netz, ob es TimeNET tauglich ist

  NodeClassP      GetNodeClassPlace(NetClassP nc);
  NodeClassP      GetNodeClassExpTrans(NetClassP nc);
  NodeClassP      GetNodeClassDetTrans(NetClassP nc);
  NodeClassP      GetNodeClassGenTrans(NetClassP nc);
  NodeClassP      GetNodeClassImmTrans(NetClassP nc);
  NodeClassP      GetNodeClassMarkPar(NetClassP nc);
  NodeClassP      GetNodeClassDelayPar(NetClassP nc);
  NodeClassP      GetNodeClassRewItem(NetClassP nc);
  ConnectorClassP GetConnectorClassArc(NetClassP nc);
  ConnectorClassP GetConnectorClassInhibitor(NetClassP nc);

  bool   IsPlace(NodeP node);
  bool   IsTransition(NodeP node);
  bool   IsGenTrans(NodeP node);
  bool   IsExpTrans(NodeP node);
  bool   IsDetTrans(NodeP node);
  bool   IsImmediate(NodeP node);
  bool   IsArc(ConnectorP conn);
  bool   IsInhibitor(ConnectorP conn);
  bool   IsParameter(NodeP node);
  bool   IsMarkPar(NodeP node);
  bool   IsDelayPar(NodeP node);
  bool   IsRewItem(NodeP node);
  TextP  GetMarking(NodeP node);
  TextP  GetDelay(NodeP node);
  ValueP GetIsMDDelay(NodeP node);
  TextP  GetPolicy(NodeP node);
  ValueP GetServerType(NodeP node);
  TextP  GetPMFunc(NodeP node);
  ValueP GetPriority(NodeP node);
  TextP  GetWeight(NodeP node);
  TextP  GetEnable(NodeP node);
  TextP  GetMultiplicity(ConnectorP c);
  ValueP GetIsMDArcMult(ConnectorP c);
  ValueP GetMarkPar(NodeP node);
  TextP  GetDelayPar(NodeP node);
  ValueP GetIsMDDelayPar(NodeP node);
  TextP  GetRewItem(NodeP node);
  TextP  GetResult(NodeP node);
  // TimeNETLink-Hilfsfunktionen

  bool    firsttime;
  ObjectP ParserObject;
  TextP   ParserText;
  bool    RecursionError;
  LinkedList<NodeP> Recursion;
  // Fuer die Fehlerausgaben des Parser

  void   ParserError(const charP message, charP string, int length);
  long   ParseMarking(charP string, int length);
  double ParseParameter(charP string, int length);
  double ParseRealParameter(charP string, int length);
  long   ParseIntegerParameter(charP string, int length);
  // Parser-Hilfsfunktionen

  Int arccnt;
  // Kantenzaehler

  NetP LoadTN(charP file);
  void SaveTN(EditorP ed, charP fname);
  // Laedt und Speichert eine .TN-Datei

  ostream& WriteTN(ostream& os);
  ostream& WriteArc(ostream& os,ConnectorP conn,int type);
  ostream& WriteInch(ostream& os,Position pos);
  ostream& WriteDelay(ostream& os,Position pos);
  ostream& OutNoSemicolon(ostream& os,charP string);
  ostream& OutSpaces(ostream& os,charP string);
  ostream& OutSpaces2(ostream& os,charP string);

  void OnFilesel(char* file, bool ok);
  FileSelector<TimeNETLink,EditorP> fsel;
  void AddCB(EditorP, charP filename, bool ok);
  bool ExportCB;
  void SaveAsCB(EditorP, charP filename, bool ok);
  void FselCallback(EditorP,charP file,bool ok);
  // Selector und CallBacks

  Position InchToPix(Coord& coord) {return Position(coord.x*135,coord.y*135);}
  Position PixToPix(Coord& coord)  {return Position(coord.x*1.35,coord.y*1.35);}
  void CreateArcs(NodeP trans,PageP prime,int type,Arc* arc,int num_arcs,
		  ConnectorClass* cc);
  // Hilfsfunktionen

  bool FlashTimer(bool);

  //---------------------------------------------------------------------

public:

  TimeNETLink(Net& n): CommModule(n,"TimeNET-Link"), 
  Wrapper<TimeNETLink,bool>(this), Parser(),
  dialog(this),
    //AH Changed Start => (WrapperID)(Widget)(Widget)
  netname(""), FlashID((WrapperID)NULL), pmfwidget((Widget)NULL), distwidget((Widget)NULL),
    //AH Changed End
  gp_logx(false), gp_logy(false), gp_title(""), gp_xdescr(""),
  gp_ydescr(""), gp_xvalue(0), gp_yvalue(0),
  pr_gdesc(aPS), pr_tdesc(aPS2),
  pr_graph(true), pr_text(false),
  pr_dest(aPRINTER), pr_cmd("-Pfr2518"), 
  exp_param(""),
  exp_from(0.0), exp_to(1.0), exp_mode(aLIN), exp_step(1.0), 
  exp2_param(""),
  exp2_from(0.0), exp2_to(1.0), exp2_mode(aLIN), exp2_step(1.0), 
  exp3_param(""),
  exp3_from(0.0), exp3_to(1.0), exp3_mode(aLIN), exp3_step(1.0), 
  sa_overall_solu(EMC_EXPLICIT), sa_distributed(SEQUENTIAL),
  sa_epsilon(1e-07), sa_max_iter(10000),
  sa_trunc_error(1e-07), sa_arith(ARBITRARY), sa_arb_bits(512),
  sa_lin_solu(DIRECT), sa_init_vec(UNIFORM),
  sa_save_vec(false), sa_seed_val(12345),
  dtsa_solu(DIRECT), dtsa_embed(aNO), dtsa_max_iter(1000), 
  dtsa_epsilon(1e-07), dtsa_scale(1),
  sp_epsilon(1e-07),sp_solu(DIRECT),sp_max_iter(10000),sp_droptol(0),
  sp_phases(2),sp_vphases(10),
  ss_distr(SERIELL), ss_method(mS), ss_detect(true), ss_conf_level(95),
  ss_error_percnt(50), ss_epsilon_max(10),
  ss_random_seed(12345), ss_max_samples(0),
  ss_min_firings(50), ss_max_modeltime(0.0),
  ss_max_realtime(0.0), ss_visu(false),
  ss_thresholds(6), ss_pilotrun(600),
  ta_variant(aB), ta_output(aCURVE), ta_transtime(10),
  ta_epsilon(1e-07), ta_stepsize(0.1), ta_internal(0.1),ta_clustersize(10),
  dtta_solu(DIRECT), dtta_embed(aNO), dtta_output(aCURVE), 
  dtta_max_iter(1000), dtta_timesteps(100), dtta_stepsize(10), 
  dtta_firststep(0), dtta_epsilon(1e-07), dtta_scale(1),
  ts_distr(SERIELL), ts_rule(true), ts_transient_time(10),
  ts_no_sampling_points(10), ts_conf_level(95),
  ts_epsilon_max(10), ts_random_seed(12345),
  ts_max_realtime(0.0),
  pidStationaryAnalysisDTSPN(0), 
  pidStationaryAnalysis(0),
  pidStationaryApproximation(0),
  pidStationarySimulation(0),
  pidRestartStSimulation(0),
  pidTransientAnalysis(0),
  pidTransientAnalysisDTSPN(0),
  pidTransientSimulation(0),
  pidExperiment(0),
  pidExperiment2(0),
  pidExperiment3(0),
  pidGnuPlot(0),
  pidEstimSE(0),
    pidTraps(0),
 pidSiphons(0),
    //AH Changed Start => (ObjectP) (TextP)
  firsttime(true),ParserObject((ObjectP)NULL),
  ParserText((TextP)NULL), RecursionError(false),
    //AH Changed End
  fsel(this) {}

  ~TimeNETLink() 
    {
      if (FlashID)
	RemoveCB(FlashID);	
      CloseFilesel();
      DestroyFilesel();
    };

  bool Init()
    {

      (net.Prime()->Class()->Name()=="EDSPN")? EDSPNen= true: EDSPNen= false;
      return true;
    };
  bool Open(Editor& ed, Page&)
    {
      FlashID = TimerCB(UPDATEINTERVALL,&TimeNETLink::FlashTimer,false);	

      CreateFilesel(model_dir.to_charPtr());

      this->ed= &ed;
      if (net.Prime()->Class()->Name()=="EDSPN")
	ed.SetDimension(3000,1500);
      return false;
    };
  bool CreatePulldown(EditorP ed, charP name, Widget parent)
    {
      if ((net.Prime()->Class()->Name() == "EDSPN") && !strcmp(name,"Module"))
	{
	  WrapperID ID;
	  Widget pulldown_ct, pulldown_dt, pulldown, pulldown1, pulldown2;

	  pulldown1= ed->CreatePulldown("Validate", parent);
	  pulldown2= ed->CreatePulldown("Evaluation", parent);

	  ed->CreateEntry("Check Net",
			  xmPushButtonWidgetClass, pulldown1, NULL, ID);
	  ed->CreateEntry("Estimate Statespace",
			  xmPushButtonWidgetClass, pulldown1, NULL, ID);
	  ed->CreateEntry("Traps",
			  xmPushButtonWidgetClass, pulldown1, NULL, ID);
	  ed->CreateEntry("Siphons",
			  xmPushButtonWidgetClass, pulldown1, NULL, ID);

	  ed->CreateEntry("Check Structure",
			  xmPushButtonWidgetClass, pulldown1, NULL, ID);
//
	  pulldown= ed->CreatePulldown("Results", 
				       ed->CreateEntry("Results",
						       xmCascadeButtonWidgetClass,
						       pulldown2, NULL, ID));
	  ed->CreateEntry("Clear",
			  xmPushButtonWidgetClass, pulldown, NULL, ID);
	  ed->CreateEntry("Get",
			  xmPushButtonWidgetClass, pulldown, NULL, ID);
	  ed->CreateEntry("GnuPlot Experiment-Results",
			  xmPushButtonWidgetClass, pulldown, NULL, ID);
//
	  pulldown= ed->CreatePulldown("Stationary Analysis", 
				       ed->CreateEntry("Stationary Analysis",
						       xmCascadeButtonWidgetClass,
						       pulldown2, NULL, ID));
	  pulldown_ct= ed->CreatePulldown("Continuous Time St.An.", 
				       ed->CreateEntry("Continuous Time St.An.",
						       xmCascadeButtonWidgetClass,
						       pulldown, NULL, ID));
	  ed->CreateEntry("Start St.An.",
			  xmPushButtonWidgetClass, pulldown_ct, NULL, ID);
	  ed->CreateEntry("Stop St.An.",
			  xmPushButtonWidgetClass, pulldown_ct, NULL, ID);
	  ed->CreateEntry("Start Experiment St.An.",
			  xmPushButtonWidgetClass, pulldown_ct, NULL, ID);
	  ed->CreateEntry("Stop Experiment St.An.",
			  xmPushButtonWidgetClass, pulldown_ct, NULL, ID);
	  pulldown_dt= ed->CreatePulldown("Discrete Time St.An.", 
				       ed->CreateEntry("Discrete Time St.An.",
						       xmCascadeButtonWidgetClass,
						       pulldown, NULL, ID));
	  ed->CreateEntry("Start St.An. DTSPN",
			  xmPushButtonWidgetClass, pulldown_dt, NULL, ID);
	  ed->CreateEntry("Stop St.An. DTSPN",
			  xmPushButtonWidgetClass, pulldown_dt, NULL, ID);
	  ed->CreateEntry("Start Experiment St.An. DTSPN",
			  xmPushButtonWidgetClass, pulldown_dt, NULL, ID);
	  ed->CreateEntry("Stop Experiment St.An. DTSPN",
			  xmPushButtonWidgetClass, pulldown_dt, NULL, ID);
//
	  pulldown= ed->CreatePulldown("Stationary Approximation", 
				       ed->CreateEntry("Stationary Approximation",
						       xmCascadeButtonWidgetClass,
						       pulldown2, NULL, ID));
	  pulldown_ct= ed->CreatePulldown("Continuous Time St.Ap.", 
				       ed->CreateEntry("Continuous Time St.Ap.",
						       xmCascadeButtonWidgetClass,
						       pulldown, NULL, ID));
	  ed->CreateEntry("Start St.Ap.",
			  xmPushButtonWidgetClass, pulldown_ct, NULL, ID);
	  ed->CreateEntry("Stop St.Ap.",
			  xmPushButtonWidgetClass, pulldown_ct, NULL, ID);
//
	  pulldown= ed->CreatePulldown("Stationary Simulation", 
				       ed->CreateEntry("Stationary Simulation",
						       xmCascadeButtonWidgetClass,
						       pulldown2, NULL, ID));
	  pulldown_ct= ed->CreatePulldown("Continuous Time St.Si.", 
				       ed->CreateEntry("Continuous Time St.Si.",
						       xmCascadeButtonWidgetClass,
						       pulldown, NULL, ID));
	  ed->CreateEntry("Start Standard St.Si.",
			  xmPushButtonWidgetClass, pulldown_ct, NULL, ID);
	  ed->CreateEntry("Stop Standard St.Si.",
			  xmPushButtonWidgetClass, pulldown_ct, NULL, ID);
	  ed->CreateEntry("Start Fast St.Si.",
			  xmPushButtonWidgetClass, pulldown_ct, NULL, ID);
	  ed->CreateEntry("Stop Fast St.Si.",
			  xmPushButtonWidgetClass, pulldown_ct, NULL, ID);
	  ed->CreateEntry("Start Experiment St.Si.",
			  xmPushButtonWidgetClass, pulldown_ct, NULL, ID);
	  ed->CreateEntry("Stop Experiment St.Si.",
			  xmPushButtonWidgetClass, pulldown_ct, NULL, ID);
//
	  pulldown= ed->CreatePulldown("Transient Analysis", 
				       ed->CreateEntry("Transient Analysis",
						       xmCascadeButtonWidgetClass,
						       pulldown2, NULL, ID));
	  pulldown_ct= ed->CreatePulldown("Contionous Time Tr.An.", 
				       ed->CreateEntry("Continuous Time Tr.An.",
						       xmCascadeButtonWidgetClass,
						       pulldown, NULL, ID));
	  ed->CreateEntry("Start Tr.An.",
			  xmPushButtonWidgetClass, pulldown_ct, NULL, ID);
	  ed->CreateEntry("Stop Tr.An.",
			  xmPushButtonWidgetClass, pulldown_ct, NULL, ID);
	  pulldown_dt= ed->CreatePulldown("Discrete Time Tr.An.", 
					  ed->CreateEntry("Discrete Time Tr.An.",
						       xmCascadeButtonWidgetClass,
						       pulldown, NULL, ID));
	  ed->CreateEntry("Start Tr.An. DTSPN",
			  xmPushButtonWidgetClass, pulldown_dt, NULL, ID);
	  ed->CreateEntry("Stop Tr.An. DTSPN",
			  xmPushButtonWidgetClass, pulldown_dt, NULL, ID);
//
	  pulldown= ed->CreatePulldown("Transient Simulation", 
				       ed->CreateEntry("Transient Simulation",
						       xmCascadeButtonWidgetClass,
						       pulldown2, NULL, ID));
	  pulldown_ct= ed->CreatePulldown("Continuous Time Tr.Si.", 
				       ed->CreateEntry("Continuous Time Tr.Si.",
						       xmCascadeButtonWidgetClass,
						       pulldown, NULL, ID));
	  ed->CreateEntry("Start Tr.Si.",
			  xmPushButtonWidgetClass, pulldown_ct, NULL, ID);
	  ed->CreateEntry("Stop Tr.Si.",
			  xmPushButtonWidgetClass, pulldown_ct, NULL, ID);
//
	  ed->CreateEntry("stop all solutions", 
			  xmPushButtonWidgetClass, pulldown2, NULL, ID);
	  return true;
	}
      else
	return false;
    };
  bool CreateEntry(EditorP ed, charP name, WidgetClass, Widget parent)
    {
      WrapperID ID;
      Widget pulldown;

      if ((net.Prime()->Class()->Name() == "SPNL") && !strcmp(name,"SPNLShowResults"))
	{
	  ed->CreateEntry("gnuplot experimental results",
			  xmPushButtonWidgetClass, parent, NULL, ID);
	}
      else if (!strcmp(name,"Add"))
	{// Import hinter Add erzeugen

	  ed->CreateEntry("Import",
			  xmPushButtonWidgetClass, parent, NULL, ID);
	}
      else if (EDSPNen && !strcmp(name,"Enter"))
	{// Enter zerstoeren

	  ed->DestroyEntry(XtNameToWidget(parent, "Enter"), NULL);
	}
      else if (EDSPNen && !strcmp(name,"Leave"))
	{// Leave zerstoeren

	  ed->DestroyEntry(XtNameToWidget(parent, "Leave"), NULL);
	}
      else if (EDSPNen && !strcmp(name,"SaveAs"))
	{// Export hinter SaveAs erzeugen

	  pulldown= ed->CreatePulldown("Remove...", 
				       ed->CreateEntry("Remove...",
						       xmCascadeButtonWidgetClass,
						       parent, NULL, ID));
	  ed->CreateEntry("Lock-File",
			  xmPushButtonWidgetClass, pulldown, NULL, ID);
	  ed->CreateEntry("Results",
			  xmPushButtonWidgetClass, pulldown, NULL, ID);

	  ed->CreateEntry("Export",
			  xmPushButtonWidgetClass, parent, NULL, ID);
	  ed->CreateEntry("Print",
			  xmPushButtonWidgetClass, parent, NULL, ID);
	}
      else if (EDSPNen && !strcmp(name,"SelectAll"))
	{// PMFEditor hinter SelectAll erzeugen

	  ed->CreateSeparator(parent);

	  pmfwidget= ed->CreateEntry("PMF Editor",
		     xmPushButtonWidgetClass, parent, NULL, ID);
	  XtSetSensitive(pmfwidget,False);
	  distwidget= ed->CreateEntry("Result Info",
		     xmPushButtonWidgetClass, parent, NULL, ID);
	  XtSetSensitive(distwidget,False);
	}
      return false;
    };
  bool Change(Editor&, Page&, Page&) {return false;};
  bool Event(Editor&,Page& pg,EventInfo&,bool before);
  bool Menu(Editor& ed, Page& pg, Ident& name, bool before);
  void OnLogwinButton(Ident& button);
  void OnDialogButton(Ident& button);
  bool NetTransform(Editor& ed);

  // Modulfunktionen
};

#endif
