/*
	Modifiziert:	Joachim Schneider	July 1994
			Anpassung an .TN-Format
*/
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <malloc.h>
#include <math.h>

#if defined(__NetBSD__) || defined(linux)
#include <float.h>
#else 
#include <values.h>
#define DBL_MAX MAXDOUBLE
#define DBL_MIN MINDOUBLE
#endif

#if defined(__hpux) || defined(ultrix) || defined(sun) &&  defined(__svr4__)
#define anint ceil
#endif

#include "netio.h"
#include "Paint.h"

/* define the following if you want a programm that makes only the curves */
/* #define CURVE_ALONE */

#ifdef __STDC__
  double nextRoundNumber(double x);
  void fprintfTicks(FILE *f, double x, double picFactor);
#else
  double nextRoundNumber();
  void fprintfTicks();
#endif

#ifdef ultrix
  /* Spezialbehandlung fuer die DEC-Kisten: */
  static char *strdup (string)
  char *string;
  {
    register char *new_string = (char *)malloc (strlen (string) + 1);
    if (new_string)
      strcpy (new_string, string);
    return new_string;
  }
#endif

#define SEPERATOR "====================\n"
/* length of pic in x direction */
#define XMAX 10.0
/* length in y direction */
#define YMAX 10.0



#ifdef __STDC__
  int readCurves (char *net, CurveT *curve)
#else
  int readCurves (net, curve)
    char *net;
    CurveT *curve;
#endif
{
  int          status = 0;
  FILE         *fd    = NULL;
  char         *line  = NULL;
  unsigned     i      = 0;
  unsigned     j      = 0;
  double       value  = 0.0;
  struct stat  buf;
  time_t       TNtime;
  char         s[MAXPATHLEN];
  char         l[MAXPATHLEN];

  if (!net) {
    fprintf(stderr,"ERROR readCurves() : no netname given.\n");
    return(1);
  } /* if */

  /* TEST FOR ACTUALITY OF <NET>.CURVES */
  sprintf(s, "%s.TN", net);
  if (lstat(s, &buf)) {
    fprintf(stderr, "readCurves() ");
    perror(s);
    return (1);
  } else 
    TNtime = buf.st_mtime;

  sprintf(s,"%s.CURVES", net);
  if (lstat(s, &buf)) {
/*
    fprintf(stderr, "readCurves() ");
    perror(s);
*/
    return (1);
  } /* if */
    
  if (buf.st_mtime < TNtime) {
    fprintf(stderr, "%s is older than <net>.TN\n", s);
    fprintf(stderr, 
      "The curves are not uptodate - You should remake the experiment.\n");
  } /* if */
    
  if ((fd = fopen(s,"r")) == NULL ) {
    fprintf(stderr, "readCurves() ");
    perror(s);
    return(2);
  } /* if */

  /* first count the NUMBER of steps */
  i = 0;
  while (fgets(l,MAXPATHLEN,fd)) {
    if (!strcmp(l,SEPERATOR)) 
      i++;
  } /* if */
  curve->count = i;

  rewind(fd);
  i = 0;
  curve->parameter = NULL;
  /* read the changed parameter name and the number of reward names */
  while (fgets(l,MAXPATHLEN,fd)) {
    if (sscanf(l, "Numerical results of reward measures for: %s", s) == 1) {
      if (curve->parameter) /* ... we have the information we want */
        break;
      curve->parameter = strdup (s);
      continue;
    } /* if */
      if (sscanf(l, "%[A-Xa-x0-9] = %lf", s, &value) == 2) 
        i++;
  } /* if */
  curve->rewcount = i;
            
  /* allocate */
  curve->rew = (CurveRewardT*) calloc(i,sizeof(CurveRewardT));  
  for (j=0; j<i; j++) 
    curve->rew[j].name = NULL;

  rewind(fd);
  i = 0;
  /* read now the values */
  while (fgets(l,MAXPATHLEN,fd)) {
    if (!strcmp(l,SEPERATOR)) {
      i++;
      j=0;
      continue;
    } /* if */

    if (sscanf(l, "Numerical results of reward measures for: %s = %lf", 
        s, &value) == 2) {
      if ( i == 1)  /* start value */
        curve->from = value;
      if ( i == 2) /* compute step size */
        curve->step = value - curve->from;
      if ( i == curve->count) /* end value */
        curve->to = value;
      continue;
    } /* if */
         
    if (sscanf(l, "%[A-Xa-x0-9] = %lf", s, &value) == 2) {
      if (curve->rew[j].name == NULL) { /* gotta initialized */
        curve->rew[j].name = strdup (s);
        curve->rew[j].value = (double*)
                              calloc(curve->count, sizeof(double));
      } /* if */
      curve->rew[j].value[i-1] = value;
      j++;
    } /* if */

  } /* while */

  if (fd) close(fd);
  return status;
}



/******************************************************************************
  makeCurves  gibt PicTex-Befehle aus
******************************************************************************/
#ifdef __STDC__ 
  int makeCurves(char *net, CurveT *curve)
#else
  int makeCurves(net, curve)
    char *net;
    CurveT *curve;
#endif  
{
  FILE           *fd    = NULL;
  int            status = 0;
  unsigned       i      = 0;
  double         value  = 0.0;
  CurveRewardT   *rew   = NULL;
  char s[MAXPATHLEN];

#ifdef CURVE_ALONE
  if (net)
    sprintf(s,"%s.pictex", net);
  else {
    fprintf(stderr,"ERROR makeCurves() : no netname given.\n");
    return(1);
  } /* if */


  if ((fd = fopen(s,"w")) == NULL ) {
    fprintf(stderr, "makeCurves() ");
    perror(s);
    return(2);
  } /* if */
#else
  fd = stdout;
#endif

  if (curve && curve->count && curve->rewcount) {
    int p;
    double max, delta, step;
    double Xmax, Ymax, Xfactor, Yfactor;

    if (status = writeProlog(fd)) 
      goto EXIT;
            
    for (p = 0, rew = (curve->rew); 
           p < curve->rewcount && rew != NULL; 
             rew = (curve->rew)+(++p)) {
      fprintf(fd, "\\vspace{1cm}\n");
      fprintf(fd, "\\beginpicture\n");
      fprintf(fd, "\\setplotarea x from 0 to %f, y from 0 to %f\n",
              XMAX, YMAX);

      /* GET MAX OF VALUE */
      for (i = 0, max = DBL_MIN; i < curve->count; i++) {
        max = rew->value[i]>max ? rew->value[i] : max;
      }
          
      /* GET X- AND Y-MAX_ON_AXIS */
      Ymax = nextRoundNumber(max);
      Xmax = nextRoundNumber(curve->to);

      Yfactor = (YMAX-0.5)/Ymax;
      Xfactor = (XMAX-0.5)/Xmax;

      /* MAKE THE AXIS */
      fprintf(fd, "\\axis bottom label {%s} \n", curve->parameter);
      fprintfTicks(fd, Xmax, Xfactor);
      fprintf(fd, "\\axis left label {%s} \n", rew->name);
      fprintfTicks(fd, Ymax, Yfactor);

      /* PLOT THE LINE */
      fprintf(fd, "\\plot ");
      for (i = 0, step = curve->from; 
             i < curve->count; 
               i++, step+=curve->step) {
        fprintf(fd, "  %lf %lf",
                (step*Xfactor /* +0.5 */),
                (rew->value[i]*Yfactor /* +0.5 */));
      } /* for */
      fprintf(fd, " /\n");

      /* PUT THE POINTS */
      for (i = 0, step = curve->from; 
             i < curve->count; 
               i++, step+=curve->step) {
        /* \\circle{2}, oae */
        fprintf(fd, "\\put {\\circle*{3}} [Bl] at %lf %lf\n",
                (step*Xfactor /* +0.5 */),
                (rew->value[i]*Yfactor /* +0.5 */));
      } /* for */

      fprintf(fd, "\\endpicture\n\n");
    }

    if (status = writeEpilog(fd))
      goto EXIT;
  } /* for */

EXIT:

  if (fd && fd != stdout)
    fclose(fd);
  return status;

} /* makeCurves */



#ifdef __STDC__
  double nextRoundNumber(double x)
#else
  double nextRoundNumber(x)
    double x;
#endif
{
  double  d = x;
  double  f = 1.0;
  double  m = 0.0;

  while (d <= 1.0) {
    f *= 10.0;
    d = f * x;
  } /* while */

  while (d > 10.0) {
    f *= 0.1;
    d = f * x;
  } /* while */

  if (d <= 1.5)
    m = 1.5;
  else if (d <= 2.0)
    m = 2.0;
  else if (d <= 2.5)
    m = 2.5;
  else if (d <= 3.0)
    m = 3.0;
  else if (d <= 4.0)
    m = 4.0;
  else if (d <= 5.0)
    m = 5.0;
  else if (d <= 7.5)
    m = 7.5;
  else 
    m = 10.0;

#ifdef debug
  fprintf(stderr, "nextRoundNumber of %f is %f.\n", x, m/f);
#endif
  return m/f;
} /* nextRoundNumber */



#ifdef __STDC__
  void fprintfTicks(FILE *f, double x, double picFactor)
#else
  void fprintfTicks(f, x, picFactor)
    FILE    *f;
    double  x;
    double  picFactor;
#endif
{
  double  d        = x;
  double  factor   = 1.0;
  double  z        = 0.0;
  double  step     = 0.0;
  double  count    = 0.0;
  long    l        = 0L;
  char    *format  = NULL;
  char    floating = 0;
  char    expo     = 0; 

  fprintf(f, "\tticks withvalues 0 ");

  while (d <= 1.0) {
    factor *= 10.0;
    d       = factor * x;
  } /* while */

  while (d > 10.0) {
    factor *= 0.1;
    d       = factor * x;
  } /* while */
    
  if (d == 1.5) 
    count = 3.0;
  else if (d == 2.0) 
    count = 4.0;
  else if (d == 2.5)
    count = 2.5;
  else if (d == 3.0)
    count = 3.0;
  else if (d == 4.0)
    count = 4.0;
  else if (d == 5.0)
    count = 5.0;
  else if (d == 7.5)
    count = 3.0;
  else if (d == 10.0)
    count = 5.0;
  else {
    fprintf(stderr, "printTicks not implemented for %f\n", d);
    count = 10.0;
  } /* if */

  floating = 0;
  /* TEST FOR FLOATING NUMBERS IN TICKS */
  for (z = step = x/count; z <= x; z += step) {
    l = z; /* ROUNDING */
    if (floating = (fabs(z - l) > 0.1)) {
      break;
    } /* if */
  } /* if */

  /* SPECIFY THE TICK VALUES FORMAT */
  if (floating)
    format = "  %.1f";
  else
    format = "  %.0f";

  expo = 0;
  if (factor<0.001 || factor>100) 
    expo = 1;

#ifdef debug
  fprintf(stderr, "fprintfTicks: floating is %d, expo is %d\n", 
          floating, expo);
#endif

  /* PRINT THE TICK VALUES */
  for (z = step = x/count; z <= x; z += step)
    if (!expo)
      fprintf(f, format, z);
    else {
      fprintf(f, format, z*factor);
      fprintf(f, "$\\cdot 10^{%.0f}$", (-1.0)*log10(factor));
    } /* if */

    /* WHERE TO PRINT */
    fprintf(f, " /\n\t from %f to %f by %f\n",
            0.0, x*picFactor, step*picFactor);

    fprintf(f, "/\n"); /* end of axis */

} /* fprintfTicks */



#ifdef __STDC__
  int writeProlog(FILE *f)
#else
  int writeProlog(f)
    FILE *f;
#endif
{

#ifdef CURVE_ALONE
  fprintf(f, 
"\\documentstyle[12pt]{dinletter}\n\
\\pagestyle{empty}\n\
\\begin{document}\n\
\n\
\\setlength{\\topmargin}{0cm}\n\
\\setlength{\\marginparwidth}{0cm}\n\
\\setlength{\\marginparsep}{0cm}\n\
\\setlength{\\marginparpush}{0cm}\n\
\\setlength{\\topmargin}{0cm}\n\
\\setlength{\\headheight}{0cm}\n\
\\setlength{\\headsep}{0cm}\n\
\\setlength{\\footheight}{0cm}\n\
\\setlength{\\footskip}{0cm}\n");
#endif

  fprintf(f,
"\n\
\\input prepictex\n\
\\input pictex\n\
\\input postpictex\n\
\n\
\\setcoordinatesystem units <1cm,1cm>\n\
\n");

  return 0;

} /* writeProlog */



#ifdef __STDC__
  int writeEpilog(FILE *f)
#else
  int writeEpilog(f)
    FILE *f;
#endif
{
#ifdef CURVE_ALONE
  fprintf(f, "\n\\end{document}\n");
#endif
  return 0;
} /* writeEpilog */



#ifdef CURVE_ALONE
  #ifdef __STDC__       
    main (int argc, char **argv)
  #else
    main (argc, argv)
      int     argc;
      char**  argv;
  #endif
{
  CurveT  curve;
  char    s[MAXPATHLEN];

  if (!argc) {
    fprintf (stderr, "%s no netname given!\n", argv[0]);
    exit(1);
  } /* if */

  if (argv[1][0] != '/') 
    sprintf(s, "%s/%s", getenv("DSPNMODELDIR"), argv[1]);
  else
    strcpy (s, argv[1]);

  if (readCurves(s, &curve))
    fprintf(stderr, "Error reading CURVES : %s\n", s);
  else
    makeCurves(s, &curve);

} /* main */

#endif

