/*$Header: /home/pdv/timenet/CVS/TimeNET/Print/src/PaintNet.c,v 1.7 2001/04/26 06:33:40 rg Exp $*/
/*******************************************************************************
Modul           :       PaintNet.c

Purpose         :       Erzeugen des PostScript-Textes zum Drucken eines 
                        DSPNexpress-Petri-Netzes
                        Erzeugen des LaTeX-Textes zur Beschreibung des 
                        DSPNexpress-Petri-Netzes 

Project Director:       Christoph Lindemann TU Berlin

Programmer      :       Christian L"uhe

History         :       release 1: November 1992

                :       release 2: Juni 1994
                        Joachim Schneider: Anpassung an .TN-Format
*******************************************************************************/

#define PAINTNET
#include <stdio.h>
#include <string.h>
#include <dirent.h>

#if defined(__NetBSD__) || defined(linux)
#include <float.h>
#define MAXFLOAT FLT_MAX
#define MINFLOAT FLT_MIN
#else
#include <values.h>  /* MAX/MIN/FLOAT MAXINT but no MININT */
#endif

#include "netio.h"
#include "Paint.h"
#include "Allocate.h"

#ifdef __STDC__
  static void printSyntax(void);
  static void scaleNet(NetDescr *net, double scale);
  static void getBoundingBox(NetDescr *net);
  static void MinMax(double *a, double *A, double x);
#else
  static void printSyntax();
  static void scaleNet();
  static void getBoundingBox();
  static void MinMax();
#endif


/* be careful: this definition depends on that in prolog.[e]*ps */
#define FONT_SIZE 10.0/(PointsPerInch)


int     marks  = 1;
int     tags   = 0;
int     spline = 0;
double  zoom   = 1.0;
int     level  = 0;
double  scale  = 1.0;   /* scales like edit (only positions, not obj) */
int     eps    = 0;     /* flag for encapsulated PostScript */

#ifdef __STDC__
  main (int argc, char **argv)
#else
  main (argc, argv)
  int argc;
  char **argv;
#endif
{
  NetDescr  *net     = NULL;
  char      *netname = NULL;
  int       error    = 0;
  int       i        = 1;
    
  if (argc>1)

    while (i<argc) {
      int l;
        
      for (l=1; (l<strlen(argv[i])) && (!error) && (argv[i][0]=='-'); l++)
        switch (argv[i][l]) {
          case 'c' :                    /* "c"urves, "s" is for scale */
            if (!spline)
	      spline=1;
            else
	      error=1;
            break;
          case 'e' :                    /* encapsulated PostScript */
            if (!eps)
	      eps=1;
            else
	      error=1;
            break;
          case 'l' : 
            if ((!level) && (sscanf(argv[++i], "%d", &level)))
	      ;
            else
	      error=1;
            break;
          case 'm' :
            if (marks)
	      marks=0;
            else
	      error=1;
            break;
          case 's' :                    /* scale */
            if ((scale==1.0) && (sscanf(argv[++i], "%f", &scale)))
	      ; /* ok */
            else
	      error=1;
            break;
          case 't' :
            if (!tags)
	      tags=1;
            else 
	      error=1;
            break;
          case 'z' :
            if ((zoom==1.0) && (sscanf(argv[++i], "%f", &zoom)))
	      ; /* ok */
            else
	      error=1;
            break;
          default: error=1;
        } /* switch */
      i++;
    } /* while */
  else
    error=1;
    
  if (error) {
    printSyntax();
    return (1); 
  } /* if */

  netname = argv[argc-1];

  /* .TN einlesen */
  net = load_net( netname, NULL );

  if (scale!=1.0)
    scaleNet(net,scale);

  if (eps)
    getBoundingBox(net);

  for ( i=0; i < net->num_places; i++)
    PaintPlace(net->place[i]);

  for ( i=0; i < net->num_trans; i++)
    PaintTrans(*net, net->trans[i]);

  return (0);

} /* main */



#ifdef __STDC__
  static void printSyntax (void)
#else
  static void printSyntax ()
#endif
{
    fprintf(stderr,"paint: Syntax-Error!\n");
    fprintf(stderr,
            "syntax: 'paint [-e] [-s scaledouble] netname'\n");
}

/* SCALE means scaling of positions/coordinates - NOT sizes/dimensions */
#ifdef __STDC__
  static void scaleNet(NetDescr *net, double scale)
#else
  static void scaleNet(net, scale)
    NetDescr  *net; 
    double    scale;
#endif
{
    int i,j,k;
  if (!net) { 
    fprintf(stderr, "scaleNet: can't scale a NULL-net\n");
    return;
  }
  fprintf (stderr, "scaleNet: scale coordinates by %f\n", scale);

  if (scale==1.0)
    return;

  for (i=0; i < net->num_places; i++) {
    net->place[i].pos.x      *= scale;
    net->place[i].pos.y      *= scale;
    net->place[i].name_pos.x *= scale;
    net->place[i].name_pos.y *= scale;
  } /* for */

  for (i=0; i < net->num_trans; i++) {
    net->trans[i].pos.x       *= scale;
    net->trans[i].pos.y       *= scale;
    net->trans[i].name_pos.x  *= scale;
    net->trans[i].name_pos.y  *= scale;
    net->trans[i].delay_pos.x *= scale;
    net->trans[i].delay_pos.y *= scale;
    for (j=0; j < net->trans[i].num_inarcs; j++) {
      for (k=0; k < net->trans[i].in_arc[j].num_pts; k++) {
        net->trans[i].in_arc[j].point[k].x *= scale;
        net->trans[i].in_arc[j].point[k].y *= scale;
      } /* for */
    } /* for */
    for (j=0; j < net->trans[i].num_outarcs; j++) {
      for (k=0; k < net->trans[i].out_arc[j].num_pts; k++) {
        net->trans[i].out_arc[j].point[k].x *= scale;
        net->trans[i].out_arc[j].point[k].y *= scale;
      } /* for */
    } /* for */
    for (j=0; j < net->trans[i].num_inharcs; j++) {
      for (k=0; k < net->trans[i].inh_arc[j].num_pts; k++) {
        net->trans[i].inh_arc[j].point[k].x *= scale;
        net->trans[i].inh_arc[j].point[k].y *= scale;
      } /* for */
    } /* for */
  } /* for */
} /* scaleNet */



#ifdef __STDC__
  static void getBoundingBox (NetDescr *net)
#else
  static void getBoundingBox (net)
    NetDescr  *net;
#endif
{
  double x=MAXFLOAT, y=MAXFLOAT, X=MINFLOAT, Y=MINFLOAT;
  double tagx, tagy;
  int i,j,k;

  if (!net) {
#ifdef DEBUG
    fprintf(stderr, "getBoundingBox: can't compute BBox from NULL-net");
#endif
    return;
  } /* if */

  for (i=0; i < net->num_places; i++) {
    MinMax (&x,&X,(net->place[i].pos.x + PixToInch(DEFAULT_SIZE/2.0)));
    MinMax (&y,&Y,(net->place[i].pos.y + PixToInch(DEFAULT_SIZE/2.0)));
    MinMax (&x,&X,(net->place[i].pos.x - PixToInch(DEFAULT_SIZE/2.0)));
    MinMax (&y,&Y,(net->place[i].pos.y - PixToInch(DEFAULT_SIZE/2.0)));
    MinMax (&x,&X,net->place[i].name_pos.x);
    MinMax (&y,&Y,net->place[i].name_pos.y);
    MinMax (&x,&X,(net->place[i].name_pos.x + 
                      strlen(net->place[i].name)*FONT_SIZE));
    MinMax (&y,&Y,(net->place[i].name_pos.y + FONT_SIZE));
  } /* for */

  for (i=0; i < net->num_trans; i++) {
    MinMax (&x,&X,(net->trans[i].pos.x + PixToInch(DEFAULT_SIZE/2)));
    MinMax (&y,&Y,(net->trans[i].pos.y + PixToInch(DEFAULT_SIZE/2)));
    MinMax (&x,&X,(net->trans[i].pos.x - PixToInch(DEFAULT_SIZE/2)));
    MinMax (&y,&Y,(net->trans[i].pos.y - PixToInch(DEFAULT_SIZE/2)));
    MinMax (&x,&X,net->trans[i].name_pos.x);
    MinMax (&y,&Y,net->trans[i].name_pos.y);
    MinMax (&x,&X,(net->trans[i].name_pos.x +
                      strlen(net->trans[i].name)*FONT_SIZE));
    MinMax (&y,&Y,(net->trans[i].name_pos.y + FONT_SIZE));    
    for (j=0; j < net->trans[i].num_inarcs; j++) {
      for (k=0; k < net->trans[i].in_arc[j].num_pts; k++) {
        MinMax (&x,&X,net->trans[i].in_arc[j].point[k].x);
        MinMax (&y,&Y,net->trans[i].in_arc[j].point[k].y);
      } /* for */
    } /* for */
    for (j=0; j < net->trans[i].num_outarcs; j++) {
      for (k=0; k < net->trans[i].out_arc[j].num_pts; k++) {
        MinMax (&x,&X,net->trans[i].out_arc[j].point[k].x);
        MinMax (&y,&Y,net->trans[i].out_arc[j].point[k].y);
      } /* for */
    } /* for */
    for (j=0; j < net->trans[i].num_inharcs; j++) {
      for (k=0; k < net->trans[i].inh_arc[j].num_pts; k++) {
        MinMax (&x,&X,net->trans[i].inh_arc[j].point[k].x);
        MinMax (&y,&Y,net->trans[i].inh_arc[j].point[k].y);
      } /* for */
    } /* for */
  } /* for */

  printf("%%%%BoundingBox: %d %d %d %d\n", /* cause of rotation in ps-code */
         ((int)InchToPoint(y))-5,          /* we've got to give the */
         ((int)InchToPoint(x))-5,          /* coordinates of BoundinBox in */
         ((int)InchToPoint(Y))+5,          /* y,x,Y,X-order */
         ((int)InchToPoint(X))+5);         /* normal is x,y,X,Y */
} /* */



/* see code */
#ifdef __STDC__
  static void MinMax (double *a, double *A, double x)
#else
  static void MinMax (a, A, x)
    double *a, *A, x;
#endif
{
  if (x<*a)
    *a= x;
  if (x>*A)
    *A= x;
} /* MinMax */





