/*$Header: /home/pdv/timenet/CVS/TimeNET/Print/src/ReadNet.c,v 1.6 1994/12/06 18:03:04 joa Exp $*/
/*******************************************************************************
Modul           :       ReadNet

Purpose         :       Einlesen der Netz-Beschreibung in interne Datenstruktur

Project Director:       Christoph Lindemann TU Berlin

Programmer      :       Christian L"uhe

History         :       release 1: November 1992
                        release 2: September 1993 
			release 3: Juni      1994
			Joachim Schneider: Anpassung an .TN-Format

*******************************************************************************/

/*#define debug*/
#define READNET

#include <stdio.h> 
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include "netio.h"
#include "Allocate.h"
#include "Paint.h"
#include "Result.h"
#include "Structure.h"

#ifndef SEEK_SET    /* under SunOS */
#define SEEK_SET 0
#endif

#ifdef ultrix
  /* Spezialbehandlung fuer die DEC-Kisten: */
  static char *strdup (string)
  char *string;
{
  register char *new_string = (char *)malloc (strlen (string) + 1);
  if (new_string)
    strcpy (new_string, string);
  return new_string;
}
#endif

/* declaration of private functions */
#ifdef __STDC__
  static int openNET(char *name, char *extension, char *mode, FILE **fd);
  static int fskipLine(FILE *fd);
#else
  static int openNET();
  static int fskipLine();
#endif

#define DEFAULT_TAG_POSITION -137.17



/*-------------------------------------- ReadResult ------------------*/
#ifdef __STDC__
  int ReadResult(NetDescr *net, char *name, ResultT *res)
#else
  int ReadResult(net, name, res)
  NetDescr *net;
  char     *name;
  ResultT  *res;
#endif
{
  FILE    *resfd   = NULL;
  FILE    *tnfd    = NULL;
  int     i        = 0;
  int     j        = 0;
  int     k        = 0;
  int     eof      = 0;
  char    name_buf[256];
  char    definition_buf[10000];
  char    format[256+6];
  char    path[MAXPATHLEN];
  time_t  TNtime;
  struct  stat buf;

  if(! name) {
    fprintf (stderr, "Read net: No Netname given!\n");
    return (NONAME_ERROR);
  } /* if */

  sprintf(path, "%s.TN", name);
  if (lstat (path, &buf)) {
    perror(path);
    exit(errno);
  } else
    TNtime = buf.st_mtime;

  sprintf(path, "%s.RESULTS", name);
  if (lstat (path, &buf)) {
/*
    dont care, let the results information away and print the def.         
    perror(path);
    exit(errno);
*/
  } else
    if (buf.st_mtime < TNtime) {
      fprintf(stderr, "%s is older than <net>.TN\n", path);
      fprintf(stderr, "The results are not uptodate - you should resolve the model.\n");
    } /* if */

  if (openNET(name,"RESULTS","r",&resfd) == OPEN_ERROR) 
    ; /* ok lets only handle the definition .. return (OPEN_ERROR); */
    
/*
  for( j = 0; j < resultCount; j++ ) {
    res->name[j]       = strdup( net->reward[j].name );
    res->definition[j] = strdup( net->reward[j].name );
  }
*/

  if (resfd) {
    double    d;
    int       next;
    unsigned  eps;
    FILE      *envfd;
    char      envName[MAXNAMLEN]; 
        
    /* get the epsilon value from Environment file */
    sprintf(envName,"%s/.TimeNET", getenv("HOME"));
    if (openNET(envName,"","r",&envfd) == OPEN_ERROR) 
      res->epsilon=(unsigned)7;
    else {
      do {
        k=fscanf(envfd, "EPSILON %*[^-]-%u", &eps);
        next=fskipLine(envfd);           /* goto next line */
      } while (k!=EOF && next!=EOF && k!=1);
      if (k==1)
        res->epsilon=eps;
      else {
        fprintf(stderr, "ReadResult: can't find EPSILON-Entry in Environment-File\n");
        res->epsilon=(unsigned)7;
      } /* else */
    } /* else */
        

    for (i=0; i < net->num_rewards; i++) {
      rewind (resfd);
      sprintf(format, "%s = %%lf", net->reward[i].name );
      do {
        k=fscanf(resfd, format, &d);     /* get the value */
        next = fskipLine(resfd);         /* goto next line */
      } while (k!=EOF && next!=EOF && k!=1);

      if (k==EOF || next==EOF) {          /* no entry for that MEASURE */
        res->availableResult[i] = (char)FALSE;
        fprintf(stderr, 
          "BAD RESULTS-FILE! no entry for MEASURE: '%s'.\n",
	  net->reward[i].name);
      } else {
        res->availableResult[i] = (char)TRUE;
        res->value[i]           = d;
      } /* if */
    } /* for */
  } else 
    /* res->availableResult[i] = 0 */
    for (i=0; i < net->num_rewards; i++) 
      res->availableResult[i] = (char)FALSE;

  return (0);

} /* ReadResult */



#ifdef __STDC__
  int ReadStructure(NetDescr *net, char *name, StructureT *structure)
#else
  int ReadStructure(net, name, structure)
  NetDescr   *net;
  char       *name;
  StructureT *structure;
#endif
{
    FILE *fd;
    char buf[1024];
    int eof;
    
    if (structure==NULL)
        return;
    if (name==(char*)NULL) {
        fprintf (stderr, "Read net: No Netname given!\n");
        return (NONAME_ERROR);
    }

    if (openNET(name,"INV", "r", &fd) == OPEN_ERROR)
        /* ok lets handle only the definition .. return (OPEN_ERROR); */
        return 0; 
    else {
        /* read over "*** Place invariants ***" */
        while ((eof=getline(fd, buf)) && 
               strcmp(buf, "*** Place invariants ***\n"));
        
        /* read the invariants upto the comment (beginns with ***) or the end */
        while ((eof=getline(fd, buf)) && 
               strncmp(buf,"***",3)) 
            if (eof>1) {
                structure->invCount++;
                if ((structure->invCount%20)==0)
                    structure->inv = (char **)realloc((char *)structure->inv, 
                                                      (structure->invCount + 20)
                                                      * sizeof(char *));
                structure->inv[structure->invCount-1] = strdup(buf);
            }
        /* now it is the comment or the end */
        if (eof>1)  /* comment ... */
            structure->comment = strdup(&buf[3]);   /* ... without "***" */
    }

    if (openNET(name,"ECS", "r", &fd) == OPEN_ERROR)
        /* ok lets handle only the definition .. return (OPEN_ERROR); */
        return 0; 
    else {
        /* read over "*** Extended conflict sets ***" */
        while ((eof=getline(fd, buf)) && 
               strcmp(buf, "*** Extended conflict sets ***\n"));        

        /* read the ECS */
        while ((eof=getline(fd, buf)))
            if (eof>1) {
                structure->ecsCount++;
                if ((structure->ecsCount%20) == 0)
                    structure->ecs = (char **)realloc((char *)structure->ecs,
                                                      (structure->ecsCount + 20)
                                                      * sizeof(char *));
                structure->ecs[structure->ecsCount-1] = strdup(buf);
            }
    }
}

#ifdef __STDC__
int getline(FILE *f, char *l)
#else
int getline(f, l)
    FILE *f;
    char *l;
#endif
{
    register i=0,eof;
    
    while ((eof=fgetc(f))!=EOF && (l[i++]=eof)!='\n');
    l[i] = '\0';
    return i;
}



/*********************************************************************/
static int openNET(name,extension,mode,fd)
  char  *name;
  char  *extension;
  char  *mode;
FILE **fd;
{
  char path[MAXNAMLEN];

  if (name[0] == '/')
    sprintf(path,"%s.%s",name,extension);
  else
    sprintf(path,"%s/%s.%s",
            getenv("MODELDIR"),name,extension);
  if ((*fd = fopen(path,mode)) == NULL) {
    if (!strcmp(extension,"TN"))
      fprintf(stderr,"openNET: can't open %s \n",path);
      return OPEN_ERROR;
  } /* if */
  return OK;
} /* openNet */



#ifdef __STDC__
  static int fskipLine(FILE *fd)
#else 
  static int fskipLine(fd)
    FILE *fd;
#endif
{
  register c;

  do
    c = fgetc(fd);
  while (c != EOF && c != '\n');
  return c;
} /* skip_line */

