/*******************************************************************************
Modul           :       options.h

Purpose         :       Definition der Options-Strukturen
			fuer die verschiedenen Evaluierungs-Verfahren
			und die jeweiligen Experimente

Programmer      :       Kolja Koischwitz
			transientes von Joerg Mitzlaff in 10.94

History         :       release 1: January 1994
*******************************************************************************/

#define NAME_BUF_LEN 64 /* Anzahl Zeichen fuer eine Namenseingabe */
#define VAL_BUF 32   /* Anzahl Zeichen fuer die Werteingabe */

#include <X11/Intrinsic.h>

/* "=1" ist sehr wichtig! XtNradioData mit Wert 0 geht nicht! */
typedef enum {SteadyState=1,Transient} SolutionType;
typedef enum {Duration=1,Point} OutputData;
typedef enum {Iterative=1,Direct} LinearSolutionMethod;
typedef enum {Ssimstandard=1,Ssimfast} SimulationTech;
typedef enum {Sequential=1,Parallel} Distribution;
typedef enum {Detect_ON=1,Detect_OFF} DetectInitial;
typedef enum {Visual_ON=1,Visual_OFF} Visualisation;
typedef enum {Marking=1,Delay} ParameterType;
typedef enum {SpectralVariance=1,ControlVariales} StatisticalMethod;
typedef enum {Rule_ON=1,Rule_OFF} Rule;
typedef enum {embno=1,embtime,embmem} Embedding;   /*new*/
typedef struct
{       /* Experiment Options */
	char parameter_name[NAME_BUF_LEN],
	  from_value[NAME_BUF_LEN],
	  to_value[NAME_BUF_LEN],
	  step_value[NAME_BUF_LEN];
	ParameterType parameter_type;
} ExperimentOptions;

typedef struct
{       /* Options for Transient Solution */
	double transient_time;
	double stepsize_method;
	OutputData output_data;
	double stepsize_result;
	int cluster_size;
	double precision;
	int keep_files;
	int graphics_desired;
	int old_solve_alg;
	int old_case_iii;
} TransientOptions;

typedef struct
{       /* Options for Transient Simulation */
        double TransientInstantOfTime;
        int NumberOfSamplingPoints;
        Distribution distribution;
        Rule PercentageRule;
	double confidence_level,
	       max_relative_error;
	long seed_value;
        long max_CPU_time;
        Boolean no_CPU_time_limit;
} TransientsimOptions;

typedef struct
{       /* Options for Analytic Solution */
	LinearSolutionMethod solution_method;
	long max_iterations;
	Distribution smc_solution;
	double precision;
} AnalyticOptions;

typedef struct
{       /* Options for Approximation */
	LinearSolutionMethod solution_method;
	long max_iterations,phases,virtual_phases;
	double precision,droptolerance;
} ApproximationOptions;


typedef struct
{       /* Options for Simulation */
	SimulationTech simulation_tech;
	StatisticalMethod statistical_method;
	Distribution distribution;
        DetectInitial detect_init;
	double confidence_level,
               errPercent,
	       max_relative_error;
	long seed_value,max_sample_number;
        long minFirings;
	long max_model_time, max_CPU_time;
        Visualisation visualisation;
	Boolean no_sample_limit,no_model_time_limit,no_CPU_time_limit;
	int opt_thresh,CPU_Time_Run;
} SimulationOptions;

typedef struct
{       /* Options for Discrete Time Analytic Solution */
	LinearSolutionMethod solution_method;
	long max_iterations;
	Embedding embedding;
	double precision;
        long scaling;
} DTSPNAnalyticOptions;

typedef struct
{       /* Options for Discrete Time Transient Solution */
	long time_steps;
	double precision;
        long step_size;
        long first_step;
        Embedding embedding;
        long scaling;
} DTSPNTransientOptions;

/* Structure for MAC-Like Keyboard-Focus-Change */
typedef enum {DOWN,UP,SET} FocusMovingDirection;

#define MAX_TEXT_WIDGETS 10
typedef struct
{
	int curr;
	int num;
	Widget *shell;
	Widget *field[MAX_TEXT_WIDGETS];
} KeyboardFocus;


/*
 RESTART 

Usage : RESTART model -D|S <max_execution_time> <max_thresholds>


 SIMULATE

Usage : SIMULATE model -D|S -S|C|RS -Ton|Toff <conf.level> <epsilon> 
                 <seed> <max.samples> <max.model time> <max.real time>

*/





