#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.h"			/* definitions and types */
#include "portab.h"

extern trans_solve_solution;

extern float mdeetaaar();			/* from _MDF.c; marking-dependent rates */

extern char *emalloc();				/* from graph_main */
extern void efree();
extern int *decode_submarking();		/* from graph_subnet */
extern sub_marking *net_end_markings();
extern int test_concession();
extern void fire_transition();
extern neis_cnt neis_mult();

int	*table;
int	*current_mark;


/*	print_global_marking shows a global marking (for error messages) */
void
print_global_marking(mark)
	int	*mark;		/* marking */
{
	int pla, tok, no, sb, bit=0;
	subnet	*sub;

	for(pla=0; pla<timed.no_places; pla++)
	{	no = timed.places[pla].number;
		tok = mark[no];
		if (tok)
		{	bit = 1;
			if (tok>1)
				printf("%d*", tok);
			printf("%s  ", timed.places[pla].name);
		}
	}
	for(sb = 0; sb < no_subnets; sb++)
	{	sub = &subnets[sb];
		for(pla=0; pla<sub->no_places; pla++)
		{	no = sub->places[pla].number;
			tok = mark[no];
			if (tok)
			{	bit = 1;
				if (tok>1)
					printf("%d*", tok);
				printf("%s  ", sub->places[pla].name);
			}
		}
	}
	if (!bit)
		printf("No Token in places.");
	printf("\n");
}

/*	global_marking obtains a global marking from a coded tangible marking	*/
static int *help_mark = NULL;
void
global_marking(sub_marks, disc_mark, marking)
	word	**sub_marks;	/* the submarkings for each subnet */
	word	*disc_mark;	/* the disconnected-places-marking */
	int	*marking;	/* marking; return parameter */
{
	int	i, s_nr;
	t_sub	*t_subnet = &timed;
	subnet 	*sub = subnets;

	if (help_mark == NULL)
		help_mark = (int *)emalloc(no_places * sizeof(int));
	if (t_subnet->no_places)
	{	decode_submarking(t_subnet->no_places, t_subnet->places, disc_mark, help_mark);
		for(i=0; i<t_subnet->no_places; i++)
			marking[t_subnet->places[i].number] = help_mark[i];
	}
	for(s_nr=0; s_nr<no_subnets; s_nr++)
	{	decode_submarking(sub[s_nr].no_places, sub[s_nr].places,
							sub_marks[s_nr], help_mark);
		for(i=0; i<sub[s_nr].no_places; i++)
			marking[sub[s_nr].places[i].number] = help_mark[i];
	}
}

/*	transition_rate returns the transition firing rate */
double /* exp_transition_rate() */
transition_rate(tsub, tra, mark)
	t_sub	*tsub;		/* timed transition's subnet */
	int	tra;		/* number of transition */
	int	*mark;		/* the current marking */
{
	extern double md_rate();
	sp_place *i_mat;
	double	rate;
	int	i, pla, value, min;

	if (tsub->transitions[tra].mark_ind)	/* marking-dependent firing delay */
	{	current_mark = mark;
		rate = md_rate(tsub->transitions[tra].mark_ind);
		if (rate > 1.0e-8)
			return rate;
		printf("ERROR: marking-dependent firing rate of transition ");
		printf("%s evaluates to 0 in marking\n", tsub->transitions[tra].name);
		print_global_marking(mark);
		exit(1);
	}
	if (tsub->transitions[tra].infinite!=IS || tsub->transitions[tra].type!=EXP)
		return tsub->transitions[tra].value;		/* normal type */
	min = MAXINT;
	i_mat = tsub->transitions[tra].i_matrix;
	for(i = 1; i <= tsub->transitions[tra].i_matrix[0].num; i++)
	{	pla = i_mat[i].num;
		value = i_mat[i].plu_id ? md_arc_mult(i_mat[i].plu_id) : i_mat[i].value;
		if (!value) continue;
		if ((value = mark[pla] / value) < min)
			min = value;
	}
	if (!min)
	{	printf("ERROR: firing rate of infinite-server transition ");
		printf("%s evaluates to 0 in marking\n", tsub->transitions[tra].name);
		print_global_marking(mark);
		exit(1);
	}
	return tsub->transitions[tra].value / (double)min;	/* infinite server */
}

/*	calc_prob calculates the probability of a path to a tangible marking */
word **
calc_prob(in, pro)
	sub_marking *in;
	double	*pro;
{
	word	**back;
	int	i;

	*pro = 1.0;
	back = (word **)emalloc(no_subnets * sizeof(word *));
	for(i=0; i<no_subnets; i++)
	{	back[i] = in[i].paths[table[i]].marking;
		*pro *= in[i].paths[table[i]].prob;
	}
	return back;
}

/*	first_tang_marking returns the first tangible marking from a set of
	reachable ones together with the probability and initializes the
	process of walking through all markings	*/
word **
first_tang_marking(input, prob)
	sub_marking *input;		/* set of reachable markings for each subnet */
	double	*prob;			/* probability (return parameter) */
{
	int	i;

	for(i=0; i<no_subnets; i++)
		table[i] = 0;		/* initialize table of counters */
	return calc_prob(input, prob);
}

/*	next_tang_marking returns the next tangible marking from a set of
	reachable ones together with the probability */
word **
next_tang_marking(sub, input, prob)
	subnet	*sub;			/* the net's subnets */
	sub_marking *input;		/* set of reachable markings for each subnet */
	double	*prob;			/* probability (return parameter) */
{
	int	i;

	for(i=0; i<no_subnets; i++)
	{	if (++(table[i]) < input[i].no_paths)
			return calc_prob(input, prob);
		table[i] = 0;
	}
	*prob = 0.0;
	return NULL;			/* no more combinations */
}

/*	merge_paths adds a new firing path */
void
merge_paths(marking, index, prob, delay, tra_no, tra_type )
	tang_marking *marking, *index;
	double prob, delay;
	int tra_no;
	transition_type tra_type;
{
	static int max_entry;

	if ((index == marking)&&(tra_type==EXP))
	{	/* dont care about exponential self-loops */
		return;
	}
	if (!marking->no_time_paths)
	{	marking->no_time_paths = 1;
		marking->time_paths = (tang_path *)emalloc(SPARSE_START * sizeof(tang_path));
		marking->time_paths[0].prob    = prob;
		marking->time_paths[0].delay   = delay;
		marking->time_paths[0].marking = index;
		marking->time_paths[0].tran_no = tra_no;
		max_entry = SPARSE_START;
		no_edges++;
		return;
	}
	if (++(marking->no_time_paths) > max_entry)
	{
		max_entry = SPARSE_NEXT(max_entry);
		if ((marking->time_paths = (tang_path *)
				realloc( marking->time_paths,
					(unsigned) (max_entry * sizeof(tang_path))))==NULL)
		{
			eperror("graph");
		}
	}
	marking->time_paths[marking->no_time_paths-1].prob   = prob;
	marking->time_paths[marking->no_time_paths-1].delay  = delay;
	marking->time_paths[marking->no_time_paths-1].marking  = index;
	marking->time_paths[marking->no_time_paths-1].tran_no = tra_no;
	no_edges++;
}

/*	enter_tang_marking enters a tangible marking into the reduced reachability tree
	and returns a pointer to the found or created marking		*/
tang_marking *
enter_tangmarking(sub_marks, disc_mark)
	word	**sub_marks;	/* pointer to subnet - markings (malloc'd!) */
	word	*disc_mark;	/* pointer to disconnected places marking */
{
	static int mark_num = 0;
	tang_marking *marking;
	tang_node   *node = graph, **insert;
	tsp_pointer *lookup;
	int	sn_nr = 0, left, right, mid, old_bound;
	word	*index;

	if (node == NULL)				/* tree is empty */
	{	node = graph = (tang_node *)emalloc(sizeof(tang_node));
		if (disc_mark == NULL)			/* no disc. places */
			node->bound = 1;
		else	node->bound = SPARSE_START;	/* sparse array */
		node->number = 1;
		node->next = (tsp_pointer *)emalloc((int)(node->bound * sizeof(tsp_pointer)));
		node->next[0].index = disc_mark;
		insert = (tang_node **)&(node->next[0].next);
		node = NULL;				/* don't go into while */
	}
	while((node != NULL) && (sn_nr < no_subnets+1))
	{	index = sn_nr ? sub_marks[sn_nr-1] : disc_mark;
		lookup = node->next;
		left = 0; right = node->number - 1;
		mid = (left + right) / 2;
		if ((void *)index <= (void *)lookup[left].index)
			mid = left;
		if ((void *)index >= (void *)lookup[right].index)
			mid = left = right;
		while(((void *)index != (void *)lookup[mid].index) && (mid != left))
		{	if ((void *)index > (void *)lookup[mid].index)
				left = mid;
			else	right = mid;
			mid = (left + right) / 2;
		}
		if ((void *)index == (void *)lookup[mid].index)		/* found */
			node = (tang_node *)lookup[mid].next;
		else
		{	mid += ((void *)index > (void *)lookup[mid].index);
			if (++(node->number) > node->bound)
			{	old_bound = node->bound;
				node->bound = SPARSE_NEXT(node->bound);
				if (node->bound <= old_bound) node->bound = SPARSE_MAX;
				if ((node->next = (tsp_pointer *) realloc(node->next,
						node->bound * sizeof(tsp_pointer)))==NULL)
				{
					eperror("graph");
				}
				lookup = (tsp_pointer *)node->next;
			}
			memmove(&(lookup[mid+1]),&(lookup[mid]), 
				sizeof(tsp_pointer) * (node->number - mid - 1));
			lookup[mid].index = index;
			insert = (tang_node **)&(lookup[mid].next);
			break;
		}
		sn_nr++;
	}
	if (sn_nr > no_subnets)			/* marking already exists */
	{	efree((char *)(sub_marks));
		marking = (tang_marking *)node;
		return marking;
	}
	/* insert new nodes into tree */
	for(sn_nr++ ; sn_nr <= no_subnets; sn_nr++)
	{	index = sn_nr ? sub_marks[sn_nr-1] : disc_mark;
		*insert = node = (tang_node *)emalloc(sizeof(tang_node));
		node->bound = SPARSE_START;	/* sparse array */
		node->number = 1;
		node->next = (tsp_pointer *)emalloc(SPARSE_START * sizeof(tsp_pointer));
		lookup = node->next;
		lookup->index = index;
		insert = (tang_node **)&(lookup->next);
	}
	marking = (tang_marking *)emalloc(sizeof(tang_marking));
	*insert = (tang_node *)marking;
	marking->number = mark_num++;
	marking->disc = disc_mark;
	marking->mark = sub_marks;
	marking->no_time_paths = 0;
	marking->time_paths = NULL;

	marking->next = new_markings;
	new_markings = marking;			/* new list entry */
	return marking;
}

/*
 * generate_rg calculates the reduced reachability graph
 */
void generate_rg(subs, tsub)
	subnet	*subs;
	t_sub	*tsub;
{
	tang_marking	*marking, *to_marking;
	sub_marking 	*end_mark;
	word		*disc, **path;
	double		prob, delay;
	tang_path	*h_path;
	timed_t *p_tra;
	int		tra, fire, first, *i_mark, *n_mark;
/*
	byte       *h_tnums, h_fire;
	h_tnums = (byte *)emalloc(no_places * sizeof(byte));
*/
	table = (int *)emalloc(no_subnets * sizeof(int));
	no_absorb_states = 0;
	i_mark = (int *)emalloc(no_places * sizeof(int));
	n_mark = (int *)emalloc(no_places * sizeof(int));
	end_mark = net_end_markings(subs, tsub, (int*)NULL, &disc);	/* evaluate M0 */
	path = first_tang_marking(end_mark, &prob);		/* get one tang. marking */
	enter_tangmarking(path, disc);				/* enter as first marking */
	if (next_tang_marking(subs, end_mark, &prob) != NULL)
		printf("Warning: vanishing initial marking with different following markings.\n");
	efree(end_mark);
	while(new_markings != NULL)
	{
		marking = new_markings;
		new_markings = new_markings->next;
		global_marking(marking->mark, marking->disc, i_mark);
		fire = /* h_fire = */ 0;
		marking->neis_mult = (neis_cnt*)emalloc(no_neis_trans * sizeof(neis_cnt));
		for(tra = 0; tra < tsub->no_trans; tra++)/* for each timed transition */
		{	current_mark = i_mark;
			p_tra = &(tsub->transitions[tra]);
			if (p_tra->type==DET && p_tra->infinite==IS)
            /* mark deterministic infinite server as not activated */
				marking->neis_mult[p_tra->det_no] = (neis_cnt)0;
			if (!test_global_conc(i_mark, p_tra->i_matrix, p_tra->h_matrix,
									p_tra->infinite==IS ))
				continue;			/* not fireable */
			fire = 1;
			if (p_tra->type==DET && p_tra->infinite==IS)
			/* calculate how often a deterministic inifinite server is activated */
				marking->neis_mult[p_tra->det_no] = neis_mult(i_mark,p_tra->i_matrix);
/*			current_mark = i_mark;    is already set */
			fire_global_trans(i_mark, n_mark, no_places,
					tsub->transitions[tra].c_matrix);
			end_mark = net_end_markings(subs, tsub, n_mark, &disc);
			path = first_tang_marking(end_mark, &prob);

			first = 1;
			delay = transition_rate(tsub, tra, i_mark);
			while(path != NULL || first)
			{	first = 0;
				to_marking = enter_tangmarking(path, disc);
				merge_paths(marking, to_marking, prob, delay, tra, p_tra->type );
				path = next_tang_marking(subs, end_mark, &prob);
			}
			efree(end_mark);
		} /* for each timed transition */

		if (!fire)
		{	no_absorb_states++;
			printf("No transition fireable in marking:  ");
			print_global_marking(i_mark);
			if(!trans_solve_solution)
			{
			  printf("ERROR: absorbing state detected,");
			  printf(" no steady-state solution possible.\n");
			  exit(1);
			}
		}
		h_path = marking->time_paths;		/* fit space */
		marking->time_paths = (tang_path *)emalloc(marking->no_time_paths * sizeof(tang_path));
		memcpy(marking->time_paths, h_path, marking->no_time_paths * sizeof(tang_path));
		efree((char *)(h_path));
/*
		marking->fire_tra_num = h_fire;
		marking->fire_nums = (byte *)emalloc(h_fire * sizeof(byte), 68);
		for (h_fire = 0; h_fire < marking->fire_tra_num; h_fire++)
			marking->fire_nums[h_fire] = h_tnums[h_fire];
		if (marking->det)
		{	i = tsub->transitions[marking->det_num].det_no;
			marking->next = det_markings[i];
			det_markings[i] = marking;
			no_nodes[i]++;
		} else
		{	marking->next = exp_markings;
			exp_markings = marking;	
			no_exp_markings++;
		}
*/
		no_tang_markings++;
	} /* while(new_markings != NULL) */
/*
	efree(h_tnums);
*/
}
