#include "types.h"	/* sp_place */
extern char *emalloc();

/*
 *	skip_line skips to the beginning of the next line
 */
char skip_line(f)
	FILE *f;
{       char buffer[100];
	fgets(buffer, 100, f);
	return buffer[0];
}

/*
 *	read_sparse reads a sparse row
 */
sp_place *read_sparse(f)
	FILE *f; 
{
	int count, i;
	sp_place *point;

	fscanf(f, "%d ", &count);	/* number of entries */
	point = (sp_place *)emalloc((count+1) * sizeof(sp_place));
	point[0].num = count;		/* first entry.num = number of entries */
	for(i = 1; i<=count; i++)
	{	fscanf(f, "%d %d", &(point[i].num), &(point[i].value));
      fscanf(f, "%d %d", &(point[i].plu_id), &(point[i].min_id));
	}
	return point;
}

/*
 *	read_set reads a set of transitions
 */
byte *read_set(f)
	FILE *f;
{
	int count, i, b;
	byte *point;

	fscanf(f, "%d ", &count);	/* read number of entries */
	point = (byte *)emalloc((count+1) * sizeof(byte));
	point[0] = count;		/* first entry: number of entries */
	for(i = 1; i <= count; i++)
	{	fscanf(f, "%d", &b);
		point[i] = b;
	}
	return point;
}

/*
 *	read_struct reads the structural description file net.STRUCT
 */
void read_struct(s_file)
	FILE *s_file;
{
	int	i, num, max, tmp;
	timed_t	*t_tra;
	imm_t	*i_tra;
	place	*pla;
	subnet	*sub;

	no_neis_trans = no_det_trans = 0;
	skip_line(s_file);		/* skip first line (comment) */
	fscanf(s_file, "%d %d", &no_trans, &no_places);
	fscanf(s_file, "%d", &max);
	timed.no_trans = max;
	timed.transitions = (timed_t *)emalloc(max * sizeof(timed_t));
	timed.no_markings = 0;
	for(num = 0; num < max; num++)
	{	t_tra = &(timed.transitions[num]);
		fscanf(s_file, "%s %d %d ", t_tra->name, &t_tra->number, &tmp );
		switch(tmp){
		case 0: t_tra->type = EXP; break;
		case 1: t_tra->type = DET; break;
		case 2: t_tra->type = GEN; break;
		default:
			fprintf(stderr,"%s illegal transition type: %d in .STRUCT\n",
					progname, tmp );
			exit(1);
			break;
		}
		fscanf(s_file, "%d %d ", &t_tra->infinite, &t_tra->mark_ind);
		fscanf(s_file, "%lf\n", &t_tra->value);
		if (ISNEIS_P(t_tra))   /* a nonexponential deterministic server */
			t_tra->det_no = no_neis_trans++;
		else if (t_tra->type==DET) /* a deterministic single server */
				t_tra->det_no = no_det_trans++;
		t_tra->i_matrix = read_sparse(s_file);
		t_tra->h_matrix = read_sparse(s_file);
		t_tra->c_matrix = read_sparse(s_file);
	}
	fscanf(s_file, "%d\n", &max);		/* disconnected places */
	timed.no_places = max;
	timed.places = (place *)emalloc(max * sizeof(place));
	for(num = 0; num < max; num++)
	{	pla = &(timed.places[num]);
		fscanf(s_file, "%s %d ", pla->name, &pla->number);
		fscanf(s_file, "%d %d ", &pla->m0, &pla->bound);
		fscanf(s_file, "%d", &pla->bits);
	}
	fscanf(s_file, "%d", &timed.memory);
	/* read immediate transition subnets */
	fscanf(s_file, "%d", &no_subnets);
	subnets = (subnet *)emalloc(no_subnets * sizeof(subnet));
	for(num = 0; num < no_subnets; num++)
	{	sub = &(subnets[num]);
		sub->end_mind = NULL;
		sub->no_markings = 0;
		fscanf(s_file, "%d", &max);
		sub->no_trans = max;
		sub->transitions = (imm_t *)emalloc(max * sizeof(imm_t));
		for(i = 0; i<sub->no_trans; i++)
		{	i_tra = &(sub->transitions[i]);
			fscanf(s_file, "%s %lf", i_tra->name, &i_tra->prob);
         fscanf(s_file, "%d %d", &i_tra->wgt_id, &i_tra->ena_id);
			i_tra->i_matrix = read_sparse(s_file);
			i_tra->h_matrix = read_sparse(s_file);
			i_tra->c_matrix = read_sparse(s_file);
			i_tra->enas = read_set(s_file);
			i_tra->diss = read_set(s_file);
		}
		fscanf(s_file, "%d", &max);
		sub->no_places = max;
		sub->places = (place *)emalloc(max * sizeof(place));
		for(i = 0; i < max; i++)
		{	pla = &(sub->places[i]);
			fscanf(s_file, "%s %d ", pla->name, &pla->number);
			fscanf(s_file, "%d %d ", &pla->m0, &pla->bound);
			fscanf(s_file, "%d", &pla->bits);
		}
		fscanf(s_file, "%d", &sub->memory);
		fscanf(s_file, "%d", &sub->no_ecs);
		sub->ecs = (byte **)emalloc(sub->no_ecs * sizeof(byte *));
		for(i = 0; i < sub->no_ecs; i++)
			sub->ecs[i] = read_set(s_file);
	}
}
