#ifndef __types_h__
#define __types_h__
#include <sys/param.h>
#include	<values.h> /* MAXINT */
#include <malloc.h>
#include <stdio.h>

/* Name of the environment file in home-dir: */
#define ENVIRONMENT	"DSPNexpress/Environment"
#define ARRAY_LIMIT	10		         /* limit for bounds to fit in an array */
#define SPARSE_START	1		      /* initial range for sparse array */
#define SPARSE_NEXT(a)	(2 * a)		/* expanding function for sparse array */
#define SPARSE_MAX    65535
#undef TRUE
#undef FALSE
#define FALSE   0
#define TRUE !FALSE

typedef unsigned char	byte;		/* 1 byte */
typedef unsigned short	word;		/* 2 bytes */
typedef byte            neis_cnt; /* eine NEIS ist max 255-mal aktiviert */
#define MAXNEIS_CNT     255

/* subnet reachability graph types */

typedef struct fire_path		/* immediate firing-path desciption */
{	double	prob;			/* probability of this firing path */
	word	*marking;		/* the last marking of the path */
}  fire_path;

typedef struct sub_node			/* immediate reachability set's tree item */
{	word	bound;			/* array: = 0; array[0..bound(place)] */
					/* sparse: > 0; array[0..bound] */
	word	number;			/* actual number of elements in sparse array */
	struct sub_node **next;		/* array of pointers to next level of tree*/
					/* (sparse or array) */
}  sub_node;

typedef struct sp_pointer		/* array element type for sparse arrays */
{	word	index;			/* index of element */
	sub_node *next;			/* next - level element in tree */
}  sp_pointer;

typedef struct sub_marking		/* (vanishing) submarking */
{	int	no_paths;		/* number of firing-paths to tangible markings */
	fire_path *paths;		/* pointer to the firing-paths (NULL = new marking) */
}  sub_marking;

/* tangible reachability graph types */

typedef struct tsp_pointer		/* array element type for sparse arrays */
{	word	*index;			/* index of element (pointer to coded marking) */
	void	*next;			/* next - level element in tree */
}  tsp_pointer;

typedef struct tang_node		/* tangible reachability set's tree item */
{	word	bound;			/* actual number of allocated elements */
	word	number;			/* actual number of elements in sparse array */
	tsp_pointer *next;		/* array of tsp_pointer to next level of tree*/
}  tang_node;

typedef	struct tang_path		/* tangible-to-tangible firing path */
{	double  prob;			/* path probability */
	double  delay;
	struct tang_marking *marking;		/* reached marking */
	byte    tran_no;
}  tang_path;

typedef struct tang_marking		/* tangible marking */
{	int	number;			/* the marking's number */
	word	*disc;			/* disconnected places submarking */
	word	**mark;			/* submarking of each subnet */
	int	no_time_paths;		/* number of time. paths */
	tang_path *time_paths;	/*markings reachable after firing of an time trans*/
/*
	byte  *fire_nums;
	byte  fire_tra_num;
*/
	/* enabling mutiplicity for nonexponential infinite servers */
	neis_cnt *neis_mult;   /* list of lenght 'no_neis_trans' */
	struct tang_marking *next;	/* next list element */
}  tang_marking;

/* net - structure types */

typedef struct sparse_place		/* element type for sparse matrices */
{	int num;			               /* number of place in row */
	int value;			            /* value */
	int plu_id, min_id;           /* id's for marking depentdent arcs */
}  sp_place;

typedef struct place			/* type for places */
{	char	name[30];		/* place name */
	int	number;			/* global number of the place */
	int	m0;			/* initial marking */
	int	bound;			/* upper bound for number of tokens in place */
	int	bits;			/* number of bits needed for this place */
}  place;

typedef enum server_type {
	SS=0, /* single   server race enabling   */
	IS=1, /* infinite server race enabling   */
	RS=2, /* single   server race resampling */
	RE=3, /* single   server race enabling   */
	RA=4  /* single   server race age        */
} server_type;

typedef enum transition_type {
	EXP=0, /* exponentielle    Transition */
	DET=1, /* deterministische Transition */
	GEN=2  /* general          Transition */
} transition_type;

#define ISNEIS_P(timed_t) (timed_t->type!=EXP && timed_t->infinite==IS)
/* determine if a timed transition is an nonexponential infinite server */

typedef struct timed_t          /* description of a timed transition */
{	char	name[30];              /* name */
	int	number;                /* global identification number */
	transition_type	type;      /* deterministic(1) / exponential(0) */
                                /* general-type(2)                   */
#ifndef OLDGENRRG
	int	det_no;                /* read_struct(73) number of det. transition */
#endif
	server_type	infinite;        /* infinite server */
	int	mark_ind;              /* marking-dependent firing rate: index */
	double	value;              /* det. or exp. delay */
	sp_place *i_matrix;          /* input matrix of the transition */
	sp_place *h_matrix;          /* inhibitor matrix of the transition */
	sp_place *c_matrix;          /* o - i matrix of the transition */
                                /* place-numbers in the matrices are global */
}  timed_t;

typedef struct timed_sub
{	int	no_trans;		/* number of timed transitions */
	timed_t	*transitions;		/* description of transitions */
	int	no_places;		/* number of disconnected places */
	place	*places;		/* description of places */
	int	memory;			/* memory space required for one marking */ 
	sub_node *tree;			/* disconnected places marking tree */
	int	no_markings;		/* number of disc. places markings */
}	t_sub;

typedef struct imm_t       /* description of an immediate transition */
{	char	name[30];         /* name */
	double	prob;          /* firing probability */
#ifndef OLDGENRRG
	int wgt_id;             /* id for marking dependent firing weight     */
	int ena_id;             /* id for marking dependent enabling function */
#endif
	sp_place *i_matrix;     /* input matrix of the transition     */
	sp_place *h_matrix;     /* inhibitor matrix of the transition */
	sp_place *c_matrix;     /* o - i matrix of the transition     */
                           /* place-numbers in the matrices are local */
	byte	*enas;            /* enabling set  */
	byte	*diss;            /* disabling set */
}  imm_t;

#ifndef OLDGENRRG
typedef struct end_mind      /* mind end marks of subnets */
{   struct end_mind *next;   /* linear list      */
	 word *mark;              /* coded submarking */
} end_mind;
#endif
	
typedef struct subnet   /* immediate transition subnet description */
{	int	no_trans;      /* number of transitions */
	imm_t	*transitions;  /* description of transitions */
	int	no_places;     /* number of places */
	place	*places;       /* description of places */
	int	memory;        /* memory space required for one marking (# words) */ 
	int	no_ecs;        /* number of ecs */
	byte	**ecs;         /* set of ecs */
	sub_node *tree;      /* subnet marking tree */
	int	no_markings;   /* number of submarkings */
#ifndef OLDGENRRG
	end_mind *end_mind;  /* list of end marks */
#endif
}	subnet;

#ifndef __main_c__
/*
 * Global Variables:
 */
extern char *progname;
extern FILE *ftmark;
extern FILE *frrg;
extern int no_det_trans;
extern int no_neis_trans;
extern int no_trans;
extern int no_places;
extern int no_subnets;
extern int no_absorb_states;
extern int no_tang_markings;
extern int no_edges;
extern t_sub          timed;
extern subnet        *subnets;
extern tang_node     *graph;
extern tang_marking  *new_markings;
#endif
#endif
