/*************************************************************************
*
*  Datei         : load_funcs.c
*
*  Aufgabe       : Hier liegen die einzelnen Funktionen, die jeweils genau
*                  ein keyword parsieren.
*                                                                          
*  Programmierer : Kolja Koischwitz
*
*  Datum         : 17. Mai 1994
*
**************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "netio.h"
#include "load_funcs.h"


/*
 * f_nettype: parse the NET_TYPE: keyword
 */
int f_nettype(net,buffer,num)
NetDescr *net;
char *buffer;
int num;
{
   char *b = (char *)skip_white(buffer);
   NetType t;

   if( !strcmp(b,"GSPN") )
      t = GSPN;
   else if( !strcmp(b,"DSPN") )
      t = DSPN;
   else if( !strcmp(b,"EDSPN") )
      t = EDSPN;
   else if( !strcmp(b,"CDSPN") )
      t = CDSPN;

   net->type = t;

   return 0;
}


/*
 * f_descr: parse the DESCRIPTION: keyword
 */
int f_descr(net,buffer,num)
NetDescr *net;
char *buffer;
int num;
{
   char *b = (char *)skip_white(buffer);

   net->comment = nStrdup(b);

   return 0;
}


/*
 * f_nplaces: parse the PLACES: keyword
 */
int f_nplaces(net,buffer,num)
NetDescr *net;
char *buffer;
int num;
{
   int i = atoi((char *)skip_white(buffer));

   if( i <= 0 )
      net->error_func("Invalid number of places", TRUE);

   net->num_places = i;
   net->place = nMalloc(i * sizeof(Place));

   return 0;
}


/*
 * f_ntrans: parse the TRANSITIONS: keyword
 */
int f_ntrans(net,buffer,num)
NetDescr *net;
char *buffer;
int num;
{
   int i = atoi((char *)skip_white(buffer));

   if( i <= 0 )
      net->error_func("Invalid number of transitions", TRUE);

   net->num_trans = i;
   net->trans = nMalloc(i * sizeof(Transition));

   return 0;
}


/*
 * f_ndelaypar: parse the DELAY_PARAMETERS: keyword
 */
int f_ndelaypar(net,buffer,num)
NetDescr *net;
char *buffer;
int num;
{
   int i = atoi((char *)skip_white(buffer));

   if( i < 0 )
      net->error_func("Invalid number of delay parameters", TRUE);

   net->num_delaypars = i;
   net->delaypar = nMalloc(i * sizeof(Delaypar));

   return 0;
}


/*
 * f_nmarkpar: parse the MARKING_PARAMETERS: keyword
 */
int f_nmarkpar(net,buffer,num)
NetDescr *net;
char *buffer;
int num;
{
   int i = atoi((char *)skip_white(buffer));

   if( i < 0 )
      net->error_func("Invalid number of marking parameters", TRUE);

   net->num_markpars = i;
   net->markpar = nMalloc(i * sizeof(Markpar));

   return 0;
}


/*
 * f_nrewards: parse the REWARD_MEASURES: keyword
 */
int f_nrewards(net,buffer,num)
NetDescr *net;
char *buffer;
int num;
{
   int i = atoi((char *)skip_white(buffer));

   if( i < 0 )
      net->error_func("Invalid number of reward measures", TRUE);

   net->num_rewards = i;
   net->reward = nMalloc(i * sizeof(Reward));

   return 0;
}


/*
 * f_markpar: parse the MARKPAR keyword
 */
int f_markpar(net,buffer,num)
NetDescr *net;
char *buffer;
int num;
{
   char *b = (char *)skip_white(buffer);
   char name[64];
   Markpar *m;
   int add;
   
   if( net->num_markpars <= 0 )
      net->error_func("No MARKING_PARAMETERS announced", TRUE);

   if( num == net->num_markpars )
      net->error_func("Markpar-table overflow", TRUE);
   else
      m = &net->markpar[num];

   if(sscanf(b,"%s %i %lg %lg",name,&m->value,&m->pos.x,&m->pos.y)!=4)
      net->error_func("Invalid MARKPAR definition", TRUE);
   m->name = nStrdup(name);
   
   return 0;
}


/*
 * f_place: parse the PLACE keyword
 */
int f_place(net,buffer,num)
NetDescr *net;
char *buffer;
int num;
{
   char *b = (char *)skip_white(buffer);
   char name[64], marking[64];
   Place *p;

   if( net->num_places <= 0 )
      net->error_func("No PLACES announced",TRUE);

   if( num == net->num_places )
      net->error_func("Place-table overflow", TRUE);
   else
      p = &net->place[num];

   sscanf(b,"%s %s %lf %lf %lf %lf", name,marking,
      &p->pos.x,&p->pos.y,
      &p->name_pos.x,&p->name_pos.y);

   p->name = nStrdup(name);

   if( isalpha(marking[0]) )
   {
      /* check if we have read marking parameters */
      if( net->num_markpars <= 0 )
         net->error_func("f_place: no marking parameters defined", TRUE);

      /* find marking parameter */
      p->is_value = FALSE;
      p->marking.markpar = find_markpar(net,marking);
   }
   else
   {
      p->is_value = TRUE;
      p->marking.value = atoi(marking);
   }

   return 0;
}


/*
 * f_delaypar: parse the DELAYPAR keyword
 */
int f_delaypar(net,buffer,num)
NetDescr *net;
char *buffer;
int num;
{
   char *b = (char *)skip_white(buffer);
   char name[64],delay[64];
   int add;
   Delaypar *d;

   if( net->num_delaypars <= 0 )
      net->error_func("No DELAY_PARAMETERS announced", TRUE);

   if( num == net->num_delaypars )
      net->error_func("Delaypar-table overflow", TRUE);
   else
      d = &net->delaypar[num];

   if(sscanf(b,"%s %s %lg %lg",name,delay,&d->pos.x,&d->pos.y)!=4)
      net->error_func("Invalid DELAYPAR definition", TRUE);

   d->name = nStrdup(name);
   d->is_value = (strcmp(delay,"<MD>") != 0);
   if(d->is_value)
     sscanf(delay,"%lg",&d->delay.value);
   
   return 0;
}


/*
 * t_get_arcs: parses the INPARCS,OUTPARCS,INHARCS keywords
 */
char *t_get_arcs(net, b, arc, num)
NetDescr *net;
char *b;
Arc *arc;
int num;
{
   int i,add;
   
   for(i=0;i<num;i++)
   {
      int j;
      char multipl[64],place[64];
      Arc *a=&arc[i];
   
      if(sscanf(b,"%s %s %i%n",multipl,place,&a->num_pts,&add)!=3)
         net->error_func("Invalid ARC definition", TRUE);
      b+=add;

      a->place = find_place(net,place);
      if(a->is_value = (strcmp(multipl,"<MD>") != 0))
         sscanf(multipl,"%i",&a->multipl.value);
		
      a->point = nMalloc(a->num_pts*sizeof(Coord));
      for(j=0;j<a->num_pts;j++)
      {
         if(sscanf(b,"%lg %lg%n",&a->point[j].x,&a->point[j].y,&add)!=2)
            net->error_func("Invalid points definition", TRUE);
         b+=add;
      }
   } 

  return b;   
}


/*
 * f_trans: parse the TRANSITION keyword
 * (Places must be read in) 
 */
int f_trans(net,buffer,num)
NetDescr *net;
char *buffer;
int num;
{
   char *b = (char *)skip_white(buffer);
   char keyword[64],name[64],delay[64],enadep[64],type[64],fpol[64];
   int add,phase;
   float weight;
   Transition *t;
   
   if( net->num_trans <= 0 )
      net->error_func("No TRANSITIONS announced", TRUE);

   if( num == net->num_trans )
      net->error_func("Transition-table overflow", TRUE);
   else
      t = &net->trans[num];

   if(sscanf(b,"%s %s %s %s %s %i %i %i %i %lg %lg %lg %lg %lg %lg %lg%n",
	     name,delay,enadep,type,fpol,&t->priority,&t->orient,
	     /*new*/
	     &phase,&t->group,&t->group_weight,
	     &t->pos.x,&t->pos.y, &t->name_pos.x,&t->name_pos.y,
	     &t->delay_pos.x,&t->delay_pos.y, &add)!=16)
   {
     b = skip_white(buffer);
     if(sscanf(b,"%s %s %s %s %s %i %i %lg %lg %lg %lg %lg %lg%n",
	       name,delay,enadep,type,fpol,&t->priority,&t->orient,
	       &t->pos.x,&t->pos.y, &t->name_pos.x,&t->name_pos.y,
	       &t->delay_pos.x,&t->delay_pos.y, &add)!=13)  
       net->error_func("Invalid TRANSITION definition", TRUE);
     else 
     {
       t->phases = DEFAULT_PHASE;
       t->group = DEFAULT_GRP;
       t->group_weight = (double)DEFAULT_GRP_WEIGHT;
     }
   }
   else t->phases = (word)phase;
   
   b+=add;
 
   t->md_enable = NULL; /* If there's no enabling function */
   t->name = nStrdup(name);
   t->inf_server = strcmp(enadep,"IS") == 0;
   t->f_policy = strcmp(fpol,"RA") == 0 ? RA : strcmp(fpol,"RS") == 0 ? RS : RE;
   t->type = strcmp(type,"EXP") == 0 ? EXP :
             strcmp(type,"DET") == 0 ? DET :
             strcmp(type,"GEN") == 0 ? GEN : IMM;

   t->delay_type = t->type == GEN ? PMFDEF :
                   isdigit(delay[0]) ? DELAY :
                   strcmp(delay,"<MD>") == 0 ? MARKDEP : DELAYPAR;
   switch( t->delay_type )
   {
      case DELAY:
         if(sscanf(delay,"%lg",&t->delay.value)!=1)
            net->error_func("Invalid transition delay", TRUE);
	 break;
      case DELAYPAR:
         /* check if we have delay parameters */
         if( net->num_delaypars <= 0 )
            net->error_func("f_trans: no delay parameters defined", TRUE);

         t->delay.delay_par=find_delaypar(net,delay);
         break;
      case MARKDEP:
      case PMFDEF:
         t->delay.md_delay = NULL;
         break;
   }
   
   /* read the inarcs */
   if(sscanf(b,"%s %i%n",keyword,&t->num_inarcs,&add)!=2)
     net->error_func("Invalid INPARC definition", TRUE);
   b+=add;

   if(strcmp(keyword,"INPARCS")!=0)
     net->error_func("Invalid keyword", TRUE);
   t->in_arc = nMalloc(t->num_inarcs*sizeof(Arc));
   b=t_get_arcs(net,b,t->in_arc,t->num_inarcs);
   
   /* read the outarcs */
   if(sscanf(b,"%s %i%n",keyword,&t->num_outarcs,&add)!=2)
     net->error_func("Invalid OUTPARC definition", TRUE);
   b+=add;
   if(strcmp(keyword,"OUTPARCS")!=0)
     net->error_func("Invalid keyword", TRUE);
   t->out_arc = nMalloc(t->num_outarcs*sizeof(Arc));
   b=t_get_arcs(net,b,t->out_arc,t->num_outarcs);

   /* read the inharcs */
   if(sscanf(b,"%s %i%n",keyword,&t->num_inharcs,&add)!=2)
     net->error_func("Invalid INHARC definition", TRUE);
   b+=add;
   if(strcmp(keyword,"INHARCS")!=0)
      net->error_func("Invalid keyword", TRUE);
   t->inh_arc = nMalloc(t->num_inharcs*sizeof(Arc));
   b=t_get_arcs(net,b,t->inh_arc,t->num_inharcs);

   return 0;
}


/*
 * f_delaypar_def: parse the DELAYPAR_DEF keyword
 * (Die Delayparameter muessen dazu schon eingelesen sein!) 
 */
int f_delaypar_def(net,buffer,num)
NetDescr *net;
char *buffer;
int num;
{
   char *b = (char *)skip_white(buffer);
   char name[64],md_str[1024];

   if( net->num_delaypars <= 0 )
      net->error_func("f_delaypar: no delay parameters defined", TRUE);

   if(sscanf(b,"%s %[^;]",name,md_str)!=2)
      net->error_func("Invalid DELAYPAR_DEF definition", TRUE);

   strcat(md_str,";");
   find_delaypar(net,name)->delay.md_delay = nStrdup(md_str);

   return 0;
}


/*
 * f_delay: parse the EXP_DELAY/DET_DELAY/GEN_DELAY/WEIGHT keywords
 * (Die Transitionen muessen dazu schon eingelesen sein!) 
 */
int f_delay(net,buffer,num)
NetDescr *net;
char *buffer;
int num;
{
   char *b = (char *)skip_white(buffer);
   char name[64],md_str[1024];
   Transition *t;

   if( net->num_trans <= 0 )
      net->error_func("f_delay: no transitions defined", TRUE);

   if(sscanf(b,"%s %[^;]",name,md_str)!=2)
      net->error_func("Invalid DELAY definition", TRUE);

   strcat(md_str,";");

   t = find_trans(net,name);

   if( t->delay_type == DELAY || t->delay_type == DELAYPAR )
   {
      char buf[128];

      sprintf(buf,"Redefining transition '%s' by '%s'\n", name,md_str);
      net->error_func(buf, FALSE);
   }

   t->delay_type = t->type == GEN ? PMFDEF : MARKDEP;
   t->delay.md_delay = nStrdup(md_str);

   return 0;
}


/*
 * f_md_enable: parse the FUNCTION keyword
 * (Die Transitionen muessen dazu schon eingelesen sein!) 
 */
int f_md_enable(net,buffer,num)
NetDescr *net;
char *buffer;
int num;
{
   char *b = (char *)skip_white(buffer);
   char name[64],md_str[1024];

   if( net->num_trans <= 0 )
      net->error_func("f_md_enable: no transitions defined", TRUE);
   
   if(sscanf(b,"%s %[^;]",name,md_str)!=2)
      net->error_func("Invalid FUNCTION definition", TRUE);
   strcat(md_str,";");
   find_trans(net,name)->md_enable = nStrdup(md_str);

   return 0;
}


/*
 * f_md_arc_mult: parse the CARDINALITY keyword
 * (Die Transitionen muessen dazu schon eingelesen sein!) 
 */
int f_md_arc_mult(net,buffer,num)
NetDescr *net;
char *buffer;
int num;
{
   int n;
   char *b = (char *)skip_white(buffer);
   char arctype[64],trans[64],place[64],md_str[1024];
   Transition *t; Arc *a;
   
   if(sscanf(b,"%s %s %s %[^;]",arctype,trans,place,md_str)!=4)
     net->error_func("Invalid CARDINALITY definition", TRUE);
   strcat(md_str,";");
   t = find_trans(net,trans);
   if(strcmp(arctype,"INPARC") == 0)
   { 
     a = t->in_arc;
	 n = t->num_inarcs;
   }
   else if(strcmp(arctype,"OUTPARC") == 0)
   { 
     a = t->out_arc;
	 n = t->num_outarcs;
   }
   else
   { 
     a = t->inh_arc;
	 n = t->num_inharcs;
   }
   
   while(n--)
     if(strcmp(place,a[n].place->name) == 0)
	 {
	   a[n].is_value = FALSE;
	   a[n].multipl.md_mult = nStrdup(md_str);
	   break;
	 }
	 
   return 0;
}


/*
 * f_reward: parse the MEASURE: keyword
 */
int f_reward(net,buffer,num)
NetDescr *net;
char *buffer;
int num;
{
   char *b = (char *)skip_white(buffer);
   Reward *r;
   char name[64],def[1024];

   if( net->num_rewards <= 0 )
      net->error_func("No REWARD_MEASURES announced", TRUE);

   if( num == net->num_rewards )
      net->error_func("Rewards-table overflow", TRUE);
   else
      r = &net->reward[num];

   if(sscanf(b,"%s %[^;]",name,def)!=2)
      net->error_func("Invalid MEASURE definition", TRUE);
   
   strcat(def,";");
   r->name = nStrdup(name);
   r->def = nStrdup(def);

   return 0;
}



