%{ 
/* ****************************************************************************
*        								      *
*  Programm         :   pars_TN.y (YACC-Quellprogramm)			      *
*                                                                             *
*  Parameter        :   Name des PN-Models
*                                                                             *
*  Rueckgabewert    :   0 --> Programm wurde erfolgreich beendet              *
*                       1 --> Fehler waehrend der Programmausfuehrung         *
*                                                                             *
*  Aufgabe          :   Das Programm verifiziert die Definitionen der         *
*                       markenabhaengigen Schaltraten, Schaltbarkeits-	      *
*			praedikate, Gewichte zeitloser Transitionen	      *
*			und Kantengewichte auf ihre syntaktische	      *
*			Korrektheit und generiert entsprechende       	      *
*                       Programmtexte in der Sprache C.          	      *
*                       Die zugehoerige lexikalische Analyse wird vom         *
*                       LEX-Quellprogramm scan_TN.x vorgenommen.    	      *
*									      *
*  Programmierer    :	Enrik Baumann					      *
*  Datum            :	Juni 1991					      *
*									      *
*  Modifiziert      :	Christian Luehe					      *
*  Datum            :	August 1992					      *
*									      *
*  Modifiziert      :	Ulrich Hertlein					      *
*  Datum            :	Oktober 1992					      *
*									      *
*    Variable Speicher-Alloziierung fuer 'buffer' durch die Hilfsfunktionen   *
*    'vlstrcpy' und 'vlstrcat' (siehe vlstring.c)			      *
*  									      *
*  Modifiziert	    :	Armin Zimmermann				      *
*  Datum            :	Februar 1993					      *
*			neue Syntax von <model>.DEF und <model>.NET	      *
*									      *
*  Modifiziert      :	Christian Luehe					      *
*  Datum            :	August 1993					      *
*    Parameter '-n' :: .DEF von stdin, .NET aus <model>.NET,kein output ausser*
*                      Errors; => syntax-Check  			      *
*									      *
*  Modifiziert      :   Christian Kelling                                     *
*  Datum            :   Feb 1994                                              *
*                       Benutzung im Simulationsmodul von DSPNexpress         *
*									      *
*  Modifiziert      :   Christian Kelling                                     *
*  Datum            :   Mai 94						      *
*                       neues Dateiformat fuer TimeNET			      *
*									      *
*  Modifiziert      :   J. Schneider                                          *
*  Datum            :   Juli 94						      *
*                       -s Schalter fuer Simulation   			      *
*                       -n Schalter fuer SyntaxCheck  			      *
*                       Aufruf ohne Schalter fuer Loesungsprozesse	      *
*									      *
*  Modifiziert      :   R. Zijal                                              *
*  Datum            :   August 96 (s. "Zijal" - comments)	      	      *
*                       Fetter bug von "pmf_fp-Einbauer"  entfernt            *
*                       Beim Aufruf mit der SyntaxOnly-Option wurde           *
*                       illegaler Weise auf ungeoeffneten pmf_fp zugegriffen  *
*                       (lesen, schliessen). Funktionierte trotzdem mit cc    *
*                       aber nicht mehr mit gcc. Effekt: neue measures konnten*
*                       nicht mehr definiert werden, da sich proc_TN aufhing. *
*                                                                             *
*  Modifiziert      :   Felix Kuehling                                        *
*  Datum            :   Maerz 2001                                            *
*                       Umstellung auf Flex und Bison.                        *
******************************************************************************/

%}
%{
#include <memory.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>

#include "portab.h"
#include "netio.h"
#include "expoly.h"
#include "pars_TN.h"
 
extern	char	yytext[];
extern	FILE	*yyin;
extern	int	fflineno;
char	*buffer = NULL;
char	*res_proc = NULL;
char	*temp_f_name = NULL;
char	error_text[1024];		/* error information */
int 	no_places = 0, no_trans = 0, no_mpar = 0, no_del_par = 0, no_helps = 0;
int	no_def_param = 0, no_md_exp = 0, no_md_det = 0, no_weights = 0, 
        no_ena_fcts = 0, no_md_arcs = 1, md_arc_nr = 0, no_gen_trans = 0;
char    def_param1 = 1;
int	*now_deter, deter_on, reward_on, pmf_on;
place	*places;
name_nr		*m_par;
name_delay	*d_par;
transition	*transitions;
md_arc		*md_arcs;
ena_fct		*ena_fcts;
weight		*weights;
int     RateParameters_checked = 0; /* SyntaxOnly and RateParameters 
				     *   => check them at the end    */

time_l *zaehler;
time_l  *zwerg;  /* eingefuehrt, da bei yacc-internen variablen $x nicht direkt
                    auf structure-komponenten zugegriffen werden kann */
power_m *zwerg2;  /* dito */


#ifdef __STDC__
int	checkParameters(name_delay *p);
int	recursive (char *name, char **open, name_delay *p);
char	*getNext(char *s);
void	my_handle(void);
int	yyparse(void);
void	yyerror(char *s);
#else
int	checkParameters();
int	recursive();
char	*getNext();
void	my_handle();
int	yyparse();
void	yyerror();
#endif

extern char *vlstrcpy();
extern char *vlstrcat();
extern char *strstr();
extern char *strdup();
extern int strncmp ();
extern char *getenv ();
extern int  getRidOf();

/* ----- */

FILE	*net, *mdf, *erc, *inf, *rdy;
char    mdfname[1024], ercname[1024], netname[1024],
	infname[1024];
char 	rdyname[1024], model_and_path[1024], modname[1024];
int	SyntaxOnly = 0;
int	Simulation = 0;


%}
%union
{	int	t_int;
	char	t_str[10000];
	double  dval;
	unsigned uval;
	struct time_l *tp;
	struct power_m *mp;
}

%token S_PARAM S_EXP S_DET S_WEIGHTS S_GEN
%token S_ENAFS S_ARC S_REWARD E_FILE
%token KW_PARAM KW_EXP KW_DET KW_GEN KW_FUNC KW_WEIGHT KW_CARD 
%token KW_MEAS KW_ARCTYPE 
%token COL SEM IF ELSE LNO LOR LAN OPA CPA EOPA ECPA COMP_OPER
%token MARKING TRA_NAME PLA_NAME INT_VALUE REAL_VALUE POS_INT POS_REAL NEG_REAL
%token PLUS MINUS MULTIPLY DIVIDE POWER
%token PROBABILITY EXPECTED CBR MEASURE_NAME PAR_NAME GEN_MDF_STR
%token R I X COM XPOT EPOT KW_PMF_DET KW_PMF_UNI KW_PMF_WEI KW_PMF_NOR
%token KW_PMF_FILE SAMP_FILE KW_FILE_SEQ KW_FILE_REV KW_FILE_UNI

%left S_DET
%left LOR
%left LAN
%left LNO
%left PLUS MINUS
%left MULTIPLY DIVIDE
%left POWER

%type <t_int>	 MARKING TRA_NAME PLA_NAME KW_ARCTYPE
%type <t_int>	 KW_FILE_SEQ KW_FILE_REV KW_FILE_UNI
%type <t_str>	 COMP_OPER INT_VALUE REAL_VALUE MEASURE_NAME PAR_NAME 
%type <t_str>    GEN_MDF_STR SAMP_FILE

%type <t_str>	 mark_func integer logic_cond comparison
%type <t_str>	 md_delay real expr if_exprs one_if_expr rew_item
%type <uval>	 POS_INT
%type <dval>	 POS_REAL NEG_REAL pmf_real
%type <tp>	 pmf_definition pmf_expression impulse rectangle
%type <mp> 	 expolynomial
%type <mp> 	 expoly_expr

%start def_file

%%

/*-----------------------------------------------------------------------*/
def_file	:	s_param s_exp s_det s_gen
			s_weights s_enafs 
			s_arc s_reward e_file
			| error
		{	exit(1);
		}
/*-----------------------------------------------------------------------*/
e_file          :       E_FILE
                {       
		if( ! SyntaxOnly || RateParameters_checked ) {
		  if (no_def_param) {
		    sprintf(error_text, 
			  "There are still undefined delay parameters (%d).\n",
			    no_def_param);
		    yyerror("");
		  }
		  
		  /* CHECK FOR CONSISTENCY */
		  if (checkParameters(d_par)) 
		    yyerror("First solve this problem : ");
		} /* if */
		}   
/*-----------------------------------------------------------------------*/
s_param		:
		| 	S_PARAM params
/*-----------------------------------------------------------------------*/
s_exp           :
                |       S_EXP md_exp_trans
/*-----------------------------------------------------------------------*/
s_det           :
                |       S_DET md_det_trans 
/*-----------------------------------------------------------------------*/
s_gen           :
                |       S_GEN pmf_def_trans
/*-----------------------------------------------------------------------*/
s_weights       :
                |       S_WEIGHTS md_weights 
/*-----------------------------------------------------------------------*/
s_enafs         :
                |       S_ENAFS md_enafs
/*-----------------------------------------------------------------------*/
s_arc           :
                |       S_ARC md_arcs 
/*-----------------------------------------------------------------------*/
s_reward        :
                |       S_REWARD rewards 
/*-----------------------------------------------------------------------*/
params		:
		| 	params param
/*-----------------------------------------------------------------------*/
param		:	KW_PARAM PAR_NAME expr SEM
                {
		    RateParameters_checked = 1;
		    if (no_def_param) { 
			int i,t,w, found = 0;

#ifdef debug
			fprintf(stderr, 
				"no_def_param = %d; no_del_par = %d\n",
				no_def_param, no_del_par);
#endif

			for (i = 0; i < no_del_par && !found; i++) 
			    if (!strcmp($2, d_par[i].name)) {
				if (d_par[i].type == Definition) {
				    found = 1;
				    d_par[i].values.def = strdup($3);

				    /* ------------------------------------
				     * search the parameter in transitions; 
				     * set there the deter value
				     */
				    for (t=0; t<no_trans; t++)
					if (transitions[t].parameter == i &&
					    transitions[t].type == IMM)
					{
					    for (w=0; w<no_weights; w++)
						if (weights[w].trans == t)
						{
						    weights[w].deter = 
							now_deter;
						    break;
						}
					    /* found ? */
					/*    if (w>=no_weights)
						yyerror ("Internal Error.");*/
					}
				    
				    /* make a new! */
				    now_deter = (int *)emalloc(no_places 
							       * sizeof(place));

#ifdef debug
				    fprintf(stderr, 
					    "Definition for '%s' is '%s'\n",
					    d_par[i].name, 
					    d_par[i].values.def);
#endif
#undef debug			      
				    {
				      char  nl_buf[ 1024 ];
				      int   i, l;

				      /* Zeilenumbruch entfernen */
				      sprintf( nl_buf, "%s\t (%s)", $2, $3 );
/*
fprintf( stderr, "\n$2 : <%s>\n", $2 );
fprintf( stderr, "$3 : <%s>\n", $3 );
fprintf( stderr, "BUF: <%s>\n", nl_buf );
fflush( stderr );
*/
				      l = strlen( nl_buf );
				      for( i = 0; i < l; i++ ) {
					if( nl_buf[ i ] == '\n' ) {
					  nl_buf[ i ] = ' ';
					}
				      }

				      fprintf(mdf, "#define %s\n", nl_buf);
				      if( Simulation ) {
				          fprintf(rdy, "#define %s\n", nl_buf);
				      } /* if */
				    }
				    break;
				} else {
				    sprintf(error_text, "Found definition for parameter '%s', which is value-type.\n",
					 $2);
				    yyerror("");
				}
			    }
			
			if (!found) 
			{
			    sprintf(error_text,
			      "The parameter '%s' is not defined.\n", 
				    $2);
			    yyerror("");
			}

			no_def_param--;
		    }
		    else 
		    {
			yyerror("Too many delay/weight definitions.");
		    }
		} 
/*-----------------------------------------------------------------------*/
md_exp_trans	:
		|	md_exp_trans one_md_exp_trans
/*-----------------------------------------------------------------------*/	
one_md_exp_trans:	KW_EXP TRA_NAME md_delay
		{
			if (!transitions[$2].id)
			{	
			    sprintf(error_text,
				"Marking dependent delay definition for independent transition '%s'.\n",
				transitions[$2].name);
			    yyerror("");
			}
			if (transitions[$2].ok)
			{
			    sprintf(error_text,
				"Redefinition of marking dependency for transition '%s'.\n",
				transitions[$2].name);
			    yyerror("");
			}
			transitions[$2].ok = 1;
			if (transitions[$2].type != EXP)
			{	
			    sprintf(error_text, 
				    "Transition '%s' is not exponential.\n",
				    transitions[$2].name);
			    yyerror("");
			}
			transitions[$2].def = strdup($3);
		}
/*-----------------------------------------------------------------------*/
md_det_trans	:
		|	md_det_trans one_md_det_trans
/*-----------------------------------------------------------------------*/
one_md_det_trans:	KW_DET TRA_NAME md_delay
		{
			if (!transitions[$2].id)
			{
			    sprintf(error_text,
				"Marking dependent delay definition for independent transition '%s'.\n",
				transitions[$2].name);
			    yyerror("");
			}
			if (transitions[$2].ok)
			{	
			    sprintf(error_text,
				"Redefinition of marking dependency for transition '%s'.\n",
				transitions[$2].name);
			    yyerror("");
			}
			transitions[$2].ok = 1;
			if (transitions[$2].type != DET)
			{
			    sprintf(error_text,
				    "Transition '%s' is not deterministic.\n",
				    transitions[$2].name);
			    yyerror("");
			}
			transitions[$2].def = strdup($3);
		}
/*-----------------------------------------------------------------------*/
pmf_def_trans	:
		|	pmf_def_trans one_pmf_def_trans
/*-----------------------------------------------------------------------*/

one_pmf_def_trans:    KW_GEN TRA_NAME pmf_definition
		{
			if (!transitions[$2].id)
			{
			    sprintf(error_text,
				"Probability mass function for independent transition '%s' defined.\n",
				transitions[$2].name);
			    yyerror("");
			}
			if (transitions[$2].ok)
			{	
			    sprintf(error_text,
				"Redefinition of probability mass function for transition '%s'.\n",
				transitions[$2].name);
			    yyerror("");
			}
			transitions[$2].ok = 1;
			if (transitions[$2].type != GEN)
			{
			    sprintf(error_text,
				    "Transition '%s' is not general.\n",
				    transitions[$2].name);
			    yyerror("");
			}
			if( Simulation ) {
				transitions[$2].pmf_def = $3;
			} else {
				transitions[$2].def = (char*) $3;
			} /* if */
		}
		|  KW_GEN TRA_NAME GEN_MDF_STR
/*-----------------------------------------------------------------------*/
md_weights	:
		|	md_weights one_md_weight
/*-----------------------------------------------------------------------*/
one_md_weight	:	KW_WEIGHT TRA_NAME md_delay
		{	int	i;
/* MACHT DAS KW_WEIGHT anstatt KW_TRANS WAS ??? */
			if (!transitions[$2].id)
			{
			    sprintf(error_text,
				"Marking dependent weight definition for independent transition '%s'.\n",
				transitions[$2].name);
			    yyerror("");
			}
			if (transitions[$2].ok)
			{	
			    sprintf(error_text,
				"Redefinition of marking dependency for transition '%s'.\n",
				transitions[$2].name);
			    yyerror("");
			}
			transitions[$2].ok = 1;
			if (transitions[$2].type != IMM)
			{	
			    sprintf(error_text,
				    "Transition '%s' is not immediate.\n",
				    transitions[$2].name);
			    yyerror("");
			}
			transitions[$2].def = strdup($3);

			for (i = 0; i < no_weights; i++)
				if (weights[i].trans == $2)
					break;
			if (i >= no_weights)
				exit(fflineno);
			weights[i].deter = now_deter;
			now_deter = (int *)emalloc(no_places * sizeof(place));
		}
/*-----------------------------------------------------------------------*/
md_enafs	:	
		|	md_enafs one_md_enaf
/*-----------------------------------------------------------------------*/
one_md_enaf	:	KW_FUNC TRA_NAME logic_cond SEM
		{	int	i;
			for (i = 0; i < no_ena_fcts; i++)
			  if (ena_fcts[i].trans == $2)
			  {	
			      sprintf(error_text, 
				"Redefinition of marking dependent enabling function in '%s'",
				 netname);
			      yyerror("");
			  }
			if (transitions[$2].type != IMM)
			{
			    yyerror(
				"Marking dependent enabling functions are only allowed \nfor immediate transitions.\n");
			}
			ena_fcts[no_ena_fcts].trans = $2;
			ena_fcts[no_ena_fcts].id = no_ena_fcts + 1;
			ena_fcts[no_ena_fcts].deter = now_deter;
			now_deter = (int *)emalloc(no_places * sizeof(place));
			fprintf(mdf, "\t\tcase %d :fire = %s;\n", ena_fcts[no_ena_fcts].id, $3);
			fprintf(mdf, "\t\t\tbreak;\n");
			no_ena_fcts++;
		}
/*-----------------------------------------------------------------------*/
md_arcs		:
		|	md_arcs one_md_arc
/*-----------------------------------------------------------------------*/
one_md_arc	:	KW_CARD KW_ARCTYPE TRA_NAME PLA_NAME md_delay
		{	md_arc	*md_walk = md_arcs;
			while (md_walk != NULL)
			{	if ((md_walk->type == $2) && (md_walk->place == $4) &&
								(md_walk->trans == $3))
					break;
				md_walk = md_walk->next;
			}
			if (md_walk == NULL)
			{
			    sprintf(error_text, 
				"No such marking dependent arc multiplicity definition in '%s'.\n",
				netname);
			    yyerror("");
			}
			if (md_walk->deter != NULL)
			{	
			    sprintf(error_text, 
				"Redefinition of marking dependent arc multiplicity in '%s'.\n",
				 netname);
			    yyerror("");
			}
			md_walk->deter = now_deter;
			/* |||md_walk->id = */ ++md_arc_nr;
			now_deter = (int *)emalloc(no_places * sizeof(place));
			fprintf(mdf, "\t\tcase %d : { \n%s }\n", md_walk->id, $5);
			fprintf(mdf, "\t\t\tbreak;\n");

		}
/*-----------------------------------------------------------------------*/
rewards		:
		|	rewards one_reward
/*-----------------------------------------------------------------------*/
one_reward	:	KW_MEAS MEASURE_NAME expr SEM
		{	
		if( Simulation ) {
			vlstrcat(&res_proc, "\n\tfor (selector=0; selector<4; selector++)\n");
			vlstrcat(&res_proc, "\tswitch (selector) {\n");
			vlstrcat(&res_proc, "\t\tcase SEL_MEAN:\n\t\t\tx0=");
			vlstrcat(&res_proc, $3);
                        vlstrcat(&res_proc, ";\n\t\t\tbreak;\n");
                        vlstrcat(&res_proc, "\t\tcase SEL_UP:\n\t\t\tx1=");
                        vlstrcat(&res_proc, $3);
                        vlstrcat(&res_proc, ";\n\t\t\tbreak;\n");
                        vlstrcat(&res_proc, "\t\tcase SEL_LO:\n\t\t\tx2=");
                        vlstrcat(&res_proc, $3);
			vlstrcat(&res_proc, ";\n\t\t\tbreak;}\n");
                        vlstrcat(&res_proc, "\n\tfprintf(file, \"");
                        vlstrcat(&res_proc, "Simulation results of measure:\\n");
                        vlstrcat(&res_proc, $2);
                        vlstrcat(&res_proc, " = %e\\n-----------------\\n");
			vlstrcat(&res_proc,
				"\\tMean Value = %e\\n\\tConfidence Interval: [%e, %e]\\n\\n\", ");
                        vlstrcat(&res_proc,"\n\t\tx0, x0, ");
                        vlstrcat(&res_proc,"MIN(x1, x2), MAX(x1, x2));\n");
                        vlstrcat(&res_proc, "\n\tif (out_value(0,0,true) > 0.0)");
                        vlstrcat(&res_proc, "\n\t\tfprintf(file, \"");
                        vlstrcat(&res_proc, "---WARNING---\\nThe measure ");
                        vlstrcat(&res_proc, $2);
                        vlstrcat(&res_proc, " has not reached the predefined accuracy.\\n\\n\\n\");");
		} else {
			vlstrcat(&res_proc, "\tvalue = ");
			vlstrcat(&res_proc, $3);
			vlstrcat(&res_proc, ";\n\tfprintf(file, \"");
			vlstrcat(&res_proc, $2);
			vlstrcat(&res_proc, " = \%.*lf\\n\", digits, value);\n");
		} /* if */
		}
/*-----------------------------------------------------------------------*/
md_delay	:	one_if_expr if_exprs ELSE expr SEM
		{	sprintf($$, "%s%s\t\trate = %s; break;\n", $1, $2, $4); }
		|	expr SEM
		{	sprintf($$, "\t\trate = %s; break;\n", $1); }
/*-----------------------------------------------------------------------*/
if_exprs	:
		{	strcpy($$, ""); }
		|	if_exprs one_if_expr
		{	strcat($$, $2); }
/*-----------------------------------------------------------------------*/
one_if_expr	:	IF logic_cond COL expr
		{	sprintf($$, "\t\tif (%s) { rate = %s; break; }\n", $2, $4); }
/*-----------------------------------------------------------------------*/
expr		:	real
		{	strcpy($$, $1);	}
		|	MINUS expr
		{	sprintf($$, "-(%s)", $2); }
		|	OPA expr CPA
		{	sprintf($$, "(%s)", $2); }
		|	expr DIVIDE expr
		{	sprintf($$, "%s/%s", $1, $3); }
		|	expr MULTIPLY expr
		{	sprintf($$, "%s*%s", $1, $3); }
		|	expr MINUS expr
		{	sprintf($$, "%s-%s", $1, $3); }
		|	expr PLUS expr
		{	sprintf($$, "%s+%s", $1, $3); }
		|	expr POWER expr
		{	sprintf($$, "my_pow(%s,%s)", $1, $3); }
/*-----------------------------------------------------------------------*/
real		:	REAL_VALUE
		{	strcpy($$, $1);	}
		|	PAR_NAME
		{	strcpy($$, $1);	}
		|	integer
		{	sprintf($$, "(double)(%s)", $1); }
		|	rew_item
		{	if (!reward_on)
			{	
			    sprintf(error_text, 
				"Reward measure items are only allowed for reward definitions in '%s'\n.",
				netname);
			    yyerror("");
			}
			strcpy($$, $1);
		}
/*-----------------------------------------------------------------------*/
pmf_definition :  KW_PMF_DET OPA POS_REAL CPA SEM
		{
			$$ = set_det_expoly($3);
		}
		| KW_PMF_UNI OPA POS_REAL COM POS_REAL CPA SEM
                {
		  if($5<=$3) {
		    fprintf(stderr, "ERROR(logic): in intervall-limits.\n");
		    exit(1);
		  } else
		    $$ = set_uniform_expoly($3,$5);
                }
                | KW_PMF_WEI OPA POS_REAL COM POS_REAL COM POS_REAL COM POS_REAL CPA SEM
                {
			if($5<=$3) {
			    fprintf(stderr, "ERROR(logic): in intervall-limits.\n");  
			    exit(1);
			}
			else {
			    printf("Warning: Weibull Distribution only implemented in diskrete time ...\nUniform Distribution will be applied in continuos time!\n");
                    	    $$ = set_uniform_expoly($3,$5);
			}
		}
		| KW_PMF_NOR OPA POS_REAL COM POS_REAL COM POS_REAL COM POS_REAL CPA SEM
                {
			if($5<=$3) {
			    fprintf(stderr, "ERROR(logic): in intervall-limits.\n");  
			    exit(1);
			} else {
			  printf("Warning: Normal Distribution only implemented in diskrete time ...\nUniform Distribution will be applied in continuos time!\n");
                    	    $$ = set_uniform_expoly($3,$5);
			}
		}
		| KW_PMF_FILE OPA SAMP_FILE COL KW_FILE_SEQ CPA SEM
                {
		  if (!strncmp(strdup($3),"$",1)){
		    vlstrcat (&temp_f_name, getenv("MODELDIR"));
		    vlstrcat (&temp_f_name, strstr($3, "/"));
		    $$ = define_file_access(temp_f_name, 0);
		    temp_f_name = NULL;
		    } else
		  $$ = define_file_access(strdup($3),0);
		}
		| KW_PMF_FILE OPA SAMP_FILE COL KW_FILE_REV CPA SEM
                {
		  if (!strncmp(strdup($3),"$",1)){
		    vlstrcat (&temp_f_name, getenv("MODELDIR"));
		    vlstrcat (&temp_f_name, strstr($3, "/"));
		    $$ = define_file_access(temp_f_name, 1);
		    temp_f_name = NULL;
		    } else
		  $$ = define_file_access(strdup($3),1);
		}		
                | KW_PMF_FILE OPA SAMP_FILE COL KW_FILE_UNI CPA SEM
                {
		  if (!strncmp(strdup($3),"$",1)){
		    vlstrcat (&temp_f_name, getenv("MODELDIR"));
		    vlstrcat (&temp_f_name, strstr($3, "/"));
		    $$ = define_file_access(temp_f_name, 2);
		    temp_f_name = NULL;
		    } else
		  $$ = define_file_access(strdup($3),2);
		}
		| pmf_expression SEM
		{
			$$ = $1;
		}
/*-----------------------------------------------------------------------*/
pmf_expression:   pmf_expression PLUS pmf_expression
		{
			$$ = melt_intervall($1,$3);
		}
		| pmf_expression MINUS pmf_expression
	     	{
			/* if any time-stru: MINUS a*I[x] dann error */
			zaehler = $3;
			do{
			  if(zaehler->p>0)
			    {
			      fprintf (stderr, "ERROR(math.): impulse with negative pointmass detected.\n");
			      exit(1);
			    }
			  zaehler = zaehler->next;
			}while(zaehler!=NULL);
			zwerg = $3;
			negate_all_c( zwerg->next->power );
			melt_intervall($1, zwerg);
		}
		| impulse
		{
			$$=$1;
		}
		| rectangle
		{
			$$=$1;
		}
/*-----------------------------------------------------------------------*/
impulse:        I EOPA POS_REAL ECPA  
                {
			if($3==0)
		      {                      /* Intervall-anfang bei Null ! */
			$$=make_l_stru(
			     0.0, (double) UNENDLICH, 1.0, zero_tail(), NULL); 
		      }
		    else
		      {
			$$=make_l_stru(0.0, $3, 0.0, zero_tail(), 
		  make_l_stru($3, (double) UNENDLICH, 1.0, zero_tail(), NULL));
		      }
		}
                | POS_REAL I EOPA POS_REAL ECPA
                {
			if($4==0)
			{                    /* Intervall-anfang bei Null ! */
			  $$=make_l_stru(
			       0.0, (double)UNENDLICH, $1, zero_tail(), NULL); 
			}
		      else
			{
			  $$=make_l_stru(0.0, $4, 0.0, zero_tail(), 
		   make_l_stru($4, (double)UNENDLICH, $1, zero_tail(), NULL)); 
			}
		}
/*-----------------------------------------------------------------------*/
rectangle:      R EOPA POS_REAL COM POS_REAL ECPA
                {
			if($5<=$3)
			 {
			   fprintf(stderr,
				   "ERROR(logic): in intervall-limits.\n");
			   exit(1);
			 }
		      else
			if($3==0)
			  {                  /* Intervall-anfang bei Null ! */
			    $$=make_l_stru(0.0, $5, 0.0, 
					   make_m_stru(0,      /* bei x^0 */
		  make_n_stru(0.0, (double)UNDEF, 1.0, NULL),  /* ist c=1 */
						       NULL),
		  make_l_stru($5, (double)UNENDLICH, 0.0, zero_tail(), NULL)); 
			  }
			else
			  {
			    $$=make_l_stru(0.0, $3, 0.0, zero_tail(), 
					   make_l_stru($3, $5, 0.0, 
					       make_m_stru(0, /* bei x^0 */
		 make_n_stru(0.0, (double)UNDEF, 1.0, NULL),  /* ist c=1 */
							   NULL),
		 make_l_stru($5, (double)UNENDLICH, 0.0, zero_tail(), NULL))); 
			  }
		}
                | expolynomial R EOPA POS_REAL COM POS_REAL ECPA
                {	 
			if($4==0)
		        {                    /* Intervall-anfang bei Null ! */
			  $$=make_l_stru(0.0, $6, 0.0, $1, 
		  make_l_stru($6, (double)UNENDLICH, 0.0, zero_tail(), NULL)); 
			}
		      else
			{
			  $$=make_l_stru(0.0, $4, 0.0, zero_tail(), 
		  make_l_stru($4, $6, 0.0, $1, make_l_stru(
			     $6, (double)UNENDLICH, 0.0, zero_tail(), NULL))); 
			}
		}
/*-----------------------------------------------------------------------*/
expolynomial:   expolynomial PLUS expolynomial
                {
			$$ = melt_power( $1,$3);
		}
                | expolynomial MINUS expolynomial
                {
			$$ = melt_power( $1, negate_all_c($3) );
		}
                | OPA expolynomial CPA
                {
			$$=$2;
		}
                | pmf_real                                   
                {
			$$=make_m_stru(0,
                                     make_n_stru(0.0, (double)UNDEF,$1,NULL),
                                     NULL);
		}
                | expoly_expr                            
                {
			$$=$1;
		}
                | pmf_real expoly_expr
                {
			zwerg2 = $2;
		       (zwerg2->expo)->c = $1;
                       $$ = $2;
		}
/*-----------------------------------------------------------------------*/
expoly_expr:    XPOT POS_INT                  
                {
			$$=make_m_stru($2,
                                       make_n_stru(0.0,(double)UNDEF,1.0,NULL),
                                          NULL);
		}
                | X
                {
			$$=make_m_stru(1,
                                     make_n_stru(0.0, (double)UNDEF,1.0,NULL),
                                     NULL);
		}
                | EPOT OPA NEG_REAL X CPA        
                {
			$$=make_m_stru(0,
                              make_n_stru($3*(-1.0), (double)UNDEF, 1.0, NULL),
				      NULL);
		}
                | XPOT POS_INT EPOT OPA NEG_REAL X CPA 
                {
			$$ = make_m_stru($2,
			       make_n_stru($5*(-1.0), (double)UNDEF,1.0, NULL),
					    NULL);
		}
                | X EPOT OPA NEG_REAL X CPA
                {
			$$ = make_m_stru(1,
			      make_n_stru($4*(-1.0), (double)UNDEF, 1.0, NULL),
					NULL);
		}
/*-----------------------------------------------------------------------*/
pmf_real:         POS_REAL
		{
			$$ = $1;
		}
                | NEG_REAL
		{
			$$ = $1;
		}
/*-----------------------------------------------------------------------*/
rew_item	:	PROBABILITY logic_cond CBR
		{	
		if( Simulation ) {
		    	fprintf(erc, "/*prob estimator*/\n");
                        fprintf(erc, "\tcase %d: \n", no_helps);
                        fprintf(erc, "\t\tif (%s)/*endif*/\n\t\t\t", $2);
                        fprintf(erc,"update_measure( %d ,ONE_TOK);\n", no_helps);
                        fprintf(erc,"\t\telse\n");
                        fprintf(erc,"\t\t\tupdate_measure( %d ,NO_TOK);\n", no_helps);
                        fprintf(erc, "\t\tbreak;\n");
                        sprintf($$, "out_value(selector, %d, false)",no_helps);
                        no_helps++;
		} else {
			fprintf(erc, "\tif (%s)\n\t\thelp[%d] += prob;\n", $2, no_helps);
			sprintf($$, "help[%d]", no_helps);
                        no_helps++;
		} /* if */
		}
		|	PROBABILITY logic_cond IF logic_cond CBR
		{	
		if( Simulation ) {
			fprintf(erc, "/*prob estimator*/\n");
                        fprintf(erc, "\tcase %d: \n", no_helps);
                        fprintf(erc, "\t\tif (%s){\n\t\t\t", $4);
                        fprintf(erc, "update_measure( %d ,ONE_TOK);\n", no_helps++);
                        fprintf(erc, "\t\t\tif (%s)/*endif*/\n\t\t\t\t", $2);
                        fprintf(erc, "update_measure( %d ,ONE_TOK);\n", no_helps);
                        fprintf(erc,"\t\t\telse\n");
                        fprintf(erc,"\t\t\t\tupdate_measure( %d ,NO_TOK);}\n", no_helps);
			fprintf(erc,"\t\telse {\n");
                        fprintf(erc,"\t\t\tupdate_measure( %d ,NO_TOK);\n", no_helps-1);
                        fprintf(erc,"\t\t\tupdate_measure( %d ,NO_TOK); }\n", no_helps);
                        fprintf(erc, "\t\tbreak;\n");
                        sprintf( $$, "out_value(selector, %d, false)/out_value(selector, %d, false)",
                          no_helps, no_helps-1);
			no_helps++;		    
		} else {
			fprintf(erc, "\tif (%s)\n\t{\thelp[%d] += prob;\n", $4, no_helps++);
			fprintf(erc, "\t\tif (%s)\n\t\t\thelp[%d] += prob;\n\t}\n", $2, no_helps);
			sprintf($$, "help[%d]/help[%d]", no_helps, no_helps-1);
			no_helps++;
		} /* if */
		}
		|	EXPECTED mark_func CBR
		{	
		if( Simulation ) {
			fprintf(erc, "/*mean estimator: %s */\n", $2);
                        fprintf(erc, "\tcase %d: \n\t\t", no_helps);
                        fprintf(erc, "update_measure( %d ,old_tok);\n", no_helps);
                        fprintf(erc, "\t\tbreak;\n");
                        sprintf($$, "out_value(selector, %d, false)", no_helps);
                        no_helps++;		    	
		} else {
		    	fprintf(erc, "\thelp[%d] += prob * (%s);\n", no_helps, $2);
		    	sprintf($$, "help[%d]", no_helps);
                        no_helps++;		    	
		} /* if */
		}
		|	EXPECTED mark_func IF logic_cond CBR
                {	
		if( Simulation ) {
			fprintf(erc, "/*mean estimator: %s */\n", $2);
                        fprintf(erc, "\tcase %d: \n", no_helps);
                        fprintf(erc, "\t\tif (%s)/*endif*/\n", $4);
                        fprintf(erc, "\t\t\tupdate_measure( %d ,old_tok);\n", no_helps);
                        fprintf(erc,"\t\telse\n");
                        fprintf(erc,"\t\t\tupdate_measure( %d ,NO_TOK);\n", no_helps);
                        fprintf(erc, "\t\tbreak;\n");
                        sprintf($$, "out_value(selector, %d, false)", no_helps);
                        no_helps++;		   
		} else {
		    	fprintf(erc, "\tif (%s)\n\t\t", $4);
		    	fprintf(erc, "help[%d] += prob * (%s);\n", no_helps, $2);
		    	sprintf($$, "help[%d]", no_helps);
                        no_helps++;		   
		} /* if */
		}
/*-----------------------------------------------------------------------*/
logic_cond	:	comparison
		{	strcpy($$, $1);	}
		|	LNO logic_cond
		{	sprintf($$, "!(%s)", $2); }
		|	OPA logic_cond CPA
		{	sprintf($$, "(%s)", $2); }
		|	logic_cond LAN logic_cond
		{	sprintf($$, "(%s)&&(%s)", $1, $3); }
		|	logic_cond LOR logic_cond	
		{	sprintf($$, "(%s)||(%s)", $1, $3); }
/*-----------------------------------------------------------------------*/
comparison	:	mark_func COMP_OPER mark_func
		{	sprintf($$, "((%s)%s(%s))", $1, $2, $3); }
/*-----------------------------------------------------------------------*/
/* if something changes here 
   you`ve got to have a look into "mdep_cardinality.c" 
*/
mark_func	:	integer
		{	strcpy($$, $1); }
		|	OPA mark_func CPA
		{	strcpy($$, "("); strcat($$, $2); strcat($$, ")"); }
		|	MINUS expr
		{	sprintf($$, "-(%s)", $2); }
		|	mark_func DIVIDE mark_func
		{	sprintf($$, "(int)(%s/%s)", $1, $3); }
		|	mark_func MULTIPLY mark_func
		{	sprintf($$, "%s*%s", $1, $3); }
		|	mark_func MINUS mark_func
		{	sprintf($$, "%s-%s", $1, $3); }
		|	mark_func PLUS mark_func
		{	sprintf($$, "%s+%s", $1, $3); }
		|	mark_func POWER mark_func
		{	sprintf($$, "(int)my_pow((double)%s,(double)%s)", $1, $3); }
/*-----------------------------------------------------------------------*/
/* if something changes here 
   you`ve got to have a look into "mdep_cardinality.c" 
*/
integer		:	MARKING
		{	
		if( Simulation ) {
			sprintf($$, "\n\t\t places[ %d ].token ", $1);
		} else {
			sprintf($$, "act_marking(%d)", $1);
		} /* if */
		}
		|	INT_VALUE
		{	strcpy($$, $1); }
/*-----------------------------------------------------------------------*/

%%

/*	emalloc tries to allocate required memory space and
	tests if it was successfully allocated; if not => exit */
char *
emalloc(bytes)
	int bytes;
{
	int *point;

	if (!bytes)
		return NULL;
	point = (int *)malloc((unsigned)(bytes));
	if (point == NULL)
	{	fprintf(stderr, "Not enough memory.\n");
		exit(-1);
	}
	return (char *)point;	
}



/*	my_handle fuer load_net-Fehlerhandle */
void
my_handle()
{
	/* tut nix */
  /* also braucht es bei load_net() auch nicht aufgerufen zu werden (R. Zijal) */
}



/*	read_net reads the net-description from net.TN */
void
read_net(netname)
	char *netname;
{	
  int	i_mpar, i_dpar, i_place, i_trans, no_arcs, i, j;
  NetDescr *description;
  md_arc	*md_help;

  char buf[256];
  FILE *pmf_fp;
  
  no_del_par = no_mpar = 0;
  no_places = no_trans = 0;
  i_mpar = i_dpar = i_place = 0;
  i_trans = -1; md_arcs = NULL;
  
  /*description = load_net( netname, SyntaxOnly ? my_handle : NULL );      (R. Zijal) */
  description = load_net( netname, NULL );

  /* for non-simulation, create .pmf file */
  if (!Simulation && !SyntaxOnly)
  {
     sprintf(buf,"%s/%s.pmf", getenv("MODELDIR"),netname);
     if (!(pmf_fp = fopen(buf,"w")))
     {
        fprintf(stderr,"Unable to write file '%s'\n", buf);
	exit(-1);
     }
  }
  
  no_places	= description->num_places;
  no_trans	= description->num_trans;
  no_del_par	= description->num_delaypars;
  no_mpar	= description->num_markpars;
  places	= (place *)	 emalloc(no_places	* sizeof(place));
  transitions	= (transition *) emalloc(no_trans	* sizeof(transition));
  weights	= (weight *)	 emalloc(no_trans	* sizeof(weight));
  ena_fcts	= (ena_fct *)	 emalloc(no_trans	* sizeof(ena_fct));
  d_par		= (name_delay *) emalloc(no_del_par	* sizeof(name_delay));
  m_par		= (name_nr *)	 emalloc(no_mpar	* sizeof(name_nr));
  memset( places,	0, (no_places	* sizeof(place)));
  memset( transitions,	0, (no_trans	* sizeof(transition)));
  memset( weights,	0, (no_trans	* sizeof(weight)));
  memset( ena_fcts,	0, (no_trans	* sizeof(ena_fct)));
  memset( d_par,	0, (no_del_par	* sizeof(name_delay)));
  memset( m_par,	0, (no_mpar	* sizeof(name_nr)));

  for (i_mpar=0; i_mpar<no_mpar; i_mpar++) {
      m_par[i_mpar].name = description->markpar[i_mpar].name;
      m_par[i_mpar].val = (double)description->markpar[i_mpar].value;
  } /* for */

  for (i_place=0; i_place<no_places; i_place++)
    places[i_place].name = description->place[i_place].name;

  for (i_dpar=0; i_dpar<no_del_par; i_dpar++) {
      d_par[i_dpar].name = strdup("DSPN_");
      vlstrcat(&d_par[i_dpar].name, description->delaypar[i_dpar].name);

      if (!(description->delaypar[i_dpar].is_value)) {
	  d_par[i_dpar].type = Definition;
	  no_def_param++;
      } else {
	  d_par[i_dpar].type = Value;
	  d_par[i_dpar].values.val = description->delaypar[i_dpar].delay.value;
      } /* if */
      if( !SyntaxOnly ) {
      	  if (description->delaypar[i_dpar].is_value) {
	      /* 
	       * write the "#define ..." into _MDF.c
	       * for use in definitions
	      */
	      fprintf(mdf,"#define %s\t (%f)\n",
		      d_par[i_dpar].name, d_par[i_dpar].values.val);
	      fprintf(erc,"#define %s\t (%f)\n",
		      d_par[i_dpar].name, d_par[i_dpar].values.val);
	      if( Simulation )
      		  fprintf(rdy,"#define %s\t (%f)\n",
		      d_par[i_dpar].name, d_par[i_dpar].values.val);
          } /* if */
      } /* if */
  } /* for */

  for (i_trans=0; i_trans<no_trans; i_trans++) {
      transitions[i_trans].number = i_trans;
      transitions[i_trans].name = description->trans[i_trans].name;
      transitions[i_trans].type = description->trans[i_trans].type;
      if ((description->trans[i_trans].delay_type == MARKDEP) ||
          (description->trans[i_trans].delay_type == PMFDEF)) {
	  transitions[i_trans].id = i_trans+1;
	  transitions[i_trans].ok = 0;
	  switch (transitions[i_trans].type) {  
	    case IMM: 
	      weights[no_weights].trans = i_trans;
	      weights[no_weights].deter = NULL;
	      no_weights++; 
	      break;
	    case EXP: 
	      no_md_exp++;
	      break;
	    case DET: 
	      no_md_det++;  
	      break;
	    case GEN: 
	      no_gen_trans++;
	      
              /* if (!Simulation)                 FETTER bug (R.Zijal) */
	      if (!Simulation && !SyntaxOnly)
	      {
	         fprintf(pmf_fp,"TRANSITION %s\n%s\n\n",
		 	description->trans[i_trans].name,
			description->trans[i_trans].delay.md_delay);
	      }
	      break;
	  } /* switch */
      } else
	if (description->trans[i_trans].delay_type == DELAYPAR) {
	  int walk;
	  char name[256];
	  
	  strcpy (name,"DSPN_");
	  strcat (name, 
		  description->trans[i_trans].delay.delay_par->name);
	  for (walk = 0; walk < no_del_par; walk++)
	    if (!strcmp(name, d_par[walk].name))
	      break;
	  if (walk >= no_del_par) {	
	      fprintf(stderr,
		      "Error: unknown delay parameter %s\n",
		      name);
	      exit(1);
	  } /* if */
	  transitions[i_trans].id = i_trans+1;
	  transitions[i_trans].def = strdup ("\t\trate = ");
	  vlstrcat(&transitions[i_trans].def, name);
	  vlstrcat(&transitions[i_trans].def,"; break;\n");
	  
	  /* mdep weight parameter */
	  if (transitions[i_trans].type == IMM &&
	      d_par[walk].type == Definition) {
	      weights[no_weights].trans = i_trans;
	      weights[no_weights].deter = NULL;
	      no_weights++;
	      transitions[i_trans].parameter = walk;
	  } /* if */
	  
	  transitions[i_trans].ok = 1;
	} else 
  	  transitions[i_trans].id = 0;
      
      no_arcs = description->trans[i_trans].num_inarcs;
      for (i=0; i<no_arcs; i++) {
	  for (j=0; j<no_places; j++)
	    if (!strcmp(description->trans[i_trans].in_arc[i].place->name, 
			places[j].name))
	      break;
	  if (j >= no_places) {
	      printf
		("ERROR: unknown place name in arc definition, place\n%s\n",
		 description->trans[i_trans].in_arc[i].place->name);
	      exit (-1);
	    }
          if (!(description->trans[i_trans].in_arc[i].is_value)) {       
	      md_help = (md_arc *)emalloc(sizeof(md_arc));
	      md_help->type = INPARC;
	      md_help->place = j;
	      md_help->trans = i_trans;
	      md_help->id = no_md_arcs;
	      md_help->deter = NULL;
	      md_help->next = md_arcs;
	      md_arcs = md_help;
	      no_md_arcs++;
	    }
	}
      no_arcs = description->trans[i_trans].num_outarcs;
      for (i=0; i<no_arcs; i++) {
	  for (j=0; j<no_places; j++)
	    if (!strcmp(description->trans[i_trans].out_arc[i].place->name, 
			places[j].name))
	      break;
	  if (j >= no_places) {
	      printf
		("ERROR: unknown place name in arc definition, place\n%s\n",
		 description->trans[i_trans].out_arc[i].place->name);
	      exit (-1);
	    }
	  if (!(description->trans[i_trans].out_arc[i].is_value)) {
	      md_help = (md_arc *)emalloc(sizeof(md_arc));
	      md_help->type = OUTPARC;
	      md_help->place = j;
	      md_help->trans = i_trans;
	      md_help->id = no_md_arcs;
	      md_help->deter = NULL;
	      md_help->next = md_arcs;
	      md_arcs = md_help;
	      no_md_arcs++;
	    }
	}

      no_arcs = description->trans[i_trans].num_inharcs;
      for (i=0; i<no_arcs; i++) {
	  for (j=0; j<no_places; j++)
	    if (!strcmp(description->trans[i_trans].inh_arc[i].place->name, 
			places[j].name))
	      break;
	  if (j >= no_places) {
	      printf
		("ERROR: unknown place name in arc definition, place\n%s\n",
		 description->trans[i_trans].inh_arc[i].place->name);
	      exit (-1);
	    }
	  if (!(description->trans[i_trans].inh_arc[i].is_value)) {
	      md_help = (md_arc *)emalloc(sizeof(md_arc));
	      md_help->type = INHARC;
	      md_help->place = j;
	      md_help->trans = i_trans;
	      md_help->id = no_md_arcs;
	      md_help->deter = NULL;
	      md_help->next = md_arcs;
	      md_arcs = md_help;
	      no_md_arcs++;
	    }
	}
    } /* for all trans */
    
    /* if (!Simulation)                 FETTER bug (R.Zijal) */
    if (!Simulation && !SyntaxOnly)
       fclose(pmf_fp);
}

/*	deter_out writes information about determining places to a file */
void
deter_out(f, p)
	FILE	*f;
	int	*p;
{
	int	i, no_deter = 0;

	for (i = 0; i < no_places; i++)
		if (p[i])
			no_deter++;
	fprintf(f, "CONNECTED %d\n", no_deter);
	for (i = 0; i < no_places; i++)
		if (p[i])
			fprintf(f, "%d\n", i);
	fprintf(f, "\n");
}

/* open files */
int main(argc, argv)
	int	argc;
	char	*argv[];
{
	int	status, i, lines = 1;
	md_arc	*a_walk;
	char buf[500];
	char *modeldir;


#ifdef	YYDEBUG
	/* set the -t option while executing yacc (Makefile) to enable debugging */
	extern	int	yydebug;

	yydebug = 1;		/* set debugging on */
#endif

	switch (argc) {
	  case 2:
	    strcpy( modname, argv[1] );
	    break;
	  case 3:
	    strcpy( modname, argv[2] );
	    if (!strcmp(argv[1], "-s")) {	  /* Option "-s" ? _N_o output */
		Simulation = 1;
		break;
	    } /* if */
	    if (!strcmp(argv[1], "-n")) {	  /* Option "-n" ? _N_o output */
		SyntaxOnly = 1;
		break;
	    } /* if */
	    /* sonst Error ... */
	  default:
	    fprintf(stderr, 
		    "%s Usage: [-<n|s>] netname\n\t\t-n : No output - syntax check\n\t\t-s : Simulation\n", 
		    argv[0]);
	    exit(-1);
	} /* switch */

	/* fetch modeldir */
	if( !(modeldir = getenv("MODELDIR")) ) {
	    fprintf(stderr,"'MODELDIR' not set\n");
	    return( NULL );
	} /* if */

	if (SyntaxOnly) {
	    sprintf(netname,"%s", modname);  /* tmp-File hat schon MODELDIR/ */
	    sprintf(infname, "/dev/null");
	    sprintf(ercname, "/dev/null");
	    sprintf(mdfname, "/dev/null");
	} else {
	    sprintf(netname,"%s/%s.TN",      modeldir,modname);
	    sprintf(infname,"%s/%s.DEFINFO", modeldir,modname);
	    sprintf(mdfname,"%s/%s_MDF.c",   modeldir,modname);
	    sprintf(ercname,"%s/%s_ERC.c",   modeldir,modname);
	    if( Simulation )
	        sprintf(rdyname,"%s/%s_RDY.c",   modeldir,modname);
	} /* if */
	
	if ((mdf = fopen(mdfname, "w")) == NULL) {	
	    fprintf(stderr, "Can't open file %s for writing\n", mdfname); 
	    exit(-1);
	} /* if */
	if ((inf = fopen(infname, "w")) == NULL) {	
	    fprintf(stderr, "Can't open file %s for writing\n", infname); 
	    exit(-1);
	} /* if */
	if ((erc = fopen(ercname, "w")) == NULL) {	
	    fprintf(stderr, "Can't open file %s for writing\n", ercname); 
	    exit(-1);
	} /* if */
	if ((net = fopen(netname, "r")) == NULL) {	
	    fprintf(stderr, "Can't open file %s for reading\n", netname); 
	    exit(-1);
	} /* if */
	if( Simulation )
	    if ((rdy = fopen(rdyname, "w")) == NULL) {	
	        fprintf(stderr, "Can't open file %s for writing\n", rdyname); 
	        exit(-1);
	    } /* if */

	if (!SyntaxOnly) {
		fprintf(mdf, "/*\tFile:    %s\n", mdfname);
		fprintf(mdf, "\tPurpose: procedures for marking dependent net parts */\n\n");
		fprintf(mdf, "#include <math.h>\n");
		if( Simulation ) {
			fprintf(mdf, "#include <stdio.h>\n");
			fprintf(mdf, "#include \"../Slave/sim_slave_types.h\"\n\n");
			fprintf(mdf, "extern place *places;\n");
			fprintf(mdf, "extern double rnd_uniform();\n");
			fprintf(mdf, "extern double get_delay_from_file();\n");
			fprintf(rdy, "#include <math.h>\n");
		} /* if */
	} /* if */
	
	read_net(modname);

	if (!SyntaxOnly) {
		if( Simulation ) {
			fprintf(mdf, "/* avoid 0^0 problem */\n");
			fprintf(mdf, "double my_pow(a,b)\n\tdouble a,b;\n");
			fprintf(mdf, "{\n\tif ((a==0.0) && (b==0.0))\n");
			fprintf(mdf, "\t\treturn (1.0);\n\treturn pow(a,b);\n}\n");
			fprintf(rdy, "/* avoid 0^0 problem */\n");
			fprintf(rdy, "double my_pow(a,b)\n\tdouble a,b;\n");
			fprintf(rdy, "{\n\tif ((a==0.0) && (b==0.0))\n");
			fprintf(rdy, "\t\treturn (1.0);\n\treturn pow(a,b);\n}\n");
		} else {
			fprintf(mdf, "\n/* this function returns the actual global marking */\n");
			fprintf(mdf, "extern int act_marking();\n\n");
			fprintf(mdf, "/* avoid 0^0 problem */\n");
			fprintf(mdf, "double my_pow(a,b)\n\tdouble a,b;\n");
			fprintf(mdf, "{\n\tif ((a==0.0) && (b==0.0))\n");
			fprintf(mdf, "\t\treturn (1.0);\n\treturn pow(a,b);\n}\n");
		} /* if */
	} /* if */

	/* main program */
	error_text[0] = '\0';
	now_deter = (int *)emalloc(no_places * sizeof(int));
	deter_on = reward_on = FALSE;
	no_helps = md_arc_nr = 0;

/*#ifdef linux Seit Bison verwendet wird tritt das Problem auch bei SunOS auf
  (Felix)*/
        /* if( SyntaxOnly ) raus, sonst geht nichts mehr (Renate) */

        if (yyin)
                fclose(yyin);
        if ((yyin = fopen(netname, "r")) == NULL) {
                fprintf(stderr, "Can't open file %s for reading\n", netname);
                   exit(-1);
	      }
/*#else
        rewind (net);
        freopen(netname,"r",yyin);*/   /* moegliche Fehlerquelle, da kein "re"open bei yyin (R. Zijal) , funktioniert unter Linux nicht (Renate) */
/*#endif*/

	while( !feof(yyin) )
	  {
	    fgets(buf,500,yyin);
	    lines ++;
	    /* skip empty lines */
	    if( buf[0] == '\n' || buf[0] == '\0' )
	      continue;
	    /* look for keyword */
	    if( !strncmp(buf,"-- DEFINITION OF PARAMETERS:",28) )
	      break;
	  }

	/* zurueckschreiben dieser zeile, da parse hier anfaengt */
	for (i=1; i<=28; i++)
	  ungetc (buf[28-i], yyin);
	fflineno = lines;
/*	
*/	
	status = yyparse();
#ifdef yydebug
	fprintf (stderr, 
		 "Parser yylineno: %d, status: %d\n", fflineno, status); 
#endif



	if( SyntaxOnly ) {
		fclose( yyin );
		fclose( net );
		exit( status );
	} /* if */



	for (i = 0; i < no_trans; i++) {
		if (transitions[i].ok || !transitions[i].id)
		 	 continue;
		fprintf(stderr,
			"Marking dependency of transition %s ", 
			transitions[i].name);
		fprintf(stderr, "is not defined in %s.\n", netname);
		exit(fflineno);
	} /* for */

	/* write information for struct to <net>.DEFINFO */
	fprintf(inf, "%sMD_ARCS %d\n",
		Simulation ? "" : "#", no_md_arcs-1);
	for (a_walk = md_arcs; a_walk != NULL; a_walk = a_walk->next) {
	   	fprintf(inf, "MDA %d %d %d %d\n", a_walk->id, 
			a_walk->type, a_walk->trans, a_walk->place);
		if (a_walk->deter == NULL)
		  {   printf("Error: Marking dependent arc multiplicity from ");
		      printf("%s to ", transitions[a_walk->trans].name);
		      printf("%s ", places[a_walk->place].name);
		      printf("not defined in %s.\n", netname);
		      exit(fflineno);
		    }
		deter_out(inf, a_walk->deter);
        } /* for */

	fprintf(inf, "%sENA_FCTS %d\n",
		Simulation ? "" : "#", no_ena_fcts);
	for (i = 0; i < no_ena_fcts; i++) {
	  	fprintf(inf, "%s %d %d \n",
			Simulation ? "FCT" : "ENA",
			ena_fcts[i].id, ena_fcts[i].trans);
		deter_out(inf, ena_fcts[i].deter);
	} /* for */

	fprintf(inf, "%sWEIGHTS %d\n",
		Simulation ? "" : "#", no_weights);
	for (i = 0; i < no_weights; i++) {
	   	fprintf(inf, "WGT %d %d \n",
			transitions[weights[i].trans].id, weights[i].trans);
		if (weights[i].deter == NULL)
		  {   printf("Error: Marking dependent firing weight definition ");
		      printf("for %s ", transitions[weights[i].trans].name);
		      printf("not defined in %s.\n", netname);
		      exit(fflineno);
		    }
		deter_out(inf, weights[i].deter);
	} /* for */

	fclose(yyin);	
	fclose(mdf);	
	fclose(inf);
	fclose(erc);
	if( Simulation )
	    fclose(rdy);

	fclose(net);
	exit(status);
}

void yyerror(s)
	char *s;
{
	FILE	*second; 
	int	i;
	char	err_buf[500], ch;

	if( !SyntaxOnly ) {
		if (s && *s) {
			fprintf(stderr, "%s", s);
			if (s[strlen(s)] != '\n')
				fprintf(stderr, "\n");
		} /* if */

		fprintf(stderr, "Syntax error before or in line %d of %s.\n", 
			fflineno, netname);

		second = fopen(netname, "r");
		for (i = 1; i < fflineno; i++)
			while(((ch = fgetc(second)) != '\n') && (ch != EOF));
		fgets(err_buf, 500, second);
		fprintf(stderr, "The error occured in line: %s\n", err_buf);
		fclose(second);

		if (*error_text)
		    fprintf(stderr, "%s\n", error_text);
	} /* if */
	
	exit( fflineno ); 
}


#ifdef __STDC__
int checkParameters(name_delay *p)
#else
int checkParameters(p)
    name_delay *p;
#endif
{
    int i;
    char *open;

    if (p == NULL)
	return 0;

    for (i = 0; i < no_del_par; i++)
    {
	if (p[i].name) 
	{
	    if (p[i].type == Definition)
		if (p[i].values.def)
		{
		    open = NULL;
		    if (recursive(p[i].name, &open, p)) 
			return 1;
		    free(open);
		}
		else
		{
		    fprintf(stderr, 
		     "checkParameters: '%s'->parameter.definition == NULL\n",
			    p[i].name);
		    return 1;
		}
	}
	else 
	{
	    fprintf(stderr, "checkParameters: parameter.name == NULL\n");
	    return 1;
	}
    }
    return 0;
}

/******************************************************************************
 * strip "DSPN_" from name, so get the original Parameter name 
 ******************************************************************************/
char *originalName(s)
     char *s;
{
  if (s)
  {
    if (strstr(s, "DSPN_") == s)
      return s+strlen("DSPN_");
    else
      return s;
  }
  else
    return "";
}

/****************************************************************************** 
 * return the names in 's' as originalName(s) in REVERSE ORDER 
 ******************************************************************************/
char *reverse(s)
     char *s;
{
  char *new = NULL;
  char *help = NULL;
  char name[50];

  while (sscanf(s, "%s", name) == 1) {
    if (new)
      help = vlstrcpy(&help, new);
    vlstrcpy (&new, " -> ");
    vlstrcat (&new, originalName(name));
    if (help)
      vlstrcat (&new, help);
    s = strstr(s,name) + strlen(name);    /* Uuuhaaaa! pointer arithmetic! */
  }
  return new;
}

/******************************************************************************
 * RECURSIVE definitions; 
 * returns : no => 0; yes => 1
 ******************************************************************************/
#ifdef __STDC__
int recursive (char *name, char **open, name_delay *p)
#else
int recursive (name, open, p)
    char *name;
    char **open;
    name_delay *p;
#endif
{
    char *def = NULL;
    char *s;
    char *next;
    int i;

    /* IN (name, open) ? */
    if (*open && (s = strstr(*open, name)) && !isalnum(s[strlen(name)])) 
    {
	sprintf(error_text, 
		"Recursive definition for \"%s\" : %s.\n", 
		originalName(name), reverse(*open));
	return 1;
    }
    
    /* GET DEFINITION */
    for (i = 0; i < no_del_par && !def; i++)
	if (!strcmp(name, p[i].name))
	    if (p[i].type == Definition)
		def = p[i].values.def;
	    else
		def = "";
      
    if (def == NULL) 
    {
	fprintf(stderr, "No definition for '%s'\n", name);
	return 1;
    }
    
    /* PUSH (name, open) */
    vlstrcat(open, " ");
    vlstrcat(open, name); 

    while (s = strstr(def, "DSPN_"))
    {
	next = getNext(s);

	if (recursive(next, open, p))
	    return 1;

	free(next);
	def = s+strlen(next);
    }
    
    /* POP (open) */
    s = strrchr(*open, ' ');
    *s = '\0';
    
    return 0;
}

#ifdef __STDC__
char *getNext(char *s)
#else
char *getNext(s)
     char* s;
#endif
{
    char *s2 = strdup(s);
    int i;
    
    for (i = 0; i < strlen(s2); i++) 
	if (! (s2[i]=='_' || 
	       isalnum(s2[i])))
	{
	    s2[i] = '\0';
	    break;
	}
    if (!strcmp(s2,"DSPN_"))
    {
	free (s2);
	s2 = NULL;
    }

    return s2;
}
