// File     ping.cpp      (C++ Implementation)
/**
 *          ping/pong Bandwidth measuring tool
 *
 *@author   Philipp Berndt
 *@version  $Id: ping.cpp,v 1.2 2000/01/12 16:02:01 pumphutt Exp $
 */

/* Gnuplot call:
set terminal postscript eps
set output "ping.eps"
set xlabel "Packet Size [doubles]"
set ylabel "Round Trip Time [s]"
plot [0:5000]  \
"out2" using 1:2 title "PvmDontRoute", \
"out" using 1:2 title "PvmRouteDirect"
*/

#include <string>
#include <iostream>

#include "pvm3cpp.h"
#include "Time.h"

namespace
{
    vector<string> hosts; // Hostname for each slave GIN
    class Exception {};
    const int TIMES=10;
}

int main()
{
    try {
        initPvm();

        // Mark this host as BUSY so no slave is started here.
        HostConfig::addVirtLoad(HostConfig::gethostname(), 99.9);

        pvm_setopt(PvmRoute, PvmAllowDirect); 
        int slavetid;

        // Start slave process
        cout << "Spawning slave..." << flush;
        string host = HostConfig::getHost();
        vector<string> argv;
        slavetid = pvm_spawn("pong", argv, PvmTaskHost, host);
        cout << "OK" << endl;

        // Dry run for initialization of TCP links
        pvm_initsend(PvmDataDefault);
        pvm_pkint(1);
        pvm_pkint(0);
        pvm_send(slavetid, 0);
        
        pvm_initsend(PvmDataInPlace);
        pvm_send(slavetid, 0);

        int bufid, bytes, msgtag, tid;
        bufid = pvm_recv(-1, -1);
        pvm_bufinfo(bufid, &bytes, &msgtag, &tid);
        if (msgtag == TAG_TERMINATE)
            throw Exception();

        for(int n=0;n<=10000;n+=100)
            {
            double* b = new double[n];

            pvm_initsend(PvmDataDefault);
            pvm_pkint(TIMES);
            pvm_pkint(n);
            pvm_send(slavetid, 0);
        
            Time start = Time::getTimeOfDay();      
            for(int i=0;i<TIMES;++i)
                {
                pvm_initsend(PvmDataInPlace);
                pvm_pkdouble(b, n, 1);
                pvm_send(slavetid, 0);

                int bufid, bytes, msgtag, tid;
                bufid = pvm_recv(-1, -1);
                pvm_bufinfo(bufid, &bytes, &msgtag, &tid);
                if (msgtag == TAG_TERMINATE)
                    throw Exception();
                pvm_upkdouble(b, n, 1);
                }
            Time dur = Time::getTimeOfDay() - start;        
            double span = (dur.get_sec() + dur.msecs()/1000.0) / TIMES;

            cout << n << " " << span << endl;
            delete b;
            }
    
        pvm_kill(slavetid);

        }
    catch(...)
        {
        }
    shutdownPvm();
}
