// File    pvm3wrap.h    (C++ Inline Implementation) 
// Module  pvm3wrap      (C++ Wrapper for PVM3.4 Functions)
// Library PVM3CPP       (C++ Extension to PVM3.4)
/**
 *         C++ Wrapper for PVM3.4 Functions
 *
 *@author  Philipp Berndt
 *@version $Id: pvm3wrap.h,v 1.2 2000/01/12 16:02:02 pumphutt Exp $
 */

#ifndef PVM3WRAP_H
#define PVM3WRAP_H

#include <stdio.h>
#include <string>
#include <sys/time.h>  

namespace PVM3 
{
#include <pvm3.h>
#undef _UID_T
}

/**
 *         Exception thrown by pvm3cpp functions.
 *
 *@author  Philipp Berndt
 *@version $Id: pvm3wrap.h,v 1.2 2000/01/12 16:02:02 pumphutt Exp $
 */
class PvmException
{
public:
/** Default Constructor */
    PvmException() : msg(0) {}

/** Constructor with user error message. */
    PvmException(const char* m) : msg(m) {}

/** Prints PVM error message, location and user error message to stderr. */
    void perror(string location);

private:
    const char* msg;
};

namespace P3PP {
using PVM3::pvmhostinfo;
using PVM3::pvmtaskinfo;
using PVM3::pvmminfo;
using PVM3::pvmmboxinfo;

// The following functions were not modified at all
using PVM3::pvm_archcode;
using PVM3::pvm_delhosts;
using PVM3::pvm_export;
using PVM3::pvm_getcontext;
using PVM3::pvm_getrbuf;
using PVM3::pvm_getsbuf;
using PVM3::pvm_newcontext;
using PVM3::pvm_packf;
using PVM3::pvm_pstat;
using PVM3::pvm_putinfo;
using PVM3::pvm_setcontext;
using PVM3::pvm_unexport;
using PVM3::pvm_unpackf;
using PVM3::pvm_getminfo;
using PVM3::pvm_getnoresets;
using PVM3::pvm_reg_tracer;
using PVM3::pvm_setminfo;
using PVM3::pvm_tickle;
using PVM3::pvm_version;

using PVM3::PvmMax;
using PVM3::PvmMin;
using PVM3::PvmSum;
using PVM3::PvmProduct;


// The following functions were modified to throw a PvmException instead of 
// returning a negative value.

inline int pvm_getinst( const char *group, int tid ) {
    int inum = PVM3::pvm_getinst( const_cast<char*>(group), tid );
    if(inum<0) throw PvmException();
    return inum; }

inline int pvm_joingroup( const char *group ) {
    int inum = PVM3::pvm_joingroup( const_cast<char*>(group) );
    if(inum<0) throw PvmException();
    return inum; }

inline int pvm_mkbuf( int encoding ) {
    int bufid = PVM3::pvm_mkbuf( encoding );
    if(bufid<0) throw PvmException();
    return bufid; }

inline int pvm_addmhf( int src, int tag, int ctx, int (*func)(int mid) ) {
    int mhid = PVM3::pvm_addmhf( src, tag, ctx, func );
    if(mhid<0) throw PvmException();
    return mhid; }

inline int pvm_getfds( int **fds ) {
    int nfds = PVM3::pvm_getfds( fds );
    if(nfds<0) throw PvmException();
    return nfds; }

inline int pvm_getopt( int what ) {
    int val = PVM3::pvm_getopt( what );
    if(val<0) throw PvmException();
    return val; }

inline int pvm_gettid( const char *group, int inum ) {
    int tid = PVM3::pvm_gettid( const_cast<char*>(group), inum );
    if(tid<0) throw PvmException();
    return tid; }

inline int pvm_gsize( const char *group ) {
    int size = PVM3::pvm_gsize( const_cast<char*>(group) );
    if(size<0) throw PvmException();
    return size; }

inline int pvm_initsend( int encoding ) {
    int bufid = PVM3::pvm_initsend( encoding );
    if(bufid<0) throw PvmException();
    return bufid; }

inline int pvm_mstat( const char *host ) {
    int mstat = PVM3::pvm_mstat( const_cast<char*>(host) );
    if(mstat == PvmSysErr) throw PvmException();
    return mstat; }

inline int pvm_mytid( void ) {
    int tid = PVM3::pvm_mytid();
    if(tid<0) throw PvmException();
    return tid; }

inline int pvm_nrecv( int tid, int msgtag ) {
    int bufid = PVM3::pvm_nrecv( tid, msgtag );
    if(bufid<0) throw PvmException();
    return bufid; }

inline int pvm_parent( void ) {
    int tid = PVM3::pvm_parent();
    if(tid<0) throw PvmException();
    return tid; }

inline int pvm_probe( int tid, int msgtag ) {
    int bufid = PVM3::pvm_probe( tid, msgtag );
    if(bufid<0) throw PvmException();
    return bufid; }

inline int pvm_recv( int tid, int msgtag ) {
    int bufid = PVM3::pvm_recv( tid, msgtag );
    if(bufid<0) throw PvmException();
    return bufid; }

inline int pvm_recvinfo( const char *name, int index, int flags ) {
    int bufid = PVM3::pvm_recvinfo( const_cast<char*>(name), index, flags );
    if(bufid<0) throw PvmException();
    return bufid; }

inline int pvm_reg_hoster() {
    int cc = PVM3::pvm_reg_hoster();
    if(cc<0) throw PvmException();
    return cc; }

inline int pvm_reg_rm( struct pvmhostinfo **hip ) {
    int cc = PVM3::pvm_reg_rm( hip );
    if(cc<0) throw PvmException();
    return cc; }

inline int pvm_reg_tasker() {
    int cc = PVM3::pvm_reg_tasker();
    if(cc<0) throw PvmException();
    return cc; }

inline int pvm_setopt( int what, int val ) {
    int oldval = PVM3::pvm_setopt( what, val );
    if(oldval<0) throw PvmException();
    return oldval; }

inline int pvm_setrbuf( int bufid ) {
    int oldbuf = PVM3::pvm_setrbuf( bufid );
    if(oldbuf<0) throw PvmException();
    return oldbuf; }

inline int pvm_setsbuf( int bufid ) {
    int oldbuf = PVM3::pvm_setsbuf( bufid );
    if(oldbuf<0) throw PvmException();
    return oldbuf; }

inline int pvm_siblings( int **tids ) {
    int ntids = PVM3::pvm_siblings( tids );
    if(ntids<0) throw PvmException();
    return ntids; }

inline int pvm_spawn( const char *task, const char **argv, int flag,
		      const char *where, int ntask, int *tids ) {
    int numt=PVM3::pvm_spawn(const_cast<char*>(task), const_cast<char**>(argv),
			     flag, const_cast<char*>(where),
			     ntask, tids);
    if(numt<0) throw PvmException();
    return numt; }

inline int pvm_tidtohost( int tid ) {
    int dtid = PVM3::pvm_tidtohost( tid );
    if(dtid<0) throw PvmException();
    return dtid; }

inline int pvm_trecv( int tid, int msgtag, const struct timeval *tmout ) {
    int bufid = PVM3::pvm_trecv( tid, msgtag, const_cast<timeval*>(tmout) );
    if(bufid<0) throw PvmException();
    return bufid; }


// The following functions were modified to return void instead of info
// and now throw a PvmException in case of an error.

inline void pvm_addhosts( const char **hosts, int nhost, int *infos ) {
    int info = PVM3::pvm_addhosts( const_cast<char**>(hosts), nhost, infos ); 
    if(info<0) throw PvmException(); }

inline void pvm_barrier( const char *group, int count ) {
    int info = PVM3::pvm_barrier( const_cast<char*>(group), count );
    if(info<0) throw PvmException(); }

inline void pvm_bcast( const char *group, int msgtag ) {
    int info = PVM3::pvm_bcast( const_cast<char*>(group), msgtag );
    if(info<0) throw PvmException(); }

inline void pvm_bufinfo( int bufid, int *bytes, int *msgtag, int *tid) {
    int info = PVM3::pvm_bufinfo( bufid, bytes, msgtag, tid);
    if(info<0) throw PvmException(); }

inline void pvm_catchout( FILE *ff ) {
    int info = PVM3::pvm_catchout( ff );
    if(info<0) throw PvmException(); }

inline void pvm_config( int *nhost, int *narch, struct pvmhostinfo **hostp ) {
    int info = PVM3::pvm_config( nhost, narch, hostp );
    if(info<0) throw PvmException(); }

inline void pvm_delinfo( const char *name, int index, int flags ) {
    int info = PVM3::pvm_delinfo( const_cast<char*>(name), index, flags );
    if(info<0) throw PvmException(); }

inline void pvm_delmhf( int mhid ) {
    int info = PVM3::pvm_delmhf( mhid );
    if(info<0) throw PvmException(); }

inline void pvm_exit( void ) {
    int info = PVM3::pvm_exit();
    if(info<0) throw PvmException(); }

inline void pvm_freebuf( int bufid ) {
    int info = PVM3::pvm_freebuf( bufid );
    if(info<0) throw PvmException(); }

inline void pvm_freecontext( int ctx ) {
    int info = PVM3::pvm_freecontext( ctx );
    if(info<0) throw PvmException(); }

inline void pvm_gather( void *result,  void *data, int count, int datatype,
			int msgtag, char *group, int rootginst)  {
    int info = PVM3::pvm_gather( result, data, count, datatype,
				 msgtag, const_cast<char*>(group), rootginst);
    if(info<0) throw PvmException(); }

inline void pvm_getmboxinfo( char *pattern, int *nclasses,
			     struct pvmmboxinfo **classes )   {
    int info = PVM3::pvm_getmboxinfo( pattern, nclasses, classes )  ;
    if(info<0) throw PvmException(); }

inline void pvm_hostsync( int host, const struct timeval *clk, 
			  const struct timeval *delta)  {
    int info = PVM3::pvm_hostsync( host, const_cast<timeval*>(clk), 
				   const_cast<timeval*>(delta));
    if(info<0) throw PvmException(); }

inline void pvm_kill( int tid ) {
    int info = PVM3::pvm_kill( tid );
    if(info<0) throw PvmException(); }

inline void pvm_lvgroup( const char *group ) {
    int info = PVM3::pvm_lvgroup( const_cast<char*>(group) );
    if(info<0) throw PvmException(); }

inline void pvm_mcast( int *tids, int ntask, int msgtag ) {
    int info = PVM3::pvm_mcast( tids, ntask, msgtag );
    if(info<0) throw PvmException(); }

inline void pvm_notify( int what, int msgtag, int cnt, const int *tids ) {
    int info = PVM3::pvm_notify( what, msgtag, cnt, const_cast<int*>(tids) );
    if(info<0) throw PvmException(); }

inline void pvm_perror( const char *msg ) {
    int info = PVM3::pvm_perror( const_cast<char*>(msg) );
    if(info<0) throw PvmException(); }

inline void pvm_pkbyte( const char *xp, int nitem, int stride ) {
    int info = PVM3::pvm_pkbyte( const_cast<char*>(xp), nitem, stride );
    if(info<0) throw PvmException(); }

inline void pvm_pkcplx( const float *cp, int nitem, int stride ) {
    int info = PVM3::pvm_pkcplx( const_cast<float*>(cp), nitem, stride );
    if(info<0) throw PvmException(); }

inline void pvm_pkdcplx( const double *zp, int nitem, int stride ) {
    int info = PVM3::pvm_pkdcplx( const_cast<double*>(zp), nitem, stride );
    if(info<0) throw PvmException(); }

inline void pvm_pkdouble( const double *dp, int nitem, int stride ) {
    int info = PVM3::pvm_pkdouble( const_cast<double*>(dp), nitem, stride );
    if(info<0) throw PvmException(); }

inline void pvm_pkfloat( const float *fp, int nitem, int stride ) {
    int info = PVM3::pvm_pkfloat( const_cast<float*>(fp), nitem, stride );
    if(info<0) throw PvmException(); }

inline void pvm_pkint( const int *ip, int nitem, int stride ) {
    int info = PVM3::pvm_pkint( const_cast<int*>(ip), nitem, stride );
    if(info<0) throw PvmException(); }

inline void pvm_pklong( const long *ip, int nitem, int stride ) {
    int info = PVM3::pvm_pklong( const_cast<long*>(ip), nitem, stride );
    if(info<0) throw PvmException(); }

inline void pvm_pkshort( const short *jp, int nitem, int stride ) {
    int info = PVM3::pvm_pkshort( const_cast<short*>(jp), nitem, stride );
    if(info<0) throw PvmException(); }

inline void pvm_pkstr( const char *sp ) {
    int info = PVM3::pvm_pkstr( const_cast<char*>(sp) );
    if(info<0) throw PvmException(); }

inline void pvm_pkuint( const unsigned int *ip, int nitem, int stride ) {
    int info = PVM3::pvm_pkuint(const_cast<unsigned int*>(ip), nitem, stride);
    if(info<0) throw PvmException(); }

inline void pvm_pkulong( const unsigned long *ip, int nitem, int stride ) {
    int info = PVM3::pvm_pkulong( const_cast<unsigned long*>(ip),
				  nitem, stride );
    if(info<0) throw PvmException(); }

inline void pvm_pkushort( const unsigned short *ip, int nitem, int stride ) {
    int info = PVM3::pvm_pkushort( const_cast<unsigned short*>(ip),
				   nitem, stride );
    if(info<0) throw PvmException(); }

inline void pvm_precv( int tid, int msgtag, char *buf,int len,
		       int datatype, int *atid, int *atag, int *alen )  {
    int info = PVM3::pvm_precv( tid, msgtag, buf, len,
				datatype, atid, atag, alen );
    if(info<0) throw PvmException(); }

inline void pvm_psend( int tid, int msgtag, void *buf, 
		       int len, int datatype ) {
    int info = PVM3::pvm_psend( tid, msgtag, buf, len, datatype );
    if(info<0) throw PvmException(); }

inline void pvm_reduce( void (*func)(int*, void*, void*, int*, int*),
			void *data, int count, int datatype,
			int msgtag, const char *group, int rootginst)  {
    int info = PVM3::pvm_reduce( func, data, count, datatype,
				 msgtag, const_cast<char*>(group), rootginst);
    if(info<0) throw PvmException(); }

inline void pvm_scatter( void *result, void *data,int count, int datatype,
			 int msgtag, char *group, int rootginst)  {
    int info = PVM3::pvm_scatter( result, data, count, datatype, msgtag,
				  const_cast<char*>(group), rootginst) ;
    if(info<0) throw PvmException(); }

inline void pvm_send( int tid, int msgtag ) {
    int info = PVM3::pvm_send( tid, msgtag );
    if(info<0) throw PvmException(); }

inline void pvm_sendsig( int tid, int signum ) {
    int info = PVM3::pvm_sendsig( tid, signum );
    if(info<0) throw PvmException(); }

inline void pvm_start_pvmd( int argc, char **argv, int block ) {
    int info = PVM3::pvm_start_pvmd( argc, argv, block );
    if(info<0) throw PvmException(); }

inline void pvm_tasks( int where, int *ntask, struct pvmtaskinfo **taskp ) {
    int info = PVM3::pvm_tasks( where, ntask, taskp );
    if(info<0) throw PvmException(); }

inline void pvm_upkbyte( char *xp, int nitem, int stride) {
    int info = PVM3::pvm_upkbyte( xp, nitem, stride);
    if(info<0) throw PvmException(); }

inline void pvm_upkcplx( float *cp, int nitem, int stride) {
    int info = PVM3::pvm_upkcplx( cp, nitem, stride);
    if(info<0) throw PvmException(); }

inline void pvm_upkdcplx( double *zp, int nitem, int stride) {
    int info = PVM3::pvm_upkdcplx( zp, nitem, stride);
    if(info<0) throw PvmException(); }

inline void pvm_upkdouble( double *dp, int nitem, int stride) {
    int info = PVM3::pvm_upkdouble( dp, nitem, stride);
    if(info<0) throw PvmException(); }

inline void pvm_upkfloat( float *fp, int nitem, int stride) {
    int info = PVM3::pvm_upkfloat( fp, nitem, stride);
    if(info<0) throw PvmException(); }

inline void pvm_upkint( int *ip, int nitem, int stride) {
    int info = PVM3::pvm_upkint( ip, nitem, stride);
    if(info<0) throw PvmException(); }

inline void pvm_upklong( long *ip, int nitem, int stride) {
    int info = PVM3::pvm_upklong( ip, nitem, stride);
    if(info<0) throw PvmException(); }

inline void pvm_upkshort( short *jp, int nitem, int stride) {
    int info = PVM3::pvm_upkshort( jp, nitem, stride);
    if(info<0) throw PvmException(); }

inline void pvm_upkstr( char *sp ) {
    int info = PVM3::pvm_upkstr( sp );
    if(info<0) throw PvmException(); }

inline void pvm_upkuint( unsigned int *ip, int nitem, int stride ) {
    int info = PVM3::pvm_upkuint( ip, nitem, stride );
    if(info<0) throw PvmException(); }

inline void pvm_upkulong( unsigned long *ip, int nitem, int stride ) {
    int info = PVM3::pvm_upkulong( ip, nitem, stride );
    if(info<0) throw PvmException(); }

inline void pvm_upkushort( unsigned short *ip, int nitem, int stride ) {
    int info = PVM3::pvm_upkushort( ip, nitem, stride );
    if(info<0) throw PvmException(); }

} // namespace P3PP


/*****************************************************************************\
*                         INLINE FUNCTION DEFINITIONS                         *
*               No classes or user interfaces below this line !               *
\*****************************************************************************/

inline void PvmException::perror(string location)
{
P3PP::pvm_perror((string("in ")+location).c_str());
if(msg)
    cerr << msg << endl; 
}

#endif // PVM3WRAP_H
