// File    pvm_util.cpp  (C++ Implementation)
// Module  pvm_util      (New Functions not found in PVM3.4)
// Library PVM3CPP       (C++ Extension to PVM3.4)
/**
 *@author  Philipp Berndt
 *@version $Id: pvm_util.cpp,v 1.2 2000/01/12 16:02:03 pumphutt Exp $
 */

#include <algorithm> // find()
#include <signal.h>  // sigignore()
#include "memory"    // use <memory> as soon as auto_ptr becomes available !!!

#include "pvm_util.h"
using namespace P3PP;

PVM_UNPACK_t PVM_UNPACK; // Dummy argument to unpacking constructors.

void pvm_halt()
{
    void (*disp)(int);
    disp = signal(SIGTERM, SIG_IGN);
    int info = PVM3::pvm_halt();
    signal(SIGTERM, disp);
    if (info<0) throw PvmException();
}

void pvm_reset()
{
    int ntask;
    int mytid = pvm_mytid();
    pvmtaskinfo *taskinfos;
    P3PP::pvm_tasks(0, &ntask, &taskinfos);
    for (int i=0; i<ntask; ++i)   // Kill all tasks
        {
        int tid = taskinfos[i].ti_tid;
        if (tid == 0 || tid == mytid || taskinfos[i].ti_a_out[0] == 0)
            continue;            // Spare zombies, myself and consoles

        cout << "Killing " << taskinfos[i].ti_a_out 
             << " Tid:" << hex << tid << dec << endl;
        pvm_kill(tid);
        }
}

void pvm_start_pvmgs(int flag, const string& where)
{
    int gstid;
    int srbuf = pvm_setrbuf(0);
  
    if ( PVM3::pvm_recvinfo("pvmgs", 0, PvmMboxDefault) == PvmNotFound )
        {
        // Don't trace output from the spawned group server
        int otid = pvm_setopt(PvmOutputTid, 0);
        int ttid = pvm_setopt(PvmTraceTid, 0);
      
        int info = PVM3::pvm_spawn("pvmgs", 0, flag | PvmMppFront,
                                   const_cast<char*>(where.c_str()),
                                   1, &gstid);
      
        pvm_setopt(PvmOutputTid, otid);
        pvm_setopt(PvmTraceTid, ttid);
      
        if (info != 1)
            throw PvmException("Failed to start group server");
      
        while ( PVM3::pvm_recvinfo("pvmgs", 0, PvmMboxDefault) == PvmNotFound )
            /* wait for it to register */ ;
        }
  
    // unpack the group server tid
    pvm_upkint(&gstid,1,1);
    pvm_setrbuf(srbuf);
}

vector<int> pvm_addhosts(const vector<string> &hosts)
{
    if (!hosts.size())
        return vector<int>();

    // Convert vector<string> to array of char*
    const char *hostarr[hosts.size()];
    vector<string>::const_iterator i= hosts.begin();
    const char **arrayp = hostarr;
    while (i != hosts.end())
        *(arrayp++) = (i++)->c_str();

    int infos[hosts.size()];
    pvm_addhosts(hostarr, hosts.size(), infos);
    return vector<int>(&infos[0],&infos[hosts.size()]);
}

vector<int> pvm_spawn(const char *task, const vector<string> &argv,
                      int flag, const string& where, int ntask)
{
    // Convert vector<string> to array of char*
    const char *myargv[argv.size()];
    vector<string>::const_iterator i= argv.begin();
    const char **arrayp = myargv;
    while (i != argv.end())
        *(arrayp++) = (i++)->c_str();

    int tids[ntask];
    int numt = PVM3::pvm_spawn(const_cast<char*>(task),
                               const_cast<char**>(myargv),
                               flag, const_cast<char*>(where.c_str()),
                               ntask, tids );
    if (numt<0) throw PvmException();
    return vector<int>(&tids[0],&tids[ntask]);
}

int pvm_spawn(const char *task, const vector<string> &argv,
              int flag, const string& where)
{
    const char *myargv[argv.size()+1];
    vector<string>::const_iterator i= argv.begin();
    const char **arrayp = myargv;
    while (i != argv.end())
        *(arrayp++) = (i++)->c_str();
    *(arrayp) = 0;

    int tid;
    int numt = PVM3::pvm_spawn(const_cast<char*>(task),
                               const_cast<char**>(myargv),
                               flag, const_cast<char*>(where.c_str()), 1,&tid);
    if (numt<0) throw PvmException();
    return tid;
}

int pvm_tid2pid(int tid)
{
    int ntask;
    pvmtaskinfo *tip;
    P3PP::pvm_tasks(tid, &ntask, &tip);
    return tip->ti_pid;
}

int pvm_addhost(const string& hostname)
{
    const char *host = hostname.c_str();
    int tid;
    P3PP::pvm_addhosts(&host, 1, &tid);
    return tid;
}
