// File    tnet_util.h    (C++ Interface)
// Module  tnet_util      (TimeNET specific functions related to PVM)
// Library PVM3CPP        (C++ Extension to PVM3.4)
/**
 *@author  Philipp Berndt
 *@version $Id: tnet_util.h,v 1.2 2000/01/12 16:02:05 pumphutt Exp $
 */

#ifndef TNET_UTIL_H
#define TNET_UTIL_H

/**
 * Initializes entire PVM environment.
 *
 * Starts PVM, reads ".TimeNET", adds hosts
 * and installs handler to transform SIGINT & SIGTERM to a TERMINATE message.
 */
void initPvm();

/**
 * Shuts down entire PVM environment.
 *
 * Terminates all child processes, but not the pvmgs.
 * If that leaves no other processes running (besides pvmgs),
 * the virtual machine will be halted, else only the hosts
 * that we added will be removed.
 */
void shutdownPvm();

/**
 * Checks whether the received message is a message regarding changes
 * to the virtual machine and takes apropriate actions.
 *
 * Use as follows:
 * <pre>
 *@@ int bufid, bytes, msgtag, tid;
 *@@ bufid = pvm_recv(-1, -1);
 *@@ pvm_bufinfo(bufid, &bytes, &msgtag, &tid);
 *@@ if(!checkSystemMessage(msgtag))
 *@@     switch(msgtag)
 *@@         {
 * </pre> 
 *
 * If the message tag is TAG_HOST_ADDED, HostConfig::hostAdded is called.
 * If the message tag is TAG_HOST_DELETED, HostConfig::hostRemoved is called.
 * If the message tag is TAG_TERMINATE an InterruptedException is thrown.
 *@returns true if message was a system message.
 */
bool checkSystemMessage(int msgtag);

/**
 * Exeption thrown by checkSystemMessage()
 * when it receives a TERMINATE message.
 */
class InterruptedException {};

#endif // TNET_UTIL_H
