/*****************************************************************************

Programm	:	callestimate

Komponente	:	
                        Zustandsraumabschaetzung auf Basis
			der Zustandsgleichung (state equation)
			Aufruf mit Netznamen, benoetigt
			zwei Dateien .inc mit Inzidenzmatrix (Stellen x Trans)
			und .m0 mit Anfangsmarkierung

Programmer	:	Alexander Huck
Date		:	1999

*****************************************************************************/

#include <malloc.h>
#include <netio.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "matrixIO.h"
#include "matrixOps.h"
#include "matrix.h"
#include "timer.h"
#include "estimate.h"

INTMATRIX *IncidenceMat, *MNull, *Mptr;
char model_name[256];
int statespace;
FILE 		*inc_file, *mnull_file;


/*******************************************************************************************
 output result of estimation to .ese- file
*******************************************************************************************/
static void output_result(double timing)
{
  FILE   *ese_file, *f;
  char	help[1000];
  
  sprintf(help,"%s.ese", model_name);
  if ((ese_file = fopen(help, "w")) == NULL)
    {
      printf("ERROR: could not open %s for writing\n", help);
      exit(1);
    }
  
  
  f = ese_file;		/* output to .ese */
  
  fprintf(f, "Result of estimation (based on state equation):\n");
  fprintf (f, "Statespace = ");
  if (statespace == -1)
    {
      fprintf(f," oo\n");
    }
  else
    fprintf(f," %i \n", statespace);
  
  fprintf (f, "Time passed with computation: %8.2f s\n", timing );
  
  fclose(ese_file);


  printf( "Result of estimation (based on state equation):\n");
  printf ( "Statespace = ");
  if (statespace == -1)
    {
      printf(" oo\n");
    }
  else
    printf(" %i \n", statespace);
  
  printf ( "Time passed with computation: %8.2f s\n", timing );
  
 
}

static void usage()
{
  printf("Usage: callestimate netname\n");
  printf("estimates the state space of a net given by\n");
  printf("untransposed incidence matrix netname.inc\n");
  printf("initial marking netname.m0\n");

  exit( -1 );
}

/*******************************************************************************************
 start estimation
*******************************************************************************************/
static void se()
{
  start_timer(20);
 
  statespace = EstimateStateEq(MNull, IncidenceMat); /*from estimate.c*/
  
  FreeMat(IncidenceMat);
  FreeMat(MNull);
  stop_timer(20);
}


/*	the main procedure handles options and file access and
	calls subordinated procedures */
int main(argc, argv)
	int argc;
	char ** argv;
{
	char net_name[256], inc_name[256], mnull_name[256];
	extern int optind;

	/* check arguments */
	if (argc == 1)
		usage();

	/* build filenames */
	sprintf(net_name,"%s", argv[optind]);
	sprintf(inc_name,"%s.inc", net_name);
	sprintf(mnull_name,"%s.m0", net_name);

	if ((inc_file = fopen(inc_name, "r")) == NULL)
	  {	printf("ERROR: could not open %s for reading\n", inc_name);
	  exit(1);
	  }
	if ((mnull_file = fopen(mnull_name, "r")) == NULL)
	  {	
	    printf("ERROR: could not open %s for reading\n", mnull_name);
	    exit(1);
	  }

	Mptr = Matrix_Read(inc_file);
	IncidenceMat=TransMat(Mptr);
	FreeMat(Mptr);
	fclose(inc_file);
	MNull = Matrix_Read(mnull_file);
	fclose(mnull_file);

	se();
	output_result(time_passed(20));
	return 0;			/* o.k */
}
