/*****************************************************************************

Programm	:	callhermite

Komponente	:	callhermite.c
                        Berechnet Spalten-Hermite-Normalform
			und Rang der Eingabe-Matrix

Programmer	:	Alexander Huck
Date		:	1999

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "matrixIO.h"
#include "matrixOps.h"
#include "matrix.h"
#include "matrixProcs.h"
#include "timer.h"
#include "invars.h"
#include "malloc.h"
#include "inequalities.h"
#include "invars.h"
#include "hermite2.h"

/*#define DEBUG*/

int main(argc, argv)
int argc;
char *argv[];
 {
   char filename[50];
   INTMATRIX *Mptr, *ROWECH, *TRANSMAT, *ZBASE;
   FILE *infile;
   int rank;
   int numsol, numcols, numrows;
   int i,j;
#ifdef DEBUG
 char debugfilename[50];
  FILE *outfile;
#endif
  
#ifdef DEBUG
  strcpy(debugfilename, "callhermite.out");
  outfile = fopen(debugfilename, "w");
#endif  

   if (argc != 2)
     {
       printf("nee, so geht das nicht\n");
       printf("bitte Dateinamen fuer Datei mit Matrix \n");
       return 1;
     }

   strcpy(filename, argv[1]); 

   infile = fopen(filename, "r");
   if (infile == NULL)
     printf("does not exist or cannot be opened");
   Mptr = Matrix_Read(infile);
   fclose(infile);

#ifdef DEBUG
   fprintf(outfile,"callhermite: INPUT MATRIX \n");
   FPRINTMATI(outfile,0,Mptr->R-1,0,Mptr->C-1,Mptr);
#endif  

   
 /*calculate Hermite Normal Form*/
  TRANSMAT = LLLHermite2(Mptr,
			 &ROWECH,
			 3,
			 4,
			 &rank);
  
#ifdef DEBUG
  printf("The transformation matrix is\n");
  PRINTMATI(0,TRANSMAT->R-1,0,TRANSMAT->C-1,TRANSMAT);
  fprintf(outfile,"The transformation matrix is\n");
  FPRINTMATI(outfile,0,TRANSMAT->R-1,0,TRANSMAT->C-1,TRANSMAT);
  
  printf("The row echelon Form is\n");
  PRINTMATI(0,ROWECH->R-1,0,ROWECH->C-1,ROWECH);
  fprintf(outfile,"The row echelon Form is\n");
  FPRINTMATI(outfile,0,ROWECH->R-1,0,ROWECH->C-1,ROWECH);
  
  fprintf(outfile,"Rank of Matrix %i\n", rank);
#endif
  
  numrows=TRANSMAT->R;
  numcols=TRANSMAT->C;
  /*number of solutions=
    columns of transformation matrix - rank*/
  numsol = numcols - rank;
  
  if(numsol==0 || rank==0)
    {
      printf("Keine Lsung\n");
      return 0;
    }
  else
    printf("Rank of Matrix %i\n", rank);
  
  /*read out the #numsol column vectors that
    form the Z-Base*/
  
  ZBASE=ZeroMatrix(numrows, numsol);
  for (i=0; i<numrows; i++)
    {
      for(j=rank; j<numcols; j++) 
	{
	  ZBASE->V[i][j-rank]=TRANSMAT->V[i][j];
	}
    }
  
  FreeMat(TRANSMAT);
  
#ifdef DEBUG
  printf("Z-Basis:\n");
  PRINTMATI(0,ZBASE->R-1,0,ZBASE->C-1,ZBASE);
  fprintf(outfile,"Z-Basis:\n");
  FPRINTMATI(outfile,0,ZBASE->R-1,0,ZBASE->C-1,ZBASE);
  
#endif	    
  
  FreeMat(ZBASE);
   return 0;
}

