/*****************************************************************************

Programm	:	descr

Komponente	:	createdescr.c
                        Erzeugt fuer ein Netz die Dateien
			.inc (Inzidenzmatrix)
			.m0 (initiale Markierung)
			.pd (place description, Namen der Stellen)
			Diese Basisinformationen werden zur Erreichbarkeitsanalyse
			verwendet.

Programmer	:	Alexander Huck
Date		:	1999

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "struct_types.h"			/* definitions and types */
#include "struct_IO.h"
#include "struct_vars.h"

#include "malloc.h"   

#include "matrix.h"
#include "matrixIO.h"
#include "matrixOps.h"
#include "intOps.h"

extern int  	minus();
extern sp_place **sparse_function();

char model_name[FILE_NAME];

/*******************************************************************************************
 create .inc, .m0, .pd - file
*******************************************************************************************/
static void output_result()
{
  FILE   *inc_file, *m0_file, *pd_file;
  INTMATRIX *Inc;
  char	help[1000];
  int i,j;
  
  /*write incidence matrix to .inc file*/
  sprintf(help,"%s.inc", model_name);
  if ((inc_file = fopen(help, "w")) == NULL)
    {	printf("ERROR: could not open %s.inc for writing\n", help);
    exit(1);
    }
  
  Inc = conv_matrix(c_matrix);
  FPRINTMATI(inc_file, 0, Inc->R-1, 0, Inc->C-1, Inc);
  
  fclose(inc_file);
  
  /*write initial marking to .m0 file*/
  sprintf(help,"%s.m0", model_name);
  if ((m0_file = fopen(help, "w")) == NULL)
    {
      printf("ERROR: could not open %s.m0 for writing\n", help);
      exit(1);
    }
  
  fprintf(m0_file,"%i 1\n", no_places);
  for(i=0; i<no_places; i++)
    fprintf(m0_file,"%d \n", places[i].m0);
  fclose(m0_file);
  
  /*write place description file .pd*/
  sprintf(help,"%s.pd", model_name);
  if ((pd_file = fopen(help, "w")) == NULL)
    {
      printf("ERROR: could not open %s.pd for writing\n", help);
      exit(1);
    }
  fprintf(pd_file, "#File generated by Shared/src/struct_tn/createdescr.c\n");
  fprintf(pd_file, "#List of places | \"original\" | \"e\"\n"); 
  fprintf(pd_file, "nbplaces %i\n", no_places);
  for(j=0;j<no_places;j++)
    {
      fprintf(pd_file,places[j].name);
      fprintf(pd_file," original e \n");
    }
  fclose(pd_file);
}

static void usage()
{
  printf("Usage: descr netname\n");
  printf("creates netname.inc (incidence matrix)\n");
  printf("        netname.m0  (initial marking)\n");
  printf("        netname.pd  (place description)\n");
  exit( -1 );
}

int main(argc, argv)
	int argc;
	char ** argv;
{
	char net_name[FILE_NAME], struct_name[FILE_NAME], inf_name[FILE_NAME];
	extern int optind;
      
	/* check arguments */
	if (argc == 1)
		usage();

	/* build filenames */
	sprintf(net_name,"%s", argv[optind]);
	sprintf(struct_name,"%s.STRUCT", net_name);
	sprintf(inf_name,"%s.DEFINFO", net_name);
	sprintf(model_name,"%s", argv[optind]);


	/* open STRUCT & DEFINFO files */
	if ((struct_file = fopen(struct_name, "w")) == NULL)
	{	printf("ERROR: could not open %s for writing\n", struct_name);
		exit(1);
	}
	if ((inf_file = fopen(inf_name, "r")) == NULL)
	{	printf("ERROR: could not open %s for reading\n", inf_name);
		exit(1);
	}

	/* begin net analysis */
	read_net(net_name);
	read_info(inf_file);		/* read additional information */

	if (no_places < 1)
	{	printf("Error: no places.\n");
		exit(1);
	}
	if (no_timed_trans < 1)
	{	printf("Error: no timed transitions.\n");
		exit(1);
	}

#ifdef DEBUG
	printf("I-Matrix:\n");
	print_matrix(i_matrix);
	printf("O-Matrix:\n");
	print_matrix(o_matrix);
	printf("M0:\n");
	for(i=0; i<no_places; i++)
	printf("%d ", places[i].m0);
#endif
					/* calculate the global c-matrix */
	c_matrix = sparse_function(o_matrix, i_matrix, no_trans, minus); 
#ifdef DEBUG
	printf("C-Matrix:\n");
 	print_matrix(c_matrix); 
#endif

	output_result();
	return 0;			/* o.k */
}

