/*****************************************************************************

Programm	:	struct, traps, siphons, estimse, reachable

Komponente	:	debugOps.c
                        Enthaelt Funktionen zur Ausgabe
			von Debugging-Informationen

Programmer	:	Alexander Huck
Date		:	1999

*****************************************************************************/

#include <stdio.h>
#include "sharedVars.h"
#include "matrix.h"

/*******************************************************************************************
       for debugging only:
       prints the ray-space partitioned in verify, saturate and not verify
*******************************************************************************************/
void PartitionPrint(Dst, Format, Mat, NbBid, equal, sup, bound, Cols)
     FILE* Dst;
     char *Format;
     INTMATRIX *Mat;
     int NbBid, equal, sup, bound, Cols;
{
  int *p, i, j;
  fprintf(Dst,"Partition of RaySpace: \n");
  
   fprintf(Dst,"Lines: \n"); 
  for (i=0;i<NbBid;i++) {
    p=*(Mat->V+i);
    for (j=0;j<=Cols;j++)
      fprintf(Dst, Format, *p++);
    fprintf(Dst, "\n");
  }
  
  fprintf(Dst,"Rays which verify the constraint: a.r>0: \n"); 
  for (i=NbBid;i<sup;i++) {
    p=*(Mat->V+i);
    for (j=0;j<=Cols;j++)
      fprintf(Dst, Format, *p++);
    fprintf(Dst, "\n");
  }
  
  fprintf(Dst,"Rays which saturate the constraint: a.r=0 \n"); 
  for (i=sup;i<equal;i++) {
    p=*(Mat->V+i);
    for (j=0;j<=Cols;j++)
      fprintf(Dst, Format, *p++);
    fprintf(Dst, "\n");
  }
  
  fprintf(Dst,"Rays which do not verify the constraint a.r<0: \n"); 
  for (i=equal;i<bound;i++) {
    p=*(Mat->V+i);
    for (j=0;j<=Cols;j++)
      fprintf(Dst, Format, *p++);
    fprintf(Dst, "\n");
  }
} /* Partition_Print */

/*******************************************************************************************
       for debugging only:
       prints the index-arrays
*******************************************************************************************/
void PrintIndex()
{
  USI i;
  USI pos1, pos2;
  printf("IndexArrays LR RL\n");
  for (i= 0 ; i< NbIndexLR; i++)
    {
      pos1 = IndexLR->V[i][0];
      pos2 = IndexRL->V[i][0];
      if (pos1 == TOP)
	{
	  if (pos2 == TOP)
	    printf("EMPTY       EMPTY\n");
	  else
	    printf("EMPTY         %i\n", pos2);
	}
      else{
	if (pos2 == TOP)
	  printf("%i         EMPTY\n", pos1);
	else
	  printf("%i           %i\n", pos1, pos2);
      }
    }
  for ( ; i< NbIndexRL; i++)
    {
      pos2 = IndexRL->V[i][0];
      if (pos2 == TOP)
	printf("        EMPTY\n");
      else
	printf("          %i\n", pos2);
    }
}

/*******************************************************************************************
       print a list of rays
       and their MinF-values
*******************************************************************************************/
void PrintRayList(int startrow, int endrow, INTMATRIX *Mat)
{
  int *p, i, j;
  unsigned NbRows, NbColumns;
  USI pos;

  fprintf(stdout,"startrow: %d, endrow: %d; Size: %d %d\n", startrow, endrow, NbRows=Mat->R, NbColumns=Mat->C);
  for (i=startrow;i<endrow;i++) {
    p=Mat->V[i];

    printf("%4d", i);

    if(i< IndexRL->R)
      pos = IndexRL->V[i][0];
    else pos = TOP;
    if (pos == TOP)
      printf("____");
    else
      printf("%4d", pos);

    pos = MinF->V[i][0];
      printf("%4d", pos);

    for (j=0;j<NbColumns;j++)
      fprintf(stdout, "%4d", *p++);
      fprintf(stdout, "\n");
  }
} /* PrintRayList */
