/*****************************************************************************

Programm	:	estimts

Komponente	:	estimTS.c
                        Zustandsraumabschaetzung auf Basis
			der Zustandsgleichung (state equation)
			+ Traps und Siphons

Programmer	:	Alexander Huck
Date		:	1999

*****************************************************************************/

#include <malloc.h>
#include <netio.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "struct_types.h"			/* definitions and types */
#include "struct_IO.h"
#include "struct_vars.h"

#include "matrixIO.h"
#include "matrixOps.h"
#include "matrix.h"
#include "timer.h"
#include "estimate.h"

extern int  	minus();
extern sp_place **sparse_function();  

char model_name[FILE_NAME];
int statespace;

/*******************************************************************************************
 output result of estimation to .ese- file
*******************************************************************************************/
static void output_result(double timing)
{
  FILE   *ets_file, *f;
  char	help[1000];
  
  sprintf(help,"%s.ets", model_name);
  if ((ets_file = fopen(help, "w")) == NULL)
    {
      printf("ERROR: could not open %s for writing\n", help);
      exit(1);
    }
  
  
  f = ets_file;		/* output to .ets */
  
  fprintf(f, "Result of estimation (based on state equation with trap/siphon):\n");
  fprintf (f, "Statespace = ");
  if (statespace == -1)
    {
      fprintf(f," oo\n");
    }
  else
    fprintf(f," %i \n", statespace);
  
  fprintf (f, "Time passed with computation: %8.2f s\n", timing );
  fclose(ets_file);

 printf("Result of estimation (based on state equation with trap/siphon):\n");
  printf ("Statespace = ");
  if (statespace == -1)
    {
      printf(" oo\n");
    }
  else
    printf(" %i \n", statespace);
  
  printf ("Time passed with computation: %8.2f s\n", timing );
  

  
}

static void usage()
{
  printf("Usage: estimts netname\n");
  printf("estimates the state space of the net\n");
  exit( -1 );
}

/*******************************************************************************************
 start estimation
*******************************************************************************************/
static void ts()
{
  int pla_nr;
  INTMATRIX *IncidenceMat, *MNull;
  INTMATRIX *PreMat, *PostMat;
  
  start_timer(20);
  
  IncidenceMat=conv_matrix(c_matrix);
  PreMat=conv_matrix(i_matrix);
  PostMat=conv_matrix(o_matrix);
  
  MNull=ZeroMatrix(no_places,1);
  for(pla_nr = 0; pla_nr < no_places; pla_nr++)
    {
      MNull->V[pla_nr][0] = places[pla_nr].m0;
    }
  
  statespace =  EstimateTrapSiphon(MNull,IncidenceMat , PostMat , PreMat);
 
  FreeMat(IncidenceMat);
  FreeMat(PostMat);
  FreeMat(PreMat);
  FreeMat(MNull);
  stop_timer(20);
}


/*	the main procedure handles options and file access and
	calls subordinated procedures */
int main(argc, argv)
	int argc;
	char ** argv;
{
	char net_name[FILE_NAME], struct_name[FILE_NAME], inf_name[FILE_NAME];

	/* check arguments */
	if (argc == 1)
		usage();

	/* build filenames */
	sprintf(net_name,"%s", argv[1]);
	sprintf(struct_name,"%s.STRUCT", net_name);
	sprintf(inf_name,"%s.DEFINFO", net_name);
	sprintf(model_name,"%s", argv[1]);

	/* open STRUCT & DEFINFO files */
	if ((struct_file = fopen(struct_name, "w")) == NULL)
	{	printf("ERROR: could not open %s for writing\n", struct_name);
		exit(1);
	}
	if ((inf_file = fopen(inf_name, "r")) == NULL)
	{	printf("ERROR: could not open %s for reading\n", inf_name);
		exit(1);
	}

	/* begin net analysis */
	printf("\nSTRUCTURAL ANALYSIS ...\n");
	read_net(net_name);
	read_info(inf_file);		/* read additional information */

	if (no_places < 1)
	{	printf("Error: no places.\n");
		exit(1);
	}
	if (no_timed_trans < 1)
	{	printf("Error: no timed transitions.\n");
		exit(1);
	}

#ifdef DEBUG
	printf("I-Matrix:\n");
	print_matrix(i_matrix);
	printf("O-Matrix:\n");
	print_matrix(o_matrix);
	printf("M0:\n");
	for(i=0; i<no_places; i++)
	printf("%d ", places[i].m0);
#endif
					/* calculate the global c-matrix */
	c_matrix = sparse_function(o_matrix, i_matrix, no_trans, minus); 
#ifdef DEBUG
	printf("C-Matrix:\n");
 	print_matrix(c_matrix); 
#endif

	ts();
	output_result(time_passed(20));
	return 0;			/* o.k */
}
