/*****************************************************************************

Programm	:	guesssolutions

Komponente	:	
                        Gueltige Belegungen der Markierungsgleichung raten
			Aufruf mit Netznamen, benoetigt
			zwei Dateien .inc mit Inzidenzmatrix (Stellen x Trans)
			und .m0 mit Anfangsmarkierung

Programmer	:	Alexander Huck
Date		:	1999

*****************************************************************************/

#include <malloc.h>
#include <netio.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "matrixIO.h"
#include "matrixOps.h"
#include "matrix.h"
#include "timer.h"

#define MAXIMUM 2

INTMATRIX *IncidenceMat, *MNull, *Mptr, *Markings;
char model_name[256];
int validmarkings=0;
int recursions=0;
int nbplaces;
int nbtransis;
FILE 		*inc_file, *mnull_file;
char net_name[256];

/*******************************************************************************************
 output result of  to .guess- file
*******************************************************************************************/
static void output_result()
{
  FILE   *guess_file, *f;
  char	help[1000];
  
  sprintf(help,"%s.guess", net_name);
  if ((guess_file = fopen(help, "w")) == NULL)
    {
      printf("ERROR: could not open %s for writing\n", help);
      exit(1);
    }
  
  f = guess_file;		/* output to .guess */
  
  fprintf(f, "Result of guess solutions:\n");
  fprintf (f, "Solutions = ");
  fprintf(f," %i \n", validmarkings);
  FPRINTMATI(guess_file, 0, validmarkings-1, 0, Markings->C-1, Markings);
 
  fclose(guess_file);
  
}

static void usage()
{
  printf("Usage: guessolution netname\n");
  printf("guess soltuions for state equation of a net given by\n");
  printf("untransposed incidence matrix netname.inc\n");
  printf("initial marking netname.m0\n");

  exit( -1 );
}

/*******************************************************************************************
 guess solutions
*******************************************************************************************/


int testsigma(INTMATRIX *Sigma)
{
  int i, j;
  INTMATRIX *Multi;

  /*  
  printf("Sigma vector to be tested: \n ");
  PRINTMATI(0, Sigma->R-1, 0, Sigma->C-1, Sigma);
  */  

  Multi=MultMat(IncidenceMat, Sigma);
  for(i=0; i<nbplaces; i++)
    {
      if(MNull->V[i][0]+Multi->V[i][0]<0)
	break;
    }
  if(i==nbplaces)
    {
      for(i=0; i<validmarkings; i++)
	{
	  for(j=0; j<nbplaces; j++)
	    {
	      if(Markings->V[i][j]!=MNull->V[j][0]+Multi->V[j][0])
		break;
	    }
	  if (j==nbplaces)
	    {
	      //     printf("Marking already contained \n");
	      FreeMat(Multi);
	      return 0;
	    }
	}

      for(i=0; i<nbplaces; i++)
	{
	  Markings->V[validmarkings][i] = MNull->V[i][0]+Multi->V[i][0];
	}
      for(i=0; i<nbtransis; i++)
	{
	  Markings->V[validmarkings][i+nbplaces] = Sigma->V[i][0];
	}
      printf("Found Valid Marking : ");
      for(i=0; i<nbplaces; i++)
	printf("%3i",  Markings->V[validmarkings][i]);
      printf(" | ");
      for(i=0; i<nbtransis; i++)
      	printf("%3i",  Markings->V[validmarkings][i+nbplaces]);
      printf("\n");
      validmarkings++;
      FreeMat(Multi);
      return 1;
    }
  else
    {
      FreeMat(Multi);
      return 0;
    }
}


void rek(INTMATRIX *Sigma, int position, int steps)
{
  int i;
  INTMATRIX *LocalSigma;
  LocalSigma = CopyMat(Sigma);
 
  steps--;
  LocalSigma->V[position][0]++;
 if (steps==0)
    {
      recursions++;
      testsigma(LocalSigma);
      return;
    }

  for(i=position; i<nbtransis; i++)
    {
      rek(LocalSigma, i, steps);
    }
  FreeMat(LocalSigma);
}

static void guess()
{
  INTMATRIX *Sigma;
  int position, steps;
  Sigma = ZeroMatrix(nbtransis, 1);
  Markings = ZeroMatrix(5000, nbplaces+nbtransis);

  for(steps=1; steps<40; steps++)
    {
      recursions = 0;
    for(position=0; position<nbtransis; position++)
      {
	rek(Sigma, position, steps);
      }
    printf("Number of recursions for %i steps: %i \n",steps, recursions);
    }

  FreeMat(Sigma);
  FreeMat(IncidenceMat);
  FreeMat(MNull);
}


/*	the main procedure handles options and file access and
	calls subordinated procedures */
int main(argc, argv)
	int argc;
	char ** argv;
{
  char inc_name[256], mnull_name[256];
	extern int optind;

	/* check arguments */
	if (argc == 1)
		usage();

	/* build filenames */
	sprintf(net_name,"%s", argv[optind]);
	sprintf(inc_name,"%s.inc", net_name);
	sprintf(mnull_name,"%s.m0", net_name);

	if ((inc_file = fopen(inc_name, "r")) == NULL)
	  {	printf("ERROR: could not open %s for reading\n", inc_name);
	  exit(1);
	  }
	if ((mnull_file = fopen(mnull_name, "r")) == NULL)
	  {	
	    printf("ERROR: could not open %s for reading\n", mnull_name);
	    exit(1);
	  }

	IncidenceMat = Matrix_Read(inc_file);
	//IncidenceMat=TransMat(Mptr);
	//FreeMat(Mptr);
	fclose(inc_file);
	MNull = Matrix_Read(mnull_file);
	fclose(mnull_file);

	nbplaces = MNull->R;
	nbtransis = IncidenceMat->C;

	guess();
	output_result();
	return 0;			/* o.k */
}
