/*****************************************************************************

Programm	:	struct

Komponente	:	inequalities.c

                        Loest ein homogenes Ungleichungssystem

Programmer	:	Alexander Huck
Date		:	1998

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <malloc.h>

#include "intOps.h"
#include "matrixOps.h"
#include "matrixProcs.h"
#include "matrixIO.h"
#include "matrix.h"
#include "timer.h"
#include "lists.h"
#include "timer.h"

#include "sharedVars.h"
#include "sharedOps.h"
#include "debugOps.h"
#include "index.h"

#define FSatEntries 2048
//#define USEFASTSAT
//#define SCOREBOARD
//#define DEBUG

/*#define HEAVYDEBUG*/
/*#define SHOWTIME*/
//#define SHOWTIME

/* RaySort -- Sorts the rays (Ray, Sat) into three tiers:
   NbBid       <= i < equal_bound : saturates the constraint
   equal_bound <= i < sup_bound   : verifies the constraint
   sup_bound   <= i < NbRay       : does not verify
   the tier order is chosen so non-verifying rays can be easily removed 
   
NbRay-> bound-> ________
                |  R-  |        R- ==> ray.eq < 0  (outside domain)
        equal-> |------|
                |  R0  |        R0 ==> ray.eq = 0  (on face of domain)
        sup  -> |------|
                |  R+  |        R+ ==> ray.eq > 0  (inside of domain)
            0-> |______|
        
*/
static void RaySort (int *equal_bound, int *sup_bound)
{ 
  static int inf_bound;
  static int eq, sup, inf;
  
  /* sort rays :0 <= i < equal_bound : saturates the constraint
     : equal_bound <= i < sup_bound : verifies the constraint
     : sup_bound <= i < NbRay : does not verify */
  /* the order is chosen so non-verifying rays can be easily removed */
  
  *sup_bound = *equal_bound = 0;
  sup = eq = 0;
  inf_bound = NbRay;
  inf = NbRay;
  
  while (inf_bound>*equal_bound)
    {
      if (Ray->V[sup][0]>0)       /* status = verify */
        {
	  if(eq!=sup)
	    {
	      RowExchange(Ray, eq, sup, raytemprow);
	      RowExchange(Sat, eq, sup, sattemprow);
	      RowExchange(MinF, eq, sup, minftemprow);
	      IndexExchange(eq, sup);
	    }
          (*equal_bound)++; eq++;
          (*sup_bound)++; sup++;
        }
      else
        {
          if (Ray->V[sup][0]<0) {    /* status = not verify */
            inf_bound--; inf--;
	    if(inf!=sup)
	      {
		RowExchange(Ray, inf, sup, raytemprow);
		RowExchange(Sat, inf, sup, sattemprow);
		RowExchange(MinF, inf, sup, minftemprow);
		IndexExchange(inf, sup);
	      }
          }
          else                  /* status = satisfy */
	    {
	      (*equal_bound)++; sup++;
	    }
        }
    }
}

static int AdjacencyTest( USI ray1, USI ray2)
{  
  static USI nbcommonconstraints;
  static USI redundant = 1;
  static USI Word, Word2;
  static int thisword;
  static USI otherRay; 
  static int *ptr1, *ptr2;
  static USI oldray1, oldray2;
  
  adjcounter++;
  if (adjcounter %500000 ==0)
    printf ("Adjacency test called %i times, nb of adj pairs %i, no. quick tests: %i\n", adjcounter, adjcounter - redcounter, quickcounter);

  /*----------------------------------------------------------------------*/
  /* compute and count the set of constraints saturated by both R+ and R- */
  /*----------------------------------------------------------------------*/
  
  nbcommonconstraints=1;
  for (thisword=0; thisword<AdjWCount; thisword++) 
    { 
      BitRow[thisword] = Sat->V[ray1][thisword] | Sat->V[ray2][thisword];/*OR the rows*/
      Word=Word2=BitRow[thisword];
      Word >>= (sizeof(int)/2 * 8);
      Word2 &= 0x0000ffff;
      nbcommonconstraints+=LookUpMat[Word];
      nbcommonconstraints+=LookUpMat[Word2];
    }
  /*last word with index [AdjWCount] has only AdjBCount bits:*/
  Word = Word2 = BitRow[AdjWCount] = Sat->V[ray1][AdjWCount] | Sat->V[ray2][AdjWCount] | AdjBitmask;
  
  Word >>= (sizeof(int)/2 * 8);
  Word2 &= 0x0000ffff;
  nbcommonconstraints+=LookUpMat[Word];
  nbcommonconstraints+=LookUpMat[Word2];
  
  
  /*-----------------------------------------------------------------*/
  /* Adjacency Test : is combination [R-,R+] a non redundant ray?    */
  /*-----------------------------------------------------------------*/
  if (nbcommonconstraints>=Dimension-2) /* dimensionality check*/
    { /* check whether a ray m saturates the same set of constraints */
      
      /*copy the two rays to the top of the ray matrix*/
      if (ray1 == TTEST)
	{
	  oldray1 = ray1;
	  ray1 = ray2;
	  ray2 = oldray1;
	}
      
      oldray2 = ray2;
      RowExchange(Sat, ray2, TTEST, sattemprow);
      ray2 = TTEST;
      
      oldray1 = ray1;
      RowExchange(Sat, ray1, TTEST+1, sattemprow);
      ray1 = TTEST+1;
      
      redundant=0;
      for (otherRay = TTEST + 2; otherRay<NbRay; otherRay++) 
	/* for (otherRay = NbRay-1; otherRay>=TTEST+2; otherRay--) */
	/*for each ray otherRay, otherRay!=ray1, otherRay!=ray2 do*/  
	{
	  /* (r+ r-) is redundant if there does NOT exist an equation */
	  /* which saturates both r+ and r- but not otherRay */
	  ptr1 = BitRow; /*contains (Sat->V[ray2] OR Sat->V[ray1]) */
	  ptr2 = Sat->V[otherRay];
	  for (thisword=0; thisword<=AdjWCount; thisword++,ptr2++,ptr1++)
	    {
	      if (*ptr2&~*ptr1) break;
	    }
	  if (thisword>AdjWCount)
	    {
	      redundant=1;
	      break;
	    }
	}
    }
  else
    {
      quickcounter++;
      redcounter++;
      return 0;
    }
  if (!redundant) 
    { 
      RowExchange(Sat, oldray1, ray1, sattemprow);
      RowExchange(Sat, oldray2, ray2, sattemprow); 
      return 1;
    }
  else 
    { 
      RowExchange(Sat, oldray1, ray1, sattemprow);
      RowExchange(Sat, oldray2, ray2, sattemprow);
      redcounter++;
      return 0;
    }
}


static void  ConditionalStoreEdge(USI ray1, USI ray2, USI k, USI i, LiList *LinkedList)
{
  static USI checked;
  static USI pos1, pos2;
  
  checked = CheckedLater(i, k, ray1, ray2);
  if (checked)
    {
      return;
    }
  pos1 = StoreIndex(ray1);
  pos2 = StoreIndex(ray2);
#ifdef DEBUG
  printf("Ray %i and %i are stored\n", ray1, ray2);
#endif
  StorePair(LinkedList, pos1, pos2);
}


static int StoreEdge(USI ray1, USI ray2, USI k, USI i, LiList *LinkedList)
{
  static USI adjacent, checked;
  static USI pos1, pos2;
  
  checked = CheckedLater(i, k, ray1, ray2);
  if (checked)
    {
      return OK;
    }

  adjacent = AdjacencyTest(ray1, ray2);
  if (adjacent)
    { 
      pos1 = StoreIndex(ray1);
      pos2 = StoreIndex(ray2);
      
#ifdef DEBUG
      printf("Ray %i and %i are stored\n", ray1, ray2);
      PRINTMATI(ray1 , ray1, 0, Ray->C-1,Ray);
      PRINTMATI(ray2 , ray2, 0, Ray->C-1,Ray);
#endif
      if (StorePair(LinkedList, pos1, pos2))
	return MEMORYERROR; /*memory exhausted*/ 
    }
  return OK;
}

static int TestNewRays(LiListElement *ListEl, USI thisrow, USI step, USI offset)
     /*combine adjacent rays*/
{
  static LiListElement *ptr;
  static USI i;
  static USI numels;
  static USI combirow;
  static USI WordNr, BitNr;
  static USI rplus, rminus;
  static USI ck, ckgen, cknew;
  numels= ListEl->NbElements;
  ckgen = MinF->V[thisrow][0];
  
  for (i = 0; i < numels; i++)
    {
      combirow = ListEl->Elements[i] - offset;
      cknew = MinF->V[combirow][0];
#ifdef HEAVYDEBUG
      printf("new ray:, minf = %i: \n", MinF->V[combirow][0]);
      PRINTMATI(combirow, combirow, 0,Ray->C-1, Ray);
#endif
      if (ckgen == cknew)
	{continue;}
      else
	if(ckgen < cknew)
	  {
	    ck = ckgen;
	    rplus = thisrow; rminus = combirow;
	  }
	else
	  {
	    ck = cknew;
	    rplus = combirow; rminus = thisrow;
	  }
      
      WordNr = ck / WSIZE; /*word / bit index of constraint ck in the saturation matrix*/
      BitNr = MSB; BitNr >>= (ck %WSIZE);
      /*MinF->V[rplus][0] < MinF->V[rminus][0] */
      /* part one of ck condition satisfied*/
      /*part two:*/
      /*rminus must not SATISFY the inequality A[ck]:*/
      /*A[ck].rminus!=0*/
      if (BitNr & Sat->V[rminus][WordNr])
	{
#ifdef HEAVYDEBUG
	  printf("ck condition satisfied\n");
#endif
	  if(StoreEdge(rplus, rminus, ck, step, LiListArray[ck]))
	    return MEMORYERROR; /*memory exhausted*/
	}
      else
	{
#ifdef HEAVYDEBUG
	  printf("FAILED SECOND PART\n");
#endif
	}
    }
  ptr = ListEl->next;
  if (ptr != NULL)
    return TestNewRays(ptr, thisrow, step, offset);
  return OK;
}


static int ConstrainR2(USI step)
{
  USI i, j, k, oldpos;
  USI rplus, rminus;
  int equal_bound, sup_bound;
  USI upper_bound;
  int WordNr;
  USI BitNr;
  USI oldrow, thisrow, newray;
  USI numrows;
  USI ck;
  USI offset;
  int low_bound1, low_bound2;

  USI checked;
  USI pos1, pos2;

  USI nbcommonconstraints;
  USI redundant;
  USI Word1, Word2;
  register USI otherRay; 
  register int *ptr1, *ptr2;
  register int thisword;
  USI oldrplus, oldrminus;
  USI rp, rm;

  int WordNr2;
  USI BitNr2;
  
  USI errorflag = 0;

  INTMATRIX *SortMat;
  LiList *GenListStart;

  register USI r0, r1, r2, r3;
  USI r4=0, r5=0, r6=0, r7=0, r8=0;
#ifdef USEFASTSAT
  USI FastSat[FSatEntries];
  USI rowcount, colcount;
#endif
#ifdef SCOREBOARD
  USI *ScoreBoard;
  USI maxScore[10];
  USI scoreredcounter=0;
  USI scoreseek=0;
#endif

#ifdef DEBUG  
  printf("**********ConstrainR2******** %i\n", step);
#endif
  
  /*needed by Combine*/
  WCount = step / WSIZE;
  BCount = MSB; BCount >>= (step %WSIZE);
  
  /*needed by Adjacency Test*/
  AdjWCount = (step + 1) / WSIZE;
  AdjBCount = MSB;
  AdjBCount >>= ((step + 1) %WSIZE);

  AdjBitmask = (AdjBCount - 1); 
  AdjBitmask = AdjBitmask<<1;
  AdjBitmask++;

  /*Sort rays*/
  start_timer (3); /*sorting*/
  RaySort (&equal_bound, &sup_bound);
  stop_timer (3); /*sorting*/
 
  /*
    NbRay-> bound-> ________
                    |  R-  |        R- ==> ray.eq < 0  (outside domain)
            equal-> |------|
                    |  R0  |        R0 ==> ray.eq = 0  (on face of domain)
            sup  -> |------|
                    |  R+  |        R+ ==> ray.eq > 0  (inside of domain)
          0      -> |______|
  */

#ifdef DEBUG
  printf("SortRays\n"); 
  PartitionPrint(stdout,"%4d", Ray, 0, equal_bound, sup_bound, NbRay, Dimension);
  printf("MinF : \n");
  PRINTMATI(0,NbRay - 1,0,MinF->C-1,MinF);
  printf("Total rays: %i\n", NbRay);
  printf("R-  R0  R+ : %i  %i  %i\n", NbRay-equal_bound, equal_bound-sup_bound, sup_bound);
#endif
  
  oldrow = NbRay; /*store position of the last ray*/
  
  if (LiListArray[step]->R!=0) 
    {
      
      /*obtain new rays: they are all in R0*/
#ifdef DEBUG
      printf("Generate new Rays\n");
#endif 
      GenList =  CreateLiList();
      GenListStart=GenList;      

      errorflag = ObtainNewRays(LiListArray[step]); /*generate new rays from the linked list*/
      if (errorflag) return errorflag;
      
  /*
            NbRay->  ______     
	            |R0 new|        R0new ==> ray.eq = 0 (new rays)
           oldrow-> |------|
                    |  R-  |        R- ==> ray.eq < 0  (outside domain)
            equal-> |------|
                    |  R0  |        R0 ==> ray.eq = 0  (on face of domain)
            sup  -> |------|
                    |  R+  |        R+ ==> ray.eq > 0  (inside of domain)
             0   -> |______|
  */
#ifdef DEBUG
      printf("Ray mit %i neuen rays\n", NbRay-oldrow); 
      PRINTMATI(0,NbRay - 1,0,Ray->C-1,Ray);
#endif 
      /*remove R- rays*/
      
      rminus = equal_bound;
      newray = oldrow;
      while ((newray<NbRay)) {
        CopyRow(Ray, newray, rminus);
	CopyRow(Sat, newray, rminus);
        rminus++;
	newray++;
      }
      NbRay=rminus;
      
#ifdef DEBUG
      printf("R- rays removed\n"); 
      PRINTMATI(0,NbRay - 1,0,Ray->C-1,Ray);
#endif
  /*
           oldrow->
            NbRay->  ______     
	            |R0 new|        R0new ==> ray.eq = 0 (new rays)
            equal-> |------|
                    |  R0  |        R0 ==> ray.eq = 0  (on face of domain)
            sup  -> |------|
                    |  R+  |        R+ ==> ray.eq > 0  (inside of domain)
             0   -> |______|
  */

      /* if this is the last constraint, return*/
      if (step == NbConstraints - 1)
	{
	  FreeLiList(GenListStart); /*free list of generators*/
	  return 0;
	}
      /*delete the indices of the R- rays*/
      offset = oldrow - equal_bound;
      
      RemoveIndices(equal_bound, oldrow);
      ResizeIndexRL(equal_bound, NbRay); /*make room in the index list for the new rays*/
      
#ifdef DEBUG
      printf("remove %i rays: \n", oldrow-equal_bound);
#endif
      
      /*for the new rays, evaluate all constraints*/
      EvalAllConstraints(equal_bound, step, NbConstraints); /*build saturation matrix, calculate MinF for new rays*/
      
      /*generator - rays durchgehen und mit generierten vergleichen*/
#ifdef DEBUG
      printf("MinF  for new Rays: \n");
      PRINTMATI(equal_bound,NbRay - 1,0,MinF->C-1,MinF);
      printf("Now new rays and their generators are tested\n");
#endif  
      TTEST=0;
      while(GenList != NULL)
	{
	  numrows = GenList->R;
	  for (i = 0; i < numrows; i++)
	    {
	      thisrow =  GenList->Indices[i];
	      errorflag = TestNewRays(GenList->AdjRows[i], thisrow, step, offset);
	      if(errorflag) return errorflag;
	    }
	  GenList = GenList ->next;
	}
      FreeLiList(GenListStart); /*free list of generators*/
    }
  else
    {/*no new rays*/
      /*throw away R- rays*/
      RemoveIndices(equal_bound, NbRay);
      NbRay = equal_bound;
      /* if this is the last constraint, return*/
      if (step == NbConstraints - 1)
	return 0;
    }
  
  FreeLiList(LiListArray[step]); /*free the linked list of this iteration*/
  LiListArray[step] = NULL;
  
#ifdef SHOWTIME
  printf("ConstrainR: current size of Ray: %i\n", NbRay);
#endif

  if (NbRay == 0) return 3; /*no solutions*/
  
  /*sort rays in R0 by MinF */
  if(sup_bound != NbRay)
    {
      SortRaysMinF(sup_bound, NbRay);
#ifdef DEBUG
      printf("R0 Rays nach MinF sortiert:\n"); 
      PRINTMATI(sup_bound ,NbRay - 1,0,Ray->C-1,Ray);
      /*the rays are now sorted : small ck values to top and big ck values to bottom*/ 
      printf("MinF for R0 rays: \n");
      PRINTMATI(sup_bound,NbRay - 1,0,MinF->C-1,MinF);
#endif
    }
  else
    {
      MinF->V[sup_bound][0] = NbConstraints;
    }
  
  BuildBorder(sup_bound); 
  upper_bound = GetUpperBound(NbConstraints);
  TTEST = sup_bound;
#ifdef DEBUG 
  printf("Zerorays: %i\n",NbRay - sup_bound);
#endif
#ifdef SHOWTIME 
  anz = 0;
  ck = MinF->V[sup_bound][0];
  rplus = sup_bound;
  
  /*calculate the number of tests*/
  while (ck!=NbConstraints)
    {
      rminus = GetBorder(ck);
      if (rminus == -1) break;
      
      anz += (rminus - rplus) * (NbRay - rminus);
      
      rplus = rminus;
      ck = MinF->V[rminus][0];
    }
  

  printf("Anzahl Tests: %i\n", anz);
#endif  

  ck = MinF->V[sup_bound][0];
  rplus = sup_bound;
  
  /*sort saturation matrix*/
  SortMat = ZeroMatrix(NbRay, 4);

  while (ck!=NbConstraints)
    {
      rminus = GetBorder(ck);
      if (rminus == -1) break;
    
      for ( i = rplus; i< rminus; i++)
	{
	  SortMat->V[i][1] = i;
	  for(j=0; j<=AdjWCount; j++)
	    {  
	      Word2 = Sat->V[i][j];
	      Word1 = Word2;
	      Word1 >>= (sizeof(int)/2 * 8);
	      Word2 &= 0x0000ffff;
	      SortMat->V[i][0] += 16 - LookUpMat[Word1];
	      SortMat->V[i][0] += 16 - LookUpMat[Word2];
	    }
	}
      QuickSort(SortMat, rplus, rminus-1, 0, 1);
      /*SortMat[i][0] = sorted values */
      /*SortMat[i][1] = position of new row i*/
      for (i = rplus; i<rminus; i++ )
	{
	  SortMat->V[i][2] = SortMat->V[i][1]; /*copy*/
	  SortMat->V[i][3] = i; /*store old position*/
	}
      QuickSort(SortMat, rplus, rminus-1, 2, 3);
      /*SortMat[i][3] = new position of row i */
      
      for (i = rplus; i<rminus; i++)
	{
	  /*exchange rays, update Index if necessary*/
	  oldpos = SortMat->V[i][1]; /*read out old position*/
	  if(i!=oldpos)
	    {
	      RowExchange(Ray, i, oldpos, raytemprow);
	      RowExchange(Sat, i, oldpos, sattemprow);
	      IndexExchange(i, oldpos);
	      SortMat->V[SortMat->V[i][3]][1] = oldpos; 
	      SortMat->V[oldpos][3] = SortMat->V[i][3];
	    }
	}
      rplus = rminus;
      ck = MinF->V[rminus][0];
    }
  
  FreeMat(SortMat);
 
#ifdef SHOWTIME	
  anztest=0;
#endif
  
  ck = MinF->V[sup_bound][0];/*first ck value*/
  low_bound1 = sup_bound; /*lower bound*/
  
#ifdef SCOREBOARD
  ScoreBoard = (USI *) malloc (Sat->R*sizeof(USI));
  for(i=0; i< Sat->R; i++)
    ScoreBoard[i] = 0;
  scoreseek = 0;
#endif

#ifdef USEFASTSAT
  /*FastSat holds the first entries of the Sat matrix 
    stored contiguously*/
  
  for(i=0, rowcount=TTEST + 2; i<FSatEntries && rowcount<NbRay; rowcount++)
    {
      if (i+AdjWCount+1>=FSatEntries) break;
      for(colcount=0; colcount<=AdjWCount; colcount++)
	FastSat[i++] = Sat->V[rowcount][colcount];	
    }
  rowcount-=(TTEST + 2);
  rowcount*=(AdjWCount+1);

  printf("AdjWCount, rowcount, elements, NbRay, i %i %i %i %i %i\n", AdjWCount, rowcount, rowcount/(AdjWCount+1), NbRay - TTEST -2, i);
#endif

  while (ck!=NbConstraints)
    {
      WordNr = ck / WSIZE; /*word / bit index of constraint ck in the saturation matrix*/
      BitNr = MSB; BitNr >>= (ck %WSIZE);
      
      k = ck - 1;
      
      BitNr2 = BitNr<<1;
      if (!(BitNr2))
	{
	  BitNr2 = 1;
	  WordNr2 = WordNr - 1;
	}
      else
	{
	   WordNr2 = WordNr;
	}
      
      for (thisword = WCount+1; thisword<=WordNr2; thisword++) {
	BitRow2[thisword] = -1; /*fill with 1111111111111...*/
      }
     
      /*BitRow2[WordNr] = BCount - 1; falsch!*/
      BitRow2[WCount] = BCount - 1;

      for (thisword = WCount; thisword< WordNr2; thisword++)
	BitRow3[thisword] = 0;
      
      BitRow3[WordNr2] = BitNr2 - 1;
      
      low_bound2 = GetBorder(ck);
      if (low_bound2 == -1) break;
      
      for (rplus=low_bound1; rplus<low_bound2; rplus++)     /* for all rays in R0*/
	{
	  for(rminus = low_bound2; rminus<NbRay; rminus++)
	    {
#ifdef SHOWTIME	
	      anztest++;
	      if (anztest %500000 ==0)
		{
		  totaltests += 500000;  
		  printf("Total of %i tests\n", totaltests);
		  printf ("Test no. %i of %i, sec.part: %i, checkedlater: %i\n", anztest, anz, secondpart, checkedlater);
		}
#endif
	      /*else :  MinF->V[rplus][0] < MinF->V[rminus][0] (the rays are sorted) */
	      /* part one of ck condition satisfied*/
	      /*part two:*/
	      /*rminus must not SATISFY the inequality A[ck]:*/
	      /*A[ck].rminus!=0*/
	      if (BitNr & Sat->V[rminus][WordNr])
		{
#ifdef SHOWTIME
		  secondpart++;
		  
		  adjcounter++;
		  if (adjcounter %500000 ==0)
		    {
		    printf ("Adjacency test called %i times, nb of adj pairs %i, no. quick tests: %i\n", adjcounter, adjcounter - redcounter, quickcounter);
		  printf("Average no. of rays scanned for 1 redundant ray: %i \n", seekray / (adjcounter - quickcounter)); 
		    }
#endif
		  /*----------------------------------------------------------------------*/
		  /* compute and count the set of constraints saturated by both R+ and R- */
		  /*----------------------------------------------------------------------*/

		  nbcommonconstraints=1;
		  for (thisword=0; thisword<AdjWCount; thisword++) 
		    { 
		      AdjBitRow[thisword] = Sat->V[rplus][thisword] | Sat->V[rminus][thisword];/*OR the rows*/
		      Word2=AdjBitRow[thisword];
		      Word1=Word2;
		      Word1 >>= (sizeof(int)/2 * 8);
		      Word2 &= 0x0000ffff;
		      nbcommonconstraints+=LookUpMat[Word1];
		      nbcommonconstraints+=LookUpMat[Word2];
		    }
		  /*last word with index [AdjWCount] has only AdjBCount bits:*/
		  AdjBitRow[AdjWCount] = Sat->V[rplus][AdjWCount] | Sat->V[rminus][AdjWCount] | AdjBitmask;
		  Word2 = AdjBitRow[AdjWCount];
		  Word1 = Word2;
		  
		  Word1 >>= (sizeof(int)/2 * 8);
		  Word2 &= 0x0000ffff;
		  nbcommonconstraints+=LookUpMat[Word1];
		  nbcommonconstraints+=LookUpMat[Word2];

		  /*-----------------------------------------------------------------*/
		  /* Adjacency Test : is combination [R-,R+] a non redundant ray?    */
		  /*-----------------------------------------------------------------*/
		  if (nbcommonconstraints<Dimension-2) /* dimensionality check*/
		    {
#ifdef SHOWTIME	
		      quickcounter++;
		      redcounter++;
#endif
		      continue;
		    }
		  else
		    { /* check whether a ray m saturates the same set of constraints*/
		      
		      checked = 0;		      
		      if(step<k)
			{
			  for (thisword=WCount; thisword<=WordNr2; thisword++) 
			    /*for (thisword=WordNr2; thisword>=WCount; thisword--) */
			    { 
			      BitRow[thisword] = ~(Sat->V[rplus][thisword] | Sat->V[rminus][thisword]);/*NOT(OR the rows)*/
			      if ( BitRow[thisword] & (BitRow3[thisword] ^ BitRow2[thisword]))/*AND(XOR) the rows*/
				{
#ifdef SHOWTIME	
				  checkedlater++;
#endif
				  checked = 1;
				  break;
				}
			    }
			}
		      
		      if (checked) continue;
		      
		      rp = rplus;
		      rm = rminus;
    
		      /*copy the two rays to the top of the ray matrix*/
		      if (rplus == TTEST)
			{
			  oldrplus = rplus;
			  rplus = rminus;
			  rminus = oldrplus;
			}
		      
		      
		      oldrminus = rminus;
		      RowExchange(Sat, rminus, TTEST, sattemprow);
		      rminus = TTEST;
		      
		      oldrplus = rplus;
		      RowExchange(Sat, rplus, TTEST+1, sattemprow);
		      rplus = TTEST+1;
		      
#ifdef USEFASTSAT
		      /*FastSat may not contain rplus/rminus
			replace by TTEST TTEST+1 */
	
		      i = (oldrminus - (TTEST +2))*(AdjWCount+1);
		      if(i<rowcount && i>=0)
			{
			  for(colcount=0; colcount<=AdjWCount; colcount++)    
			    FastSat[i++] = Sat->V[oldrminus][colcount];	    
			}
		      i = (oldrplus - (TTEST +2))*(AdjWCount+1);
		      if(i<rowcount && i>=0)
			{
			  for(colcount=0; colcount<=AdjWCount; colcount++)
			    FastSat[i++] = Sat->V[oldrplus][colcount];
			}
#endif
		      
		      redundant=0;
#ifdef SHOWTIME	
		      seek = 0;
#endif

		     
		      if(AdjWCount==0)
			{
			  r4++;
#include "adjtest_macro0.h"
			} 
		      else if (AdjWCount==1)
			{
			  r5++;
#include "adjtest_macro1.h"	  
			}
		      else if (AdjWCount==2)
			{
			  r6++;
#include "adjtest_macro2.h"	  
			}
		      else if (AdjWCount==3)
			{
			  r7++;
#include "adjtest_macro3.h"	  
			}
		      else
			{
			  r8++;
#include "adjtest_macro.h"
		 	}
		      
		      //#include "adjtest_macro.h"
 

		    }
		  
		  if (!redundant) 
		    {
		 
#ifdef USEFASTSAT
		      /*restore entries in FastSat*/
		      
		      i = (oldrminus - (TTEST +2))*(AdjWCount+1);
		      if(i<rowcount && i>=0)
			{
			  for(colcount=0; colcount<=AdjWCount; colcount++)
			    FastSat[i++] = Sat->V[rminus][colcount];
			}

		      i = (oldrplus - (TTEST +2))*(AdjWCount+1);
		       if(i<rowcount && i>=0)
			{
			  for(colcount=0; colcount<=AdjWCount; colcount++)
			    FastSat[i++] = Sat->V[rplus][colcount];
			}
#endif
		       RowExchange(Sat, oldrplus, rplus, sattemprow);
		       RowExchange(Sat, oldrminus, rminus, sattemprow);
		      
		       rplus = rp; rminus = rm;
		    }
		  else 
		    { 
		    is_redundant:
#ifdef USEFASTSAT
		      /*restore entries in FastSat*/
		      
		      i = (oldrminus - (TTEST +2))*(AdjWCount+1);
		      if(i<rowcount && i>=0)
			{
			  for(colcount=0; colcount<=AdjWCount; colcount++)
			    FastSat[i++] = Sat->V[rminus][colcount];
			}

		      i = (oldrplus - (TTEST +2))*(AdjWCount+1);
		       if(i<rowcount && i>=0)
			{
			  for(colcount=0; colcount<=AdjWCount; colcount++)
			    FastSat[i++] = Sat->V[rplus][colcount];
			}
#endif
		       RowExchange(Sat, oldrplus, rplus, sattemprow);
		       RowExchange(Sat, oldrminus, rminus, sattemprow); 

		       rplus = rp; rminus = rm;
#ifdef SHOWTIME	
		      redcounter++;
#endif
		      continue;
		    }
		  pos1 = StoreIndex(rplus);
		  pos2 = StoreIndex(rminus); 
		  if(StorePair( LiListArray[ck], pos1, pos2))
		    return MEMORYERROR; /*memory exhausted*/
		}
	    }
	}

#ifdef SCOREBOARD
  for(i=0; i< 10; i++)
    maxScore[i]=0;
  for(i=0; i< Sat->R; i++)
    {
      for(j=0; j<10; j++)
	{
	  if(ScoreBoard[i]>ScoreBoard[maxScore[j]])
	    {
	      for(k=9; k>j; k--) 
		maxScore[k]=maxScore[k-1];
	      maxScore[j]=i;
	      break;
	    }
	}
    }
  printf("Maxscore: TTEST=%i lowbound1=%i lowbound2=%i \n", TTEST, low_bound1, low_bound2);
  printf("Number of adj. Tests since last output: %i \n", scoreseek);
  printf("Number of red. Pairs since last output: %i \n", scoreredcounter);
  if(scoreredcounter)
  printf("Average no. of rays scanned for 1 redundant ray: %i \n", 
	 (scoreseek)/(scoreredcounter));   

  for(j=0; j<10; j++)
    scoreseek-= (maxScore[j] - TTEST - 2) * ScoreBoard[maxScore[j]];
  
  printf("Moegliche Ersparnis: \n");
  printf("Number of adj. Tests since last output: %i \n", scoreseek);
  printf("Number of red. Pairs since last output: %i \n", scoreredcounter);
  if(scoreredcounter)
  printf("Average no. of rays scanned for 1 redundant ray: %i \n", 
	 (scoreseek)/(scoreredcounter));

  scoreseek = 0;
  scoreredcounter=0;

  for(j=0; j<10; j++)
    printf(" %i : %i \n", maxScore[j], ScoreBoard[maxScore[j]]);
 for(i=0; i< Sat->R; i++)
    ScoreBoard[i] = 0;
#endif

      low_bound1 = low_bound2;
      ck = MinF->V[low_bound2][0]; 
    }

  // printf("%i %i %i %i %i \n", r4, r5, r6, r7, r8);

#ifdef SCOREBOARD
  for(i=0; i< 10; i++)
    maxScore[i]=0;
  for(i=0; i< Sat->R; i++)
    {
      for(j=0; j<10; j++)
	{
	  if(ScoreBoard[i]>ScoreBoard[maxScore[j]])
	    {
	      for(k=9; k>j; k--) 
		maxScore[k]=maxScore[k-1];
	      maxScore[j]=i;
	      break;
	    }
	}
    }
  printf("Maxscore: TTEST=%i lowbound1=%i lowbound2=%i \n", TTEST, low_bound1, low_bound2);
  printf("Number of adj. Tests since last output: %i \n", scoreseek);
  printf("Number of red. Pairs since last output: %i \n", scoreredcounter);
  if(scoreredcounter)
  printf("Average no. of rays scanned for 1 redundant ray: %i \n", 
	 (scoreseek)/(scoreredcounter));   

  for(j=0; j<10; j++)
    scoreseek-= (maxScore[j] - TTEST - 2) * ScoreBoard[maxScore[j]];
  
  printf("Moegliche Ersparnis: \n");
  printf("Number of adj. Tests since last output: %i \n", scoreseek);
  printf("Number of red. Pairs since last output: %i \n", scoreredcounter);
  if(scoreredcounter)
  printf("Average no. of rays scanned for 1 redundant ray: %i \n", 
	 (scoreseek)/(scoreredcounter));

  scoreseek = 0;
  scoreredcounter=0;

  for(j=0; j<10; j++)
    printf(" %i : %i \n", maxScore[j], ScoreBoard[maxScore[j]]);

  free(ScoreBoard);
#endif

  return 0;
}



static void AugmentR(int rnew, USI step)
{ 
  USI i;
#ifdef DEBUG
  printf("***********AugmentR********\n");
#endif
  for( i = NbLines + 1 ; i<NbRay; i++)
    {
      if(Ray->V[i][0] != 0)
	{
#ifdef DEBUG
	  printf("combine row %i with row %i\n", i, rnew);
#endif
	  SimpleCombine(i, rnew, i, Ray);
	}      
    }
}


static void InitList(USI thisConstraint)
{
  USI nbcommonconstraints;
  USI rayonly, redundant;
  USI Word, thisbit, thisword;
  USI rplus, rminus, otherRay;
  USI bound;
  USI ck;
  USI WordNr, BitNr;
  int *ptr1, *ptr2;
  USI bitmask;
  
  USI wordcount = NbConstraints / WSIZE;
  USI bitcount = MSB;
  bitcount >>= (NbConstraints % WSIZE);

  bitmask = (bitcount - 1); 
  bitmask = bitmask<<1;
  bitmask++;
 
#ifdef DEBUG
  printf("InitList\n");
#endif
  /*compute complete saturation matrix for all rays and all constraints*/
  EvalAllConstraints(0, 0, NbConstraints);

#ifdef DEBUG
  printf("Saturation matrix filled: \n");
  printsat(Sat, 0, NbRay - 1);
  printf("MinF: \n");
  PRINTMATI(0,MinF->R-1,0,MinF->C-1,MinF);
#endif

  /*perform adjacency test for all pairs of rays */
  bound = NbRay;

#ifdef DEBUG
  printf("Adjacency Test\n");
#endif
 for (rplus=0; rplus<NbRay; rplus++)     /* for all pairs of R+*/
   for(rminus=rplus + 1; rminus<NbRay; rminus++) {        /* for all pairs of R-*/
#ifdef DEBUG
     printf("Betrachte Reihen %i und %i:\n", rplus, rminus);
#endif
     /*----------------------------------------------------------------------*/
     /* compute and count the set of constraints saturated by both R+ and R- */
     /*----------------------------------------------------------------------*/
     nbcommonconstraints=0;
     for (thisword=0; thisword<wordcount; thisword++) 
       { 
	 BitRow[thisword] = Sat->V[rplus][thisword] | Sat->V[rminus][thisword];/*OR the rows*/
	 Word=BitRow[thisword];
	 for (thisbit=MSB; thisbit!=0; thisbit>>=1)
	   {/*scan the bits of each word*/
	     if (!(Word&thisbit)) nbcommonconstraints++;
	   }
       }
     /*last word with index [wordcount] has only bitcount bits:*/
     Word = BitRow[wordcount] = Sat->V[rplus][wordcount] | Sat->V[rminus][wordcount] | bitmask;
     for (thisbit=MSB; thisbit!=bitcount; thisbit>>=1)
       {
	 if (!(Word&thisbit)) nbcommonconstraints++;
       }     
     if (Ray->V[rplus][Dimension]==0 && Ray->V[rminus][Dimension]==0)
       nbcommonconstraints++;      /* account for pos constr */
     /*-----------------------------------------------------------------*/
     /* Adjacency Test : is combination [R-,R+] a non redundant ray?    */
     /*-----------------------------------------------------------------*/
     if (nbcommonconstraints>=Dimension-2) /* dimensionality check*/
       { /* check whether a ray m saturates the same set of constraints */
	 redundant=0;
	 rayonly  =(Ray->V[rplus][Dimension]==0 && Ray->V[rminus][Dimension]==0); /*no vertices*/
	 for (otherRay=0; otherRay<bound; otherRay++) if ((otherRay!=rplus)&&(otherRay!=rminus))
	   /*for each ray otherRay, otherRay!=rplus, otherRay!=rminus do*/  
	   {
	     /* 2 rays (r+ r-) are never made redundant by a vertex   */
	     /* because the positivity constraint saturates both rays */
	     /* but not the vertex                                    */
	     if (rayonly && Ray->V[otherRay][Dimension]!=0) continue;
	     
	     /* (r+ r-) is redundant if there does NOT exist an equation */
	     /* which saturates both r+ and r- but not otherRay */
	     ptr1 = BitRow; /*contains (Sat->V[rminus] OR Sat->V[rplus]) */
	     ptr2 = Sat->V[otherRay];
	     for (thisword=0; thisword<=wordcount; thisword++,ptr2++,ptr1++)
	       {
		 if (*ptr2&~*ptr1) break;
	       }
	     if (thisword>wordcount)
	       {
		 redundant=1;
		 break;
	       }
	   }
	 if (!redundant)
	   {/*found adjacent pair*/
	     /*check ck condition*/
#ifdef DEBUG
	     printf("Found adjacent pair: %i %i\n", rplus, rminus);
#endif
	     if (MinF->V[rplus][0] < MinF->V[rminus][0])
	       {
		 ck = MinF->V[rplus][0];
		 WordNr = ck / WSIZE;
		 BitNr = MSB; BitNr >>= (ck %WSIZE);
		 /*rminus must VERIFY not SATISFY the inequality A[ck]:*/
		   /*A[ck].rminus>0*/
		 if (BitNr & Sat->V[rminus][WordNr])
		     {
#ifdef DEBUG
		       printf("ck condition satisfied\n");
#endif
		       ConditionalStoreEdge(rplus, rminus, ck, thisConstraint, LiListArray[ck]);
		     }
	       }
	    if (MinF->V[rplus][0] > MinF->V[rminus][0])
	       {
		 ck = MinF->V[rminus][0];
		 WordNr = ck / WSIZE;
		 BitNr = MSB; BitNr >>= (ck %WSIZE);
		 /*rplus must VERIFY not SATISFY the inequality A[ck]:*/
		   /*A[ck].rminus>0*/
		 if (BitNr & Sat->V[rplus][WordNr])
		     {
#ifdef DEBUG
		       printf("ck condition satisfied\n");
#endif	
		       ConditionalStoreEdge(rminus, rplus, ck, thisConstraint, LiListArray[ck]);
		     }
	       }
	   }
       }
   }

}

static int BuildBasis()
{
  int *BASIS;
  USI counter =0;
  int i, j, k;
  USI step, rnew;

/*build basis*/
 BASIS = (int *)malloc(NbConstraints * sizeof(int));
 for (step=0; step<NbConstraints; step++)
   {
     BASIS[step] = 1;
   }
  for (step=0; step<NbConstraints; step++)
  {
     EvalConstraint(Constraints->V[step]); /*evaluate this constraint*/
    
     /*check if it slices the lineality space*/
     rnew = ConstrainL(step);
     if(rnew==-1)
       {
	 /*this constraint did not slice the lineality space -
	   select another constraint*/
	 BASIS[step] = 0; /*mark as unused*/
       }
     else
       {
	 BASIS[step] = 1; /*used to build the basis*/
	 AugmentR(rnew, step);
	 counter++;
	 /*	printf("AugmentCount: %i", counter);
		printf("Number of Lines: %i\n", NbLines);*/
	 
       }
    if (NbLines == 0) break; /*no more lines - basis built*/
  }
 
  if (NbLines !=0)
    {
      printf("DUAL(): no basis found!\n");
      return 1;
    }

  /*basis built, now sort the constraints
    constraints used for building the basis 
    are sorted to the top of the constraint - matrix*/
 
  for (i = 0; i < NbConstraints; i++)
    {
      if (!BASIS[i])
	{
	  for (j = i; j < NbConstraints; j++)
	    {
	      if(BASIS[j]) break;
	    }
	  if (j == NbConstraints) break;
	  /*slide up basis row*/
	  for(k=j; k>i; k--)
	    RowExchange( Constraints, k, k-1, raytemprow );
	      
	  BASIS[i] = 1;
	  BASIS[j] = 0;
	}
    }



#ifdef DEBUG
  printf("Basis built\n");
  printf("reordered constraints: \n");
  PRINTMATI(0,NbConstraints - 1,0,Constraints->C-1,Constraints);
#endif
  free(BASIS);
  return counter;
}


int Dual ()
{
  USI i;
  USI result;
  USI errorflag = 0;
  USI step, counter;

#ifdef DEBUG
  FreeLi *ptr;
  USI flrsize = 0;
#endif

#ifdef DEBUG
  printf("Dual\n");
#endif

  Dimension     = Constraints->C - 1;	/* dimension = number of variables = columns - 1  */
  NbConstraints = Constraints->R; /*number of constraints*/
  maxMinF = NbConstraints; /*fix maxmin infeasible value */
  nc  = ( NbConstraints-1 ) / (sizeof(int)*8) + 1; /*number of words needed when each constraint is one bit*/

  /*temporary rows needed for copy and compare operations*/
  AdjBitRow=(int *)malloc(nc * sizeof(int));
  BitRow=(int *)malloc(nc * sizeof(int)); 
  BitRow2=(int *)malloc(nc * sizeof(int));
  BitRow3=(int *)malloc(nc * sizeof(int));


  /*used for storing the starting row of different ck-values*/
  Border=(int *)malloc((NbConstraints+1) * sizeof(int));
  /*Lookup-table used for adjacency testing*/
  CreateLookUpMat();

  NbLines = Init(); /*initialize the Rayspace /Saturation matrix*/
  NbRay = NbLines + 1;

#ifdef DEBUG
  printf("Init Rayspace /Sat:\n");
  PRINTMATI(0,Ray->R-1,0,Ray->C-1,Ray);
#endif  

  MinF = ZeroMatrix(NbRay, 4); /*MinF matrix holds the minimum infeasible index for each ray*/
  /*for translating LinkedList -> Ray*/
  IndexLR = ZeroMatrix(NbRay, 2); 
  NbIndexLR = 0;
  /*for translating Ray -> LinkedList*/
  IndexRL = ZeroMatrix(NbRay, 2);
  NbIndexRL = NbRay;
  for ( i = 0; i< NbRay; i++)
    {
      IndexRL->V[i][0] = TOP; /*mark as empty*/
    }
  
  /*temporary rows used for copy operations*/
  raytemprow = (int *)malloc((Ray->C)  * sizeof(int));
  sattemprow = (int *)malloc((Sat->C)  * sizeof(int));
  minftemprow = (int *)malloc((MinF->C)  * sizeof(int));
  
  /*initialize the linked list array*/
  LiListArray = (LiList **)malloc(NbConstraints * sizeof(LiList *)); 
  
  for(i = 0; i < NbConstraints; i++)
    {
      LiListArray[i] = CreateLiList();
    }
  
  /*list for the first free index entry*/
  FLR = InitFreeList();
  blockflag = 1; /*memory block available*/
  
  counter = BuildBasis();
  
  /*initialize the linked lists*/
  InitList(counter - 1);
  
#ifdef DEBUG
  printf("LiLists ausgeben\n");
  for(i=0; i< NbConstraints; i++)
    {
      if (LiListArray[i] !=NULL)
	if(LiListArray[i]->R != 0){
	  printf("Liste Nr. %i\n", i);
	  printLiList(LiListArray[i]);
	}
    }
  PRINTMATI(0,NbRay - 1,0,Ray->C-1,Ray);
#endif
  
  NbRay--; /*remove positivity constraint*/
  
  /*start algorithm*/
  for (step = counter; step<NbConstraints; step++)
    { 


#ifdef SHOWTIME
      printf("##############################\n");
      printf("     Constraint Nr. %i of %i\n", step, NbConstraints-1);
      printf("##############################\n");
      
      for (j=0; j<=Dimension; j++) printf("%i ", Constraints->V[step][j]);
      
      printf("\n");
      printf("\n");
#endif
#ifdef SHOWTIME
      printf ("Time passed with computation: %8.2f s\n", time_passed(0) );
      printf ("Time spent in init phase: %8.2f s\n", time_passed(1) );
      printf ("Time spent with adjacency testing: %8.2f s\n", time_passed(2) );
      printf ("Time spent with sorting: %8.2f s\n", time_passed(3) );
      printf ("Time spent in combine: %8.2f s\n", time_passed(4) );
      printf ("Time spent with evaluation : %8.2f s\n", time_passed(5) );
      printf ("Time spent with list administration: %8.2f s\n", time_passed(6) );
      printf ("Time for testing the new rays : %8.2f s\n", time_passed(7) );
      printf ("Time for checked later : %8.2f s\n", time_passed(8) );
      printf ("Time in ConstrainR2 : %8.2f s\n", time_passed(9) );
      printf ("Schleife in ConstrainR2 : %8.2f s\n", time_passed(10) );
      printf ("StoreEdge : %8.2f s\n", time_passed(11) );
#endif   
      
      EvalConstraint(Constraints->V[step]); /*evaluate this constraint*/  
      result = ConstrainR2(step);
      
      if (result)
	{
	  printf("Error!");
	  errorflag = 1;
	  break;
	}
      
#ifdef DEBUG
      printf("Aktive Listen\n");
      for(i=step; i< NbConstraints; i++)
	{
	  if (LiListArray[i] !=NULL)
	    if(LiListArray[i]->R != 0){
	      printf("Liste Nr. %i\n", i);
	      ScanList(LiListArray[i], &entrys, &elems);
	      printf("Eintraege : %i   Elemente: %i\n", entrys, elems);
	    }
	}
      ptr = FLR;
      flrsize = 0;
      while (ptr!=NULL)
	{
	  flrsize += maxElementsFL;
	  ptr = ptr->next;
	}
      printf("Size of FreeList: %i\n", flrsize);
#endif
      
#ifdef DEBUG
      PRINTMATI(0,NbRay - 1,0,Ray->C-1,Ray);
#endif
    }
  /*printf("\n\n FERTIG!\n");*/
  
  PartFreeMat(&Ray, NbRay);
  PartFreeMat(&Sat, NbRay);
  FreeMat(MinF);
  FreeMat(IndexLR);
  FreeMat(IndexRL);
  for(i=0; i< NbConstraints; i++)
    {
      if (LiListArray[i] !=NULL)
	FreeLiList(LiListArray[i]);
    }
  free(LiListArray);
  DelFreeList(FLR);
  free(AdjBitRow);
  
  free(BitRow);
  free(BitRow2);
  free(BitRow3);
  
  free(Border);
  free(raytemprow);
  free(sattemprow); 
  free(minftemprow);
  
  free(LookUpMat);
#ifdef DEBUG
  printf("Result:\n");
  printf("% i Solutions:\n", NbRay);
  PRINTMATI(0,Ray->R-1,0,Ray->C-1,Ray);
#endif
  if(errorflag)
    return 1;
  else
    return 0;
}


int SolveInequationSystem(INTMATRIX **Solution, INTMATRIX *Constr, unsigned NbMaxR)
{
  
  USI numsol;
  USI errorflag;
  
  NbMaxRays = NbMaxR; /*maximum size of ray matrix*/
  Constraints = Constr; /*constraint matrix*/
  
  reset_all_timer ();
  start_timer (0);
  
  errorflag = Dual(); /*compute ray matrix from constraints*/
  if (errorflag) 
    {
      printf("an error occured\n");
      FreeMat(Ray);
      FreeMat(Sat);
      return 0;
    }
  
  /* printf("Size of pure ray matrix: %i\n", Ray->R);*/
  numsol=CountRays(Ray); 
  /*   printf("numsol: %i \n", numsol);*/
  (*Solution) = ExtractRays(Ray, numsol); /*delete status information*/
  
  FreeMat(Ray);
  FreeMat(Sat);
  
  stop_timer (0);
  
  /*printf ("Time passed with computation: %8.2f s\n", time_passed(0) );*/
#ifdef SHOWTIME
  printf ("Time passed with computation: %8.2f s\n", time_passed(0) );
  printf ("Time spent in init phase: %8.2f s\n", time_passed(1) );
  printf ("Time spent with adjacency testing: %8.2f s\n", time_passed(2) );
  printf ("Time spent with sorting: %8.2f s\n", time_passed(3) );
  printf ("Time spent in combine: %8.2f s\n", time_passed(4) );
  printf ("Time spent with evaluation : %8.2f s\n", time_passed(5) );
  printf ("Time spent with list administration: %8.2f s\n", time_passed(6) );
  printf ("Time for testing the new rays : %8.2f s\n", time_passed(7) );
  printf ("Time for checked later : %8.2f s\n", time_passed(8) );
  printf ("Time in ConstrainR2 : %8.2f s\n", time_passed(9) );
  printf ("Schleife in ConstrainR2 : %8.2f s\n", time_passed(10) );
  printf ("StoreEdge : %8.2f s\n", time_passed(11) );
  
  printf ("Adjacency test called %i times, nb of not adj. pairs: %i\n", adjcounter, redcounter);
#endif
  
  return numsol;
}
