/*****************************************************************************

Programm	:	struct

Komponente	:	intOps.c

                        Integer-Operationen, die von hermite2.c und inequalities.c 
			benoetigt werden

Programmer	:	Alexander Huck
Date		:	1998

*****************************************************************************/

#include "matrix.h"
#include <malloc.h>


/*******************************************************************************************
      vergleicht die absoluten Betrge von a und b,
       gibt fr 	
       |a|>|b|  =>1
       |a|=|b|  =>0
       |a|<|b|  =>-1
*******************************************************************************************/
int RSV(int a, int b)
{
  a=abs(a);
  b=abs(b);
  if (a>b) return 1;
  if (a==b) return 0;
  return -1;
} /*RSV*/

/*******************************************************************************************
    INT(a/b)
    examples:
    INT(1, 3)= 0
    INT(-1, 3)= -1 
*******************************************************************************************/
int INT(int a, int b)
{
  int divab=a/b;
  int negflag=(a*b<0);
  int modab=abs(a)%abs(b);
  if (negflag && (modab!=0 || divab==0) )
    {
      return divab-1;
    }
  return divab;
} /*INT*/

/*******************************************************************************************
    NEAREST_INT(a/b)
    examples:
    NEAREST_INT(1, 3)= 0
    NEAREST_INT(-1, 3)= 0
*******************************************************************************************/
int  NEAREST_INT(int a, int b)
{
  int y=INT(a,b);
  int x=b*y;
  int z=a-x;
  z=z*2;
  if (RSV(z, b) == 1)                       
    {             
      y=y+1;                                                          
    }                                         
  return y;                          
} /*NEAREST_INT*/

/*******************************************************************************************
    greatest common divisor of a and b
*******************************************************************************************/
int Gcd(int a, int b)
{
  int aux;
  
  while (a) { aux=b%a; b=a; a=aux; }
  return (b<0 ? -b : b);
} /* Gcd */

/*******************************************************************************************
  greatest common divisor of all numbers
  in vector p
*******************************************************************************************/
int VectorGcd(int *p, USI length)
{
  int a=0, b, i, j;
  for(i=1; i<length;i++)
    {
      a=p[i];
      if(a) break;
    }
  a=abs(a);
  if (a==1) return 1;
  for(j=i+1; j<length; j++)
    {
      b=p[j];
      if (b) a=Gcd(a,abs(b));
      if (a==1) return 1;
    }
  return a;
} /*VectorGcd*/

/*******************************************************************************************
  gcd reduce a vector 
*******************************************************************************************/
void Vector_Normalize(int *p, USI length)
{
  int g, i;
  g = VectorGcd( p, length);	/* get gcd of all elements in that row*/
  if (!g) return;
  if ( g!=1 ) for (i=1; i<length; i++) p[i]/=g;		/* normalize */
} /* Vector_Normalize */

