/*****************************************************************************

Programm	:	struct

Komponente	:	invars.c
                        Bestimmt die P-Invarianten (semi-flows) fuer gegebene Inzidenzmatrix
			Dazu wird zunaechst durch Berechnung der
			Hermite-Normalform eine Basis fuer die flows bestimmt.
			Um auf die semi-flows zu kommen, wird ein Ungleichungssystem 
			geloest, das aus der Basis fuer die flows besteht


Programmer	:	Alexander Huck
Date		:	1998

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "matrix.h"
#include "matrixIO.h"
#include "matrixOps.h"
#include "matrixProcs.h"
#include "hermite2.h"
#include "inequalities.h"
#include "timer.h"

#define MAXROWSIZE 1000000   /*Maximum number of matrix-rows allowed during inequality solving*/
/*#define DEBUG*/

/*******************************************************************************************
      Invariants: calculate invariants of a net
      from incidence matrix A
      invariants are returned in the columns of Invars
      returns number of invariants
*******************************************************************************************/
int INVARIANTS(INTMATRIX **INVARS, INTMATRIX *A)
{
  INTMATRIX *TRANSMAT, *ROWECH, *ZBASE, *INEQ, *SOLUTION;
  USI numrows, numcols, numsol;
  int rank;
  USI i,j;
  double zeit;
#ifdef DEBUG
 char filename[50], filename2[50], filename3[50];
  FILE *outfile;
  FILE *outfile2;
  FILE *outfile3;
#endif
  
#ifdef DEBUG
  strcpy(filename, "test.out");
  printf("filename : %s\n", filename);
  outfile = fopen(filename, "w");
  
  strcpy(filename2, "zbasis.out");
  printf("filename : %s\n", filename2);
  outfile2 = fopen(filename2, "w");
  
  strcpy(filename3, "ohnebasis.out");
  printf("filename : %s\n", filename3);
  outfile3 = fopen(filename3, "w");
  TEMP=ZeroMatrix(A->R+A->C, A->C + 2);
  
  for (i=0; i<A->R; i++)
    {
      TEMP->V[i][0]=0;
      for(j=1; j<A->C+2-1; j++)  
	{
	  TEMP->V[i][j]=A->V[i][j-1];
	}
    }
  for (i=A->R; i<A->R+A->C; i++)
    {
      TEMP->V[i][0]=1;
      for(j=1; j<A->C+2-1; j++)  
	{
	  if((j-1)==(i-A->R))
	     TEMP->V[i][j]=1;
	}
    }
  FPRINTMATI(outfile3,0,TEMP->R-1,0,TEMP->C-1,TEMP);
  fclose(outfile3);

  printf("INVARIANTS: INPUT MATRIX \n");
  PRINTMATI(0,A->R-1,0,A->C-1,A);
  fprintf(outfile,"INVARIANTS: INPUT MATRIX \n");
  FPRINTMATI(outfile,0,A->R-1,0,A->C-1,A);
#endif
	
  reset_timer (0);
  start_timer (0);
  //printf("Invariant calculation starts :\n");
  
  /*calculate Hermite Normal Form*/
  TRANSMAT = LLLHermite2(A, &ROWECH, 3, 4, &rank);
  
#ifdef DEBUG
  printf("The transformation matrix is\n");
  PRINTMATI(0,TRANSMAT->R-1,0,TRANSMAT->C-1,TRANSMAT);
  fprintf(outfile,"The transformation matrix is\n");
  FPRINTMATI(outfile,0,TRANSMAT->R-1,0,TRANSMAT->C-1,TRANSMAT);
  
  printf("The row echelon Form is\n");
  PRINTMATI(0,ROWECH->R-1,0,ROWECH->C-1,ROWECH);
  fprintf(outfile,"The row echelon Form is\n");
  FPRINTMATI(outfile,0,ROWECH->R-1,0,ROWECH->C-1,ROWECH);
  
  printf("Rank of Matrix %i\n", rank);
  fprintf(outfile,"Rank of Matrix %i\n", rank);
#endif
  
  numrows=TRANSMAT->R;
  numcols=TRANSMAT->C;
  /*number of solutions=
    columns of transformation matrix - rank*/
  numsol = numcols - rank;
  
  if(numsol==0 || rank==0)
    {
#ifdef DEBUG
      printf("Keine Lsung\n");
      fprintf(outfile,"Keine Lsung\n");
#endif
      return 0;
    }
  
  /*read out the #numsol column vectors that
    form the Z-Base*/
  
  ZBASE=ZeroMatrix(numrows, numsol);
  for (i=0; i<numrows; i++)
    {
      for(j=rank; j<numcols; j++) 
	{
	  ZBASE->V[i][j-rank]=TRANSMAT->V[i][j];
	}
    }
  
  FreeMat(TRANSMAT);
  
#ifdef DEBUG
  printf("Z-Basis:\n");
  PRINTMATI(0,ZBASE->R-1,0,ZBASE->C-1,ZBASE);
  fprintf(outfile,"Z-Basis:\n");
  FPRINTMATI(outfile,0,ZBASE->R-1,0,ZBASE->C-1,ZBASE);
  
  fprintf(outfile2,"Z-Basis:\n");
  FPRINTMATI(outfile2,0,ZBASE->R-1,0,ZBASE->C-1,ZBASE);
  fclose(outfile2);
#endif	    
  
  
  INEQ=ZeroMatrix(ZBASE->R, ZBASE->C+2);
  for (i=0; i<INEQ->R; i++)
    {
      INEQ->V[i][0]=1;
      for(j=1; j<INEQ->C-1; j++)  
	{
	  INEQ->V[i][j]=ZBASE->V[i][j-1];
	}
    }

#ifdef DEBUG
  printf("Vor Sortieren /  eliminieren:\n ");
  PRINTMATI(0,INEQ->R-1,0,INEQ->C-1, INEQ);
  fprintf(outfile,"Vor Sortieren /  eliminieren:\n ");
  FPRINTMATI(outfile,0,INEQ->R-1,0,INEQ->C-1,INEQ);
#endif

#ifdef DEBUG	    
  printf("eliminieren: \n");
#endif
  ElimEqualRows(&INEQ);
  ElimSimpleConstraints(&INEQ);
  
#ifdef DEBUG
  printf("sortieren: \n");
#endif
  SortRows(INEQ);
  
#ifdef DEBUG
  printf("Sortiert und eliminiert:\n ");
  PRINTMATI(0,INEQ->R-1,0,INEQ->C-1, INEQ);
  fprintf(outfile,"Sortiert und eliminiert:\n ");
  FPRINTMATI(outfile,0,INEQ->R-1,0,INEQ->C-1,INEQ);
  fflush(outfile);
#endif
  
  /*Solve the system of inequalities ZBASE.x>=0 */
  numsol=SolveInequationSystem(&SOLUTION,INEQ, MAXROWSIZE);
  if (numsol==0) return 0;
  
#ifdef DEBUG
  printf("Lsung des Ungleichungssystems:\n ");
  PRINTMATI(0,SOLUTION->R-1,0,SOLUTION->C-1,SOLUTION);
  fprintf(outfile,"Lsung des Ungleichungssystems:\n ");
  FPRINTMATI(outfile,0,SOLUTION->R-1,0,SOLUTION->C-1,SOLUTION);
#endif	    
  
  /*multiply ZBASE with SOLUTION to obtain the invariants*/
  
  (*INVARS) = MultMat(ZBASE, SOLUTION);
  FreeMat(ZBASE);
  FreeMat(SOLUTION);
  
#ifdef DEBUG
  printf("Multipliziere mit Z-Basis:\n");
  PRINTMATI(0,(*INVARS)->R-1,0,(*INVARS)->C-1,(*INVARS));
  fprintf(outfile,"Multipliziere mit Z-Basis:\n");
  FPRINTMATI(outfile,0,(*INVARS)->R-1,0,(*INVARS)->C-1,(*INVARS));
  
  
  fprintf(outfile,"Das wars\n");

#endif
  
  stop_timer (0);
  zeit=time_passed(0);

#ifdef DEBUG 
  fprintf (outfile,"Time passed with computation: %8.2f s\n", zeit);
  fclose(outfile);
  
#endif
  
  /*printf ("Time passed with computation: %8.2f s\n", zeit);
    printf("Invariant calculation done.\n");*/
  
  return ((*INVARS)->C);
}
