/*****************************************************************************

Programm	:	struct

Komponente	:       lists.c

                        Funktionen zum Erzeugen, Loeschen und Verwalten von Listen
			Wird von inequalities.c benutzt

Programmer	:	Alexander Huck
Date		:	1998

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <malloc.h>

#include "intOps.h"
#include "matrixOps.h"
#include "matrixIO.h"
#include "matrix.h"
#include "timer.h"
#include "lists.h"
#include "sharedVars.h"

#define NEGATIVE -1

/*******************************************************************************************
create an empty linked list      
*******************************************************************************************/
LiList *CreateLiList()
{
  LiList *ptr;

  ptr=(LiList *)malloc(sizeof(LiList));
  ptr->Indices[0] = NEGATIVE;
  ptr->R = 0; /*number of rows*/
  ptr->next = NULL;

  return ptr;
}

/*******************************************************************************************
free a linked list      
*******************************************************************************************/
void FreeLiListElement(LiListElement *ListEl)
{
  LiListElement *ptr;
  ptr = ListEl;
  ListEl= ListEl->next;
  if ( ListEl != NULL)
    FreeLiListElement(ListEl);
  free(ptr);
  return;
}

void FreeLiList(LiList *List)
/* free the memory allocated by a linked list*/
{
  USI i;
  USI numrows;
  LiList *Temp;
  while(List != NULL)
    {
      numrows = List->R;
      for (i = 0; i < numrows; i++)
	{
	  FreeLiListElement(List->AdjRows[i]);
	}
      Temp = List;
      List = List ->next;
      free(Temp);
    }
  return;
}

/*******************************************************************************************
      
*******************************************************************************************/
int CreateLLElement(LiListElement **ptr)
{
  	LiListElement *N;
	N=(LiListElement *)malloc(sizeof(LiListElement));
	if( N==NULL ) 
	  {
	    printf("CreateLLElement: no more memory available!\n");
	    return MEMORYERROR;
	  }
	N->NbElements = 1;
	N->next = NULL;
	if ((*ptr) == NULL)
	  {
	    (*ptr) = N;
	    return OK;
	  }
	else
	  {
	    (*ptr)->next = N;
	    (*ptr) = N;
	    return OK;
	  }
}
/*******************************************************************************************
      store a pair of rays in the linked list
*******************************************************************************************/
int StorePair(LiList *List, int ray1, int ray2)
     /*store a pair in the linked list*/
{
USI numrows, NbElems;
USI i;
LiListElement *pointer;
USI contained = 0;
USI position=0;

    do
      {
	numrows = List->R;
	for (i = 0; i < numrows; i++)
	  {
	     if (List->Indices[i] == ray1)
	       {
		 position = i;
		 contained = 1;
		 break;
	       }
	  }
	if (contained) break;
	if (List->next != NULL)
	  List = List->next;
	else break;
      }
    while(1);

if (contained)
  {

    pointer = List->AdjRows[position];
    while(pointer->next != NULL)
      {
	pointer = pointer->next;
      }
    NbElems = (pointer->NbElements);
    if (NbElems == maxElements)
      {
	if(CreateLLElement(&pointer))
	  return 1;
	pointer->Elements[0] = ray2;
      }
    else
      {
	pointer->Elements[NbElems] = ray2;
	(pointer->NbElements)++;
      }
      

    return 0;
  }
else /*ray1 not yet contained in linked List*/
  {

    if (numrows == maxList)
      /* list filled, append a new one*/
      {
	List->next = CreateLiList();
	List = List->next;
	numrows = 0;
      }

    List->Indices[numrows] = ray1;
    pointer = NULL;
    if(CreateLLElement(&pointer))
      return 1;
    pointer->Elements[0] = ray2;
    List->AdjRows[numrows] = pointer;
    (List->R)++;
    return 0;
  }
}

/*******************************************************************************************
 used for debbuging: count entrys and elements in the list     
*******************************************************************************************/
void ScanList(LiList *List, int *entrys, int *elems)
{
  USI i;
  USI numrows, numels;
  LiListElement *ListEl;
  
  (*entrys) = 0;
  (*elems) = 0;
  
  while(List != NULL)
    {
      numrows = List->R;
      (*entrys) +=numrows;
      for (i = 0; i < numrows; i++)
	{
	  ListEl = List->AdjRows[i];
    while(ListEl != NULL)
      {  
	numels= ListEl->NbElements;
	(*elems) += numels;
	ListEl = ListEl->next;
      }
	}
      List = List ->next;
    }
}

/*******************************************************************************************
     print linkedlist-element
*******************************************************************************************/
void printLiListElement(LiListElement *ListEl)
{
  LiListElement *ptr;
  USI i;
  USI numels;
  numels= ListEl->NbElements;
  for (i = 0; i < numels; i++)
    {
      printf(" %i ", ListEl->Elements[i]);
    }
  ptr = ListEl->next;
  if (ptr != NULL)
    printLiListElement(ptr);
  else
    printf("\n");
  return;
}

/*******************************************************************************************
      print a linked list
*******************************************************************************************/
void printLiList(LiList *List)
/* print a linked list*/
{
  USI i;
  USI numrows;
  USI count = 0;
 
  printf("************DISPLAYLIST***********\n");
while(List != NULL)
  {
  numrows = List->R;
  for (i = 0; i < numrows; i++)
    {
      printf("Row number %i ", List->Indices[i]);
      printf("is adjacent to rows\n");
      printLiListElement(List->AdjRows[i]);
    }
  List = List ->next;
  count += numrows;
}
  printf("*********%i ELEMENTS***********\n", count);
  return;
}

/*******************************************************************************************
     initialize list of free row indices
*******************************************************************************************/
FreeLi *InitFreeList()
{
  	FreeLi *N;
	N=(FreeLi *)malloc(sizeof(FreeLi));
	N->Bottom = 0;
	N->Top = 0;
	N->next = NULL;
	return N;
}

/*******************************************************************************************
     delete list of free row indices
*******************************************************************************************/
void DelFreeList(FreeLi *FreeList)
{
  FreeLi *ptr;
  ptr = FreeList;
  FreeList= FreeList->next;
  if ( FreeList != NULL)
    DelFreeList(FreeList);
  free(ptr);
  return;
}

/*******************************************************************************************
       return a pointer to a free row index
*******************************************************************************************/
int GetFreeRow(FreeLi **FreeList, USI *blockflag)
{
  FreeLi *ptr2, *ptr3;
  int pos1;
  USI bottom, top;
  
  ptr2 = (*FreeList);
  
  /*gib den Index an Position bottom zurck*/
  
  bottom = ptr2->Bottom;
  top = ptr2->Top;
  
  pos1 = ptr2->FreeRows[bottom];
  
  bottom ++;
  if (bottom == top)
    /*Liste leer, weil abgearbeitet*/
    {
      ptr3 = ptr2;
      ptr2 = ptr2->next;
      if (ptr2 == NULL)
	{/*kein Nachfolger mehr da*/
	  /*d.h. keine vereinzelten Reihen*/
	  (*blockflag) = 1;
	  
	  ptr3->Bottom=0;
	  ptr3->Top=0;
	  (*FreeList) = ptr3;
	  
	} 
      else
	{
	  free(ptr3); /*Liste freigeben*/
	  (*FreeList) = ptr2;
	}
    } 
  else
    {
      ptr2->Bottom = bottom;
    }
  
  return pos1;
}

/*******************************************************************************************
    wird benutzt um Reihen mit Index != NbRay zu lschen
*******************************************************************************************/
void DelSingleRow(FreeLi **FreeList, USI *blockflag, USI row)
{
  FreeLi *ptr2;
  USI top;

  ptr2 = (*FreeList);
  while (ptr2->next != NULL)
    ptr2 = ptr2->next;

  top = ptr2->Top;
 
  if (top == maxElementsFL)
    /* list full*/
    {
      ptr2->next = InitFreeList();
      ptr2 = ptr2->next;
      top = 0;
    }

  ptr2->FreeRows[top] = row;
  (ptr2->Top)++;
  (*blockflag) = 0;
}
