#include "matrix.h"
#include "sharedVars.h"
#include <malloc.h>
#include <string.h>
#include <stdio.h>


INTMATRIX *BuildMat(unsigned int m, unsigned int n)
/*
 * Allocates space for an m x n matrix
 */
{
	INTMATRIX *N;
	int i;

	N=(INTMATRIX *)malloc(sizeof(INTMATRIX));
	N->R = m;     	/*Number of Rows*/
	N->C = n;       /*Number of Columns*/
	N->V = (int **)malloc(m * sizeof(int *)); /*Values*/
	for (i = 0; i < m; i++)
		N->V[i] = (int *)malloc(n * sizeof(int));
	return (N);
}


INTMATRIX *ZeroMatrix(USI m, USI n)
/*
 * returns the zero  m x n matrix.
 */
{
	int i, j;
	INTMATRIX *Mptr;
	Mptr = BuildMat(m, n);
	for (i = 0; i < m; i++)
	{
		for (j = 0; j <n; j++)
			Mptr->V[i][j] = 0;
	}
	return (Mptr);
}


INTMATRIX *IdentityMat(USI n)
/*
 * returns the identity matrix of size n.
 */
{
	unsigned int i, j;
	INTMATRIX *Mptr;

	Mptr = BuildMat(n, n);
	for (i = 0; i <= n - 1; i++)
	{
		for (j = 0; j <= n - 1; j++)
		{
			if ( i == j)
				Mptr->V[i][j] = 1;
			else
				Mptr->V[i][j] = 0;
		}
	}
	return (Mptr);
}

int ResizeRows(INTMATRIX **A, int numrows)
     /*add numrows rows to a Matrix*/
{
  int rows, newrows, cols;
  USI i;
  int **dummy;
  rows = (*A)->R;
  cols = (*A)->C;
  newrows = rows + numrows;
  dummy = (int **)malloc(newrows * sizeof(int *));
  if (dummy == NULL) 
    {
      printf("RESIZEROWS: no more memory available!\n");
      return MEMORYERROR;
    }
  for (i = 0; i < rows; i++)
    dummy[i] = (*A)->V[i];
  for (i = rows; i < newrows; i++)
    dummy[i] = (int *)malloc(cols * sizeof(int));
  free((*A)->V);
  (*A)->V=dummy;
 (*A)->R = newrows;
  return OK;
}


void FreeMat(INTMATRIX *Mptr)
/*
 * frees the storage allocated to the two-dimensional array Mptr->V.
 */
{
	unsigned int i, m;

	if (Mptr == NULL)
		return;
	m = Mptr->R;
	for (i = 0; i < m; i++)
	{
		free(Mptr->V[i]);
	}
	free(Mptr->V);
	free(Mptr);
	return;
}

void PartFreeMat(INTMATRIX **Mptr, USI startrow)
     /*free rows beginning at startrow*/
{
	unsigned int i, m;
	if ((*Mptr) == NULL)
		return;
	m = (*Mptr)->R;
	for (i = startrow; i < m; i++)
	{
		free((*Mptr)->V[i]);
	}
	(*Mptr)->R=startrow;
	return;
}

INTMATRIX *CopyMat(INTMATRIX *Mptr)
/*
 * copy Matrix.
 */
{
	unsigned int i, j, m, n;
	INTMATRIX *Nptr;

	m = Mptr->R;
	n = Mptr->C;
	Nptr = BuildMat(m, n);
	for (i = 0; i < m; i++)
		for (j = 0; j < n; j++)
			Nptr->V[i][j] = Mptr->V[i][j];
	return (Nptr);
}

int SIZEOFMATI(INTMATRIX *Mptr)
     /*returns size of matrix*/
{
	USI size = 0;
	USI n, m;

	if (Mptr == NULL)
	  {
	    printf("size of matrix: NULL-pointer???\n");
	    return 0;
	  }
	  
	m= Mptr->R;
	n= Mptr->C;

	/*space occupied by data*/
	size += n * m * sizeof(int);
	/*space occupied by pointers*/
	size += m * sizeof(int *);
	/*space occupied by struct*/
	size += sizeof(INTMATRIX);
	printf("size of matrix: %i bytes\n", size);
	return size;
}
