/*****************************************************************************

Programm	:	struct

Komponente	:	matrixCombine.c

                        Kombinieren von zwei Matrixreihen

Programmer	:	Alexander Huck
Date		:	1998

*****************************************************************************/
#include "intOps.h"
#include "matrix.h"
#include "sharedVars.h"
#include <malloc.h>
#include <stdio.h>
#include <math.h>
/*#define DEBUG*/

long double GGcd(long double a, long double b)
{
  long double aux;
  while (a) { aux = fmod(b, a); b=a; a=aux; }
  return (b<0 ? -b : b);
} 

long double VGcd(long double *p, USI length)
{
  long double a=0, b;
  int i, j;
  for(i=1; i<length;i++)
    {
      a=p[i];
      if(a) break;
    }
  a=abs(a);
  if (a==1) return 1;
  for(j=i+1; j<length; j++)
    {
      b=p[j];
      if (b) a=GGcd(a,abs(b));
      if (a==1) return 1;
    }
  return a;
}

void VNormalize(long double *p, USI length)
{
  long double g;
  int i;
  g = VGcd( p, length);	
  if ( g!=1 ) for (i=1; i<length; i++) p[i]/=g;
} 

/*the same as CombineRow, only with floating point arithmetic.
If CombineRow fails due to an overflow, CombineRowFloat is called.
(type long double has more bits than integer).
If CombineRowFloat fails too, OVERFLOWERROR is returned.
*/
int CombineRowFloat(USI p, USI q, USI r, int A, int B, INTMATRIX *Mptr)
{
	unsigned int j;
	long double U1, V1, X1, Y1;
	long double *temp;
	int numcols = Mptr->C;
	temp = (long double *)malloc( numcols * sizeof(long double));

	for (j = 1; j <= Mptr->C - 1; j++)
	{

	  U1 = (long double)(Mptr->V[p][j]);
	  V1 = (long double)(Mptr->V[q][j]);
	  X1 = (long double)A * U1;
	  Y1 = (long double)B * V1;

	  if ( (U1 && (A != X1/U1)) ||
	       (V1 && (B != Y1/V1)) )
	    {
	      printf("? Vector_Combine:float multiplication overflow\n");
	      printf("X1, Y1 %Lg %Lg\n", X1, Y1);
	      return OVERFLOWERROR;
	    }
	
	  temp[j] = X1 + Y1;
	}

	VNormalize(temp, numcols);

	for (j = 1; j <= Mptr->C - 1; j++)
	{
	  if (temp[j] - (int)temp[j] != 0)
	    {
	      printf("? Vector_Combine: overflow, number too big\n");
	      return OVERFLOWERROR;
	    }
	  Mptr->V[r][j] = (int)temp[j];
	}
	Mptr->V[r][0] = 0;
	return OK;
}

int CombineRow(USI p, USI q, USI r, int A, int B, INTMATRIX *Mptr)
/*
 *  A times row p plus B times row q -> row r of *Mptr.
 *  First column of row r stays unchanged
 */
{
	unsigned int j;
	int U, V, X, Y, Z;
	int numcols = Mptr->C;

	for (j = 1; j <= Mptr->C - 1; j++)
	{
	  
	  U = (Mptr->V[p][j]);
	  V = (Mptr->V[q][j]);
	  X = A * U;
	  Y = B * V;
	  
	  if ( (U && (A != X/U)) ||
	       (V && (B != Y/V)) )
	    {
	      printf("? Vector_Combine: multiplication overflow\n");
	      printf("retrying with float\n");
	      printf("X, Y %x %x\n", X, Y);
	      if (CombineRowFloat(p, q, r, A, B, Mptr)) /*write new row r3*/
		return OVERFLOWERROR; /*overflow error: number too big*/
	      else
		return OK; /*ok*/
	    }

	  Z = X+Y;
	  
	  if ((Z-Y) != X)
	    {
	        printf("? Vector_Combine: addition overflow\n");
		printf("retrying with float\n");
		printf("X, Y %x %x\n", X, Y);
		if(CombineRowFloat(p, q, r, A, B, Mptr)) /*write new row r3*/
		  return OVERFLOWERROR;
		else
		  return OK;
	    }

	  Mptr->V[r][j] = Z;
	}

	Vector_Normalize(Mptr->V[r], numcols); /*gcd reduce r3*/

	Mptr->V[r][0] = 0;

	return OK;
}
