/*****************************************************************************

Programm	:	struct

Komponente	:	matrixOps.c

                        Operationen auf Matrizen

Programmer	:	Alexander Huck
Date		:	1998

*****************************************************************************/
#include "intOps.h"
#include "matrix.h"
#include "sharedVars.h"
#include <malloc.h>
#include <stdio.h>
#include <math.h>
/*#define DEBUG*/

INTMATRIX *SwapCols(USI p, USI q, INTMATRIX *Mptr)
/*
 * Updates Mptr by interchanging cols p and q 
 */
{
	unsigned int i;
	int Temp;

	for (i = 0; i <= Mptr->R - 1; i++)
	{
		Temp = Mptr->V[i][p];
		Mptr->V[i][p] = Mptr->V[i][q];
		Mptr->V[i][q] = Temp;
	}
	return (Mptr);
}

INTMATRIX *SwapRows(USI p, USI q, INTMATRIX *Mptr)
/*
 * Updates Mptr by pointer-swap rows p and q 
 */
{
        int *temp;

        temp = Mptr->V[p];
        Mptr->V[p] = Mptr->V[q];
        Mptr->V[q] = temp;
        return (Mptr);
}

void RowExchange(INTMATRIX *Mptr, USI a, USI b, int *temp)
     /*swap rows a and b of Mptr
       by memory copy
       */
{
  int *aptr, *bptr;
  USI RowSize= Mptr->C * sizeof(int);
  aptr = Mptr->V[a];
  bptr = Mptr->V[b];
  memcpy((char *)(temp), (char *)(aptr), (int)(RowSize));
  memcpy((char *)(aptr), (char *)(bptr), (int)(RowSize));
  memcpy((char *)(bptr), (char *)(temp), (int)(RowSize));
}

void CopyRow(INTMATRIX *Mptr, USI a, USI b)
     /*copy row a to row b*/
{
  int *aptr, *bptr;
  USI numcols = Mptr->C;
  USI RowSize= numcols * sizeof(int);
  aptr = Mptr->V[a];
  bptr = Mptr->V[b];
  memcpy((char *)(bptr), (char *)(aptr), (int)(RowSize));
}

INTMATRIX *AddMultCol(USI p, USI q, int A, INTMATRIX *Mptr)
/*
 * adding A times col p to col q of *Mptr.
 */
{
	unsigned int j;
	int X;
	INTMATRIX *Nptr;

	Nptr = CopyMat(Mptr);
	for (j = 0; j <= Nptr->R - 1; j++)
	{
		X = A * (Nptr->V[j][p]);
		Nptr->V[j][q] = Nptr->V[j][q] + X;
	}
	return (Nptr);
}

void RowOr(USI p1, USI p2, USI p3, INTMATRIX *Mptr)
     /*
       row p3 = row p1 OR row p2
       */
{
       	unsigned int j;
  	int X, Y;

 	for (j = 0; j <= Mptr->C - 1; j++)
	{
		X = (Mptr->V[p1][j]);
		Y = (Mptr->V[p2][j]);
		Mptr->V[p3][j] = X | Y;
	}
	return;
} 

INTMATRIX *ScalarCol(USI p, int A, INTMATRIX *Mptr)
/*
 * multiplying the pth column of *Mptr by A.
 */
{
	unsigned int j;
	INTMATRIX *Nptr;

	Nptr = CopyMat(Mptr);
	for (j = 0; j <= Nptr->R - 1; j++)
	{
		Nptr->V[j][p] = (Nptr->V[j][p]) * A;
	}
	return (Nptr);
}


USI RowSeek(INTMATRIX *A, USI i, USI j)
/* Scans rows >= i, cols >= j for first nonzero row if any. Returns 
 * this row number if there is one, else returns A->R;
 */
{
	USI k, l, n;

	n = A->R;
	for (k = j; k < n; k++)
	{
		for (l = 0; l <= i; l++)
		{
			if ((A->V[k][l])!=0)
				return k;
		}
	}
	return k;
}

INTMATRIX *MultMat(INTMATRIX *Mptr, INTMATRIX *Nptr)
/*
 * Here Mptr->C = Nptr->R. 
 * returns (*Mptr) * (*Nptr)
 */
{
	int X, Y ;
	int i, j, k;
	INTMATRIX *Lptr;

	Lptr = BuildMat(Mptr->R, Nptr->C);
	for (i = 0; i < Mptr->R; i++)
	{
		for (j = 0; j < Nptr->C; j++)
		{
			X = 0;
			for (k = 0; k < Mptr->C; k++)
			{
				Y = (Mptr->V[i][k])*( Nptr->V[k][j]);
				X = X + Y;
			}
			Lptr->V[i][j] = X;
		}
	}
	return (Lptr);
}

INTMATRIX *TransMat(INTMATRIX *Mptr)
     /*Transpose Matrix*/
{
        int i, j;
	INTMATRIX *Lptr;

	Lptr = BuildMat(Mptr->C, Mptr->R);
	for (i = 0; i <= Mptr->R - 1; i++)
	{
		for (j = 0; j < Mptr->C; j++)
		{
			Lptr->V[j][i] = Mptr->V[i][j];
		}
	}
	return (Lptr);
}

/*is row findrow of SRC in SEEK?
  yes => return row of SEEK
  no => return -1*/
int ISIN(int findrow, INTMATRIX *SRC, INTMATRIX *SEEK)
{
  USI i,j;
  int numcols = SEEK->C;
  int numrows = SEEK->R;
  if (SRC->C!=numcols) return -1;
  for(i=0; i<numrows; i++)
    {
      for(j=0; j<numcols; j++)
	{
	  if(SEEK->V[i][j]!=SRC->V[findrow][j])
	    break;
	}
      if (j==numcols) return i;
    }
  return -1;
}

