/*****************************************************************************

Programm	:	struct

Komponente	:	matrixProcs.c

                        Sortieren von Matrizen
			Routinen zur Vorverarbeitung

Programmer	:	Alexander Huck
Date		:	1998

*****************************************************************************/
#include "intOps.h"
#include "matrixIO.h"
#include "matrix.h"
#include "sharedVars.h"
#include <malloc.h>
#include <stdio.h>
#include <math.h>

//#define DEBUG

void ElimEqualRows(INTMATRIX **Mptr)
     /*eliminate equal rows, leaving only the first one*/
{
  INTMATRIX *Lptr;
  USI i,j,k,equal;
  int numcols = (*Mptr)->C;
  int numrows = (*Mptr)->R;
  int rowcount = 0;
  int *obsolet;
  obsolet = (int *)malloc( numrows * sizeof(int));
  for(i=0; i<numrows; i++)
    obsolet[i]=0;
  
  Lptr = BuildMat(numrows, numcols);
  
  for (i=0; i<numrows; i++)
    {
      if (obsolet[i]) continue;
      /*copy into new matrix*/
      for (k=0; k<numcols; k++)
	{
	  Lptr->V[rowcount][k]=(*Mptr)->V[i][k];
	}
      rowcount++;
      /*compare with all other rows*/
      for (k=i+1; k<numrows; k++)
	{
	  if(obsolet[i]) continue;
	  for (j=0, equal=1; (j<numcols) && equal; j++)
	    {
	      equal=((*Mptr)->V[i][j]==(*Mptr)->V[k][j]);
	    }
	  if (equal) obsolet[k]=1;
#ifdef DEBUG
	  if (equal)
	    {
	      printf("rows %i and %i are identical\n", i, k);
	      PRINTMATI(i,i,0,(*Mptr)->C-1, (*Mptr));
	      PRINTMATI(k,k,0,(*Mptr)->C-1, (*Mptr));
	    }
#endif
	}
    }
  PartFreeMat(&Lptr, rowcount);
  free(obsolet);
  
#ifdef DEBUG
  printf("ELIMEQUALROWS: %i rows removed.\n", numrows - rowcount);
#endif
  
  FreeMat((*Mptr));
  /*
  (*Mptr)=ZeroMatrix(rowcount, numcols);
  for (i=0; i<rowcount; i++)
    {
      for (k=0; k<numcols; k++)
	{
	  (*Mptr)->V[i][k]=Lptr->V[i][k];
	}
    }
  */
  (*Mptr) = Lptr;
  return;
}

void RemoveZeroRows(INTMATRIX **Mptr)
{
  INTMATRIX *Lptr;
  int numcols = (*Mptr)->C;
  int numrows = (*Mptr)->R;
  int rowcount = 0;
  int testrow, column;
  int nozerorow;

  Lptr = BuildMat(numrows, numcols);

   for(testrow=0;  testrow< numrows; testrow++)
    {/*for all rows*/
      nozerorow = 0;
      for (column=0; column< numcols; column++)
	    {/*for all columns*/
	      nozerorow |= (*Mptr)->V[testrow][column];
	    }
      if (nozerorow)
	{
	for (column=0; column<numcols; column++)
	   {
	     Lptr->V[rowcount][column]=(*Mptr)->V[testrow][column];
	   }
	 rowcount++;
	}
    }
   PartFreeMat(&Lptr, rowcount);
   FreeMat((*Mptr));
   (*Mptr) = Lptr;
   return;
}

void ElimNonMinimal(INTMATRIX **Mptr)
{
  /*find minimal supports*/
  /*a minimal support does not contain another support*/
int testrow, comparerow, column, minimal=0;	
 INTMATRIX *Lptr;
 int rowcount = 0;
 int numcols = (*Mptr)->C;
 int numrows = (*Mptr)->R;

 Lptr = BuildMat(numrows, numcols);

  for(testrow=0;  testrow< (*Mptr)->R; testrow++)
    {/*for all rows*/
      for(comparerow=0; comparerow< (*Mptr)->R; comparerow++)
	{/*for all rows*/
	  if (comparerow == testrow) continue;
	  minimal = 0;
	  for (column=0; column< (*Mptr)->C; column++)
	    {/*for all columns*/
	      /*testrow does not contain comparerow,
		if it has an 0-element where comparerow has an element !=0*/ 
	      if((*Mptr)->V[comparerow][column] && !(*Mptr)->V[testrow][column])
		minimal = 1;
	    }
	  if (!minimal) break;
	}
      if(minimal)
	{
	/*copy into new matrix*/
	 for (column=0; column<numcols; column++)
	   {
	     Lptr->V[rowcount][column]=(*Mptr)->V[testrow][column];
	   }
	 rowcount++;
	}
    }
  PartFreeMat(&Lptr, rowcount);
  FreeMat((*Mptr));
  (*Mptr) = Lptr;
  return;
}

void ElimSimpleConstraints(INTMATRIX **Mptr)
{
  INTMATRIX *Lptr;
  USI i,j,k;
  int numcols = (*Mptr)->C;
  int numrows = (*Mptr)->R;
  int rowcount = 0;
  int firstnonzero, a, b, multi, multiflag;
  int *obsolet;
  obsolet = (int *)malloc( numrows * sizeof(int));
  for(i=0; i<numrows; i++)
    obsolet[i]=0;
  
  Lptr = BuildMat(numrows, numcols);
  
  for (i=0; i<numrows; i++)
	  {
	    if (obsolet[i]) continue;
	    /*copy into new matrix*/
	    for (k=0; k<numcols; k++)
	      {
		Lptr->V[rowcount][k]=(*Mptr)->V[i][k];
	      }
	    rowcount++;
	    /*check if row k is a positive multiple of row i*/
	    for (k=i+1; k<numrows; k++)
	      {
		if(obsolet[i]) continue;
		if((*Mptr)->V[i][0]!=(*Mptr)->V[k][0]) continue; /*dont compare ineqs with eqs*/ 
		for (j=1; j<numcols; j++)
		  {
		    if((*Mptr)->V[i][j]!=0) break;
		  }
		if (j == numcols) continue;
		firstnonzero = j;
		a = (*Mptr)->V[k][firstnonzero];
		b = (*Mptr)->V[i][firstnonzero];
		if (!a) continue;
		if ((a%b)!=0) continue; 
		multi = a/b;
		if (multi<0) continue;
		/*printf("compare row %i with row %i, multi=%i\n", i, k, multi);*/
		multiflag = 1;
		for (j=1; j<numcols; j++)
		  {
		    a = (*Mptr)->V[k][j];
		    b = (*Mptr)->V[i][j];
		    if (a!= multi * b)
		      {
			multiflag = 0;
			break;
		      }
		  }
		if (multiflag) 
		  {
#ifdef DEBUG
		    printf("Row %i is a multiple of row %i \n", k, i);
#endif
		    obsolet[k]=1;
		  }
	      }
	  }
  PartFreeMat(&Lptr, rowcount);
  free(obsolet);
  
#ifdef DEBUG
  printf("ELIMSIMPLECONSTRAINTS: %i rows removed.\n", numrows - rowcount);
#endif
  
  FreeMat((*Mptr));
  (*Mptr) = Lptr;
  return;
}

int LexSmaller(INTMATRIX *Mptr, int v1, int v2)
{ /* nmax is the size of vectors v1,v2 */
  int determined, smaller;
  int j;
  int nmax = Mptr->C;

  smaller = 0;
  determined = 0;
  j = 2;
  do {
    if (abs(Mptr->V[v1][j - 1] -Mptr->V[v2][j - 1])>0) {
      if (Mptr->V[v1][j - 1] < Mptr->V[v2][j - 1]) {
	    smaller = 1;
	  }
      determined = 1;
    } else
      j++;
  } while (!(determined) && (j <= nmax));
  return smaller;
}


void SortRows(INTMATRIX *Mptr)
{
  int i,j,swapped;
  int numrows = Mptr->R;
  int *TT;
 
  /*perform Bubble-sort*/
  for (i = numrows; --i>=0; ) {
    swapped = 0;
    for ( j = 0; j<i; j++) {
      if (!LexSmaller(Mptr, j, j+1)) {	
	TT = Mptr->V[j];
	Mptr->V[j] = Mptr->V[j+1];
	Mptr->V[j+1] = TT;
	
	swapped = 1;
      }
    }
    if (!swapped)
      break;
  }

   /*slide equations up to the top of the matrix*/
  for (i = numrows; --i>=0; ) {
    swapped = 0;
    for ( j = 0; j<i; j++) {
      if (Mptr->V[j][0]> Mptr->V[j+1][0]) {	
	TT = Mptr->V[j];
	Mptr->V[j] = Mptr->V[j+1];
	Mptr->V[j+1] = TT;
	
	swapped = 1;
      }
    }
    if (!swapped)
      break;
  }
  
}


/*sort a constraint matrix.
A good order of constraints leads to a dramatical decrease of computation time.
The order chosen here is lexmin, rows with only one entry come first
*/
void _SortRows(INTMATRIX *Mptr)
{
  int i,j, positiveentry, negativeentry, zeroentry, onlyone, swapped;
  int numcols = Mptr->C;
  int numrows = Mptr->R;
  int *rowstat, *TT;
  int T;
  rowstat = (int *)malloc( numrows * sizeof(int));
  onlyone = 0;
  for (i=0; i<numrows; i++)
    {
      positiveentry = 0;
      negativeentry = 0;
      zeroentry = 0;
      for(j=0; j<numcols; j++)
	{
	  if(Mptr->V[i][j]>0) positiveentry++;
	  if(Mptr->V[i][j]>0) negativeentry++;
	  if(Mptr->V[i][j]==0) zeroentry++;
	}
      rowstat[i] = zeroentry;
      if (numcols - rowstat[i] == 2)
	{
	  onlyone++;
	}
    }
#ifdef DEBUG
  printf("Zeilen mit nur einem Eintrag: %i\n", onlyone);
#endif
  /*perform Bubble-sort*/
  for (i = numrows; --i>=0; ) {
    swapped = 0;
    for ( j = 0; j<i; j++) {
      if (rowstat[j] < rowstat[j+1]) { /*lowest expansion-factor to top*/
	
	T = rowstat[j];
	rowstat[j] = rowstat[j+1];
	rowstat[j+1] = T;
	
	TT = Mptr->V[j];
	Mptr->V[j] = Mptr->V[j+1];
	Mptr->V[j+1] = TT;
	
	swapped = 1;
      }
    }
    if (!swapped)
      break;
  }
  
  for (i=0; i<onlyone; i++)
    {
      zeroentry = 0;
      for(j=0; j<numcols; j++)
	{
	  if(Mptr->V[i][j]!=0) zeroentry+= j;
	}
      rowstat[i] = zeroentry;
    }
  
  for (i = onlyone-1; --i>=0; ) {
    swapped = 0;
    for ( j = 0; j<i; j++) {
      if (rowstat[j] > rowstat[j+1]) 
	{
	  
	  T = rowstat[j];
	  rowstat[j] = rowstat[j+1];
	  rowstat[j+1] = T;
	  
	  TT = Mptr->V[j];
	  Mptr->V[j] = Mptr->V[j+1];
	  Mptr->V[j+1] = TT;
	  
	  swapped = 1;
	}
    }
    if (!swapped)
      break;
  }
  
  for (i=onlyone; i<numrows; i++)
    {
      zeroentry = 0;
      for(j=0; j<numcols; j++)
	{
	  if(Mptr->V[i][j]!=0) zeroentry+= j;
	}
      rowstat[i] = zeroentry;
     }
  
  for (i = numrows; --i>=onlyone+1; ) {
    swapped = 0;
    for ( j = 0; j<i; j++) {
      if (rowstat[j] > rowstat[j+1]) 
	{
	  
	  T = rowstat[j];
	  rowstat[j] = rowstat[j+1];
	  rowstat[j+1] = T;
	  
	  TT = Mptr->V[j];
	  Mptr->V[j] = Mptr->V[j+1];
	  Mptr->V[j+1] = TT;
	  
	  swapped = 1;
	}
    }
    if (!swapped)
      break;
  }
  free(rowstat);
}

int Partition(INTMATRIX *OV, USI p, USI r, USI index1, USI index2)
{
  USI x;
  USI i,j,temp;
  x=OV->V[p][index1];
  i=p-1;
  j=r+1;
  while (1){
    do{
      j--;
    } while (OV->V[j][index1] > x );
    do{
      i++;
    } while (OV->V[i][index1] < x);
    if (i<j){
      temp=OV->V[i][index1];
      OV->V[i][index1]=OV->V[j][index1];
      OV->V[j][index1]=temp;

      temp=OV->V[i][index2];
      OV->V[i][index2]=OV->V[j][index2];
      OV->V[j][index2]=temp;
    }
    else{
      return j;
    }
  }
}

void QuickSort(INTMATRIX *OV, USI p, USI r, USI index1, USI index2)
     /*sort OV by its component OV->V[index1]
       OV->V[index2] is swapped accordingly
       start with index p 
       up to index r*/
{
  int q;
  
  if (p < r){
    q = Partition(OV, p, r, index1, index2);
    QuickSort(OV, p, q, index1, index2);
    QuickSort(OV, q+1, r, index1, index2);
  }
}
