/*****************************************************************************

Programm	:	estimse, reachable

Komponente	:	polycount.c
                        Bestimmen des Inhalts eines Polytops

Programmer	:	Alexander Huck
Date		:	1999

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <malloc.h>

#include "intOps.h"
#include "matrixOps.h"
#include "matrixIO.h"
#include "matrix.h"
#include "timer.h"
#include "remove_redundants.h"

#include "sharedVars.h"
#include "sharedOps.h"
#include "debugOps.h"
#include "index.h"
#include "programOutput.h"
#include "simplifyBounds.h"

/*
#define PRINTDOMAINS
#define DEBUG
#define DEBUGCOUNT
#define DEBUGSIMPLIFY
#define SHOWMARKING
#define DEBUGCOUNT
#define DEBUGCOUNT3
#define SHOWTIME
*/


#define TESTSTART TTEST

INTMATRIX **Marker;
int recursions;
int markingcount;

int Distance(USI ray1, USI ray2)
{
  USI j;
  USI dist=0;
  for(j=1; j<Dimension; j++)
    dist+=abs(Ray->V[ray1][j]-Ray->V[ray2][j]);
  return dist;
}

     
int Coarsify(USI start, USI end )
{
  USI redundant = 1;
  USI Word, Word2;
  int thisword;
  USI otherRay; 
  USI oldrplus, oldrminus;
  USI rp, rm;
  USI RowSize;
  USI rplus, rminus;
  USI bound;
  int i;
  USI nbcommonconstraints;
  int *ptr1, *ptr2;
  int *incidence;
  int inccounter;
  int maxdist;
  OldRays=0;

  bound=NbRay;
  RowSize=Sat->C * sizeof(int);
  BitRow=(int *)malloc(RowSize);
  incidence=(int *)malloc((end-start)*sizeof(int));

  TTEST = NbLines;
 
  printf("Coarsify testing rays from %i to %i \n", start, end);

  for(rplus=start;rplus<end;rplus++)
    {
      if (Ray->V[rplus][Dimension]==0) continue; /*dont test rays*/
      //  printf("Ecke %i \n", rplus);
      inccounter=0;
      for(rminus=rplus+1;rminus<end;rminus++)
	{
	  if (Ray->V[rminus][Dimension]==0) continue; /*dont test rays*/
	  /*----------------------------------------------------------------------*/
	  /* compute and count the set of constraints saturated by both R+ and R- */
	  /*----------------------------------------------------------------------*/
	  
	  nbcommonconstraints=0;
	  for (thisword=0; thisword<AdjWCount; thisword++) 
	    { 
	      BitRow[thisword] = Sat->V[rplus][thisword] | Sat->V[rminus][thisword];/*OR the rows*/
	      Word=Word2=BitRow[thisword];
	      Word >>= (sizeof(int)/2 * 8);
	      Word2 &= 0x0000ffff;
	      nbcommonconstraints+=LookUpMat[Word];
	      nbcommonconstraints+=LookUpMat[Word2];
	    }
	  /*last word with index [AdjWCount] has only AdjBCount bits:*/
	  Word = Word2 = BitRow[AdjWCount] = Sat->V[rplus][AdjWCount] | Sat->V[rminus][AdjWCount] | AdjBitmask;
	  
	  Word >>= (sizeof(int)/2 * 8);
	  Word2 &= 0x0000ffff;
	  nbcommonconstraints+=LookUpMat[Word];
	  nbcommonconstraints+=LookUpMat[Word2];
	  
	  /*-----------------------------------------------------------------*/
	  /* Adjacency Test : is combination [R-,R+] a non redundant ray?    */
	  /*-----------------------------------------------------------------*/
	  if (nbcommonconstraints + NbLines>=Dimension-2) /* dimensionality check*/
	    { /* check whether a ray m saturates the same set of constraints */
	      
	      rp = rplus;
	      rm = rminus;
	      
	      /*copy the two rays to the top of the ray matrix*/
	      
	      if (rplus == TTEST)
		{
		  oldrplus = rplus;
		  rplus = rminus;
		  rminus = oldrplus;
		}
	      
	      oldrminus = rminus;
	      RowExchange(Sat, rminus, TTEST, sattemprow);
	      RowExchange(Ray, rminus, TTEST, raytemprow);
	      rminus = TTEST;
	      
	      oldrplus = rplus;
	      RowExchange(Sat, rplus, TTEST+1, sattemprow);
	      RowExchange(Ray, rplus, TTEST+1, raytemprow);
	      rplus = TTEST+1;
	      
	      redundant=0;
	      for (otherRay = TTEST + 2; otherRay<bound; otherRay++) 
		if ((otherRay!=rplus)&&(otherRay!=rminus))
		  /* for (otherRay = NbRay-1; otherRay>=TTEST+2; otherRay--) */
		  /*for each ray otherRay, otherRay!=rplus, otherRay!=rminus do*/  
		  {
		    /* (r+ r-) is redundant if there does NOT exist an equation */
		    /* which saturates both r+ and r- but not otherRay */
		    ptr1 = BitRow; /*contains (Sat->V[rminus] OR Sat->V[rplus]) */
		    ptr2 = Sat->V[otherRay];
		    
		    thisword=0;
		    do{
		      if (*ptr2&~*ptr1) break;
		      thisword++;
		      if (thisword>AdjWCount)
			goto is_redundant;
		      ptr2++,ptr1++;
		    } while (1); 
		  }
	      
	      if (!redundant)  {
		/*------------------------------------------------------------*/
		/* Add new ray generated by [R+,R-]                           */
		/*------------------------------------------------------------*/
		
		stop_timer (2); /*adjtest*/
		
		
		RowExchange(Sat, oldrplus, rplus, sattemprow);
		RowExchange(Sat, oldrminus, rminus, sattemprow);
		RowExchange(Ray, oldrplus, rplus, raytemprow);
		RowExchange(Ray, oldrminus, rminus, raytemprow);
		
		rplus = rp; rminus = rm;
		
		//	printf("Rays %i and %i are adjacent.\n", rminus, rplus);
		incidence[inccounter] = rminus; 
		inccounter++;
	      } 
	      else /*redundant*/
		{
		is_redundant:
		  RowExchange(Sat, oldrplus, rplus, sattemprow);
		  RowExchange(Sat, oldrminus, rminus, sattemprow);
		  RowExchange(Ray, oldrplus, rplus, raytemprow);
		  RowExchange(Ray, oldrminus, rminus, raytemprow);
		  
		  rplus = rp; rminus = rm;
		  
		}
	    } /*if dimensionality check*/
	}
      // printf("Rays adjacent to %i :\n", rplus);
      maxdist=0;
      for(i=0; i<inccounter; i++)
	{
	  if(Distance(rplus, incidence[i])>maxdist)
	    maxdist = Distance(rplus, incidence[i]);
	  // printf("%i distance: %i \n", incidence[i], Distance(rplus, incidence[i]));
	}
      if(maxdist<10 && inccounter>0)
	printf("Ecke %i mit kleinem Abstand %i zu %i anderen Ecken \n", rplus, maxdist, inccounter);
    }
  free(BitRow);
  free(incidence);
  return 0;
}

static void AugmentR(int rnew, USI Equality)
{
  USI i;
  /*
  printf("***********AugmentR********\n");
  printf("rnew, NbRay %i %i \n", rnew, NbRay);
  */

  for( i = NbLines + 1 ; i<NbRay; i++)
    {
      if(Ray->V[i][0] != 0)
	{
	  /* printf("combine row %i with row %i\n", i, rnew);*/
	  SimpleCombine(i, rnew, i, Ray);
	}      
    }
  if (Equality)
    {
      RowExchange(Ray, rnew, NbRay-1, raytemprow);
      RowExchange(Sat, rnew, NbRay-1, sattemprow);
    }
}


/* RaySort -- Sorts the rays (Ray, Sat) into three tiers:
         NbBid       <= i < equal_bound : saturates the constraint
         equal_bound <= i < sup_bound   : verifies the constraint
         sup_bound   <= i < NbRay       : does not verify
   the tier order is chosen so non-verifying rays can be easily removed 

*/
static void RaySort (int *equal_bound,int *sup_bound)
{ 
  int inf_bound;
  int eq, sup, inf;
  
  /* sort rays : NbLines <= i < equal_bound : saturates the constraint
     : equal_bound <= i < sup_bound : verifies the constraint
     : sup_bound <= i < NbRay : does not verify */
  /* the order is chosen so non-verifying rays can be easily removed */
  
  *sup_bound = *equal_bound = NbLines;
  sup = eq = NbLines;
  inf_bound = NbRay;
  inf = NbRay;
  
  while (inf_bound>*sup_bound)
    {
      if (Ray->V[sup][0]==0)       /* status = satisfy */
        {
	  RowExchange(Ray, eq, sup, raytemprow);
          RowExchange(Sat, eq, sup, sattemprow);
          (*equal_bound)++; eq++;
          (*sup_bound)++; sup++;
        }
      else
        {
          if (Ray->V[sup][0]<0) {    /* status = verify */
            inf_bound--; inf--;
            RowExchange(Ray, inf, sup, raytemprow);
            RowExchange(Sat, inf, sup, sattemprow);
          }
          else                  /* status = does not verify */
	    {
	      (*sup_bound)++; sup++;
	    }
        }
    }
}

static int ConstrainR2(USI step, USI wordcount, USI bitcount, USI *numrays, USI Equality )
{
  USI redundant = 1;
  USI Word, Word2;
  int thisword;
  int otherRay; 
  USI oldrplus, oldrminus;
  USI rp, rm;
  USI i,k;
  USI plusrays, minusrays, zerorays;
  USI rminus_lowbound;
  USI RowSize;
  USI rplus, rminus, newray;
  USI equal_bound, sup_bound, bound;
  USI nbcommonconstraints;
  int *ptr1, *ptr2;
  register USI r0, r1, r2, r3;
#ifdef SHOWTIME
  float oldpercent=0.0, percent=0.0;
#endif
  OldRays=0;
 
  /* printf("**********ConstrainR********\n");*/
  
  /*needed by Adjacency Test*/
  AdjWCount = (step  +1) / WSIZE;
  AdjBCount = MSB;
  AdjBCount >>= ((step +1 ) %WSIZE);
  
  AdjBitmask = (AdjBCount - 1); 
  AdjBitmask = AdjBitmask<<1;
  AdjBitmask++;
  
  start_timer (3);
  
  RaySort (&equal_bound, &sup_bound);
  
  stop_timer (3);
  
  /*        Ray 
  NbRay-> bound-> ________
                  |  R-  |        R- ==> ray.eq < 0  (outside domain)
	    sup-> |------|
	          |  R+  |        R+ ==> ray.eq > 0  (inside domain)
          equal-> |------|
                  |  R0  |        R0 ==> ray.eq = 0  (on face of domain)
        NbLines-> |______|
        */

  // Coarsify(NbLines, NbRay );

  for(k=equal_bound; k< NbRay; k++)
    {
      Sat->V[k][wordcount]  |=bitcount; 
      /*set saturation bit to 1 for rays 
	that do not satisfy the constraint*/
    }
  
#ifdef DEBUG
 printf("SortRays\n"); 
 PartitionPrint(stdout,"%3d", Ray, NbLines, equal_bound, sup_bound, NbRay, Dimension);
#endif


 bound=NbRay;
 RowSize=Sat->C * sizeof(int);
 BitRow=(int *)malloc(RowSize);

#ifdef SHOWTIME 
 printf("Performing %i x %i tests on %i rays\n", sup_bound-equal_bound, bound-sup_bound, NbRay);
#endif
 
 /*separate rays from vertices:
   2 rays (r+ r-) are never made redundant by a vertex 
  */

 zerorays = NbLines;
 for (i = NbLines; i<equal_bound; i++)
   {
     if (Ray->V[i][Dimension]==0)
       {
	 RowExchange(Ray, i, zerorays, raytemprow);
	 RowExchange(Sat, i, zerorays, sattemprow);
	 zerorays++;
       }
   }

 plusrays = equal_bound;
 for (i = equal_bound; i<sup_bound; i++)
   {
     if (Ray->V[i][Dimension]==0)
       {
	 RowExchange(Ray, i, plusrays, raytemprow);
	 RowExchange(Sat, i, plusrays, sattemprow);
	 plusrays++;
       }
   }

 minusrays = sup_bound;
 for (i = sup_bound; i<bound; i++)
   {
     if (Ray->V[i][Dimension]==0)
       {
	 RowExchange(Ray, i, minusrays, raytemprow);
	 RowExchange(Sat, i, minusrays, sattemprow);
	 minusrays++;
       }
   }
 
 /* printf("There are %i/%i rays in r0, %i/%i rays in rplus, and %i/%i rays in rminus\n",
	zerorays-NbLines, equal_bound-NbLines,
	plusrays-equal_bound, sup_bound-equal_bound,
	minusrays-sup_bound, bound-sup_bound);
 */

 /*first check rays only*/

 TTEST=equal_bound;
 for (rplus=equal_bound; rplus<plusrays; rplus++){     /* for all rays of R+*/
#ifdef SHOWTIME 
   percent = (float)(rplus-equal_bound)/(float)(plusrays-equal_bound)*100.0;
   if(percent-oldpercent>5.0)
     {
       printf(" %2.2f%%", percent);
       fflush(stdout);
       oldpercent = percent;
     }
#endif
   for(rminus=sup_bound; rminus<minusrays; rminus++) {        /* for all rays of R-*/
	//for(rminus=minus_lowbound; rminus<minus_upbound; rminus++) {        /* for all pairs of R-*/
     /*printf("Betrachte Reihen %i und %i:\n", rplus, rminus);*/
     
     adjcounter++;
     
     /*----------------------------------------------------------------------*/
     /* compute and count the set of constraints saturated by both R+ and R- */
     /*----------------------------------------------------------------------*/
     
     nbcommonconstraints=1; /* account for pos constr */
     for (thisword=0; thisword<AdjWCount; thisword++) 
       { 
	 BitRow[thisword] = Sat->V[rplus][thisword] | Sat->V[rminus][thisword];/*OR the rows*/
	 Word=Word2=BitRow[thisword];
	 Word >>= (sizeof(int)/2 * 8);
	 Word2 &= 0x0000ffff;
	 nbcommonconstraints+=LookUpMat[Word];
	 nbcommonconstraints+=LookUpMat[Word2];
       }
     /*last word with index [AdjWCount] has only AdjBCount bits:*/
     Word = Word2 = BitRow[AdjWCount] = Sat->V[rplus][AdjWCount] | Sat->V[rminus][AdjWCount] | AdjBitmask;
     
     Word >>= (sizeof(int)/2 * 8);
     Word2 &= 0x0000ffff;
     nbcommonconstraints+=LookUpMat[Word];
     nbcommonconstraints+=LookUpMat[Word2];
	 
     /*-----------------------------------------------------------------*/
     /* Adjacency Test : is combination [R-,R+] a non redundant ray?    */
     /*-----------------------------------------------------------------*/
     if (nbcommonconstraints + NbLines>=Dimension-2) /* dimensionality check*/
       { /* check whether a ray m saturates the same set of constraints */
	 
	 /*copy the two rays to the top of the ray matrix*/
 
	 rp = rplus;
	 rm = rminus;
	 
	 RowExchange(Sat, rminus, sup_bound, sattemprow);
	 RowExchange(Ray, rminus, sup_bound, raytemprow);
	 
	 RowExchange(Sat, rplus, equal_bound , sattemprow);
	 RowExchange(Ray, rplus, equal_bound, raytemprow);

	 rplus = equal_bound;
	 rminus = sup_bound;

	 redundant=0;

#include "adjtestpolyrays_macro.h"
  	 
	 if (!redundant)  {
	   /*------------------------------------------------------------*/
	   /* Add new ray generated by [R+,R-]                           */
	   /*------------------------------------------------------------*/
	   
	   stop_timer (2); /*adjtest*/
	   
	   RowExchange(Sat, rm, sup_bound, sattemprow);
	   RowExchange(Ray, rm, sup_bound, raytemprow);
	   RowExchange(Sat, rp, equal_bound , sattemprow);
	   RowExchange(Ray, rp, equal_bound, raytemprow);

	   rplus = rp; rminus = rm;
	   
	   redcounter++;
	   /*Check the size of the RaySpace
	     allocate new memory if necessary*/
	   if(abs(NbRay-OldRays)>RESIZE_ADDROWS)
	     {
	       OldRays=NbRay;
	       printf("Currently %i Rays\n", NbRay);
#ifdef DEBUG
	       printf("Currently %i Rays\n", NbRay);
	       printf ("Adjacency test called %i times, nb of not adj. pairs: %i\n", adjcounter, adjcounter - redcounter);
#endif
	     }
	   /*printf("numrays, NbRay %i %i \n", (*numrays), NbRay);*/
	   if (NbRay>=(*numrays))
	     {
	       if (NbRay>=NbMaxRays) 
		 {
		   printf("too many rays\n");
		   printf("NbRay %i NbMaxRays %i\n", NbRay, NbMaxRays);
		   return 1; /*too many rays, return with error*/
		 }
	       ResizeRows(&Ray, RESIZE_ADDROWS); /*add rows to Ray and Sat*/
	       ResizeRows(&Sat, RESIZE_ADDROWS);
	       (*numrays)+=RESIZE_ADDROWS;
	     }
	   
	   /* computes the new ray */
	   /*NbRay = combination of rminus and rplus*/
	   
	   start_timer (4); /*combine*/
	   
	   SimpleCombine(rminus, rplus, NbRay, Ray);
	     
#ifdef DEBUG
	   printf("Rays %i and %i are adjacent.\n", rminus, rplus);
	   printf("Compute Combine: Combine Ray %i with %i -> %i\n",rminus, rplus, NbRay );
#endif	      
	   /*printf("rayspace, sat:\n");
	     PRINTMATI(0,NbRay-1,0,Ray->C-1,Ray);
	     
	     Rays_Print(stdout,"%3d", Ray, NbRay, Dimension);
	   */
	   
	   RowOr(rminus, rplus, NbRay, Sat); 
	   Sat->V[NbRay][wordcount]&=~bitcount;
	   
	   stop_timer (4); /*combine*/
	   
	   NbRay++; 
	 } 
	 else /*redundant*/
	   {
	   is_redundant2:
	 
	     RowExchange(Sat, rm, sup_bound, sattemprow);
	     RowExchange(Ray, rm, sup_bound, raytemprow);
	     RowExchange(Sat, rp, equal_bound , sattemprow);
	     RowExchange(Ray, rp, equal_bound, raytemprow);
	     
	     rplus = rp; rminus = rm;
	     
	   }
       } /*if dimensionality check*/
   }/*for rminus*/
 }/*for rplus*/
#ifdef SHOWTIME 
 printf("\n");
#endif 

 /*now check combinations ray x vertex , vertex x ray, vertex x vertex*/
 TTEST = NbLines;
 for (rplus=equal_bound; rplus<sup_bound; rplus++){     /* for all pairs of R+*/
#ifdef SHOWTIME 
   percent = (float)(rplus-equal_bound)/(float)(sup_bound-equal_bound)*100.0;
   if(percent-oldpercent>5.0)
     {
       printf(" %2.2f%%", percent);
       fflush(stdout);
       oldpercent = percent;
     }
#endif
   if(rplus<plusrays) rminus_lowbound = minusrays; /*compare +rays with -vertices only*/
   else rminus_lowbound = sup_bound; /*compare +vertices with all from - */
   for(rminus=rminus_lowbound; rminus<bound; rminus++) {        /* for all pairs of R-*/
	//for(rminus=minus_lowbound; rminus<minus_upbound; rminus++) {        /* for all pairs of R-*/
     /*printf("Betrachte Reihen %i und %i:\n", rplus, rminus);*/
     
     adjcounter++;
     
     /*----------------------------------------------------------------------*/
     /* compute and count the set of constraints saturated by both R+ and R- */
     /*----------------------------------------------------------------------*/
     
     nbcommonconstraints=0;
     for (thisword=0; thisword<AdjWCount; thisword++) 
       { 
	 BitRow[thisword] = Sat->V[rplus][thisword] | Sat->V[rminus][thisword];/*OR the rows*/
	 Word=Word2=BitRow[thisword];
	 Word >>= (sizeof(int)/2 * 8);
	 Word2 &= 0x0000ffff;
	 nbcommonconstraints+=LookUpMat[Word];
	 nbcommonconstraints+=LookUpMat[Word2];
       }
     /*last word with index [AdjWCount] has only AdjBCount bits:*/
     Word = Word2 = BitRow[AdjWCount] = Sat->V[rplus][AdjWCount] | Sat->V[rminus][AdjWCount] | AdjBitmask;
     
     Word >>= (sizeof(int)/2 * 8);
     Word2 &= 0x0000ffff;
     nbcommonconstraints+=LookUpMat[Word];
     nbcommonconstraints+=LookUpMat[Word2];
	 
     //   if (Ray->V[rplus][Dimension]==0 && Ray->V[rminus][Dimension]==0)
     //   nbcommonconstraints++;      /* account for pos constr */

     /*-----------------------------------------------------------------*/
     /* Adjacency Test : is combination [R-,R+] a non redundant ray?    */
     /*-----------------------------------------------------------------*/
     if (nbcommonconstraints + NbLines>=Dimension-2) /* dimensionality check*/
       { /* check whether a ray m saturates the same set of constraints */
	 
	 //	 rayonly  =(Ray->V[rplus][Dimension]==0 && Ray->V[rminus][Dimension]==0); /*no vertices*/
	 
	 rp = rplus;
	 rm = rminus;
	 
	 /*copy the two rays to the top of the ray matrix*/
	 
	 if (rplus == TTEST)
	   {
	     oldrplus = rplus;
	     rplus = rminus;
	     rminus = oldrplus;
	   }
	 
	 oldrminus = rminus;
	 RowExchange(Sat, rminus, TTEST, sattemprow);
	 RowExchange(Ray, rminus, TTEST, raytemprow);
	 rminus = TTEST;
	 
	 oldrplus = rplus;
	 RowExchange(Sat, rplus, TTEST+1, sattemprow);
	 RowExchange(Ray, rplus, TTEST+1, raytemprow);
	 rplus = TTEST+1;
	 
	 redundant=0;
	 
	 if(AdjWCount==0)
	   {
#include "adjtestpoly_macro0.h"
	   }  
 	 else if (AdjWCount==1)
	   {
#include "adjtestpoly_macro1.h"	   
	   } 
	 else if (AdjWCount==2) 
	   { 
#include "adjtestpoly_macro2.h"	   
	   } 
	 else if (AdjWCount==3) 
	   { 
#include "adjtestpoly_macro3.h" 
	   } 
	 else 
	   { 
#include "adjtestpoly_macro.h" 
	   } 
	 
	 if (!redundant)  {
	   /*------------------------------------------------------------*/
	   /* Add new ray generated by [R+,R-]                           */
	   /*------------------------------------------------------------*/
	   
	   stop_timer (2); /*adjtest*/
	   
	   
	   RowExchange(Sat, oldrplus, rplus, sattemprow);
	   RowExchange(Sat, oldrminus, rminus, sattemprow);
	   RowExchange(Ray, oldrplus, rplus, raytemprow);
	   RowExchange(Ray, oldrminus, rminus, raytemprow);
	   
	   rplus = rp; rminus = rm;
	   
	   redcounter++;
	   /*Check the size of the RaySpace
	     allocate new memory if necessary*/
	   if(abs(NbRay-OldRays)>RESIZE_ADDROWS)
	     {
	       OldRays=NbRay;
	       printf("Currently %i Rays\n", NbRay);
#ifdef DEBUG
	       printf("Currently %i Rays\n", NbRay);
	       printf ("Adjacency test called %i times, nb of not adj. pairs: %i\n", adjcounter, adjcounter - redcounter);
#endif
	     }
	   /*printf("numrays, NbRay %i %i \n", (*numrays), NbRay);*/
	   if (NbRay>=(*numrays))
	     {
	       if (NbRay>=NbMaxRays) 
		 {
		   printf("too many rays\n");
		   printf("NbRay %i NbMaxRays %i\n", NbRay, NbMaxRays);
		   return 1; /*too many rays, return with error*/
		 }
	       ResizeRows(&Ray, RESIZE_ADDROWS); /*add rows to Ray and Sat*/
	       ResizeRows(&Sat, RESIZE_ADDROWS);
	       (*numrays)+=RESIZE_ADDROWS;
	     }
	   
	   /* computes the new ray */
	   /*NbRay = combination of rminus and rplus*/
	   
	   start_timer (4); /*combine*/
	   
	   SimpleCombine(rminus, rplus, NbRay, Ray);
	     
#ifdef DEBUG
	   printf("Rays %i and %i are adjacent.\n", rminus, rplus);
	   printf("Compute Combine: Combine Ray %i with %i -> %i\n",rminus, rplus, NbRay );
#endif	      
	   /*printf("rayspace, sat:\n");
	     PRINTMATI(0,NbRay-1,0,Ray->C-1,Ray);
	     
	     Rays_Print(stdout,"%3d", Ray, NbRay, Dimension);
	   */
	   
	   RowOr(rminus, rplus, NbRay, Sat); 
	   Sat->V[NbRay][wordcount]&=~bitcount;
	   
	   stop_timer (4); /*combine*/
	   
	   NbRay++; 
	 } 
	 else /*redundant*/
	   {
	   is_redundant:
	     RowExchange(Sat, oldrplus, rplus, sattemprow);
	     RowExchange(Sat, oldrminus, rminus, sattemprow);
	     RowExchange(Ray, oldrplus, rplus, raytemprow);
	     RowExchange(Ray, oldrminus, rminus, raytemprow);
	     
	     rplus = rp; rminus = rm; 
	   }
       } /*if dimensionality check*/
   }/*for rminus*/
 }/*for rplus*/
#ifdef SHOWTIME 
 printf("\n");
#endif 

 /* Eliminates all rays from R-
    moves the new rays */

 if (Equality)
   rminus = equal_bound; /*remove R- and R+ */
 else
   rminus = sup_bound; /*remove R- */
 newray = NbRay;
 while ((rminus<bound)&&(newray>bound)) {
   newray--;
   CopyRow(Ray, newray, rminus);
   CopyRow(Sat, newray, rminus);
   rminus++;
 }
 if (rminus==bound) NbRay=newray;
 else NbRay=rminus;
 free(BitRow);
 return 0;
}

static int Dual(int startrow, int endrow, int ISSOLVABLE)
{
  USI k;
  USI rnew;
  int result = 0;
  int j;
  USI firsttime = 1;
  USI Equality;
  USI numrays = NbRay;
  USI validvertex;
  USI wordcount =   startrow/WSIZE;
  USI  bitcount = MSB; bitcount >>= startrow%WSIZE;
  
  Dimension =  Constraints->C - 1;
  NbMaxRays = 500000;
  
  /*allocate memory for temporary rows*/
  raytemprow = (int *)malloc(Ray->C  * sizeof(int));
  sattemprow = (int *)malloc(Sat->C  * sizeof(int)); 
  
  for (k=startrow; k<endrow; k++)
    {
#ifdef SHOWTIME 
      printf("Constraint no. %i of %i\n", k, endrow-1);
#endif

      EvalConstraint(Constraints->V[k]);    
      rnew = ConstrainL(k);
      
      Equality = (Constraints->V[k][0] == 0);
      if(rnew!=-1)
	{
	  /*update the saturation matrix for the new ray*/
	  for (j=0; j<Sat->C; j++)
	    {
	      Sat->V[rnew][j]=0;
	    }
	  /* new ray saturates everything except last inequality */
	  Sat->V[rnew][wordcount] |= bitcount;
	  
	  AugmentR(rnew, Equality);
	  
	if(Equality) NbRay--;
	}
      else
	{
	  if ((NbLines == 0) && firsttime)
	    {
	      firsttime = 0;
	    }
	  result = ConstrainR2(k, wordcount, bitcount, &numrays, Equality);
	  
	  if(ISSOLVABLE)
	    {
	      validvertex = CheckForStartingVertex(0, startrow, endrow);
	      if(validvertex != -1)
		{
#ifdef DEBUG
		  printf("valid vertex found: step %i ray %i\n", k, validvertex);
		  PRINTMATI(validvertex, validvertex, 0, Ray->C-1, Ray);
#endif
		  return 1;

		}
	}
	}
      if (result == 1)
	{
	  printf("fehler!");
	  break;
	}
      NEXT(wordcount, bitcount); /*increase counter by one bit*/
    }
  
  PartFreeMat(&Ray, NbRay);
  PartFreeMat(&Sat, NbRay);
  
  if(Ray->R == 0)
    {
      printf("contradiction!? no solution\n");
      return 0;
    }
  
  free(raytemprow);
  free(sattemprow);
  return 1;
}


int PolyEnum (INTMATRIX *InputConstraints, int vars)
{
  INTMATRIX *ExtendedRay;
  INTMATRIX *SimpleConstraints;
  int solution;
  USI count;
  USI i;
  int errorflag = 0;
  int startrow, endrow;
  USI NbOldRays = 0;
  
  int thisdomain;
  int thisline;
  
  INTMATRIX **LoopDomain;

#ifdef PRINTDOMAINS
  char filename[50];
  FILE *outfile;
   strcpy(filename, "Constraints.out");
  outfile = fopen(filename, "w");
#endif  

  reset_all_timer ();
  start_timer (0);
   
  Constraints = CopyMat(InputConstraints);

#ifdef PRINTDOMAINS
  fprintf(outfile,"POLYCOUNT: CONSTRAINTS (INPUT MATRIX) \n");
  FPRINTMATI(outfile,0,Constraints->R-1,0,Constraints->C-1,Constraints);
  fclose(outfile);
#endif
  
  /*
    printf("++++++++++++++++++++PolyEnum INPUT:++++++++++++++++++++++++\n");
    PRINTMATI(0,Constraints->R-1,0,Constraints->C-1,Constraints);
    printf("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
  */
  
  Dimension =  Constraints->C - 1;
  nc  = ( Constraints->R - 1 ) / (sizeof(int)*8) + 1;

  /*Lookup-table used for adjacency testing*/
  CreateLookUpMat();
  
  /*maximum of Dimension - 1 variables*/
  if (vars > Dimension - 1 )
    vars = Dimension -1;
  /*
    printf("Variables: %i \t Markings: %i \t Constraints: %i \n", Constraints->C, vars, Constraints->R);
  */
  NbLines = Init(); /*initialize the Rayspace /Saturation matrix*/
  NbRay = NbLines +1;
  
  LoopDomain =  (INTMATRIX **)malloc((Dimension + 2) * sizeof(INTMATRIX *));
  Marker = (INTMATRIX **)malloc((Dimension + 2) * sizeof(INTMATRIX *));
 
  startrow = 0;
  endrow = Constraints->R;

   start_timer(1);

  Dual(startrow, endrow, 0);
  if (Ray->R == 0)
    {
      printf("Error! Set of constraints has no solution!\n");
      return 0;
    }
  
  /*  printf("Anzahl rays: %i \n", CountRays(Ray));  */
  
#ifdef DEBUGCOUNT
  printf("1. Step: from the input-constraints generate the ray matrix \n");
  printf("First step (ray creation): %8.2f s\n", time_passed(1) );
  printf("Size of ray matrix: %i \n", Ray->R);
  PRINTMATI(0,Ray->R-1,0,Ray->C-1,Ray);
#endif  
  
  
  /*write status words for the ray matrix*/
  WriteStatus(Constraints, Ray, Sat);
  
  stop_timer(1);
  

  /*generate simplified constraints
    returns no. of lines in the new constraint matrix*/
    
  Remove_Redundant_Constraints(&Constraints, Ray, Sat);
  SimpleConstraints = Constraints;
  Constraints = NULL;

  NbLines = 0;
  for ( i = 0; i < SimpleConstraints->R; i++)
    {
      if (!(SimpleConstraints->V[i][0]))
	NbLines++;
    }
  
#ifdef DEBUGCOUNT
  printf("2. Step: Simplify Constraints \n");
  PRINTMATI(0,SimpleConstraints->R-1,0,SimpleConstraints->C-1,SimpleConstraints);
#endif
  
  /*add new lines to ray matrix*/
  ExtendedRay = AddLines(Ray, Dimension - 2);
  
  NbOldRays = Ray -> R;
  
  FreeMat(Ray);
  FreeMat(Sat);
  /*build new sat matrix extended by Dimension - 2 columns 
    (one for each new line in the ray matrix)*/
  
  Sat = BuildSat(ExtendedRay, SimpleConstraints, NbOldRays);
  
  /*write loop bounds for innermost loop*/
  LoopDomain[Dimension - 1] = CopyMat(SimpleConstraints);
  LoopDomain[Dimension] = NULL; 
  
  count = Dimension - 2;
  NbRay = SimpleConstraints->R;
  /*for each new line
    from inner loops to outer loops
  */
  for(thisline=NbOldRays; thisline<ExtendedRay->R; thisline++)
    { 
      /*augment the Constraints and Sat - matrix
	ExtendedRay contains the original ray matrix plus Dimension - 2 lines
	and stays unchanged by Dual
	each step, one new line (thisline) is considered; the Constraint and Sat
	matrix are changed accordingly to fit the new constraint 
	NbRay contains the number of rows of the Constraints - matrix, NbLines the
	number of lines (equations) in the constraints - matrix 
      */  

      Constraints = ExtendedRay;
      Ray = SimpleConstraints;
      Dual(thisline, thisline + 1, 0);
      
      /*compute the status words for the constraint matrix*/
      WriteStatusConstraints2(SimpleConstraints, ExtendedRay, Sat, ExtendedRay -> R);
      
#ifdef DEBUGCOUNT
      printf("Constraint - matrix for loop %i \n", count);
      PRINTMATI(0, SimpleConstraints->R - 1, 0, SimpleConstraints->C-1, SimpleConstraints);
#endif
      LoopDomain[count] = CopyMat(SimpleConstraints);
      count--;
    }
  
  for(thisdomain =vars + 1; thisdomain<= Dimension ; thisdomain++)
    {
      Constraints = LoopDomain[thisdomain];
      FreeMat(Constraints);
    }

#ifdef PRINTDOMAINS
  strcpy(filename, "LoopDomain.out");
  outfile = fopen(filename, "w");  
  fprintf(outfile,"*******************+LOOP_DOMAIN+**********************\n"); 
#endif
  for(thisdomain = 1; thisdomain<= vars ; thisdomain++)
    {
#ifdef PRINTDOMAINS
      fprintf(outfile,"----------------------------------------------------\n");
#endif
      Constraints = LoopDomain[thisdomain];
#ifdef PRINTDOMAINS
      FPRINTMATI(outfile, 0, Constraints->R - 1, 0, Constraints->C-1, Constraints);
#endif
    }
  
  SimplifyLoopBounds(LoopDomain, vars);
  SimplifyLoopBounds2(LoopDomain, &vars);
  SimplifyLoopBounds(LoopDomain, vars);
  SimplifyLoopBounds2(LoopDomain, &vars);
  
  SimplifyLoopBounds4(LoopDomain, vars);
  SimplifyLoopBounds2(LoopDomain,  &vars);
  SimplifyLoopBounds5(LoopDomain, vars);
  
#ifdef PRINTDOMAINS
  fprintf(outfile,"*******************+LOOP_DOMAIN+**********************\n"); 
#endif
  for(thisdomain = 1; thisdomain<= vars ; thisdomain++)
    {
#ifdef PRINTDOMAINS
      fprintf(outfile,"----------------------------------------------------\n");
#endif
      Constraints = LoopDomain[thisdomain];
#ifdef PRINTDOMAINS
      FPRINTMATI(outfile, 0, Constraints->R - 1, 0, Constraints->C-1, Constraints);
#endif
    }
#ifdef PRINTDOMAINS
  fclose(outfile);
#endif
  
  solution = ProgramOutput(LoopDomain, vars);

  for(thisdomain =1; thisdomain <= vars ; thisdomain++)
    {
      Constraints = LoopDomain[thisdomain];
      if(Constraints->V[0][0] == 0)
	Marker[thisdomain] = ZeroMatrix(Constraints->R, 2);
      else
	Marker[thisdomain] = ZeroMatrix(Constraints->R, 1);
    }
  
  if (errorflag)
    {
      solution = -1;
    }
  else
    
    /*free memory*/
    for(thisdomain =1; thisdomain<= vars ; thisdomain++)
      {
	Constraints = LoopDomain[thisdomain];
	FreeMat(Constraints);
      }
  free(LoopDomain);
  
  stop_timer (0);
  
  printf ("Time passed with computation: %8.2f s\n", time_passed(0) );
  /*
    printf ("First step (ray creation): %8.2f s\n", time_passed(1) );
    printf ("Second step (simplify constraints): %8.2f s\n", time_passed(8) );
    printf ("third step (build loop domains): %8.2f s\n", time_passed(9) );
    
    printf ("3. Init                  : %8.2f s\n", time_passed(11) );
    printf ("3. Dual                  : %8.2f s\n", time_passed(12) );
    printf ("3. WriteStatusConstraints: %8.2f s\n", time_passed(13) );
    
    
    printf ("final step (counting): %8.2f s\n", time_passed(10) );
    printf ("4. lowerupperbounds: %8.2f s\n", time_passed(14) );
    printf ("4. scalar product: %8.2f s\n", time_passed(15) );
  */
  return solution;
  
}


int TestConstraints(INTMATRIX *InputConstraints)
{
  USI i;
  int startrow, endrow;
  Constraints = CopyMat(InputConstraints);
  Dimension =  Constraints->C - 1;
  nc  = ( Constraints->R - 1 ) / (sizeof(int)*8) + 1;

  /*Lookup-table used for adjacency testing*/
  CreateLookUpMat();
  
  NbLines = Init(Constraints, &Ray, &Sat); /*initialize the Rayspace /Saturation matrix*/
  NbRay = NbLines +1;

  startrow = 0;
  endrow = Constraints->R;

  Dual(startrow, endrow, 1);

#ifdef DEBUG
  printf("Ray matrix calculated by TestConstraints \n");
  PRINTMATI(0,Ray->R-1,0,Ray->C-1,Ray);
#endif  
  
  /*check if there are any solutions (vertices)
   solutions must be in IN */
  for(i=0;i<Ray->R;i++)
    {
      if(Ray->V[i][Ray->C-1]==1)
	return 1;
    }
  
  return 0;
}


