/*****************************************************************************

Programm	:	reachable

Komponente	:	reachable.c
                        Test auf Nichterreichbarkeit

Programmer	:	Alexander Huck
Date		:	1999

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "struct_types.h"			/* definitions and types */
#include "struct_vars.h"			/* global definitions */

#include "matrix.h"
#include "matrixIO.h"
#include "matrixOps.h"
#include "matrixProcs.h"
#include "hermite2.h"
#include "inequalities.h"
#include "timer.h"
#include "polycount.h"

#define MAXROWSIZE 1000000   /*Maximum number of matrix-rows allowed during inequality solving*/

/*#define DEBUG*/

FILE 		*inc_file, *mnull_file, *md_file;

/*******************************************************************************************
check if marking MD is reachable from Mnull with incidence matrix A 
*******************************************************************************************/
/*obsolete - this test (Reachable) is based on the cutting plane algorithm implemented 
  in inequalities.c, it is now done by a call of lp_solve via CallSolver in this file*/
int Reachable(INTMATRIX *A, INTMATRIX *Mnull, INTMATRIX *MD)
{
  INTMATRIX *MDelta;
  INTMATRIX *Constraints;
  USI i,j;
  double zeit = 0.0;
  int reachable = 0;
  int row;
  USI nbtransitions = 0, nbplaces = 0;
  USI  nbvars = 0, nbconstraints = 0;
  int docare = 0;   
#ifdef DEBUG
  char filename[50], filename2[50];
  FILE *outfile;
  FILE *outfile2;
  strcpy(filename, "test.out");
  printf("filename : %s\n", filename);
  outfile = fopen(filename, "w");
  
  strcpy(filename2, "zbasis.out");
  printf("filename : %s\n", filename2);
  outfile2 = fopen(filename2, "w");
  
  printf("REACHABLE: INPUT INCIDENCE MATRIX \n");
  PRINTMATI(0,A->R-1,0,A->C-1,A);
  fprintf(outfile,"REACHABLE: INPUT INCIDENCE MATRIX \n");
  FPRINTMATI(outfile,0,A->R-1,0,A->C-1,A);
#endif
  
  reset_timer (0);
  start_timer (0);
  printf("Starting Reachability Computation :\n");
  
  if(Mnull->R == 0 || MD->R ==0 || MD->R != Mnull->R)
    {
      printf("problem with mnull/md\n");
      exit(1);
    }
  
  
  MDelta=ZeroMatrix(MD->R, 1);
  for (i=0; i<MD->R; i++)
    {
      MDelta->V[i][0]= Mnull->V[i][0];
      if( MD->V[i][0] != -1)
	docare++;
    }
  
#ifdef DEBUG
  printf("MDelta MATRIX \n");
  PRINTMATI(0,MDelta->R-1,0,MDelta->C-1,MDelta);
  fprintf(outfile,"MDelta MATRIX \n");
  FPRINTMATI(outfile,0,MDelta->R-1,0,MDelta->C-1,MDelta);
#endif	
  
  nbtransitions = A -> R;
  nbplaces = A -> C;
  
  /*derive the number of variables*/
  
  nbvars = nbplaces; /*one variable for each place*/
  nbvars += nbtransitions; /*characteristic vector sigma*/
  
  /*derive the number of constraints*/
  
   nbconstraints += nbplaces; /*equations for M = M0 +C.Sigma =0*/
  nbconstraints += 2*nbplaces; /*equations for M = M0 +C.Sigma =0*/

  nbconstraints += nbplaces + nbtransitions; /*inequalities for M>=0 and Sigma>= 0*/
  nbconstraints += docare;
  
  /*init matrix for constraints*/
  Constraints = ZeroMatrix(nbconstraints, nbvars + 2);
  
  /*fix markings*/
  row = 0;
  for(i=0; i< nbplaces; i++)
    {
      if( MD->V[i][0] != -1)
	{
	  Constraints->V[row][1 + i] = 1; /*mark for mi*/
	  Constraints->V[row][nbvars + 1] =  -MD->V[i][0];
	  row++;
	}
    }
  /*equations for M - C.Sigma - M0 =0*/
  
  for(i=0; i< nbplaces; i++)
    {
      Constraints->V[row][1 + i] = 1; /*mark for mi*/
      for(j=0; j< nbtransitions; j++)
	{
	  Constraints->V[row][1 + nbplaces + j] =  -A->V[j][i];
	}
      Constraints->V[row][nbvars + 1] =  -MDelta->V[i][0];
      row++;
    }
  
  /*M>=0, Sigma>= 0, firing_vectors >= 0*/
  
  for(i=0; i< nbplaces+nbtransitions; i++)
    {
      Constraints->V[row][0] = 1; /*mark for inequality*/
      Constraints->V[row][1 + i] = 1; /*mark for S_i*/
      row++;
    }
  
  
#ifdef PRINTCONSTRAINTS
  printf("Constraint-matrix:\n");
  PRINTMATI(0,Constraints->R-1,0,Constraints->C-1,Constraints);
#endif
  
#ifdef DEBUG
  printf("Constraint-matrix:\n");
  PRINTMATI(0,Constraints->R-1,0,Constraints->C-1,Constraints);
#endif
  RemoveZeroRows(&Constraints);
  
  
#ifdef DEBUG	    
  printf("eliminieren: \n");
#endif
  ElimEqualRows(&Constraints);
  ElimSimpleConstraints(&Constraints);
  
#ifdef DEBUG
  printf("sortieren: \n");
#endif
  SortRows(Constraints);
  
#ifdef DEBUG
  printf("Sortiert und eliminiert:\n ");
  PRINTMATI(0,Constraints->R-1,0,Constraints->C-1, Constraints);
  fprintf(outfile,"Sortiert und eliminiert:\n ");
  FPRINTMATI(outfile,0,Constraints->R-1,0,Constraints->C-1,Constraints);
#endif
  
  reachable = TestConstraints(Constraints);
  stop_timer(0);
  zeit=time_passed(0);
#ifdef DEBUG
  
  fprintf (outfile,"Time passed with computation: %8.2f s\n", zeit);
#endif
  printf("Reachability Computation done.\n");
  printf ("Time passed with computation: %8.2f s\n", zeit);
  
  return reachable;
}

/*******************************************************************************************
check if marking MD is reachable from Mnull with incidence matrix A 
calls the program lp_solve to solve linear program
*******************************************************************************************/
int CallSolver(INTMATRIX *A, INTMATRIX *Mnull, INTMATRIX *MD, char *basename)
{
  INTMATRIX *MDelta;
  USI i,j;
  double zeit = 0.0;
  int row;
  USI nbtransitions = 0, nbplaces = 0;
  USI  nbvars = 0, nbconstraints = 0;
  int docare = 0;  
  char lp_name[500], progname[500], resultfilename[500], command[500];
  char *tnetdir;
  FILE *lpfile, *resultfile;
  int first, found = 0;
  char buffer[10000];

#ifdef DEBUG
  printf("REACHABLE: INPUT INCIDENCE MATRIX \n");
  PRINTMATI(0,A->R-1,0,A->C-1,A);
#endif
  tnetdir = getenv("TNETHOME");
  sprintf(progname, "%s/Agnes/scripts/lp_solvescript", tnetdir);
  sprintf(resultfilename, "%s.lpout", basename);
  sprintf(lp_name,"%s.lp", basename);
  sprintf(command, "%s  %s  %s", progname, lp_name, resultfilename);

  if ((lpfile = fopen(lp_name, "w")) == NULL)
    {	
      printf("ERROR: could not open %s for writing \n", lp_name);
      exit(1);
    }

  reset_timer (0);
  start_timer (0);
  printf("Starting Reachability Computation :\n");
  
  if(Mnull->R == 0 || MD->R ==0 || MD->R != Mnull->R)
    {
      printf("problem with mnull/md\n");
      exit(1);
    }
  
  MDelta=ZeroMatrix(MD->R, 1);
  for (i=0; i<MD->R; i++)
    {
      MDelta->V[i][0]= Mnull->V[i][0]; /*copy initial marking to MDelta*/
      if( MD->V[i][0] != -1) /*count entries that are not don't care*/
	docare++;
    }
  
#ifdef DEBUG
  printf("MDelta MATRIX \n");
  PRINTMATI(0,MDelta->R-1,0,MDelta->C-1,MDelta);
#endif	
  
  /*get number of transitions/places from incidence matrix*/
  nbtransitions = A -> R;
  nbplaces = A -> C;
  
  /*derive the number of variables*/
  nbvars = nbplaces; /*one variable for each place*/
  nbvars += nbtransitions; /*characteristic vector sigma*/
  
  /*derive the number of constraints*/
  nbconstraints += nbplaces; /*equations for M = M0 +C.Sigma =0*/
  nbconstraints += 2*nbplaces; /*equations for M = M0 +C.Sigma =0*/

  nbconstraints += nbplaces + nbtransitions; /*inequalities for M>=0 and Sigma>= 0*/
  nbconstraints += docare;
  
  /*write a dummy objective function: minimize m0*/
  fprintf(lpfile, "min: m0;\n");

  /*write equations for fixed markings, e.g. m2 = 5 ;*/
 
  /*
    for(i=0; i< nbplaces; i++)
    {
    if( MD->V[i][0] != -1)
    fprintf(lpfile, " m%i = %i; \n", i, MD->V[i][0]);
    } */
  
  /*equations for M = C.Sigma + M0 */
  first = 1;
  for(i=0; i< nbplaces; i++)
    {
      if( MD->V[i][0] != -1) /*is not don't care*/
	fprintf(lpfile, " %i = ", MD->V[i][0]);
      else
	fprintf(lpfile, " m%i = ", i); /*mark for mi*/
      for(j=0; j< nbtransitions; j++)
	{ /*write C.Sigma */
	  if (A->V[j][i])
	    {
	      first = 0;
	      if(first || A->V[j][i] < 0 )
		fprintf(lpfile, " %i s%i ", A->V[j][i], j);
	      else
		fprintf(lpfile, " + %i s%i ", A->V[j][i], j);
	    }
	}
      fprintf(lpfile, " + %i ; \n ", MDelta->V[i][0]);
      row++;
    }
  
  /*declare variables*/
  fprintf(lpfile, "int ");
  for(i=0; i< nbplaces; i++)
    if( MD->V[i][0] == -1 ) /*is dont care*/
      fprintf(lpfile, " m%i,", i);
  for(j=0; j< nbtransitions-1; j++)
    fprintf(lpfile, " s%i,", j);
  fprintf(lpfile, " s%i;\n", j);
  fclose(lpfile);
  
  system(command);
  
  while( !(resultfile = fopen(resultfilename, "r")))
    {
    }
  
  if(fscanf(resultfile,"%s",buffer)!=1)
    exit(1);
  
  while(fscanf(resultfile,"%s",buffer))
    {
      if (strstr(buffer,"function:"))
	{found = 1; break;}
      if (strstr(buffer,"infeasible"))
	{found = 0; break;}
    }
  stop_timer(0);
  zeit=time_passed(0);
  printf("Reachability Computation done.\n");
  printf ("Time passed with computation: %8.2f s\n", zeit);
  
  return found;
}

static void usage()
{
  printf("Please give name of netfile\n");
  exit(-1);
}

int main(argc, argv)
     int argc;
     char ** argv;
{
  INTMATRIX *Aptr, *Mnullptr, *Mdptr;
  int reachable;
  char inc_name[FILE_NAME], mnull_name[FILE_NAME], md_name[FILE_NAME];
  
  /* check arguments */
  if (argc != 2)
    usage();
  
  /* build filenames */
  sprintf(inc_name,"%s.inc", argv[1]);
  sprintf(mnull_name,"%s.m0", argv[1]);
  sprintf(md_name,"%s.md", argv[1]);
  
  /* open STRUCT & DEFINFO files */
  if ((inc_file = fopen(inc_name, "r")) == NULL)
    {	printf("ERROR: could not open %s for reading\n", inc_name);
    exit(1);
    }
  if ((mnull_file = fopen(mnull_name, "r")) == NULL)
    {	
      printf("ERROR: could not open %s for reading\n", mnull_name);
      exit(1);
    }
  if ((md_file = fopen(md_name, "r")) == NULL)
    {
      printf("ERROR: could not open %s for reading\n", md_name);
      exit(1);
    }
  
  Aptr = Matrix_Read(inc_file);
  fclose(inc_file);
  Mnullptr = Matrix_Read(mnull_file);
  fclose(mnull_file);
  Mdptr = Matrix_Read(md_file);
  fclose(md_file);
  
  //  reachable = Reachable(Aptr, Mnullptr, Mdptr);
  reachable = CallSolver(Aptr, Mnullptr, Mdptr, argv[1]);
  if(reachable)
    printf("The marking is possibly reachable\n");
  else
    printf("The marking is unreachable\n");

  FreeMat(Aptr); FreeMat(Mnullptr); FreeMat(Mdptr);
  
  return 0;			/* o.k */
}
