#ifndef __SHAREDVARS_H__
#define __SHAREDVARS_H__
#include "matrix.h"
#include "lists.h"

/*******************************************************************************************
       definitions
*******************************************************************************************/

#define WSIZE (8*sizeof(int))
#define RESIZE_ADDROWS 1000


/* put a one in the MSB of an int (portable) */
#define MSB ((unsigned)1<<(sizeof(int)*8-1))
#define MSBHALF ((unsigned)1<<(sizeof(int)/2*8-1))
/*next word j, next bit b*/
#define NEXT(j, b) { if (!((b)>>=1)) { (b)=MSB; (j)++; } }
/*previous word j, previous bit b*/
#define PREV(j, b) { if (!((b)<<=1)) { (b)=1; (j)--; } }
#define TOP ((int)(MSB-1))
#define TOPHALF (unsigned)(MSBHALF - 1 + MSBHALF)

/*error types*/
#define OVERFLOWERROR 1
#define MEMORYERROR 2
#define OK 0

/*******************************************************************************************
       global variables
*******************************************************************************************/
extern INTMATRIX *Ray; /*rays*/
extern INTMATRIX *Constraints;/*Constraints*/
extern INTMATRIX *Sat; /*saturation bit-matrix : BITi=0 means Ai.x =0, BITi=1 means Ai.x!=0 for Constraint Ai*/ 
 
extern USI NbRay; /*last used Element in Ray*/
extern USI NbLines; /*number of Lines in Ray*/
extern USI NbConstraints; /*number of constraints*/
extern USI Dimension; /*homogenous Dimension*/
extern USI nc;

extern INTMATRIX *MinF;/*minimum infeasible index: MinF[i] = Index of first constraint for which A.RAYi<0*/


extern INTMATRIX *IndexLR; /*translate ray-index from LinkedList to Ray*/
extern INTMATRIX *IndexRL; /*translate ray-index from Ray to LinkedList*/


extern LiList **LiListArray;/*Array for the linked lists*/
extern LiList *GenList;/*store the index of the feasible generator of a new ray and the index of the new ray*/
extern FreeLi *FLR; /*free entrys in the index list*/
extern USI blockflag; /*=0 if there are single rows to be filled*/
/*=1 if there are no single rows : allocate a new memory block for new rows*/
extern USI NbIndexLR; /*last used Element in IndexLR*/
extern USI NbIndexRL;

extern USI OldRays; /*for debugging output only*/
extern USI NbMaxRays; /*maximum number of rays allowed during computation*/
 
extern USI WCount, BCount;
extern USI AdjWCount, AdjBCount, AdjBitmask; /*needed by adjacency test*/
extern int *AdjBitRow, *BitRow, *BitRow2, *BitRow3;
extern int *raytemprow, *sattemprow,*minftemprow;
extern int *Border;
extern int *LookUpMat;
extern USI TTEST;
extern USI anztest, checkedlater, secondpart;
extern USI adjcounter, redcounter, quickcounter;
extern USI seekray, seek;
extern USI totaltests;
extern int maxMinF;

#endif
