/*****************************************************************************

Programm	:	estimse

Komponente	:	simplifyBounds.c
                        loopbounds vereinfachen

Programmer	:	Alexander Huck
Date		:	1999

*****************************************************************************/

#include <malloc.h>
#include "sharedVars.h"
#include "intOps.h"
#include "matrixOps.h"
#include "remove_redundants.h"

void SimplifyLoopBounds(INTMATRIX **LoopDomain, int vars)
     /*delete all constraints of LoopDomain i
       that have coefficient a_i=0 and therefore 
       does not affect the LoopBound.
       Then sort inequalities and equations and perform
       Gaussian elimination on both
     */
{  
  INTMATRIX *LocalConstr;
  int thisvar, row;
  int bottom;
  
  raytemprow = (int *)malloc(LoopDomain[1] -> C  * sizeof(int));
  /*for all loop-domains
    each loop domain belongs to one variable*/
#ifdef DEBUGSIMPLIFY
  printf("Simplify Loop Bounds\n");
#endif
  for(thisvar = 1; thisvar<= vars ; thisvar++)
    {
      LocalConstr = LoopDomain[thisvar];
      /*
	printf("Unreduced constraint, variable = %i \n", thisvar);
	PRINTMATI(0, LocalConstr->R - 1, 0, LocalConstr->C-1, LocalConstr);
      */
      bottom = LocalConstr->R;
      /*for all constraints of this loop-domain*/
      for(row=0; row< LocalConstr->R; row++)
	{
	  if (row==bottom) break;
	  
	  if(LocalConstr->V[row][thisvar]==0)
	    /*if the coefficient of variable thisvar is 0,
	      then the constraint row does not affect thisvar
	      and can be deleted*/
	    {
	      /*swap to the bottom*/
	      RowExchange(LocalConstr, row, bottom-1, raytemprow);
	      bottom--;
	      row--;
	    }
	}
      /*free rows at bottom*/
      PartFreeMat(&LocalConstr,bottom);
      Remove_Redundant_Constraints2(&LocalConstr);
    }
  free(raytemprow);
}

void SimplifyLoopBounds2(INTMATRIX **LoopDomain, int *vars)
{  
  INTMATRIX *LocalConstr;
  int thisvar, j, k, row;
  int eq_coeff,ineq_coeff;
  
  /*for all loop-domains
    each loop domain belongs to one variable*/
#ifdef DEBUGSIMPLIFY
  printf("Simplify Loop Bounds 2\n");
#endif
  for(thisvar = 1; thisvar<= (*vars) ; thisvar++)
    {
      LocalConstr = LoopDomain[thisvar];
      /*
	printf("Unreduced constraint, variable = %i \n", thisvar);
	PRINTMATI(0, LocalConstr->R - 1, 0, LocalConstr->C-1, LocalConstr);
      */
      
      if(LocalConstr->V[0][0]==1) /*inequality*/
	continue;
      else{ /*equality*/
	for(row = thisvar + 1; row<= (*vars); row++)
	  {/*remaining constraints*/
	    for(j = 0; j< LoopDomain[row]->R; j++)
	      {
		if(LoopDomain[row]->V[j][thisvar]!=0)
		  {
#ifdef DEBUGSIMPLIFY
		    printf("Inequality %i, Constraint %i dependent on equality-constraint %i\n", j, row, thisvar);
		    PRINTMATI(0, 0, 0, LocalConstr->C-1, LocalConstr);
		    PRINTMATI(j, j, 0,  LoopDomain[row]->C-1,  LoopDomain[row]);
		    printf("remove dependency: \n");
#endif
		    eq_coeff = LocalConstr->V[0][thisvar]; 
		    ineq_coeff = LoopDomain[row]->V[j][thisvar];
		    if(eq_coeff*ineq_coeff>0) /*-- / ++ : sub*/
		      {
			for(k=1; k< LocalConstr->C; k++)
			  {
			    LoopDomain[row]->V[j][k] = abs(eq_coeff)* LoopDomain[row]->V[j][k]\
			      - 	abs(ineq_coeff)* LocalConstr->V[0][k];
			  }
			
			Vector_Normalize( LoopDomain[row]->V[j],LocalConstr->C ); /*gcd reduce*/
		      }
		    else/*+1/-1  add*/
		      {
			for(k=1; k< LocalConstr->C; k++)
			  {
			    LoopDomain[row]->V[j][k] = abs(eq_coeff)* LoopDomain[row]->V[j][k]\
			      +	abs(ineq_coeff)* LocalConstr->V[0][k];
			  }
			Vector_Normalize( LoopDomain[row]->V[j],LocalConstr->C ); /*gcd reduce*/
		      }
#ifdef DEBUGSIMPLIFY
		    PRINTMATI(j, j, 0,  LoopDomain[row]->C-1,  LoopDomain[row]);
#endif
		  }
	      }
	  }
      }
    }
}


void SimplifyLoopBounds4(INTMATRIX **LoopDomain, int vars)
     /*when there is an equation defining the bounds of a loop,
       delete all other constraints */
{  
  INTMATRIX *LocalConstr, *temp;
  int thisvar;
  int col, dindex=0, i, j;
#ifdef DEBUGSIMPLIFY
  int thisdomain;
#endif
  
  for(thisvar = 1; thisvar<= vars ; thisvar++)
    {
      LocalConstr = LoopDomain[thisvar];
      if(LocalConstr->V[0][0]==0)
	{
	  /*calculate dependency index*/
	  for(col = 1; col< thisvar; col++)
	    {
	      if (LocalConstr->V[0][col])
		dindex = col;
	    }
#ifdef DEBUGSIMPLIFY
	  printf("equation %i depends on %i\n", thisvar, dindex);
#endif
	  PartFreeMat(&LocalConstr,1);
	  LoopDomain[thisvar] = LocalConstr;
	}
    }
  
  for(thisvar = 1; thisvar<= vars ; thisvar++)
    {
      LocalConstr = LoopDomain[thisvar];
      if(LocalConstr->V[0][0]==0)
	{
	  /*calculate dependency index*/
	  for(col = 1; col< thisvar; col++)
	    {
	      if (LocalConstr->V[0][col])
		dindex = col;
	    }
	  
	  for(i = dindex+1; i< thisvar; i++)
	    {
	      if(LoopDomain[i]->V[0][0] == 0)
		dindex++;
	      else
		break;
	    }
	  
	  if(dindex<thisvar-1)
	    {
#ifdef DEBUGSIMPLIFY
	      printf("equation %i depends on %i\n", thisvar, dindex);
	      printf("*******************+LOOP_DOMAIN+**********************\n"); 
	      for(thisdomain = dindex; thisdomain<= vars ; thisdomain++)
		{
		  printf("-------Domain: %i------------------------------------------\n", thisdomain);
		  LocalConstr = LoopDomain[thisdomain];
		  PRINTMATI(0, LocalConstr->R - 1, 0, LocalConstr->C-1, LocalConstr);
		}
#endif
	      
	      /*slide equation up */
	      temp =  LoopDomain[thisvar];
	      for(i = thisvar; i>=dindex + 2; i--)
		{
		  LoopDomain[i] = LoopDomain[i - 1];
#ifdef DEBUGSIMPLIFY
		  printf(" LoopDomain[%i] = LoopDomain[%i]\n", i, i - 1); 
#endif
		}
	      LoopDomain[dindex + 1] = temp;
#ifdef DEBUGSIMPLIFY
	      printf(" LoopDomain[%i] = LoopDomain[%i] \n", dindex + 1 , thisvar);
	      
	      printf("slide equation up \n");
	      printf("*******************+LOOP_DOMAIN+**********************\n"); 
	      for(thisdomain = dindex; thisdomain<= vars ; thisdomain++)
		{
		  printf("-------Domain: %i------------------------------------------\n", thisdomain);
		  LocalConstr = LoopDomain[thisdomain];
		  PRINTMATI(0, LocalConstr->R - 1, 0, LocalConstr->C-1, LocalConstr);
		}
	      
	      printf("for all rows %i -- %i :\n", dindex +1, vars -1);
#endif
	      for(i = dindex + 1; i<=vars; i++)
		{
		  for(j = thisvar; j>=dindex+2; j--)
		    {
#ifdef DEBUGSIMPLIFY
		      if(i == dindex +1)
			printf("SwapCols(%i, %i)\n", j, j-1);
#endif
		      SwapCols(j, j - 1, LoopDomain[i] );
		    }
		}
#ifdef DEBUGSIMPLIFY		
	      printf("exchange variables \n");
	      printf("*******************+LOOP_DOMAIN+**********************\n"); 
	      for(thisdomain = dindex; thisdomain<= vars ; thisdomain++)
		{
		  printf("-------Domain: %i------------------------------------------\n", thisdomain);
		  LocalConstr = LoopDomain[thisdomain];
		  PRINTMATI(0, LocalConstr->R - 1, 0, LocalConstr->C-1, LocalConstr);
		}
#endif
	    }
	}
    }
}

void SimplifyLoopBounds5(INTMATRIX **LoopDomain, int vars)
     /*make bounded variable positive for each equation loop bound 
       sort constraints that belong to lower bounds to the top
     */
{  
  INTMATRIX *LocalConstr;
  int thisvar, i, row, col;
  
  raytemprow = (int *)malloc(LoopDomain[1] -> C  * sizeof(int));
  /*for all loop-domains
    each loop domain belongs to one variable*/
  
  for(thisvar = 1; thisvar<= vars ; thisvar++)
    {
      LocalConstr = LoopDomain[thisvar];
      if(LocalConstr->V[0][0] == 0)
	{
	  for(row=0; row< LocalConstr->R; row++)
	    { 
	      if(LocalConstr->V[row][thisvar]<0)
		{
		  for(col = 1; col< LocalConstr->C; col++)
		    {
		      LocalConstr->V[row][col] = -LocalConstr->V[row][col];
		    }
		}
	    }
	}
      else
	{
	  for(row=0; row< LocalConstr->R; row++)
	    { 
	      if(LocalConstr->V[row][thisvar]<0)
		{
		  for(i=row +1; i< LocalConstr->R; i++)
		    {
		      if(LocalConstr->V[i][thisvar]>0)
			{
			  RowExchange(LocalConstr, row, i, raytemprow);
			  break;
			}
		    }
		}
	    }
	}
    }
  free(raytemprow);
}
