#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "polycount.h"
#include "matrix.h"
#include "matrixIO.h"
#include "matrixOps.h"
#include "matrixProcs.h"
#include "hermite2.h"
#include "inequalities.h"
#include "intOps.h"
#include "invars.h"
#include "timer.h"
#include "siphons.h"

#define MAXROWSIZE 100000000   /*Maximum number of matrix-rows allowed during inequation solving*/
#define MIN(a,b) a<b ? a:b

//#define DEBUG

INTMATRIX *Siphon, *SiphonMarking;

/*******************************************************************************************
 Siphons: calculate the siphons of a net given by its initial marking M,
 its incidence-, post-, and pre-matrix.
 The result is stored in the matrix Siphon, where each row is one Siphon.
 Additionaly, the marking of each siphon is stored in SiphonMarking.
 Returns the number of Siphons.
*******************************************************************************************/
int Siphons(INTMATRIX *_M, INTMATRIX *_IncidenceMat, INTMATRIX *_PostMat, INTMATRIX *_PreMat)
{
  INTMATRIX *INEQ, *SOLUTION;
  INTMATRIX *M, *IncidenceMat, *PostMat, *PreMat;
  INTMATRIX *NSigma;
  int sum;
  int count, numsol;
  int i,j;
  USI nbtransitions = 0, nbplaces = 0;
#ifdef DEBUG
  char filename[50];
  FILE *outfile;
#endif
  
  M = CopyMat(_M);
  IncidenceMat = CopyMat(_IncidenceMat);
  PostMat = CopyMat(_PostMat);
  PreMat = CopyMat(_PreMat);
  
  nbtransitions = IncidenceMat -> R;
  nbplaces = IncidenceMat -> C;
  
#ifdef DEBUG
  printf("The net has %i places and %i transitions\n", nbplaces, nbtransitions);
#endif
  
  if (nbplaces == 0) return 0; /*no places, no marking*/
  if (nbtransitions == 0) return 1; /*no transitions: only initial marking*/
  
#ifdef DEBUG
  printf("Incidence matrix: transis ------>\n"); 
  PRINTMATI(0,IncidenceMat->C-1,0,IncidenceMat->R-1,TransMat(IncidenceMat));
  printf("POST matrix: transis ------>\n"); 
  PRINTMATI(0,PostMat->C-1,0, PostMat->R-1,TransMat(PostMat));
  printf("PRE matrix: transis ------>\n"); 
  PRINTMATI(0,PreMat->C-1,0,PreMat->R-1,TransMat(PreMat));
#endif
  
  /*from PRE and POST generate the modified Incidence-matrix NSigma (->siphons)
    in NSigma all entrys belonging to the input-arcs of a transition are
    replaced by the sum of its output-arcs*/
  
  /*build NSigma*/
  NSigma = ZeroMatrix(nbtransitions, nbplaces);
  for(i=0; i<nbtransitions; i++)
    for(j=0; j<nbplaces; j++)
      {
	NSigma->V[i][j] = PostMat->V[i][j];
      }
  
  for(i=0; i<nbtransitions; i++)
    {
      sum = 0;
      /*calculate sum of output-arcs*/
      for(j=0; j<nbplaces; j++)
	sum += PostMat->V[i][j];
      for(j=0; j<nbplaces; j++)
	{
	  if (PreMat->V[i][j] != 0) /*input-arc*/
	    {
	      NSigma->V[i][j] = -sum;
	    }
	}
    }
  
#ifdef DEBUG
  printf("NSigma matrix: transis ------>\n"); 
  PRINTMATI(0,NSigma->C-1,0,NSigma->R-1,TransMat(NSigma));
#endif
  
  INEQ=ZeroMatrix(NSigma->R + nbplaces, NSigma->C+2);
  for (i=0; i<NSigma->R; i++)
    {
      INEQ->V[i][0]=1;
      for(j=1; j<INEQ->C-1; j++)  
	{
	  INEQ->V[i][j]= - NSigma->V[i][j-1];
	}
    }
  /*add non-negativity constraints*/
  count = 1;
  for (i=NSigma->R; i<INEQ->R; i++)
    {
      INEQ->V[i][0] = 1;
      INEQ->V[i][count] = 1;
      count++;
    }
  
 /*eliminate identical rows and simple constraints*/
  ElimEqualRows(&INEQ);
  ElimSimpleConstraints(&INEQ);
  /*sort the matrix*/
  SortRows(INEQ);
  
#ifdef DEBUG
  printf("Sortiert und eliminiert:\n ");
  PRINTMATI(0,INEQ->R-1,0,INEQ->C-1, INEQ);
#endif
  
  /*Solve the system of inequalities ZBASE.x>=0 */
  numsol=SolveInequationSystem(&SOLUTION,INEQ, MAXROWSIZE);
  
  if (numsol)
    {
#ifdef DEBUG
      printf("Lsung des Ungleichungssystems:\n ");
      PRINTMATI(0,SOLUTION->R-1,0,SOLUTION->C-1,SOLUTION);
#endif	    
      
      /*the columns of SOLUTION are the solutions of the siphon equation
	the siphons are the supports of SOLUTION
      */
      Siphon = CopyMat(TransMat(SOLUTION));
      
      /*calculate supports: 1 if != 0*/
      for (i=0; i< Siphon->R; i++)
	for(j=0; j< Siphon->C; j++)
	  if (Siphon->V[i][j]) Siphon->V[i][j] = 1;
      
      ElimEqualRows(&Siphon);
#ifdef DEBUG
      printf("Siphons, nonminimal:\n "); 
      PRINTMATI(0,Siphon->R-1,0,Siphon->C-1, Siphon);
#endif
      /*remove siphons that are not minimal*/
      ElimNonMinimal(&Siphon);
#ifdef DEBUG
      printf("Siphons, minimal:\n "); 
      PRINTMATI(0,Siphon->R-1,0,Siphon->C-1, Siphon);
#endif 
    }/*if (numsol)*/
  else
    Siphon = BuildMat(0,0);
  
  /*calculate initial Marking:
    Siphon.M0*/
  SiphonMarking = MultMat(Siphon, M);
#ifdef DEBUG	      
  printf("SiphonMarking:\n ");
  PRINTMATI(0,SiphonMarking->R-1,0,SiphonMarking->C-1, SiphonMarking);
#endif
  return Siphon->R;
}
