#include "struct_types.h"
#include <stdlib.h>
#include "matrix.h"
#include "netio.h"
#include "struct_vars.h"

/*	The following procedure returns the [row, col]-value
	of a sparse-matrix. Rows and cols count from 0      */	
int
sp_access(matrix, sp_row, sp_col)
	sp_place **matrix;
	int sp_row, sp_col;
{  sp_place *now;

	now = matrix[sp_row];
	while(now)
	{	if (now->num == sp_col)
			return now->value;
		if (now->num > sp_col)
			return 0;
		now = now->next;
	}
	return 0;
}

/*	The following procedure writes newvalue to the [row, col]-value
	of a sparse-matrix. Rows and cols count from 0      */	
void
sp_input(matrix, sp_row, sp_col, newvalue)
	sp_place **matrix;
	int sp_row, sp_col, newvalue;
{  sp_place *now;
 
 now = matrix[sp_row];
 while(now)
   {	
     if (now->num == sp_col)
       now->value = newvalue;
     now = now->next;
   }
}


INTMATRIX *conv_matrix(mat)
     sp_place **mat;
{
        INTMATRIX *Mptr;
	int i, j;

	Mptr=ZeroMatrix(no_trans,no_places);

	for(i=0; i<no_trans; i++)           /* via sp_access */
	{ 	for(j=0; j<no_places; j++)
  		Mptr->V[i][j]=sp_access(mat, i, j);
	}
	return Mptr;
}

/*
 * config_arcs: convert arc format
 */
int config_arcs(net, i, type)
NetDescr *net;
int i,type;
{
   int no_arcs=0;
   Arc *net_arc=0;
   sp_place **ptr=0;

   sp_place *last_p, *new_p;
   int j, point;
   char pl_name[50];

   /* determine arc type */
   switch( type )
   {
   case 0:
      no_arcs = net->trans[i].num_inarcs;
      net_arc = net->trans[i].in_arc;
      ptr = &(i_matrix[i]);
      break;
   case 1:
      no_arcs = net->trans[i].num_outarcs;
      net_arc = net->trans[i].out_arc;
      ptr = &(o_matrix[i]);
      break;
   case 2:
      no_arcs = net->trans[i].num_inharcs;
      net_arc = net->trans[i].inh_arc;
      ptr = &(h_matrix[i]);
      break;
   }

   for( j = 0; j < no_arcs; j++ )
   {
      new_p = (sp_place *)nMalloc(sizeof(sp_place));

      /* find index of place */
      strcpy(pl_name,net_arc[j].place->name);
      for (point = 0; point < no_places; point++)
         if (!strcmp(pl_name, places[point].name))
            break;
      if (point >= no_places)
      {
         printf("Unknown place %s in arc definition\n", pl_name);
         exit(1);
      }
      new_p->num = point;
      new_p->min_id = new_p->plu_id = 0;

      /*arc is not marking-dependent (has a value)*/
      if( net_arc[j].is_value )
         new_p->value = net_arc[j].multipl.value;
      else
	  places[point].mdconnect = 1;

      /* sort-by-insert */
      last_p = *ptr;
      if (last_p == NULL || last_p->num > new_p->num)
      {
         new_p->next = *ptr;
         *ptr = new_p;             /* first entry */
      }
      else
      {
         while( last_p->next != NULL && last_p->next->num < new_p->num )
            last_p = last_p->next;

         new_p->next = last_p->next;
         last_p->next = new_p;		/* normal entry */
      }
   }

   return( 0 );
}


/*
 * config_net: convert from netio-dataformat to local format
 */
int config_net(net)
NetDescr *net;
{
   int no_mpar, no_rpar;

   int i;

   /* create places list */
   no_places = net->num_places;
   places = (place *)nMalloc(no_places * sizeof(place));

   /* create transitions list */
   no_trans = net->num_trans;
   transitions = (transition *)nMalloc(no_trans * sizeof(transition));

   i_matrix = (sp_place **)nMalloc(no_trans * sizeof(sp_place *));
   o_matrix = (sp_place **)nMalloc(no_trans * sizeof(sp_place *));
   h_matrix = (sp_place **)nMalloc(no_trans * sizeof(sp_place *));
   for( i = 0; i < no_trans; i++ )
   {
      i_matrix[i] = NULL;
      o_matrix[i] = NULL;
      h_matrix[i] = NULL;
   }

   /* create delay parameters list */
   no_rpar = net->num_delaypars;
   r_par = (name_delay *)nMalloc(no_rpar * sizeof(name_delay));

   /* create marking parameters list */
   no_mpar = net->num_markpars;
   m_par = (name_nr *)nMalloc(no_mpar * sizeof(name_nr));


   /* config marking parameters */
   for( i = 0; i < no_mpar; i++ )
   {
      strcpy(m_par[i].name,net->markpar[i].name);
      m_par[i].val = net->markpar[i].value;
   }

   /* config places */
   for( i = 0; i < no_places; i++ )
   {
      strcpy(places[i].name,net->place[i].name);
      if( net->place[i].is_value )
      {
         places[i].m0 = net->place[i].marking.value;
         places[i].m0_str = (char *)nMalloc(16);
         sprintf(places[i].m0_str,"%d", places[i].m0);
      }
      else
      {
         places[i].m0 = net->place[i].marking.markpar->value;
         places[i].m0_str = net->place[i].marking.markpar->name;
      }
      places[i].mdconnect = 0; /*not connected to mark. dep. arc, updated later*/
   }


   /* config delay parameters */
   for( i = 0; i < no_rpar; i++ )
   {
      r_par[i].name = net->delaypar[i].name;
      if( net->delaypar[i].is_value )
      {
         r_par[i].type = Value;
         r_par[i].values.val = net->delaypar[i].delay.value;
      }
      else
         r_par[i].type = Definition;
   }

   /* config transitions */
   for( i = 0; i < no_trans; i++ )
   {
      transitions[i].number = i;
      transitions[i].mark_dep_nr = 0;
      strcpy(transitions[i].name,net->trans[i].name);
      transitions[i].priority = net->trans[i].priority;

      switch( net->trans[i].delay_type )
      {
      case DELAY:
         transitions[i].value = net->trans[i].delay.value;
         break;
      case DELAYPAR:
         if( net->trans[i].delay.delay_par->is_value )
            transitions[i].value = net->trans[i].delay.delay_par->delay.value;
         else
            transitions[i].mark_dep_nr = i+1;
         break;
      case MARKDEP:
         transitions[i].mark_dep_nr = i+1;
         break;
      case PMFDEF:
         break;
      }

      /* Definitively buggy! inf_server is either True or False, by definition.
       * This is, however, how it was done in the original file ...
       */
      transitions[i].infinite = -1;
      if( net->trans[i].inf_server )
         transitions[i].infinite = s_IS;
      else
         transitions[i].infinite = s_SS;

      if( transitions[i].infinite == -1 )
         switch( net->trans[i].f_policy )
         {
         case RS:
            transitions[i].infinite = s_RS;
            break;
         case RA:
            transitions[i].infinite = s_RA;
            break;
         case RE:
            transitions[i].infinite = s_RE;
            break;
         }

      switch( net->trans[i].type )
      {
      case IMM:
         transitions[i].type = s_IMM;
         break;
      case EXP:
         transitions[i].type = s_EXP;
         break;
      case DET:
         transitions[i].type = s_DET;
         break;
      case GEN:
         transitions[i].type = s_GEN;
         break;
      }
      no_timed_trans += (transitions[i].type != s_IMM);

      transitions[i].rate_id = 0;
      transitions[i].w_deter = NULL;
      transitions[i].ena_id = 0;
      transitions[i].e_deter = NULL;

      /* config arcs */
      config_arcs(net,i,0);        /* input arcs */
      config_arcs(net,i,1);        /* output arcs */
      config_arcs(net,i,2);        /* inhibitor arcs */
   }

   return 0;
}

void generate_pmf(filename,net)
char *filename;
NetDescr *net;
{
   char buf[256];
   FILE *fp;
   
   int i;
   
   /* build filename, open file */
   sprintf(buf,"%s.pmf", filename);
   if ((fp = fopen(buf,"w")) == NULL)
   {
      printf("ERROR: unable to create file '%s'\n", buf);
      exit(1);
   }

   /* write out all general transitions */
   for (i = 0; i < net->num_trans; i++)
      if (net->trans[i].type == GEN)
      {
         fprintf(fp,"TRANSITION %s\n", net->trans[i].name);
	 fprintf(fp,"%s\n\n", net->trans[i].delay.md_delay);
      }
   
   /* close file */
   fclose(fp);
}


void read_net(filename)
char *filename;
{
   NetDescr *net;
   net = load_net(filename,NULL);
   if (net)
   {
      config_net(net);
      
      if (pmf_bit)
         generate_pmf(filename,net);
   }
   else
   {
      printf("ERROR: could not read %s properly\n", filename);
      exit(1);
   }

   return;
}


/*	deter_in reads information about determining places from a file */
void
deter_in(f, p)
	FILE	*f;
	int	**p;
{
	int	i, no_deter = 0;
	char	dum[50];

	fscanf(f, "%s", dum);
	fscanf(f, "%s", dum); no_deter = atoi(dum);
	*p = (int *)nMalloc(no_places * sizeof(int));
	for (i = 0; i < no_deter; i++)
	{	fscanf(f, "%s", dum);
		(*p)[atoi(dum)] = 1;
	}
}


/*	read_info reads additional informations about marking dependent net parts,
	<net>.DEFINFO is provided by proc_MDF */
void
read_info(inf)
	FILE	*inf;
{
	int 	i, id, tra, no_ena_fcts, no_weights;
	md_arc	*p;
	char	dum[50];
	sp_place *p_w;

	/* read information from <net>.DEFINFO */
	fscanf(inf, "%s", dum);
	fscanf(inf, "%s", dum);
	no_md_arcs = atoi(dum);
	md_arcs = (md_arc *)nMalloc(no_md_arcs * sizeof(md_arc));
	for (i = 0; i < no_md_arcs; i++)
	{	p = &(md_arcs[i]);
		fscanf(inf, "%s", dum);
		fscanf(inf, "%s", dum); p->id = atoi(dum);
		fscanf(inf, "%s", dum); p->type = atoi(dum);
		fscanf(inf, "%s", dum); p->trans = atoi(dum);
		fscanf(inf, "%s", dum); p->place = atoi(dum);
		deter_in(inf, &(p->deter));
		switch (p->type)
		{ case INPUT:	p_w = i_matrix[p->trans]; break;
		  case OUTPUT:	p_w = o_matrix[p->trans]; break;
		  case INHIBIT:	p_w = h_matrix[p->trans]; break;
		  default:	exit(1);
		}
		while (p_w != NULL && p_w->num != p->place)
			p_w = p_w->next;
		if (p_w == NULL)
			exit(1);
		p_w->plu_id = p->id;
	}
	fscanf(inf, "%s", dum);
	fscanf(inf, "%s", dum);
	no_ena_fcts = atoi(dum);
	for (i = 0; i < no_ena_fcts; i++)
	{	fscanf(inf, "%s", dum);
		fscanf(inf, "%s", dum); id = atoi(dum);
		fscanf(inf, "%s", dum); tra = atoi(dum);
		transitions[tra].ena_id = id;		
		deter_in(inf, &(transitions[tra].e_deter));
	}
	fscanf(inf, "%s", dum);
	fscanf(inf, "%s", dum);
	no_weights = atoi(dum);
	for (i = 0; i < no_weights; i++)
	{	fscanf(inf, "%s", dum);
		fscanf(inf, "%s", dum); id = atoi(dum);
		fscanf(inf, "%s", dum); tra = atoi(dum);
		transitions[tra].rate_id = id;
		deter_in(inf, &(transitions[tra].w_deter));
	}
}
