/*****************************************************************************

	Programm	:	struct

	Komponente	:	struct_ecs.c
				Ermitteln der ECS der Subnetze
				Test auf Confusions

	Programmierer	:	Armin Zimmermann

	Datum		:	Juni 1992

*****************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include "struct_types.h"			/* definitions and types */
#include "struct_IO.h"
#include "struct_vars.h"

#include <netio.h>


extern sp_place *sparse_function_row();		/* from struct_subnet */
extern int	is_connected();


/*	this procedure returns a value from a sparse row */
int
sp_row(one_row, index)
	sp_place *one_row;
	int	index;
{     
	while(one_row && one_row->num < index)
		one_row = one_row->next;
	if (one_row && one_row->num == index)
		return (IS_MD(one_row) ? 0 : one_row->value);
	else	return 0;
}

/*	sp_row_free free's a sparse row */
void
sp_row_free(pointer)
	sp_place *pointer;
{
	while(pointer)
	{	sp_place *merk=pointer->next;
		free(pointer);
		pointer=merk;
	}
}

/*	is_in_packed returns true if b is in the packed set of a */
int
is_in_packed(a, b)
	byte *a;
	int  b;
{
	int look;

	for(look = 1; look <= a[0]; look++)
		if (a[look] == b)
			return 1;
		else if (a[look] > b)
			return 0;
	return 0;
}

/*	structural properties defined by membership in the relation-sets (packed) */

#define ME(t1, t2)	is_in_packed(sub->me[t1], t2)		/* mutual exclusion */
#define SC(t1, t2)	is_in_packed(sub->diss[t2], t1)		/* structural conflict */
#define ISC(t1, t2)	is_in_packed(sub->idiss[t2], t1)	/* ind. structural conflict */
#define ENA(t1, t2)	is_in_packed(sub->ena[t1], t2)		/* enabling relation */
#define PIENA(t1, t2)	is_in_packed(sub->pienas[t1], t2)	/* ind. enabling relation */
#define CI(t1, t2)	is_in_packed(sub->cis[t1], t2)		/* can influence */
#define ICI(t1, t2)	is_in_packed(sub->icis[t1], t2)		/* can ind. influence */
#define PEQU(t1, t2)	(r_pri(t1) == r_pri(t2))		/* equal priority level */
#define PGR(t1, t2)	(r_pri(t1) > r_pri(t2))			/* higher priority level */
#define SSC(t,s)	(PEQU(t,s) && !ME(t,s) && (SC(t,s) || SC(s,t) || ISC(t,s) || ISC(s,t)))
							/* symmetrical structural conflict */

/*	print_ecs prints one ecs */
void
print_ecs(sub, nr)
	subnet *sub;
	int	nr;
{
	int	i;

	printf("\nECS: ");
	for(i = 1; i <= sub->ecs[nr][0]; i++)
		printf(" %s", r_name(sub->ecs[nr][i]));
	printf("\n");
}

/*	h_min is used as the parameter-function for sparse_function_row
	to calculate the H-k(tm) - row					*/
void
h_min(hmp, hkp, resp)
	sp_place *hmp, *hkp, *resp;
{
	int hm, hk;

	if (((hmp != NULL) && IS_MD(hmp)) || ((hkp != NULL) && IS_MD(hkp)))
	{	resp->value = 1;
		return;			/* md: might be any value */
	}
	hm = (hmp == NULL) ? 0 : hmp->value;
	hk = (hkp == NULL) ? 0 : hkp->value;
	resp->value = (hm ? (hk ? hk > hm : 1) : 0);
	resp->min_id = resp->plu_id = 0;
}

/*	mini is used as the parameter-function for sparse_function_row
	to calculate the intersection between 2 bags (componentwise minimum)
	for the 'causally connected' test */
void
mini(a, b, r)
	sp_place *a, *b, *r;
{
	int	av, bv;

	av = (a == NULL) ? 0 : IS_MD(a) || a->value;	/* md: might be >0 */
	bv = (b == NULL) ? 0 : IS_MD(b) || b->value;
	r->value = min(av, bv);
	r->min_id = r->plu_id = 0;
}

/*	s_min is used as the parameter-function for sparse_function_row for bag-minus */
void
s_min(a, b, r)
	sp_place *a, *b, *r;
{
	int	av, bv;

	av = (a == NULL) ? 0 :  IS_MD(a) || a->value;	/* md: might be >0 */
	bv = (b == NULL) ? 0 : !IS_MD(b) && b->value;	/* md: might be =0 */
	r->value = max(av - bv, 0);
	r->min_id = r->plu_id = 0;
}

/*	max is used as a function parameter for sparse_function_row for MME:
	I(t1) + I(t2) - component_min(I(t1), I(t2)) 	*/
void
maxi(a, b, r)
	sp_place *a, *b, *r;
{
	if (a == NULL && b == NULL)
	{	r->value = 0;
		return;
	}
	if (a == NULL)
	{	r->value = IS_MD(b) ? 0 : b->value;
		return;
	}
	if (b == NULL)
	{	r->value = IS_MD(a) ? 0 : a->value;
		return;
	}
	r->value = max(a->value, b->value);
	r->min_id = r->plu_id = 0;
}
	
/*	confusion tests a pair of immediate transitions for external confusions */
void
confusion(sub, ecs, tk, tm, bit)
	subnet *sub;
	int	ecs, tk, tm, *bit;
{
	int tl, tx;
	sp_place *Il_minus_Ol, *Ol_minus_Il, *Im_minus_Ik, *Hm_minus_Hk;
	sp_place *TestInhibit, *TestInput;

	if (ME(tk, tm) || tk == tm)
		return;			/* no confusion possible */
	Im_minus_Ik = sparse_function_row(sub->i_matrix[tm], sub->i_matrix[tk], s_min);
	Hm_minus_Hk = sparse_function_row(sub->h_matrix[tm], sub->h_matrix[tk], h_min);
	for(tl = 0; tl < sub->no_trans; tl++)
	{	if (is_in_packed(sub->ecs[ecs], tl))
			continue;	/* ecs-member */
		if (ME(tk, tl))
			continue;			/* no confusion */
		if (PEQU(tl, tk) && (CI(tl, tk) || ICI(tl, tk)))
		{	if (!*bit)
			{	print_ecs(sub, ecs);
				*bit = 1;
			}
			printf("%s ",  conf_bit ? "Warning:" : "Exiting:" );
			printf("%sdirect external ",  CI(tl, tk) ? "" : "in" );
			printf("confusion: %s influences ", r_name(tl));
			printf("%s.\n", r_name(tk));
			if (!conf_bit)
				exit(1);
		}
		Ol_minus_Il = sparse_function_row(sub->o_matrix[tl], sub->i_matrix[tl], s_min);
		Il_minus_Ol = sparse_function_row(sub->i_matrix[tl], sub->o_matrix[tl], s_min);
		TestInput = sparse_function_row(Ol_minus_Il, Im_minus_Ik, mini);
		TestInhibit = sparse_function_row(Il_minus_Ol, Hm_minus_Hk, mini);
		if (TestInput == NULL && TestInhibit == NULL)
		{	sp_row_free(TestInput);
			sp_row_free(TestInhibit);
			sp_row_free(Il_minus_Ol);
			continue;
		}
		if (PEQU(tl, tk))
		{	if (!*bit)
			{	*bit = 1;
				print_ecs(sub, ecs);
			}
			printf("%s direct external ",  conf_bit ? "Warning:" : "Exiting:" );
			printf("(%s-)", TestInput != NULL ? "input" : "inhibitor");
			printf("confusion between ");
			printf("%s and %s ", r_name(tk), r_name(tm));
			printf("caused by %s.\n", r_name(tl));
			if (!conf_bit)
				exit(1);
		}
		if (PGR(tl, tk))
		{	for(tx = 0; tx < sub->no_trans; tx++)
			{	if (is_in_packed(sub->ecs[ecs], tx))
					continue;
				if (r_pri(tx) != r_pri(tk))
					continue;
				if (!PIENA(tx, tl))
					continue;
				if (!*bit)
				{	*bit = 1;
					print_ecs(sub, ecs);
				}
				printf("%s indirect ",  conf_bit ? "Warning:" : "Exiting:" );
				printf("external (%s", TestInput != NULL ?
								 "input" : "inhibitor");
				printf("-)confusion between ");
				printf("%s and %s ", r_name(tk), r_name(tm));
				printf("caused by %s via %s.\n", r_name(tx), r_name(tl));
				if (!conf_bit)
					exit(1);
			}
		}
		sp_row_free(TestInput);
		sp_row_free(TestInhibit);
		sp_row_free(Il_minus_Ol);
	}
	sp_row_free(Im_minus_Ik);		/* release data */
	sp_row_free(Hm_minus_Hk);
}

/*	test_confusions tests one ecs for confusions */
void
test_confusions(sub, num)
	subnet *sub;
	int	num;
{
	int	tki, tmi, tk, tm;
	int	bit = 0;

	for(tki = 1; tki <= sub->ecs[num][0]; tki++)
	{	tk = sub->ecs[num][tki];
		for(tmi = 1; tmi <= sub->ecs[num][0]; tmi++)
		{	tm = sub->ecs[num][tmi];
			if (tmi != tki)
			{	/* for all pairs of transitions tk and tm in an ecs */
				tm = sub->ecs[num][tmi];
				/* test tm and tk for inner confusion */
				if (!ME(tk, tm) && !SC(tk, tm) && !ISC(tk, tm))
				{	if (!bit)
					{	bit = 1;
						print_ecs(sub, num);
					}
					printf("%s inner confusion between %s and %s.\n",
						conf_bit ? "Warning:" : "Exiting:" ,
							 r_name(tk), r_name(tm));
					if (!conf_bit)
						exit(1);
				}
			}
			/* test for external confusion */
			confusion(sub, num, tk, tm, &bit);
		}
	}
}

/*	sme returns true if t1 and t2 are in structural mutual exclusion relation */
int
sme(sub, t1, t2)
	subnet	*sub;
	int t1, t2;
{
	int	pl_nr;
	int	ht1, it2, ht2, it1;

	for(pl_nr = 0; pl_nr < sub->no_places; pl_nr++)			/* for each place */
	{	ht1 = sp_row(sub->h_matrix[t1], pl_nr);
		ht2 = sp_row(sub->h_matrix[t2], pl_nr);
		it1 = sp_row(sub->i_matrix[t1], pl_nr);
		it2 = sp_row(sub->i_matrix[t2], pl_nr);
		if ((ht1 && it2 && (ht1 <= it2)) || (ht2 && it1 && (ht2 <= it1)))
		{
#ifdef DEBUG
			printf("%s SME %s\n", r_name(t1), r_name(t2));
#endif
			return 1;
		}
	}
	return 0;
}


/*	mme returns true if t1 and t2 are in marking mutual exclusion relation */
int
mme(sub, t1, t2)
	subnet *sub;
	int t1, t2;
{
	pinvar	*pin_now;
	sp_place *inter_row, *now;
	int	sum;

	inter_row = sparse_function_row(sub->i_matrix[t1], sub->i_matrix[t2], maxi);
	pin_now = pinvars;
	while(pin_now)						/* for each p-invariant */
	{	sum = 0;
		now = inter_row;
		while(now)					/* for each inter_row entry */
		{	sum += sp_row(pin_now->first, p_number(now->num)) * now->value;
			now = now->next;
		}
		if (sum > pin_now->token_count)
		{
#ifdef DEBUG
			printf("%s MME %s\n", r_name(t1), r_name(t2));
#endif
			sp_row_free(inter_row);
			return 1;
		}
		pin_now = pin_now->next;
	}
	return 0;
}

/*	calc_me_relation calculates the set of transitions which are mutual exclusive
	for each transition in the subnet (without priority checking; = ME Concession) */
void
calc_me_relation(sub)
	subnet	*sub;
{
	int	entry, t1, t2;

#ifdef DEBUG
	printf("    ===> mutual exclusions\n");
#endif
	sub->me = (byte **)nMalloc(sub->no_trans * sizeof(byte *));
	for(t1 = 0; t1 < sub->no_trans; t1++)
	{	sub->me[t1] = (byte *)nMalloc((sub->no_trans+2) * sizeof(byte));
		entry = 1;
		for(t2 = 0; t2 < sub->no_trans; t2++)		/* ME if SME or MME */
		{	if (t1 == t2)
				continue;
			if (sme(sub, t1, t2) || mme(sub, t1, t2))
				sub->me[t1][entry++] = t2;
		}
		sub->me[t1][0] = entry - 1;			/* number of entries */
	}
}

/*	calc_enas_diss calculates the enabling and disabling sets
	for transitions in subnets */
void
calc_enas_diss(sub)
	subnet *sub;
{
	byte		*max_diss, *max_enas, *max_ci;
	int		tra, test_tra, o_val, i_val, pr, tr;
	sp_place	*c_pla;

	sub->enas = (byte **)nMalloc(sub->no_trans * sizeof(byte *));
	sub->diss = (byte **)nMalloc(sub->no_trans * sizeof(byte *));
	sub->cis  = (byte **)nMalloc(sub->no_trans * sizeof(byte *));
	max_diss  = (byte  *)nMalloc(sub->no_trans * sizeof(byte  ));
	max_enas  = (byte  *)nMalloc(sub->no_trans * sizeof(byte  ));
	max_ci    = (byte  *)nMalloc(sub->no_trans * sizeof(byte  ));
#ifdef DEBUG
	printf("subnet %d\n", sub->number-1);
	printf("direct links\n");
#endif
	/* for each transition in the subnet */
	for(tra = 0; tra < sub->no_trans; tra++)
	{	for(test_tra = 0; test_tra < sub->no_trans; test_tra++)		/* clear */
			max_diss[test_tra] = max_enas[test_tra] = max_ci[test_tra] = 0;
		/* calculate the set of transitions, which could be directly 	
		enabled or disabled or influenced due to the firing of this transition */
#ifdef DEBUG
		printf("transition %s:\n", r_name(tra));
#endif
		c_pla = sub->c_matrix[tra];
		while(c_pla)
		{	pr = p_number(c_pla->num);
			o_val = max( c_pla->value, 0) || c_pla->plu_id;	/* O - I */
			i_val = max(-c_pla->value, 0) || c_pla->min_id;	/* I - O */
			for(test_tra = 0; test_tra < sub->no_trans; test_tra++)
			{	if (test_tra == tra)
					continue;
				tr = r_number(test_tra);
				if ((o_val && is_connected(tr, pr, INPUT)) ||
				    (i_val && is_connected(tr, pr, INHIBIT)) ||
				    is_connected(tr, pr, ENABLE) ||
				    is_connected(tr, pr, ARC_IND) )
					max_enas[test_tra] = 1;
				if ((i_val && is_connected(tr, pr, INPUT)) ||
				    (o_val && is_connected(tr, pr, INHIBIT)) ||
				    is_connected(tr, pr, ENABLE) ||
				    is_connected(tr, pr, ARC_IND) )
					max_diss[test_tra] = 1;
				if (is_connected(tr, pr, WEIGHT))
					max_ci[test_tra] = 1;
			}
			c_pla = c_pla->next;
		}
		i_val = o_val = pr = 1;
		for(test_tra = 0; test_tra < sub->no_trans; test_tra++)
		{	i_val += max_enas[test_tra];
			o_val += max_diss[test_tra];
			pr    += max_ci[test_tra];	/* count sets */
		}
		sub->enas[tra] = (byte *)nMalloc(i_val * sizeof(byte));
		sub->diss[tra] = (byte *)nMalloc(o_val * sizeof(byte));
		sub->cis[tra]  = (byte *)nMalloc(pr    * sizeof(byte));
		sub->enas[tra][0] = i_val - 1;		/* number of entries */
		sub->diss[tra][0] = o_val - 1;
		sub->cis[tra][0]  = pr    - 1;
		i_val = o_val = pr = 1;
		for(test_tra = 0; test_tra < sub->no_trans; test_tra++)
		{	if (max_enas[test_tra])
				sub->enas[tra][i_val++] = test_tra;
			if (max_diss[test_tra])
				sub->diss[tra][o_val++] = test_tra;
			if (max_ci[test_tra])
				sub->cis[tra][pr++] = test_tra;
		}
#ifdef DEBUG
		printf("ENAS: ");
		for(test_tra = 1; test_tra <= sub->enas[tra][0]; test_tra++)
			printf("%s ",r_name(sub->enas[tra][test_tra]));
		printf("\nDISS: ");
		for(test_tra = 1; test_tra <= sub->diss[tra][0]; test_tra++)
			printf("%s ",r_name(sub->diss[tra][test_tra]));
		printf("\nCIS: ");
		for(test_tra = 1; test_tra <= sub->cis[tra][0]; test_tra++)
			printf("%s ",r_name(sub->cis[tra][test_tra]));
		printf("\n\n");
#endif
	}
}

/*	calc_ienas_idiss calculates the indirect enabling, disabling
	and influence sets for transitions in subnets */
void
calc_ienas_idiss(sub)
	subnet *sub;
{
	byte	*max_diss, *max_enas, *max_icis, *test_set;
	int	tra, test_tra, tneu, i_val, o_val, c_val;
#ifdef DEBUG
	printf("indirect links\n");
#endif
	test_set    = (byte  *)nMalloc(sub->no_trans * sizeof(byte  ));
	max_diss    = (byte  *)nMalloc(sub->no_trans * sizeof(byte  ));
	max_enas    = (byte  *)nMalloc(sub->no_trans * sizeof(byte  ));
	max_icis    = (byte  *)nMalloc(sub->no_trans * sizeof(byte  ));
	sub->pienas = (byte **)nMalloc(sub->no_trans * sizeof(byte *));
	sub->idiss  = (byte **)nMalloc(sub->no_trans * sizeof(byte *));
	sub->icis   = (byte **)nMalloc(sub->no_trans * sizeof(byte *));
	for(tra = 0; tra < sub->no_trans; tra++)
	{	for(test_tra = 0; test_tra < sub->no_trans; test_tra++)	/* clear */
		{	max_diss[test_tra] = max_enas[test_tra] = 0;
			test_set[test_tra] = max_icis[test_tra] = 0;
		}
		/* calculate the set of transitions, which could be indirectly 	
		   enabled, disabled or influenced due to the firing of this transition */
#ifdef DEBUG
		printf("transition %s:\n", r_name(tra));
#endif
		for(test_tra = 1; test_tra <= sub->enas[tra][0]; test_tra++)
			if (r_pri(sub->enas[tra][test_tra]) > r_pri(tra))
				test_set[sub->enas[tra][test_tra]] = 1;	/* mark */
		while(1)
		{	/* look for first transition in remaining set */
			for(tneu = 0; tneu < sub->no_trans; tneu++)
				if (test_set[tneu] == 1)
					break;			/* found */
			if (tneu >= sub->no_trans)
				break;				/* exit while-loop */
			for(test_tra = 1; test_tra <= sub->diss[tneu][0]; test_tra++)
				if (r_pri(sub->diss[tneu][test_tra]) == r_pri(tra))
					max_diss[sub->diss[tneu][test_tra]] = 1;
			for(test_tra = 1; test_tra <= sub->cis[tneu][0]; test_tra++)
				if (r_pri(sub->cis[tneu][test_tra]) == r_pri(tra))
					max_icis[sub->cis[tneu][test_tra]] = 1;
			for(test_tra = 1; test_tra <= sub->enas[tneu][0]; test_tra++)
				if (r_pri(sub->enas[tneu][test_tra]) > r_pri(tra))
					if (!test_set[sub->enas[tneu][test_tra]])
						test_set[sub->enas[tneu][test_tra]] = 1;
			test_set[tneu] = 2;			/* mark o.k. */
		}
		i_val = o_val = c_val = 1;	/* need 1 more indix than entries */
		for(test_tra = 0; test_tra < sub->no_trans; test_tra++)
		{	i_val += (test_set[test_tra] > 0);
			o_val += max_diss[test_tra];
			c_val += max_icis[test_tra];		/* count sets */
		}
		sub->pienas[tra] = (byte *)nMalloc(i_val * sizeof(byte));
		sub->idiss[tra]  = (byte *)nMalloc(o_val * sizeof(byte));
		sub->icis[tra]   = (byte *)nMalloc(c_val * sizeof(byte));
		sub->pienas[tra][0] = i_val - 1;	/* real number of entries */
		sub->idiss[tra][0]  = o_val - 1;
		sub->icis[tra][0]   = c_val - 1;
		i_val = o_val = c_val = 1;
		for(test_tra = 0; test_tra < sub->no_trans; test_tra++)
		{	if (test_set[test_tra])
				sub->pienas[tra][i_val++] = test_tra;
			if (max_diss[test_tra])
				sub->idiss[tra][o_val++] = test_tra;
			if (max_icis[test_tra])
				sub->icis[tra][c_val++] = test_tra;
		}
#ifdef DEBUG
		printf("PIENAS: ");
		for(test_tra = 1; test_tra <= sub->pienas[tra][0]; test_tra++)
	 		printf("%s ",r_name(sub->pienas[tra][test_tra]));
		printf("\nIDISS: ");
		for(test_tra = 1; test_tra <= sub->idiss[tra][0]; test_tra++)
			printf("%s ",r_name(sub->idiss[tra][test_tra]));
		printf("\nICIS: ");
		for(test_tra = 1; test_tra <= sub->icis[tra][0]; test_tra++)
			printf("%s ",r_name(sub->icis[tra][test_tra]));
		printf("\n\n");
#endif
	}
	free(test_set);
	free(max_diss);
	free(max_enas);
}

/*	calc_ecs calculates the extended conflict sets of a subnet */
void
calc_ecs(sub)
	subnet *sub;
{
	int	tra, test_tra, tneu, members;
	byte	*tr_left, *test_set;

	tr_left = (byte *)nMalloc(no_trans * sizeof(byte));
	test_set = (byte *)nMalloc(no_trans * sizeof(byte));
	sub->no_ecs = 0;
	sub->ecs = (byte **)nMalloc(sub->no_trans * sizeof(byte *));
	tra = sub->no_trans;
	for(test_tra = 0; test_tra < tra; test_tra++)
		tr_left[test_tra] = 0;			/* clear */
	while(tra)					/* transitions left */
	{	tneu = -1;
		test_tra = 0;
		/* look for left transition with highest priority */
		for(test_tra = 0; test_tra < sub->no_trans; test_tra++)
		{	if (tr_left[test_tra] == 1)
				continue;		/* not left */
			if (tneu == -1 || r_pri(tneu) < r_pri(test_tra))
				tneu = test_tra;	/* higher priority */
		}
		for(test_tra = 0; test_tra < sub->no_trans; test_tra++)
			test_set[test_tra] = 0;		/* clear */
		test_set[tneu] = 1;			/* initial ecs-member */
		while(1)
		{	tneu = 0;
			while(tneu < sub->no_trans && 
					(tr_left[tneu] || !test_set[tneu]))
				tneu++;
			if (tneu >= sub->no_trans)
				break;			/* ecs completed */
			for(test_tra = 0; test_tra < sub->no_trans; test_tra++)
				if (tr_left[test_tra] || test_set[test_tra])
					continue;		/* uninteresting */
				else if (SSC(tneu, test_tra))
				{	test_set[test_tra] = 1;	/* new ecs member */
					if (!SC(tneu, test_tra) && !SC(test_tra, tneu))
		printf("Warning: %s joins ecs because of indirect conflict with %s.\n",
			r_name(test_tra), r_name(tneu));
				}
			tr_left[tneu] = 1;			/* mark tneu o.k. */
		}
		for(members = 0, test_tra = 0; test_tra < sub->no_trans; test_tra++)
			if (test_set[test_tra])
				members++;			/* count ecs members */
		sub->ecs[sub->no_ecs++] = (byte *)nMalloc((members + 2)*sizeof(byte));
#ifdef DEBUG
		printf("ECS %d contains %d transition(s):\n", sub->no_ecs, members);
#endif
		for(tneu = 1, test_tra = 0; test_tra < sub->no_trans; test_tra++)
			if (test_set[test_tra])
			{	sub->ecs[sub->no_ecs-1][tneu++] = test_tra;
#ifdef DEBUG
				printf("%s ", r_name(test_tra));
#endif
			}
		sub->ecs[sub->no_ecs-1][0] = members;	  /* number of ecs members */
		sub->ecs[sub->no_ecs-1][members + 1] = 255; /* stop index */
		tra -= members;		/* number of left transitions in subnet */
#ifdef DEBUG
		printf("\n");
#endif
	}
	free(tr_left);
	free(test_set);
}

/*	calculate structural properties of a subnet */
void
calc_sn_struct(sub)
	subnet	*sub;
{
	int	ecs_num;

#ifdef DEBUG
	printf("=======> Subnet %d\n", sub->number-1);
#endif
	calc_me_relation(sub);
	calc_enas_diss(sub);		/* calculate enas and diss */
	calc_ienas_idiss(sub);		/* calculate ienas and idiss */
	calc_ecs(sub);			/* calculate ext. conflict sets */
	for (ecs_num = 0; ecs_num < sub->no_ecs; ecs_num++)
		test_confusions(sub, ecs_num);	/* test for confusions */
}

/*	main procedure of this program part */
void
calc_struct(first_subnet)
	subnet	*first_subnet;
{
	subnet	*sub = first_subnet;

#ifdef DEBUG
	printf("=======> structural properties\n");
#endif
	while(sub != NULL)			/* for each subnet */
	{	calc_sn_struct(sub);
		sub = sub->next;
	}
}

