/*****************************************************************************

Programm	:	struct

Komponente	:	struct_pin.c
			Einlesen des <netname>.NET - Files,
			Ermitteln der Place-Invarianten des DSPN
	     

Programmer	:	Armin Zimmermann
Date		:	Juni 1992

Change		:	Thu Mar  3 17:12:50 MET 1994

		IMM immediate (zeitlose)           transition.type = IMM (0)
		DET deterministische               transition.type = DET (1)
		EXP exponentielle                  transition.type = EXP (2)
		GEN general (beliebig verteilt)    transition.type = GEN (3)

		SS single   server race enabling   transition.infinite = SS  (0)
		IS infinite server race enabling   transition.infinite = IS  (1)

		rs single   server race resampling transition.infinite = RS  (2)
		re single   server race enabling   transition.infinite = RE  (3)
		ra single   server race age        transition.infinite = RA  (4)

Change		:	May/June 1994 (Ulrich Hertlein)

		adaption to .TN file format, uses load_net library function;
		NetDescr is converted to local data format, after that, everything
		unchanged

Change          :      1998 Alexander Huck  

                old invariant algorithm replaced

*****************************************************************************/


#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>

#include "struct_types.h"			/* definitions and types */
#include "struct_IO.h"
#include "struct_vars.h"
#include "portab.h"

#include "matrix.h"
#include "matrixOps.h"
#include "matrixIO.h"
#include "invars.h"
#include "estimate.h"

#include <netio.h>

/*#define DEBUG*/

row 	*zero, *positive, *negative;		/* rows sorted by active column */
						/*       = no_places + no_trans */

#ifdef DEBUG
/*	print a sparse matrix (test only) */	
void
print_matrix(mat)
	sp_place **mat;
{	int i, j;
	sp_place *now;

	printf("\nMatrix:\n");
	for(i=0; i<no_trans; i++)           /* via sp_access */
	{ 	for(j=0; j<no_places; j++)
  		printf(" %2d", sp_access(mat, i, j));
  		printf("\n");
	}
	return;
}

/*	print a sparse row (test only) */	
void
print_rows(rowp)
	row *rowp;
{  int i;

	if (!rowp)
	{	printf("\nNo Rows found.\n");
		return;
	}
	printf("\nRows:\n");
	while(rowp)
	{	for(i=0; i<no_places+no_trans; i++)
			printf(" %2d", rowp->data[i]);
		printf("\n");
		rowp = rowp->next;
	}
}
#endif


/*	print the obtained p-invariants (option -p) and calculate the upper bounds
	for number of tokens in each place 						*/
void
p_output()
{	sp_place *act_place;
	pinvar *pin;
	int pin_num;

	for(pin_num = 0; pin_num < no_places; pin_num++)
   	{	places[pin_num].bound = MAX_TOKEN;		/* maximum upper bound */
		places[pin_num].covered = 0;
	}
	if (pin_bit)
		printf("\nThe net contains %d P-invariants.", no_pinvar);
	pin = pinvars;  pin_num = 0;
	while(pin)
	{	if (pin_bit)
			printf("\n==> Number %d (token-count = %d):\n", 
	                	++pin_num, pin->token_count);
		act_place = pin->first;
		while(act_place)
		{	if (pin_bit)
			{	if (act_place->value != 1)
					printf("%dx", act_place->value);
				printf("%s ", places[act_place->num].name);
			}
	     
			if (places[act_place->num].bound > pin->token_count / act_place->value)
				places[act_place->num].bound = 
							pin->token_count / act_place->value;
			places[act_place->num].covered = 1;	/* place is covered */
			act_place = act_place->next;
		}
		pin = pin->next;
	}
	if (pin_bit)
		printf("\n\n");
	for(pin_num = 0; pin_num < no_places; pin_num++)
	if (places[pin_num].covered)
	{	if (pin_bit)
			printf("Place %s is %d-bounded.\n", 
				places[pin_num].name, places[pin_num].bound);
	}
	else	if (pin_bit)
   	   		printf("Warning-place %s is not contained in any P-invariant.\n",
				places[pin_num].name);
}


/*	release the allocated memory of a sparse row */
void
free_rows(rowp)
	row **rowp;
{  row *now, *next;

	now = *rowp;
	*rowp = NULL;
	if (!now)
   	return;
	while(now)
	{	next = now->next;
		free(now->data);
		free(now);
		now = next;
	}
}

/*	main procedure to calculate the minimal-support p-invariants */
void
p_invariants()
{	int pla_nr, *data, i, j;
	row *r_now=0, *r_last;
	pinvar *pin_now, *pin_last;
	sp_place *place_now=0, *place_last;
	char	inter[30];
	USI outercol, mdarcplace = 0;

	INTMATRIX *IncidenceMat, *InvariantsMat, *OldMat, *MNull;
	INTMATRIX *PreMat, *PostMat;
	int numinv, numcols, MERKDIRS;
   
	pinvars = NULL;   zero = NULL;
	positive = NULL;  negative = NULL;
	/* generate the first set of rows */
	row_width = no_places + no_trans;
	r_last = NULL;
	
	IncidenceMat=conv_matrix(c_matrix);
	PreMat=conv_matrix(i_matrix);
	PostMat=conv_matrix(o_matrix);
	
#ifdef DEBUG
	printf("Incidence matrix: transis ------>\n"); 
	PRINTMATI(0,IncidenceMat->C-1,0,IncidenceMat->R-1,TransMat(IncidenceMat));
	printf("POST matrix: transis ------>\n"); 
	PRINTMATI(0,PostMat->C-1,0, PostMat->R-1,TransMat(PostMat));
	printf("PRE matrix: transis ------>\n"); 
	PRINTMATI(0,PreMat->C-1,0,PreMat->R-1,TransMat(PreMat));
#endif	


	mdarcplace = 0;
	for(i=0;i<no_places;i++)	  
	  if(places[i].mdconnect) /*connected to md ?*/
	    mdarcplace++;
		
	numcols =  no_places - mdarcplace;
	OldMat = IncidenceMat;
	IncidenceMat = BuildMat(no_trans, numcols);
	
	outercol = 0;
	for (i = 0; i <  no_places; i++)
	  {	    
	    if(places[i].mdconnect)
	      {
		outercol++;
		continue;
	      }
	    for (j = 0; j < no_trans; j++) 
	      IncidenceMat->V[j][i-outercol] = OldMat->V[j][i];
	  }
	
	FreeMat(OldMat);
	if(IncidenceMat->C) 
	  numinv=INVARIANTS(&InvariantsMat, IncidenceMat);
	else numinv =0;
	
	if(numinv)
	  {
	    /*re-insert places connected to md-arc into the invariants-matrix*/
	    ResizeRows( &InvariantsMat, mdarcplace);
	    for(i=0; i<no_places; i++)
	      {
		if(places[i].mdconnect)
		  {
		    /*move rows downwards*/
		    for(j=no_places-1; j>i; j--)
		      {
			CopyRow(InvariantsMat,j-1,j);
		      }
		    /*insert zero row*/
		    for(j=0; j<numinv; j++)
		      {
			InvariantsMat->V[i][j] = 0;
		      }
		  }
	      }
	  }

	MNull=ZeroMatrix(no_places,1);
	for(pla_nr = 0; pla_nr < no_places; pla_nr++)
	  {
	    MNull->V[pla_nr][0] = places[pla_nr].m0;
	  }
		
#ifdef DEBUG
	if(numinv)
	  {
	    printf("Invariants: invars ------>\n"); 
	    PRINTMATI(0,InvariantsMat->R-1,0,InvariantsMat->C-1, InvariantsMat);
	  }
#endif	

	if (numinv == 0)
	  InvariantsMat = ZeroMatrix(0,0);

	for(j = 0; j < numinv; j++)
	{
		data = (int *)nMalloc(InvariantsMat->R * sizeof(int)); 

		for(i = 0; i < InvariantsMat->R; i++)
		  {
			data[i] = InvariantsMat->V[i][j]; 
		  }

		r_now = (row *)nMalloc(sizeof(row));
		r_now->data = data;
		r_now->no_entries = 1;

		if (r_last)
			r_last->next = r_now;
		else
			zero = r_now;
		r_last = r_now;
	}
		if (r_last != NULL) r_now->next = NULL; 
		if (numinv!=0) 	
		  {
		    FreeMat(InvariantsMat);
		  }

		FreeMat(IncidenceMat);
		FreeMat(PostMat);
		FreeMat(PreMat);
		FreeMat(MNull);

		MERKDIRS=no_trans; no_trans=0;

	/* get p-invars from last set of rows */
	no_pinvar = 0; pin_last = pinvars;
	r_now = zero;
	pin_now = NULL;
	while(r_now)
	{
	  no_pinvar++;
		pin_now = (pinvar *)nMalloc(sizeof(pinvar));
		pin_now->token_count = 0;
		pin_now->count_str[0] = 0;
		if (pin_last)
			pin_last->next = pin_now;
		else
			pinvars = pin_now;
		pin_last = pin_now;
 		pin_now->first = NULL;
		data = &(r_now->data[no_trans]);
		place_last = NULL;
		for(pla_nr = 0; pla_nr < no_places; pla_nr++)
		{
		  if (!data[pla_nr])
				continue;
			place_now = (sp_place *)nMalloc(sizeof(sp_place));
			if (place_last)
				place_last->next = place_now;
			else
				pin_now->first = place_now;
			place_last = place_now;
			place_now->num = pla_nr;
			place_now->value = data[pla_nr];
			place_now->min_id = place_now->plu_id = 0;
			if (data[pla_nr] * places[pla_nr].m0)
			{	sprintf(inter, "%d*", data[pla_nr]);
				if (pin_now->token_count)
					strcat(pin_now->count_str, " + ");
				if (data[pla_nr] > 1)
					strcat(pin_now->count_str, inter);
				strcat(pin_now->count_str, places[pla_nr].m0_str);
			}
			pin_now->token_count += data[pla_nr] * places[pla_nr].m0;
		}
		place_now->next = NULL;
		r_now = r_now->next;
	}
	if (pin_now)
		pin_now->next = NULL;
	free_rows(&zero);
	no_trans=MERKDIRS;

}

