#ifndef __STRUCT_TYPES_H__
#define __STRUCT_TYPES_H__

/*****************************************************************************

	Programm	:	struct

	Komponente	:	struct_types.h
				Typdeklarationen und Definitionen

	Programmierer	:	Armin Zimmermann

	Datum		:	Juni 1992

*****************************************************************************/

#define	MAX_TOKEN	65535	/* maximum number of token in a place (2 bytes) */
#define FILE_NAME	200	/* maximal length of a filename */

#define max(a, b)	(a > b ? a : b)
#define min(a, b)	(a < b ? a : b)

#define p_number(a)	sub->places[a]
#define p_name(a)	places[p_number(a)].name

#define	r_number(a)	sub->transitions[a]
#define r_name(a)	transitions[r_number(a)].name
#define r_pri(a)	transitions[r_number(a)].priority

#define s_SS	0
#define s_IS	1
#define s_RS	2
#define s_RE	3
#define s_RA	4

#define	s_IMM	0
#define	s_DET	1
#define	s_EXP	2
#define s_GEN   3

#define	INPUT	0		/* type codes for marking dep. arcs */
#define OUTPUT	1		/* and connections */
#define	INHIBIT	2
#define WEIGHT	10
#define ENABLE	11
#define ARC_IND	12
#define	ANY	255

#define IS_MD(a)	(a->min_id || a->plu_id)

typedef unsigned char	byte;

typedef struct sparse_place		/* element type for sparse matrices */
{	int	num;			/* number of place in row */
	int	value;			/* value */
	int	min_id, plu_id;		/* marking dependency id */
	struct	sparse_place *next;
}  sp_place;

typedef struct pinvar			/* type for one P-invariant */
{	sp_place *first;		/* first place entry */
	int	token_count;		/* token count of the P-inv. */
	char	count_str[1000];		/* description of the token count */
	struct pinvar *next;
}  pinvar;

typedef struct place			/* type for places */
{	char	name[30];		/* place name */
	int	m0;			/* initial marking */
	char	*m0_str;		/* description of m0 */
	int	bound;			/* upper bound for #of tokens in place */
	int	covered;		/* place is covered by a P-invariant */
	int	subnet;			/* number of the subnet */
  int     mdconnect;              /* connected to at least one marking dependent arc */
}  place;

typedef struct name_nr
{	char	name[30];		/* object name */
	int	val;			/* number or value */
}  name_nr;

typedef enum {Value, Definition} delay_type;

typedef struct name_delay
{	char	*name;			/* delay parameter name */
	delay_type type;		/* value, definition */
	union {
	    double	val;		/* value */
	    char        *def;
	} values;
}  name_delay;

typedef struct transition		/* type for transitions */
{	char	name[30];		/* place name */
	int	type;			/* transition type, see above */
	int	mark_dep_nr;		/* marking dep. rate / weight; index */
	int	infinite;		/* infinite server bit */
	double	value;			/* firing time / probability value */	
	int	priority;		/* priority (immediate only) */
	int	subnet;			/* number of the subnet (immediate only) */
	int	number;			/* global identification number */
	int	rate_id;		/* weight is marking dependent; index */
	int	*w_deter;		/* weight determining places */
	int	ena_id;			/* index for enabling function */
	int	*e_deter;		/* enabling function determining places */
}  transition;

typedef struct subnet
{	int	number;			/* number of the subnet, 1 = det. trans. */
	int	no_trans, no_places, no_ecs;
					/* number of transitions, places and ecs */
	int	*transitions, *places;	/* transition's and place's real numbers */
	byte	**diss, **enas;		/* disabling and enabling sets */
	byte	**pienas, **idiss;	/* indirect enabling and disabling sets */
	byte	**cis, **icis;		/* direct and ind. 'can influence' sets */
	byte	**me;			/* mutual exclusion sets */
	byte	**ecs;			/* set of ecs */
	sp_place **i_matrix;		/* input matrix of the subnet */
	sp_place **o_matrix;		/* output matrix of the subnet */
	sp_place **h_matrix;		/* inhibitor matrix of the subnet */
	sp_place **c_matrix;		/* o - i matrix of the subnet */
	struct subnet *next;
}	subnet;


typedef struct row              	/* row type for M/S-algorithm */
{       int *data;              	/* row data, integer array [row_width] */
        int no_entries;         	/* number of no-null entries in D-matrix */
        struct row *next;
}  row;

typedef struct md_arc			/* type for marking dependent arcs */
{	int	type;			/* arc type, see above */
	int	place;			/* number of connected place */
	int	trans;			/* number of connected transition */
	int	id;			/* global identification number */
	int	*deter;			/* determining places */
}  md_arc;

#endif
