/*****************************************************************************

Programm	:	traps

Komponente	:	trapsdisplay.c
                        Berechnung von siphons

Programmer	:	Alexander Huck
Date		:	1999

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "polycount.h"
#include "matrix.h"
#include "matrixIO.h"
#include "matrixOps.h"
#include "timer.h"

#include "struct_types.h"			/* definitions and types */
#include "struct_IO.h"
#include "struct_vars.h"

#include "traps.h"

//#define DEBUG

extern int  	minus();
extern sp_place **sparse_function();  
char model_name[FILE_NAME];
int numtraps = 0;

/*******************************************************************************************
 output traps to .traps - file
*******************************************************************************************/
static void output_result(double timing)
{
  FILE   *traps_file, *f;
  char	help[1000];
  int i,j;
  
  sprintf(help,"%s.traps", model_name);
  if ((traps_file = fopen(help, "w")) == NULL)
    {
      printf("ERROR: could not open %s for writing\n", help);
      exit(1);
    }
  
  f = traps_file;		/* output to .trap */
  
  fprintf(f, "Calculation of Traps:\n");
  fprintf (f, "Time passed with computation: %8.2f s\n", timing );
  if (numtraps == 0)
    {
      fprintf(f," There are no traps\n");
    }
  else
    {
      fprintf(f,"No. of traps = %i \n", numtraps);
      for(i=0;i<numtraps;i++)
	{
	  fprintf(f,"MARKING: %4i ", TrapMarking->V[i][0]);
	  fprintf(f,"{ ");
	  
	  for(j=0;j<no_places;j++)
	    {
	      if(Trap->V[i][j])
		{
		  fprintf(f,places[j].name);
		  fprintf(f," ");
		}
	    }
	  fprintf(f,"}\n");
	}
      
#ifdef DEBUG
      fprintf(traps_file,"Traps: \n");
      FPRINTMATI(traps_file,0,Trap->R-1,0,Trap->C-1,Trap);
#endif
    }
  fclose(traps_file);	
}

/*******************************************************************************************
 start trap calculation
*******************************************************************************************/
static void CalcTraps()
{
  int pla_nr;
  INTMATRIX *IncidenceMat, *MNull;
  INTMATRIX *PreMat, *PostMat;
  
  start_timer(20);
  
  IncidenceMat=conv_matrix(c_matrix);
  PreMat=conv_matrix(i_matrix);
  PostMat=conv_matrix(o_matrix);
  
  MNull=ZeroMatrix(no_places,1);
  for(pla_nr = 0; pla_nr < no_places; pla_nr++)
    {
      MNull->V[pla_nr][0] = places[pla_nr].m0;
    }
  
  numtraps=Traps(MNull, IncidenceMat, PostMat, PreMat);
  
  FreeMat(IncidenceMat);
  FreeMat(PostMat);
  FreeMat(PreMat);
  FreeMat(MNull);
  stop_timer(20);
}

static void usage()
{
	printf("Usage: traps netname\n");
	exit( -1 );
}


/*	the main procedure handles options and file access and
	calls subordinated procedures */
int main(argc, argv)
	int argc;
	char ** argv;
{
	char net_name[FILE_NAME], struct_name[FILE_NAME], inf_name[FILE_NAME];
	extern int optind;

	/* check arguments */
	if (argc == 1)
		usage();

	/* build filenames */
	sprintf(net_name,"%s", argv[optind]);
	sprintf(struct_name,"%s.STRUCT", net_name);
	sprintf(inf_name,"%s.DEFINFO", net_name);
	sprintf(model_name,"%s", argv[optind]);


	/* open STRUCT & DEFINFO files */
	if ((struct_file = fopen(struct_name, "w")) == NULL)
	{	printf("ERROR: could not open %s for writing\n", struct_name);
		exit(1);
	}
	if ((inf_file = fopen(inf_name, "r")) == NULL)
	{	printf("ERROR: could not open %s for reading\n", inf_name);
		exit(1);
	}

	/* begin net analysis */
	read_net(net_name);
	read_info(inf_file);		/* read additional information */

	if (no_places < 1)
	{	printf("Error: no places.\n");
		exit(1);
	}
	if (no_timed_trans < 1)
	{	printf("Error: no timed transitions.\n");
		exit(1);
	}

#ifdef DEBUG
	printf("I-Matrix:\n");
	print_matrix(i_matrix);
	printf("O-Matrix:\n");
	print_matrix(o_matrix);
	printf("M0:\n");
	for(i=0; i<no_places; i++)
	printf("%d ", places[i].m0);
#endif
					/* calculate the global c-matrix */
	c_matrix = sparse_function(o_matrix, i_matrix, no_trans, minus); 
#ifdef DEBUG
	printf("C-Matrix:\n");
 	print_matrix(c_matrix); 
#endif

	CalcTraps();
	output_result(time_passed(20));
	return 0;			/* o.k */
}
