/*****************************************************************************

Programm	:	traps

Komponente	:	traps.c
                        Berechnung von traps

Programmer	:	Alexander Huck
Date		:	1999

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "polycount.h"
#include "matrix.h"
#include "matrixIO.h"
#include "matrixOps.h"
#include "matrixProcs.h"
#include "hermite2.h"
#include "inequalities.h"
#include "intOps.h"
#include "invars.h"
#include "timer.h"

#include "traps.h"

//#define DEBUG
#define MAXROWSIZE 100000000   /*Maximum number of matrix-rows allowed during inequality solving*/
#define MIN(a,b) a<b ? a:b

INTMATRIX *Trap, *TrapMarking;

int Traps(INTMATRIX *_M, INTMATRIX *_IncidenceMat, INTMATRIX *_PostMat, INTMATRIX *_PreMat)
     /*input: initial marking M
       Incidence-matrix, Post-matrix (output-arcs), Pre-Matrix (Input arcs)
       */
{
INTMATRIX *Ineq, *Solution;
INTMATRIX *M, *IncidenceMat, *PostMat, *PreMat;
INTMATRIX *NTheta;
int sum;
int count, numsol;
int i,j;
USI nbtransitions = 0, nbplaces = 0;

 M = CopyMat(_M);
 IncidenceMat = CopyMat(_IncidenceMat);
 PostMat = CopyMat(_PostMat);
 PreMat = CopyMat(_PreMat);
 
 nbtransitions = IncidenceMat -> R;
 nbplaces = IncidenceMat -> C;
 
#ifdef DEBUG
 printf("The net has %i places and %i transitions\n", nbplaces, nbtransitions);
#endif
 
 if (nbplaces == 0) return 0; /*no places, no marking*/
 if (nbtransitions == 0) return 1; /*no transitions: only initial marking*/
 
 /*from PRE and POST generate the modified Incidence-matrix NTheta (->traps)
   in NTheta all entrys belonging to the output-arcs of a transition are
   replaced by the sum of its input-arcs*/
 
#ifdef DEBUG
 printf("Incidence matrix: transis ------>\n"); 
 PRINTMATI(0,IncidenceMat->C-1,0,IncidenceMat->R-1,TransMat(IncidenceMat));
 printf("POST matrix: transis ------>\n"); 
 PRINTMATI(0,PostMat->C-1,0, PostMat->R-1,TransMat(PostMat));
 printf("PRE matrix: transis ------>\n"); 
 PRINTMATI(0,PreMat->C-1,0,PreMat->R-1,TransMat(PreMat));
#endif

 /*build NTheta*/
 NTheta = ZeroMatrix(nbtransitions, nbplaces);
 for(i=0; i<nbtransitions; i++)
   for(j=0; j<nbplaces; j++)
     {
       NTheta->V[i][j] = - PreMat->V[i][j];
     }
 
 for(i=0; i<nbtransitions; i++)
   {
     sum = 0;
     /*calculate sum of input-arcs*/
     for(j=0; j<nbplaces; j++)
       sum += PreMat->V[i][j];
     for(j=0; j<nbplaces; j++)
       {
	 if (PostMat->V[i][j] != 0) /*output-arc*/
	   {
	     NTheta->V[i][j] = sum;
	   }
       }
   }
 
#ifdef DEBUG
 printf("NTheta matrix: transis ------>\n"); 
 PRINTMATI(0,NTheta->C-1,0,NTheta->R-1,TransMat(NTheta));
#endif
 
 Ineq=ZeroMatrix(NTheta->R + nbplaces, NTheta->C+2);
 for (i=0; i<NTheta->R; i++)
   {
     Ineq->V[i][0]=1;
     for(j=1; j<Ineq->C-1; j++)  
       {
	 Ineq->V[i][j]=NTheta->V[i][j-1];
       }
   }
 /*add non-negativity constraints*/
 count = 1;
 for (i=NTheta->R; i<Ineq->R; i++)
   {
     Ineq->V[i][0] = 1;
     Ineq->V[i][count] = 1;
     count++;
   }
 
#ifdef DEBUG	    
 printf("eliminieren: \n");
#endif
 ElimEqualRows(&Ineq);
 ElimSimpleConstraints(&Ineq);
 
#ifdef DEBUG
 printf("sortieren: \n");
#endif
 SortRows(Ineq);
 
#ifdef DEBUG
 printf("Sortiert und eliminiert:\n ");
 PRINTMATI(0,Ineq->R-1,0,Ineq->C-1, Ineq);
#endif
 
 /*Solve the system of inequalities ZBASE.x>=0 */
 numsol=SolveInequationSystem(&Solution,Ineq, MAXROWSIZE);
 
 if (numsol)
   {
#ifdef DEBUG  
     printf("Lsung des Ungleichungssystems:\n ");
     PRINTMATI(0,Solution->R-1,0,Solution->C-1,Solution);	       
#endif	    
     
     /*the columns of Solution are the solutions of the trap equation
       calculate traps = supports of Solution
     */
     Trap = CopyMat(TransMat(Solution));
     
     for (i=0; i< Trap->R; i++)
       for(j=0; j< Trap->C; j++)
	 if (Trap->V[i][j]) Trap->V[i][j] = 1;
     
#ifdef DEBUG
     printf("Traps: (unreduced)\n ");
     PRINTMATI(0,Trap->R-1,0,Trap->C-1, Trap);
#endif
    
     ElimEqualRows(&Trap);
      
#ifdef DEBUG
     printf("Traps: (elimequalrows)\n ");
     PRINTMATI(0,Trap->R-1,0,Trap->C-1, Trap);
#endif
     
     ElimNonMinimal(&Trap);
 
#ifdef DEBUG
     printf("Traps:\n ");
     PRINTMATI(0,Trap->R-1,0,Trap->C-1, Trap);
#endif
   } /*if (numsol)*/
 else
   Trap = BuildMat(0,0);
 
 /*calculate initial Marking:
   Trap.M0
 */
 TrapMarking = MultMat(Trap, M);
#ifdef DEBUG	      
 printf("Marking:\n ");
 PRINTMATI(0,TrapMarking->R-1,0,TrapMarking->C-1, TrapMarking);
#endif

 return Trap->R;
}
