// File    : TimeWrapper.cpp      (C++ Implementation) 
// Module  : TimeWrapper          (Ressource Monitoring Functions)
//
// Purpose : Time and memory measuring functions.
//
// Author  : Oliver Hesse
// Date    : September 1999


#include <fstream>
#include "Time.h"


extern "C" {


// get time usages
Time* getCPUTime_C()
{
   return new Time(Time::getCPUTime());
}

Time* getTimeOfDay_C()
{
   return new Time(Time::getTimeOfDay());
}

// output of time in format "hh:mm:ss.msec"
void printTime_C(const Time *time)
{
   cout << *time << endl;
}

// output of time in format "sec.msec"
void printTimeSec_C(const Time *time)
{
   Time t = *time;
   cout << t.get_sec() << "." << setw(3) << setfill('0') << t.msecs() << endl;
}

// output of time difference in format "hh:mm:ss.msec"
void printTimeDiff_C(const Time *endTime, const Time *startTime)
{
   cout << *endTime - *startTime << endl;
}

// output of time difference in format "sec.msec" 
void printTimeDiffSec_C(const Time *endTime, const Time *startTime)
{
   Time t = *endTime - *startTime;
   cout << t.get_sec() << "." << setw(3) << setfill('0') << t.msecs() << endl;
}

// time operations
Time* addTimes_C(const Time *firstTime, const Time *secTime)
{
   return new Time(*firstTime + *secTime);
}

Time* subTimes_C(const Time *endTime, const Time *startTime)
{
   return new Time(*endTime - *startTime);
}

// write and load times
void saveTime_C(const Time *time, char* netname, char* timeType)
{
   string filename = string(netname) + "." + string(timeType) + ".time";

   ofstream os(filename.c_str());
   time->print_hmsm(os);
}

Time* loadTime_C(char* netname, char* timeType)
{   
   Time time;
   string filename = string(netname) + "." + string(timeType) + ".time";

   ifstream is(filename.c_str());
   is >> time;

   if (is.fail())
   {
      cout << "Warning: loadTime_C() of " << filename << " failed." << endl;
      return new Time(0.0);
   }
   return new Time(time);
}

// delete time pointer
void freeTime_C(Time *time)
{
   delete time;
}




// get memory usage
long getUsedMem_C()
{
   return getUsedMem();
}

// output of memory
void printMem_C(long mem)
{
   cout << mem << " KB" << endl;
}

// output of memory difference
long printMemDiff_C(long memAtEnd, long memAtStart)
{
   cout << memAtEnd - memAtStart << " KB" << endl;
}

// write and load memory
void saveMem_C(const long mem, char* netname, char* memType)
{
   string filename = string(netname) + "." + string(memType) + ".mem";

   ofstream os(filename.c_str());
   os << mem;
}

long loadMem_C(char* netname, char* memType)
{   
   long mem;
   string filename = string(netname) + "." + string(memType) + ".mem";

   ifstream is(filename.c_str());
   is >> mem;

   if (is.fail())
   {
      cout << "Warning: loadMem_C() of " << filename << " failed." << endl;
      return (long)0;
   }

   return mem;
}


}
