/************************************************************************
Modul	    :	fast_rusers.c

Purpose		: 	portable rusersd and rstatd-Requests
                (both Protocol-Version 2) with fast Timeout

Project Director:	Christoph Lindemann TU Berlin

Programmer	:	Kolja Koischwitz

History		:	release 1: August 1993
*                                                                           *
*  Modifiziert	  :	Joachim Schneider				    *
*									    *
*  Datum	  :	July 1994 					    *
************************************************************************/

#include <stdio.h> /* defines NULL */
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#define PORTMAP
#include <rpc/rpc.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <rpc/pmap_prot.h>
#include <netdb.h>
#include <sys/param.h>

#include "rusers.h" /* <rpcsvc/rusers.h> not identical on all systems */
#include "rstat.h"  /* <rpcsvc/rstat.h> as well */

/* portmapper_alive(): Gibt 1 zurueck, wenn der Portmapper auf "host"
   innerhalb von 1 sec antwortet.
 ----------------------------------------------------------------------*/
static int portmapper_alive(host)
char *host;
{	int ret, pmap_s=RPC_ANYSOCK; 
	struct timeval tout; CLIENT *cl;
	struct sockaddr_in on;
	struct hostent *he=gethostbyname(host);

	tout.tv_sec=1; tout.tv_usec=0;
	if(he==NULL) return(0);
	on.sin_family=AF_INET;
	on.sin_port=htons(PMAPPORT);
	memcpy(&on.sin_addr,he->h_addr,he->h_length);
	if((cl=clntudp_create(&on,PMAPPROG,PMAPVERS,tout,&pmap_s))==NULL)
	{
#ifdef debug
		clnt_perror(cl,"clntudp_create(PORTMAPPER)");
#endif
		return(0);
	}
	ret=clnt_call(cl,PMAPPROC_NULL,xdr_void,NULL,xdr_void,NULL,tout)==RPC_SUCCESS;
#ifdef debug
	if(!ret) clnt_perror(cl,"clnt_call(PORTMAPPER)");
#endif
	clnt_destroy(cl);
	return(ret);
}

/* rusers2(): rusersd-Request nach Protokoll-Version 2 (RUSERSVERS_IDLE),
   mit Timeout 2 Sec. Das c-library-Aequivalent rusers() verwendet unter
   Solaris2 leider schon Version 3, und hat ausserdem ein weitaus groesseres
   Timeout. Gibt 0 zurueck, wenn alles OK ist! (Wie die lib-Routine..)
 ----------------------------------------------------------------------*/
int rusers2(host, up)
char *host;
utmpidlearr *up;
{	CLIENT *cl; int success;
	struct timeval tout;
	
	tout.tv_sec=2; tout.tv_usec=0;
	if(!portmapper_alive(host)) return(-1);
	if((cl=clnt_create(host,RUSERSPROG,RUSERSVERS_IDLE,"udp"))==NULL)
		return(-1);
	memset(up,0,sizeof(utmpidlearr));
	success=clnt_call(cl,RUSERSPROC_NAMES,xdr_void,NULL,xdr_utmpidlearr,
		(caddr_t)up,tout)==RPC_SUCCESS;
	clnt_destroy(cl);
	return(success?0:-1);
}

/* rstat2(): rstatd-Request nach Protokoll-Version 2 (RSTATVERS_SWTCH),
   mit Timeout 2 Sec. Das c-library-Aequivalent rstat() benutzt unter
   Solaris 2 ebenfalls ein anderes Protokoll, und hat natuerlich auch
   ein groesseres Timeout. Gibt 0 zurueck, wenn alles OK ist!
 ----------------------------------------------------------------------*/
int rstat2(host, st)
char *host;
struct statsswtch *st;
{	CLIENT *cl; int success;
	struct timeval tout;
	
	tout.tv_sec=2; tout.tv_usec=0;
	if(!portmapper_alive(host)) return(-1);
	if((cl=clnt_create(host,RSTATPROG,RSTATVERS_SWTCH,"udp"))==NULL)
	{
#ifdef debug
		clnt_perror(cl,"clnt_create");
#endif
		return(-1);
	}
	memset(st,0,sizeof(struct statsswtch));
	success=clnt_call(cl,RSTATPROC_STATS,xdr_void,NULL,xdr_statsswtch,
		(caddr_t)st,tout)==RPC_SUCCESS;
#ifdef debug
	if(!success) clnt_perror(cl,"clnt_call");
#endif
	clnt_destroy(cl);
	return(success?0:-1);
}
