/******************************************************************************
Modul       : getusage.c

Purpose     :  determine and output rusage

Project Director:	Christoph Lindemann TU Berlin

Programmer	:	Kolja Koischwitz

History		:	release 1: October 1992
                release 2: August 1993
******************************************************************************/

#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "portab.h"

#ifdef __hpux
/* for HP-UX compatibility: */
#include <syscall.h>

/* Systemcalls */
#define getpagesize()				sysconf(_SC_PAGE_SIZE)
/* Der Systemcall SYS_getrusage liefert leider 0 fuer rusage.ru_maxrss */
#define getrusage(who,rusage)		syscall(SYS_getrusage,who,rusage)
#endif

/* Prozedur gibt verbrauchte Rechenzeit auf ms genau nach stdout */
void show_rusage()
{
#if defined(sun) && !defined(SOL2)
  long hours, mins, secs;
  struct rusage ru;

  if (getrusage(RUSAGE_SELF,&ru))
  {
  	perror("getrusage");
	return;
  }
  ru.ru_utime.tv_sec += ru.ru_stime.tv_sec;
  ru.ru_utime.tv_usec += ru.ru_stime.tv_usec;
  while (ru.ru_utime.tv_usec >= 1000000)
  {
    ru.ru_utime.tv_sec++;
   ru.ru_utime.tv_usec -= 1000000;
  }

  mins = hours = secs = ru.ru_utime.tv_sec;
  hours /= 3600; secs %= 60; mins /= 60; mins %= 60;

  printf ("Elapsed CPU time: %ld:%02ld:%02ld\n",
	  hours, mins, secs);
  printf ("Required memory: %lu KB\n", (ru.ru_maxrss * getpagesize ()) >> 10);
#else
  long secs, hours, mins;

#ifndef CLOCKS_PER_SEC
#define     CLOCKS_PER_SEC          1000000
#endif

  mins = hours = secs = (long) ( clock()/CLOCKS_PER_SEC );

  hours /= 3600;
  secs %= 60;
  mins /= 60; mins %= 60;
  printf ("Elapsed CPU time: %ld:%02ld:%02ld\n", hours, mins, secs);
  
#endif
}
