// File    Elem.h          (C++ class definition and inline implementation)
// Module  Elem            (Element access support for Vectors and Matrices)
// Library VMMATH          (Library for Vector and Matrix Algebra)

#ifndef ELEM_H
#define ELEM_H

/**
 *         Abstract base class of an object with a value that can be read.
 * 
 *@author  Philipp Berndt
 *@version $Id: Elem.h,v 1.2 2000/01/12 16:02:10 pumphutt Exp $
 *@see     ConstElemRef
 *@see     Elem
 */
class ConstElem
{
public:
    /** Destructor */
    virtual ~ConstElem() {};

    /** Reads the value. */
    virtual double get() const = 0;
};


/**
 *         Abstract base class of an object with a value that can be read
 *         and written.
 *
 *@author  Philipp Berndt
 *@version $Id: Elem.h,v 1.2 2000/01/12 16:02:10 pumphutt Exp $
 *@see     ElemRef
 *@see     ConstElem
 */
class Elem : public ConstElem
{
public:
    /** Writes the value. */
    virtual void set(double d) = 0;
};


/**
 *         Base class of a reference to a
 *@$       ConstElem
 *         .
 *
 *@author  Philipp Berndt
 *@version $Id: Elem.h,v 1.2 2000/01/12 16:02:10 pumphutt Exp $
 *@see     ConstElem
 *@see     ElemRef
 */
class ConstElemRef
{
public:
    /** Constructor */
    ConstElemRef(const ConstElem &r) : ref(r) {}

    /** Converts this reference to the referred value. */
    operator double() const      { return ref.get();         }
private:
    const ConstElem& ref;
};


/**
 *         Base class of a reference to an 
 *@$       Elem
 *         .
 *
 *@author  Philipp Berndt
 *@version $Id: Elem.h,v 1.2 2000/01/12 16:02:10 pumphutt Exp $
 *@see     ConstElem
 *@see     ElemRef
 */
class ElemRef
{
public:
    /** Constructor */
    ElemRef(Elem &r) : ref(r) {}

    /** Converts this reference to the referred value. */
    operator double() const      { return ref.get();         }

    double operator=  (double d) { ref.set(d); return d;     }
    double operator+= (double d) { return *this = *this + d; }
    double operator-= (double d) { return *this = *this - d; }
    double operator*= (double d) { return *this = *this * d; }
    double operator/= (double d) { return *this = *this / d; }
private:
    Elem& ref;
};

#endif // not ELEM_H
