// File    OnesVector.h    (C++ class definition and inline implementation)
// Library VMMATH          (Library for Vector and Matrix Algebra)

#ifndef ONESVECTOR_H
#define ONESVECTOR_H

#include "Vector.h"

/**
 *         A Vector of ones.
 *
 *         All elements of this vector always contain "1.0".
 *
 *@author  Philipp Berndt
 *@version $Id: OnesVector.h,v 1.2 2000/01/12 16:02:11 pumphutt Exp $
 *@see     OVRep
 */
class OnesVector : public Vector
{
    class OVRep;
public:
    /** Constructor */
    explicit OnesVector(int dim);

    /** Copy constructor */
    OnesVector(const OnesVector &v);

#ifdef PVM_SUPPORT
    /** Constructor that unpacks from active PVM receive buffer. */
    OnesVector(PVM_UNPACK_t);
#endif

private: // forbidden
    OnesVector& operator*= (const FilterMatrix &){ throw ConstException();}
    OnesVector& operator*= (const SparseMatrix &){ throw ConstException();}
    OnesVector& operator+= (const class SparseVector &) 
        { throw ConstException(); };
};


/**
 *         The representation of a OnesVector.
 *
 *@author  Philipp Berndt
 *@version $Id: OnesVector.h,v 1.2 2000/01/12 16:02:11 pumphutt Exp $
 *@see     OnesVector
 */
class OnesVector::OVRep :  public Vector::VRep
{
public:
    /** Constructor */
    explicit OVRep(int dim) : VRep(dim) {}

    /** Copy constructor */
    OVRep(const OVRep& r) : VRep(r.getCols()) {}

    /** Returns a new copy of this OVRep. */
    OVRep* clone() const { return new OVRep(*this); }

    /** Gets the value of one element, namely 1. */
    double get(int col) const { return 1.0; }

#ifdef PVM_SUPPORT
    /** Packs this DVRep into the active PVM send buffer. */
    void pvm_pack() const;
#endif

private: // forbidden
    void set(int col, double d) { throw ConstException(); }
};


/*****************************************************************************\
*                         INLINE FUNCTION DEFINITIONS                         *
*               No classes or user interfaces below this line !               *
\*****************************************************************************/

// Class OnesVector:

inline OnesVector::OnesVector(int dim)
    : Vector(new OnesVector::OVRep(dim)) {}

inline OnesVector::OnesVector(const OnesVector& v) 
    : Vector(v) {}

#ifdef PVM_SUPPORT
inline OnesVector::OnesVector(PVM_UNPACK_t)
    : Vector(new DenseVector::DVRep(pvm_upkint())) {}
#endif


// Class OVRep:

#ifdef PVM_SUPPORT
inline void OnesVector::OVRep::pvm_pack() const 
{ pvm_pkint(getCols()); }
#endif

#endif // not ONESVECTOR_H
