// File    vmmath.h        (Public C++ Header)
// Library VMMATH          (Library for Vector and Matrix Algebra)
/**
 *         Library for Vector and Matrix Algebra
 * 
 * This library offers various vector- and matrix-types and their common
 * algebraic operations and is designed for universal applicability and
 * easy extensibility.
 * 
 *@author  Philipp Berndt
 *@version 1.0
 *@pkgdoc  vmmath
 */

#ifndef VMMATH_H
#define VMMATH_H

// Check dimensions of vectors and matrices ? Set to 'true' or 'false'!
#define DIMCHECK false
#include <iostream>

// Send big Objects in multiple PVM-messages ?
//#define PVM_FRAGMENT

/**
 *         Exception thrown when an operation 
 *         with two Objects of incompatible dimensions is attempted.
 *         Define DIMCHECK as 'true' to make this work.
 *
 *@author  Philipp Berndt
 *@version $Id: vmmath.h,v 1.2 2000/01/12 16:02:14 pumphutt Exp $
 */
struct TypeMismatchException
{
    TypeMismatchException()
        { cerr << "TypeMismatchException raised!" << endl; }
};

/**
 *         Exception thrown when a 
 *         write access to a constant Element is attempted.
 *
 *@author  Philipp Berndt
 *@version $Id: vmmath.h,v 1.2 2000/01/12 16:02:14 pumphutt Exp $
 */
struct ConstException
{
    ConstException()
        { cerr << "ConstException raised!" << endl; }
};

inline void assertDim(bool assertion)
{
    if (DIMCHECK && !assertion)
        throw TypeMismatchException();
}

// Vector types:
#include "SparseVector.h"
#include "DenseVector.h"
#include "OnesVector.h"

// Matrix types:
#include "DiagMatrix.h"
#include "SparseMatrix.h"
#include "FilterMatrix.h"

#include "mixed.h" // Algebraic operators with mixed types 

#include "Rand.h"  // Random number generator

#endif // not VMMATH_H

