/*******************************************************************************
Modul:              showStatSimCurves.c

Purpose:            Routinen fuer plot des stat. Verhaltens definierter measures

Project Director:   Christian Kelling

Programmer:         Henning Ziegler

History:            10.96 (Dipl.Arbeit)
*******************************************************************************/

/* #define DEBUG */

#include <X11/Intrinsic.h>  /* Intrinsics Definitions*/
#include <X11/StringDefs.h>    /* Standard Name-String definitions*/

#include <X11/Shell.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/AsciiText.h>

#include <sys/param.h>
#include <math.h>
#include <stdio.h>
#include <signal.h>
#include <memory.h>
#include "environment.h"
#include "frontend.h"
#include "showStatSimCurves.h"
#include "toggle.bm"

/* file name buffer size */
#define FNAME_SIZE  256

#define TIMEOUT          500

#define BUTTONWIDTH      110     /* alle Werte in Pixel, Erfahrungswerte */
#define BUTTONWIDTH2      70
#define BUTTONHEIGHT      25
#define INIT_GC_BREITE   600    
#define INIT_GC_HOEHE    300
#define SCHOENER_RAND    20
#define MIN_ZAHL_ABSTAND_HORI 10
#define MIN_ZAHL_ANSTAND_VERT 10
#define MAX_ZAHL_BREITE  40
#define MAX_ZAHL_HOEHE   15
#define BIG_TICS         5
#define LITTLE_TICS      3
#define ANZ_LITTLE_TICS  4
#define FONTHALBEHOEHE   6
#define FONTHALBEBREITE   2
#define TOGGLE_LEFT_MARGIN 16
#define STR_BUTTONWIDTH      126


/* lokale Variablen */
static XtAppContext app_context;  
static Widget topLevel;
static Pixmap toggle_pixmap;      /* Haekchen */
static Drawing_Basis drabas;      /* Daten aus .curves file */
static FILE *fcurves;
static char statusline[1000];
static int no_data_blocks;
static char s_or_d[4];		  /*stop or done */

static Display *disp;
static Colormap cmap;
static int depth;
static GC gc;

static Widget CpannedWin;
static Widget CmainForm;
static Widget CbuttonForm;
static Widget CmswitchButton, CmswitchMenu, *CmswitchEntry;
static Widget CmbreakButton, CmbreakMenu, *CmbreakEntry;
static Widget ClswitchButton, ClswitchMenu, ClswitchEntry[3];
static Widget ClxswitchButton, ClxswitchMenu, ClxswitchEntry[2];
static Widget CbreakButton;     /* nur waehrend des loesungsvorgangs zu benutzen */
static Widget ChelpButton;
static Widget CstatusLabel;
static Widget CplotCore;

static Widget helpPopup, helpBox, helpdoneButton, helpText;

static Widget CcswitchButton, CcswitchMenu, CcswitchEntry[3];

static int gc_breite=INIT_GC_BREITE; 
static int gc_hoehe=INIT_GC_HOEHE;

/* folgende Vars. gelten nur fuer akt. Kachel. */
static int ka_breite;
static int ka_hoehe;
static int offsetx;
static int offsety;
static double minX;
static double maxX;
static double minY;
static double maxY;
static int scale;
static int scale_x;

static double udiffX;    /* pixelbildgroesse / reale argumentbereichsgroesse */
static double udiffY;    /*   -||-  */
static double rmaxX;     /* aufgerundetes maxX ist aber kleiner maxX+rticstepX */
static double rminX;     /* abgerundetes maxX ist aber groesser  minX-rticstepX */
static double lrminX;    /* log10(rminX); */
static double rmaxY;     /* aufgerundetes maxY ist aber kleiner maxY+rticstepy */
static double rminY;     /* abgerundetes maxY ist aber groesser  minY-rticstepy */
static double lrminY;    /* log10(rminY); */
static int    startx;    /* zum schnellen unrealx berechnen */
static int    starty;    /* zum schnellen unrealy berechnen */
static int    Sonderfall=false; /* wird nicht mehr gebraucht : immer false */

static int refresh_visu = True;
static Conf_Type confidence = Lines;
static XFontStruct *str_font_struct;
bool aktD_reduced;

/* Prototyping */
static void initDrawingBasis();
static void help_setup();
static void initmswitch();
/*static void initmbreak();*/
static void get_input();
static void spool();
static char *next_word();
static void mswitchCB();        /* Callback auf Meaure-Auswahl */
static void cswitchCB();        /* Callback auf Confidence ON/OFF-Auswahl */
static void mbreakCB();
static void lswitchCB();
static void lxswitchCB();
static void abbrechenCB();
static void ChelpCB();
static void helpDone();
static void exposeCurves();
static void redrawAll();
static void redrawKachel();
static void drawAxes();
static int unrealX();
static int unrealY();
static void drawNextStep();

/* actions & translations for the CplotCore widget */
static XtActionsRec graphActions[1] = {{ "Crefresh", exposeCurves }};
static String graphTrans = "<Expose>: Crefresh()";



/********************************************************************************/
int main(argc,argv)     /* einzige Argument ist der volle Name der .curves File */
/********************************************************************************/
int argc;
char **argv;
{
 topLevel = XtVaAppInitialize(
                 &app_context,"transsol",
                 NULL, 0,
                 &argc,argv,
                 NULL,
		 NULL);

  toggle_pixmap = XCreateBitmapFromData(XtDisplay(topLevel),
			RootWindowOfScreen(XtScreen(topLevel)),
			toggle_bits,toggle_width,toggle_height);


  CpannedWin = XtVaCreateManagedWidget(
                        "CpannedWin",
		        panedWidgetClass,
		        topLevel,
		        NULL);

  CmainForm = XtVaCreateManagedWidget(
                        "CmainForm",
		        formWidgetClass,
			CpannedWin,
		 XtNshowGrip,False,
		        NULL);

  CmswitchButton = XtVaCreateManagedWidget(
                        "CmswitchButton",
		        menuButtonWidgetClass,
			CmainForm,
		        XtNlabel,"switch measure",
			XtNwidth,BUTTONWIDTH,
		        XtNheight,BUTTONHEIGHT,
                        XtNright,XtChainLeft,
		        NULL);
  CmswitchMenu = XtVaCreatePopupShell(
        		"menu",
		        simpleMenuWidgetClass,
        		CmswitchButton,
		        NULL);

  ClxswitchButton = XtVaCreateManagedWidget(
        		"ClswitchButton",
		        menuButtonWidgetClass,
        		CmainForm,
        		XtNfromHoriz,CmswitchButton,
		        XtNlabel,"scale x",
			XtNwidth,BUTTONWIDTH2,
		        XtNheight,BUTTONHEIGHT,
                        XtNleft,XtChainLeft,
                        XtNright,XtChainLeft,
		        NULL);
  ClxswitchMenu = XtVaCreatePopupShell(
        		"menu",
		        simpleMenuWidgetClass,
        		ClxswitchButton,
		        NULL);

  ClxswitchEntry[0] = XtVaCreateManagedWidget(
        		"adaptive linear",
		        smeBSBObjectClass,
        		ClxswitchMenu,
			XtNleftMargin,TOGGLE_LEFT_MARGIN,
		        NULL);
  XtAddCallback(ClxswitchEntry[0],XtNcallback,(XtCallbackProc)lxswitchCB,(XtPointer) Adaptive_x);
  ClxswitchEntry[1] = XtVaCreateManagedWidget(
        		"absolute linear",
		        smeBSBObjectClass,
        		ClxswitchMenu,
			XtNleftMargin,TOGGLE_LEFT_MARGIN,
		        NULL);
  XtAddCallback(ClxswitchEntry[1],XtNcallback,(XtCallbackProc)lxswitchCB,(XtPointer) Absolute_x);
 
  XtVaSetValues(ClxswitchEntry[0],XtNleftBitmap,toggle_pixmap,NULL);


  ClswitchButton = XtVaCreateManagedWidget(
        		"ClswitchButton",
		        menuButtonWidgetClass,
        		CmainForm,
        		XtNfromHoriz,ClxswitchButton,
		        XtNlabel,"scale y",
			XtNwidth,BUTTONWIDTH2,
		        XtNheight,BUTTONHEIGHT,
                        XtNleft,XtChainLeft,
                        XtNright,XtChainLeft,
		        NULL);

  ClswitchMenu = XtVaCreatePopupShell(
        		"menu",
		        simpleMenuWidgetClass,
        		ClswitchButton,
		        NULL);

  ClswitchEntry[0] = XtVaCreateManagedWidget(
        		"adaptive linear",
		        smeBSBObjectClass,
        		ClswitchMenu,
			XtNleftMargin,TOGGLE_LEFT_MARGIN,
		        NULL);
  XtAddCallback(ClswitchEntry[0],XtNcallback,(XtCallbackProc)lswitchCB,(XtPointer) Adaptive);
  ClswitchEntry[1] = XtVaCreateManagedWidget(
        		"absolute linear",
		        smeBSBObjectClass,
        		ClswitchMenu,
			XtNleftMargin,TOGGLE_LEFT_MARGIN,
		        NULL);
  XtAddCallback(ClswitchEntry[1],XtNcallback,(XtCallbackProc)lswitchCB,(XtPointer) Absolute);

  ClswitchEntry[2] = XtVaCreateManagedWidget(
        		"logarithmic",
		        smeBSBObjectClass,
        		ClswitchMenu,
			XtNleftMargin,TOGGLE_LEFT_MARGIN,
		        NULL);
  XtAddCallback(ClswitchEntry[2],XtNcallback,(XtCallbackProc)lswitchCB,(XtPointer) Logarithmic);

  XtVaSetValues(ClswitchEntry[1],XtNleftBitmap,toggle_pixmap,NULL);

  CcswitchButton = XtVaCreateManagedWidget(
			"CcswitchButton",
			menuButtonWidgetClass,
			CmainForm,
			XtNfromHoriz,ClswitchButton,
			XtNlabel,"confidence",
			XtNwidth,STR_BUTTONWIDTH,
			XtNheight,BUTTONHEIGHT,
			XtNleft,XtChainLeft,
			XtNright,XtChainLeft,
			NULL);
  CcswitchMenu = XtVaCreatePopupShell(
			"menu",
			simpleMenuWidgetClass,
			CcswitchButton,
			NULL);

  CcswitchEntry[0] = XtVaCreateManagedWidget(
			"lines",
			smeBSBObjectClass,
			CcswitchMenu,
			XtNleftMargin,TOGGLE_LEFT_MARGIN,
			NULL);
  XtAddCallback(CcswitchEntry[0],XtNcallback,(XtCallbackProc)cswitchCB,(XtPointer) Lines);
  CcswitchEntry[1] = XtVaCreateManagedWidget(
			"dashes",
			smeBSBObjectClass,
			CcswitchMenu,
			XtNleftMargin,TOGGLE_LEFT_MARGIN,
			NULL);
  XtAddCallback(CcswitchEntry[1],XtNcallback,(XtCallbackProc)cswitchCB,(XtPointer) Dashes);
  CcswitchEntry[2] = XtVaCreateManagedWidget(
			"off",
			smeBSBObjectClass,
			CcswitchMenu,
			XtNleftMargin,TOGGLE_LEFT_MARGIN,
			NULL);
  XtAddCallback(CcswitchEntry[2],XtNcallback,(XtCallbackProc)cswitchCB,(XtPointer) Off);

  XtVaSetValues(CcswitchEntry[0],XtNleftBitmap,toggle_pixmap,NULL);

 strcpy(s_or_d,"stop");
 CbreakButton = XtVaCreateManagedWidget(
                        "CbreakButton",
		        commandWidgetClass,
                        CmainForm,
                        XtNfromHoriz,ClswitchButton,
 		        XtNlabel,s_or_d,
			XtNsensitive,True,
			XtNwidth,BUTTONWIDTH2,
		        XtNheight,BUTTONHEIGHT,
                        XtNleft,XtChainRight,
                        XtNright,XtChainRight,
		        NULL);
  XtAddCallback(CbreakButton,XtNcallback,(XtCallbackProc)abbrechenCB,NULL);

  ChelpButton = XtVaCreateManagedWidget(
                        "ChelpButton",
		        commandWidgetClass,
                        CmainForm,
			XtNlabel,"Help",
                        XtNfromHoriz,CbreakButton, 
			XtNwidth,BUTTONWIDTH,
		        XtNheight,BUTTONHEIGHT,
		        XtNleft,XtChainRight,
		        NULL);
  XtAddCallback(ChelpButton,XtNcallback,(XtCallbackProc)ChelpCB,NULL);

  memset(statusline,'\0',1000);
  CstatusLabel = XtVaCreateManagedWidget(
                       "CstatusLabel",
                       asciiTextWidgetClass,
		       CpannedWin,
		       XtNheight,BUTTONHEIGHT,
                       XtNdisplayCaret,False,
                       XtNeditType,XawtextRead,
                       XtNfromVert,CmswitchButton,
 		       XtNwidth,gc_breite,	
		       XtNstring,statusline,
		       XtNjustify,XtJustifyLeft,
		 XtNshowGrip,False,
		       NULL);
  /* parse action table */
  XtAppAddActions(app_context,graphActions,XtNumber(graphActions));
  CplotCore = XtVaCreateManagedWidget(
                       "CplotCore",
		       coreWidgetClass,
		       CpannedWin,
                       XtNfromVert,CstatusLabel,
	               XtNwidth,gc_breite,
                       XtNheight,gc_hoehe,
                       XtNtranslations,XtParseTranslationTable(graphTrans),
		 XtNshowGrip,False,
		       NULL);

  
  disp = XtDisplay(topLevel);
  gc = DefaultGC(disp,0);
  depth = DefaultDepth(disp,0);
  cmap = DefaultColormap(disp,0);


  initDrawingBasis(argv[1]);

  no_data_blocks = 2;

  initmswitch();

  XtRealizeWidget(topLevel);
  
  XtAppAddTimeOut(app_context,TIMEOUT,(XtTimerCallbackProc)get_input,NULL);
  XtAppMainLoop(app_context);

  return 0;
}

/******************************************************************************/
static void initDrawingBasis(arg1)
/******************************************************************************/
char *arg1;
{
  int i,j;
  int dummy;
  char buffer[BUF_SIZE];
  char *bs, *be;
  String modelpath;
  char fullfile[FNAME_SIZE];


  /* get environment variables */
  if ((modelpath = (String)getenv(MODELDIR_ENV)) == NULL)
  {
    printf("Environment variable %s not set; can't specified model",MODELDIR_ENV);
    exit(-1);
  }


  /* open file */
  sprintf(fullfile,"%s/%s",modelpath,arg1);
  if (!(fcurves=fopen(fullfile,"r"))) 
  {
      fprintf(stderr,"Unable to open file '%s' for reading.\n",fullfile );
      exit(-1);
  }

  /* read header info */
  /* anzahl x werte */
  spool(fcurves,buffer);
  sscanf(buffer,"%d", &(drabas.numDatas));

  for (i=0;i<drabas.numMeasures;i++) drabas.mdata[i].aktData = 0;

  /* anzahl measures */
  spool(fcurves,buffer);
  sscanf(buffer,"%d", &(drabas.numMeasures));

  /* allen moeglichen speicher allozieren */
  drabas.mdata = (Measure_Datas *)malloc(sizeof(Measure_Datas)*(drabas.numMeasures));
  for(i=0;i<drabas.numMeasures;i++)
  {
    drabas.mdata[i].time = (double *)malloc(sizeof(double)*(drabas.numDatas));
    drabas.mdata[i].Datas = (double *)malloc(sizeof(double)*(drabas.numDatas));
    drabas.mdata[i].conf_datas = (double *)malloc(sizeof(double)*(drabas.numDatas));
    drabas.mdata[i].stopped = False;
    str_font_struct = XQueryFont(disp,XGContextFromGC(gc));
    for(j=0;j<drabas.numDatas;j++)
    {
     drabas.mdata[i].time[j] = 0.0;
     drabas.mdata[i].Datas[j] = 0.0;
     drabas.mdata[i].conf_datas[j] = 0.0; 
    }
  }
  
  /* alle namen der measures */
  spool(fcurves,buffer);
  bs = buffer;
  for(i=0;i<drabas.numMeasures;i++)
  {
    be = next_word(bs);
    drabas.mdata[i].name = (char *)strdup(bs);
    bs = be;
  }

  /* maxX wert */
  spool(fcurves,buffer);
  sscanf(buffer,"%lf", &(drabas.mdata[0].maxX));
  for(i=0;i<drabas.numMeasures;i++)
  {
    drabas.mdata[i].maxX = drabas.mdata[0].maxX;
  }
  /* minX values at the beginning 0 */
  for(i=0;i<drabas.numMeasures;i++)
  {
    drabas.mdata[i].minX = 0.0;
  }

  /* alle maxY je measure */
  spool(fcurves,buffer);
  if(buffer[0]!='\n')
  {
    bs = buffer;
    for(i=0;i<drabas.numMeasures;i++)
    {
      sscanf(bs,"%lf", &(drabas.mdata[i].maxY));

      /* this spools buf to the next char after a the first space */
      if(i<drabas.numMeasures)
	while( *bs++ != ' ' )
	  ;
    }
  }
  else
  {
    for(i=0;i<drabas.numMeasures;i++)
    {
      drabas.mdata[i].maxY = 0.0;
    }
  }

  /* alle minY je measure */
  spool(fcurves,buffer);
  if(buffer[0]!='\n')
  {
    bs = buffer;
    for(i=0;i<drabas.numMeasures;i++)
    {
      sscanf(bs,"%lf", &(drabas.mdata[i].minY));

      /* this spools buf to the next char after a the first space; ausser beim letzten Eintrag */
      if(i<drabas.numMeasures)
	while( *bs++ != ' ' )
	  ;
    }  
  }
  else
  {
    for(i=0;i<drabas.numMeasures;i++)
    {
      drabas.mdata[i].minY = 1.0e300;
    }
  }


  /* number of runs (noch nicht verwendet) */
  spool(fcurves,buffer);
  sscanf(buffer,"%d", &dummy);

  /* letzten Kommentar weglesen */
  fgets(buffer,BUF_SIZE,fcurves);

  /* set sonstiges fuer alle measures */
  for(i=0;i<drabas.numMeasures;i++)
  {
    drabas.mdata[i].scale = Absolute;
    drabas.mdata[i].scale_x = Adaptive_x;
    drabas.mdata[i].visible = True;
    drabas.mdata[i].rminY = 1.0e300;
    drabas.mdata[i].rmaxY = 0;
  }
}

/******************************************************************************/
static void help_setup()
/******************************************************************************/
{
  Dimension width,height,t_width, wbuf,off_set;
  int distance;
  XFontStruct *fontStruct;

  helpPopup = XtVaCreatePopupShell(
			"Transient Solution Output Help Popup",
			transientShellWidgetClass,
			topLevel,
			NULL);
  helpBox = XtVaCreateManagedWidget(
		"helpBox",
		boxWidgetClass,
		helpPopup,
		XtNborderWidth,0,
		XtNhSpace,4,
		XtNvSpace,4,
		NULL);
  helpdoneButton = XtVaCreateManagedWidget(
		   "doneButton",
                    commandWidgetClass,
                    helpBox,
		    XtNwidth, BUTTONWIDTH,
                    XtNlabel,"Done",
                    XtNjustify,XtJustifyCenter,
                    NULL);


  XtAddCallback(helpdoneButton,XtNcallback,helpDone,NULL);

  helpText = XtVaCreateManagedWidget(
		"helpText",
		asciiTextWidgetClass,
		helpBox,
		XtNheight, 300,
                XtNwidth, 600,
		XtNstring,"There is no help-text available.\nSorry.",
		XtNlength, 300,
	        XtNdisplayCaret,False,
		XtNscrollVertical,XawtextScrollWhenNeeded,
		XtNwrap, XawtextWrapWord,
		XtNeditType, XawtextRead,
		NULL);

  XtVaGetValues (helpText,
		 XtNfont, &fontStruct,
		 XtNheight, &height,
		 NULL);  
}

/******************************************************************************/
static void initmswitch()
/******************************************************************************/
{
  int i;

  CmswitchEntry = (Widget *)emalloc( sizeof(Widget) * (drabas.numMeasures + 1) ); 

  for(i=0;i<drabas.numMeasures;i++)
  {
    CmswitchEntry[i] = XtVaCreateManagedWidget(
        		drabas.mdata[i].name,
		        smeBSBObjectClass,
        		CmswitchMenu,
			XtNleftMargin,TOGGLE_LEFT_MARGIN,
		        NULL);
    XtAddCallback(CmswitchEntry[i],XtNcallback,(XtCallbackProc)mswitchCB,(XtPointer) i);

    XtVaSetValues(CmswitchEntry[i],XtNleftBitmap,None,NULL);
    
    drabas.mdata[i].visible = True;
  }
  CmswitchEntry[i] = XtVaCreateManagedWidget(
        	     "show all",
		     smeBSBObjectClass,
        	     CmswitchMenu,
		     XtNleftMargin,TOGGLE_LEFT_MARGIN,
		     NULL);
  XtAddCallback(CmswitchEntry[i],XtNcallback,(XtCallbackProc)mswitchCB,(XtPointer) i);

  XtVaSetValues(CmswitchEntry[i],XtNleftBitmap,toggle_pixmap,NULL);
}


/******************************************************************************/
static void get_input()
/******************************************************************************/
{
  char buf[BUF_SIZE];
  char zeichen;
  char *bs;
  char *newl;
  int i,j,k, measure_index;
  static int offset = 0;
  int measure_id;
  double aktTime;
  double help_minY, help_maxY;
  Drawable drawon;
  Arg args[1];

  memset(buf,'\0',BUF_SIZE);
  clearerr(fcurves);
  while(!feof(fcurves))
  {
    zeichen = getc(fcurves);
    if(!feof(fcurves))
    {
      buf[offset++] = zeichen;

      if(zeichen == '\n')
      {
	if( (buf[0] == '#') || (buf[0] == '&') )
	{
          if( buf[0] == '#' )
	    {
	      memset(statusline,'\0',1000);
	      memccpy(statusline,&(buf[1]),'\n',1000);
	      newl=memchr(statusline,'\n',1000);
	      *newl = '\0';
	      XtVaSetValues(CstatusLabel,XtNstring,statusline,NULL);
	      XawTextEnableRedisplay(CstatusLabel);
              strcpy(s_or_d,"done");
              XtSetArg(args[0], XtNlabel,"done");  
              XtSetValues(CbreakButton,args,1);
              fclose(fcurves);
              return;
	    }
            else
	      {
                bs = buf; 
                *bs++;  /* '&' ueberlesen */  
                sscanf(bs,"%d",&measure_index);
                drabas.mdata[measure_index].stopped = True;
		i = 1;
                bs = buf;
                while( *bs++ != ' ')
                  i++; 
                memset(statusline,'\0',1000);
	        memccpy(statusline,&(buf[--i]),'\n',1000); /* Leerzeichen */
	        newl=memchr(statusline,'\n',1000);
	        *newl = '\0';
	        XtVaSetValues(CstatusLabel,XtNstring,statusline,NULL);
	        XawTextEnableRedisplay(CstatusLabel);
                for(i=drabas.mdata[measure_index].aktData;i<=drabas.numDatas;i++) 
                 drabas.mdata[measure_index].Datas[i]=-1;
                if (drabas.mdata[measure_index].aktData==0) 
                {
 		 drabas.mdata[measure_index].visible=false;
                 /* unbedingt noch Haekchen neu setzen: */
                 for(i=0;i<drabas.numMeasures;i++)
                  if(drabas.mdata[i].visible) 
                   XtVaSetValues(CmswitchEntry[i],XtNleftBitmap,toggle_pixmap,NULL);
                  else 
                  {
                   XtVaSetValues(CmswitchEntry[i],XtNleftBitmap,None,NULL);
                   XtSetSensitive(CmswitchEntry[i],false);
                  }
                 XtVaSetValues(CmswitchEntry[drabas.numMeasures],XtNleftBitmap,None,NULL);
                 /* XtSetSensitive(CmswitchEntry[drabas.numMeasures],false);*/
                }
	      }
	}
	else
	{
	  bs = buf;
	  /* measure_id lesen */
	  sscanf(bs,"%d",&measure_id);	  
          while(*bs++ != ' ');
	  /* zeit lesen */
	  sscanf(bs,"%lf",&aktTime);
          /* probs der measures und Konfidenzwerte lesen */
          if( drabas.mdata[measure_id].stopped == False )
	    {
              drabas.mdata[measure_id].time[drabas.mdata[measure_id].aktData]=aktTime;
              if(drabas.mdata[measure_id].aktData==1)
               drabas.mdata[measure_id].minX=drabas.mdata[measure_id].time[0];
              while(aktTime > drabas.mdata[measure_id].maxX)
              drabas.mdata[measure_id].maxX=drabas.mdata[measure_id].maxX*1.1;
              while(*bs++ != ' ');
              sscanf(bs,"%lf",&(drabas.mdata[measure_id].Datas[drabas.mdata[measure_id].aktData]));    
              while(*bs++ != ' ');
              sscanf(bs,"%lf",&(drabas.mdata[measure_id].conf_datas[drabas.mdata[measure_id].aktData]));	      
	    }
          if(!finite(drabas.mdata[measure_id].Datas[drabas.mdata[measure_id].aktData]))
	  {
	    printf("warning: measure %s has an infinite value t time %g.\n",drabas.mdata[measure_id].name,drabas.mdata[measure_id].time[drabas.mdata[measure_id].aktData]);
	    drabas.mdata[measure_id].Datas[drabas.mdata[measure_id].aktData] = 0.0;
	  }
          drabas.mdata[measure_id].aktData++; 
          if(drabas.mdata[measure_id].aktData>(drabas.numDatas))
          {
           for(j=1;j<drabas.mdata[measure_id].aktData;j+=2)
           { 
            if(!drabas.mdata[measure_id].stopped)
	     {
              drabas.mdata[measure_id].Datas[(j-1)/2]=0;
              drabas.mdata[measure_id].conf_datas[(j-1)/2]=0;
	      
              drabas.mdata[measure_id].Datas[(j-1)/2]=drabas.mdata[measure_id].Datas[j];
              drabas.mdata[measure_id].conf_datas[(j-1)/2]=drabas.mdata[measure_id].conf_datas[j];

              drabas.mdata[measure_id].Datas[j]=0;
              drabas.mdata[measure_id].conf_datas[j]=0;

	      drabas.mdata[measure_id].time[(j-1)/2]=0;
              drabas.mdata[measure_id].time[(j-1)/2]=drabas.mdata[measure_id].time[j];
              drabas.mdata[measure_id].time[j]=0;
             }
           }
           if (!drabas.mdata[measure_id].stopped) 
             drabas.mdata[measure_id].minX=drabas.mdata[measure_id].time[0];
           drabas.mdata[measure_id].aktData=drabas.numDatas/2;
           aktD_reduced=true;
          }
  	  /* max und min veraendert?  */ 
          if((drabas.mdata[measure_id].aktData == drabas.numDatas) 
             || (drabas.mdata[measure_id].aktData && !drabas.mdata[measure_id].time[drabas.mdata[measure_id].aktData])
             /* hezi: Indikator fuer keine weiteren Daten <- trotzdem darstellen*/
             || (aktD_reduced==true)                            )
	    {
             if(!drabas.mdata[measure_id].stopped)             
	      {
                help_minY = drabas.mdata[measure_id].Datas[0] - 
                            drabas.mdata[measure_id].conf_datas[0];
                help_maxY = drabas.mdata[measure_id].Datas[0] +
                            drabas.mdata[measure_id].conf_datas[0];
                for(j=0;j<drabas.mdata[measure_id].aktData;j++)
		  {
	            if( (drabas.mdata[measure_id].Datas[j] - 
                         drabas.mdata[measure_id].conf_datas[j] )
                                   < help_minY)
	               help_minY = drabas.mdata[measure_id].Datas[j] - 
                                  drabas.mdata[measure_id].conf_datas[j];
                    if( (drabas.mdata[measure_id].Datas[j] + 
                         drabas.mdata[measure_id].conf_datas[j] )
                                   > help_maxY)
	               help_maxY = drabas.mdata[measure_id].Datas[j] + 
                                  drabas.mdata[measure_id].conf_datas[j];
		  }
                drabas.mdata[measure_id].minY = help_minY;
                XtVaSetValues(ClswitchEntry[ drabas.mdata[measure_id].scale],XtNleftBitmap,None,NULL);           
                if(help_minY<0.0) drabas.mdata[measure_id].scale=Adaptive;
                /*printf("help_minY=%g\t..minX=%g\t..minY=%g\n",
                       help_minY,drabas.mdata[measure_id].minX,drabas.mdata[measure_id].minY);*/
                drabas.mdata[measure_id].maxY = help_maxY;
	      }           
            redrawAll();
            aktD_reduced=false;
	  }
	}
        offset = 0;
        memset(buf,'\0',BUF_SIZE);
      }
    }
  }
#ifdef DEBUG
  printf("--> leave\n");
#endif

   XtAppAddTimeOut(app_context,TIMEOUT,(XtTimerCallbackProc)get_input,NULL);
}

/*******************************************************************************/
static void spool(fp, buf)
/*******************************************************************************/
FILE *fp;
char *buf;
{
   do
   {
      fgets(buf,BUF_SIZE,fp);
   } while( buf[0] == COMMENT );

   if( strlen(buf) > 1 )
      buf[strlen(buf) - 1] = '\0';
}


/*******************************************************************************/
static char *next_word(start)
/*******************************************************************************/
char *start;
{
   char *end;

   /* make sure we don't segmentation fault */
   if( start == NULL )
      return( NULL );

   /* search for end of word */
   for( end = start; (*end != ' ') && (*end != '\t') && (*end != '\0'); end++ )
      ;

   /* check if we are at the end of the string */
   if( *end == '\0' )
      return( NULL );
   else
   {
      *end++ = '\0';
      return( end );
   }
}

/******************************************************************************/
static void mswitchCB(w, clientData, callData)
/******************************************************************************/
Widget w;
XtPointer clientData, callData;
{
  int i,j;
  int num;
  int findex;
  Drawable drawon;
  
  num =  (int)((long)clientData);
  refresh_visu = True;
  if(num == drabas.numMeasures)         /* want all measures */
  {
    for(i=0;i<drabas.numMeasures;i++)
    {
      XtVaSetValues(CmswitchEntry[i],XtNleftBitmap,None,NULL);
      drabas.mdata[i].visible = True;
    }
    XtVaSetValues(CmswitchEntry[ num ],XtNleftBitmap,toggle_pixmap,NULL);
  }
  else                                  /* want special measure */
  {  
    for(i=0;i<drabas.numMeasures;i++)
      if(drabas.mdata[i].visible == False)
	break;
    if(i==drabas.numMeasures)           /* was  previously everything enabled ? */
    {
      for(i=0;i<drabas.numMeasures;i++)
      {
	XtVaSetValues(CmswitchEntry[i],XtNleftBitmap,None,NULL);
	drabas.mdata[i].visible = False;
      }

      XtVaSetValues(CmswitchEntry[drabas.numMeasures],XtNleftBitmap,None,NULL);
      
      XtVaSetValues(CmswitchEntry[ num ],XtNleftBitmap,toggle_pixmap,NULL);
      drabas.mdata[num].visible = True;
    }
    else                                /* there where still some measures to enable */
    {
      if(drabas.mdata[num].visible)        /* want to switch "off" special measure */
      {
	XtVaSetValues(CmswitchEntry[num],XtNleftBitmap,None,NULL);
	drabas.mdata[num].visible = False;
      }
      else                                 /* want to switch "on" special measure */
      {
	XtVaSetValues(CmswitchEntry[num],XtNleftBitmap,toggle_pixmap,NULL);
	drabas.mdata[num].visible = True;
      }

      for(i=0;i<drabas.numMeasures;i++)    /* last measures switched off ? */
	if(drabas.mdata[i].visible == True)
	  break;
      if(i==drabas.numMeasures)            /* yes --> wanna see all measures */
      {
	for(i=0;i<drabas.numMeasures;i++)
        {
	  drabas.mdata[i].visible = True;

	}
	XtVaSetValues(CmswitchEntry[ drabas.numMeasures ],XtNleftBitmap,toggle_pixmap,NULL);
      }

      for(i=0;i<drabas.numMeasures;i++)   /* last measures switched on ? */
	if(drabas.mdata[i].visible == False)
	  break;
      if(i==drabas.numMeasures)           /* yes --> wanna see all measures */
      {
	for(i=0;i<drabas.numMeasures;i++)
        {
	  XtVaSetValues(CmswitchEntry[i],XtNleftBitmap,None,NULL);
	}

	XtVaSetValues(CmswitchEntry[ drabas.numMeasures ],XtNleftBitmap,toggle_pixmap,NULL);
      }
    }
  }

  /* set logscale haekchen */
  findex = 0;
  while( !(drabas.mdata[findex].visible) )
    findex++;

  for(i=0;i<3;i++)
  {
    XtVaSetValues(ClswitchEntry[i],XtNleftBitmap,None,NULL);
    XtVaSetValues(ClxswitchEntry[i],XtNleftBitmap,None,NULL);
  }
  XtVaSetValues(ClswitchEntry[ drabas.mdata[findex].scale ],XtNleftBitmap,toggle_pixmap,NULL);
  XtVaSetValues(ClxswitchEntry[ drabas.mdata[findex].scale_x ],XtNleftBitmap,toggle_pixmap,NULL);  

  for(i=0;i<drabas.numMeasures;i++)
  {
    if( (drabas.mdata[i].scale != drabas.mdata[findex].scale) && drabas.mdata[i].visible )
      for(j=0;j<3;j++)
	  XtVaSetValues(ClswitchEntry[j],XtNleftBitmap,None,NULL);
    if( (drabas.mdata[i].scale_x != drabas.mdata[findex].scale_x) && drabas.mdata[i].visible )
      for(j=0;j<3;j++)
	  XtVaSetValues(ClxswitchEntry[j],XtNleftBitmap,None,NULL);
  }

  /* clear all */
  drawon = XtWindow(CplotCore);
  XSetForeground(disp,gc,WhitePixel(disp,0));
  XFillRectangle(disp,drawon,gc,0,0,gc_breite,gc_hoehe);
  redrawAll();
}

/******************************************************************************/
static void mbreakCB(w, clientData, callData)
/******************************************************************************/
Widget w;
XtPointer clientData, callData;
{
  /* Simulation der einzelnen Slaves beenden */
}

/******************************************************************************/
static void lswitchCB(w, clientData, callData)
/******************************************************************************/
Widget w;
XtPointer clientData, callData;
{
  int i;
  int log;

  log = (int)((long)clientData);
  refresh_visu = True;
  /* set logscale fuer alle sichtbaren measures */
  for(i=0;i<drabas.numMeasures;i++)
  {
    if(drabas.mdata[i].visible == True)
    {
      drabas.mdata[i].scale = log;
      if(log==Adaptive)
	drabas.mdata[i].rminY = drabas.mdata[i].minY;
    }
  }

  /* Haekchen verwalten */
  for(i=0;i<3;i++)
    XtVaSetValues(ClswitchEntry[i],XtNleftBitmap,None,NULL);

  XtVaSetValues(ClswitchEntry[log],XtNleftBitmap,toggle_pixmap,NULL);

  redrawAll();
}

/******************************************************************************/
static void lxswitchCB(w, clientData, callData)
/******************************************************************************/
Widget w;
XtPointer clientData, callData;
{
  int i;
  int log;

  log = (int)((long)clientData);
  refresh_visu = True;
  /* set logscale fuer alle sichtbaren measures */
  for(i=0;i<drabas.numMeasures;i++)
  {
    if(drabas.mdata[i].visible == True)
    {
      drabas.mdata[i].scale_x = log;
      if(log==Adaptive_x)
	drabas.mdata[i].rminX = drabas.mdata[i].minX;
    }
  }

  /* Haekchen verwalten */
  for(i=0;i<2;i++)
    XtVaSetValues(ClxswitchEntry[i],XtNleftBitmap,None,NULL);
  XtVaSetValues(ClxswitchEntry[log],XtNleftBitmap,toggle_pixmap,NULL);

  redrawAll();
}

/******************************************************************************/
static void cswitchCB(w, clientData, callData)
/******************************************************************************/
Widget w;
XtPointer clientData, callData;
{
  int i;

  refresh_visu = True; 
  confidence = (int)((long)clientData);
  /* Haekchen verwalten */
  for(i = 0; i < 3; i++)
    XtVaSetValues(CcswitchEntry[i],XtNleftBitmap,None,NULL);
  XtVaSetValues(CcswitchEntry[confidence],XtNleftBitmap,toggle_pixmap,NULL);
  redrawAll();
}

/******************************************************************************/
static void abbrechenCB(w,clientData,callData)
/******************************************************************************/
Widget w;
XtPointer clientData,callData;
{ 
  Arg args[1];

  XtPopdown(topLevel);
  fprintf(stderr,"%s\n",stderr,s_or_d);
  if (!strcmp(s_or_d,"stop"))
  {
    system ("$TNETHOME/Simulation/scripts/KillMaster");
    strcpy(s_or_d,"done");
    XtSetArg(args[0], XtNlabel,"done");  
    XtSetValues(CbreakButton,args,1);
  }
  else
  {
    fclose(fcurves);
    exit(1);
  }
}

/******************************************************************************/
static void ChelpCB(w,clientData,callData)
/******************************************************************************/
Widget w;
XtPointer clientData,callData;
{
   String tnethome;
   char path[MAXPATHLEN];
   Position x,y;


   /* get environment variables */
   if ((tnethome = (String)getenv(DSPNPACKAGE_ENV)) == NULL)
   {
     printf("Environment variable %s not set; can't find multi-solve program",DSPNPACKAGE_ENV);
     exit(-1);
   } 
   sprintf(path,"%s/TransientSolution/help/TransSimCurves.help", tnethome);
   /* check if widget exists */
   if( !helpPopup )
    help_setup();
   /* initialize box */
   XtVaSetValues(helpText, 
		 XtNstring, path, 
		 XtNtype, XawAsciiFile,
		 NULL);
   XtTranslateCoords(topLevel,0,200,&x,&y);
   XtVaSetValues(helpPopup,
                 XtNx,x-5,
                 XtNy,y+110,
                 NULL);
   /* set parent insensitive, popup shell */
   XtSetSensitive(  ChelpButton,False); 
   XtPopup(helpPopup, XtGrabNone);
}

/******************************************************************************/
static void helpDone(w, clientData, callData)
/******************************************************************************/
    Widget w;
    XtPointer clientData, callData;
{
    XtSetSensitive(  ChelpButton, True);
    XtPopdown(helpPopup);
}


/******************************************************************************/
static void exposeCurves(w, event, params, num_params)
/******************************************************************************/
Widget w;
XEvent *event;
String *params;
Cardinal *num_params;
{
   Dimension b,h;
   int i,zwerg;

   /* only process last event */
   if( event==NULL || event->xexpose.count == 0 )  /* achtung bus error provoziert */
   {
     
     /* get present size of window */
     XtVaGetValues(w,
		   XtNwidth,&b,
		   XtNheight,&h,
		   NULL);
     gc_breite=(int)b;
     gc_hoehe=(int)h;
     refresh_visu = True;
     redrawAll();
   }
}



/******************************************************************************/
static void redrawAll()  
/******************************************************************************/
{
  int i,j;
  int n;
  int columns,rows,kachel_index,akt_kachel,rest;
  int kachel_breite;
  int kachel_hoehe;
  int x,y;
  int str_mitte;
  char *buf;
  int same_scale=true;


  /* berechnung der kacheln */
  n = 0;
  for(i=0;i<drabas.numMeasures;i++)
    if(drabas.mdata[i].visible == True)
      n++;

  columns = ceil(sqrt((double)n));  /* maximal benoetigte anzahl an kacheln */
  rows = columns - (int)( (columns * columns - n ) / columns ); /* kann ich zeilen einsparen ? */
  kachel_breite = gc_breite / columns;
  kachel_hoehe = gc_hoehe / rows;
 
  akt_kachel = 0;
  kachel_index = 0;
  while( !(drabas.mdata[kachel_index].visible) )
    kachel_index++;
  for(i=0;i<rows;i++)  
  {
    if(  (n - akt_kachel) >= columns )
    {
      for(j=0;j<columns;j++) 
      {
	drabas.mdata[kachel_index].kachel_breite = kachel_breite;
	drabas.mdata[kachel_index].kachel_hoehe = kachel_hoehe;
	drabas.mdata[kachel_index].kachel_x = j*kachel_breite;
	drabas.mdata[kachel_index].kachel_y = i*kachel_hoehe;
	akt_kachel++;
	kachel_index++;
	while( !(drabas.mdata[kachel_index].visible) )
	  kachel_index++;
      }
    }
    else /* weniger kacheln uebrig als spaltenplaetze */
    {
      rest = n - akt_kachel;

      /* clear freien rest */
      XSetForeground(disp,gc,WhitePixel(disp,0));
      XFillRectangle(disp,XtWindow(CplotCore),gc,
		     0, 
		     i*kachel_hoehe,
		     kachel_breite*(columns-rest)/2,
		     gc_hoehe);
      XFillRectangle(disp,XtWindow(CplotCore),gc,
		     gc_breite - kachel_breite*(columns-rest)/2,
		     i*kachel_hoehe,
		     gc_breite,
		     gc_hoehe);

      for(j=0;j<rest;j++) 
      {
	drabas.mdata[kachel_index].kachel_breite = kachel_breite;
	drabas.mdata[kachel_index].kachel_hoehe = kachel_hoehe;
	drabas.mdata[kachel_index].kachel_x = j*kachel_breite + kachel_breite*(columns-rest)/2;
	drabas.mdata[kachel_index].kachel_y = i*kachel_hoehe;
	akt_kachel++;
	kachel_index++;
	while( !(drabas.mdata[kachel_index].visible) )
	  kachel_index++;
      }
    }
  }
  /* gehe alle kacheln durch und lasse sie zeichnen */
  for(i=0;i<drabas.numMeasures;i++)
     if( drabas.mdata[i].visible == True )
        if( (refresh_visu == True )  || 
            (drabas.mdata[i].stopped == False) ) 
          redrawKachel(i); 
    refresh_visu = False; 
  j=0;
  while(!drabas.mdata[j].visible) j++;
  for(i=j;i<drabas.numMeasures;i++)    
   if(drabas.mdata[i].visible && (drabas.mdata[j].scale != drabas.mdata[i].scale))
   {
     same_scale=false;
     break;
   }
  if (same_scale)
  {
   for(i=0;i<3;i++)
    XtVaSetValues(ClswitchEntry[i],XtNleftBitmap,None,NULL);
   XtVaSetValues(ClswitchEntry[drabas.mdata[j].scale],XtNleftBitmap,toggle_pixmap,NULL);
  }
   
}

/******************************************************************************/
static void redrawKachel(num)
/******************************************************************************/
int num;
{
  int i;
  Drawable drawon;
  int x,y;
  int str_mitte;
  char *buf;

  ka_breite = drabas.mdata[num].kachel_breite;
  ka_hoehe = drabas.mdata[num].kachel_hoehe;
  offsetx = drabas.mdata[num].kachel_x;
  offsety = drabas.mdata[num].kachel_y;
  minX = drabas.mdata[num].minX;
  maxX = drabas.mdata[num].maxX;
  minY = drabas.mdata[num].minY;
  maxY = drabas.mdata[num].maxY;
  scale = drabas.mdata[num].scale;
  scale_x = drabas.mdata[num].scale_x;
  /*Sonderfall = False;*/

/*   printf("Kachel:%d   maxY %lf minY %lf scale %d sonderfall\n",num,maxY,minY,scale,Sonderfall); */
 
  drawon = XtWindow(CplotCore);


   if(scale==Logarithmic && minY<=0)
   {
       printf("logarithmic output for measure %s makes no sense : switch to linear-adaptive output.\n",
              drabas.mdata[num].name);
       scale = Adaptive;
       drabas.mdata[num].scale = Adaptive;
     } 
   if(scale==Logarithmic && minY == 0)
      {
        minY = 1.0e300;
        for(i=0;i<drabas.numDatas;i++)
        {
          if( ((drabas.mdata[num].Datas[i] - drabas.mdata[num].conf_datas[i]) > 0) && ( (drabas.mdata[num].Datas[i] - drabas.mdata[num].conf_datas[i]) < minY) )
          minY = drabas.mdata[num].Datas[i];
        }
        if(minY==1.0e300)
        {
          printf("logarithmic output for measure %s makes no sense : switch to linear-adaptive output.\n",
                 drabas.mdata[num].name);
          scale = Adaptive;
          drabas.mdata[num].scale = Adaptive;
	}
	} 
  drawAxes(drawon,num); 

  if(scale != drabas.mdata[num].scale)
  {
   /*  printf("scale von %d nach %d bei measure %d\n",drabas.mdata[num].scale,scale,num);  */
    for(i=0;i<3;i++)
      XtVaSetValues(ClswitchEntry[i],XtNleftBitmap,None,NULL);
  }
  drabas.mdata[num].scale = scale;
  drabas.mdata[num].udiffX = udiffX;
  drabas.mdata[num].udiffY = udiffY;
  if(!Sonderfall)
  {
    drabas.mdata[num].rminX = rminX;
    drabas.mdata[num].rmaxX = rmaxX;
    drabas.mdata[num].rminY = rminY;
    drabas.mdata[num].rmaxY = rmaxY;
  }
  drabas.mdata[num].lrminY = lrminY;
  drabas.mdata[num].startx = startx;
  drabas.mdata[num].starty = starty;

/*   printf("Kachel:%d   rmaxY %lf rminY %lf scale %d sonderfall\n", 
 	 num, 
 	 drabas.mdata[num].rmaxY, 
 	 drabas.mdata[num].rminY, 
 	 drabas.mdata[num].scale, 
 	 Sonderfall); */


 /* Titel zeichnen */
  x = offsetx + ka_breite/2;
  y = offsety +  MAX_ZAHL_HOEHE;
  buf = drabas.mdata[num].name;
  str_mitte = XTextWidth(str_font_struct,buf,strlen(buf))/2;
  XDrawImageString(disp,drawon,gc,x-str_mitte,y,buf,strlen(buf));
  for(i=0;i<drabas.mdata[num].aktData;i++)
  {
   if(drabas.mdata[num].Datas[i]!=-1) /* Indikator ob gueltiger Wert */
    drawNextStep(i,num);
  }
}

/******************************************************************************/
static void drawAxes(drawon, num)  
/******************************************************************************/
Drawable drawon;
int num;
{
    double rmaxX;
    double lmaxX;
    double lmaxY;     /* log10(maxY) oder 0 */
    double lminX;
    double lminY;
    double ticstep;   /* genaue beschriftungsabstand */
    double fticstep;  /* Groessenordnung von ticstep z.B. bei 965 -> 2 */
    double rticstepx;  /* auf eine stelle aufgerundet fuer x Achse */
    double rticstepy;  /* auf eine stelle aufgerundet fuer x Achse */
    double max_zahlen;
    int anz_ytics,anz_xtics;   
    int tic,x,y;            /* zum zeichnen der tics */
    double i;
    double teile;
    char buf[BUF_SIZE];
    int str_mitte;
    double  zae,nen;
    /* x-achse beschriften vorbereiten */
    sprintf(buf,"%.3g\0",maxX);
    str_mitte = XTextWidth(str_font_struct,buf,strlen(buf))/2;
    zae = (double) (ka_breite - (2*SCHOENER_RAND - MAX_ZAHL_BREITE - str_mitte));
    nen = (double) (MIN_ZAHL_ABSTAND_HORI + MAX_ZAHL_BREITE);
    max_zahlen = floor(zae / nen);
    if(max_zahlen < 1)
      max_zahlen = 1;
    if(scale_x==Absolute_x)
      minX = 0.0;

    ticstep = (maxX - minX)/max_zahlen;     /* ticstep leider rational --> ticstep runden */    

    fticstep = pow(10.0, floor(log10(ticstep)) + 1);  /* max_zahlen ist mind. 1 und diffY ungleich 0 */
    rticstepx = ceil(ticstep/fticstep)*fticstep;  /* aufrunden */
    rminX = floor(minX/fticstep)*fticstep;   /* maxY ist immer groesser ticstep */
	
    ticstep = (maxX-rminX)/max_zahlen;     /* ticstep leider rational --> ticstep runden */    
    fticstep = pow(10.0, floor(log10(ticstep)) + 1);  /* max_zahlen ist mind. 1 und diffY ungleich 0 */
    rticstepx = ceil(ticstep/fticstep)*fticstep;  /* aufrunden */

    anz_xtics = ceil( (maxX-rminX)/rticstepx );
    rmaxX = rminX+anz_xtics*rticstepx;

    if( (rmaxX/maxX > 1.2) || (max_zahlen/anz_xtics > 2.0) ) /* zuviel Platz verschenkt, oder zu wenig Zahlen ? */
    {
      fticstep = pow(10.0,floor(log10(ticstep)));  /* max_zahlen ist mind. 1 und diffY ungleich 0 */
      rticstepx = ceil(ticstep/fticstep)*fticstep;  /* aufrunden */
      rminX = floor(minX/fticstep)*fticstep;   /* maxY ist immer groesser ticstep */
	
      ticstep = (maxX-rminX)/max_zahlen;     /* ticstep leider rational --> ticstep runden */    
      fticstep = pow(10.0,floor(log10(ticstep)));  /* max_zahlen ist mind. 1 und diffY ungleich 0 */
      rticstepx = ceil(ticstep/fticstep)*fticstep;  /* aufrunden */

      anz_xtics = ceil( (maxX-rminX)/rticstepx );
      rmaxX = rminX+anz_xtics*rticstepx;
    }
    udiffX = (ka_breite-2*SCHOENER_RAND-MAX_ZAHL_BREITE) / (rmaxX-rminX);
    startx=SCHOENER_RAND+MAX_ZAHL_BREITE;      
    
    /* y-achse beschriften vorbereiten */
    zae = (double)(ka_hoehe-SCHOENER_RAND-2*MAX_ZAHL_HOEHE);
    nen = (double)(MIN_ZAHL_ANSTAND_VERT+MAX_ZAHL_HOEHE);
    max_zahlen = ceil(zae / nen );
    if(max_zahlen < 1)
      max_zahlen = 1;

    if(minY >= maxY)                           /****** Sonderfall *******/
    {
        Sonderfall = false;                    /* bisher 01/97 True */
        if (scale!= Absolute) 
        {
          printf("selected output for measure %s makes no sense  :switch to linear-absolute output.\n",
                drabas.mdata[num].name);
	  scale = Absolute;
        }
        if(maxY==0.0)    /* sind alle Werte gleich und 0 */
	{
	  rminY = 0.0;
	  rmaxY = 1.0;
	  anz_ytics = 1;
	  rticstepy = 1.0;
	}
	else           /* sind alle Werte nur gleich */
	{	
	  rminY = 0.0;
	  rmaxY = 2.0*maxY;
	  anz_ytics = 2;
	  rticstepy = maxY;
	}
	udiffY = (ka_hoehe-2*SCHOENER_RAND-MAX_ZAHL_HOEHE) / (rmaxY-rminY);
    }
    else if(scale==Logarithmic )                          /****** lorarithmic *****/
    {
        Sonderfall=false;
        lminY = floor(log10(minY));                    
        lmaxY = ceil(log10(maxY));                     
        rticstepy = ceil((lmaxY-lminY) / max_zahlen);  /* exponentenschrittweite aufgerundet*/
	
	rmaxY = pow(10.0,lmaxY);  
        anz_ytics = ceil((lmaxY-lminY) / rticstepy);    
	rminY = pow(10.0,lmaxY-anz_ytics*rticstepy);
	lrminY=log10(rminY);                           /* 10 hoch irgendwas ist nie null */
        udiffY = (ka_hoehe-2*SCHOENER_RAND-MAX_ZAHL_HOEHE) / (log10(rmaxY)-lrminY); /* 10 hoch irgendwas ist nie null */
    }
    else                                                                 /****** NOT lorarithmic *****/
    {
        Sonderfall=false;
        if(scale==Absolute) minY = 0.0;
	ticstep = (maxY - minY)/max_zahlen;     /* ticstep leider rational --> ticstep runden */    

	fticstep = pow(10.0, floor(log10(ticstep)) + 1);  /* max_zahlen ist mind. 1 und diffY ungleich 0 */
	rticstepy = ceil(ticstep/fticstep)*fticstep;  /* aufrunden */
	rminY = floor(minY/fticstep)*fticstep;   /* maxY ist immer groesser ticstep */
	
	ticstep = (maxY-rminY)/max_zahlen;     /* ticstep leider rational --> ticstep runden */    
	fticstep = pow(10.0, floor(log10(ticstep)) + 1);  /* max_zahlen ist mind. 1 und diffY ungleich 0 */
	rticstepy = ceil(ticstep/fticstep)*fticstep;  /* aufrunden */

	anz_ytics = ceil( (maxY-rminY)/rticstepy );
	rmaxY = rminY+anz_ytics*rticstepy;

	if( (rmaxY/maxY > 1.2) || (max_zahlen/anz_ytics > 2.0) ) /* zuviel Platz verschenkt, oder zu wenig Zahlen ? */
        {
	  fticstep = pow(10.0,floor(log10(ticstep)));  /* max_zahlen ist mind. 1 und diffY ungleich 0 */
	  rticstepy = ceil(ticstep/fticstep)*fticstep;  /* aufrunden */
	  rminY = floor(minY/fticstep)*fticstep;   /* maxY ist immer groesser ticstep */
	
	  ticstep = (maxY-rminY)/max_zahlen;     /* ticstep leider rational --> ticstep runden */    
	  fticstep = pow(10.0,floor(log10(ticstep)));  /* max_zahlen ist mind. 1 und diffY ungleich 0 */
	  rticstepy = ceil(ticstep/fticstep)*fticstep;  /* aufrunden */

	  anz_ytics = ceil( (maxY-rminY)/rticstepy );
	  rmaxY = rminY+anz_ytics*rticstepy;
	}
        udiffY = (ka_hoehe-2*SCHOENER_RAND-MAX_ZAHL_HOEHE) / (rmaxY-rminY);
    }
    starty=SCHOENER_RAND+MAX_ZAHL_HOEHE;

    /* clear Kachel */
    XSetForeground(disp,gc,WhitePixel(disp,0));
    XFillRectangle(disp,drawon,gc,offsetx,offsety,ka_breite,ka_hoehe);

    /* achsen zeichnen */
    XSetForeground(disp,gc,BlackPixel(disp,0));
    XSetLineAttributes(disp,gc,2,LineSolid,0,0);
    XDrawLine(disp,drawon,gc,unrealX(rminX),
                             unrealY(rmaxY),
                             unrealX(rminX),
                             unrealY(rminY));  /* y achse */
    XDrawLine(disp,drawon,gc,unrealX(rminX),
                             unrealY(rminY),
                             unrealX(rmaxX),
                             unrealY(rminY));   /* x achse */
    
    /* y tics und text zeichnen */
    x=unrealX(rminX);
    if(scale==Logarithmic)                                               /****** lorarithmic *****/
    {
      for(tic=0;tic<anz_ytics;tic++)
      {
	XSetLineAttributes(disp,gc,2,LineSolid,0,0);
	y = unrealY(rminY*pow(10.0,tic*rticstepy));
	sprintf(buf,"10e%d\0",(int)(lrminY+tic*rticstepy));
        XDrawLine(disp,drawon,gc,x,y,x+BIG_TICS,y);
        XDrawImageString(disp,drawon,gc,offsetx+SCHOENER_RAND,y+FONTHALBEHOEHE,buf,strlen(buf));
	XSetLineAttributes(disp,gc,1,LineSolid,0,0);
	teile = (double)(ANZ_LITTLE_TICS*2);
	teile++;
	for(i=2; i<teile; i++)
        {
	  y = unrealY(i/teile*rminY*pow(10.0,(tic+1)*rticstepy));
          XDrawLine(disp,drawon,gc,x,y,x+LITTLE_TICS,y);
	}
      }
      XSetLineAttributes(disp,gc,2,LineSolid,0,0);  /* letzte Zahl aber keine Tics mehr */
      y = unrealY(rminY*pow(10.0,tic*rticstepy));    /* oder rmaxY */
      XDrawLine(disp,drawon,gc,x,y,x+BIG_TICS,y);
      sprintf(buf,"10e%d\0",(int)(lrminY+tic*rticstepy));  /* oder rmaxY */
      XDrawImageString(disp,drawon,gc,offsetx+SCHOENER_RAND,y+FONTHALBEHOEHE,buf,strlen(buf));

    }
    else                                                      /****** NOT lorarithmic *****/
    {
      for(tic=0;tic<anz_ytics;tic++)
      {
	XSetLineAttributes(disp,gc,2,LineSolid,0,0);
	y = unrealY(rminY+tic*rticstepy);
	sprintf(buf,"%5.4g\0",rminY+tic*rticstepy);
	XDrawLine(disp,drawon,gc,x,y,x+BIG_TICS,y);
        XDrawImageString(disp,drawon,gc,offsetx+SCHOENER_RAND,y+FONTHALBEHOEHE,buf,strlen(buf));
	XSetLineAttributes(disp,gc,1,LineSolid,0,0);
	teile = (double)ANZ_LITTLE_TICS;
	teile++;
	for(i=1; i<teile; i++)
	{
	  y = unrealY(rminY+tic*rticstepy + i/teile*rticstepy);
	  XDrawLine(disp,drawon,gc,x,y,x+LITTLE_TICS,y);
	}
      }
      XSetLineAttributes(disp,gc,2,LineSolid,0,0);  /* letzte Zahl aber keine Tics mehr */
      y = unrealY(rminY+tic*rticstepy);    /* oder rmaxY */
      XDrawLine(disp,drawon,gc,x,y,x+BIG_TICS,y);
      sprintf(buf,"%5.4g\0",rminY+tic*rticstepy);
      XDrawImageString(disp,drawon,gc,offsetx+SCHOENER_RAND,y+FONTHALBEHOEHE,buf,strlen(buf));
    }
    /* x tics und text zeichnen */
    y = unrealY(rminY);
    for(tic=0;tic<anz_xtics;tic++)
    {
        XSetLineAttributes(disp,gc,2,LineSolid,0,0);
        x = unrealX(rminX + tic*rticstepx);
        XDrawLine(disp,drawon,gc,x,y,x,y-BIG_TICS);
        sprintf(buf,"%.4g\0",rminX+tic*rticstepx);
        str_mitte = XTextWidth(str_font_struct,buf,strlen(buf))/2;
        XDrawImageString(disp,drawon,gc,x-str_mitte,y+MAX_ZAHL_HOEHE,buf,strlen(buf));
        XSetLineAttributes(disp,gc,1,LineSolid,0,0);
	teile= (double)ANZ_LITTLE_TICS;
        teile++;
        for(i=1;i<teile;i++)
        {
            x = unrealX(rminX + tic*rticstepx+i/teile*rticstepx);
            XDrawLine(disp,drawon,gc,x,y,x,y-LITTLE_TICS);
        }
     }
     XSetLineAttributes(disp,gc,2,LineSolid,0,0);  /* letzte Zahl aber keine Tics mehr */
     x = unrealX(rminX+tic*rticstepx);   /* oder rmaxY */
     XDrawLine(disp,drawon,gc,x,y,x,y-BIG_TICS);
     sprintf(buf,"%.4g\0",rminX+tic*rticstepx);
     XDrawImageString(disp,drawon,gc,x-str_mitte,y+MAX_ZAHL_HOEHE,buf,strlen(buf));
  
}

/******************************************************************************/
static int unrealX(x)
/******************************************************************************/
double x;
{
  if(scale_x==Adaptive_x && !Sonderfall)
  {
    /* udiffX=(ka_hoehe-unwichtig)/rdiffX; rdiffX=rmaxX-rminX */
    return(offsetx + startx + (x-rminX)*udiffX); 
  }
  else
  {
    /* udiffX=(ka_hoehe-unwichtig)/rdiffX; rdiffX=rmaxX-rminX */
    return(offsetx + startx + x*udiffX); 
  }
}

/******************************************************************************/
static int unrealY(y)
/******************************************************************************/
double y;
{

  if(scale==Logarithmic)
  {
    if(y>0)
      /* udiffY=(ka_hoehe-unwichtig)/lrdiffY;       */ 
      /* lrdiffY=lrmaxY-lrminY; lrminY=log10(rminY) */
      return(offsety + ka_hoehe - starty - (log10(y)-lrminY)*udiffY);
    else
      return(offsety + ka_hoehe - starty);
  }
  else if(scale==Adaptive && !Sonderfall)
  {
    /* udiffY=(ka_hoehe-unwichtig)/rdiffY; rdiffY=rmaxY-rminY */
    return(offsety + ka_hoehe - starty - (y-rminY)*udiffY); 
  }
  else
  {
    /* udiffY=(ka_hoehe-unwichtig)/rdiffY; rdiffY=rmaxY-rminY */
    return(offsety + ka_hoehe - starty - y*udiffY); 
  }
}

/******************************************************************************/
static void drawNextStep(index,num)
/******************************************************************************/
int index;
int num;
{  
  Drawable drawon;

  drawon = XtWindow(CplotCore);

  if(drabas.mdata[num].time[index]==0.0 )
     return;
  
  scale = drabas.mdata[num].scale;
  ka_hoehe = drabas.mdata[num].kachel_hoehe;
  offsetx = drabas.mdata[num].kachel_x;
  offsety = drabas.mdata[num].kachel_y;
  udiffX = drabas.mdata[num].udiffX;
  udiffY = drabas.mdata[num].udiffY;
  rminY = drabas.mdata[num].rminY;
  rmaxY = drabas.mdata[num].rmaxY;
  lrminY = drabas.mdata[num].lrminY;
  startx = drabas.mdata[num].startx;
  starty = drabas.mdata[num].starty;

  XSetLineAttributes(disp,gc,3,LineSolid,0,0);

  if( !(scale==Logarithmic && drabas.mdata[num].Datas[index-1]==0 && drabas.mdata[num].Datas[index]==0) )
    { 
      if(index == 0 && drabas.mdata[num].aktData==1)
      {
         XDrawLine(disp,drawon,gc, 
               unrealX(drabas.mdata[num].time[index])-5,
               unrealY(drabas.mdata[num].Datas[index]) -5,
               unrealX(drabas.mdata[num].time[index])+5,
               unrealY(drabas.mdata[num].Datas[index]) +5);  
         XDrawLine(disp,drawon,gc, 
               unrealX(drabas.mdata[num].time[index])-5,
               unrealY(drabas.mdata[num].Datas[index]) + 5,
               unrealX(drabas.mdata[num].time[index])+5,
               unrealY(drabas.mdata[num].Datas[index]) -5);
      }
      if(index > 0)
         XDrawLine(disp,drawon,gc, 
	        unrealX(drabas.mdata[num].time[index - 1]),
	        unrealY(drabas.mdata[num].Datas[index-1]) - 1,
	        unrealX(drabas.mdata[num].time[index]),
	        unrealY(drabas.mdata[num].Datas[index]) - 1);

      if( confidence != Off )
	{

          if(confidence == Lines)        
            XSetLineAttributes(disp,gc,1,LineSolid,0,0);
          else
            XSetLineAttributes(disp,gc,1,LineOnOffDash,0,0);
      
          if(index == 0 && drabas.mdata[num].aktData==1)
          {
	    /* XDrawLine(disp,drawon,gc, 
	      unrealX(drabas.mdata[num].time[index]),
	      unrealY(drabas.mdata[num].Datas[index] + 
                      drabas.mdata[num].conf_datas[index]) -1,
	      unrealX(drabas.mdata[num].time[index]),
	      unrealY(drabas.mdata[num].Datas[index] - 
                      drabas.mdata[num].conf_datas[index]) -1);*/
            XDrawLine(disp,drawon,gc, 
	      unrealX(drabas.mdata[num].time[index])-3,
	      unrealY(drabas.mdata[num].Datas[index] - 
                      drabas.mdata[num].conf_datas[index]) -3,
	      unrealX(drabas.mdata[num].time[index])+3,
	      unrealY(drabas.mdata[num].Datas[index] - 
                      drabas.mdata[num].conf_datas[index]) +3);
            XDrawLine(disp,drawon,gc, 
	      unrealX(drabas.mdata[num].time[index])-3,
	      unrealY(drabas.mdata[num].Datas[index] - 
                      drabas.mdata[num].conf_datas[index]) +3,
	      unrealX(drabas.mdata[num].time[index])+3,
	      unrealY(drabas.mdata[num].Datas[index] - 
                      drabas.mdata[num].conf_datas[index]) -3);
            XDrawLine(disp,drawon,gc, 
	      unrealX(drabas.mdata[num].time[index])-3,
	      unrealY(drabas.mdata[num].Datas[index] + 
                      drabas.mdata[num].conf_datas[index]) -3,
	      unrealX(drabas.mdata[num].time[index])+3,
	      unrealY(drabas.mdata[num].Datas[index] + 
                      drabas.mdata[num].conf_datas[index]) +3);
            XDrawLine(disp,drawon,gc, 
	      unrealX(drabas.mdata[num].time[index])-3,
	      unrealY(drabas.mdata[num].Datas[index] + 
                      drabas.mdata[num].conf_datas[index]) +3,
	      unrealX(drabas.mdata[num].time[index])+3,
	      unrealY(drabas.mdata[num].Datas[index] + 
                      drabas.mdata[num].conf_datas[index]) -3);
	  }

          if(index > 0)  /* alle Werte sind Mittelwerte und besitzen eine Varianz  */
          {  
            XDrawLine(disp,drawon,gc, 
	      unrealX(drabas.mdata[num].time[index - 1]),
	      unrealY(drabas.mdata[num].Datas[index-1] - 
                      drabas.mdata[num].conf_datas[index-1]) -1,
	      unrealX(drabas.mdata[num].time[index]),
	      unrealY(drabas.mdata[num].Datas[index] - 
                      drabas.mdata[num].conf_datas[index]) -1);


             XDrawLine(disp,drawon,gc, 
	      unrealX(drabas.mdata[num].time[index - 1]),
	      unrealY(drabas.mdata[num].Datas[index-1] + 
                      drabas.mdata[num].conf_datas[index-1]) -1,
	      unrealX(drabas.mdata[num].time[index]),
	      unrealY(drabas.mdata[num].Datas[index] + 
                      drabas.mdata[num].conf_datas[index]) -1);
	  }
      }

    }
}



    

