/*************************************************************************

filename:	random.c

purpose:	It implements the base random number generator 
		and a linear congruence generator     
		See R. Jain: "The Art of Computer Systems Performance
		Modeling, John Wiley 1992.
		
		It provides some help functions for the generation of 
		generally distributed random numbers

project director: Christoph Lindemann
		  Technical University of Berlin
		  Computer Science Department

author:	Ulrich Hertlein, Ch. Kelling

date:		June 1992, May 94


***************************************************************************/


#include	<stdio.h>
#include    <stdlib.h>
#include	<math.h>
#include	"sim_slave_types.h"

extern	int	initial_seed;
extern  double (*pmf_fctn[])();
extern transition *transitions;    /* importiert aus sim_slave.c */

void rand_init ()
{
  void srand48();
  
  srand48((long)initial_seed);
  
}

double rnd ()
{
  double drand48();

  return drand48();

}


double	rnd_uniform (a, b)

double	a, b;

{
  double	res;
  double	rnd();

  res = a + (b - a) * rnd (1);
  
  return (res);
}


double	rnd_exp (a)

double	a;

{
  double	b;
  b = rnd (2);
  
  return (-a * log(b));
}

/*************************************************************************

  function:     get_delay_from_file

  parameter:    filename,trans_id, kindOfSelection
  
  purpose:      determine the next delay for generel Transition with
                File-Interface
 
  return:       next delay

  author:       Henning Ziegler

  date:         February 1997

*************************************************************************/
double get_delay_from_file(filename,mark_dep_id, kindOfSelection)
char *filename;
int mark_dep_id;
int kindOfSelection;

{
 transition *dataFileTrans;
 FILE *dat;
 double nextRate;
 char buffer[255];
 int i;  
 int trans_id;
 
 trans_id=0;       /* geht alle Transitionen durch, um richtige zu finden */
 while (transitions[trans_id].mark_dep_id!=mark_dep_id) trans_id++;
 dataFileTrans=&transitions[trans_id];
 if (dataFileTrans==NULL)
 {
  printf("ERROR:transition[%d] not exist\n",trans_id);
  exit(0);
 }
 if (dataFileTrans->buffer==NULL)
 {
  if ((dat=fopen(filename,"r"))==NULL)
  {
   printf("ERROR:Needed File %s does not exist or unable to open\n",filename);
   exit(0);
  }
  do fgets(buffer,255,dat); while( buffer[0]=='#');  
                                            /* Kommentarzeilen oder Leerzeilen*/
  dataFileTrans->noEntries=atoi(buffer);    /* Zahl der Eintraege */
  dataFileTrans->lastEntry=0;               /* initialisieren */
                                            /* alloziieren */ 
  dataFileTrans->buffer=(double*)malloc(dataFileTrans->noEntries*sizeof(double)); 
  do fgets(buffer,255,dat); while( buffer[0]=='#');  
                                            /* Kommentarzeilen oder Leerzeilen*/
  dataFileTrans->buffer[0]=atof(buffer);    /* erster Wert */
  for(i=1;i<dataFileTrans->noEntries;i++)
  {
   fgets(buffer,255,dat);                
   dataFileTrans->buffer[i]=atof(buffer);   /* naechster Wert */
  }
  fclose(dat);
 }
#ifdef debug
 printf("noEntries=%d\tlastEntry=%d\n",dataFileTrans->noEntries,dataFileTrans->lastEntry);
#endif
 switch (kindOfSelection)
 {
   case SEQUENCE: /* von vorn nach hinten */
                  if(dataFileTrans->lastEntry==dataFileTrans->noEntries-1)
                   dataFileTrans->lastEntry=0; else dataFileTrans->lastEntry+=1;
   break;

   case SEQU_REV: /* von hinten nach vorn */
                  if(dataFileTrans->lastEntry==0) 
                   dataFileTrans->lastEntry==dataFileTrans->noEntries-1; 
                  else dataFileTrans->lastEntry-=1;
   break;
   case RAND_UNI: /* uniform */
                  dataFileTrans->lastEntry=floor((dataFileTrans->noEntries-1)*drand48()+0.5);
   break;
 }

 nextRate=dataFileTrans->buffer[dataFileTrans->lastEntry];
#ifdef debug
 printf("nextRate=%g\n",nextRate);
#endif

 return (nextRate);
}

  

/*************************************************************************

  function:     rnd_exp2

  parameter:    mean value a, random number r
  
  purpose:      generate a exponential distributed sample
 
  return:       --

  author:       Matthias Hoffmann

  date:         March 1995

*************************************************************************/
double	rnd_exp2 (a, r)
  double a; double r;
{
  return (-a * log(r));
}

/****************************************************************************/
/*									    */
/*	functions for random number transformation			    */
/*									    */
/****************************************************************************/

/********************   brac_root()  *****************************************/
/* Das eingegebe Intervall X1 - X2 wird eingeschraenkt auf den Bereich um der*/
/* Nullstelle in der gegeben Fnutkion. Das Bracketing hoert dann auf, wenn   */
/* ein Vorzeichenwechsel stattfindet. Das Intervall ist jetzt schmaler und   */
/* und die neuen X1- und X2-Werte werden zurueckgegeben.                     */
/*****************************************************************************/
void brac_root(trans_id, interval, x1, x2, y, f_offset)
int trans_id;	/* transiton id for function selection */
int interval;	/* number of the interval for function selection */
double *x1;
double *x2;
double  y;	/* given value y=f(x) */
int f_offset;   /* max. number of intervals, used to compute the offset for 
		  finding the according pmf-function */
{

  double       temp;
  double       dx;
  double       fp,fc;
  register int i;
  int 		dest;

  dest = (trans_id-1)*f_offset+interval-1;
  temp = *x1;
  dx = (*x2 - *x1)/MAX_TRIES_ROOT_FINDING;                 /* Genauigkeit */
  fp = (pmf_fctn[dest](*x1)) - y;  
              
  for(i= 0; i<MAX_TRIES_ROOT_FINDING; i++){
     fc = (pmf_fctn[dest](*x1+dx)) - y;
     if(fc * fp < 0.0 ){    
        *x2 = *x1 + dx;
        return;
   
      }
      *x1 = *x1 + dx;
   }
  *x1 = temp;
}
/*****************************************************************************/


/************************  root_bisection() **********************************/
/* Nachedem das Intervall eingeschraenkt wurde, wird an Hand der root        */
/* bisection Methode die Nullstelle der gegebenen Funktion an dem gegebenem  */
/* Y-Wert gefunden. Der X-Wert der zu dieser Nullstelle gefuehrt hat wird    */
/* dann zurueck gegeben.                                                     */
/*****************************************************************************/
double root_bisection(trans_id, interval, x1, x2, y, f_offset)
int trans_id;	/* transiton id for function selection */
int interval;	/* number of the interval for function selection */
double x1;
double x2;
double y;
int f_offset;  /* max. number of intervals, used to compute the offset for 
		  finding the according pmf-function */
{

  register int  i;
  int		dest;
  double        f,fmid,xmid,rtb,dx;
  
  dest = (trans_id-1)*f_offset+interval-1;
  f = (pmf_fctn[dest](x1)) - y;
  fmid = (pmf_fctn[dest](x2)) - y;
  rtb = f < 0.0 ? (dx =x2-x1, x1) : (dx = x1-x2, x2);
  for ( i=0; i<MAX_TRIES_ROOT_FINDING; i++){
     xmid = rtb + (dx *= 0.5);
     fmid = (pmf_fctn[dest](xmid)) - y;
     if (fmid <= 0.0) rtb = xmid;
     if (fabs(dx) < ACCURACY || fmid == 0.0) return rtb; 
   }
  printf("ERROR: Too many bisection operations for transition %d in interval %d.\n", trans_id, interval);
  exit (-1);
}






