/******************************************************************************
filename :simu_kernel.c

purpose  :simulation engine for DSPN-models in TimeNET-descriptionformat
  
author   :    Matthias Hoffmann

date     :November 1994

  exported functions
bool    count_firings
void    sim_net
void    save_simstate
void    load_simstate
void    record_all_cvs

  local functions:
        void    test_cvs_trans
	int     test_enabled
        double  get_delay
	double  get_exp_sample
        double  get_same_sample
        void    no_changes
        void    replace_imm_md
        void    replace_tim_RS
        void    change_tim_md
        void    add_RA_MD
        int     init_sim
        void    calculate_delta_marking
        void    change_marking
        void    check_affected
        int     update_affected_imm
        void    update_affected_tim
        void    reset_thresholds				     HeZi 12/95

******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "sim_slave_types.h"
#include "../Master/stat_utils.h"

/* global variables used by the modul simulation */

extern double       current_time, transient_time;
extern int          StatSema;
extern int          no_md_arcs;
extern int          no_places;
extern int          no_trans;
extern int          no_mark_par;
extern int          no_del_par;
extern int          no_md_exp;
extern int          no_md_det;
extern int          no_weights;
extern int          no_ena_fcts;
extern place        *places;
extern transition   *transitions;
extern char       *er_malloc();
extern int          min_firings;

static int          *delta_marking;
static int          *save_marking[NO_SAVESTATES];
static int          *save_no_in_list[NO_SAVESTATES];
/************************************/
/* global variables used by RESTART */
/************************************/

#ifdef RST

/*#define DEBUG_RST	/* activate them for (min.) debuging (printf) for RESTART*/
/*#define DEBUG_RST_1     /* zeigt jede Beendigung eines set_of_retrials an */
/*#define DEBUG_RST_2     /* zeigt jede Ueber-und Unterschreitung von Schwellen an */
int optimiere=1;         /* 0: nicht optimieren             1: optimieren */

extern void  my_exit();
extern int   direction;
extern int   event;
extern int   border;
extern int   no_thresholds;
extern int   no_opt_thresholds;     
extern int   *t_value;
extern int   *retrials;
extern int   Satz;
extern int   Satz_changed;
extern double *prob_opt;
extern double Prob_rare_event;
extern char  measure_place[100];
extern long  rare_event_count;
extern double *first_k_trial_time;
extern double *current_k_time;
extern int   ns;

int    *optimized_threshold;
double *first_trial_time_t;
double *current_time_t;

int no_more_test=0;
int ra_sample=0;


#endif

/* end of global variables */

/*************************************************************************

  function:     record_all_cvs

  parameter:    measure id
  
  purpose:      record all cvs related to the measure given in the parameter
 
  return:       --

  author:       Christian Kelling
                Technical University of Berlin

  date:         March 1994

*************************************************************************/

#ifdef CVR
void record_all_cvs (i_measure)
  int i_measure;
{
  void record_cv();
  int i, j;
  cv_list_elem *p_cv_list;

  for (i=0; i<no_trans; i++)
    {
      for (j=0; j<transitions[i].no_cv_meas ; j++)
        {
          p_cv_list = transitions[i].cv_list[j];
          if (( i_measure == p_cv_list->meas_number) &&
                transitions[i].no_in_list>0 && 
              ( p_cv_list->recorded == false ))
            {
	      record_cv (i_measure, p_cv_list->cv_number,
			 (get_CVsample(i) - transitions[i].delay));
              p_cv_list->recorded = true;
            }
        }
    }
}
#endif


/*************************************************************************

  function:     count_firings

  parameter:    none
  
  purpose:      count the number of firings of each transition in order to 
                ensure that each transition has MIN_FIRINGS times fired 
                This should lead to reliable simulation results!!!
 
  author:       Christian Kelling
		Technical University of Berlin

  date:         November 1994

*************************************************************************/
bool count_firings()
{
  int i;
  bool done;

  done = true;
  for (i=0; i<no_trans; i++)
    if ((transitions[i].no_firings < min_firings) && 
        (transitions[i].type != IMM))
      { 
/*	printf ("ERROR: Transition %s has not fired %d times\n", 
                transitions[i].name, MIN_FIRINGS);*/
        done = false;
        return done;
      }
  return done;
}


/*************************************************************************

function:test_cvs_trans

parameter:-
  
purpose:check all cvs-trans whether their fire policy is RE.
                EXP-trans are automatic set to RE 
 
return:number of enabled transitions
considering the server type

author:Matthias Hoffmann

date:October 1994

*************************************************************************/
#ifdef CVR
void test_cvs_trans()
{
  int i;

  for( i=0 ; i<no_trans ; i++ )
    if( transitions[i].no_cv_meas > 0 )
      if( transitions[i].fire_policy != RE )
        if( transitions[i].type == EXP )
        { transitions[i].fire_policy = RE;
		       printf( "ERROR: EXP-trans %s used as control variate\nrace policy changed to RE\n", transitions[i].name );
        }
        else
	{ printf( "ERROR: non-EXP-trans %s used as control variate\nbut hasn't race policy RE\n", transitions[i].name );
          exit( FAIL );
        }
}
#endif


/*************************************************************************

function:test_enabled

parameter:transition number
  
purpose:check whether the transition is enabled in the current marking 
 
return:number of enabled transitions
considering the server type

author:Matthias Hoffmann

date:October 1994

*************************************************************************/
static int test_enabled(tra_nr)
     int tra_nr;
{
  int md_arc_mult(), md_ena_function();
  int j, no_enabled, arc_ena, arc_mult;

  for( j=0; j<transitions[tra_nr].no_inh ; j++ )
  { if( transitions[tra_nr].inh_from_place[j].id )  /* m-d arc mult.*/
      arc_mult =  md_arc_mult(transitions[tra_nr].inh_from_place[j].id);
    else
      arc_mult = transitions[tra_nr].inh_from_place[j].multiplicity;
/*    if((arc_mult == 0) && (transitions[tra_nr].serv_type==INF_SERV)) 
      return 0;*/
    if(( places[transitions[tra_nr].inh_from_place[j].element].token 
         >= arc_mult) && (arc_mult != 0)) 
      return 0;
  }
  no_enabled = MAX_TOKEN+1;   /* noch nicht getestet */ 
  for( j=0 ; j<transitions[tra_nr].no_inputs ; j++ )
  { if( transitions[tra_nr].in_from_place[j].id ) /* m-d arc mult.*/
      arc_mult =  md_arc_mult(transitions[tra_nr].in_from_place[j].id);
    else
      arc_mult = transitions[tra_nr].in_from_place[j].multiplicity;
    if(arc_mult == 0)
      if( transitions[tra_nr].serv_type==INF_SERV ) 
        arc_ena = MAX_TOKEN+1;  /* INF_SERV with input: <md>-arc_mult = 0 */
      else 
        arc_ena = 1;
    else
      arc_ena = places[transitions[tra_nr].in_from_place[j].element].token
                / arc_mult;
    if( arc_ena < no_enabled )
      no_enabled = arc_ena;
  }
  /* check the transition wether it will be enabled by an 
     enabling function */
  if( no_enabled )
    if( transitions[tra_nr].ena_id && 
		   (!md_ena_function(transitions[tra_nr].ena_id)))
      no_enabled = 0;
  if( (transitions[tra_nr].serv_type == SING_SERV) && (no_enabled>0))
    return 1;
  if( no_enabled > MAX_TOKEN )  /*  source transition is IMM_SERV */
    return 0;
  return no_enabled; 
}


/*************************************************************************

function:get_delay

parameter:transition number
  
purpose:returns a delay according to the transition-type
 
return:--

author:Matthias Hoffmann

date:March 1995

*************************************************************************/
static double get_delay( tra_nr )
  int tra_nr;
{
  double md_rate();
  double delay;

  if( transitions[tra_nr].mark_dep_id )    /* delay is m-d */
    transitions[tra_nr].delay = md_rate( transitions[tra_nr].mark_dep_id );
  delay = transitions[tra_nr].delay;
  return( delay );
}


/*************************************************************************

function:get_exp_sample

parameter:transition number
  
purpose:returns a delay according to the transition-type

return:--

author:Matthias Hoffmann

date:March 1994

*************************************************************************/
static double get_exp_sample( tra_nr )
  int tra_nr;
{
  double rnd_exp2();

  return( rnd_exp2( transitions[tra_nr].delay, 
                    transitions[tra_nr].samp_rnd ));
}


/*************************************************************************

function:get_same_sample

parameter:transition number
  
purpose:returns the delay
 
return:--

author:Matthias Hoffmann

date:March 1994

*************************************************************************/
static double get_same_sample( tra_nr )
  int tra_nr;
{
  return( transitions[tra_nr].delay );
}


/*************************************************************************

function:no changes

parameter:transition number
  
purpose:dummy function
 
return:--

author:Matthias Hoffmann

date:March 1994

*************************************************************************/
static void no_changes( tra_nr )
  int tra_nr;
{;}


/*************************************************************************

function:replace_imm_md

parameter:transition number
  
purpose:overwrites all ttf-elements
 
return:--

author:Matthias Hoffmann

date:March 1994

*************************************************************************/
static void replace_imm_md( tra_nr )
  int tra_nr;
{
  double sample;
  int i;
  
  transitions[tra_nr].delay = get_delay(tra_nr);
  sample = (*transitions[tra_nr].get_sample)(tra_nr);
  start_changing (tra_nr);

  for( i=0 ; i<transitions[tra_nr].no_in_list ; i++ )
      replace_imm_ttf( transitions[tra_nr].prio,
				     transitions[tra_nr].samp_rnd, sample );
}


/*************************************************************************

function:replace_tim_RS

parameter:transition number
  
purpose:overwrites all ttf_elements

return:--

author:Matthias Hoffmann

date:March 1994

*************************************************************************/
static void replace_tim_RS( tra_nr )
  int tra_nr;
{ 
  double sample;
  int i;
  double rnd();

  start_changing( tra_nr );
  for( i=0 ; i<transitions[tra_nr].no_in_list ; i++ )
  { transitions[tra_nr].samp_rnd = rnd(1);
    transitions[tra_nr].delay = get_delay(tra_nr);
    sample = (*transitions[tra_nr].get_sample)(tra_nr);
    replace_RS_ttf( transitions[tra_nr].prio,
				     transitions[tra_nr].samp_rnd, sample );
  } 
}


/*************************************************************************

function:change_tim_md

parameter:transition number
  
purpose:changes all ttf-elements
 
return:--

author:Matthias Hoffmann

date:March 1994

*************************************************************************/
static void change_tim_md( tra_nr )
  int tra_nr;
{
  double sample;
  int i;

  start_changing( tra_nr );
  for( i=0 ; i<transitions[tra_nr].no_in_list ; i++ )
  { transitions[tra_nr].samp_rnd = get_samp_rnd(tra_nr);
    transitions[tra_nr].delay = get_delay(tra_nr);
    sample = (*transitions[tra_nr].get_sample)(tra_nr);
    change_tim_ttf( transitions[tra_nr].prio,
			           transitions[tra_nr].samp_rnd, sample );
  } 
}


/*************************************************************************

function:add_RA_md

parameter:--
  
  purpose:      enables transition and changes rem_delay or add new sample
 
return:like in all add-functions

author:Matthias Hoffmann

date:April 1995

*************************************************************************/
void add_RA_md( tra_nr, prio, samp_rnd, sample )
  int tra_nr; int prio; double samp_rnd; double sample;
{
  add_rag_ttf( tra_nr, prio, samp_rnd, sample );
  transitions[tra_nr].no_in_list++;
  change_tim_md( tra_nr );
  transitions[tra_nr].no_in_list--;
}

    
/*************************************************************************

function:init_sim

parameter:--
  
  purpose:      initialises the simulator and the ttf-list
 
return:number of enabled immediate transitions in the
                initial marking

author:Matthias Hoffmann

date:April 1995

*************************************************************************/
static int init_sim()
{
  double get_exp_sample();
  double get_same_sample();
  void   no_changes();
  void   replace_imm_md();
  void   replace_tim_RE();
  void   change_tim_md();
  void   del_ttf();
  void   off_rag_ttf();

  double rnd();
  /*char *er_malloc();*/
  int i, j, no_enabled, enabled_imm;
  double sample;
  for( i=0 ; i<no_trans ; i++ )
  { if( transitions[i].type == IMM )
    { transitions[i].get_sample = get_same_sample;
      transitions[i].add_ttf    = add_imm_ttf;
      transitions[i].del_ttf    = del_ttf;
      if( transitions[i].mark_dep_id )
	transitions[i].change_ttf = replace_imm_md;
      else
        transitions[i].change_ttf = no_changes;
    }
    else   /* timed */
    { if( transitions[i].fire_policy == RA )
      { if( transitions[i].mark_dep_id )
          transitions[i].add_ttf = add_RA_md;
        else
          transitions[i].add_ttf = add_rag_ttf;
        transitions[i].del_ttf = off_rag_ttf;
      }
      else
      { transitions[i].add_ttf = add_tim_ttf;        
        transitions[i].del_ttf = del_ttf;
      }
      if( transitions[i].type == EXP )
        transitions[i].get_sample = get_exp_sample;
      else
        transitions[i].get_sample = get_same_sample;
      if( transitions[i].fire_policy == RS )
	transitions[i].change_ttf = replace_tim_RS;
      else
      { if( transitions[i].mark_dep_id )
          transitions[i].change_ttf = change_tim_md;
        else
          transitions[i].change_ttf = no_changes;
      }
    }
  }
  delta_marking = (int *)er_malloc( no_places*sizeof(int) );
  for( i=0 ; i<NO_SAVESTATES ; i++ )
    { save_marking[i] = (int *)er_malloc( no_places*sizeof(int) );
      save_no_in_list[i] = (int *)er_malloc( no_trans*sizeof(int) );
  }
  for( i=0 ; i<no_places ; i++ )
    *(delta_marking+i) = 0;
  init_ttf();
  enabled_imm = 0;
  for (i=0; i<no_trans; i++)
  { transitions[i].no_in_list = 0;
    no_enabled = test_enabled(i);
    for( j = 0 ; j < no_enabled ; j++ )
    { if( transitions[i].type == IMM )
        enabled_imm++;
      else
        transitions[i].samp_rnd = rnd(0);
      transitions[i].delay = get_delay( i );
      sample = (*transitions[i].get_sample)(i);
      (*transitions[i].add_ttf)(i, transitions[i].prio,
                                transitions[i].samp_rnd, sample );
      transitions[i].no_in_list++;
    }
    transitions[i].affected = false;
  }
  return( enabled_imm );
}


/*************************************************************************

  function:     free_simu_kernel_pointer

  parameter:    -
  
  purpose:      free allocated memory for global pointers

  author:       Ralf Putz

  date:         Dezember 1995

*************************************************************************/
void free_simu_kernel_pointer()
{
  int i;

  free( delta_marking );
  for( i = 0; i < NO_SAVESTATES; i++)
    {
      free( save_marking[i] );
      free( save_no_in_list[i] );
    }
}


/*************************************************************************

function:calculate_delta_marking

parameter:transition number
  
purpose:calculates the changes in the marking caused by
                the firing of the transition firtra
 
return:--

author:Matthias Hoffmann

date:October 1994

*************************************************************************/
static void calculate_delta_marking( firtra )
  int firtra;
{
  int md_arc_mult();
  int i;
  
  for( i=0 ; i<transitions[firtra].no_inputs ; i++ )
    if( transitions[firtra].in_from_place[i].id )  /* md-mult */
      *(delta_marking+transitions[firtra].in_from_place[i].element) -=
	md_arc_mult( transitions[firtra].in_from_place[i].id );
    else
      *(delta_marking+transitions[firtra].in_from_place[i].element) -=
	transitions[firtra].in_from_place[i].multiplicity;
  for( i=0 ; i<transitions[firtra].no_outputs ; i++ )
    if( transitions[firtra].out_to_place[i].id )  /* md-mult */
      *(delta_marking+transitions[firtra].out_to_place[i].element) +=
	md_arc_mult( transitions[firtra].out_to_place[i].id );
    else
      *(delta_marking+transitions[firtra].out_to_place[i].element) +=
	transitions[firtra].out_to_place[i].multiplicity;
}


/*************************************************************************

function:change_marking

parameter:--
  
purpose:initiates the recording of the measures, which are
                affected by the changed_marking and  changes the
                marking in the net
 
return:--

author:Matthias Hoffmann

date:October 1994

*************************************************************************/
static void change_marking()
{
  int i, j, meas_id;
  
  for( i=0 ; i<no_places ; i++ )
    { if( *(delta_marking+i) != 0 )
      { if( places[i].no_measures != NO_MEASURE )
        { for( j=0 ; j<places[i].no_measures ; j++ )
        { meas_id = places[i].meas_list[j];
#ifdef CVR
   record_all_cvs( meas_id );
#endif
#ifndef STR
          select_measure( meas_id, places[i].token );
#endif
        }
      }
      places[i].token += delta_marking[i];
      delta_marking[i] = 0;
    }
  }
}


/*************************************************************************

function:check_affected

parameter:transition number
  
purpose:marks all transitions, which are (posssibly) affected
                by the firing of the transition firtra
 
return:--

author:Matthias Hoffmann

date:October 1994

*************************************************************************/
static void check_affected( firtra )
  int firtra;
{
  int i;

  for( i=0 ; i<transitions[firtra].no_aff ; i++ )
    transitions[transitions[firtra].affected_trans[i]].affected = true;
}


/*************************************************************************

function:update_affected_imm

parameter:--
  
purpose:updates the number of ttf in the ttf-list for all affected
                immediate transition according to the actual marking

return:number of added or deleted ttf-list-elements

author:Matthias Hoffmann

date:March 1995

*************************************************************************/
static int update_affected_imm()
{
  int i, delta_imm, no_enabled;
  double sample;

  delta_imm = 0;
  for( i=0 ; i<no_trans ; i++ ) 
    {
      /* ck RS*/
      if((transitions[i].type != IMM ) && (transitions[i].fire_policy == RS ) 
	 && (transitions[i].affected == true) && (transitions[i].no_in_list > 0))
	(*transitions[i].change_ttf)(i);
      /*ck RS */
      if((transitions[i].type == IMM) && (transitions[i].affected == true))
	{ 
	  no_enabled = test_enabled( i );
	  while( no_enabled < transitions[i].no_in_list )
	    { (*transitions[i].del_ttf)( i, transitions[i].no_in_list, &sample );
	    transitions[i].no_in_list--;
	    delta_imm--;
	    }
	  if( transitions[i].no_in_list > 0 )
	    (*transitions[i].change_ttf)(i);
	  while( no_enabled > transitions[i].no_in_list )  /* add some */
	    { 
	      transitions[i].delay = get_delay( i );
	      sample = (*transitions[i].get_sample)(i);
	      (*transitions[i].add_ttf)(i, transitions[i].prio,
					transitions[i].samp_rnd, sample );
	      transitions[i].no_in_list++;
	      delta_imm++;
	    }
	  transitions[i].affected = false;
	}
    }
  return( delta_imm );
}



/*************************************************************************

function:update_affected_timed

parameter:--
  
purpose:updates the number of ttf 
                for all affected timed transition according to the actual
                marking
 
return:--

author:Matthias Hoffmann

date:March 1995

*************************************************************************/
static void update_affected_timed()
{
  int i, no_enabled;
  double sample;
  double rnd();
#ifdef CVR
  int j;
  cv_list_elem *p_cv_list;
#endif

  for( i=0 ; i<no_trans ; i++ ) 
    if((transitions[i].type != IMM) && (transitions[i].affected == true))
    { no_enabled = test_enabled( i );
#ifdef CVR
      if((transitions[i].no_cv_meas > 0) && (transitions[i].no_in_list == 0)
          && (no_enabled == 1))
      transitions[i].just_enabled = true;
#endif
      while( no_enabled < transitions[i].no_in_list )
      { (*transitions[i].del_ttf)( i, transitions[i].no_in_list, &sample );
        transitions[i].no_in_list--;
      }
      if( transitions[i].no_in_list > 0 )
        (*transitions[i].change_ttf)(i);
      while( no_enabled > transitions[i].no_in_list )  /* add some */
      { transitions[i].samp_rnd = rnd(0);
		   transitions[i].delay = get_delay( i );
        sample = (*transitions[i].get_sample)(i);
        (transitions[i].add_ttf)(i, transitions[i].prio,
                                  transitions[i].samp_rnd, sample );
        transitions[i].no_in_list++;
#ifdef CVR
        if(transitions[i].just_enabled == true)
	  for( j=0 ; j < transitions[i].no_cv_meas ; j++ )
	    { p_cv_list = transitions[i].cv_list[j];
            p_cv_list->recorded = false;
          }
#endif
      }
    transitions[i].affected = false;
    }
}

/* end of global variables */

/*************************************************************************

function:save_simstate

parameter:imm_in_list
  
purpose:save the current simulation state, update imm_in_list

author:Matthias Hoffmann

date:November 1994

*************************************************************************/
#ifdef RST
void save_simstate( imm_in_list, nr )
  int *imm_in_list; int nr;
{
  int i;

  *imm_in_list += update_affected_imm();
  update_affected_timed();
  for( i=0 ; i<no_places ; i++ )
    save_marking[nr][i] = places[i].token;
  for( i=0 ; i<no_trans ; i++ )
    save_no_in_list[nr][i] = transitions[i].no_in_list;
  save_list(nr);
}
#endif

/*************************************************************************

function:load_simstate

parameter:imm_in_list

purpose:load a saved simulation state, update imm_in_list

author:Matthias Hoffmann

date:November 1994

*************************************************************************/
#ifdef RST
void load_simstate( imm_in_list, nr )
  int *imm_in_list; int nr;
{
  int i;

  for( i=0 ; i<no_places ; i++ )
     places[i].token = save_marking[nr][i];
  *imm_in_list = 0;
  for( i=0 ; i<no_trans ; i++ )
  { transitions[i].no_in_list = save_no_in_list[nr][i];
    if( transitions[i].type == IMM )
      *imm_in_list += transitions[i].no_in_list;
  }
  load_list(nr);
}
#endif


/*************************************************************************
function:           reset_thresholds
global Variable: 
parameter:        mode, rare_event, P_A
return:               1: resetting has finished;  2: retrials has been optimize
author:              Henning Ziegler
date:                  1/96
*************************************************************************/
#ifdef RST
int reset_thresholds(mode, rare_event_t, P_A, pnot)
int mode; int rare_event_t; double P_A; int pnot;
{
 double Prob[MAX_THRESHOLDS+1];
 double new_Prob[MAX_THRESHOLDS+1];
 int i,j,v,t=0;
 int reset_no_t=0; 
 int opt_threshold_nr,missing_t,diff_t;
 double m,n;
 double diff_Prob;
 int new_t_value[MAX_THRESHOLDS+1];

 /* determine optimal number of thresholds */
 if (P_A!=0.0) opt_threshold_nr=(int)(-0.5*log(P_A)); 
  else opt_threshold_nr=no_thresholds;

 for(j=1;j<=no_thresholds;j++) new_t_value[j]=t_value[j]; 
 do t++; while(optimized_threshold[t]==1);
 new_t_value[0]=border;
 Prob[0]=1.0;
 new_Prob[0]=1.0;
 for (j=1;j<=pnot;j++) Prob[j]=new_Prob[j]=prob_opt[j];
 for (j=1;j<=no_thresholds;j++) first_trial_time_t[j]+=first_trial_time_t[j-1];
 
 /* determine the probabilities of the thresholds */
 for (j=pnot+1;j<=no_thresholds;j++) 
 { 
   Prob[j]=current_time_t[j]/first_trial_time_t[j];
   for (i=1;i<j;i++) Prob[j]=Prob[j]/(double)retrials[i];
   new_Prob[j]=Prob[j];
#ifdef DEBUG_RST
   printf("ERROR:current_time_t[%d]=%g\tfirst_trial_time_t[%d]=%g\tP[T%d=%d]=%g\n"
          ,j,current_time_t[j],j,first_trial_time_t[j],j,t_value[j],Prob[j]);
#endif
 }

 /* determine the retrials */
 for(i=1;i<no_thresholds;i++) 
 { 
   retrials[i]=((int)(1.0/sqrt(Prob[i+1]/Prob[i-1])+0.5) < 1) ? 1 : (int)(1.0/sqrt(Prob[i+1]/Prob[i-1])+0.5);
#ifdef DEBUG_RST
   printf("ERROR:ret[%d]=%d\t",i,retrials[i]);
#endif
 }
 retrials[no_thresholds]=((int)(1.0/sqrt(P_A/Prob[no_thresholds-1])+0.5) < 1) ? 1 :(int)(1.0/sqrt(P_A/Prob[no_thresholds-1])+0.5);
#ifdef DEBUG_RST
 printf("ERROR:ret[%d]=%d\n",no_thresholds,retrials[no_thresholds]);
#endif
 /* mode=1: threre are non optimal thresholds => reset them */
 if (mode==1) 
 {
   j=t;
   while(j<=no_thresholds)
   {
     /* determine the new position of threshold j */
     if (!optimized_threshold[j])
     { 
       diff_Prob=-1.1;
       for (i=t;i<=no_thresholds;++i) 
         if (fabs(Prob[i]/new_Prob[j-1]-exp((double)(-2)))<fabs(diff_Prob))
         { 
           diff_Prob=Prob[i]/new_Prob[j-1]-exp((double)(-2));
           v=i;
         }  
       if (diff_Prob>0.0 && v < no_thresholds) ++v;
       m=(double)(t_value[v]-t_value[v-1])/((Prob[v]-Prob[v-1])/new_Prob[j-1]);
       n=(double)(t_value[v-1])-m*Prob[v-1]/new_Prob[j-1];
       new_t_value[j]=(int)(m*exp((double)(-2))+n+0.5);
       new_Prob[j]= new_Prob[j-1]*exp((double)(-2));
#ifdef DEBUG_RST
       printf("ERROR:m=%g\t n=%g\t P[new_t_value[%d]=%d]=%g\n"
                  ,m,n,j,new_t_value[j],new_Prob[j]);
#endif
       if (new_t_value[j]==new_t_value[j-1]) new_t_value[j]+=direction;
     }

     /* exit, if threshold behind the threshold of rare event */
     if (((direction==1) && ((new_t_value[j]<=new_t_value[j-1])
         ||(new_t_value[j]>=rare_event_t)))|| 
         ((direction==-1) && ((new_t_value[j]>=new_t_value[j-1])
         ||(new_t_value[j]<=rare_event_t)))) 
     {
       no_thresholds=j-1;
       reset_no_t=1;
     }
     ++j;
   }

   /* more thresholds necessary */
   if (!reset_no_t && (no_thresholds < opt_threshold_nr)
       && (new_t_value[no_thresholds]<(rare_event_t-1)))
   {
     missing_t=opt_threshold_nr-no_thresholds;
     diff_t=abs(rare_event_t-direction-new_t_value[no_thresholds])/missing_t;
     if (diff_t>0)
       do 
       {   
         no_thresholds+=1;
         new_t_value[no_thresholds]=new_t_value[no_thresholds-1]+diff_t;
         retrials[no_thresholds]=7;
       } while(no_thresholds<opt_threshold_nr); 
     reset_no_t=1; 
   }

   /* check: threshold optimal */
   for (j=t;((j<=no_thresholds) && optimized_threshold[j-1]);j++)
     if (new_t_value[j]==t_value[j] && !optimized_threshold[j])
       optimized_threshold[j]=1;

   /* store new_value  of the thresholds in t_value[ ] */
   for (j=t;j<=no_thresholds;j++) t_value[j]=new_t_value[j];

   /* determine the thresholds which are optimize now */
   for (j=t;optimized_threshold[j];j++);
   no_opt_thresholds=j-1;

   for (j=0;j<=no_thresholds;j++)
   {
     first_trial_time_t[j]=0.0;
     current_time_t[j]=0.0;
   }
   if (no_opt_thresholds==no_thresholds) return (1);
   else return(0);
 }
 return (2);
}


/*************************************************************************
function:           sum_ftt
global Variable: first_trial_time_t
parameter:        non
purpose:            compute the first_trial_time for threshhold no
return:               result (just those first_trial_time)
author:              Henning Ziegler
date:                 1/96
*************************************************************************/
double sum_ftt(no)
int no;
{ 
  int z;
  double result=0;

  for (z=0; z<=no; z++) result+=first_trial_time_t[z]; 
  return result;
}


/*************************************************************************
function:           Prob_opt_OK
global Variable: current_time_t ,  prob_opt ,  retrials 
parameter:        T , *p_o
purpose:            compare the Probability for threshhold T with the
                         value from the pre-simulation
return:               OHKAY (=1, all estimated Prob. nearby the values)  
author:              Henning Ziegler
date:                 1/96
*************************************************************************/
int Prob_opt_t_OK(T, p_o)
int T; double *p_o;
{
  int y, p_r=1, OHKAY=1;
  double Probi;
  double rel_error;

  for (y=1;(y<=T) && OHKAY ;y++)
  {
    Probi=current_time_t[y]/sum_ftt(y)/p_r;
    p_r=p_r*retrials[y];
    rel_error=100.0*fabs((Probi-prob_opt[y])/prob_opt[y]);
    if (rel_error <= 20.0) OHKAY=OHKAY*1;
     else OHKAY=0;
#ifdef DEBUG_RST
    printf("ERROR:Probi%d=%g\tprob_opt[%d]=%g\trel. Fehler=%g Prozent\tOHKAY=%d\n"
           ,y,Probi,y,prob_opt[y],rel_error,OHKAY);
#endif
  }
  return(OHKAY);
}



/*************************************************************************
function:             new_start
global Variable: 
parameter:          t, *sor
purpose:           
                        
return:              
author:               Henning Ziegler
date:                  4/96
*************************************************************************/
void new_start(t, sor)
int t; int *sor;
{
  int i;
  int diff_t;

  if ((t-no_opt_thresholds)!=0)
  {
    diff_t=abs((event-t_value[no_opt_thresholds]-direction)/(t-no_opt_thresholds));
    if (diff_t!=0)
    {
      for (i=no_opt_thresholds+1;i<=t;i++)
      {
        t_value[i]=t_value[i-1]+diff_t*direction;
        retrials[i]=7;
        if (t>i) diff_t=abs((event-t_value[i]-direction)/(t-i));
      }
      no_thresholds=t;
    }
    else /* no more thresholds available => increase the retrials*/
    {
      for (i=no_opt_thresholds+1;i<=no_thresholds;i++) retrials[i]*=2;
      if (ra_sample) no_more_test=1;
    }
    current_time=0.0;
    for(i=0;i<=MAX_THRESHOLDS;i++)
    {
      first_trial_time_t[i]=0.0;
      current_time_t[i]=0.0;
    }
  }
  else /* do nothing: Prob_rare_event too inaccurate */;
  *sor=0;
#ifdef DEBUG_RST
            printf("ERROR: Sim. has restarted with no_thresholds=%d\tretrials[%d]=%d\n"
                   ,no_thresholds,no_thresholds,retrials[no_thresholds]);
#endif
}
#endif

/*************************************************************************

  function:     d_greater

  parameter:    two doubles x and y
  
  purpose:      call x only greater as y if the difference is at least 1e-10

  author:       Ralf Putz

  date:         Dezember 1995

*************************************************************************/
int d_greater(x, y)
double x, y;
{
  if( (x - y) > 1e-10 )
    return true;
  else 
    return false;
}

/*************************************************************************

function:sim_net

parameter:--

purpose:perform the main simulation loop until the stop condition
becomes true (StatSema == 0)

author:Matthias Hoffmann

date:November 1994 update: Dezember 1995 (RESTART)

*************************************************************************/
void sim_net (
#ifdef STR
	      no_intervals
#endif
			   )
#ifdef STR
  int no_intervals;
#endif
{

  int  firtra, imm_in_list;
  double sample, time_in_list, passed_time;

#ifdef STR
  double interval_length;
#endif
#ifdef RST
  long MAX_OCCUR_OF_RARE_EVENT=RESTART_BATCH_SIZE_DEF*5;
  int MAX_SET_OF_RETRIALS=5000;
  int *retrial_count;
  int i;
  int place_no;
  int state=0,
       set_of_retrials=0,
       alles_optimiert=0,
       pre_no_opt_thresholds=no_opt_thresholds;

  optimized_threshold=(int *)er_malloc((MAX_THRESHOLDS+1)*sizeof(int));
  first_trial_time_t=(double *)er_malloc((MAX_THRESHOLDS+1)*sizeof(double));
  current_time_t=(double *)er_malloc((MAX_THRESHOLDS+1)*sizeof(double));
  retrial_count=(int *)er_malloc((MAX_THRESHOLDS+1)*sizeof(int));

  /*determine the optimized thresholds */
  for (i=0; i<=no_opt_thresholds; ++i) optimized_threshold[i]=1;
  for (i=no_opt_thresholds+1; i<=no_thresholds; ++i) optimized_threshold[i]=0;
  
  /* get the number of measure_place */
  for (i=0; i<no_places; ++i)
    if (!strcmp(measure_place,places[i].name))
      place_no=i;
  
  /* jump in the right starting interval */
  for (i=1;i<=no_thresholds;++i)
    if (((places[place_no].token>=t_value[i])&&(direction==1))||
        ((places[place_no].token<=t_value[i])&&(direction==-1)))
      state=i;

#endif

#ifdef CVR
  test_cvs_trans();
#endif
  imm_in_list = init_sim();
/* now start the simulation loop */
  current_time = 0.0;
#ifdef STR
  Measure_Update();
  interval_length = transient_time / no_intervals;
   while( !d_greater(current_time, transient_time ))
#else
  while( StatSema )
#endif
  { 
    firtra = trans_to_fire( &sample, &time_in_list, &passed_time );
    transitions[firtra].no_firings++;
    current_time += passed_time;
#ifdef RST
    current_k_time[Satz] += passed_time;
    for(i=state;retrial_count[i]==1;--i) current_time_t[i]+=passed_time;

    i=state;
    while (i>=0)
    {
      if (i==0)
      { 
        first_k_trial_time[Satz]+=passed_time;          /* for P(rare_event) */
        first_trial_time_t[state]+=passed_time;         /* for P(t_value[..])*/
      }
      else
       if ((retrial_count[i]!=1)&&(retrial_count[i]!=0)) break;
      --i;
    }
#endif

#ifdef STR
    while( d_greater(current_time, interval_length)  && 
           d_greater(transient_time, interval_length)  )
        {
	  Measure_Update();
	  interval_length += transient_time / no_intervals;
         }
    if( !d_greater(current_time, transient_time))
    {
#endif
      calculate_delta_marking( firtra );
      change_marking();
      check_affected( firtra );
      del_fir_tra();
      transitions[firtra].no_in_list--;
      if( transitions[firtra].type == IMM )
        imm_in_list--;
      imm_in_list += update_affected_imm();
      if( imm_in_list == 0 )
        update_affected_timed();
#ifdef STR
    }
#endif
#ifdef RST
    if (state<no_thresholds)
      if (((places[place_no].token>=t_value[state+1])&&(direction==1))||
          ((places[place_no].token<=t_value[state+1])&&(direction==-1)))
        if ((((places[place_no].token>=t_value[state+2])&&(direction==1))||
             ((places[place_no].token<=t_value[state+2])&&(direction==-1)))&&
             state!=(no_thresholds-1))
        {
          fprintf(stderr,"error: more then one threshold passed \n");
          my_exit(1);
        }
       else
        {
          ++state;
#ifdef DEBUG_RST_2
	  printf("ERROR:%d.Schwelle(%d Marken) ueberschritten\n"
                 ,state,t_value[state]);
#endif
          retrial_count[state]=1;
          save_simstate( &imm_in_list, state );
         }
    if (state>0)
      if (((places[place_no].token<t_value[state])&&(direction==1))||
          ((places[place_no].token>t_value[state])&&(direction==-1)))
        if ((retrial_count[state]<retrials[state])&&(retrial_count[state]!=0))
        {
#ifdef DEBUG_RST_2
        printf("ERROR:%d.Schwelle(%d Marken) unterschritten - zuruecksetzen\n"
               ,state,t_value[state]);
#endif
          ++retrial_count[state];
	  load_simstate( &imm_in_list, state );
        }
        else
        {
#ifdef DEBUG_RST_2
          printf("ERROR:%d.Schwelle(%d Marken) unterschritten - fertig => weiter\n"
                 ,state,t_value[state]);
#endif
          --state;
	  if ((state==0) && (retrial_count[1]==retrials[1]))
          {
            set_of_retrials++;
#ifdef DEBUG_RST_1
            printf("ERROR:===> sor=%d  rare_event_count=%d Satz=%d Prob_rare_event=%g nt=%d not=%d\n"
                   ,set_of_retrials
                   ,rare_event_count
                   ,Satz
                   ,Prob_rare_event
                   ,no_thresholds
                   ,no_opt_thresholds);
#endif
          }
        }
    if (optimiere)
    {
      if (!no_more_test)
       /* check if simulation should new started */
       if (!ns && (no_thresholds!=no_opt_thresholds) &&  (Satz==1)
	    &&(((set_of_retrials>=1000) && (rare_event_count <= 10))
	       || (rare_event_count>= MAX_OCCUR_OF_RARE_EVENT*5)))
       {

#ifdef DEBUG_RST
            printf("ERROR:===> sor=%d  rare_event_count=%d Satz=%d Prob_rare_event=%g nt=%d not=%d\n"
                   ,set_of_retrials
                   ,rare_event_count
                   ,Satz
                   ,Prob_rare_event
                   ,no_thresholds
                   ,no_opt_thresholds);
#endif

          /* probably too many thresholds */
          if (Prob_rare_event!=0.0)           
          {
            i=(int)(-0.5*log(Prob_rare_event));
            if (!ra_sample || (ra_sample && i>no_thresholds))
            {
              ra_sample=1;
              if (i!=no_thresholds) 
              {
#ifdef DEBUG_RST
               printf("ERROR:Probably too many thresholds => call new_start with %d thresholds\n",i);
#endif
               new_start(i,&set_of_retrials);
               ns=1;
               rare_event_count=0;
               MAX_OCCUR_OF_RARE_EVENT=5000;
               for (i=1;i<=no_thresholds;++i)
                  if (((places[place_no].token>=t_value[i])&&(direction==1))||
                      ((places[place_no].token<=t_value[i])&&(direction==-1)))
                    state=i;
                for (i=0; i<=MAX_THRESHOLDS;++i) retrial_count[i]=0;
              }
            }
            else 
            {
             no_more_test=1;
            }
          }
          /* probably not enough thresholds */
          else if(!ra_sample)                  
          {
#ifdef DEBUG_RST
            printf("ERROR:Probably not enough thresholds => call new_start with %d thresholds\n"
		   ,((int)(ceil((double)(no_thresholds)*1.2))));
#endif
            new_start((int)(ceil((double)(no_thresholds)*1.2)),&set_of_retrials);
            ns=1;
            rare_event_count=0;
            MAX_OCCUR_OF_RARE_EVENT=5000;
            for (i=1;i<=no_thresholds;++i)
              if (((places[place_no].token>=t_value[i])&&(direction==1))||
                  ((places[place_no].token<=t_value[i])&&(direction==-1)))
	 state=i;
              for (i=0; i<=MAX_THRESHOLDS;++i) retrial_count[i]=0;
          } 
       }

       /* check if simulation can stopped for reset_thresholds */
       if (!alles_optimiert && (Satz <= MAX_RESETTING_OF_THRESHOLDS))
        if ((pre_no_opt_thresholds && (set_of_retrials>=1)) || (!pre_no_opt_thresholds&&(set_of_retrials>=MAX_SET_OF_RETRIALS)))
          if ((state==0) && (retrial_count[1]==retrials[1]) && (rare_event_count >= MAX_OCCUR_OF_RARE_EVENT))
          {

#ifdef DEBUG_RST
            printf("ERROR:READY for reset_thresholds:\n===> sor=%d  rare_event_count=%d Satz=%d Prob_rare_event=%g nt=%d not=%d\n"
                   ,set_of_retrials
                   ,rare_event_count
                   ,Satz
                   ,Prob_rare_event
                   ,no_thresholds
                   ,no_opt_thresholds);
#endif

            if (  Prob_opt_t_OK( (pre_no_opt_thresholds  < 2  ) ? pre_no_opt_thresholds: 2 ,prob_opt))
            {
              alles_optimiert=reset_thresholds(1,event,Prob_rare_event,pre_no_opt_thresholds);
              Satz++;
              Satz_changed=1;
              set_of_retrials=0;
              MAX_OCCUR_OF_RARE_EVENT=RESTART_BATCH_SIZE_DEF*5;
              rare_event_count=0;
            }
            else 
            { 
              set_of_retrials=0; 
              MAX_OCCUR_OF_RARE_EVENT=rare_event_count+1;
            }
          }

       /* all thresholds has been optimize: last optimizing of retrials */
       if ((alles_optimiert==1) && (Satz <= MAX_RESETTING_OF_THRESHOLDS))
        if ((pre_no_opt_thresholds  && (set_of_retrials>=1)) || (!pre_no_opt_thresholds&&(set_of_retrials>=MAX_SET_OF_RETRIALS)))
          if ((state==0) && (retrial_count[1]==retrials[1]) && (rare_event_count >= MAX_OCCUR_OF_RARE_EVENT))
          {

#ifdef DEBUG_RST
            printf("ERROR:===> sor=%d  rare_event_count=%d Satz=%d Prob_rare_event=%g nt=%d not=%d\n"
                   ,set_of_retrials
                   ,rare_event_count
                   ,Satz
                   ,Prob_rare_event
                   ,no_thresholds
                   ,no_opt_thresholds);
#endif

            if  (Prob_opt_t_OK((pre_no_opt_thresholds  < 2  ) ? pre_no_opt_thresholds: 2 ,prob_opt))
            {
              alles_optimiert=reset_thresholds(0,event,Prob_rare_event,pre_no_opt_thresholds);
              Satz++;
              Satz_changed=1;
            }
            set_of_retrials=0;
            MAX_OCCUR_OF_RARE_EVENT=RESTART_BATCH_SIZE_DEF*5;
            rare_event_count=0;
          }
    } 
#endif
  }
#ifdef STR
  Measure_Update();
  free_simu_kernel_pointer();
  free_sim_list_pointer();
#endif
}



