/* #define DEBUG_RANDS */

/******************************************************************************
Programm	:	exec_rands.c

Parameter	:	none, because this file is part of the library ar.a

Purpose		: 	Execution of rand processes either local or
			distributed on different hosts

Project Director:	Christoph Lindemann TU Berlin

Programmer	:	Tobias Bading

History		:	release 1: October 1992
			release 2: July 1993, Armin Zimmermann
				   each rand process solves one isomorphism
				   class instead of a SMC
*                                                                           *
*  Modifiziert	  :	Joachim Schneider				    *
*										    *
*  Datum	  :	July 1994 					    *

Aenderung       :       Juni 1995 (Patrick Drescher)

******************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <fcntl.h>
#include <setjmp.h>
#include <unistd.h>
#include <ctype.h>

#include "portab.h"
#define TNETHOME_ENV "TNETHOME"
#define BIN_DIR "StatAnalysis/bin"
#define TNET_RC ".TimeNET"

#if !defined(ultrix) && !defined(__hpux) && \
    !defined(SOL2) && !defined(__osf__) && !defined(linux)
#  include <vfork.h>
#endif

#ifdef SOL2
#include <sys/filio.h>
#endif

#include <rpc/rpc.h>
#include "rusers.h" /* <rpcsvc/rusers.h> not identical on all systems */
#include "rstat.h"  /* <rpcsvc/rstat.h> as well */

#include "graph_types.h"
#include "graph_extern.h"
#include "csmc_types.h"
#include "exec_types.h"
#include "matrices_Q_delta.h"
#include "matrices_P_C.h"

#define false 0
#define true 1

randinfo *randinfos;            /* Informationen ueber die zugeordneten Markoff- Ketten */
	/* eta und q Werte von allen Markoff-Ketten.
	Diese Daten werden in graph_timed.c erzeugt. Dieses Array hat die
	Groesse no_det_trans + 1, wobei randinfos[0] immer dem Prozess,
	welcher fuer die Menge Mexp zustaendig ist, zugeordnet ist. Die
	Eintraege randinfos[1] bis randinfos[no_det_trans] werden bei ver-
	teilter Ausfuehrung nach ihrem Eintrag effort absteigend sortiert,
	d.h. randinfos[1] ist mit dem groesstem Aufwand behaftet. */

/* #define DEBUG_RANDS 1 */

#define OUT_OF_MEMORY { \
  perror (NULL); \
  fprintf (stderr, "error in %s at line %u: out of memory !\n", \
           __FILE__, __LINE__); \
  exit (1); \
}

#define FILE_WRITE_ERROR(filename) { \
  perror (NULL); \
  fprintf (stderr, "error in %s at line %u: can't write to file '%s'!\n", \
	   __FILE__, __LINE__, filename); \
  exit (1); \
}

#define MAX_KB_PER_SEC_SUM 150
#define MIN_PARALLEL 1
#define MAX_PARALLEL 5


/************************/
/*** Typdefinitionen: ***/
/************************/
/* Typ des Uebergabeparameters der Funktion receive_messages: */
typedef enum receive_mode {
  wait_for_one, wait_for_all
} receive_mode;


/*******************************/
/*** Funktionsdeklarationen: ***/
/*******************************/
/* extern: */
extern char *getenv ();
extern double det_transition_delay ();
extern int rusers2();
extern int rstat2();

void execute_rand_processes ();	/* Hauptroutine, wird von graph_main.c aus
				   aufgerufen */
/* modul-lokal: */
void my_fwrite ();		/* Senden von Daten an rand-Prozesse */
static void receive_messages ();	/* Lesen von Nachrichten der rand-Prozesse */
static void start_rand_process ();	/* rand-Prozess starten */
static void sigchld_handler ();	/* Handler fuer SIGCHLD-Signale */
static void local_execution ();	/* rand-Prozesse lokal starten */
static void distributed_execution ();	/* rand-Prozesse verteilt starten */


/******************/
/*** Variablen: ***/
/******************/
/* extern: */
extern char **environ;
/* aus graph_main.c: */
extern bool write_files;		/* .MDET_<k> und .DELTA_<k>-Files
					   schreiben ? */
/* modul-lokal: */
static char *netname;			/* Name des EDSPN */
static jmp_buf jmp_env;			/* fuer setjmp () und longjmp () */
time_t start_time;			/* Startzeit der Funktion
					   execute_rand_processes */
static int server_fd;			/* Filedescriptor fuer die
					   Server-Socket */
static unsigned short portnumber;	/* Portnummer dieser Socket */
static char *local_hostname;		/* Name des lokalen Rechners */
static hostinfo *local_host;		/* Informationen ueber den lokalen
					   Rechner */
static double sum_of_rusage_times;	/* Summe der CPU-Zeiten der rand-
					   Prozesse */
static unsigned running_processes;	/* Anzahl der gerade laufenden
					   Prozesse ohne den fuer Mexp */
static unsigned terminated_processes;	/* Anzahl in einer bestimmten Zeit
					   terminierten rand-Prozesse */
static unsigned kb_per_sec_sum;		/* Geschaetze Platten-Schreibge-
					   schwindigkeit der laufenden
					   Prozesse */
static enum {
  local, distributed
} execution_mode;		/* Ausfuehrungsart: lokal oder verteilt */
static unsigned free_hosts;		/* Anzahl freier Rechner */

/* Informationen aus dem Environtment-File: */
static double start_distribution_at;	/* Aufwand, ab welchem rand-Prozesse
					   verteilt werden */
static hostinfo *hostlist;		/* Liste der Rechner, auf denen rand-
					   Prozesse gestartet werden koennen */
static hostinfo **sorted_hosts;		/* Array von freien Rechnern, sortiert
					   nach Verwendungsreihenfolge */
static unsigned max_processes_per_host;	/* Maximale Anzahl der rand-Prozesse
					   pro Rechner */
static bool force_local_execution,	/* lokale Ausfuehrung erzwingen ? */
	    force_distributed_execution;/* verteilte Ausfuehrung erzwingen ? */

/* externe Variablen */
extern int no_normal_classes;		/* Anzahl der zu startenden rand-prozesse */
extern Iso_Class *first_iso_class;	/* first 

/******************************/
/*** Funktionsdefinitionen: ***/
/******************************/
/******************************************************************************

 Funktion:	my_fwrite

 Parameter:	data:	Zeiger auf die zu schreibenden Daten
		length: Anzahl der zu schreibenden Bytes
		filepointer: Filepointer (einer Socket zu einem rand-Prozess)

 Rueckgabewert:	---

 Zweck:		Senden von Datenpaketen (ueber eine Socket)

******************************************************************************/
void my_fwrite (data, length, filepointer)
char *data;
unsigned length;
FILE *filepointer;
{
  int rt;

nochmal:
  errno = 0;
  if ((rt = write ((int)fileno (filepointer), data, length)) != length)
  {
    if (rt >= 0)
    {
      data += rt;
      length -= rt;
      goto nochmal;
    }
    perror (__FILE__);
    fprintf (stderr, "error in %s: can't write to socket!\n", __FILE__);
    exit (1);
  }
}


/******************************************************************************

 Funktion:	sigchld_handler

 Parameter:	---

 Rueckgabewert:	---

 Zweck:		Erloesen von Zombi-Prozessen

******************************************************************************/
static void sigchld_handler ()
{
  int status;

  waitpid (-1, &status, WNOHANG);
}

/******************************************************************************

 Funktion:	establish_server_socket

 Parameter:	---

 Rueckgabewert:	---

 Zweck:		Erstellen einer Socket, an der sich spaeter die rand-Prozesse
		melden. Die Portnummer wird dabei vom Kernel bestimmt.

******************************************************************************/
static void establish_server_socket ()
{
  struct sockaddr_in server;
  int length = sizeof (server);

  /* Server-Socket oeffnen: */
  if ((server_fd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
  {
    perror ("socket");
    exit (1);
  }
  memset (&server,0, sizeof (server));
  server.sin_family = AF_INET;
  server.sin_port = 0;
  server.sin_addr.s_addr = INADDR_ANY;
  if (bind (server_fd, (struct sockaddr *) &server, sizeof (server)))
  {
    perror ("bind");
    exit (1);
  }

  /* Portnummer ermitteln: */
  getsockname (server_fd, (struct sockaddr *) &server, &length);
  if (length == sizeof (server))
    portnumber = ntohs (server.sin_port);
  else
  {
    perror ("getsockname");
    exit (1);
  }

  if (listen (server_fd, 5))
  {
    perror ("listen");
    exit (1);
  }
}

/******************************************************************************

 Funktion:	start_rand_process

 Parameter:	Zeiger auf eine randinfo-Struktur, epsilon

 Rueckgabewert:	Die Eintraege randinfo->status und
		randinfo->host->running_processes werden modifiziert

 Zweck:		Starten eines rand-Prozesses auf dem Rechner, der durch
		randinfo->host->hostname spezifiziert ist

******************************************************************************/
static void start_rand_process (info, epsilon)
struct randinfo *info;
double epsilon;
{
  static char *dspnhome, port[10],epsilon_string[64];
  char commandpath[200];
  int stderr_backup_fd;
  register hostinfo *host = info->host;

  if (!port[0])
    sprintf (port, "%u", portnumber);

  sprintf(epsilon_string,"%lg",epsilon);

  /* Pfad fuer das Executable rand bestimmen: */
  if (!dspnhome && !(dspnhome = getenv (TNETHOME_ENV)))
  {
    fprintf (stderr, "error in %s: environment variable '%s' not set!\n", __FILE__,TNETHOME_ENV);
    exit (1);
  }
  sprintf (commandpath, "%s/%s/rand", dspnhome,BIN_DIR);

  /* vfork () statt fork (), da dieser Prozess extrem viel Speicher schluckt.
     Laestiger Nebeneffekt: Bei lokaler Ausfuehrung der rand-Prozesse stimmen
     wegen vfork () die Speicherbelegungsangaben von getrusage () nicht mehr */
#if !defined(ultrix) && !defined(__hpux) && \
    !defined(SOL2) && !defined(__osf__)
  switch (vfork ())
#else
  switch (fork ())
#endif
  {
    case -1: perror ("vfork");
             exit (1);
    case 0:
    {
#ifndef ultrix
      /* Standarderror-Filedescriptor merken: */
      stderr_backup_fd = dup (2);
#endif

      if (host != local_host)
      {
	/* Rand-Prozess auf einem anderen Rechner mittels rsh starten: */
	execl (RSH_CMD_PATH, RSH_CMD, host->hostname, "-n", "nice", "+4",
	       commandpath, local_hostname, port, netname,epsilon_string, NULL);
	perror ("can't exec rsh process");
	exit (1);
      }
      else
      {
	/* Rand-Prozess lokal starten: */
	execl (commandpath, "rand", local_hostname, port, netname, epsilon_string, NULL);
	perror ("can't exec rand process");
	exit (1);
      }
    }
  }

  ++host->running_processes;
  info->status = waiting_for_connection;

  printf ("Starting transient analysis of isomorphism class %u\n", info->Class->class_id);
  printf ("with %d isomorphic component%s on station %s.\n\n", 
	  info->Class->no_total_csmcs,
	  info->Class->no_total_csmcs > 1 ? "s" : "", host->hostname);
}


/******************************************************************************

 Funktion:	wait_for_connection

 Parameter:	---

 Rueckgabewert:	Zeiger auf die randinfo-Struktur des rand-Prozesses, zu
		welchem eine Socketverbindung aufgebaut wurde. Ausserdem
		werden die Felder status, connection_fd und connection_fp
		des Rueckgabewertes modifiziert.

 Zweck:		Auf die Verbindung zu einem rand-Prozess warten. Ist eine
		Verbindungsaufnahme erfolgt, wird die zur Verbindung gehoerende
		randinfo-Struktur modifiziert.

******************************************************************************/
/* Auf die Verbindung zu einem rand-Prozess warten: */
/* Rueckgabewert: Index in die Arrays connection_fds, connection_fps
   und connection_hosts */
static randinfo *wait_for_connection ()
{
  struct sockaddr_in connection_addr;
  struct hostent *hostentry;
  int fd, connection_addrlen;
  struct timeval timeout;
  fd_set fd_readset;
  int select_return;
  unsigned i;

  while (1)
  {
    FD_ZERO (&fd_readset);
    FD_SET (server_fd, &fd_readset);
    timeout.tv_sec = 20L;
    timeout.tv_usec = 0L;

    /* Maximal 20 Sekunden auf die Verbindung zu einem rand-Prozess warten: */
    do
#ifdef DEBUG_RANDS
      select_return = select (FD_SETSIZE, &fd_readset, NULL, NULL, NULL);
#else
      select_return = select (FD_SETSIZE, &fd_readset, NULL, NULL, &timeout);
#endif
    while (select_return < 0 && errno == EINTR);
    if (select_return < 0)
    {
      perror ("select");
      exit (1);
    }
    if (!select_return)
      return NULL;

    /* Verbindung herstellen: */
    memset(&connection_addr,0, sizeof (connection_addr));
    do
    {
      connection_addrlen = sizeof (connection_addr);
      fd = accept (server_fd, (struct sockaddr *) &connection_addr, &connection_addrlen);
    }
    while (fd < 0 && errno == EINTR);
    if (fd < 0)
    {
      perror ("accept");
      exit (1);
    }

    /* Adresse des Rechners, auf welchem der rand-Prozesse laeuft, ermitteln: */
    if (!(hostentry = gethostbyaddr ( (char *)&connection_addr.sin_addr.s_addr, 4,
				     AF_INET)))
    {
      perror ("gethostbyaddr");
      exit (1);
    }
    {
      register char *dot = strchr (hostentry->h_name, '.');
      if (dot)
	*dot = '\0';
    }

    for (i = 1; i <= no_normal_classes; ++i)
    {
      register randinfo *info = &randinfos[i];
      if (info->status == waiting_for_connection && !strcmp (hostentry->h_name,
							     info->host->hostname))
      {
	info->status = connected;
	info->connection_fd = fd;
	/* Buffermechanismen der C-Library an die Socketverbindung haengen: */
	if (!(info->connection_fp = fdopen (fd, "w")))
	{
	  perror ("fdopen");
	  exit (1);
	}
#ifdef DEBUG_RANDS
	if (execution_mode == distributed)
	  printf ("connected with %s.\n", hostentry->h_name);
#endif
	return info;
      }
    }

    /* Fehler: Verbindung zu einem unbekannten Prozess wurde hergestellt */
    fprintf (stderr, "unkown process on station %s has tried to connect.\n",
	     hostentry->h_name);
    close (fd);
  }
}


/******************************************************************************

 Funktion:	receive_messages

 Parameter:
 		mode == wait_for_one -> Warten, bis ein rand-Prozess fertig ist
		mode == wait_for_all -> Warten, bis alle rand-Prozesse
					fertig sind

 Rueckgabewert:	Einige Felder von randinfo-Strukturen terminierter Prozesse
		werden modifiziert.

 Zweck:		Lesen einer Nachricht (Bildschirmausgabe) eines rand-Prozesses.
		Die Nachrichten eines rand-Prozesses werden gesammelt aus-
		gegeben, sobald der Prozess terminiert ist. Je nach mode
		wartet die Funktion evtl. bis ein oder alle rand-Prozesse
		terminiert sind.

******************************************************************************/
static void receive_messages (mode)
receive_mode mode;
{
  static char **outputs;		/* Array von Strings, in welchen die
					   Ausgaben jedes rand-Prozesses ge-
					   speichert werden */
  fd_set fd_readset;			/* Menge von Read-Filedescriptoren
					   fuer select */
  char message_buffer[200], *message;
  bool ready, once_again, rand_abgebrochen;
  int select_return, read_return;
  unsigned i;

  if (!outputs)
    if (!(outputs = (char **)calloc (no_normal_classes+1, sizeof (char *))))
      OUT_OF_MEMORY;

  do
  {
    /* Filedescriptor readset erstellen: */
    FD_ZERO (&fd_readset);
    for (ready = true, i = 1; i <= no_normal_classes; ++i)
      if (randinfos[i].status == running)
      {
	FD_SET (randinfos[i].connection_fd, &fd_readset);
	ready = false;
      }
    if (ready)
      /* Alle rand-Prozesse fertig */
      return;

    once_again = false;
    do
    {
	/* Auf Nachrichten der rand-Prozesse warten: */
        select_return = select (FD_SETSIZE, &fd_readset, NULL, NULL, NULL);
    }
    while (select_return < 0 && errno == EINTR);
    if (select_return < 0)
    {
      perror ("select");
      exit (1);
    }
    if (select_return > 0)
      for (i = 1; i <= no_normal_classes; ++i)
      {
	register randinfo *info = &randinfos[i];
        if (info->status == running &&
	    FD_ISSET (info->connection_fd, &fd_readset))
	{
	  /* Von rand-Prozess i ist eine Nachricht da */
	  message = message_buffer - 1;
	  rand_abgebrochen = false;
	  do
	  {
	    if ((read_return = read (info->connection_fd, ++message, 1)) < 0)
	    {
	      perror ("read on socket after select");
	      exit (1);
	    }
	    else
	      if (!read_return)
	      {
		rand_abgebrochen = true;
		break;
	      }
	  }
	  while (*message && *message != '\n' &&
		 message - message_buffer < 199);
	  *message = '\n';
	  *++message = '\0';
	  if (rand_abgebrochen || !strncmp (message_buffer, "ready.", 6))
	  {
	    /* rand-Prozess i ist fertig,
	       alle gesammelten Nachrichten ausgeben: */
	      if (rand_abgebrochen && execution_mode == distributed)
	      {
		printf ("Transient analysis of isomorphism class %u failed!\n%s",
			info->Class->class_id, outputs[i]);
		puts ("trying again...\n");
	      }
	      else
	      {
	  	printf ("Transient analysis of isomorphism class %u terminated.\n",
		        info->Class->class_id);
		printf ("%s\n", outputs[i]);
	      }
	    /* Socket-Verbindung schliessen: */
	    fclose (info->connection_fp);
	    info->connection_fp = (FILE *)0;
	    info->connection_fd = 0;
	    --info->host->running_processes;
	    info->host = (hostinfo *)0;
	    if (rand_abgebrochen && execution_mode == distributed)
	      info->status = no_process;
	    else
	      info->status = terminated;
	    if (i)
	    {
	      ++terminated_processes;
	      --running_processes;
	      kb_per_sec_sum -= info->kb_per_sec;
	    }
	    free (outputs[i]);
	    outputs[i] = (char *)0;
	    if (mode == wait_for_one)
	      return;
	  }
	  else
	  {
	    if (!strncmp (message_buffer, "Elapsed CPU time", 16))
	    {
	      /* Summe der CPU-Zeiten korrigieren: */
	      double rt1, rt2, rt3;
	      sscanf (message_buffer + 17, "%lf:%lf:%lf", &rt1, &rt2, &rt3);
	      sum_of_rusage_times += rt3 + 60 * (rt2 + 60 * rt1);
	    }
	    /* Die Nachricht des rand-Prozesses bis zu seiner
	       Beendigung speichern: */
	    if (!outputs[i])
	    {
	      if (!(outputs[i] = strdup (message_buffer)))
		OUT_OF_MEMORY;
	    }
	    else
	    {
	      if (!(outputs[i] = realloc (outputs[i], strlen (outputs[i]) +
					  strlen (message_buffer) + 1)))
		OUT_OF_MEMORY;
	      strcat (outputs[i], message_buffer);
	    }
	  }
	  once_again = true;
	}
      }
  }
  while (once_again);
}


/******************************************************************************

 Funktion:	read_environment

 Parameter:	Name des Environment-Files

 Rueckgabewert:	---

 Zweck:		Einlesen der fuer dieses Modul relevanten Eintraege des
		Environment-Files, z.B.: Hostliste, ...

******************************************************************************/
static void read_environment (filename)
char *filename;
{
  FILE *fp;
  char line[300];
  hostinfo *info = NULL, *prev_hostinfo = NULL;
  enum { NONE = 0, SUN, DEC } reading_hostlist = NONE;
  unsigned priority = 1;

  if (!(fp = fopen (filename, "r")))
  {
    fprintf (stderr, "error in %s: can't open file '%s' for reading!\n", __FILE__, filename);
    exit (1);
  }

  /* Default-Werte: */
  start_distribution_at = 200.0;
  max_processes_per_host = 2;

  do
  {
    /* Zeile aus dem Environment-File lesen: */
    fgets (line, 298, fp);

    /* Kommentarzeilen ignorieren */
    if (line[0] == '#' || line[0] == '!' || line[0] == '\n' || line[0] == '\0')
      continue;

    /* eventuelles Newline am Zeilenende entfernen: */
    if (line[strlen (line) - 1] == '\n')
      line[strlen (line) - 1] = '\0';
/*
    if (!strncasecmp (line, "SUNHOSTLIST", 11))
    {
      reading_hostlist = SUN;
      continue;
    }
*/
    if (!strncasecmp (line, "HOSTLISTSTART", 13))
    {
#ifdef ultrix
      reading_hostlist = DEC;
#else
      reading_hostlist = SUN;
#endif
      continue;
    }
    if (!strncasecmp (line, "HOSTLISTEND", 11))
    {
      reading_hostlist = NONE;
      continue;
    }

#ifdef ultrix
    if (reading_hostlist == DEC)
#else
    if (reading_hostlist == SUN)
#endif
    {
      if (!strncasecmp (line, "PRIORITY", 8))
      {
	sscanf (line + 9, "%d", &priority);
	continue;
      }

      /* Den gelesenen Rechnernamen in der Hostliste vermerken: */
      if (!(info = (hostinfo *)malloc (sizeof (hostinfo))))
	OUT_OF_MEMORY;
      {
	register char *dot = strchr (line, '.');
	if (dot)
	  *dot = '\0';
      }

      if (!strncasecmp (line, "USE ALWAYS", 10))
      {
	/* Test mittels rusers, ob jemand am Rechner sitzt, unterdruecken */
	register char *character;

	for (character = line + 10; *character && !isalpha (*character); ++character);
	if (!*character)
	{
	  fprintf (stderr, "missing hostname after USE ALWAYS in Environmentfile!\n");
	  free (info);
	  continue;
	}
	info->no_rusers = true;
	if (!(info->hostname = strdup (character)))
	  OUT_OF_MEMORY;
      }
      else
      {
	info->no_rusers = false;
	if (!(info->hostname = strdup (line)))
	  OUT_OF_MEMORY;
      }
      info->running_processes = 0;
      info->free = false;
      info->priority = priority;
      if (prev_hostinfo)
	prev_hostinfo->next = info;
      else
	hostlist = info;
      prev_hostinfo = info;
    }
    else if (!strncasecmp (line, "MAX PROCESSES PER HOST", 22))
    {
      /* Maximale Anzahl der gleichzeitig auf einem Rechner laufenden rand-
	 Prozesse lesen: */
      sscanf (line + 23, "%u", &max_processes_per_host);
      if (max_processes_per_host < 1)
	max_processes_per_host = 1;
    }
    else if (!strncasecmp (line, "START DISTRIBUTION AT", 21))
    {
      /* Mindest-Aufwand fuer die Verteilung der rand-Prozesse lesen: */
      sscanf (line + 22, "%lf", &start_distribution_at);
      if (start_distribution_at < 0.0)
	start_distribution_at = 200.0;
    }
  }
  while (!feof (fp));

  if (info)
    info->next = (hostinfo *)0;
  else
  {
    fprintf (stderr, "error in %s: No hostlist found in file '%s'!\n",
	     __FILE__, filename);
    exit (1);
  }

  /* Testen, ob der Rechner, auf dem dieser Prozess gerade laeuft, in
     der Hostliste vermerkt ist: */
  {
    char hostname[200];
    hostinfo *info;

    if (gethostname (hostname, 199))
    {
      perror ("gethostname");
      exit (1);
    }
    {
      register char *dot = strchr (hostname, '.');
      if (dot)
	*dot = '\0';
    }

    for (info = hostlist; info; info = info->next)
      if (!strcmp (hostname, info->hostname))
	break;
    if (info)
      local_host = info;
    else
    {
      if (!(info = (hostinfo *)malloc (sizeof (hostinfo))))
	OUT_OF_MEMORY;
      if (!(info->hostname = strdup (hostname)))
	OUT_OF_MEMORY;
      info->running_processes = 0;
      info->free = false;
      info->priority = 0;
      info->next = hostlist;
      hostlist = local_host = info;
    }
    local_hostname = local_host->hostname;
  }

  fclose (fp);
}


/******************************************************************************

 Funktion:	next_free_host

 Parameter:	---

 Rueckgabewert:	Zeiger auf eine hostinfo-Struktur oder NULL, falls kein
		freier Rechner mehr gefunden wurde

 Zweck:		Bestimmung des naechsten freien Rechners aus der Hostliste.
		Frei bedeutet, dass niemand an der Console eingeloggt ist.

******************************************************************************/
/* Den naechsten freien Rechner aus der Hostliste bestimmen: */
static hostinfo *next_free_host ()
{
  static hostinfo *next_host = (hostinfo *)-1;
  unsigned j;
  register hostinfo *host;

  if ((void *)next_host == (void *)-1)
    next_host = hostlist;

  if (!next_host)
    return NULL;

  for (host = next_host; host; host = host->next)
  {
    utmpidlearr rusersinfo;
    register struct utmpidle *array;
    short found = false;

    if (!host->no_rusers)
    {
	  found = rusers2 (host->hostname, &rusersinfo)==0 ? true : false;
      if (found)
	  {
	  /* Ein Rechner gilt als frei, wenn niemand
	     an der Konsole eingeloggt ist */
	  array = rusersinfo.utmpidlearr_val;
	  for (j = 0; j < rusersinfo.utmpidlearr_len; j++)
	    if (!strncmp (array[j].ui_utmp.ut_line, "console", 8))
	    {
	      found = false;
	      break;
	    }
	  }
    }
	else
	  found = true;

	if (found)
    {
      host->free = true;
      ++free_hosts;
      next_host = host->next;
      return host;
    }
  }
  next_host = NULL;
  return NULL;
}


/******************************************************************************

 Funktion:	determine_load

 Parameter:	Zeiger auf eine hostinfo-Struktur

 Rueckgabewert:	Das Feld load der uebergebenen Struktur wird modifiziert

 Zweck:		Bestimmung der load (Auslastung) des durch 'host'
		spezifizierten Rechners

******************************************************************************/
static void determine_load (host)
hostinfo *host;
{
  struct statsswtch rstatinfo;

    if (rstat2 (host->hostname, &rstatinfo))
      host->load = host->no_rusers ? 0 : 2557;	/* 9.99 * 256 */
    else
      host->load = rstatinfo.avenrun[0];
}


/******************************************************************************

 Funktion:	compare_hostinfos

 Parameter:	2 Adressen von Zeigern auf hostinfo-Strukturen

 Rueckgabewert:	Integer, welcher angibt, ob item1 logisch kleiner, gleich
		oder groesser als item2 ist

 Zweck:		Vergleichsfunktion zum Sortieren der freien Rechner
		(Array sorted_hosts) nach ihrer Auslastung und
		Prioritaet mittels qsort()

******************************************************************************/
static int compare_hostinfos (item1, item2)
register hostinfo **item1, **item2;
{
  register int value1 = ((*item1)->load + 256) / ((*item1)->priority + 1),
	       value2 = ((*item2)->load + 256) / ((*item2)->priority + 1);

  return (value1 - value2);
}


/******************************************************************************

 Funktion:	sort_hosts_by_load

 Parameter:	---

 Rueckgabewert:	---

 Zweck:		Sortieren des Arrays sorted_hosts nach Auslastung und
		Prioritaet der Rechner

******************************************************************************/
static void sort_hosts_by_load ()
{
  register hostinfo *host, **host_in_array;

  if (!(sorted_hosts = (hostinfo **)malloc (free_hosts * sizeof (hostinfo *))))
    OUT_OF_MEMORY;

  host_in_array = sorted_hosts;
  for (host = hostlist; host; host = host->next)
    if (host->free)
      *host_in_array++ = host;

  if (host_in_array - sorted_hosts != free_hosts)
  {
    fprintf (stderr, "error in %s: free_hosts incorrect!\n", __FILE__);
    exit (1);
  }

  /* Rechner nach ihrer Prioritaet und ihrer Auslastung sortieren: */
  qsort (sorted_hosts, free_hosts, sizeof (hostinfo *), compare_hostinfos);
}


/******************************************************************************

 Funktion:	all_rands_active

 Parameter:	---

 Rueckgabewert:	true oder false

 Zweck:		Testet, ob alle rand-Prozesse aktiviert wurden, d.h. entweder
		laufen oder bereits fertig sind

******************************************************************************/
static bool all_rands_active ()
{
  register unsigned i;

  for (i = 0; i <= no_normal_classes; ++i)
    if (randinfos[i].status != running && randinfos[i].status != terminated)
      return (bool)false;

  return (bool)true;
}


/******************************************************************************

 Funktion:	all_rands_terminated

 Parameter:	---

 Rueckgabewert:	true oder false

 Zweck:		Testet, ob bereits alle rand-Prozesse fertig sind

******************************************************************************/
static bool all_rands_terminated ()
{
  register unsigned i;

  for (i = 0; i <= no_normal_classes; ++i)
    if (randinfos[i].status != terminated)
      return (bool)false;

  return (bool)true;
}


/******************************************************************************

 Funktion:	distributed_execution

 Parameter:	epsilon

 Rueckgabewert:	---

 Zweck:		Diese Funktion uebernimmt das Management der Prozessverteilung
		im Falle einer verteilten Ausfuehrung der rand-Prozesse, also:
		- benutzbare Rechner bestimmen
		- rand-Prozesse verteilen (Neustart durch Fehler abgebrochener
		  Prozesse)
		- Daten an rand-Prozesse senden
		- warten, bis alle rand-Prozesse terminiert sind

******************************************************************************/
static void distributed_execution (epsilon)
double epsilon;
{
  unsigned j;
  hostinfo *host, **next_host;
  bool process_started, one_process_per_host;

  printf ("Free hosts found:\n       ");
  do
  {
    host = next_free_host ();
    if (host)
    {
      printf ("%s ", host->hostname);
      fflush (stdout);
    }
  }
  while (host);
  putchar ('\n');

  if (free_hosts < 2)
  {
    puts ("\nLess than 2 free hosts found, switching to local execution...\n");
    local_execution ();
    return;
  }

  printf ("loads: ");
  for (host = hostlist; host; host = host->next)
    if (host->free)
    {
      register unsigned len = strlen (host->hostname);
      char format[20];

      determine_load (host);
      sprintf (format, "%%-%d.2f ", len);
      printf (format, (float)host->load / (float)256.0);
      fflush (stdout);
      if (host->load > 3 * 256)
      {
	/* Rechner mit einer load > 3.0 werden nicht verwendet */
	host->free = false;
	--free_hosts;
      }
    }
  puts ("\n");

  if (free_hosts < 2)
  {
    puts ("Less than 2 usable hosts found, switching to local execution...\n");
    local_execution ();
    return;
  }

  sort_hosts_by_load ();

  /* old MEXP */
  randinfos->status = terminated;

  one_process_per_host = false;
  next_host = sorted_hosts;
  while (!all_rands_terminated ())
  {
    /* Soviel rand-Prozesse wie moeglich verteilt parallel laufen lassen: */
    process_started = false;
    for (; next_host < &sorted_hosts[free_hosts] && !all_rands_active ();
	 ++next_host)
    {
      host = *next_host;
      if (host->free && host->running_processes < max_processes_per_host &&
	  (!one_process_per_host || !host->running_processes))
      {
	/* Auf dem Rechner host->hostname kann ein rand-Prozess
	   gestartet werden: */
	register unsigned j;

	for (j = 1; j <= no_normal_classes; ++j)
	  if (randinfos[j].status == no_process &&
	      running_processes < MAX_PARALLEL &&
	      (kb_per_sec_sum + randinfos[j].kb_per_sec <= MAX_KB_PER_SEC_SUM
	       || running_processes < MIN_PARALLEL))
	  {
	    randinfos[j].host = host;
	    start_rand_process (&randinfos[j], epsilon);
	    process_started = true;
	    ++running_processes;
	    kb_per_sec_sum += randinfos[j].kb_per_sec;
	    break;
	  }

	if (j > no_normal_classes)
	{
	  /* Es konnte kein rand-Prozess gestartet werden */
	  next_host = sorted_hosts;
	  one_process_per_host = true;
	  break;
	}
      }
    }
    if (next_host >= &sorted_hosts[free_hosts])
    {
      next_host = sorted_hosts;
      one_process_per_host = false;
    }
    if (process_started && !one_process_per_host && !all_rands_active ())
      continue;

    terminated_processes = 0;

    /* Auf eine Verbindung mit den gerade gestarteten rand-Prozessen warten: */
    for (j = 1; j <= no_normal_classes; ++j)
      if (randinfos[j].status == waiting_for_connection)
      {
	j = 0;
	if (!wait_for_connection ())
	{
	  /* Fehler aufgetreten: Ein rand-Prozess meldet sich nicht */
	  register unsigned k;

	  for (k = 1; k <= no_normal_classes; ++k)
	    if (randinfos[k].status == waiting_for_connection)
	    {
	      fprintf (stderr, "can't connect to rand process on station '%s' !\n",
		       randinfos[k].host->hostname);
	      randinfos[k].status = no_process;
	      randinfos[k].host->free = false;
	      --randinfos[k].host->running_processes;
	      randinfos[k].host = (hostinfo *)0;
	    }
	  putchar ('\n');
	  break;
	}
      }

    /* Daten an die gestarteten Prozesse senden: */
    for (j = 1; j <= no_normal_classes; ++j)
    {
      register randinfo *info = &randinfos[j];
      if (info->status == connected)
      {
	send_data_to_rand(info->connection_fp, info->Class);
	if (info->connection_fp)
	  fflush (info->connection_fp);
	info->status = running;
      }
    }
    /* Eventuell warten, bis ein rand-Prozess terminiert ist: */
    if (!terminated_processes)
    {
      /* Auf die Beendigung eines rand-Prozesses warten, aber den fuer
	 die Menge Mexp zustaendigen Prozess nicht beachten: */
      do
	receive_messages (wait_for_one);
      while (!terminated_processes);
    }
  }
}


/******************************************************************************

 Funktion:	local_execution

 Parameter:	epsilon

 Rueckgabewert:	---

 Zweck:		Diese Funktion steuert den Ablauf im Falle einer lokalen
		Ausfuehrung der rand-Prozesse. Entsprechend dem Eintrag
		MAX_PROCESSES_PER_HOST aus dem Environment-File koennen
		evtl. mehrere Prozesse parallel laufen.

******************************************************************************/
static void local_execution (epsilon)
double epsilon;
{
  unsigned rand_count, processes_to_start = max_processes_per_host + 1;
  FILE *rand_fp;

  terminated_processes = 0;
  for (rand_count = 1; rand_count <= no_normal_classes; ++rand_count)
  {
    register struct randinfo *info = &randinfos[rand_count];

    if (info->status == terminated)
      continue;

    /* Auf die Verbindung zum rand-Prozess warten: */
    while (1)
    {
      /* Einen rand-Prozess lokal starten: */
      info->host = local_host;
      start_rand_process (info, epsilon);

      /* Auf die Verbindung zum rand-Prozess warten: */
      if (wait_for_connection () != info)
	fprintf (stderr, "can't connect to rand process on station '%s' !\n\n",
		 local_hostname);
      else
	break;
    }
    rand_fp = info->connection_fp;

    send_data_to_rand(rand_fp, info->Class);

    if (info->connection_fp)
      fflush (rand_fp);
    info->status = running;

    if (!--processes_to_start)
    {
      /* eventuell warten, bis ein rand-Prozess fertig ist: */
      if (terminated_processes)
      {
	processes_to_start = terminated_processes;
	terminated_processes = 0;
      }
      else
      {
	/* Auf die Beendigung eines rand-Prozesses warten, aber den fuer
	   die Menge Mexp zustaendigen Prozess nicht beachten: */
	do
	  receive_messages (wait_for_one);
	while (!terminated_processes);
	processes_to_start = terminated_processes;
	terminated_processes = 0;
      }
    }
  }
  /* warten, bis alle rand-Prozesse fertig sind: */
  receive_messages (wait_for_all);
}


/******************************************************************************

 Funktion:	compare_randinfos

 Parameter:	2 Adressen von Zeigern auf randinfo-Strukturen

 Rueckgabewert:	Integer, welcher angibt, ob item1 logisch kleiner, gleich
		oder groesser als item2 ist

 Zweck:		Vergleichsfunktion zum Sortieren der randinfo-Strukturen
		nach dem geschaetzen Aufwand zur Loesung der Markoff-Ketten

******************************************************************************/
static int compare_randinfos (item1, item2)
register randinfo *item1, *item2;
{
  return (int)(item2->effort - item1->effort);
}


/******************************************************************************

 Funktion:	write_tmpfiles

 Parameter:	---

 Rueckgabewert:	---

 Zweck:		write_tmpfiles legt die Generatormatrizen und Delta-Matrizen
		der zugeordneten Markoff-Ketten in Files ab. Dies geschieht
		nur, wenn <netname>.graph mit der Option -f gestartet wird.
		Defaultmaessig geschieht dies nicht, da diese Option nur
		zum Testen, zur Entwicklung neuer Algorithmen etc. gedacht ist.

******************************************************************************/
void write_tmpfiles (netname)
	char	*netname;
{
  char mdet_filename[300], delta_filename[300];
  register int path;
  register tang_marking *mark;
  unsigned filenumber;
  FILE *mdet_fp, *delta_fp;

  for (filenumber = 0; filenumber <= no_det_trans; ++filenumber)
  {
    sprintf (mdet_filename, "%s.MDET_%02u", netname, filenumber);
    if (!(mdet_fp = fopen (mdet_filename, "w")))
    {
      fprintf (stderr, "error in %s: can't open file '%s' for writing!\n",
	       __FILE__, mdet_filename);
      exit (1);
    }
    if (filenumber)
    {
      sprintf (delta_filename, "%s.DELTA_%02u", netname, filenumber);
      if (!(delta_fp = fopen (delta_filename, "w")))
      {
	fprintf (stderr, "error in %s: can't open file '%s' for writing!\n",
		 __FILE__, delta_filename);
	exit (1);
      }
    }

    for (mark = filenumber ? det_markings[filenumber - 1] : exp_markings;
	 mark; mark = mark->next)
    {
      /* Zeilenheader fuer Matrix Q schreiben: */
      if (filenumber)
      {
	Q_tau_row_header.tau = det_transition_delay (mark);
	Q_tau_row_header.neg_sum = -(double)mark->sum_of_rates;
	Q_tau_row_header.row_index = mark->number;
	Q_tau_row_header.entries_in_row = mark->no_exp_paths;
	if (fwrite (&Q_tau_row_header, SIZEOF_Q_tau_row_header, 1, mdet_fp) != 1)
	  FILE_WRITE_ERROR (mdet_filename);
      }
      else
      {
	Q_row_header.neg_sum = -(double)mark->sum_of_rates;
	Q_row_header.row_index = mark->number;
	Q_row_header.entries_in_row = mark->no_exp_paths;
	if (fwrite (&Q_row_header, SIZEOF_Q_row_header, 1, mdet_fp) != 1)
	  FILE_WRITE_ERROR (mdet_filename);
      }

      /* Nicht-Null-Elemente der Zeile aus Matrix Q schreiben: */
      for (path = 0; path < mark->no_exp_paths; ++path)
      {
	Q_row_entry.rate = (double)mark->exp_paths[path].prob;
	Q_row_entry.column_index = ((tang_marking *)
				    mark->exp_paths[path].marking)->number;
	if (fwrite (&Q_row_entry, SIZEOF_Q_row_entry, 1, mdet_fp) != 1)
	  FILE_WRITE_ERROR (mdet_fp);
      }

      if (mark->no_det_paths)
      {
	/* Zeilenheader fuer Matrix Delta senden: */
	delta_row_header.row_index = mark->number;
	delta_row_header.entries_in_row = mark->no_det_paths;
	if (fwrite (&delta_row_header, SIZEOF_delta_row_header, 1, delta_fp)
	    != 1)
	  FILE_WRITE_ERROR (delta_filename);

	/* Nicht-Null-Elemente der Zeile aus Matrix Delta senden: */
	for (path = 0; path < mark->no_det_paths; ++path)
	{
	  delta_row_entry.probability = (double)mark->det_paths[path].prob;
	  delta_row_entry.column_index = ((tang_marking *)mark->det_paths[
					 path].marking)->number;
	  if (fwrite (&delta_row_entry, SIZEOF_delta_row_entry, 1, delta_fp)
	      != 1)
	    FILE_WRITE_ERROR (delta_filename);
	}
      }
    }
    fclose (mdet_fp);
    if (filenumber)
      fclose (delta_fp);
  }
}


/******************************************************************************

 Funktion:	execute_rand_processes

 Parameter:	Name des NETZES

 Rueckgabewert:	---

 Zweck:		Diese Funktion wird am Ende des Moduls graph_main.c aufgrufen,
		nachdem der reduzierte Erreichbarkeitsgraph generiert wurde.

******************************************************************************/
void execute_rand_processes (net_name, epsilon, exec_mode)
char *net_name;
double epsilon;
char *exec_mode;
{
  int det;
  char filename[400], *home;

  fprintf(stderr,"exec_rands: epsilon=%lg\n",epsilon);
  fprintf(stderr,"            exec_mode=%s\n",exec_mode);    

  setvbuf(stdout,NULL,_IOLBF,BUFSIZ);
  setvbuf(stderr,NULL,_IOLBF,BUFSIZ);

  netname = net_name;

  /* SIGCHLD- und SIGIO-Signale abfangen: */
  if ((void *)signal (SIGCHLD, sigchld_handler) == (void *)-1L)
  {
    perror ("signal");
    exit (1);
  }

  /* Server-Socket erzeugen: */
  establish_server_socket ();

  /* Environment-File einlesen: */
  if (!(home = getenv ("HOME")))
  {
    fprintf (stderr, "error in %s: environment variable 'HOME' not set!\n", __FILE__);
    exit (1);
  }
  sprintf (filename, "%s/%s", home, TNET_RC);
  read_environment (filename);

  force_local_execution = false;
  force_distributed_execution = false;
  if (!(strcmp(exec_mode, "parallel")))
      force_distributed_execution = true;
  else if (!(strcmp(exec_mode, "sequential")))
	   force_local_execution = true;
       else
	   {
	   fprintf (stderr, "error in %s: Unknown execution mode '%s' !\n",
	    __FILE__, exec_mode);
           exit (1);
	   }
  puts ("\n\nDERIVING THE EMBEDDED MARKOV CHAIN OF THE NET ...\n");
    {
      register double sum_of_efforts = 0.0;

      /* Aufwand abschaetzen: */
      for (det = 1; det <= no_normal_classes; ++det)
      {
	register randinfo *info = &randinfos[det];

	info->effort = info->max_qii * (double)info->eta * 0.00261;
	info->kb_per_sec = (unsigned)(sqrt ((double)info->eta) / info->max_qii);
	if (info->effort < 1.0)
	  info->kb_per_sec = (unsigned)((double)info->kb_per_sec * info->effort);
	if (info->kb_per_sec > MAX_KB_PER_SEC_SUM)
	  info->kb_per_sec = MAX_KB_PER_SEC_SUM;
	sum_of_efforts += info->effort;
      }

      /* Die zugeordneten Markoff-Ketten absteigend nach
	 geschaetztem Aufwand sortieren: */
      qsort (&randinfos[1], no_normal_classes, sizeof (randinfo), compare_randinfos);

      /*** Entscheiden, ob die rand-Prozesse verteilt werden sollen: ***/
      if (!(no_normal_classes < 2 || ((sum_of_efforts < start_distribution_at ||
				       force_local_execution)
				      && !force_distributed_execution)))
      {
	/*** rand-Prozesse parallel verteilt ausfuehren: ***/
	execution_mode = distributed;

	printf ("Starting all rand processes parallel on different hosts");
	if (max_processes_per_host > 1)
	  printf (" with max. %u concurrent processes.\n\n",
		  max_processes_per_host);
	else
	  puts (".\n");
	distributed_execution (epsilon);
      }
      else
      {
	/*** rand-Prozesse lokal ausfuehren: ***/
	execution_mode = local;

	printf ("Starting all rand processes locally ");
	if (max_processes_per_host > 1)
	   printf ("with max. %u concurrent processes.\n\n", max_processes_per_host);
	else
	   puts ("with sequential execution.\n");
	local_execution (epsilon);
      }
    }

  /* Zeiten ausgeben: */
    printf ("Total CPU time: %.2f sec.\n", sum_of_rusage_times);

    {
      long seconds, minutes = 0, hours = 0;
      seconds = time (NULL) - start_time;
      if (seconds > 59)
      {
	minutes = seconds / 60;
	seconds %= 60;
      }
      if (minutes > 59)
      {
	hours = minutes / 60;
	minutes %= 60;
      }
      printf ("Total overall time: %02lu:%02lu:%02lu\n",
	      hours, minutes, seconds);
    }
}
