/**************************************************************************
 File:		expoly_types.h

 Purpose:	include-file for expoly.c and rspnrand.c
		contains data structure of an expolynomial distribution

 Programmer:	Reinhard German

 History:	Release 1: June 2, 1993 (Reinhard German)
                Release 2: May, 1999 (Oliver Hesse)    
**************************************************************************/

#ifndef EXPOLY_TYPES_H
#define EXPOLY_TYPES_H

/**************************************************************************
  internal representation of an expolynomial distribution:

  recursive structure consisting of three hierarchy levels
  intervals on the real time axis: 0 =< l =< L
  powers of x with exponent m    : 0 =< m =< M
  exponentials with rate lambda_n: n =< n =< N
**************************************************************************/

typedef struct time_l {        /* time interval l */
  double x1;                   /* beginning of the interval */
  double x2;                   /* end of the interval (if < infinity) */
  double p;                    /* point mass */
  struct power_m *power;       /* list of powers */
  struct time_l *next;         /* next interval */
} time_l;

typedef struct power_m {       /* power with exponent m */
  unsigned m;                  /* exponent */  
  struct expo_n *expo;         /* list of exponentials */
  struct power_m *next;        /* next power */
} power_m;

typedef struct expo_n {        /* exponential with rate lambda_n */
  double lambda;               /* rate */ 
  double b;                    /* constant corresponding to 1-PDF */
  double c;                    /* constant corresponding to pdf */
  struct expo_n *next;         /* next exponential */
} expo_n;


#endif
