/****************************************************************************
** Evaluation : STATIONARY ANALYSIS                                        **
*****************************************************************************
** Program    : graph                                                      **
** Component  : graph_main                                                 **
** File       : graph_main.c (C source)                                    **
**                                                                         **
** Purpose    : main comonent of <Model>.graph with global variables       **
**              generation of the reduced reachability graph and write     **
**		the RRG to files                                           **
**              calls calculation of the CSMCs                             **
**              calls the selected solution method                         **
**                                                                         **
** Programmer :	Armin Zimmermann                                           **
** Date	      :	june of 1992                                               **
**                                                                         **
** Changed by : Patrick Drescher                                           **
**              june of 1995                                               **
**                                                                         **
** Changed by : Oliver Hesse                                               **
** Master's T.: realis. of an iterative solution method for not marc. SPNs **
** Date       : june of 1999                                               **
****************************************************************************/

/* #define GRAPHONLY */		  /* obtain only the reachability graph */
/* #define DEBUG_RG_RESORTED */   /* show the resorted reachability graph */
/* #define MEMDEBUG */		  /* obtain information about mem. alloc. */
/* #define EMEMDEBUG */
/* #define SHOW_SPECIAL_FORMAT */ /* show used times, mem. in spec. format */

#undef _POSIX_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <time.h>
#if ! defined (__hpux) && ! defined (__NetBSD__)
#include <alloca.h>
#endif
#include <string.h>
#include <sys/types.h>
#include <errno.h>

#include "graph_types.h"		 /* definitions and types */
#include "portab.h"

#include "TimeWrapper.h"                 /* time and memory usage */   

#define true  1
#define false 0

#define COUNT_DYNAMIC			 /* count dynamic memory allocation */

extern void generate_rg();		 /* from graph_timed.c */
extern void set();
extern void execute_rand_processes ();	 /* from execute_rand_processes.c */
extern int execute_erand_processes ();   /* from execute_erand.c */
extern time_t start_time;
void write_tmpfiles();
extern void csmc_main();		 /* from csmc_main */
extern void obtain_randinfos();
extern int  no_normal_classes;
extern void show_rusage();


void solve_main(const char* net_name, const char* exec_mode,
                const long arb_bits, const double trunc_error,
                const char* init_vec, const long seed_value,
                const char* save_vec,
		const int max_iter, const double epsilon);



/******* global variables *******/

int no_places;			/* number of places in the net */
int no_trans;			/* number of transitions in the net */
int no_subnets;			/* number of immediate transition subnets */
int no_det_trans;		/* number of deterministic transitions */
int no_gen_trans;               /* number of general transitions */
int last_class_id;              /* important for .EMCii & .CONVii for gen. trans. */
int no_tang_markings;		/* number of tangible markings */
int no_exp_markings;            /* number of exponential tangible markings */
int no_absorb_states;		/* number of absorbing markings */
int *no_van_submarks;		/* number of vanishing submarkings */

t_sub		timed;		/* timed transitions and disconnected places */
subnet		*subnets;	/* immediate transition subnets */
tang_node 	*r_graph;	/* reduced reachability graph */
tang_marking 	*new_markings;	/* list of all new markings */
tang_marking 	*exp_markings;	/* list of exponential markings */
tang_marking 	**det_markings;	/* lists of markings with enabled det. trans */
int             *no_nodes;      /* number of markings of the corresponding EMC */


FILE 	*struct_file;		/* input file with structural informations */
FILE	*rates_file;		/* RATES - file (firing rates of timed trans. in markings) */
FILE	*rg_file;		/* RG - file (description of tangible markings) */
FILE	*mexp_file;		/* MEXP - file (firing of exponential transitions) */
FILE	*mdet_file;		/* MDET - files (markoff chains of deterministic trans.) */
FILE	*delta_file;		/* DELTA - files (immediate trans. firing probabilities) */
int	out_bit;		/* information output option */
int	mem_bit;		/* memory save option from environment */

/* Globale Variablen, welche von exec_rands.c importiert werden: */
char		write_files = false;	/* Flag, Files .MDET_<k> und .DELTA_<k> schreiben ?
					   (fuer execute_rand_processes.c) */

#ifdef MEMDEBUG

#define	NO_ACCOUNTS	100

typedef struct
{	int num, val;
}  sort_type;

int	max[NO_ACCOUNTS], now[NO_ACCOUNTS], max_sum;

/*	comp is the comparison function for sortspace */
int
comp(a, b)
	sort_type *a, *b;
{
	if (a->val > b->val)
		return -1;
	return (a->val < b->val);
}
#endif

static mem_list_t memlist = {NULL,NULL};

/*	emalloc tries to allocate required memory space and
	tests if it was successfully allocated; if not => exit */
char *
emalloc(bytes, account)
	int bytes, account;
{
	int *point;

	if (!bytes)
		return NULL;
#ifdef MEMDEBUG
	point = (int *)malloc((unsigned)(bytes + 2*sizeof(int)));
#else
	point = (int *)malloc((unsigned)(bytes));
#endif
	if (point == NULL)
	{	printf("The reduced reachability graph of this DSPN\n");
		printf("cannot be generated because of lacking main memory.\n");
		if (mem_bit)
			printf("Retry on a workstation with more main memory.\n");
		exit(2);
	}

#ifdef MEMDEBUG
	*(point++) = account;
	*(point++) = bytes;
	now[account] += bytes;
	sum = 0;
	for(i = 0; i < NO_ACCOUNTS; i++)
		sum += now[i];
	if (sum > max_sum)
	{	max_sum = sum;
		for (i = 0; i < NO_ACCOUNTS; i++)
			max[i] = now[i];
	}
#endif
	return (char *)point;	
}

/*	efree frees a memory pointer */
void
efree(point)
	char	*point;
{
	int	*hpoint = (int *)point;

	if (point == NULL)
		return;
#ifdef MEMDEBUG
	bytes = *(--hpoint);
	acc = *(--hpoint);
	now[acc] -= bytes;
	if (now[acc] < 0)
	{	printf("ERROR in free (MEMDEBUG)\n");
		exit(1);
	}
	
#endif
	free(hpoint);
}

/*	skip_line skips to the beginning of the next line */
char
skip_line(f)
	FILE *f;
{       char buffer[100];
	fgets(buffer, 100, f);
	return buffer[0];
}

/*	read_sparse reads a sparse row */
sp_place *
read_sparse(f)
	FILE *f; 
{
	int count, i;
	sp_place *point;

	fscanf(f, "%d ", &count);	/* number of entries */
	point = (sp_place *)emalloc((count+1) * sizeof(sp_place), 1);
	point[0].num = count;		/* first entry.num = number of entries */
	for(i = 1; i<=count; i++)
	{	fscanf(f, "%d %d", &(point[i].num), &(point[i].value));
		fscanf(f, "%d %d", &(point[i].plu_id), &(point[i].min_id));
	}
	return point;
}

/*	read_set reads a set of transitions */
byte *
read_set(f)
	FILE *f;
{
	int count, i, b;
	byte *point;

	fscanf(f, "%d ", &count);	/* read number of entries */
	point = (byte *)emalloc((count+1) * sizeof(byte), 2);
	point[0] = count;		/* first entry: number of entries */
	for(i = 1; i <= count; i++)
	{	fscanf(f, "%d", &b);
		point[i] = b;
	}
	return point;
}

/*	read_struct reads the structural description file net.STRUCT */
void
read_struct(s_file)
	FILE *s_file;
{
	int	i, num, max;
	timed_t	*t_tra;
	imm_t	*i_tra;
	place	*pla;
	subnet	*sub;

	no_det_trans = no_gen_trans = 0;
	skip_line(s_file);		/* skip first line (comment) */
	fscanf(s_file, "%d %d", &no_trans, &no_places);
	fscanf(s_file, "%d", &max);
	timed.no_trans = max;
	timed.transitions = (timed_t *)emalloc(max * sizeof(timed_t), 3);
	timed.no_markings = 0;
	for(num = 0; num < max; num++)
	{	t_tra = &(timed.transitions[num]);
		fscanf(s_file, "%s %d %d ", t_tra->name, &t_tra->number, &t_tra->type);
		if (t_tra->type) /* DETERMINISTIC (or general) */
			t_tra->det_no = no_det_trans++;
                if (t_tra->type==2) /* GENERAL DISTRIBUTED */
                        no_gen_trans++;
                if (t_tra->type==1 && t_tra->infinite) /* DETERMINISTIC and INFINITE SERVER */
		  {
                  printf("Cannot compute infinite server of deterministic type ! \n");
                  exit(1);
		  }
		fscanf(s_file, "%d %d ", &t_tra->infinite, &t_tra->mark_ind);
		fscanf(s_file, "%lf\n", &t_tra->value);
		t_tra->i_matrix = read_sparse(s_file);
		t_tra->h_matrix = read_sparse(s_file);
		t_tra->c_matrix = read_sparse(s_file);
	}
	fscanf(s_file, "%d\n", &max);		/* disconnected places */
	timed.no_places = max;
	timed.places = (place *)emalloc(max * sizeof(place), 4);
	for(num = 0; num < max; num++)
	{	pla = &(timed.places[num]);
		fscanf(s_file, "%s %d ", pla->name, &pla->number);
		fscanf(s_file, "%d %d ", &pla->m0, &pla->bound);
		fscanf(s_file, "%d", &pla->bits);
	}
	fscanf(s_file, "%d", &timed.memory);
	/* read immediate transition subnets */
	fscanf(s_file, "%d", &no_subnets);
	subnets = (subnet *)emalloc(no_subnets * sizeof(subnet), 5);
	for(num = 0; num < no_subnets; num++)
	{	sub = &(subnets[num]);
		sub->end_mind = NULL;
		sub->no_markings = 0;
		fscanf(s_file, "%d", &max);
		sub->no_trans = max;
		sub->transitions = (imm_t *)emalloc(max * sizeof(imm_t), 6);
		for(i = 0; i<sub->no_trans; i++)
		{	i_tra = &(sub->transitions[i]);
			fscanf(s_file, "%s %lf", i_tra->name, &i_tra->prob);
			fscanf(s_file, "%d %d", &i_tra->wgt_id, &i_tra->ena_id);
			i_tra->i_matrix = read_sparse(s_file);
			i_tra->h_matrix = read_sparse(s_file);
			i_tra->c_matrix = read_sparse(s_file);
			i_tra->enas = read_set(s_file);
			i_tra->diss = read_set(s_file);
		}
		fscanf(s_file, "%d", &max);
		sub->no_places = max;
		sub->places = (place *)emalloc(max * sizeof(place), 7);
		for(i = 0; i < max; i++)
		{	pla = &(sub->places[i]);
			fscanf(s_file, "%s %d ", pla->name, &pla->number);
			fscanf(s_file, "%d %d ", &pla->m0, &pla->bound);
			fscanf(s_file, "%d", &pla->bits);
		}
		fscanf(s_file, "%d", &sub->memory);
		fscanf(s_file, "%d", &sub->no_ecs);
		sub->ecs = (byte **)emalloc(sub->no_ecs * sizeof(byte *), 8);
		for(i = 0; i < sub->no_ecs; i++)
			sub->ecs[i] = read_set(s_file);
	}
	
}

/*	print_info shows some information about the reachability graph */
void
print_info(net_name)
	char	*net_name;
{
	int	sn;
	char	aux_name[FILE_NAME];
	FILE	*aux_f;

	printf("\nThis %sSPN ", no_det_trans ? "D" : "G");
	printf("contains %d subnet%s.\n", no_subnets, no_subnets!=1 ? "s" : "");
	if (!mem_bit)
	{	int h = 0;
		printf("Number of significant submarkings:\n");
		for (sn = 0; sn < no_subnets; sn++)
		{	printf("\tsubnet %d : %d\n", sn+1, no_van_submarks[sn]);
			h += no_van_submarks[sn];
		}
		printf("Overall number of significant submarkings: %d\n", h);
	}

	printf("Number of tangible markings: %d\n", no_tang_markings);
	if (no_det_trans)
	{	printf("In %d markings are ", no_exp_markings);
		printf("only exponential transitions enabled.\n");
	}

	if (no_absorb_states)
	{	printf("%d absorbing state%s detected.\n", no_absorb_states,
			no_absorb_states > 1 ? "s" : "");
		strcpy(aux_name, net_name);
		strcat(aux_name, ".AUX");
		if ((aux_f = fopen(aux_name, "a")) == NULL)
		{	printf("ERROR: could not open %s for appending.\n", aux_name);
			exit(1);
		}
		fprintf(aux_f, "ABSORBING STATES %d\n", no_absorb_states);
		fclose(aux_f);
	}
	if (!out_bit)
		return;
	if (timed.no_places)
		printf("disconnected places: %d submarkings.\n", timed.no_markings);
	if (mem_bit)
		return;
	for(sn = 0; sn < no_subnets; sn++)
		printf("subnet %d (with place %s): %d submarkings.\n",
			sn, subnets[sn].places[0].name, subnets[sn].no_markings);
}



/****************************************************************************
** FUNCTION : show usage                                                   **
*****************************************************************************
** Purpose  : show usage to start <Model>.graph process                    **
****************************************************************************/

void show_usage()
{
  puts("Usage:");
  puts("MODEL.graph -ofm MODEL SOL_METHOD EXEC_MODE ARB_BITS EPSILON");
  puts("                 TRUNC_ERROR MAX_ITER INIT_VEC SEED_VAL SAVE_VEC");
  puts("");
  puts("-o = print informations");
  puts("-f = write .MDET_<k> and .DELTA_<k>-files");
  puts("-m = save memory");
  puts("");
}


/****************************************************************************
** FUNCTION : main                                                         **
*****************************************************************************
** Purpose  : main function of <Model>.graph process                       **
**                                                                         **
**            the main procedure handles options and file access           **
**            and calls subordinated procedures                            **
**                                                                         **
** Input    : argv[1]        special options [-ofm]                        **
**            argv[2]        name of the corresponding DSPN                **
**            argv[3]        solution method                               **
**            argv[4]        sequential or parallel execution              **
**            argv[5]        bits for arbitrary precision (calc. factors)  **
**            argv[6]        specified error tolerance                     **
**            argv[7]        truncation error (calc. factors)              **
**            argv[8]        maximal number of outer iterations            **
**            argv[9]        kind of initial iteration vector              **
**            argv[10]       seed value for randomisation of init. vector  **
**            argv[11]       option if stationary vector shall be saved    **
**                                                                         **
** Output   : exit(0)        programm exited normaly                       **
**            exit(1)        a failure occurred while programm execution   **
****************************************************************************/

int main(argc, argv)

	int argc;
	char ** argv;
{
        enum args { NETNAME, SOL_METHOD, EXEC_MODE,
                    ARB_BITS, EPSILON, TRUNC_ERROR,
                    MAX_ITER, INIT_VEC, SEED_VAL, SAVE_VEC };

	char struct_name[FILE_NAME], net_name[FILE_NAME];
	char rg_name[FILE_NAME], rates_name[FILE_NAME];

	int    i, arg_num;

        long   max_iter;
        long   arb_bits;
	double epsilon;
	double trunc_error;
        long   seed_val;

        void   *cpu_startTimeGraph;    // start time for graph process
        void   *day_startTimeGraph;    // start time for graph process
	void   *cpu_startTimeCSMC;     // start time for calculate CSMCs
	void   *day_startTimeCSMC;     // start time for calculate CSMCs
	void   *cpu_startTimeSol;      // start time for solution method
	void   *day_startTimeSol;      // start time for solution method

        long    mem_atStartGraph;      // used memory at start of calc. RG
        long    mem_atStartCSMC;       // used memory at start of calc. CSMCs
        long    mem_atStartSol;        // used memory at start of sol. method

	clock();                       // start timer for show_rusage()
 
#       ifdef MEMDEBUG
	    for (i=0; i<NO_ACCOUNTS; i++)
                now[i] = max[i] = 0;
	    max_sum = 0;
#       endif


	/**********************************
        *** check options and arguments ***
        **********************************/

        /* check number of arguments */
	if (!((argc = 6) || (argc != 10)))
	{	
	        printf("ERROR: illegal number of arguments\n");
		show_usage();
		exit(1);
	}

        /* check special options */
	out_bit = write_files = mem_bit = 0;

        arg_num = 1;
	if (argv[1][0] == '-')
	{
	        arg_num++;
	        for(i = 1; i < strlen(argv[1]); i++)
		        switch(argv[1][i])
			{	
                                case 'o':	out_bit = 1;
						break;
				case 'm':	mem_bit = 1;
						break;
				case 'f':	write_files = 1;
						break;
				default:	show_usage();
						exit(1);
			}
	}

        /* get begin of rest arguments */
	argv = &argv[arg_num];

        /* netname */
	strcpy(net_name, argv[NETNAME]);
	strcpy(struct_name, argv[NETNAME]);
	strcpy(rates_name, argv[NETNAME]);
	strcpy(rg_name, argv[NETNAME]);

        /* check solution method */
        if (strcmp(argv[SOL_METHOD], "FI_avoidance") &&
            strcmp(argv[SOL_METHOD], "EMC_explicit"))
	{    
                printf("ERROR: unknown SOL_METHOD\n");
                printf("use: [EMC_explicit | FI_avoidance]\n");
		exit(1);
	}

        /* check execution mode */
        if (strcmp(argv[EXEC_MODE], "sequential") &&
            strcmp(argv[EXEC_MODE], "parallel"))
	{       
                printf("ERROR: unknown EXEC_MODE\n");
                printf("use: [sequential | parallel]\n");
                exit(1);
        }	
	
        /* check arbitrary precision */
	sscanf(argv[ARB_BITS], "%ld", &arb_bits);
        if ((arb_bits > 0) && (arb_bits < 128))
	{       
                printf("ERROR: illegal value for ARB_BITS\n");
                printf("ARB_BITS: 128 ... 100000\n");
                printf("ARB_BITS: 0 (precision of double)\n");
                exit(1);
        }	
	
        /* check epsilon */
	sscanf(argv[EPSILON], "%lg", &epsilon);
        if ((epsilon <= 0.0) || (epsilon >= 1.0))
	{       
                printf("ERROR: illegal value for EPSILON\n");
                printf("use: 0 <= 1e-07 <= 1.0\n");
		exit(1);
	}

	

        /* check truncation error */
	sscanf(argv[TRUNC_ERROR], "%lg", &trunc_error);
        if ((trunc_error <= 0.0) || (trunc_error >= 1.0))
	{       
                printf("ERROR: illegal value for TRUNC_ERROR\n");
                printf("use: 0 <= 1e-07 <= 1.0\n");
		exit(1);
	}


        /* in case of fill-in avoidance solution method ... */
        if (!strcmp(argv[SOL_METHOD], "FI_avoidance"))
	{	
	        /* check maximal iterations */
	        sscanf(argv[MAX_ITER], "%ld", &max_iter);
                if (max_iter < 1)
	        {       
                        printf("ERROR: illegal value for MAX_ITER\n");
                        printf("MAX_ITER: 1 <= max. iter.\n");
                exit(1);
		}	

	        /* check initial vector */
                if (strcmp(argv[INIT_VEC], "uniform") &&
                    strcmp(argv[INIT_VEC], "random") &&
                    strcmp(argv[INIT_VEC], "load"))
		{
                        printf("ERROR: unknown option for INIT_VEC\n");
                        printf("INIT_VEC: [uniform | random | load ]\n");
                        exit(1);
		}

                /* check seed value */
                sscanf(argv[SEED_VAL], "%ld", &seed_val);
                if (seed_val < 0)
	        {       
                        printf("ERROR: illegal value for SEED_VAL\n");
                        printf("SEED_VAL: 0 <= max. iter.\n");
                        exit(1);
                }	


	        /* check option to save stationary vector */
                if (strcmp(argv[SAVE_VEC], "yes") &&
                    strcmp(argv[SAVE_VEC], "no"))
                {
                        printf("ERROR: unknown SAVE_VEC option\n");
                        printf("SAVE_VEC: [yes | no]\n");
                        exit(1);
		}
	}


        /******************
	*** check files ***
        ******************/

	strcat(struct_name, ".STRUCT");
	if ((struct_file = fopen(struct_name, "r")) == NULL)
	{      
                printf("ERROR: could not open %s for reading\n", struct_name);
		exit(1);
	}
	strcat(rates_name, ".RATES");
	if ((rates_file = fopen(rates_name, "w")) == NULL)
	{	
                printf("ERROR: could not open %s for writing\n", rates_name);
		exit(1);
	}
	strcat(rg_name, ".RG");
	if ((rg_file = fopen(rg_name, "w")) == NULL)
	{	
                printf("ERROR: could not open %s for writing\n", rg_name);
		exit(1);
	}

        /************
        *** start ***
        ************/

	printf("\n");
	printf("\nGENERATING THE REDUCED REACHABILITY GRAPH ");

	if (mem_bit)
		printf("USING MEMORY EFFICIENT ALGORITHM ...\n");
	else	printf("USING TIME EFFICIENT ALGORITHM ...\n");
	fflush(stdout);

        /* get start times for total time */
	cpu_startTimeGraph = getCPUTime_C();
        day_startTimeGraph = getTimeOfDay_C();
        mem_atStartGraph = getUsedMem_C();


        /* read the structural description file */
	read_struct(struct_file);

        /* initialisations */
	new_markings = NULL;
	exp_markings = NULL;
	no_tang_markings = 0;
	no_van_submarks = (int *)emalloc(no_subnets * sizeof(int));

	for (i=0; i<no_subnets; i++)
		no_van_submarks[i] = 0;

	r_graph = NULL;
	det_markings = (tang_marking **)
                          emalloc(no_det_trans * sizeof(tang_marking *), 9);
	no_nodes = (int *)emalloc(no_det_trans * sizeof(int));

	for(i = 0; i < no_det_trans; i++)
	{	det_markings[i] = NULL;
		no_nodes[i] = 0;
	}


        /* generation of the reachability graph */
	generate_rg(subnets, &timed);


	/* show some informations */
	print_info(net_name);

#       ifdef GRAPHONLY
	    show_graph_tree(subnets, &timed);
            printf("\n");
	    show_rusage();
	    printf("Obtaining RG only; stop.\n");
	    exit(0);
#       endif

        /* get time and memory usage ***/
	cpu_startTimeCSMC = getCPUTime_C();
	day_startTimeCSMC = getTimeOfDay_C();
	mem_atStartCSMC = getUsedMem_C();

	/* show statistics about generate RG */
	printf("\n");
	printf("generate RG : Elapsed CPU time  : ");
	printTimeDiff_C(cpu_startTimeCSMC, cpu_startTimeGraph);
	printf("generate RG : Elapsed real time : ");
	printTimeDiff_C(day_startTimeCSMC, day_startTimeGraph);
	printf("generate RG : Required memory : ");
	printMemDiff_C(mem_atStartCSMC, mem_atStartGraph);
	printf("\n");

#       ifdef SHOW_SPECIAL_FORMAT
	    printTimeDiffSec_C(cpu_startTimeCSMC, cpu_startTimeGraph);
	    printTimeDiffSec_C(day_startTimeCSMC, day_startTimeGraph);
	    printf("%ld", mem_atStartCSMC - mem_atStartGraph);
	    printf("\n");
#       endif


        /* calculate CSMCs */
        /* test of special structurs in case of explicit solution method */
	csmc_main(net_name, argv[SOL_METHOD]);


        /* get start times for solution method */
	cpu_startTimeSol = getCPUTime_C();
	day_startTimeSol = getTimeOfDay_C();
        mem_atStartSol = getUsedMem_C();

	/* write intermediate files, if -f option is set (debugging etc) */
	if (write_files)
		write_tmpfiles(net_name);

	/* obtain informations about rand efforts */
	obtain_randinfos();

#       ifdef DEBUG_RG_RESORTED
	   show_graph_tree(subnets, &timed);
#       endif


	/************************************
	*** call selected solution method ***
        ************************************/

        if ( !(strcmp(argv[SOL_METHOD], "FI_avoidance")) )
        {
           /* show statistics about calculate csmcs */
           printf("\n");
           printf("calculate CSMCs : Elapsed CPU time  : ");
	   printTimeDiff_C(cpu_startTimeSol, cpu_startTimeCSMC);
           printf("calculate CSMCs : Elapsed real time : ");
	   printTimeDiff_C(day_startTimeSol, day_startTimeCSMC);
           printf("calculate CSMCs : Required memory : ");
           printMemDiff_C(mem_atStartSol, mem_atStartCSMC);
           printf("\n");

#          ifdef SHOW_SPECIAL_FORMAT
	       printTimeDiffSec_C(cpu_startTimeSol, cpu_startTimeCSMC);
	       printTimeDiffSec_C(day_startTimeSol, day_startTimeCSMC);
	       printf("%ld", mem_atStartSol - mem_atStartCSMC);
	       printf("\n");
#          endif

           /* close files */
	   fclose(rates_file);
	   fclose(struct_file);
	   fclose(rg_file);

           /* save start times */
           saveTime_C(cpu_startTimeGraph, argv[NETNAME], "cpu_start");
           saveTime_C(day_startTimeGraph, argv[NETNAME], "day_start");
           saveTime_C(cpu_startTimeSol, argv[NETNAME], "cpu_startSol");
           saveTime_C(day_startTimeSol, argv[NETNAME], "day_startSol");

           /* save memory at start points */
           saveMem_C(mem_atStartGraph, argv[NETNAME], "atStart");
           saveMem_C(mem_atStartSol, argv[NETNAME], "atStartSol");


           /* start iterative solution method with avoiding fill-in */
	   solve_main(argv[NETNAME], argv[EXEC_MODE],
                      arb_bits, trunc_error,
	  	      argv[INIT_VEC], seed_val, argv[SAVE_VEC],
                      max_iter, epsilon);

	}
        else
	{
           void   *cpu_endTimeGraph;  // end time of graph process
           void   *day_endTimeGraph;  // end time of graph process
           void   *cpu_TimeRand;      // time for execute rand processes
           void   *cpu_TimeGraph;     // time for graph process

           long    mem_atEndGraph;    // used memory at end of graph process
	   long    mem_diff;          // difference of used memory

           /* show statistics about calc. csmcs, iso. and top. tests */
           printf("\n");
           printf("calc. CSMCs, iso. and top. tests : Elapsed CPU time  : ");
	   printTimeDiff_C(cpu_startTimeSol, cpu_startTimeCSMC);
           printf("calc. CSMCs, iso. and top. tests : Elapsed real time : ");
	   printTimeDiff_C(day_startTimeSol, day_startTimeCSMC);
           printf("calc. CSMCs, iso. and top. tests : Required memory : ");
           printMemDiff_C(mem_atStartSol, mem_atStartCSMC);
           printf("\n");

#          ifdef SHOW_SPECIAL_FORMAT
	       printTimeDiffSec_C(cpu_startTimeSol, cpu_startTimeCSMC);
	       printTimeDiffSec_C(day_startTimeSol, day_startTimeCSMC);
	       printf("%ld", mem_atStartSol - mem_atStartCSMC);
	       printf("\n");
#          endif


           /* start first part of EMC explicit solution method */
	   if (no_normal_classes)
	      execute_rand_processes(argv[NETNAME], epsilon,
                                     argv[EXEC_MODE]);

           if (no_gen_trans)      
              execute_erand_processes(argv[NETNAME], epsilon,
                                      arb_bits, trunc_error);


           /* get time and memory usage ***/
           cpu_endTimeGraph = getCPUTime_C();
           day_endTimeGraph = getTimeOfDay_C();
           mem_atEndGraph = getUsedMem_C();

           /* show statistics about first part of EMC explicit sol. method */
           printf("\n");
           printf("Execute rand processes : Elapsed CPU time  : ");
	   printTimeDiff_C(cpu_endTimeGraph, cpu_startTimeSol);
           printf("Execute rand processes : Elapsed real time : ");
	   printTimeDiff_C(day_endTimeGraph, day_startTimeSol);
           printf("Execute rand processes : Required memory : ");
           printMemDiff_C(mem_atEndGraph, mem_atStartSol);
           printf("\n");


#          ifdef SHOW_SPECIAL_FORMAT
	       printTimeDiffSec_C(cpu_endTimeGraph, cpu_startTimeSol);
	       printTimeDiffSec_C(day_endTimeGraph, day_startTimeSol);
	       printf("%ld", mem_atEndGraph - mem_atStartSol);
	       printf("\n");
#          endif

           /* save elapsed times */
           cpu_TimeGraph = subTimes_C(cpu_endTimeGraph, cpu_startTimeGraph);
           saveTime_C(cpu_TimeGraph, argv[NETNAME], "cpu_Graph");
           saveTime_C(day_startTimeGraph, argv[NETNAME], "day_start");

           cpu_TimeRand = subTimes_C(cpu_endTimeGraph, cpu_startTimeSol);
           saveTime_C(cpu_TimeRand, argv[NETNAME], "cpu_Rand");
           saveTime_C(day_startTimeSol, argv[NETNAME], "day_startSol");

           /* save required memory */
           mem_diff = mem_atEndGraph - mem_atStartGraph;
           saveMem_C(mem_diff, argv[NETNAME], "Graph");
           mem_diff = mem_atEndGraph - mem_atStartSol;
           saveMem_C(mem_diff, argv[NETNAME], "Rand");

           /* delete time pointers */
	   freeTime_C(cpu_endTimeGraph);
	   freeTime_C(day_endTimeGraph);
	   freeTime_C(cpu_TimeRand);
	   freeTime_C(cpu_TimeGraph);

           /* close files */
	   fclose(rates_file);
	   fclose(struct_file);
	   fclose(rg_file);
	}

        /* delete time pointers */
	freeTime_C(cpu_startTimeGraph);
	freeTime_C(day_startTimeGraph);
	freeTime_C(cpu_startTimeCSMC);
	freeTime_C(day_startTimeCSMC);
	freeTime_C(cpu_startTimeSol);
	freeTime_C(day_startTimeSol);


#       ifdef MEMDEBUG
	{	sort_type sortspace[NO_ACCOUNTS];
		int p;
		printf("\nGraph MEMDEBUG information:\n");
		printf("Maximum memory space allocated: %d\n", max_sum);

		for (i = 0; i < NO_ACCOUNTS; i++)
		{	sortspace[i].num = i;
			sortspace[i].val = max[i];
		}
		qsort((char *)(sortspace), NO_ACCOUNTS, sizeof(sort_type),
                                           comp);

		printf(" Account     Space   Percent\n");

		for (i = 0; (i < NO_ACCOUNTS) && sortspace[i].val; i++)
		{	p = (int)((100.0 * sortspace[i].val) / max_sum);
			if (!p)
				break;
			printf("%8d  %8d  %6d%%\n", sortspace[i].num,
                                                    sortspace[i].val, p);
		}
	}
#       endif
	exit(0);
}
