/*****************************************************************************

	Programm	:	graph

	Komponente	:	graph_types.h
				Typdefinitionen und Konstanten

	Programmierer	:	Armin Zimmermann

	Datum		:	Juni 1992

*****************************************************************************/

#define	ENVIRONMENT	".TimeNET"
					/* Name of the environment file in home-dir */

#define	ARRAY_LIMIT	10		/* limit for bounds to fit in an array */
#define	SPARSE_START	1		/* initial range for sparse array */
#define	SPARSE_NEXT(a)	(2 * a)		/* expanding function for sparse array */
#define SPARSE_MAX	65535		/* maximal range */
#define FILE_NAME	200		/* maximal length of a filename */

#ifndef MAX
#  define MAX(a, b)	(a > b ? a : b)	/* maximum macro */
#endif
#ifndef MIN
#  define MIN(a, b)	(a > b ? b : a)	/* minimum macro */
#endif

typedef unsigned char	byte;		/* 1 byte */
typedef unsigned word;		        /* 4 bytes  aus short word wurde word*/

/* subnet reachability graph types */

typedef struct fire_path		/* immediate firing-path desciption */
{	double	prob;			/* probability of this firing path */
	word	*marking;		/* the last marking of the path */
}  fire_path;

typedef struct sub_node			/* immediate reachability set's tree item */
{	word	bound;			/* array: = 0; array[0..bound(place)] */
					/* sparse: > 0; array[0..bound] */
	word	number;			/* actual number of elements in sparse array */
	struct sub_node **next;		/* array of pointers to next level of tree*/
					/* (sparse or array) */
}  sub_node;

typedef struct sp_pointer		/* array element type for sparse arrays */
{	word	index;			/* index of element */
	sub_node *next;			/* next - level element in tree */
}  sp_pointer;

typedef struct sub_marking		/* (vanishing) submarking */
{	int	no_paths;		/* number of firing-paths to tangible markings */
	fire_path *paths;		/* pointer to the firing-paths (NULL = new marking) */
}  sub_marking;

typedef struct mem_sub_marking		/* (vanishing) submarking in memory - save mode */
{	double	prob;			/* probability of the firing path to this marking */
	word	*mark;			/* packed marking  */
	struct mem_sub_marking *forward, *backward;
					/* twice concatenated list pointer */
	int	no_pred;		/* number of direct predecessor markings */
}  mem_sub_marking;

typedef struct end_mind			/* mind end marks of subnets */
{	struct	end_mind *next;		/* linear list */
	word	*mark;			/* coded submarking */
}  end_mind_t;

/* tangible reachability graph types */

typedef struct tsp_pointer		/* array element type for sparse arrays */
{	word	*index;			/* index of element (pointer to coded marking) */
	void	*next;			/* next - level element in tree */
}  tsp_pointer;

typedef struct tang_node		/* tangible reachability set's tree item */
{	word	bound;			/* actual number of allocated elements */
	word	number;			/* actual number of elements in sparse array */
	tsp_pointer *next;		/* array of tsp_pointer to next level of tree*/
}  tang_node;

typedef	struct tang_path		/* tangible-to-tangible firing path */
{	double	prob;			/* path probability */
	void	*marking;		/* reached marking */
}  tang_path;

typedef struct tang_marking		/* tangible marking */
{	double	sum_of_rates;		/* sum of all exp. firing rates */
	int	number;			/* the marking's number */
	word	*disc;			/* disconnected places submarking */
	word	**mark;			/* submarking of each subnet */
	byte	*fire_nums;		/* numbers of all enabled transitions */
	byte	fire_tra_num;		/* number of fireable transitions */
	byte	det;			/* Bit 0: a deterministic transition is fireable */
					/* Bit 1: the marking is initial in MDET */
	byte	det_num;		/* number of fireable det. trans. (if any) */
	int	no_det_paths;		/* number of det. paths */
	tang_path *det_paths;		/* markings reachable after firing of a det. trans */
	int	no_exp_paths;		/* number of exp. paths */
	tang_path *exp_paths;		/* markings reachable after firing of an exp. trans */
	struct tang_marking *next;	/* next list element */
}  tang_marking;

/* net - structure types */

typedef struct sparse_place		/* element type for sparse matrices */
{	int	num;			/* number of place in row */
	int	value;			/* value */
	int	plu_id, min_id;		/* id's for marking dependent arcs */
}  sp_place;

typedef struct place			/* type for places */
{	char	name[30];		/* place name */
	int	number;			/* global number of the place */
	int	m0;			/* initial marking */
	int	bound;			/* upper bound for number of tokens in place */
	int	bits;			/* number of bits needed for this place */
}  place;

typedef struct timed_t			/* description of a timed transition */
{	char	name[30];		/* name */
	int	number;			/* global identification number */
	int	type;			/* deterministic(1) / exponential(0) */
	int	det_no;			/* number of det. transition */
	int	infinite;		/* infinite server */
	int	mark_ind;		/* marking-dependent firing rate: index */
	double	value;			/* firing rate / time */
	sp_place *i_matrix;		/* input matrix of the transition */
	sp_place *h_matrix;		/* inhibitor matrix of the transition */
	sp_place *c_matrix;		/* o - i matrix of the transition */
					/* place-numbers in the matrices are global */
}  timed_t;

typedef struct timed_sub
{	int	no_trans;		/* number of timed transitions */
	timed_t	*transitions;		/* description of transitions */
	int	no_places;		/* number of disconnected places */
	place	*places;		/* description of places */
	int	memory;			/* memory space required for one marking */ 
	sub_node *tree;			/* disconnected places marking tree */
	int	no_markings;		/* number of disc. places markings */
}	t_sub;

typedef struct imm_t			/* description of an immediate transition */
{	char	name[30];		/* name */
	double	prob;			/* firing probability */
	int	wgt_id;			/* id for marking dependent firing weight */
	int	ena_id;			/* id for marking dependent enabling function */
	sp_place *i_matrix;		/* input matrix of the transition */
	sp_place *h_matrix;		/* inhibitor matrix of the transition */
	sp_place *c_matrix;		/* o - i matrix of the transition */
					/* place-numbers in the matrices are local */
	byte	*enas;			/* enabling set */
	byte	*diss;			/* disabling set */
}  imm_t;
	
typedef struct subnet			/* immediate transition subnet description */
{	int	no_trans;		/* number of transitions */
	imm_t	*transitions;		/* description of transitions */
	int	no_places;		/* number of places */
	place	*places;		/* description of places */
	int	memory;			/* memory space required for one marking (# words) */ 
	int	no_ecs;			/* number of ecs */
	byte	**ecs;			/* set of ecs */
	sub_node *tree;			/* subnet marking tree */
	int	no_markings;		/* number of submarkings */
	end_mind_t *end_mind;			/* list of end marks */
	mem_sub_marking	*list1, *list2, *e_list;
					/* list pointers for memory save option */
}	subnet;

/* types for files */

typedef struct mexp_type
{	double		value;
	unsigned	number1;
	unsigned	number2;
}  mexp_type;

typedef struct delta_type
{	double		value;
	unsigned	number;
}  delta_type;

typedef struct mem_list
{       struct mem_list *next;
        struct mem_list *prev;
}  mem_list_t;
