/******************************************************************************
 Program:           mem.c

 Purpose:           Provide a fault-checking memory allocation interface.
                    Also to handle realloc's with a NULL pointer passed.

 Project Director:  Reinhard German

 Programmer:        Ulrich Hertlein

 History:           Release 1: May 18, 1993
                    Release 2: Patrick Drescher, June 1995
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#include "mem.h"


/*
 * Function:        Malloc
 *                  allocates a chunk of memory; fails if there is no more
 *                  virtual memory available
 *
 * Arguments:       size of the memory chunk
 *
 * Result:          pointer to a block of at lease size bytes
 */
void *Malloc(size)
unsigned size;
{
   void *ptr;

   if( (ptr = (void *)malloc(size)) == (void *)NULL )
   {
      fprintf(stderr,"Malloc: failed for %d byte\n", size);
      exit(1);
   }

   return( ptr );
}


/*
 * Function:        Calloc
 *                  allocates a number of chunks of memory of a given size;
 *                  fails if there is no more memory available
 *
 * Arguments:       number of chunks to be allocated
 *                  size of each chunk
 *
 * Result:          pointer to a chunk of memory
 */
void *Calloc(nelem, size)
unsigned nelem;
unsigned size;
{
   void *ptr;

   if( (ptr = (void *)calloc(nelem,size)) == (void *)NULL )
   {
      fprintf(stderr,"Calloc: failed for %d chunks of %d byte\n", nelem,size);
      exit(1);
   }

   return( ptr );
}


/*
 * Function:        Realloc
 *                  enlarges an existing chunk to a given size; if the original
 *                  chunk is zero, it allocates a new chunk of the desired
 *                  size
 *
 * Arguments:       pointer to the old memory block
 *                  size of the new block in bytes
 *
 * Result:          pointer to the new memory block
 */
void *Realloc(ptr, size)
void *ptr;
unsigned size;
{
   if( ptr == (void *)NULL )
      ptr = Malloc(size);
   else
      if( (ptr = (void *)realloc((char *)ptr,size)) == (void *)NULL )
      {
         fprintf(stderr,"Realloc: failed for %d byte\n", size);
         exit(1);
      }

   return( ptr );
}
