/******************************************************************************

Programm	:	rand

Komponente	:	rand_procs.c

Aufgabe		:	Prozeduren fuer die Verarbeitung von duenn besetzten
			Vektoren und Matritzen
			Prozeduren fuer die Berechnung von Poisson - Wahrscheinlichkeiten

Projektleiter	:	Christoph Lindemann TU Berlin

Programmierer	:	Armin Zimmermann

Datum		:	Juni 1993

******************************************************************************/

#include <stdio.h>		/* for NULL */
#include <math.h>
#include "rand_types.h"
#include "portab.h"


#define	COMP_DOUBLE_EPS	1e-10

extern char	*emalloc();
extern void	efree();



/* compare two doubles with predefined epsilon */
int
d_comp(d1, d2)
	double	d1, d2;
{
	double	diff = (d1 - d2);

	if ((diff < COMP_DOUBLE_EPS) && (diff > -COMP_DOUBLE_EPS))
		return 1;
	else	return 0;
}

/* get the global number of a csmc state or preempted state */
int
global_state_num(local_num, i_csmc)
	int		local_num;
	iso_csmc	*i_csmc;
{
	register int	num;

	if (local_num >= no_total_states)
	{	fprintf(stderr, "local number exceeded.\n");
		my_exit(1);
	}
	num = local_num >= no_csmc_states ? i_csmc->orig_no[local_num - no_csmc_states]
			: no_csmcs * local_num + i_csmc->csmc_no + first_state_no;
	return num;
}

/* get the number of states in one level of the branching tree */
int
no_states_in_level(n, k)
	int n, k;
{
	int	i;
	double	val = 1.0;

	for (i = 1; i <= k; i++)
		val *= (double)(n+i-1) / (double)i;
	return (int)val;
}

/* copy results of branch solution via a depth first search */
double	*new_C_row = NULL, *new_P_row = NULL;
double	*old_C_row, *old_P_row;
void
dfs_copy_branch(old, new)
	int		old, new;
{
	int	e_no;

	new_C_row[new] = old_C_row[old];
	new_P_row[new] = old_P_row[old];
	visited[old] = 1;
	if (branch_level[old] >= branch_k-1)
		return;
	scan_array(e_no, branch_n)
		if (!visited[state_succs[old][e_no]])
			dfs_copy_branch(state_succs[old][e_no], state_succs[new][e_no]);
}

/* map the results of branch solutions */
void
map_branch(act_tau_set, old_state_no, new_state_no)
	tau_set		*act_tau_set;
	int		old_state_no, new_state_no;
{
	int		state_num;
	sp_entry	*p_walk = NULL, *c_walk = NULL;

	/* initializations */
	if (new_C_row == NULL)
	{	new_C_row = (double *)emalloc(no_csmc_states * sizeof(double));
		new_P_row = (double *)emalloc(no_csmc_states * sizeof(double));
	}
	memset(new_C_row,0, no_csmc_states * sizeof(double));
	memset(new_P_row,0, no_csmc_states * sizeof(double));
	memset(visited,0, no_csmc_states * sizeof(char));
	old_C_row = act_tau_set->C_row_copy;
	old_P_row = act_tau_set->P_row_copy;

	/* depth first search and copy values */
	dfs_copy_branch(old_state_no, new_state_no);

	/* get the new vectors */
	scan_array(state_num, no_csmc_states)
	{	if (!is_zero(new_C_row[state_num]))
		{	if (c_walk == NULL)
				act_tau_set->C_row->entries = c_walk = new_entry();
			else	c_walk = c_walk->next = new_entry();
			c_walk->num = state_num;
			c_walk->val = new_C_row[state_num];
		}
		if (!is_zero(new_P_row[state_num]))
		{	if (p_walk == NULL)
				act_tau_set->P_row->entries = p_walk = new_entry();
			else	p_walk = p_walk->next = new_entry();
			p_walk->num = state_num;
			p_walk->val = new_P_row[state_num];
		}
	}
}

/* vector / matrix entry operations */

static sp_entry *first_free_item = NULL;

/* get a new entry  (tobias bading) */
sp_entry *
new_entry()
{
	static sp_entry		**buffers;
	static unsigned		number_of_buffers, last_used_buffer;

	if (!first_free_item)
	{	number_of_buffers = MEM_BUFFERS;
		buffers = (sp_entry **)emalloc (MEM_BUFFERS * sizeof (sp_entry *));
		buffers[0] = (sp_entry *)calloc (MEM_ITEMS, sizeof (sp_entry));
		last_used_buffer = 0;
		first_free_item = buffers[0] + 1;
		buffers[0]->next = NULL;
		return buffers[0];
	}
	{	register sp_entry *returnvalue = first_free_item;

		if (first_free_item->next)
			first_free_item = first_free_item->next;
		else
			if (++first_free_item - buffers[last_used_buffer] == MEM_ITEMS)
			{	if (++last_used_buffer == number_of_buffers)
				{	number_of_buffers += MEM_BUFFERS;
					buffers = (sp_entry **)realloc (buffers, 
							number_of_buffers * sizeof (sp_entry *));
				}
				buffers[last_used_buffer] = 
					(sp_entry *)calloc(MEM_ITEMS, sizeof (sp_entry));
				first_free_item = buffers[last_used_buffer];
			}
		returnvalue->next = NULL;
		return returnvalue;
	}
}

/* delete entry */
#define del_entry(item) { \
  item->next = first_free_item; \
  first_free_item = item; \
}

/* vector operations */

/* get a new vector */
sp_vector *
new_vector()
{
	sp_vector		*new_vec;

	new_vec = (sp_vector *)emalloc(sizeof(sp_vector));
	new_vec->num = 0;
	new_vec->no_entries = 0;
	new_vec->entries = NULL;
	new_vec->next = NULL;
	return new_vec;
}

/* clear vector */
void
clear_vector(vec)
	sp_vector	*vec;
{
	sp_entry	*e_walk, *h_walk;

	if ((vec == NULL) || !vec->entries)
		return;
	for (e_walk = vec->entries; e_walk->next != NULL; e_walk = e_walk->next);
	e_walk->next = first_free_item;
	first_free_item = vec->entries;
	vec->entries = NULL;
	vec->no_entries = 0;
}

/* delete vector */
void
del_vector(vec)
	sp_vector	**vec;
{
	clear_vector(*vec);
	efree(*vec);
	*vec = NULL;
}

/* quick adding of values to a vector */
/* written by Tobias Bading */
sp_entry *
add_to_vector(vector, prev_item, index, value)
	sp_vector	*vector;
	register sp_entry *prev_item;
	register int	index;
	double		value;
{
	register sp_entry *prev_entry = NULL;
	register sp_entry *entry = vector->entries;

	if (is_zero(value))
		return prev_item;

	if (prev_item)
		entry = prev_item->next;

	while (entry)
		if (entry->num < index)
		{	prev_entry = entry;
			entry = entry->next;
		}
		else
			if (entry->num == index)
			{	entry->val += value;
				return entry;
			}
			else
				break;

	{	register sp_entry *new_e = new_entry();

		vector->no_entries++;
		new_e->num = index;
		new_e->val = value;
		new_e->next = entry;
		if (prev_entry)
			prev_entry->next = new_e;
		else
			if (prev_item)
				prev_item->next = new_e;
			else
				vector->entries = new_e;
		return new_e;
	}
}

/* copy vector */
sp_vector *
copy_vector(vec)
	sp_vector	*vec;
{
	sp_vector	*new_vec = new_vector();
	sp_entry	*entry_walk, *n_entry;
	sp_entry	*copy_walk = NULL;

	new_vec->no_entries = vec->no_entries;
	new_vec->num = vec->num;
	scan_list(entry_walk, vec->entries)
	{	n_entry = new_entry();
		n_entry->num = entry_walk->num;
		n_entry->val = entry_walk->val;
		if (copy_walk == NULL)
			new_vec->entries = n_entry;
		else	copy_walk->next = n_entry;
		copy_walk = n_entry;
	}
	return new_vec;
}

/* compare two vectors */
int
comp_vector(vec1, vec2)
	sp_vector	*vec1, *vec2;
{
	sp_entry	*e_walk_1 = vec1->entries;
	sp_entry	*e_walk_2 = vec2->entries;

	while((e_walk_1 != NULL) || (e_walk_2 != NULL))
	{	if ((e_walk_1 == NULL) || 
                    ((e_walk_1 && e_walk_2)?(e_walk_2->num < e_walk_1->num):0))
		{	if (!d_comp(e_walk_2->val, 0.0))
				return 0;
			e_walk_2 = e_walk_2->next;
		} else
		if ((e_walk_2 == NULL) || 
                    ((e_walk_1 && e_walk_2)?(e_walk_2->num > e_walk_1->num):0))
		{	if (!d_comp(e_walk_1->val, 0.0))
				return 0;
			e_walk_1 = e_walk_1->next;
		} else
		{	/* e_walk_2->num == e_walk_1->num */
			if (!d_comp(e_walk_2->val, e_walk_1->val))
				return 0;
			e_walk_1 = e_walk_1->next;
			e_walk_2 = e_walk_2->next;
		}
	}
	return 1;
}

/* matrix operations */

/* get a new matrix */
sp_matrix *
new_matrix()
{
	sp_matrix	*new_mat;

	new_mat = (sp_matrix *)emalloc(sizeof(sp_matrix));
	new_mat->rows = NULL;
	return new_mat;
}

/* clear matrix */
void
clear_matrix(mat)
	sp_matrix	*mat;
{
	sp_vector	*v_walk, *h_walk;

	if (mat == NULL)
		return;
	for (v_walk = mat->rows; v_walk != NULL;)
	{	h_walk = v_walk->next;
		del_vector(&v_walk);
		v_walk = h_walk;
	}
	mat->rows = NULL;
}

/* delete matrix */
void
del_matrix(mat)
	sp_matrix	**mat;
{
	clear_matrix(*mat);
	efree(*mat);
	*mat = NULL;
}

/* add an entry to a matrix; if the entry already exist, add the values */
void
add_entry_to_matrix(mat, row_nr, col_nr, value)
	sp_matrix	*mat;
	int		row_nr, col_nr;
	double		value;
{
	sp_vector	*vec_walk, *h_vec_walk = NULL, *new_vec;
	static sp_vector *old_vec;
	static sp_matrix *old_mat;

	if (is_zero(value) || (mat == NULL))
		return;

	if ((old_mat != mat) || !old_vec || (old_vec->num > row_nr))
		old_vec = mat->rows;
	old_mat = mat;

	for (vec_walk = old_vec; ; vec_walk = vec_walk->next)
	{	if ((vec_walk == NULL) || (vec_walk->num > row_nr))
		{	new_vec = new_vector();
			new_vec->num = row_nr;
			new_vec->no_entries = 1;
			new_vec->next = vec_walk;
			if (h_vec_walk == NULL)
				mat->rows = new_vec;
			else	h_vec_walk->next = new_vec;
			new_vec->entries = new_entry();
			new_vec->entries->num = col_nr;
			new_vec->entries->val = value;
			old_vec = new_vec;
			return;
		}
		if (vec_walk->num == row_nr)
		{	add_to_vector(vec_walk, NULL, col_nr, value);
			old_vec = vec_walk;
			return;
		}
		h_vec_walk = vec_walk;
	}
}

/* delete rows from a matrix */
void
del_rows_from_matrix(mat, row_num_to)
	sp_matrix	*mat;
	int		row_num_to;
{
	sp_vector	*v_walk, *hv_walk;
	sp_entry	*e_walk;
	int		bit;

	if (mat == NULL)
		return;
	hv_walk = NULL;
	scan_list(v_walk, mat->rows)
	{	if (v_walk->num <= row_num_to)
		{	efree(v_walk);
			mat->rows = v_walk->next;
			continue;
		}
		bit = 1;
		while (e_walk = v_walk->entries)
		{	if (e_walk->num > row_num_to)
				break;
			v_walk->entries = e_walk->next;
			del_entry(e_walk);
			if (!--v_walk->no_entries)
			{	if (hv_walk == NULL)
					mat->rows = v_walk->next;
				else	hv_walk->next = v_walk->next;
				efree(v_walk);
				bit = 0;
				break;
			}
		}
		if (bit)
			hv_walk = v_walk;
	}
}

/* copy matrix */
sp_matrix *
copy_matrix(mat)
	sp_matrix	*mat;
{
	sp_matrix	*new_mat = new_matrix();
	sp_vector	*new_vec, *v_walk;
	sp_vector	*copy_walk = NULL;

	scan_list(v_walk, mat->rows)
	{	new_vec = copy_vector(v_walk);
		if (copy_walk == NULL)
			new_mat->rows = new_vec;
		else	copy_walk->next = new_vec;
		copy_walk = new_vec;
	}
	return new_mat;
}

/* calculations */

/* vector1 += vector2 * factor */
void
vector_add_mult(vec1, vec2, factor)
	sp_vector	*vec1, *vec2;
	double		factor;
{
	sp_entry	*entry1 = vec1->entries;
	sp_entry	*hentry1 = NULL, *hentry;
	sp_entry	*entry2 = vec2->entries;
	int		num1, num2;

	while((entry1 != NULL) || (entry2 != NULL))
	{	num1 = entry1 == NULL ? 10000000 : entry1->num;
		num2 = entry2 == NULL ? 10000000 : entry2->num;
		if (num1 < num2)
		{	hentry1 = entry1;
			entry1 = entry1->next;
		}
		else if (num1 == num2)
		{	entry1->val += entry2->val * factor;
			hentry1 = entry1;
			entry1 = entry1->next;
			entry2 = entry2->next;
		}
		else /* num1 > num2 */
		{	hentry = new_entry();
			hentry->num = num2;
			hentry->val = entry2->val * factor;
			hentry->next = entry1;
			if (hentry1 == NULL)
				vec1->entries = hentry;
			else	hentry1->next = hentry;
			vec1->no_entries++;
			hentry1 = hentry;
			entry2 = entry2->next;
		}
	}
}

/* multiply a vector with a matrix */
/* vec = vec * mat */
void
mult_vector_matrix(vec, mat)
	sp_vector	*vec;
	sp_matrix	*mat;
{
	sp_vector	*pre_vec = new_vector();
	sp_vector	*vec_walk;
	register sp_entry *vec_e_walk;

	memmove(pre_vec,vec, sizeof(sp_vector));
	vec->no_entries = 0;
	vec->entries = NULL;
	vec_walk = mat->rows;

	/* efficient calculation of leaf states */
	if (mat == A_matrix)
	{	register sp_entry *prev_entry = NULL;

		scan_list(vec_e_walk, pre_vec->entries)
		    if (leaf_state[vec_e_walk->num])
		    {	if (prev_entry)
				prev_entry = prev_entry->next = new_entry();
			else	vec->entries = prev_entry = new_entry();
			memmove(prev_entry, vec_e_walk, sizeof(sp_entry));
			vec->no_entries++;
		    }
		if (prev_entry)
		    prev_entry->next = NULL;
	}

	scan_list(vec_e_walk, pre_vec->entries)
	{	register double	phi_factor = vec_e_walk->val;
		register sp_entry *prev_entry = NULL, *mat_e_walk;

		/* special algorithm for A matrix */
		if (mat == A_matrix)
		{	register unsigned walk_no;

			if (leaf_state[vec_e_walk->num])
				continue;

			vec_walk = A_matrix_rows[vec_e_walk->num];
			if (!vec_walk)
				continue;
		} else
		{	while (vec_e_walk && vec_walk && (vec_walk->num < vec_e_walk->num))
				vec_walk = vec_walk->next;
			if (!vec_walk || !vec_e_walk)
				break;
			if (vec_walk->num > vec_e_walk->num)
				continue;
		}
		scan_list(mat_e_walk, vec_walk->entries)
			prev_entry = add_to_vector(vec, prev_entry, mat_e_walk->num, 
					mat_e_walk->val * phi_factor);
	}
	del_vector(&pre_vec);
}

/* for tests */
void show_entry(e)
	sp_entry *e;
{
	printf("[%d] = %lf\n", e->num, e->val);
}

void show_vector(v)
	sp_vector *v;
{
	sp_entry *e;

	printf("vector [%d] %d entries\n", v->num, v->no_entries);
	scan_list(e, v->entries)
		show_entry(e);
}

void show_matrix(m)
	sp_matrix *m;
{
	sp_vector *v;

	printf("*** Matrix ***\n");
	scan_list(v, m->rows)
		show_vector(v);
}

/* procedures for the efficient and stable calculation of poisson probabilities */

/******************************************************************************

 Funktion:	d

 Parameter:	k und lambda

 Rueckgabewert:	double-Wert d

 Zweck:		Wird von find_r() benoetigt

******************************************************************************/
double d (k, lambda)
double k, lambda;
{
  return 1.0 / (1.0 - exp (-(2.0/9.0) * (k * sqrt (2.0 * lambda) + 3.0/2.0)));
}


/******************************************************************************

 Funktion:	find_r

 Parameter:	lambda, epsilon, Mode m, k_p

 Rueckgabewert:	Der rechte Abschnittspunkt

 Zweck:		Berechnet den rechten Abschnittspunkt

******************************************************************************/
double find_r (lambda, epsilon, m, k_p)
double lambda, epsilon, m, *k_p;
{
	double a_lambda;
	double k;
	double q_k_lambda;
	double R;

	a_lambda = (1.0 + 1.0/lambda) * exp(1.0/16.0) * sqrt(2.0);

	/* corollary 1 of section 4 : k (bzw. R) berechnen */
	for (k = 3.0; /*k < 10.0*/; k++)
	{
		/* q_k_lambda berechnen */
		q_k_lambda = a_lambda * d(k,lambda) * exp(-pow(k,2.0)/2.0) / (k * sqrt(2.0*M_PI));
		/* R berechnen */
		R = ceil(m + k*sqrt(2.0*lambda) + 3.0/2.0);
		if (q_k_lambda < epsilon/2.0)
			break;
	}
	*k_p = k;
	return R;
} /* find_r */


/******************************************************************************

 Funktion:	finder

 Parameter:	lambda --> Wert von lambda
		epsilon --> Gewuenschte Genauigkeit
		tau --> Kleinster darstellbarer Wert
		omega --> Groesster darstellbarer Wert
		m --> Mode von Lambda
		L_p --> Linker Abschnittspunkt
		R_p --> Rechter Abschnittspunkt
		w_m_p --> Der Wert von w(m)
		F_p --> Flag ob Fehler aufgetreten sind

 Rueckgabewert:	---

 Zweck:		Berechnet den linken und rechten Abschnittspunkt und
		den Wert von w(m).

******************************************************************************/
void finder (lambda, epsilon, tau, omega, m, L_p, R_p, w_m_p, F_p)
double lambda, epsilon, tau, omega, m, *L_p, *R_p, *w_m_p;
char *F_p;
{
	double L, R, w_m;
	char F = 1;
	double b_lambda;
	double t_k_lambda;
	double k;
	double c_m;
	double q_dach;
	double p_lambda_k;

	if (lambda <= 0.0)
	{
		L = R = 0.0;
		F = 0;
		goto EXIT;
	}

	R = find_r(lambda, epsilon, m, &k);

	if (lambda < 25.0)
	{
		L = 0.0;
	}
	else /* lambda >= 25.0 */
	{
		/* b_lambda berechnen */
		b_lambda = (1.0 + 1.0/lambda) * exp(1.0/(8.0*lambda));

		/* corollary 2 of section 4 : k (bzw. L) berechnen */
		t_k_lambda = b_lambda * exp(-pow(k,2.0)/2.0) / (k*sqrt(2.0*M_PI));
		if (t_k_lambda >= epsilon/2.0)
		{
			printf ("FINDER: WARNING: t_k_lambda isn't less then epsilon/2 !\n");
		}
		L = floor(m - k*sqrt(lambda) - 3.0/2.0);
		if (L<0.0)
		   L = 0.0;
	}

	/* w(m) berechnen */
	w_m = omega / (1.0e10 * (R - L));

	if (lambda >= 25.0)
	{
		/* corollary 3 of section 5 : p_lambda_k berechnen */
		c_m = (1.0 / sqrt(2.0*M_PI*m)) * exp(m - lambda - 1.0/(12.0*m));
		q_dach = k * sqrt(2.0) + 3.0/(2.0*sqrt(lambda));
		p_lambda_k = c_m * exp(- pow(q_dach + 1.0,2.0) / 2.0);
		if (p_lambda_k * w_m < tau)
		{
			F = 0;
			goto EXIT;
		}

		/* corollary 4 of section 5 : p_lambda_k berechnen */
		q_dach = k + 3.0/(2.0*sqrt(lambda));
		if (q_dach <= sqrt(lambda)/2.0)
		{
			p_lambda_k = c_m * exp( -(pow(q_dach,2.0)/2.0) - 
					       pow(q_dach,3.0)/(3.0*sqrt(lambda)) );
		}
#if 0
		else
		  puts ("ACHTUNG: corollary 4 of section 5 (ii) und (iii) noch nicht implementiert !!");
#endif

		if (p_lambda_k * w_m > tau)
			F = 1;
		else
			F = 0;
	}

EXIT:
	*L_p = L;
	*R_p = R;
	*w_m_p = w_m;
	*F_p = F;
} /* finder */


/******************************************************************************

 Funktion:	weighter

 Parameter:	lambda -> Wert von Lambda
		epsilon --> geforderte Genauigkeit
		tau --> kleinster darstellbarer Wert
		omega --> groesster darstellbarer Wert
		L --> Linker Abschnittspunkt
		R --> Rechter Abschnittspunkt
		w --> Die Gewichtungen
		W --> Summe der Gewichte
		F --> Flag ob Fehler aufgetreten sind

 Rueckgabewert:	---

 Zweck:		Berechnung der Poisson Wahrscheinlichkeiten fuer
		grosse Werte von lambda.

******************************************************************************/
void rweighter (lambda, epsilon, tau, omega, L, R, w, W, F)
	double	lambda;	/* in */
	double	epsilon;
	double	tau;
	double	omega;
	double	*L;	/* out */
	double	*R;
	double	**w;	/* w(L),...,w(R) */
	double	*W;
	char	*F;
{
	double m;
	double w_m;
	double j;
	unsigned i, s, t;
	double q;

/*Initialize:*/
	m = floor(lambda);	/* mode */

#ifdef DEBUG
	printf ("before finder-call: lambda = %lg, epsilon = %lg, tau = %lg, m = %lg\n",
		lambda, epsilon, tau, m);
#endif

	finder (lambda, epsilon, tau, omega, m, L, R, &w_m, F);
	if (*F == 0)
	{
		printf ("FINDER fails !!\n");
		printf ("L=%.0f R=%.0f w_m=%e F=%d\n", *L, *R, w_m, (int)*F);
		return;
	}
	*w = (double *) emalloc ((int)(*R - *L + 1) * sizeof(double));
	(*w)[(int)(m-*L)] = w_m;
/*Down:*/
	for (j = m, i = (unsigned)(j - *L); j > *L; j--, i--)
	{
		(*w)[i-1] = (j / lambda) * (*w)[i];
	}
/*Up:*/
	if (lambda < 400) /* Special */
	{
#ifdef DEBUG
		puts ("Special:");
#endif
		if (*R > 600.0)
		{
			printf ("Special: R > 600 !\n");
			*F = 0;
			return;
		}
		j = m; i = (unsigned)(j - *L);
		while (j < *R)
		{
			q = lambda / (j + 1.0);
			if ((*w)[i] > tau / q)
			{
				(*w)[i+1] = q * (*w)[i];
				j++; i++;
			}
			else
			{
				*R = j;
				goto Compute_W;
			}
		}
	}
	else
	{
		for (j = m, i = (unsigned)(j - *L); j < *R; j++, i++)
		{
			(*w)[i+1] = (lambda / (j + 1)) * (*w)[i];
		}
	}
Compute_W:
	*W = 0.0;
	s = 0;
	t = (unsigned)(*R - *L);
	while (s < t)
	{
		if ((*w)[s] <= (*w)[t])
			*W += (*w)[s++];
		else
			*W += (*w)[t--];
	}
	*W += (*w)[s];
} /* weighter */
