/****************************************************************************
** Evaluation : STATIONARY ANALYSIS                                        **
*****************************************************************************
** Program    : solve                                                      **
** Component  : solve_base                                                 **
** File       : solve_base.h (C++ header)                                  **
**                                                                         **
** Purpose    : definition of the class SolveBase which is used by the     **
**              iterative solution method with fill-in avoidance           **
**                                                                         **
** Master's T.: realis. of an iterative solution method for not marc. SPNs **
**                                                                         **
** Programmer : Oliver Hesse                                               **
**                                                                         **
** Date       : june of 1999                                               **
****************************************************************************/

#ifndef SOLVE_BASE_H
#define SOLVE_BASE_H

/****************************************************************************
** LIST OF IMPORTS                                                         **
****************************************************************************/

#include <string>

#include "solve_gen.h"            // GenMarkingsSmc and GenMarkings
#include "solve_exp.h"            // ExpMarkings
#include "vmmath.h"               // vector and matrics representations
                                  // and operations

/****************************************************************************
** DEFINITIONS OF TYPES                                                    **
****************************************************************************/

class SolveBase                   // base parameters and functions of solve
                                  // process
{
public:

  string        Netname;          // name of the corresponding DSPN
  
  long          ArbBits;          // = 0: using doubles for calc. factors
                                  // > 0: using arb. prec. for calc. factors
  double        TruncError;       // max. truncation error for calc. factors

  string        uSource;          // kind of initial vector
  bool          uSave;            // save option for stat. iteration vector 

  long          SeedValue;        // value to randomize initial vector

  long          MaxIter;          // maximal number of outer iterations
  double        Epsilon;          // prespecified error tolerance

  double        aperiodicFact;    // aperiodic factor for P - matrix
                                  // prevents alterniering behaviour of
                                  // iteration vector u(m)

  GenMarkings   Gen;              // general markings S^G
  ExpMarkings   Exp;              // exponential markings S^E

  int           dim;              // dimension corresp. to 'no_tang_markings'
                                  // vector(dim), matrix(dim, dim)


  DenseVector get_u_init() const; // initialisation of iteration vector u(0)

  DenseVector read_u() const;     // read initial iteration vector u(0)
                                  // from file
  void write_u(const Vector& u) const;
                                  // write stationary iteration vector u(M)
                                  // to file
  void delete_u_file() const;     // delete the file <model>.u if it exists

  void calculate_pi_phi(const DenseVector&  u,
			const DenseVector&  a,
			const DenseVector&  b) const;
                                  // calulation of state probabilities pi
                                  // and firing freqencies phi

  void write_probmark(SparseVector pi) const;
                                  // save state probabilities

  void write_firefreq(SparseVector phi) const;
                                  // save firing frequencies


protected:

  SolveBase(int dim, int no_gen); // constructor definition
};


/****************************************************************************
** CONSTRUCTOR OF 'SOLVE BASE'                                             **
****************************************************************************/

inline SolveBase::SolveBase(int _dim, int no_gen)
  : Gen(no_gen, GenMarkingsSmc(_dim)), Exp(_dim), dim(_dim) {}
   

#endif
