/****************************************************************************
** Evaluation : STATIONARY ANALYSIS                                        **
*****************************************************************************
** Program    : solve                                                      **
** Component  : solve_gen                                                  **
** File       : solve_gen.h (C++ header)                                   **
**                                                                         **
** Purpose    : definition of the structure GenMarkingsSmc (S^g) for the   **
**              SMC of each general transition                             **
**              GenMarkings (S^G) contains the set of all GenMarkinsSmcs   **
**              at regeneration points of the EMC                          **
**              Both are used by the iterative solution method with        **
**              fill-in avoidance                                          **
**                                                                         **
** Master's T.: realis. of an iterative solution method for not marc. SPNs **
**                                                                         **
** Programmer : Oliver Hesse                                               **
**                                                                         **
** Date       : june of 1999                                               **
****************************************************************************/

#ifndef SOLVE_GEN_H
#define SOLVE_GEN_H

/****************************************************************************
** LIST OF IMPORTS                                                         **
****************************************************************************/

#include <string>
#include <vector>      
using namespace std;

#include "vmmath.h"                 // vector and matrics representations
                                    // and operations

/****************************************************************************
** DEFINITIONS OF TYPES                                                    **
****************************************************************************/

struct GenMarkingsSmc               // for a single general trans. 'gen'
{
  string         name;              // name of the transition
  string         pmf_def;           // definition of density function

  FilterMatrix   I;                 // filter matrix
  DiagMatrix     D_tau;
  SparseMatrix   Q;                 // generator matrix
  SparseMatrix   Q_bar;             // generator matrix of preemptions
  SparseMatrix   A;                 // normalisized generator matrix
  double         q;                 // maximal diagonal value of Q
  double         effort;            // estimated effort (q * eta)
  SparseMatrix   DELTA;             // branching probability matrix

  vector<double> alpha_f;           // vector of factors with transient probs. 
  vector<double> alpha_F_bar;       // vector of factors with sojourn times
  unsigned       R;                 // right truncation point
  unsigned       k_min;

  bool operator< (const GenMarkingsSmc&) const;

  void calculate_q_and_A();         // calculates renewal rate q and
                                    // norm. matrix A

  void calculate_alphas(double TruncErrort, long ArbBits);
                                    // gets expoly distribution and calculates
                                    // alpha_f and alpha_F_bar_factors

  void jensens_method_a_b(const SparseVector& gen_u,
                                SparseVector &gen_a, SparseVector &gen_b);
                                    // jensen's method
                                    // calculates gen_a(m) and gen_b(m)

  void jensens_method_a(const SparseVector& gen_u,
	                      SparseVector &gen_a);
                                    // jensen's method calculates gen_a(m)

#if 0
  void jensens_method_dynamic(const SparseVector& gen_u, double Epsilon,
                                    SparseVector &gen_a, SparseVector &gen_b);

  void jensens_method_orig(const SparseVector& gen_u, double Epsilon,
                                 SparseVector &gen_a, SparseVector &gen_b);
#endif

  int getDim() const;               // dimension of the matrices and vectors

  void estimateEffort();            // call this before using operator<
  GenMarkingsSmc(int dim);          // constructor definition

#ifdef PVM_SUPPORT
  GenMarkingsSmc(PVM_UNPACK_t);
  void pvm_pack();
  static GenMarkingsSmc pvm_unpack();
#endif
};



typedef vector<GenMarkingsSmc> GenMarkings;
                                    // 'Gen' is a set of 'gen's        

#endif
