//#define DEBUG_CONV
/****************************************************************************
** Evaluation : STATIONARY ANALYSIS                                        **
*****************************************************************************
** Program    : solve                                                      **
** Component  : solve main                                                 **
** File       : solve_main.cpp (C++ source)                                **
**                                                                         **
** Purpose    : main component of the iterative solution method with        **
**              fill-in avoidance                                          **
**              convertion of pointer structures of 'C' to new             **
**              representations of 'C++' for exponential, general          **
**              (and deterministic) markings                               **
**              calling of sequential or parallel solve process            **
**                                                                         **
** Master's T.: realis. of an iterative solution method for not marc. SPNs **
**                                                                         **
** Programmer : Oliver Hesse                                               **
**                                                                         **
** Date       : june of 1999                                               **
****************************************************************************/

/****************************************************************************
** DEFINITIONS                                                             **
****************************************************************************/

//#define DEBUG_CONV              // show convertions of data

/****************************************************************************
** LIST OF IMPORTS                                                         **
****************************************************************************/

#include <iostream>               // stream outputs
#include <string>                 // string
#include <stdio.h>		  // old C for FILE, remove

#include "solve_gen.h"            // struct. GenMarkingsSmc and GenMarkings
#include "solve_exp.h"            // struct. ExpMarkings
#include "solve_base.h"           // parameters and functions of solveBase

#include "vmmath.h"               // vector and matrics representations
                                  // and operations
#include "Time.h"                 // time and memory usage
using namespace std;


extern "C" {
#include "graph_types.h"
#include "csmc_types.h"
#include "exec_types.h"
#include "parsdfunc.h"            // BUFLEN for parse_pmf file
#include "graph_extern.h"

extern SMC       **ESMCs;         // array of subordinated Markov chains (gen)
extern SMC       **SMCs;          // array of subordinated Markov chains (det)
extern randinfo  *randinfos;      // infos about MCs
extern int       no_smcs;
extern int       no_esmcs;
extern int       max_edges;
extern edge      *sort_edges;

int edgecomp(const void *a, const void *b);
                                  // means int edgecomp(edge *a, edge *b)
int file_read(char *filename, char *tname, char *formbuf);

//void efree_all();
}

/****************************************************************************
** GLOBAL FUNCTIONS                                                        **
****************************************************************************/

extern "C" void solve_main(const char* net_name, const char* exec_mode,
			   const long arb_bits, const double trunc_error,
                           const char* init_vec, const long seed_value,
                           const char* save_vec,
			   const long max_iter, const double epsilon);


extern void solve_sequential(SolveBase&);  
extern void solve_parallel(SolveBase&);


/****************************************************************************
** GLOBAL DEFINITION OF TYPES                                              **
****************************************************************************/

class ConvSolveBase : public SolveBase
{
public:
   /*** constructor declaration ***/
   ConvSolveBase(const char* net_name,
                 const long arb_bits, const double trunc_error,
                 const char* init_vec, const long seed_value,
                 const char* save_vec,
		 const long max_iter, const double epsilon);

private:
   /*** convertion of exponential markings of the EMC ***/
   void convert_exp(tang_marking *exp_markings, ExpMarkings &exp);

   /*** convertion of general markings of SMCs ***/
   void convert_gen(SMC *esmc, GenMarkingsSmc &gen);

   /*** convertion of deterministic markings of SMCs ***/
   void convert_det(Iso_Class *class_p, GenMarkingsSmc &gen);
};


/****************************************************************************
** FUNCTION : constructor definition of ConvertSolveBase                   **
*****************************************************************************
** Purpose  : initialisation of SolveBase                                  **
**            main function to convert pointer structures of 'C' to new    **
**            representations of 'C++' for exponential, general            **
**            (and deterministic) markings                                 **
**            (gen. trans. contains det. trans. and gen. trans.)           **
**                                                                         **
** Input    : net_name       name of the corresponding DSPN                **
**            arb_bits       Bits for arbitrary precision (calc. factors)  **
**            trunc_error    truncation error (calc. factors)              **
**            init_vec       kind of initial iteration vector              **
**            seed_value     seed value for randomisation of init. vector  **
**            save_vec       option if stationary vector shall be saved    **
**            max_iter       maximal number of outer iterations            **
**            epsilon        accuracy of vector equality                   **
**                                                                         **
**            exp_markings   (hidden) 'C' struct of exp. marks. of the EMC **
**            ESMC           (hidden) 'C' struct of general marks. of SMCs **
**            randinfos      (hidden) 'C' struct of determ. marks. of SMCs **
**                                                                         **
** Output   : SB   (hidden)   member variables of SolveBase                **
**            Exp  (hidden)   'C++' structs. for exponential markings      **
**            Gen  (hidden)   'C++' structs. for general markings          **
****************************************************************************/

ConvSolveBase::ConvSolveBase(const char* net_name,
                             const long arb_bits, const double trunc_error,
                             const char* init_vec, const long seed_value,
                             const char* save_vec,
			     const long max_iter, double epsilon)

  : SolveBase(no_tang_markings, no_gen_trans + no_normal_classes)
{
   /*** store parameters to SolveBase ... ***/
   Netname       = net_name;
   ArbBits       = arb_bits;
   TruncError    = trunc_error;
   uSource       = init_vec;
   uSave         = (strcmp(save_vec, "yes")==0);
   SeedValue     = seed_value;
   MaxIter       = max_iter;
   Epsilon       = epsilon;

   /*** store aperiodic factor (0..1) to SolveBase ***/
   aperiodicFact = 0.99;


#  ifdef DEBUG_CONV
      cout << endl;
      cout << "Converting data ..."
           << endl;
      cout << "All:   no_trans          : " << no_trans << endl;
      cout << "Exp: > exp_markings      : " << exp_markings << endl;
      cout << "Gen: > no_gen_trans      : " << no_gen_trans << endl;
      cout << "       no_esmcs          : " << no_esmcs << endl;
      cout << "Det:   no_det_trans      : " << no_det_trans << endl;
      cout << "       no_smcs           : " << no_smcs << endl;
      cout << "     > no_normal_classes : " << no_normal_classes << endl;
      cout << "       no_iso_classes    : " << no_iso_classes << endl;
      cout << endl;
      cout << "Dim: > no_tang_markings  : " << no_tang_markings << endl;
      cout << endl;
#  endif

   /*** convert exponential markings -> Exp ***/
   convert_exp(exp_markings, Exp);

#  ifdef DEBUG_CONV
      cout << "Exp.I:\n" << Exp.I << endl;
      cout << "Exp.P:\n" << Exp.P << endl;
      cout << "Exp.C:\n" << Exp.C << endl;
#  endif

   /*** convert general markings -> Gen ***/
   int g = 0;
   for(int i = 0; i < no_gen_trans; i++, g++)
   {
      convert_gen(ESMCs[i], Gen[g]);

#     ifdef DEBUG_CONV
         cout << "Gen: Gen["<< g <<"].name: "    << Gen[g].name    << endl;
         cout << "Gen: Gen["<< g <<"].pmf_def: " << Gen[g].pmf_def << endl;
         cout << "Gen: Gen["<< g <<"].I:\n"      << Gen[g].I       << endl;
         cout << "Gen: Gen["<< g <<"].Q:\n"      << Gen[g].Q       << endl;
         cout << "Gen: Gen["<< g <<"].Q_bar:\n"  << Gen[g].Q_bar   << endl;
         cout << "Gen: Gen["<< g <<"].DELTA:\n"  << Gen[g].DELTA   << endl;
         cout << "Gen: Gen["<< g <<"].D_tau:\n"  << Gen[g].D_tau   << endl;
#     endif
   }

   /*** convert deterministic markings -> Gen ***/
   for(int i = 1; i <= no_normal_classes; i++, g++)
   {
      struct randinfo *info = &randinfos[i];
      convert_det(info->Class, Gen[g]);

#     ifdef DEBUG_CONV
         cout << "Det: Gen["<< g <<"].name: "    << Gen[g].name    << endl;
         cout << "Det: Gen["<< g <<"].pmf_def: " << Gen[g].pmf_def << endl;
         cout << "Det: Gen["<< g <<"].I:\n"      << Gen[g].I       << endl;
         cout << "Det: Gen["<< g <<"].Q:\n"      << Gen[g].Q       << endl;
         cout << "Det: Gen["<< g <<"].Q_bar:\n"  << Gen[g].Q_bar   << endl;
         cout << "Det: Gen["<< g <<"].DELTA:\n"  << Gen[g].DELTA   << endl;
         cout << "Det: Gen["<< g <<"].D_tau:\n"  << Gen[g].D_tau   << endl;
#     endif
   }

   for(GenMarkings::iterator i = Gen.begin(); i != Gen.end();)
   {
      GenMarkingsSmc &gen = *i;
      if (gen.Q.empty() && gen.Q_bar.empty())
      {
         cout << "Erasing component with no egdes." << endl;
	 Gen.erase(i);
      }
      else
	 ++i;
   }

#  ifdef DEBUG_CONV
      cout << "Convertion of data has finished." << endl;
#  endif

   //efree_all();
}



/****************************************************************************
** FUNCTION : convert exp                                                  **
*****************************************************************************
** Purpose  : convertion of data corresponding to exponential markings     **
**            of the EMC                                                   **
**                                                                         **
** Input    : exp_markings    'C' struct. of exponential markings          **
**                                                                         **
** Output   : Exp             'C++' structs. of exponential markings       **
****************************************************************************/

void ConvSolveBase::convert_exp(tang_marking *exp_markings, ExpMarkings &Exp)
{    
   tang_marking  *t_walk;
   int            f_count;
  
   /*** scan all exponential markings ***/
   scan_list(t_walk, exp_markings)
   {
      double div_rate_sum = 1.0 / t_walk->sum_of_rates;

      /*** current row ***/
      int row = t_walk->number -1;

#     ifdef DEBUG_CONV
//         cout << "Exp: entry in row: " << row << endl;
#     endif

      /*** entry in Exp.C ***/
      Exp.C(row, row) = div_rate_sum;


      /*** entry in Exp.I ***/
      Exp.I.setDiagElem(row, true);

      
      /*** entry in Exp.P ***/
      scan_array(f_count, t_walk->no_exp_paths)
      {
         tang_path   *p_walk = &(t_walk->exp_paths[f_count]);
	 int          col = ((tang_marking *)(p_walk->marking))->number -1;

	 Exp.P(row, col) = p_walk->prob * div_rate_sum;
      }
   }
}



/****************************************************************************
** FUNCTION : convert gen                                                  **
*****************************************************************************
** Purpose  : convertion of data corresponding to general markings of SMCs **
**                                                                         **
**            see execute_erand.c::send_Q()                                **
**            see execute_erand.c::send_DELTA()                            **
**                                                                         **
** Input    : esmc            'C' struct. of general markings of SMC       **
**                                                                         **
** Output   : gen             'C++' structs. of general markings of SMC    **
****************************************************************************/

void ConvSolveBase::convert_gen(SMC *esmc, GenMarkingsSmc &gen)
{
   tang_marking    *mark;

   /*** entry in gen.name ***/
   gen.name = esmc->trans->name;


   /*** entry in gen.D_tau ***/
   gen.D_tau = DiagMatrix(OnesVector(dim));


   /*** scan all general SMCs ***/
   mark = esmc->states;
   for(int r = 0; r < esmc->no_states; r++, mark = mark->next)
   {
      /*** current row ***/
      int row = mark->number -1;

#     ifdef DEBUG_CONV
//          cout << "Gen: entry in row: " << row << endl;
#     endif


      /*** entry in gen.I ***/
      gen.I.setDiagElem(row, true);


      /*** entry in gen.Q (Q and Q_bar) ***/
      gen.Q(row, row) = -(double) mark->sum_of_rates; 

      for (int path = 0; path < mark->no_exp_paths; ++path)
      {
         int col = ((tang_marking *) mark->exp_paths[path].marking)->number -1;
	 gen.Q(row, col) = (double) mark->exp_paths[path].prob;
      }
      

      /*** entry in gen.DELTA ***/
      for (int path = 0; path < mark->no_det_paths; ++path)
      {
         int col = ((tang_marking *) mark->det_paths[path].marking)->number -1;
	 gen.DELTA(row, col) = (double) mark->det_paths[path].prob;
      }
   }
  
   /*** extend gen.DELTA by diagonal ones outside S^g ***/
//    gen.DELTA += gen.I.getComplement(); 


   /*** split gen.Q into gen.Q and gen.Q_bar ***/
   gen.Q_bar = gen.Q * gen.I.getComplement();
   gen.Q *= gen.I;


   /*** entry in gen.pmf_def ***/
   char formbuf[BUFLEN];
   char *filename = const_cast<char*>((string(Netname) + ".pmf").c_str());

   if(!file_read(filename, const_cast<char*>(gen.name.c_str()), formbuf))
   {
      cerr << "file_read (" << filename << ", " << gen.name << ", " 
	   << "...) failed." << endl;
      exit(1);
   }
   gen.pmf_def = formbuf;
}



/****************************************************************************
** FUNCTION : convert det                                                  **
*****************************************************************************
** Purpose  : convertion of data corresponding to determ. markings of SMCs **
**                                                                         **
**            see csmc_main.c::send_data_to_rand()                         **
**            see csmc_main.c::send_iso_data_to_rand()                     **
**                                                                         **
** Input    : class_p         'C' struct. of deterministic markings of SMC **
**                                                                         **
** Output   : gen             'C++' structs. of general markings of SMC    **
****************************************************************************/

void ConvSolveBase::convert_det(Iso_Class *class_p, GenMarkingsSmc &gen)
{
   Tau_CSMC	 *tau_walk;
   node		 *n_walk, *tn_walk;
   edge		 *e_walk;
   tang_marking	 *t_walk, *tt_walk;
   tang_path	 *path;
   int		  j, e_num, p_num;

#  ifdef DEBUG_CONV
      int act_preempt_no = class_p->graph->no_states;
      cout << "Det: act_preempt_no: " << act_preempt_no -1 << endl;
      cout << "Det: first_state_no: " << class_p->first_state_no -1 << endl;
#  endif
   
   /*** entry in gen.pmf_def ***/ 
   gen.pmf_def = "I[1.0];";


   /*** scan all deterministic SMCs ***/
   for(n_walk = class_p->graph->first_node; n_walk != 0;
       n_walk = n_walk->next)
   {	
      t_walk = n_walk->orig_mark;

      /*** current row ***/
      int row = t_walk->number -1;

#     ifdef DEBUG_CONV
//          cout << "Det (SMC): entry in row: " << row << endl;
#     endif


      /*** entry in gen.I ***/
      gen.I.setDiagElem(row, true);

      
      /*** gen.Q  ***/
      if (sort_edges == NULL)
      {
         max_edges = 100;
	 sort_edges = (edge *)malloc(max_edges * sizeof(edge));
      }

      if (max_edges <= 2 * t_walk->no_exp_paths)
      {
	 max_edges = 2 * (t_walk->no_exp_paths + 1);
	 sort_edges = (edge *)realloc(sort_edges, max_edges * sizeof(edge));
      }
      
      /*** gen.Q: diagonal entry ***/
      sort_edges[0].to_node = (void *)((long)t_walk->number-1);
      sort_edges[0].bits = 0;
      sort_edges[0].rate = -t_walk->sum_of_rates;
      
      /*** gen.Q: assemble edges ***/
      e_num = 1;
      p_num = n_walk->no_succ;

      scan_list(e_walk, n_walk->successors)
      {
         tn_walk = (node *)e_walk->to_node;
	 t_walk = tn_walk->orig_mark;

	 if (e_walk->bits & PREEMPTING)
	    j = p_num--;
	 else 
	 {
	    j = e_num++;

	    /* global number of state */
	    sort_edges[j].to_node = (void *)((long)t_walk->number-1);
	 }
	 /* rate */
	 sort_edges[j].rate = e_walk->rate;
      }
      qsort(sort_edges, e_num, sizeof(edge), edgecomp);

      /*** entries in gen.Q: scan number of normal edges ***/
      for(int i = 0; i < e_num; ++i)
      {
  	 /* current column (never store (int)s in (void*) !) */
	 int col = reinterpret_cast<int>(sort_edges[i].to_node);

         /* rate */
	 gen.Q(row, col) = sort_edges[i].rate;
      }

#     ifdef DEBUG_CONV
         /* number of preempting edges */
//          p_num = n_walk->no_succ-p_num;
//          cout << "number of preempting edges : " << p_num << endl;

//          for (int i = n_walk->no_succ; i < p_num + n_walk->no_succ; i++)
//          {
            /* rate */
// 	   cout << "preempting edges : " << sort_edges[i].rate << endl;;
//          }
#     endif
   }
  

   /*** scan all deterministic CSMCs ***/
   scan_list(tau_walk, class_p->Tau_CSMCs)
   {
#     ifdef DEBUG_CONV
      /* number of csmcs with the same tau */
      int i = tau_walk->no_isom_csmcs + 1;
      cout << "number of csmcs with the same tau : " << i << endl;

      /* tau marking dependent ? */
      char md_tau = tau_walk->bits & MARK_DEP_TAU;
  
   
      /* get absolute tau, if not marking dependent */
      if (!md_tau)
      {
         n_walk = tau_walk->graph->first_node;
         t_walk = n_walk->orig_mark;

         /* insert here for multiple tau transient solution */
         /* all tau values relative to tau_i (tau_x/tau_i) */

            double tau_i = det_transition_delay(t_walk);
            cout << "relative tau_i (not md_tau) : " << tau_i << endl;
      }
#     endif

      /*** scan CSMC states (rows) ***/
      for(n_walk = tau_walk->graph->first_node; n_walk; n_walk = n_walk->next)
      {
         t_walk = n_walk->orig_mark;

         /* current row */
         int row = t_walk->number -1;

#        ifdef DEBUG_CONV
//            cout << "Det (CSMC): entry in row : " << row << endl;
#        endif

    
         /*** entry in gen.DIAG_tau ***/
         gen.D_tau(row, row) = det_transition_delay(t_walk);

#        ifdef DEBUG_CONV
//       if (md_tau)
//          cout << "tau_i (md_tau) : " << gen.D_tau(row, row) << endl;
//       else
//          cout << "tau_i (not md_tau) : " << gen.D_tau(row, row) << endl;
#        endif


         /*** gen.DELTA ***/
         if (max_edges <= t_walk->no_det_paths)
         {
            max_edges = t_walk->no_det_paths + 1;
	    sort_edges = (edge *)realloc(sort_edges, max_edges * sizeof(edge));
         }
    
         /*** gen.DELTA: assemble entries ***/
	 int i; // !!!
         scan_array(i, t_walk->no_det_paths)
         {
	    path = &(t_walk->det_paths[i]);
	    tt_walk = (tang_marking *)path->marking;

	    /* state number */
	    sort_edges[i].to_node = (void *)((long)tt_walk->number);

	    /* probability */
	    sort_edges[i].rate = path->prob;
         }
         qsort(sort_edges, t_walk->no_det_paths, sizeof(edge), edgecomp);
    
         /*** entries in gen.DELTA: scan state numbers */
         scan_array(i, t_walk->no_det_paths)
         {
            /* current column (never store (int)s in (void*)s !) */
	    int col = reinterpret_cast<int>(sort_edges[i].to_node) -1;

	    /* probability */
	    gen.DELTA(row, col) = sort_edges[i].rate;
         }

    
         /*** entry in gen.Q_bar: look for preempted states ***/
         scan_list(e_walk, n_walk->successors)
         {
            if (!(e_walk->bits & PREEMPTING))
	       continue;
	    t_walk = (tang_marking *)e_walk->to_node;

            /* rate */
	    gen.Q_bar(row, t_walk->number -1) = e_walk->rate;
         }
      }
   }


   /*** Extend gen.DELTA by diagonal ones outside S^g ***/
//    gen.DELTA += gen.I.getComplement(); 
}



/****************************************************************************
** FUNCTION : solve main                                                   **
*****************************************************************************
** Purpose  : main function to solve the stationary solution               **
**            calling convertion of data with initialisation of Solve Base **
**            calling of sequential or parallel solve process              **
**                                                                         **
** Input    : net_name       name of the corresponding DSPN                **
**            exec_mode      sequential or parallel execution              **
**            arb_bits       Bits for arbitrary precision (calc. factors)  **
**            trunc_error    truncation error (calc. factors)              **
**            init_vec       kind of initial iteration vector              **
**            seed_value     seed value for randomisation of init. vector  **
**            save_vec       option if stationary vector shall be saved    **
**            max_iter       maximal number of outer iterations            **
**            epsilon        accuracy of vector equality                   **
**                                                                         **
** Output   : nothing                                                      **
****************************************************************************/

void solve_main(const char* net_name, const char* exec_mode,
                const long arb_bits, const double trunc_error,
                const char* init_vec, const long seed_value,
                const char* save_vec,
		const long max_iter, const double epsilon)
{
  /******************************************
  *** convert data and init Solve Base SB ***
  ******************************************/

  Time   day_startTime = Time::getTimeOfDay();
  Time   cpu_startTime = Time::getCPUTime();
  long   mem_atStart = getUsedMem();
  cout << "CONVERTING DATA TO FORMAT OF ITERATIVE SOLUTION METHOD" << endl;

   /*** call constructor of ConvSolveBaseto to init SB ***/
   ConvSolveBase SB(net_name, arb_bits, trunc_error, init_vec,
                    seed_value, save_vec, max_iter, epsilon);


   Time   day_diffTime = Time::getTimeOfDay() - day_startTime;
   Time   cpu_diffTime = Time::getCPUTime() - cpu_startTime;
   long   mem_diff = getUsedMem() - mem_atStart;

   cout << endl;
   cout << "Convert data structures : Elapsed CPU time  : "
	<< cpu_diffTime << endl;
   cout << "Convert data structures : Elapsed real time : "
	<< day_diffTime << endl;
   cout << "Convert data structures : Required memory : "
	<< mem_diff << " KB" << endl;
   cout << endl;

   /*************************************************
   *** start sequential or parallel solve process ***
   *************************************************/

   cout << endl;
   cout << "STARTING ITERATIVE SOLUTION METHOD WITH FILL-IN AVOIDANCE ..."
        << endl;
   cout << endl;

   if( string(exec_mode) == "parallel")
   {
#     ifdef PVM_SUPPORT
         solve_parallel(SB);   // start parallel solve process
#     else
         cout << "ERROR: Cannot run parallel." << endl;
         cout << "Set PVM_ROOT to compile with PVM_SUPPORT" << endl;
         exit(1);
#     endif
   }
   else
     solve_sequential(SB);     // start sequential solve process


   /**********************
   *** show statistics ***
   **********************/

   Time day_endTime = Time::getTimeOfDay();
   Time cpu_endTime = Time::getCPUTime();

   /*** read start times ***/
   Time cpu_totalTime = cpu_endTime - loadTime(SB.Netname, "cpu_start");
   Time day_totalTime = day_endTime - loadTime(SB.Netname, "day_start");
   Time cpu_TimeSol   = cpu_endTime - loadTime(SB.Netname, "cpu_startSol");
   Time day_TimeSol   = day_endTime - loadTime(SB.Netname, "day_startSol");

   /*** read required memory at start ***/
   long mem_total = getUsedMem() - loadMem(SB.Netname, "atStart");
   long mem_Sol   = getUsedMem() - loadMem(SB.Netname, "atStartSol");

   /* show statistics about solution method */
   cout << endl;
   cout << "Fill-in avoidance solution method : Elapsed CPU time  : ";
   cout << cpu_TimeSol << endl;
   cout << "Fill-in avoidance solution method : Elapsed real time : ";
   cout << day_TimeSol << endl;
   cout << "Fill-in avoidance solution method : Required memory : ";
   cout << mem_Sol << " KB" << endl;
   cout << endl;

   /* show statistics about overall solution method */
   cout << endl;
   cout << "Overall solution method : Elapsed CPU time  : ";
   cout << cpu_totalTime << endl;
   cout << "Overall solution method : Elapsed real time : ";
   cout << day_totalTime << endl;
   cout << "Overall solution method : Required memory : ";
   cout << mem_total << " KB" << endl;
   cout << endl;
}


