/* emcfilt.c - Copyright (c)1993, Author: U.Hertlein
 *
 * reads the .EMCxx files created by the rand processes and shows their
 * contents in readable form
 *
 * $Id: emcfilt.c,v 1.1 1994/08/24 14:34:09 rspn Exp $
 */

#include <stdio.h>
#include <fcntl.h>


#define HEADER_SIZE (4 + 4)
#define ENTRY_SIZE  (8 + 4)

struct row_header
{
   unsigned row_index;
   unsigned row_entries;
} head;

struct row_entry
{
   double probability;
   unsigned column_index;
} entry;


static int usage()
{
   fprintf(stderr,"Usage: emcfilt filename ...\n");
   fprintf(stderr,"\tfilename is just the basename of the file, like:\n");
   fprintf(stderr,"\tfcs2 for files fcs2.EMC00, fcs2.EMC01, fcs2.EMC02 ...\n");

   exit( 1 );
}

int parse_emc(fp)
FILE *fp;
{
   int i,old = 0;

   while( !feof(fp) )
   {
      fread(&head,HEADER_SIZE,1,fp);

      if( head.row_index == old )
         break;
      else
         old = head.row_index;

      fprintf(stdout,"row.index = %d, entries = %d\n", head.row_index,head.row_entries);

      for( i = 0; i < head.row_entries; i++ )
      {
         fread(&entry,ENTRY_SIZE,1,fp);
         fprintf(stdout," col.index = %d, prob = %lf\n", entry.column_index,entry.probability);
      }
      fprintf(stdout,"\n");
   }

   return( 0 );
}

int main(argc, argv)
int argc;
char **argv;
{
   int i,running;

   int filenum;
   char filename[256];
   FILE *fp;

   if( argc < 2 )
      usage();

   for( i = 1; --argc; i++ )
   {
      running = 1;
      filenum = 0;

      while( running )
      {
         sprintf(filename,"%s.EMC%02u", argv[i],filenum);

         if( (fp = fopen(filename,"r")) == (FILE *)NULL )
	 {
            running = (filenum == 0);
	 }
         else
	 {
            fprintf(stdout,"Parsing file '%s' ...\n", filename);

            parse_emc(fp);

            fclose(fp);
	 }

         filenum++;
      }
   }

   exit( 0 );
}
