/*****************************************************************************

	File Name	:	showRESULTS.c


	Parameter	:	Options [sot], Net Name

	Return Value	:	0 -> Program successfully terminated
				1 -> Error occurred

	Purpose		:	showRESULTS displays the probabilities
				associated with the tangible markings of a DSPN
 
	Input Files	:	<net>.TN, <net>.RG, <net>.PROBMARK, <net>.PROBTOK

	Project Director:	Christoph Lindemann
				Technical University of Berlin

	Programmer	:	Armin Zimmermann

	History		:	First version September 1992
				April 1993: new file format, option t

*                                                                           *
*  Modifiziert	  :	Joachim Schneider				    *
*									    *
*  Datum	  :	July 1994 					    *
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>

#define	USAGE	"Usage: showRESULTS [-sot] <netname>\n"
#define FILE_NAME	100

#include "portab.h"

typedef struct rg_type
{	unsigned	number;
}  rg_type;

typedef	struct m_type
{	int	number;
	int	*token;
	double	prob;
	struct m_type *next;
}  m_type;

int	no_places;			/* number of places in the net */
m_type	*markings = NULL;		/* tangible markings */
int	no_markings;			/* number of tangible markings */
char	**place_names;			/* names of places */
int	*place_bounds;			/* upper bounds of tokens in places */
FILE 	*net_file;			/* net definition file .TN */
FILE 	*rg_file;			/* reachability set file .RG */
FILE 	*prob_file;			/* marking probability file .PROBMARK */
FILE 	*tok_file;			/* token distribution probability file .PROBTOK */
int	sort_bit;			/* sort markings by probability */
int	zero_bit;			/* show only markings with prob > 0 */
int	tok_bit;			/* show only token distribution */

/*	next_line skips to the beginning of the next data line
	and returns the number of the token; 0 = EOF */
#define	NEXT_LINE_BUF	500
#define NO_TOKENS	12
char	tokens[NO_TOKENS][30] = {"eof", "PLACES:", "TRANSITIONS:", 
			"DELAY_PARAMETERS:", "MARKING_PARAMETERS:",
			"MARKPAR", "PLACE", "DELAYPAR", "TRANSITION", 
			"INPARCS", "OUTPARCS", "INHARCS" };
int
next_line(in_file, buffer)
	FILE *in_file;
	char *buffer;
{	
	int	beg, tok, copy;
	char	in_token[50];

	while(1)
	{	if (fgets(buffer, NEXT_LINE_BUF, in_file) == NULL)
			return 0;		/* EOF reached */
		for (beg = 0; beg < NEXT_LINE_BUF; beg++)
			if (!isspace(buffer[beg]))
				break;
		if (buffer[beg] && (beg < NEXT_LINE_BUF) && (buffer[beg] != '#'))
			break;		/* line is not empty nor a comment */
	}
	for (copy = 0; buffer[beg] && !isspace(buffer[beg]); )
		in_token[copy++] = buffer[beg++];
	in_token[copy] = 0;
	for (tok = 1; tok < NO_TOKENS; tok++)
		if (!strcmp(in_token, tokens[tok]))
		{	for (copy = 0; buffer[beg]; )
				buffer[copy++] = buffer[beg++];
			buffer[copy] = 0;
			return tok;
		}
	return -1;			/* not found */
}

/*	read_net reads the net-description from net.TN */
void
read_net(net)
	FILE *net;
{	
	static	char buf[NEXT_LINE_BUF];
	char	help[50], h1[50], h2[50];
	int	code, i_place;

	no_places = 0;
	i_place = 0;
	while(code = next_line(net, buf))
	{	switch(code)
		{	case 1:	sscanf(buf, "%d", &no_places);
				place_names = (char **)malloc(no_places * sizeof(char *));
				break;
			case 6:	if (i_place >= no_places)
				{	printf("too many places, line %s\n", buf);
					exit(1);
				}
				sscanf(buf, "%s ", help);
				place_names[i_place] = strdup(help);
				i_place++;
				break;
			default:break;
		}
	}
}

/*	sort markings by probability */
void
sortiere(anker)
	m_type **anker;
{
	m_type *halde, *walk, *last;

	halde = *anker;
	*anker = NULL;
	while (halde != NULL)
	{	walk = *anker;		/* Insert-Sort */
		last = NULL;
		while (walk != NULL)
		{	if (halde->prob > walk->prob)
				break;
			last = walk;
			walk = walk->next;
		}
		if (walk == *anker)	/* first element */
		{	*anker = halde;
			halde = halde->next;
			(*anker)->next = walk;
			continue;
		}
		last->next = halde;	/* inner or last element */
		halde = halde->next;
		last->next->next = walk;
	}
}

/*	show token distribution probablilities */
void
tok_output()
{
	double	in_d;
	int	place, toks, min, max;

	/* .PROBTOK lesen */
	for (place = 0; place < no_places; place++)
	{	fread(&in_d, sizeof(double), 1, tok_file);
		min = (int)in_d;
		fread(&in_d, sizeof(double), 1, tok_file);
		max = (int)in_d;
		printf("\n=> Place %s (%d..%d token)\n", place_names[place], min, max);
		for (toks = min; toks <= max; toks++)
		{	fread(&in_d, sizeof(double), 1, tok_file);
			if (in_d != 0.0 || !zero_bit)
				printf("   %5d  %12.12lg\n", toks, in_d);
		}
	}
	exit(0);
}

/*	main procedure */
main(argc, argv)
	int argc;
	char ** argv;
{
	char	net_name[FILE_NAME], rg_name[FILE_NAME];
	char	prob_name[FILE_NAME], tok_name[FILE_NAME];
	char	options[10], topline[1000], help[100];
	int	arg_num = 1, i, ende, k;
	rg_type	rg;
	m_type	*new_m, *walk;
	double	dummy;

	if (argc == 1)
	{	printf("ERROR: no netname found\n");
		printf(USAGE);
		printf("-s : sort markings according to their");
		printf(" probabilities in descending order\n");
		printf("-o : show only markings with nonzero probability\n");
		printf("-t : show the token distribution probabilities\n");
		exit(1);
	}
	sort_bit = zero_bit = tok_bit = 0;
	if (argv[1][0] == '-')
	{	arg_num++;
		strcpy(options, argv[1]+1);
		for(i = 0; i < strlen(options); i++)
			switch(options[i])
			{	case 's': sort_bit = 1; break;
				case 'o': zero_bit = 1; break;
				case 't': tok_bit  = 1; break;
				default : printf(USAGE);
					  printf("-s : sort markings according to their");
					  printf(" probabilities in descending order\n");
			printf("-o : show only markings with nonzero probability\n");
			printf("-t : show the token distribution probabilities\n");
					  exit(1);
			}
	}
	if (argc <= arg_num)
	{	printf("ERROR: no netname found\n");
		printf(USAGE);
		exit(1);
	}
	strcpy(net_name, argv[arg_num]);
	strcpy(rg_name, argv[arg_num]);
	strcpy(prob_name, argv[arg_num]);
	strcpy(tok_name, argv[arg_num]);
	strcat(net_name, ".TN");
	strcat(rg_name, ".RG");
	strcat(prob_name, ".PROBMARK");
	strcat(tok_name, ".PROBTOK");
	if ((net_file = fopen(net_name, "r")) == NULL)
	{	printf("ERROR: could not open %s for reading\n", net_name);
		exit(1);
	}
	if (tok_bit)
	{	if ((tok_file = fopen(tok_name, "r")) == NULL)
		{	printf("ERROR: could not open %s for reading\n", tok_name);
			exit(1);
		}
	} else
	{	if ((rg_file = fopen(rg_name, "r")) == NULL)
		{	printf("ERROR: could not open %s for reading\n", rg_name);
			exit(1);
		}
		if ((prob_file = fopen(prob_name, "r")) == NULL)
		{	printf("ERROR: could not open %s for reading\n", prob_name);
			exit(1);
		}
	}
	printf("\nResults for %s:\n", argv[arg_num]);
	read_net(net_file);		/* read the net description file */
	if (no_places < 1)
	{	printf("Error: no places.\n");
		exit(1);
	}
	printf("the net contains %d places.\n", no_places);

	if (tok_bit)
		tok_output();

	printf("\nMapping numbers to places of the DSPN:\n");
	for (i=0; i<no_places; i++)
		printf("%3d => %s\n", i, place_names[i]);
	printf("\n");

	/* .RG lesen */
	place_bounds = (int *)malloc(no_places * sizeof(int));
	for (i=0; i<no_places; i++)
		place_bounds[i] = 0;
	ende = no_markings = 0;
	walk = markings;
        while (!ende)
	{	new_m = (m_type *)malloc(sizeof(m_type));
		new_m->number = ++no_markings;
		new_m->token = (int *)malloc(no_places * sizeof(int));
		for(i=0; i<no_places; i++)
		{	if(fread(&rg, sizeof(rg), 1, rg_file)!=1)
                        {	printf("the reachability set contains");
				printf(" %d markings.\n", --no_markings);
				ende = 1;
                                break;
                        }
			new_m->token[i] = rg.number;
			new_m->next = NULL;
			if (place_bounds[i] < rg.number)
				place_bounds[i] = rg.number;
		}
		if (!ende)
			if (walk == NULL)
				markings = new_m;
			else	walk->next = new_m;
		walk = new_m;
        }
	fclose(rg_file);

	/* .PROBMARK lesen */
	walk = markings;
	fread(&dummy, sizeof(double), 1, prob_file);
	fread(&dummy, sizeof(double), 1, prob_file);
	for (i=0; i<no_markings; i++)
	{	if (!fread(&(walk->prob), sizeof(double), 1, prob_file))
		{	printf("error in %s.\n", prob_name);
			exit(1);
		}
		walk = walk->next;
	}

	if (sort_bit)
		sortiere(&markings);

	/* Ausgabe vorbereiten */
	for (i=0; i<no_places; i++)
	{	if (place_bounds[i] < i)
			place_bounds[i] = i;
		if (!place_bounds[i])
			place_bounds[i] = 1;
		else place_bounds[i] = 1 + (int)floor(log10((double)place_bounds[i]));
	}
	strcpy(topline, "\n         PLACES\nMarking ");
	for (i=0; i<no_places; i++)
	{	sprintf(help, "%d", i);
		for (k=0; k<=place_bounds[i]-strlen(help); k++)
			strcat(topline, " ");
		strcat(topline, help);
	}
	strcat(topline, "   Probability\n");
	ende = strlen(topline) - 9;
	for (i=0; i<ende; i++)
		strcat(topline, "-");
	strcat(topline, "\n");
	/* Ausgabe */
	printf("%s", topline);
	for (walk = markings; walk != NULL; walk = walk->next)
	{	if (zero_bit && walk->prob == 0.0)
			continue;
		printf("%7d ", walk->number);
		for (i=0; i<no_places; i++)
		{	sprintf(help, "%d", walk->token[i]);
			for (k=0; k<=place_bounds[i]-strlen(help); k++)
				printf(" ");
			printf("%s", help);
		}
		printf("   %.15g\n", walk->prob);
	}
}
